/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.dgws.http;

import dk.sundhedsdatastyrelsen.cbs.CPRFilter;
import dk.sundhedsdatastyrelsen.cbs.exceptions.HttpException;
import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.entity.StringEntity;
import org.apache.http.util.EntityUtils;

import java.io.IOException;
import java.io.UncheckedIOException;
import java.nio.charset.StandardCharsets;

public final class HttpUtil {

    private HttpUtil() {
    }

    public static HttpPost createSoapRequest(String endpoint, String soapAction, String requestXml) {
        HttpPost httpPost = new HttpPost(endpoint);
        httpPost.addHeader("SOAPAction", soapAction);
        httpPost.addHeader("Content-type", "text/xml");
        httpPost.setEntity(new StringEntity(requestXml, StandardCharsets.UTF_8));
        return httpPost;
    }

    public static void checkSoapResponseStatus(String endpoint, String soapAction, HttpResponse httpResponse) {
        int statusCode = httpResponse.getStatusLine().getStatusCode();
        if (statusCode != HttpStatus.SC_OK) {
            String body = httpResponse.getEntity() != null ? CPRFilter.filter(extractResponse(httpResponse)) : "<NO_RESPONSE_BODY>";
            throw new HttpException(String.format("Got abnormal status code (%d) from %s, action=%s, body=%s", statusCode, endpoint, soapAction, body));
        }
    }

    public static String extractResponse(HttpResponse httpResponse) {
        try {
            HttpEntity httpEntity = httpResponse.getEntity();
            String responseXml = EntityUtils.toString(httpEntity, StandardCharsets.UTF_8);
            EntityUtils.consume(httpEntity);
            return responseXml;
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        }
    }
}
