/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.kafka;

import dk.sds.nsp.kafka.clients.config.NSPKafkaConfig;
import org.apache.commons.lang3.StringUtils;
import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.apache.kafka.clients.producer.ProducerConfig;
import org.apache.log4j.Logger;
import org.springframework.core.env.ConfigurableEnvironment;
import org.springframework.core.env.EnumerablePropertySource;
import org.springframework.core.env.MutablePropertySources;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.concurrent.atomic.AtomicInteger;

public class CbsKafkaUtils {
    private static final Logger log = Logger.getLogger(CbsKafkaUtils.class);
    private static final Map<String, AtomicInteger> CLIENT_ID_SUFFIX_MAP = new HashMap<>();
    private static final String WARN_PROP_NOT_SUPPORTED = "Kafka property %s not supported. Ignoring";

    private CbsKafkaUtils() {
    }

    public static Properties getKafkaProperties(String kafkaPropertyPrefix, ConfigurableEnvironment env) {
        Properties kafkaProperties = new Properties();
        MutablePropertySources propSources = env.getPropertySources();
        propSources.stream()
            .filter(EnumerablePropertySource.class::isInstance)
                .map(ps -> ((EnumerablePropertySource<?>) ps).getPropertyNames())
                .flatMap(Arrays::stream)
                .filter(propName -> propName.startsWith(kafkaPropertyPrefix) || NSPKafkaConfig.isNSPProperty(propName))
                .forEach(propName -> kafkaProperties.setProperty(StringUtils.removeStartIgnoreCase(propName, kafkaPropertyPrefix), env.getProperty(propName))); //lookup value using env.getProperty, which respects Spring property overriding, see https://stackoverflow.com/a/42521523

        if (kafkaProperties.contains(ProducerConfig.KEY_SERIALIZER_CLASS_CONFIG))
            log.warn(String.format(WARN_PROP_NOT_SUPPORTED, ProducerConfig.KEY_SERIALIZER_CLASS_CONFIG));
        if (kafkaProperties.contains(ProducerConfig.VALUE_SERIALIZER_CLASS_CONFIG))
            log.warn(String.format(WARN_PROP_NOT_SUPPORTED, ProducerConfig.VALUE_SERIALIZER_CLASS_CONFIG));

        if (kafkaProperties.contains(ConsumerConfig.KEY_DESERIALIZER_CLASS_CONFIG))
            log.warn(String.format(WARN_PROP_NOT_SUPPORTED, ConsumerConfig.KEY_DESERIALIZER_CLASS_CONFIG));
        if (kafkaProperties.contains(ConsumerConfig.VALUE_DESERIALIZER_CLASS_CONFIG))
            log.warn(String.format(WARN_PROP_NOT_SUPPORTED, ConsumerConfig.VALUE_DESERIALIZER_CLASS_CONFIG));

        return kafkaProperties;
    }

    public static synchronized Properties getProducerInstanceProperties(Properties kafkaProperties) {
        String clientId = String.valueOf(kafkaProperties.getOrDefault(ConsumerConfig.CLIENT_ID_CONFIG, "CbsKafkaProducer"));
        String uniqueClientId = clientId + "-" + CLIENT_ID_SUFFIX_MAP.computeIfAbsent(clientId, id -> new AtomicInteger(1)).getAndIncrement();
        Properties instanceProperties = new Properties();
        instanceProperties.putAll(kafkaProperties);
        instanceProperties.setProperty(ProducerConfig.CLIENT_ID_CONFIG, uniqueClientId);
        return instanceProperties;
    }


    public static synchronized Properties getConsumerInstanceProperties(Properties kafkaProperties) {
        String clientId = String.valueOf(kafkaProperties.getOrDefault(ConsumerConfig.CLIENT_ID_CONFIG, "CbsKafkaConsumer"));
        String uniqueClientId = clientId + "-" + CLIENT_ID_SUFFIX_MAP.computeIfAbsent(clientId, id -> new AtomicInteger(1)).getAndIncrement();
        Properties instanceProperties = new Properties();
        instanceProperties.putAll(kafkaProperties);
        instanceProperties.setProperty(ConsumerConfig.CLIENT_ID_CONFIG, uniqueClientId);
        return instanceProperties;
    }
}
