/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.kafka.config;

import dk.sds.nsp.kafka.clients.consumer.NSPKafkaConsumer;
import dk.sds.nsp.kafka.clients.producer.NSPKafkaProducer;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.kafka.DataFetcherJobImpl;
import dk.sundhedsdatastyrelsen.cbs.dto.kafka.DataFetcherJob;
import dk.sundhedsdatastyrelsen.cbs.kafka.CbsKafkaConsumer;
import dk.sundhedsdatastyrelsen.cbs.kafka.serialization.JacksonDeserializer;
import dk.sundhedsdatastyrelsen.cbs.kafka.serialization.JacksonSerializer;
import org.apache.commons.pool2.impl.GenericObjectPoolConfig;
import org.apache.kafka.common.serialization.Deserializer;
import org.apache.kafka.common.serialization.Serializer;
import org.apache.kafka.common.serialization.StringDeserializer;
import org.apache.kafka.common.serialization.StringSerializer;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Lazy;

import java.util.Properties;
import java.util.function.Function;

@Configuration
public class CbsKafkaConfig {

    @Bean
    @Lazy
    public Serializer<String> stringSerializer() {
        return new StringSerializer();
    }

    @Bean
    @Lazy
    public Serializer<DataFetcherJob> dataFetcherJobSerializer() {
        return new JacksonSerializer<>();
    }

    @Bean
    @Lazy
    public Deserializer<String> stringDeserializer() {
        return new StringDeserializer();
    }

    @Bean
    @Lazy
    public Deserializer<DataFetcherJob> dataFetcherJobDeserializer() {
        return new JacksonDeserializer<>(DataFetcherJobImpl.class);
    }

    @Bean
    @Lazy
    public Function<Properties, NSPKafkaConsumer<String, DataFetcherJob>> nspKafkaConsumerSupplier(Deserializer<String> stringDeserializer,
                                                                                                   Deserializer<DataFetcherJob> dataFetcherJobDeserializer) {
        return consumerInstanceProperties -> new NSPKafkaConsumer<>(consumerInstanceProperties, stringDeserializer, dataFetcherJobDeserializer);
    }

    @Bean
    @Lazy
    public GenericObjectPoolConfig<CbsKafkaConsumer> poolConfig(@Value("${cbs.kafka.consumer.max-pool-size:10}") int maxPoolSize,
                                                                @Value("${cbs.kafka.consumer.max-wait-millis:20000}") int maxWaitMillis) {
        GenericObjectPoolConfig<CbsKafkaConsumer> poolConfig = new GenericObjectPoolConfig<>();
        poolConfig.setLifo(false); //use FIFO for evenly consumption of partitions (each consumer is assigned 1 or more partitions, so consumers should be invoked in round-robin fashion to evenly consume from all partitions)
        poolConfig.setFairness(true); //waiting servlet threads are served FIFO
        poolConfig.setBlockWhenExhausted(true);
        poolConfig.setMaxWaitMillis(maxWaitMillis);
        poolConfig.setMaxTotal(maxPoolSize);
        return poolConfig;
    }

    @Bean
    @Lazy
    public Function<Properties, NSPKafkaProducer<String, DataFetcherJob>> nspKafkaProducerFunction(Serializer<String> stringSerializer,
                                                                                                   Serializer<DataFetcherJob> dataFetcherJobSerializer) {
        return producerInstanceProperties -> new NSPKafkaProducer<>(producerInstanceProperties, stringSerializer, dataFetcherJobSerializer);
    }

}
