/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.kafka.pooling;

import dk.sundhedsdatastyrelsen.cbs.dto.kafka.DataFetcherJob;
import dk.sundhedsdatastyrelsen.cbs.kafka.CbsKafkaConsumer;
import org.apache.commons.pool2.PooledObjectFactory;
import org.apache.commons.pool2.impl.GenericObjectPool;
import org.apache.log4j.Logger;

import javax.annotation.PreDestroy;
import java.io.Closeable;
import java.util.Collections;
import java.util.NoSuchElementException;
import java.util.Set;

public class CbsKafkaConsumerPool implements Closeable {
    private static final Logger log = Logger.getLogger(CbsKafkaConsumerPool.class);
    private final GenericObjectPool<CbsKafkaConsumer> consumerPool;

    public CbsKafkaConsumerPool(GenericObjectPool<CbsKafkaConsumer> consumerObjectPool) {
        consumerPool = consumerObjectPool;
    }

    public Set<DataFetcherJob> pollDataFetcherJobs() throws Exception {
        CbsKafkaConsumer consumer;
        try {
            consumer = consumerPool.borrowObject();
        } catch (NoSuchElementException noSuchElementException) {
            log.debug("Timeout waiting for idle CbsKafkaConsumer - Returning empty set of jobs");
            return Collections.emptySet();
        }
        try {
            return consumer.pollDataFetcherJobs();
        } finally {
            consumerPool.returnObject(consumer);
        }
    }

    @PreDestroy
    @Override
    public void close() {
        consumerPool.close(); //only destroys idle objects and interrupts threads waiting for an object

        PooledObjectFactory<CbsKafkaConsumer> factory = consumerPool.getFactory();
        if (factory instanceof PooledClosableObjectFactory) {
            ((PooledClosableObjectFactory<CbsKafkaConsumer>) factory).closeAllCreatedObjects();
        } else {
            log.warn("could not close non-idle objects");
        }
    }

}
