/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.kafka.status;

import com.google.common.base.Suppliers;
import dk.sundhedsdatastyrelsen.cbs.status.StatusCheck;
import dk.sundhedsdatastyrelsen.cbs.status.StatusChecker;
import org.apache.kafka.clients.admin.AdminClient;
import org.apache.kafka.clients.admin.DescribeClusterOptions;
import org.apache.log4j.Logger;

import java.util.HashMap;
import java.util.Map;
import java.util.function.Supplier;

public class KafkaStatusChecker implements StatusChecker {

    private static final Logger LOGGER = Logger.getLogger(KafkaStatusChecker.class);

    private final String title;
    private final Supplier<AdminClient> adminClientSupplier;

    public KafkaStatusChecker(Supplier<AdminClient> adminClientSupplier, String title) {
        this.title = title;
        this.adminClientSupplier = Suppliers.memoize(adminClientSupplier::get);
    }

    @Override
    public StatusCheck checkStatus() {

        final Map<String, String> info = new HashMap<>();

        boolean isOk = false;
        try {
            final AdminClient adminClient = this.adminClientSupplier.get();
            isOk = !adminClient.describeCluster(new DescribeClusterOptions().timeoutMs(500)).nodes().get().isEmpty();
        } catch (InterruptedException e) {
            info.put("Exception", e.getMessage());
            Thread.currentThread().interrupt();
        } catch (Exception e) {
            info.put("Exception", e.getMessage());
        }

        info.put("Is OK", String.valueOf(isOk));

        return new StatusCheck(title, info, isOk ? StatusCheck.Status.OK : StatusCheck.Status.ERROR);
    }

    // Called by spring on shutdown. (inferred by name)
    public void close() {
        try {
            adminClientSupplier.get().close();
        } catch (Exception e) {
            LOGGER.error("Failed to close Admin Client", e);
        }
    }
}
