/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.kafka;

import dk.sds.nsp.kafka.clients.consumer.NSPKafkaConsumer;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.kafka.DataFetcherJobBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.kafka.DataFetcherJob;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import org.apache.kafka.clients.consumer.ConsumerRecord;
import org.apache.kafka.clients.consumer.ConsumerRecords;
import org.apache.kafka.common.TopicPartition;
import org.apache.kafka.common.errors.WakeupException;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;

import java.util.*;

import static com.google.common.truth.Truth.assertThat;
import static dk.sundhedsdatastyrelsen.cbs.test.KafkaTestUtils.createRandomDataFetcherJobWithMultipleFetchCprs;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.mockito.Mockito.*;

@CbsTest
class CbsKafkaConsumerTest {
    @Mock
    private NSPKafkaConsumer<String, DataFetcherJob> mockNspConsumer;
    private CbsKafkaConsumer cbsKafkaConsumer;
    private static final String topic = "topic";

    @BeforeEach
    void setUp() {
        cbsKafkaConsumer = new CbsKafkaConsumer(topic, "my-clientId", mockNspConsumer, 100);
    }

    @Test
    void test_get_single_job() {
        Map<TopicPartition, List<ConsumerRecord<String, DataFetcherJob>>> records = new HashMap<>();
        ArrayList<ConsumerRecord<String, DataFetcherJob>> record = new ArrayList<>();
        DataFetcherJob job = createRandomDataFetcherJobWithMultipleFetchCprs();
        record.add(new ConsumerRecord<>(topic, 0, 0, job.getCpr(), job));
        records.put(new TopicPartition(topic, 0), record);
        when(mockNspConsumer.poll(anyLong())).thenReturn(new ConsumerRecords<>(records));

        Set<DataFetcherJob> jobs = cbsKafkaConsumer.pollDataFetcherJobs();
        assertThat(jobs).containsExactly(job);
    }

    @Test
    void test_get_multiple_jobs() {
        Map<TopicPartition, List<ConsumerRecord<String, DataFetcherJob>>> records = new HashMap<>();
        ArrayList<ConsumerRecord<String, DataFetcherJob>> record = new ArrayList<>();
        DataFetcherJob job1 = createRandomDataFetcherJobWithMultipleFetchCprs();
        DataFetcherJob job2 = createRandomDataFetcherJobWithMultipleFetchCprs();
        record.add(new ConsumerRecord<>(topic, 0, 0, job1.getCpr(), job1));
        record.add(new ConsumerRecord<>(topic, 0, 0, job2.getCpr(), job2));
        records.put(new TopicPartition(topic, 0), record);
        when(mockNspConsumer.poll(anyLong())).thenReturn(new ConsumerRecords<>(records));

        Set<DataFetcherJob> jobs = cbsKafkaConsumer.pollDataFetcherJobs();
        assertThat(jobs).containsExactly(job1, job2);

    }

    @Test
    void identical_records_are_collapsed() {
        Map<TopicPartition, List<ConsumerRecord<String, DataFetcherJob>>> recordMap = new HashMap<>();
        ArrayList<ConsumerRecord<String, DataFetcherJob>> records = new ArrayList<>();
        DataFetcherJob job1 = createRandomDataFetcherJobWithMultipleFetchCprs();
        String cpr = job1.getCpr();
        DataFetcherJob job2 = DataFetcherJobBuilder.aDataFetcherJob()
            .withCpr(cpr)
            .withMsgId(job1.getMsgId())
            .withJobId(job1.getJobId())
            .withFetchCprs(job1.getFetchCprs())
            .build();

        records.add(new ConsumerRecord<>(topic, 0, 0, cpr, job1));
        records.add(new ConsumerRecord<>(topic, 0, 0, cpr, job2));
        recordMap.put(new TopicPartition(topic, 0), records);
        when(mockNspConsumer.poll(anyLong())).thenReturn(new ConsumerRecords<>(recordMap));

        Set<DataFetcherJob> jobs = cbsKafkaConsumer.pollDataFetcherJobs();
        assertThat(jobs).hasSize(1);
        assertThat(jobs).containsExactly(job1);
        assertThat(jobs).containsExactly(job2);
    }

    @Test
    void WakeupException_during_poll_closes_KafkaConsumer_if_CbsKafkaConsumer_has_been_closed() {
        when(mockNspConsumer.poll(anyLong())).thenThrow(new WakeupException());
        cbsKafkaConsumer.close();
        Set<DataFetcherJob> jobs = cbsKafkaConsumer.pollDataFetcherJobs();
        assertThat(jobs).isEmpty();
        verify(mockNspConsumer, atLeastOnce()).close(anyLong(), any());
    }

    @Test
    void WakeupException_during_poll_does_not_close_KafkaConsumer_if_CbsKafkaConsumer_has_not_been_closed() {
        when(mockNspConsumer.poll(anyLong())).thenThrow(new WakeupException());
        assertThrows(WakeupException.class, () -> cbsKafkaConsumer.pollDataFetcherJobs());
        verify(mockNspConsumer, never()).close();
    }
}
