/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.kafka;

import com.google.common.collect.ImmutableMap;
import com.google.common.truth.Truth;
import dk.sds.nsp.kafka.clients.producer.NSPKafkaProducer;
import dk.sundhedsdatastyrelsen.cbs.dto.kafka.DataFetcherJob;
import dk.sundhedsdatastyrelsen.cbs.dto.kafka.FetchType;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import dk.sundhedsdatastyrelsen.cbs.test.TestUtils;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Mock;

import java.util.Properties;
import java.util.UUID;

import static org.junit.jupiter.api.Assertions.assertDoesNotThrow;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.mockito.Mockito.verify;

@CbsTest
class CbsKafkaProducerTest {
    @Mock
    private NSPKafkaProducer<String, DataFetcherJob> mockNspProducer;
    @Captor
    private ArgumentCaptor<ProducerRecord<String, DataFetcherJob>> producerRecordArgumentCaptor;

    @Test
    void test_send_job() {
        String topic = "topic";
        CbsKafkaProducer cbsKafkaProducer = new CbsKafkaProducer(topic, new Properties(), () -> mockNspProducer);

        String cpr = TestUtils.randomCprNumber();
        FetchType fetchType = FetchType.FETCH_VACCINATION_STATUS;
        String jobId = UUID.randomUUID().toString();
        String msgId = UUID.randomUUID().toString();

        cbsKafkaProducer.sendDataFetcherJob(cpr, jobId, msgId, ImmutableMap.of(cpr, fetchType));

        verify(mockNspProducer).send(producerRecordArgumentCaptor.capture());

        ProducerRecord<String, DataFetcherJob> producerRecord = producerRecordArgumentCaptor.getValue();
        assertEquals(topic, producerRecord.topic());
        assertEquals(cpr, producerRecord.key());
        assertEquals(cpr, producerRecord.value().getCpr());
        Truth.assertThat(producerRecord.value().getFetchCprs()).containsExactly(cpr, fetchType);
        assertEquals(jobId, producerRecord.value().getJobId());
        assertEquals(msgId, producerRecord.value().getMsgId());
    }

    @Test
    void test_send_job_multiple_cpr() {
        String topic = "topic";
        CbsKafkaProducer cbsKafkaProducer = new CbsKafkaProducer(topic, new Properties(), () -> mockNspProducer);

        String mainCpr = TestUtils.randomCprNumber();
        String child1cpr = TestUtils.randomCprNumber();
        String child2cpr = TestUtils.randomCprNumber();
        FetchType fetchType1 = FetchType.FETCH_VACCINATION_STATUS;
        FetchType fetchType2 = FetchType.FETCH_ALL;
        FetchType fetchType3 = FetchType.FETCH_TEST_STATUS;
        String jobId = UUID.randomUUID().toString();
        String msgId = UUID.randomUUID().toString();

        cbsKafkaProducer.sendDataFetcherJob(mainCpr, jobId, msgId, ImmutableMap.of(mainCpr, fetchType1, child1cpr, fetchType2, child2cpr, fetchType3));

        verify(mockNspProducer).send(producerRecordArgumentCaptor.capture());

        ProducerRecord<String, DataFetcherJob> producerRecord = producerRecordArgumentCaptor.getValue();
        assertEquals(topic, producerRecord.topic());
        assertEquals(mainCpr, producerRecord.key());
        assertEquals(mainCpr, producerRecord.value().getCpr());
        Truth.assertThat(producerRecord.value().getFetchCprs()).containsExactly(mainCpr, fetchType1, child1cpr, fetchType2, child2cpr, fetchType3);
        assertEquals(jobId, producerRecord.value().getJobId());
        assertEquals(msgId, producerRecord.value().getMsgId());
    }

    @Test
    void test_shutdown() {
        CbsKafkaProducer cbsKafkaProducer = new CbsKafkaProducer("topic", new Properties(), () -> mockNspProducer);
        assertDoesNotThrow(cbsKafkaProducer::close);
    }
}
