/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.kafka.pooling;

import com.google.common.collect.Iterables;
import com.google.common.truth.Correspondence;
import dk.sds.nsp.kafka.clients.consumer.NSPKafkaConsumer;
import dk.sundhedsdatastyrelsen.cbs.dto.kafka.DataFetcherJob;
import dk.sundhedsdatastyrelsen.cbs.kafka.CbsKafkaConsumer;
import dk.sundhedsdatastyrelsen.cbs.kafka.config.CbsKafkaConfig;
import dk.sundhedsdatastyrelsen.cbs.test.AssertionTestUtil;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import org.apache.commons.lang3.concurrent.BasicThreadFactory;
import org.apache.commons.pool2.impl.GenericObjectPool;
import org.apache.commons.pool2.impl.GenericObjectPoolConfig;
import org.apache.kafka.clients.consumer.ConsumerRecords;
import org.apache.log4j.Appender;
import org.apache.log4j.Level;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.apache.log4j.spi.LoggingEvent;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.stubbing.VoidAnswer1;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.context.junit.jupiter.SpringJUnitConfig;

import java.util.*;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.ReentrantLock;
import java.util.function.Supplier;

import static com.google.common.truth.Truth.assertThat;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.mockito.AdditionalAnswers.answerVoid;
import static org.mockito.Mockito.*;

@CbsTest
@SpringJUnitConfig(CbsKafkaConfig.class)
class CbsKafkaConsumerPoolTest {
    @Autowired
    GenericObjectPoolConfig<CbsKafkaConsumer> poolConfig;
    @Mock
    private NSPKafkaConsumer<String, DataFetcherJob> mockConsumer1;
    @Mock
    private NSPKafkaConsumer<String, DataFetcherJob> mockConsumer2;

    private final int maxPoolSize = 2;
    private CbsKafkaConsumerPool consumerPool;
    private final ReentrantLock lock = new ReentrantLock();
    private ExecutorService executorService;
    private GenericObjectPool<CbsKafkaConsumer> consumerObjectPool;

    @Mock
    private Appender mockAppender;
    private final List<LoggingEvent> capturedLogEvents = new ArrayList<>();
    private Logger cbsKafkaConsumerPoolLogger;
    private final Correspondence<LoggingEvent, String> compareLogMessages = Correspondence.from((loggingEvent, message) -> loggingEvent.getMessage().equals(message), "LoggingEvent contains log message");

    @BeforeEach
    void reset() {
        executorService = Executors.newFixedThreadPool(
            maxPoolSize + 5,
            new BasicThreadFactory.Builder()
                .namingPattern("TestThread-%d")
                .daemon(false)
                .priority(Thread.NORM_PRIORITY)
                .build());

        lenient().when(mockConsumer1.poll(anyLong())).thenAnswer(invocation -> tryLock());
        lenient().when(mockConsumer2.poll(anyLong())).thenAnswer(invocation -> tryLock());

        Iterator<CbsKafkaConsumer> cycle = Iterables.cycle(
                new CbsKafkaConsumer("my-topic", "my-clientId-1", mockConsumer1, 100),
                new CbsKafkaConsumer("my-topic", "my-clientId-2", mockConsumer2, 100))
            .iterator();
        Supplier<CbsKafkaConsumer> cbsKafkaConsumerSupplier = cycle::next;
        PooledClosableObjectFactory<CbsKafkaConsumer> pooledObjectFactory = new PooledClosableObjectFactory<>(cbsKafkaConsumerSupplier);
        poolConfig.setMaxWaitMillis(500);
        poolConfig.setMaxTotal(maxPoolSize);
        consumerObjectPool = Mockito.spy(new GenericObjectPool<>(pooledObjectFactory, poolConfig));
        consumerPool = new CbsKafkaConsumerPool(consumerObjectPool);

        lenient().when(mockAppender.getName()).thenReturn("MockAppender");
        lenient().doAnswer(answerVoid((VoidAnswer1<LoggingEvent>) capturedLogEvents::add)).when(mockAppender).doAppend(any());
        cbsKafkaConsumerPoolLogger = LogManager.getLogger(CbsKafkaConsumerPool.class);
        cbsKafkaConsumerPoolLogger.addAppender(mockAppender);
        cbsKafkaConsumerPoolLogger.setLevel(Level.DEBUG);
    }

    @AfterEach
    void tearDown() {
        cbsKafkaConsumerPoolLogger.removeAppender(mockAppender);
    }

    @Test
    void test_max_pool_size() {
        lock.lock();
        for (int i = 0; i < maxPoolSize + 2; i++) {
            executorService.submit(() -> consumerPool.pollDataFetcherJobs());
        }
        AssertionTestUtil.ensureAssertedFor500Millis(() -> assertThat(lock.getQueueLength()).isAtMost(maxPoolSize));
        lock.unlock();
    }

    @Test
    void caller_thread_blocked_when_pool_is_exhausted() throws Exception {
        lock.lock();
        for (int i = 0; i <= maxPoolSize; i++) {
            executorService.submit(() -> consumerPool.pollDataFetcherJobs()); //blocks
        }
        try {
            //noinspection ResultOfMethodCallIgnored
            executorService.awaitTermination(100, TimeUnit.MILLISECONDS); //let the threads work for 100 ms to make sure the pool is exhausted
        } catch (InterruptedException ignored) {
            //always happens
        }

        //all consumers in use
        Set<DataFetcherJob> jobs = consumerPool.pollDataFetcherJobs();
        assertThat(jobs).isEmpty();
        assertThat(capturedLogEvents).comparingElementsUsing(compareLogMessages).contains("Timeout waiting for idle CbsKafkaConsumer - Returning empty set of jobs");
        lock.unlock();
    }

    @Test
    void consumers_are_invoked_round_robin() throws Exception {
        //setup: Force 2 consumers to be created
        lock.lock();
        executorService.submit(() -> consumerPool.pollDataFetcherJobs()); //start 1st consumer which will block
        executorService.submit(() -> consumerPool.pollDataFetcherJobs()); //start 2nd consumer which will also block
        AssertionTestUtil.waitForAsserted(() -> assertEquals(2, lock.getQueueLength()));
        lock.unlock();
        AssertionTestUtil.waitForAsserted(() -> assertEquals(0, lock.getQueueLength()));
        verify(mockConsumer1, times(1)).poll(anyLong());
        verify(mockConsumer2, times(1)).poll(anyLong());

        //test: we do not know which consumer is next (depends on thread scheduling) so we just check that after x iterations, both consumers have been invoked evenly
        for (int i = 0; i < 10; i++) {
            consumerPool.pollDataFetcherJobs();
        }
        //should be invoked 5 times each
        verify(mockConsumer1, times(6)).poll(anyLong());
        verify(mockConsumer2, times(6)).poll(anyLong());
    }

    @Test
    void completing_job_removes_from_object_factory() {
        consumerObjectPool.setMaxIdle(0);
        lock.lock();
        executorService.submit(() -> consumerPool.pollDataFetcherJobs()); //start 1st consumer which will block
        executorService.submit(() -> consumerPool.pollDataFetcherJobs()); //start 2nd consumer which will also block
        AssertionTestUtil.waitForAsserted(() -> verify(consumerObjectPool, times(2)).borrowObject());
        lock.unlock();
        AssertionTestUtil.waitForAsserted(() -> verify(consumerObjectPool, times(2)).returnObject(any()));
    }

    @Test
    void closing_pool_closes_all_consumers() {
        doThrow(ConcurrentModificationException.class).when(mockConsumer1).close(anyLong(), any(TimeUnit.class));
        doThrow(ConcurrentModificationException.class).when(mockConsumer2).close(anyLong(), any(TimeUnit.class));

        lock.lock();
        executorService.submit(() -> consumerPool.pollDataFetcherJobs()); //start 1st consumer which will block
        executorService.submit(() -> consumerPool.pollDataFetcherJobs()); //start 2nd consumer which will also block
        AssertionTestUtil.waitForAsserted(() -> assertEquals(2, lock.getQueueLength()));

        consumerPool.close();
        lock.unlock();

        AssertionTestUtil.waitForAsserted(() -> assertEquals(0, lock.getQueueLength()));
        verify(mockConsumer1).close(anyLong(), any(TimeUnit.class));
        verify(mockConsumer1).wakeup();
        verify(mockConsumer2).close(anyLong(), any(TimeUnit.class));
        verify(mockConsumer2).wakeup();
    }


    private ConsumerRecords<String, DataFetcherJob> tryLock() {
        lock.lock(); //blocks if lock is already held
        lock.unlock();
        return new ConsumerRecords<>(Collections.emptyMap());
    }
}
