/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.kafka.serialization;

import com.google.common.collect.ImmutableMap;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.config.DtoImplConfiguration;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.kafka.DataFetcherJobImpl;
import dk.sundhedsdatastyrelsen.cbs.dto.kafka.DataFetcherJob;
import dk.sundhedsdatastyrelsen.cbs.dto.kafka.FetchType;
import dk.sundhedsdatastyrelsen.cbs.exceptions.CbsKafkaException;
import dk.sundhedsdatastyrelsen.cbs.kafka.config.CbsKafkaConfig;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import dk.sundhedsdatastyrelsen.cbs.test.TestUtils;
import org.apache.kafka.common.serialization.Deserializer;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.context.junit.jupiter.SpringJUnitConfig;

import java.util.UUID;

import static com.google.common.truth.Truth.assertThat;
import static org.junit.jupiter.api.Assertions.*;

@CbsTest
@SpringJUnitConfig({CbsKafkaConfig.class, DtoImplConfiguration.class})
class JacksonDeserializerTest {
    @Autowired
    Deserializer<DataFetcherJob> deserializer;

    @Test
    void test_deserialization() {
        String msgId = "MyMessageId";
        FetchType fetchType = FetchType.FETCH_TEST_STATUS;
        String jobId = UUID.randomUUID().toString();
        String cpr = TestUtils.randomCprNumber();

        byte[] bytes = ("{" +
            "\"msgId\":\"" + msgId + "\"," +
            "\"jobId\":\"" + jobId + "\"," +
            "\"cpr\":\"" + cpr + "\"," +
            "\"fetchCprs\":{\"" + cpr + "\":\"" + fetchType + "\"}" +
            "}").getBytes();

        DataFetcherJob job = deserializer.deserialize("topic", bytes);
        assertThat(job.getFetchCprs()).containsExactly(cpr, fetchType);
        assertEquals(msgId, job.getMsgId());
        assertEquals(jobId, job.getJobId());
        assertEquals(cpr, job.getCpr());
    }

    /**
     * @deprecated remove when support for 'type' has been removed in {@link DataFetcherJobImpl#DataFetcherJobImpl(String, FetchType, String, String, ImmutableMap)}
     */
    @Test
    @Deprecated
    void test_backwards_compatible_deserialization() {
        String msgId = "MyMessageId";
        FetchType fetchType = FetchType.FETCH_TEST_STATUS;
        String jobId = UUID.randomUUID().toString();
        String cpr = TestUtils.randomCprNumber();

        byte[] oldBytes = ("{\"msgId\":\"" + msgId + "\",\"type\":\"" + fetchType + "\",\"jobId\":\"" + jobId + "\",\"cpr\":\"" + cpr + "\"}").getBytes();

        DataFetcherJob job = deserializer.deserialize("topic", oldBytes);
        assertEquals(msgId, job.getMsgId());
        assertThat(job.getFetchCprs()).containsExactly(cpr, fetchType);
        assertEquals(jobId, job.getJobId());
        assertEquals(cpr, job.getCpr());
    }

    @Test
    void test_deserialization_with_additional_cprs() {
        String msgId = "MyMessageId";
        FetchType fetchType1 = FetchType.FETCH_TEST_STATUS;
        FetchType fetchType2 = FetchType.FETCH_ALL;
        String jobId = UUID.randomUUID().toString();
        String cpr1 = TestUtils.randomCprNumber();
        String cpr2 = TestUtils.randomCprNumber();

        byte[] bytes = ("{" +
            "\"msgId\":\"" + msgId + "\"," +
            "\"jobId\":\"" + jobId + "\"," +
            "\"cpr\":\"" + cpr1 + "\"," +
            "\"fetchCprs\":{\"" + cpr1 + "\":\"" + fetchType1 + "\",\"" + cpr2 + "\":\"" + fetchType2 + "\"}" +
            "}").getBytes();

        DataFetcherJob job = deserializer.deserialize("topic", bytes);
        assertEquals(msgId, job.getMsgId());
        assertEquals(jobId, job.getJobId());
        assertEquals(cpr1, job.getCpr());
        assertThat(job.getFetchCprs()).containsExactly(cpr1, fetchType1, cpr2, fetchType2);
    }

    @Test
    void throws_exception_if_not_configured() {
        JacksonDeserializer<DataFetcherJob> deserializer = new JacksonDeserializer<>(null);

        String msgId = "MyMessageId";
        FetchType fetchType = FetchType.FETCH_TEST_STATUS;
        String jobId = UUID.randomUUID().toString();
        String cpr = TestUtils.randomCprNumber();

        byte[] bytes = ("{" +
            "\"msgId\":\"" + msgId + "\"," +
            "\"jobId\":\"" + jobId + "\"," +
            "\"cpr\":\"" + cpr + "\"," +
            "\"fetchCprs\":{\"" + cpr + "\":\"" + fetchType + "\"}" +
            "}").getBytes();


        assertThrows(CbsKafkaException.class, () -> deserializer.deserialize("topic", bytes));
    }

    @Test
    @DisplayName("Ignore unknown properties - for making it possible to do rolling upgrades that introduces new JSON properties")
    void ignores_unknown_properties() {
        String msgId = "MyMessageId";
        FetchType fetchType = FetchType.FETCH_TEST_STATUS;
        String jobId = UUID.randomUUID().toString();
        String cpr = TestUtils.randomCprNumber();

        byte[] bytes = ("{" +
            "\"fakeProp\":\"blah\"," +
            "\"msgId\":\"" + msgId + "\"," +
            "\"jobId\":\"" + jobId + "\"," +
            "\"cpr\":\"" + cpr + "\"," +
            "\"fetchCprs\":{\"" + cpr + "\":\"" + fetchType + "\"}" +
            "}").getBytes();


        DataFetcherJob job = deserializer.deserialize("topic", bytes);
        assertEquals(msgId, job.getMsgId());
        assertEquals(jobId, job.getJobId());
        assertEquals(cpr, job.getCpr());
    }

    @Test
    void returns_null_on_null_input() {
        //like the other Kafka deserializers
        assertNull(deserializer.deserialize("topic", null));
    }
}
