/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.kafka.status;

import dk.sundhedsdatastyrelsen.cbs.status.StatusCheck;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import org.apache.kafka.clients.admin.AdminClient;
import org.apache.kafka.common.Node;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Answers;
import org.mockito.Mock;

import java.util.Collections;
import java.util.concurrent.ExecutionException;

import static org.junit.jupiter.api.Assertions.assertAll;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

@CbsTest
class KafkaStatusCheckerTest {

    @Mock(answer = Answers.RETURNS_DEEP_STUBS)
    private AdminClient adminClient;
    private KafkaStatusChecker kafkaStatusChecker;

    @BeforeEach
    void setUp() {
        kafkaStatusChecker = new KafkaStatusChecker(() -> adminClient, "KAFKA");
    }

    @Test
    void ok_if_there_are_nodes() throws ExecutionException, InterruptedException {
        when(adminClient.describeCluster(any()).nodes().get()).thenReturn(Collections.singleton(mock(Node.class)));

        final StatusCheck status = kafkaStatusChecker.checkStatus();

        assertAll(
            () -> assertEquals("true", status.getInfo().get("Is OK")),
            () -> assertNull(status.getInfo().get("Exception")),
            () -> assertEquals(StatusCheck.Status.OK, status.getStatus()),
            () -> assertEquals("KAFKA", status.getTitle())
        );
    }

    @Test
    void not_ok_if_there_are_no_nodes() throws ExecutionException, InterruptedException {
        when(adminClient.describeCluster(any()).nodes().get()).thenReturn(Collections.emptyList());

        final StatusCheck status = kafkaStatusChecker.checkStatus();

        assertAll(
            () -> assertEquals("false", status.getInfo().get("Is OK")),
            () -> assertNull(status.getInfo().get("Exception")),
            () -> assertEquals(StatusCheck.Status.ERROR, status.getStatus()),
            () -> assertEquals("KAFKA", status.getTitle())
        );
    }

    @Test
    void not_ok_if_admin_client_interrupted() throws ExecutionException, InterruptedException {
        when(adminClient.describeCluster(any()).nodes().get()).thenThrow(new InterruptedException("INTERRUPT!"));

        final StatusCheck status = kafkaStatusChecker.checkStatus();

        assertAll(
            () -> assertEquals("false", status.getInfo().get("Is OK")),
            () -> assertEquals("INTERRUPT!", status.getInfo().get("Exception")),
            () -> assertEquals(StatusCheck.Status.ERROR, status.getStatus()),
            () -> assertEquals("KAFKA", status.getTitle())
        );
    }

    @Test
    void not_ok_if_admin_client_fails() throws ExecutionException, InterruptedException {
        when(adminClient.describeCluster(any()).nodes().get()).thenThrow(new ExecutionException("ERROR!", new RuntimeException()));

        final StatusCheck status = kafkaStatusChecker.checkStatus();

        assertAll(
            () -> assertEquals("false", status.getInfo().get("Is OK")),
            () -> assertEquals("ERROR!", status.getInfo().get("Exception")),
            () -> assertEquals(StatusCheck.Status.ERROR, status.getStatus()),
            () -> assertEquals("KAFKA", status.getTitle())
        );
    }
}
