/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.nas.facade;

import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.sundhedsdatastyrelsen.cbs.dgws.DGWSHeaders;
import dk.sundhedsdatastyrelsen.cbs.dgws.GenericDGWSHeaderProvider;
import dk.sundhedsdatastyrelsen.cbs.exceptions.NasException;
import org.oasis_open.docs.wsn.b_2.FilterType;
import org.oasis_open.docs.wsn.b_2.GetMessages;
import org.oasis_open.docs.wsn.b_2.GetMessagesResponse;
import org.oasis_open.docs.wsn.b_2.NotificationMessageHolderType;
import org.oasis_open.docs.wsn.b_2.ObjectFactory;
import org.oasis_open.docs.wsn.b_2.Subscribe;
import org.oasis_open.docs.wsn.b_2.SubscribeResponse;
import org.oasis_open.docs.wsn.b_2.TopicExpressionType;
import org.oasis_open.docs.wsn.bw_2.CreatePullPoint;
import org.oasis_open.docs.wsn.bw_2.NotificationProducer;
import org.oasis_open.docs.wsn.bw_2.PullPoint;
import org.oasis_open.docs.wss._2004._01.oasis_200401_wss_wssecurity_secext_1_0.Security;

import javax.xml.ws.Holder;
import javax.xml.ws.WebServiceFeature;
import javax.xml.ws.wsaddressing.W3CEndpointReference;
import java.math.BigInteger;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import static dk.sundhedsdatastyrelsen.cbs.metrics.TimerMetric.timedCall;

public class NasClientImpl implements NasClient {

    private final ObjectFactory b2ObjectFactory = new ObjectFactory();

    private final GenericDGWSHeaderProvider<Security, Header> headerProvider;
    private final NotificationProducer notificationProducer;
    private final CreatePullPoint createPullPointPort;
    private final PullPointFactory pullPointFactory;
    private final Map<String, PullPoint> cachedPullPoints = new HashMap<>();

    public NasClientImpl(GenericDGWSHeaderProvider<Security, Header> headerProvider, NotificationProducer notificationProducer, CreatePullPoint createPullPointPort, PullPointFactory pullPointFactory) {
        this.headerProvider = headerProvider;
        this.notificationProducer = notificationProducer;
        this.createPullPointPort = createPullPointPort;
        this.pullPointFactory = pullPointFactory;
    }

    @Override
    public W3CEndpointReference createPullPoint() {
        try {
            DGWSHeaders<Security, Header> headers = headerProvider.getHeaders();
            return timedCall("nas_createPullPoint", () -> createPullPointPort.createPullPoint(b2ObjectFactory.createCreatePullPoint(), headers.getSecurity(), new Holder<>(headers.getMedComHeader())).getPullPoint());
        } catch (Exception e) {
            throw new NasException("Failed to create Pull point", e);
        }
    }

    @Override
    public W3CEndpointReference subscribe(W3CEndpointReference consumerReference, String topic) {
        TopicExpressionType topicExpressionType = b2ObjectFactory.createTopicExpressionType();
        topicExpressionType.setDialect(Constants.TOPIC_EXPRESSION_SIMPLE);
        topicExpressionType.getContent().add(topic);

        FilterType filterType = b2ObjectFactory.createFilterType();
        filterType.getAny().add(b2ObjectFactory.createTopicExpression(topicExpressionType));

        Subscribe subscribe = b2ObjectFactory.createSubscribe();
        subscribe.setConsumerReference(consumerReference);
        subscribe.setFilter(filterType);

        try {
            DGWSHeaders<Security, Header> headers = headerProvider.getHeaders();
            SubscribeResponse response = timedCall("nas_subscribe", () -> notificationProducer.subscribe(subscribe, headers.getSecurity(), new Holder<>(headers.getMedComHeader())));
            return response.getSubscriptionReference();
        } catch (Exception e) {
            throw new NasException(String.format("Failed to create subscription for topic '%s' on pull point '%s'", topic, EndpointReferences.extractAddress(consumerReference)), e);
        }
    }

    @Override
    public List<NotificationMessageHolderType> getMessages(W3CEndpointReference consumerReference, long maxNumberOfMessages, UUID messageId, WebServiceFeature... features) {
        final String address = EndpointReferences.extractAddress(consumerReference);
        PullPoint pullPoint = cachedPullPoints.computeIfAbsent(address, r -> pullPointFactory.getPullPoint(address, features));

        GetMessages getMessages = b2ObjectFactory.createGetMessages();
        getMessages.setMaximumNumber(BigInteger.valueOf(maxNumberOfMessages));

        try {
            DGWSHeaders<Security, Header> headers = headerProvider.getHeaders();
            Header medComHeader = headers.getMedComHeader();
            medComHeader.getLinking().setMessageID(messageId.toString());
            GetMessagesResponse response = timedCall("nas_getMessages", () -> pullPoint.getMessages(getMessages, headers.getSecurity(), new Holder<>(medComHeader)));
            return response.getNotificationMessage();
        } catch (Exception e) {
            throw new NasException(String.format("Failed to get messages on pull point '%s'", address), e);
        }
    }


}
