/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.nas.facade;

import org.apache.log4j.Logger;
import org.oasis_open.docs.wsn.b_2.NotificationMessageHolderType;

import javax.xml.ws.wsaddressing.W3CEndpointReference;
import java.util.Collections;
import java.util.List;
import java.util.UUID;

/**
 * Consumes messages from Nas based on NasConsumerProperties
 * Messages are processed in chunks by a NasNotificationProcessor
 * NasSubscriptionRepository handles persistence of PullPoints and messageIds
 */
public class NasConsumer {

    private static final Logger log = Logger.getLogger(NasConsumer.class);

    private final NasClient nasClient;
    private final NasSubscriptionRepository repository;
    private final NasNotificationProcessor processor;

    private final long chunkSize;
    private final int chunkMaxLimit;
    private final String topic;

    public NasConsumer(NasConsumerProperties properties, NasClient nasClient, NasSubscriptionRepository repository, NasNotificationProcessor processor) {
        topic = properties.getTopic();
        chunkSize = properties.getChunkSize();
        chunkMaxLimit = properties.getChunkMaxLimit();

        this.nasClient = nasClient;

        this.repository = repository;
        this.processor = processor;
    }

    public void trigger() {
        W3CEndpointReference pullPoint = repository.getPullPoint(topic);

        boolean hasMoreMessages = true;
        for (int chunks = 0; chunks < chunkMaxLimit && hasMoreMessages; chunks++) {
            hasMoreMessages = process(pullPoint);
        }
    }

    private boolean process(W3CEndpointReference pullPoint) {
        UUID messageId = repository.getNextMessageId(topic);

        log.debug(String.format("Requesting messages from topic '%s' with message id '%s'", topic, messageId));

        List<NotificationMessageHolderType> messages = nasClient.getMessages(pullPoint, chunkSize, messageId);
        boolean hasMessages = !messages.isEmpty();

        if (hasMessages) {
            log.debug(String.format("Processing messages from topic '%s' with message id '%s'", topic, messageId));

            processor.process(Collections.unmodifiableList(messages));
            log.debug(String.format("Processing of messages from topic '%s' with message id '%s' completed", topic, messageId));
        }

        repository.commit(topic);

        return hasMessages;
    }
}
