/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.nas.facade;

import dk.nsi.advis.v10.NotifyContent;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import dk.sundhedsdatastyrelsen.coronapass.ws.JobStatus;
import dk.sundhedsdatastyrelsen.coronapass.ws.JobStatusNotification;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.oasis_open.docs.wsn.b_2.NotificationMessageHolderType;
import org.oasis_open.docs.wsn.b_2.TopicExpressionType;

import javax.xml.bind.JAXBElement;
import java.util.ArrayList;
import java.util.List;
import java.util.UUID;

import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.mockito.Mockito.*;

@CbsTest
@SuppressWarnings({"java:S5979", "unchecked"}) /*SonarLint cannot see that @CbsTest has @ExtendWith(MockitoExtension.class)*/
class NasConsumerTest {

    private final org.oasis_open.docs.wsn.b_2.ObjectFactory b2ObjectFactory = new org.oasis_open.docs.wsn.b_2.ObjectFactory();
    private final dk.nsi.advis.v10.ObjectFactory nsiObjectFactory = new dk.nsi.advis.v10.ObjectFactory();
    private final dk.sundhedsdatastyrelsen.coronapass.ws.ObjectFactory cbsObjectFactory = new dk.sundhedsdatastyrelsen.coronapass.ws.ObjectFactory();

    @Mock
    private NasClient nasClient;

    @Mock
    private NasSubscriptionRepository repository;

    @Mock
    private NasNotificationProcessor nasNotificationProcessor;

    private NasConsumer nasConsumer;

    @BeforeEach
    void setUp() {
        NasConsumerProperties properties = new NasConsumerProperties("myTopic", 10L, 3);
        nasConsumer = new NasConsumer(properties, nasClient, repository, nasNotificationProcessor);

    }

    @Test
    void should_not_process_when_there_are_no_messages() {
        when(nasClient.getMessages(any(), anyLong(), any())).thenReturn(getMessages(0));

        nasConsumer.trigger();

        verifyNoInteractions(nasNotificationProcessor);
    }

    @Test
    void should_process_one_chunk_when_there_is_one_chunk_of_messages() {
        when(nasClient.getMessages(any(), anyLong(), any())).thenReturn(getMessages(10), getMessages(0));

        nasConsumer.trigger();

        verify(nasNotificationProcessor).process(any());
    }

    @Test
    void should_process_two_chunks_when_there_are_two_chunks_of_messages() {
        when(nasClient.getMessages(any(), anyLong(), any())).thenReturn(getMessages(10), getMessages(10), getMessages(0));

        nasConsumer.trigger();

        verify(nasNotificationProcessor, times(2)).process(any());
    }

    @Test
    void should_stop_processing_after_reaching_chunk_max_limit() {
        // keeps returning messages
        when(nasClient.getMessages(any(), anyLong(), any())).thenReturn(getMessages(10));

        nasConsumer.trigger();

        verify(nasNotificationProcessor, times(3)).process(any());
    }

    @Test
    void should_commit_after_each_chunk() {
        // keeps returning messages
        when(nasClient.getMessages(any(), anyLong(), any())).thenReturn(getMessages(10));

        nasConsumer.trigger();

        verify(repository, times(3)).commit(any());
    }

    @Test
    void should_fail_if_processor_fails() {
        // keeps returning messages
        when(nasClient.getMessages(any(), anyLong(), any())).thenReturn(getMessages(10));
        doThrow(new RuntimeException("error")).when(nasNotificationProcessor).process(any());

        assertThrows(RuntimeException.class, () -> nasConsumer.trigger());
    }

    @Test
    void should_not_commit_if_processor_fails() {
        when(nasClient.getMessages(any(), anyLong(), any())).thenReturn(getMessages(10));
        doThrow(new RuntimeException("error")).when(nasNotificationProcessor).process(any());

        assertThrows(RuntimeException.class, () -> nasConsumer.trigger());

        verify(repository, never()).commit(any());
    }

    @Test
    void should_commit_until_processor_fails() {
        // scenario: First chunk is processed successfully, second chunk fails
        // behaviour: Should commit first chunk

        when(nasClient.getMessages(any(), anyLong(), any())).thenReturn(getMessages(10));
        // first succeed, then fail
        doNothing()
            .doThrow(new RuntimeException("error"))
            .when(nasNotificationProcessor).process(any());

        assertThrows(RuntimeException.class, () -> nasConsumer.trigger());

        verify(repository).commit(any());
    }

    private List<NotificationMessageHolderType> getMessages(int count) {
        ArrayList<NotificationMessageHolderType> messages = new ArrayList<>();
        for (int i = 0; i < count; i++) {
            messages.add(newMessage());
        }
        return messages;
    }

    private NotificationMessageHolderType newMessage() {
        JobStatusNotification jobStatusNotification = cbsObjectFactory.createJobStatusNotification();
        jobStatusNotification.setJobId(UUID.randomUUID().toString());
        jobStatusNotification.setJobStatus(JobStatus.OK);

        JAXBElement<JobStatusNotification> payload = cbsObjectFactory.createJobStatusNotification(jobStatusNotification);

        NotifyContent notifyContent = nsiObjectFactory.createNotifyContent();
        notifyContent.setAny(payload);
        notifyContent.setId("0506504003");
        notifyContent.setIdType(Constants.ID_TYPE_CPR);
        notifyContent.setIsSystemNotification(false);

        NotificationMessageHolderType.Message message = b2ObjectFactory.createNotificationMessageHolderTypeMessage();
        message.setAny(notifyContent);

        TopicExpressionType topicExpressionType = b2ObjectFactory.createTopicExpressionType();
        topicExpressionType.setDialect(Constants.TOPIC_EXPRESSION_SIMPLE);
        topicExpressionType.getContent().add("myTopic");

        NotificationMessageHolderType notificationMessageHolderType = b2ObjectFactory.createNotificationMessageHolderType();
        notificationMessageHolderType.setMessage(message);
        notificationMessageHolderType.setTopic(topicExpressionType);
        return notificationMessageHolderType;
    }
}
