/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.logging;

import dk.sds.nsp.audit.Audit;
import dk.sds.nsp.audit.AuditBuilder;
import dk.sds.nsp.security.Security;
import dk.sds.nsp.security.SecurityContext;
import dk.sundhedsdatastyrelsen.cbs.CbsLogging;
import org.apache.log4j.MDC;

import java.util.Collections;
import java.util.Map;

public class AuditLogger {

    private final String component;

    private final Map<String, Object> extraConstantEntries;

    public AuditLogger(String component) {
        this(component, Collections.emptyMap());
    }

    public AuditLogger(String component, Map<String, Object> extraConstantEntries) {
        this.component = component;
        this.extraConstantEntries = extraConstantEntries;
    }

    public void log(AuditData auditData, String context) {
        AuditBuilder auditBuilder = Audit.createAuditBuilder();
        SecurityContext securityContext = Security.getSecurityContext();

        securityContext.getActingUser().ifPresent(actingUser -> {
            auditData.withSensitivePersonalInformation("actingUserIdentifier", actingUser.getIdentifier())
                .withRegularPersonalInformation("actingUserIdentifierFormat", actingUser.getIdentifierFormat() != null ? actingUser.getIdentifierFormat().name() : null)
                .withRegularPersonalInformation("actingUserType", actingUser.getUserType() != null ? actingUser.getUserType().name() : null)
                .withSensitivePersonalInformation("actingUserGivenName", actingUser.getGivenName().orElse(null))
                .withSensitivePersonalInformation("actingUserSurName", actingUser.getSurname().orElse(null));

            actingUser.getCredentials().ifPresent(credentials ->
                auditData.withRegularPersonalInformation("actingUserEducationCode", credentials.getEducationCode().orElse(null))
                    .withSensitivePersonalInformation("actingUserAuthorizationCode", credentials.getAuthorizationCode().orElse(null))
            );
        });

        securityContext.getPrincipalUser().ifPresent(principalUser -> {
            auditData.withSensitivePersonalInformation("principalUserIdentifier", principalUser.getIdentifier())
                .withRegularPersonalInformation("principalUserIdentifierFormat", principalUser.getIdentifierFormat() != null ? principalUser.getIdentifierFormat().name() : null)
                .withRegularPersonalInformation("principalUserType", principalUser.getUserType() != null ? principalUser.getUserType().name() : null)
                .withSensitivePersonalInformation("principalUserGivenName", principalUser.getGivenName().orElse(null))
                .withSensitivePersonalInformation("principalUserSurName", principalUser.getSurname().orElse(null));

            principalUser.getCredentials().ifPresent(credentials ->
                auditData.withRegularPersonalInformation("principalUserEducationCode", credentials.getEducationCode().orElse(null))
                    .withSensitivePersonalInformation("principalUserAuthorizationCode", credentials.getAuthorizationCode().orElse(null))
                    .withSensitivePersonalInformation("principalUserNationalRole", credentials.getNationalRole() != null ? credentials.getNationalRole().orElse(null) : null));
        });

        securityContext.getOrganisation().ifPresent(organisation ->
            auditData.withNonPersonalInformation("orgName", organisation.getName().orElse(null))
                .withNonPersonalInformation("orgIdentifier", organisation.getIdentifier())
                .withNonPersonalInformation("orgIdentifierFormat", organisation.getIdentifierFormat() != null ? organisation.getIdentifierFormat().name() : null));

        securityContext.getClient().ifPresent(client ->
            auditData.withRegularPersonalInformation("clientName", client.getName().orElse(null))
                .withRegularPersonalInformation("clientPersistentUniqueKey", client.getPersistentUniqueKey().map(SecurityContext.PersistentUniqueKey::value).orElse(null))
        );

        auditData.withNonPersonalInformation("traceId", MDC.get(CbsLogging.TRACE_ID_PROP));
        auditData.withNonPersonalInformation("msgId", MDC.get(CbsLogging.MESSAGE_ID_PROP));

        extraConstantEntries.forEach(auditData::withNonPersonalInformation);

        auditData.write(auditBuilder, component, context);
    }

}
