/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.monitoring;

import dk.sundhedsdatastyrelsen.cbs.status.StatusCheck;
import dk.sundhedsdatastyrelsen.cbs.status.StatusChecker;
import org.apache.log4j.Logger;

import javax.servlet.ServletContext;
import javax.servlet.annotation.WebServlet;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.UncheckedIOException;
import java.net.HttpURLConnection;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;
import java.util.Properties;
import java.util.function.BinaryOperator;
import java.util.stream.Collectors;

import static dk.sundhedsdatastyrelsen.cbs.monitoring.IsAliveServletContextListener.START_TIME;

@WebServlet("/isalive")
public class IsAliveServlet extends HttpServlet {
    private static String serviceName = null;
    private static final List<StatusChecker> statusCheckers = new ArrayList<>();

    private static final Logger log = Logger.getLogger(IsAliveServlet.class);
    private static BuildData buildData;

    private static final String CONTENT_TYPE = "text/plain";
    private static final String CR = "\n";

    @Override
    protected void doGet(HttpServletRequest request, HttpServletResponse response) {

        try {
            BuildData bd = getBuildData(request.getSession().getServletContext());

            StringBuilder s = new StringBuilder();
            response.setContentType(CONTENT_TYPE);
            s.append("Service name: ").append((serviceName)).append(CR);
            s.append("Current time: ").append(Instant.now()).append(CR);
            s.append("Service start time: ").append(START_TIME).append(CR);
            s.append("Build-Date: ").append(bd.builtDate).append(CR);
            s.append("Build-Version: ").append(bd.builtVersion).append(CR);

            final List<StatusCheck> statusChecks = statusCheckers.stream()
                .map(StatusChecker::checkStatus)
                .collect(Collectors.toList());

            int status = statusChecks.stream()
                .map(StatusCheck::getStatus)
                .reduce(BinaryOperator.maxBy(Comparator.comparing(StatusCheck.Status::severity)))
                .map(StatusCheck.Status::httpCode)
                .orElse(200); // if statusCheckers is empty

            response.getWriter().println("Status: " + status);
            response.getWriter().println(s);

            statusChecks.forEach(check -> writeToResponse(response, check));

            response.setHeader("X-SERVICE-VERSION", bd.builtVersion);
            response.setStatus(status);
        } catch (Exception ex) {
            response.setStatus(HttpURLConnection.HTTP_INTERNAL_ERROR);
            log.error("Error in isAlive", ex);
        }
    }

    private void writeToResponse(HttpServletResponse response, StatusCheck check) {
        final StringBuilder builder = new StringBuilder(check.getTitle()).append(": ").append(CR);

        check.getInfo().forEach(
            (key, value) -> builder.append("  ").append(key).append(": ").append(value).append(CR)
        );

        try {
            response.getWriter().println(builder);
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        }
    }

    private static BuildData getBuildData(ServletContext loader) throws IOException {
        if (buildData == null) {
            InputStream stream = loader.getResourceAsStream("/META-INF/MANIFEST.MF");
            if (stream != null) {
                Properties prop = new Properties();
                prop.load(stream);
                stream.close();
                buildData = new BuildData(prop.getProperty("Built-Version"), prop.getProperty("Built-Date"));

                log.info("Got build data: " + buildData);
            } else {
                throw new FileNotFoundException("/META-INF/MANIFEST.MF could not be found");
            }
        }
        return buildData;
    }

    public static void setServiceName(String serviceName) {
        IsAliveServlet.serviceName = serviceName;
    }

    public static void addStatusChecks(StatusChecker... checks) {
        IsAliveServlet.statusCheckers.addAll(Arrays.asList(checks));
    }
}

