/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs;

import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import java.util.Arrays;
import java.util.Comparator;
import java.util.function.Supplier;

public final class Objects {

    private Objects() {}

    @SafeVarargs
    @Nullable
    public static <T> T max(@NonNull final Comparator<T> comparator, @NonNull T... values) {
        return Arrays.stream(values)
            .filter(java.util.Objects::nonNull)
            .max(comparator)
            .orElse(null);
    }

    @SafeVarargs
    @NonNull
    public static <T> T firstNonNullOrDefault(@NonNull final T defaultValue, @NonNull final T... values) {
        return Arrays.stream(values)
            .filter(java.util.Objects::nonNull)
            .findFirst()
            .orElse(defaultValue);
    }

    @NonNull
    public static <T> T nonNullOrDefault(@Nullable final T value, @NonNull final T defaultValue) {
        return value == null ? defaultValue : value;
    }


    @SafeVarargs
    @NonNull
    public static <T> T firstNonNullSupplierOrDefault(@NonNull final T defaultValue, @NonNull final Supplier<T>... suppliers) {
        return Arrays.stream(suppliers)
            .filter(java.util.Objects::nonNull)
            .map(Supplier::get)
            .filter(java.util.Objects::nonNull)
            .findFirst()
            .orElse(defaultValue);
    }

    @SafeVarargs
    @NonNull
    public static <T> T firstNonNullOrDefaultSupplier(@NonNull final Supplier<T> defaultSupplier, @NonNull final Supplier<T>... suppliers) {
        return Arrays.stream(suppliers)
            .filter(java.util.Objects::nonNull)
            .map(Supplier::get)
            .filter(java.util.Objects::nonNull)
            .findFirst()
            .orElseGet(defaultSupplier);
    }

    @SafeVarargs
    @Nullable
    public static <T> T firstNonNull(@NonNull final T... values) {
        return Arrays.stream(values)
            .filter(java.util.Objects::nonNull)
            .findFirst()
            .orElse(null);
    }

    @SafeVarargs
    @Nullable
    public static <T> T firstNonNullSupplier(@NonNull final Supplier<T>... suppliers) {
        return Arrays.stream(suppliers)
            .filter(java.util.Objects::nonNull)
            .map(Supplier::get)
            .filter(java.util.Objects::nonNull)
            .findFirst()
            .orElse(null);
    }

}
