/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.status;

import org.apache.log4j.Logger;

import java.security.cert.CertificateExpiredException;
import java.security.cert.CertificateNotYetValidException;
import java.security.cert.X509Certificate;
import java.text.SimpleDateFormat;
import java.time.Duration;
import java.time.Instant;
import java.util.LinkedHashMap;
import java.util.Map;


public class CertificateStatusChecker implements StatusChecker {

    private static final Logger LOGGER = Logger.getLogger(CertificateStatusChecker.class);

    private final SimpleDateFormat format = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.SSS Z");

    private final X509Certificate certificate;
    private final String title;

    public CertificateStatusChecker(X509Certificate certificate, String title) {
        this.certificate = certificate;
        this.title = title;
    }

    @Override
    public StatusCheck checkStatus() {
        final Map<String, String> info = new LinkedHashMap<>();
        info.put("Subject", certificate.getSubjectDN().getName());
        info.put("Public Key Algorithm", certificate.getPublicKey().getAlgorithm());
        info.put("Not Before", format.format(certificate.getNotBefore()));
        info.put("Not After", format.format(certificate.getNotAfter()));
        boolean isValid = false;
        try {
            certificate.checkValidity();
            isValid = true;
        } catch (CertificateExpiredException | CertificateNotYetValidException e) {
            info.put("Exception", e.getMessage());
        }

        info.put("Is Valid", String.valueOf(isValid));


        StatusCheck.Status status = StatusCheck.Status.ERROR;

        if (isValid) {
            status = StatusCheck.Status.OK;
            if (Instant.now().plus(Duration.ofDays(30)).isAfter(certificate.getNotAfter().toInstant())) {
                LOGGER.warn(String.format("Certificate '%s' will expire in less than one month (notAfter='%s')", title, format.format(certificate.getNotAfter())));
                status = StatusCheck.Status.WARN;
            }
        }

        return new StatusCheck(title, info, status);
    }

}
