/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs;

import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import java.util.Comparator;
import java.util.function.Supplier;
import java.util.stream.Stream;

import static java.util.Comparator.naturalOrder;
import static org.junit.jupiter.api.Assertions.assertSame;
import static org.junit.jupiter.params.provider.Arguments.arguments;

class ObjectsTest {

    @SafeVarargs
    @ParameterizedTest
    @MethodSource("maxSource")
    final <T> void max(@Nullable final T max, @NonNull final Comparator<T> comparator, @NonNull final T... values) {
        assertSame(max, Objects.max(comparator, values));
    }

    @NonNull
    private static Stream<Arguments> maxSource() {
        final String a = "a";
        final String z = "z";
        final Integer i1 = 1;
        final int i2 = 2;
        final Integer i3 = 3;
        return Stream.of(
            arguments(i3, naturalOrder(), new Integer[] { i2, i1, i3 }),
            arguments(i1, naturalOrder().reversed(), new Integer[] { i2, i1, i3 }),
            arguments(z, naturalOrder(), new String[] { a, null, z, "o" }),
            arguments(a, naturalOrder().reversed(), new String[] { a, null, z,"o" }),
            arguments(null, naturalOrder(), new Object[2]));
    }

    @ParameterizedTest
    @MethodSource("firstNonNullOrDefaultSource")
    <T> void firstNonNullOrDefault(
        @NonNull final T expected,
        @NonNull final T defaultValue,
        @Nullable final T value1,
        @Nullable final T value2,
        @Nullable final T value3) {
        assertSame(
            expected,
            Objects.firstNonNullOrDefault(
                defaultValue,
                value1,
                value2,
                value3));
    }

    @ParameterizedTest
    @MethodSource("firstNonNullOrDefaultSource")
    <T> void firstNonNullOrDefaultSupplier(
        @NonNull final T expected,
        @NonNull final T defaultValue,
        @Nullable final T value1,
        @Nullable final T value2,
        @Nullable final T value3) {
        assertSame(
            expected,
            Objects.firstNonNullOrDefaultSupplier(
                supplier(defaultValue),
                supplier(value1),
                supplier(value2),
                supplier(value3)));
    }

    @ParameterizedTest
    @MethodSource("firstNonNullOrDefaultSource")
    final <T> void firstNonNullSupplierOrDefault(
        @NonNull final T expected,
        @NonNull final T defaultValue,
        @Nullable final T value1,
        @Nullable final T value2,
        @Nullable final T value3) {
        assertSame(
            expected,
            Objects.firstNonNullSupplierOrDefault(
                defaultValue,
                supplier(value1),
                supplier(value2),
                supplier(value3)));
    }

    @NonNull
    private static Stream<Arguments> firstNonNullOrDefaultSource() {
        final Object defaultValue = new Object();
        final Object firstNonNull = new Object();
        final Object nonNull = new Object();
        return Stream.of(
            arguments(defaultValue, defaultValue, null, null, null),
            arguments(firstNonNull, defaultValue, null, firstNonNull, nonNull));
    }

    @ParameterizedTest
    @MethodSource("firstNotNullSource")
    <T> void firstNonNull(
        @NonNull final T expected,
        @Nullable final T value1,
        @Nullable final T value2,
        @Nullable final T value3) {
        assertSame(
            expected,
            Objects.firstNonNull(
                value1,
                value2,
                value3));
    }

    @ParameterizedTest
    @MethodSource("firstNotNullSource")
    <T> void firstNonNullSupplier(
        @NonNull final T expected,
        @Nullable final T value1,
        @Nullable final T value2,
        @Nullable final T value3) {
        assertSame(
            expected,
            Objects.firstNonNullSupplier(
                supplier(value1),
                supplier(value2),
                supplier(value3)));
    }

    @NonNull
    private static Stream<Arguments> firstNotNullSource() {
        final Object firstNonNull = new Object();
        final Object nonNull = new Object();
        return Stream.of(
            arguments(null, null, null, null),
            arguments(firstNonNull, null, firstNonNull, nonNull));
    }

    @ParameterizedTest
    @MethodSource("nonNullOrDefaultSource")
    <T> void nonNullOrDefault(
        @NonNull final T expected,
        @Nullable final T value,
        @NonNull final T defaultValue) {
        assertSame(
            expected,
            Objects.nonNullOrDefault(value, defaultValue));
    }

    @NonNull
    private static Stream<Arguments> nonNullOrDefaultSource() {
        final Object defaultValue = new Object();
        final Object nonNull = new Object();
        return Stream.of(
            arguments(defaultValue, null, defaultValue),
            arguments(nonNull, nonNull, defaultValue));
    }


    @NonNull
    private static <T> Supplier<T> supplier(@Nullable final T value) {
        return () -> value;
    }

}
