/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.metrics;

import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import org.apache.log4j.Appender;
import org.apache.log4j.Logger;
import org.apache.log4j.spi.LoggingEvent;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Mock;
import org.mockito.Mockito;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

@CbsTest
class MetricsTest {

    @Mock
    private Appender appender;

    @Captor
    private ArgumentCaptor<LoggingEvent> loggingEventArgumentCaptor;

    @BeforeEach
    void setUp() {
        Logger logger = Logger.getLogger("METRICS");
        logger.addAppender(appender);
    }

    @Test
    void should_log_metrics() {
        MetricsBuilder.newBuilder("metrics_test")
            .addLabel("label1", "value1")
            .addLabel("label2", "value2")
            .emit();

        Mockito.verify(appender).doAppend(loggingEventArgumentCaptor.capture());

        LoggingEvent loggingEvent = loggingEventArgumentCaptor.getValue();

        assertEquals("metric=\"metrics_test\" label1=\"value1\" label2=\"value2\"", loggingEvent.getRenderedMessage());
    }

    @Test
    void should_time_supplier() {
        String suppliedValue = TimerMetric.timedGet("supplier", () -> "hello");
        Mockito.verify(appender).doAppend(loggingEventArgumentCaptor.capture());
        LoggingEvent loggingEvent = loggingEventArgumentCaptor.getValue();

        assertTrue(loggingEvent.getRenderedMessage().contains("metric=\"supplier\""));
        assertTrue(loggingEvent.getRenderedMessage().contains("durationMillis=\""));

        assertEquals("hello", suppliedValue);
    }

    @Test
    void should_time_callable() {
        String suppliedValue = TimerMetric.timedCall("callable", () -> "hello");
        Mockito.verify(appender).doAppend(loggingEventArgumentCaptor.capture());
        LoggingEvent loggingEvent = loggingEventArgumentCaptor.getValue();

        assertTrue(loggingEvent.getRenderedMessage().contains("metric=\"callable\""));
        assertTrue(loggingEvent.getRenderedMessage().contains("durationMillis=\""));

        assertEquals("hello", suppliedValue);
    }

    @Test
    void should_time_throwing_callable() {
        RuntimeException exception = new RuntimeException();
        RuntimeException thrown = assertThrows(RuntimeException.class, () -> TimerMetric.timedCall("thrower", () -> {
            throw exception;
        }));

        Mockito.verify(appender).doAppend(loggingEventArgumentCaptor.capture());
        LoggingEvent loggingEvent = loggingEventArgumentCaptor.getValue();

        assertTrue(loggingEvent.getRenderedMessage().contains("metric=\"thrower\""));
        assertTrue(loggingEvent.getRenderedMessage().contains("durationMillis=\""));

        assertEquals(exception, thrown);
    }
}
