/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.status;

import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Answers;
import org.mockito.Mock;

import java.security.cert.CertificateExpiredException;
import java.security.cert.CertificateNotYetValidException;
import java.security.cert.X509Certificate;
import java.time.Duration;
import java.time.Instant;
import java.util.Date;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.doThrow;
import static org.mockito.Mockito.when;

@CbsTest
class CertificateStatusCheckerTest {

    @Mock(answer = Answers.RETURNS_DEEP_STUBS)
    private X509Certificate certificate;
    private CertificateStatusChecker certificateStatusChecker;

    @BeforeEach
    void setUp() {
        certificateStatusChecker = new CertificateStatusChecker(certificate, "CERT");
    }

    @Test
    void is_ok_if_cert_is_valid() {
        when(certificate.getNotBefore()).thenReturn(new Date(Instant.now().minus(Duration.ofDays(10L)).toEpochMilli()));
        when(certificate.getNotAfter()).thenReturn(new Date(Instant.now().plus(Duration.ofDays(40L)).toEpochMilli()));
        when(certificate.getPublicKey().getAlgorithm()).thenReturn("EC");
        when(certificate.getSubjectDN().getName()).thenReturn("test-cert");

        final StatusCheck statusCheck = certificateStatusChecker.checkStatus();

        assertEquals(StatusCheck.Status.OK, statusCheck.getStatus());
    }

    @Test
    void is_warn_if_cert_expires_soon() {
        when(certificate.getNotBefore()).thenReturn(new Date(Instant.now().minus(Duration.ofDays(10L)).toEpochMilli()));
        when(certificate.getNotAfter()).thenReturn(new Date(Instant.now().plus(Duration.ofDays(29L)).toEpochMilli()));
        when(certificate.getPublicKey().getAlgorithm()).thenReturn("EC");
        when(certificate.getSubjectDN().getName()).thenReturn("test-cert");

        final StatusCheck statusCheck = certificateStatusChecker.checkStatus();

        assertEquals(StatusCheck.Status.WARN, statusCheck.getStatus());
    }

    @Test
    void is_error_if_cert_expired() throws CertificateNotYetValidException, CertificateExpiredException {
        when(certificate.getNotBefore()).thenReturn(new Date(Instant.now().minus(Duration.ofDays(10L)).toEpochMilli()));
        when(certificate.getNotAfter()).thenReturn(new Date(Instant.now().plus(Duration.ofDays(29L)).toEpochMilli()));
        when(certificate.getPublicKey().getAlgorithm()).thenReturn("EC");
        when(certificate.getSubjectDN().getName()).thenReturn("test-cert");

        doThrow(new CertificateExpiredException()).when(certificate).checkValidity();

        final StatusCheck statusCheck = certificateStatusChecker.checkStatus();

        assertEquals(StatusCheck.Status.ERROR, statusCheck.getStatus());
    }

    @Test
    void is_error_if_cert_not_yet_valid() throws CertificateNotYetValidException, CertificateExpiredException {
        when(certificate.getNotBefore()).thenReturn(new Date(Instant.now().minus(Duration.ofDays(10L)).toEpochMilli()));
        when(certificate.getNotAfter()).thenReturn(new Date(Instant.now().plus(Duration.ofDays(29L)).toEpochMilli()));
        when(certificate.getPublicKey().getAlgorithm()).thenReturn("EC");
        when(certificate.getSubjectDN().getName()).thenReturn("test-cert");

        doThrow(new CertificateNotYetValidException()).when(certificate).checkValidity();

        final StatusCheck statusCheck = certificateStatusChecker.checkStatus();

        assertEquals(StatusCheck.Status.ERROR, statusCheck.getStatus());
    }


}
