/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datacache.invalidator.impl;

import dk.sundhedsdatastyrelsen.cbs.datacache.invalidator.CacheInvalidationRecord;
import dk.sundhedsdatastyrelsen.cbs.datacache.invalidator.RecordInvalidator;
import dk.sundhedsdatastyrelsen.cbs.datacache.invalidator.exceptions.DataCacheInvalidationException;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.NasNotificationProcessor;
import org.oasis_open.docs.wsn.b_2.NotificationMessageHolderType;

import java.util.List;
import java.util.Objects;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Function;
import java.util.function.Predicate;

public class CacheInvalidatingNasNotificationProcessor implements NasNotificationProcessor {

    private final RecordInvalidator recordInvalidator;
    private final RecordMapper recordMapper;

    public CacheInvalidatingNasNotificationProcessor(RecordInvalidator recordInvalidator, RecordMapper recordMapper) {
        this.recordInvalidator = recordInvalidator;
        this.recordMapper = recordMapper;
    }

    @Override
    public void process(List<NotificationMessageHolderType> messages) {
        messages.stream()
            .map(recordMapper::toRecord)
            .filter(Objects::nonNull)
            .filter(distinctByKey(CacheInvalidationRecord::getCpr))
            .forEach(this::invalidateRecord);
    }

    private void invalidateRecord(CacheInvalidationRecord invalidationRecord) {
        boolean invalidated = recordInvalidator.invalidate(invalidationRecord);

        if (!invalidated) {
            handleRecordNotInvalidated();
        }
    }

    private void handleRecordNotInvalidated() {
        throw new DataCacheInvalidationException("Could not invalidate record.");
    }

    private static <T> Predicate<T> distinctByKey(Function<? super T, ?> keyExtractor) {
        Set<Object> seen = ConcurrentHashMap.newKeySet();
        return t -> seen.add(keyExtractor.apply(t));
    }

}
