/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datacache.invalidator.impl;

import dk.sundhedsdatastyrelsen.cbs.dao.NasLock;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.NasLockImpl;
import dk.sundhedsdatastyrelsen.cbs.datacache.Reader;
import dk.sundhedsdatastyrelsen.cbs.datacache.WriterWithDelete;
import org.apache.log4j.Logger;
import org.springframework.dao.EmptyResultDataAccessException;

import java.time.Clock;
import java.time.LocalDateTime;

public class LockProviderImpl implements LockProvider {

    private static final Logger log = Logger.getLogger(LockProviderImpl.class);

    private final Reader<NasLock, String> reader;
    private final WriterWithDelete<NasLock, String> writer;
    private final long lockForAtLeastMinutes;
    private final Clock clock;

    public LockProviderImpl(Reader<NasLock, String> reader, WriterWithDelete<NasLock, String> writer, long lockForAtLeastMinutes, Clock clock) {
        this.reader = reader;
        this.writer = writer;
        this.lockForAtLeastMinutes = lockForAtLeastMinutes;
        this.clock = clock;
    }

    @Override
    public boolean acquireLock(String topic) {
        log.debug("Acquiring lock");

        NasLock lock = read(topic);

        if (lock != null) {
            log.debug("Lock present");

            if (isLockExpired(lock)) {
                log.debug("Lock expired");
                releaseLock(topic);
            } else {
                log.debug("Lock valid");
                return false;
            }
        }

        log.debug("Creating lock");

        return writer.write(new NasLockImpl().topic(topic).lockedUntil(now().plusMinutes(lockForAtLeastMinutes)));
    }

    private NasLock read(String topic) {
        try {
            return reader.search(topic);
        } catch (EmptyResultDataAccessException e) {
            return null;
        }
    }

    private LocalDateTime now() {
        return LocalDateTime.now(clock);
    }

    private boolean isLockExpired(NasLock lock) {
        return now().isAfter(lock.lockedUntil());
    }

    @Override
    public boolean releaseLock(String topic) {
        log.debug("Releasing Lock");
        return writer.delete(topic);
    }
}
