/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datacache.invalidator.servlet;

import dk.sundhedsdatastyrelsen.cbs.CbsLogging;
import dk.sundhedsdatastyrelsen.cbs.datacache.invalidator.impl.LockProvider;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.NasConsumer;
import org.apache.log4j.Logger;
import org.springframework.context.ApplicationContext;

import javax.servlet.annotation.WebServlet;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.util.UUID;

import static dk.sundhedsdatastyrelsen.cbs.CbsLogging.TRACE_ID_PROP;
import static dk.sundhedsdatastyrelsen.cbs.datacache.invalidator.DataCacheInvalidationConfiguration.LAB_RESULT_NAS_CONSUMER_BEAN;
import static dk.sundhedsdatastyrelsen.cbs.datacache.invalidator.DataCacheInvalidationConfiguration.VACCINATION_NAS_CONSUMER_BEAN;
import static dk.sundhedsdatastyrelsen.cbs.datacache.invalidator.servlet.DataCacheInvalidationServletContextListener.SPRING_APPLICATION_CONTEXT;

@WebServlet("/trigger")
public class DataCacheInvalidationServlet extends HttpServlet {
    private static final Logger log = Logger.getLogger(DataCacheInvalidationServlet.class);

    @Override
    protected void doGet(HttpServletRequest request, HttpServletResponse response) {
        CbsLogging.withMDC(TRACE_ID_PROP, UUID.randomUUID().toString(), () -> doTrigger(request, response));
    }

    private void doTrigger(HttpServletRequest request, HttpServletResponse response) {
        log.debug("/trigger called");
        ApplicationContext applicationContext = (ApplicationContext) request.getServletContext().getAttribute(SPRING_APPLICATION_CONTEXT);
        NasConsumer vaccinationMessageConsumer = applicationContext.getBean(VACCINATION_NAS_CONSUMER_BEAN, NasConsumer.class);
        NasConsumer labResultMessageConsumer = applicationContext.getBean(LAB_RESULT_NAS_CONSUMER_BEAN, NasConsumer.class);
        LockProvider lockProvider = applicationContext.getBean(LockProvider.class);

        try {
            boolean lockAcquired = lockProvider.acquireLock("trigger");
            if (!lockAcquired) {
                log.warn("Could not acquire lock. Will not run job");
                sendResponse(423, response, "Job Locked");
                return;
            }
        } catch (Exception e) {
            log.error("Exception occurred trying to acquire lock", e);
            send500Response(response, e);
            return;
        }

        try {
            vaccinationMessageConsumer.trigger();
            labResultMessageConsumer.trigger();
        } catch (Exception e) {
            log.error("Exception occurred during invalidation", e);
            send500Response(response, e);
            return;
        } finally {
            lockProvider.releaseLock("trigger");
        }

        int status = 200;
        sendResponse(status, response);
    }

    private void send500Response(HttpServletResponse response, Exception e) {
        response.setContentType("text/plain; charset=utf-8");
        response.setStatus(500);
        try {
            response.getWriter().write("Exception occurred during invalidation" + e.getMessage());
        } catch (IOException ioException) {
            log.warn("Could not send response back to client", e);
        }
    }

    private void sendResponse(int status, HttpServletResponse response) {
        sendResponse(status, response, "Job Completed");
    }

    private void sendResponse(int status, HttpServletResponse response, String payload) {
        response.setContentType("text/plain; charset=utf-8");
        response.setStatus(status);
        try {
            response.getWriter().write(payload);
        } catch (IOException e) {
            log.warn("Could not send response back to client", e);
        }
    }
}
