/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datacache.invalidator;

import dk.sundhedsdatastyrelsen.cbs.dao.*;
import dk.sundhedsdatastyrelsen.cbs.datacache.Reader;
import dk.sundhedsdatastyrelsen.cbs.datacache.Writer;
import dk.sundhedsdatastyrelsen.cbs.datacache.WriterWithDelete;
import dk.sundhedsdatastyrelsen.cbs.datacache.status.JdbcStatusChecker;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import org.apache.log4j.Logger;
import org.mockito.Answers;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Primary;
import org.springframework.context.annotation.PropertySource;
import org.springframework.lang.NonNull;

import javax.sql.DataSource;
import java.util.HashMap;
import java.util.Map;
import java.util.function.Function;

import static org.mockito.Mockito.mock;

@Configuration
@PropertySource("classpath:application-datacache-invalidator.properties")
public class TestConfig {

    private static final Logger log = Logger.getLogger("TestWriter");

    private final SimpleReaderWriter<String, NasLock> nasLockSimpleReaderWriter = new SimpleReaderWriter<>(NasLock::topic);
    private final SimpleReaderWriter<String, NasStatus> nasStatusSimpleReaderWriter = new SimpleReaderWriter<>(NasStatus::topic);
    private final SimpleReaderWriter<String, NasSubscription> nasSubscriptionSimpleReaderWriter = new SimpleReaderWriter<>(NasSubscription::topic);

    @Bean
    @NonNull
    public Writer<VaccinationRecord<Vaccination>> vaccinationWriter() {
        return record -> {
            log.info("Writing VaccinationRecord " + record.traceId() + " :: " + record.invalidationTime());
            return true;
        };
    }

    @Bean
    @NonNull
    public Writer<LabResultRecord<LabResults>> labResultWriter() {
        return record -> {
            log.info("Writing LabResultRecord " + record.traceId() + " :: " + record.invalidationTime());
            return true;
        };
    }

    @Bean
    @NonNull
    public Reader<NasLock, String> nasLockReader() {
        return nasLockSimpleReaderWriter;
    }

    @Bean
    @NonNull
    public WriterWithDelete<NasLock, String> nasLockWriter() {
        return nasLockSimpleReaderWriter;
    }

    @Bean
    @NonNull
    public Reader<NasStatus, String> nasStatusReader() {
        return nasStatusSimpleReaderWriter;
    }

    @Bean
    @NonNull
    public WriterWithDelete<NasStatus, String> nasStatusWriter() {
        return nasStatusSimpleReaderWriter;
    }

    @Bean
    @NonNull
    public Reader<NasSubscription, String> nasSubscriptionReader() {
        return nasSubscriptionSimpleReaderWriter;
    }

    @Bean
    @NonNull
    public Writer<NasSubscription> nasSubscriptionWriter() {
        return nasSubscriptionSimpleReaderWriter;
    }

    @Bean
    @Primary
    public DataSource dataSource() {
        return mock(DataSource.class, Answers.RETURNS_DEEP_STUBS);
    }

    @Bean
    JdbcStatusChecker cbsJdbcStatusChecker() {
        return mock(JdbcStatusChecker.class);
    }

    private static class SimpleReaderWriter<K, T> implements Reader<T, K>, WriterWithDelete<T, K> {

        private final Map<K, T> map = new HashMap<>();
        private final Function<T, K> keyFunction;

        private SimpleReaderWriter(Function<T, K> keyFunction) {
            this.keyFunction = keyFunction;
        }

        @Override
        public T search(K key) {
            return map.get(key);
        }

        @Override
        public boolean write(T obj) {
            map.put(keyFunction.apply(obj), obj);

            return true;
        }

        @Override
        public boolean delete(K key) {
            return map.remove(key) != null;
        }
    }
}
