/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datacache.invalidator.impl;

import dk.sundhedsdatastyrelsen.cbs.dao.NasLock;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.NasLockImpl;
import dk.sundhedsdatastyrelsen.cbs.datacache.Reader;
import dk.sundhedsdatastyrelsen.cbs.datacache.WriterWithDelete;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Mock;

import java.time.Clock;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;

import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.*;

@CbsTest
class LockProviderImplTest {

    @Mock
    private Reader<NasLock, String> reader;

    @Mock
    private WriterWithDelete<NasLock, String> writer;

    @Captor
    private ArgumentCaptor<NasLock> nasLockArgumentCaptor;

    private LockProviderImpl lockProvider;
    private Clock clock;

    @BeforeEach
    void setUp() {
        clock = Clock.fixed(Instant.now(), ZoneId.systemDefault());
        lockProvider = new LockProviderImpl(reader, writer, 5L, clock);
    }

    @Test
    void should_acquire_and_persist_lock_if_none_is_present() {
        when(writer.write(any())).thenReturn(true);

        boolean myLock = lockProvider.acquireLock("myLock");
        assertTrue(myLock);

        verify(writer).write(nasLockArgumentCaptor.capture());

        NasLock lock = nasLockArgumentCaptor.getValue();

        assertTrue(now().isBefore(lock.lockedUntil()));
    }

    @Test
    void should_not_acquire_lock_if_present_is_valid() {
        when(reader.search(any())).thenReturn(new NasLockImpl().topic("mylock").lockedUntil(now().plusMinutes(5L)));

        boolean myLock = lockProvider.acquireLock("myLock");
        assertFalse(myLock);

        verifyNoInteractions(writer);
    }

    @Test
    void should_acquire_lock_if_present_is_expired() {
        when(reader.search(any())).thenReturn(new NasLockImpl().topic("mylock").lockedUntil(now().minusSeconds(10L)));
        when(writer.write(any())).thenReturn(true);

        boolean myLock = lockProvider.acquireLock("myLock");
        assertTrue(myLock);

        verify(writer).delete("myLock");
        verify(writer).write(any());
    }

    @Test
    void should_delete_lock_on_release() {
        lockProvider.releaseLock("myLock");
        verify(writer).delete("myLock");
    }

    private LocalDateTime now() {
        return LocalDateTime.now(clock);
    }
}
