/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datacache.invalidator.impl;

import dk.sundhedsdatastyrelsen.cbs.dao.NasStatus;
import dk.sundhedsdatastyrelsen.cbs.dao.NasSubscription;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.NasStatusImpl;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.NasSubscriptionImpl;
import dk.sundhedsdatastyrelsen.cbs.datacache.Reader;
import dk.sundhedsdatastyrelsen.cbs.datacache.Writer;
import dk.sundhedsdatastyrelsen.cbs.datacache.WriterWithDelete;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.EndpointReferences;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.NasClient;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.springframework.dao.EmptyResultDataAccessException;

import javax.xml.ws.wsaddressing.W3CEndpointReference;
import java.util.UUID;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.*;

@CbsTest
class PersistedNasRepositoryTest {

    @Mock
    private NasClient nasClient;
    @Mock
    private Reader<NasSubscription, String> subscriptionReader;
    @Mock
    private Writer<NasSubscription> subscriptionWriter;
    @Mock
    private Reader<NasStatus, String> statusReader;
    @Mock
    private WriterWithDelete<NasStatus, String> statusWriter;

    private PersistedNasRepository persistedNasRepository;

    @BeforeEach
    void setUp() {
        persistedNasRepository = new PersistedNasRepository(nasClient, subscriptionReader, subscriptionWriter, statusReader, statusWriter);
    }

    @Test
    void should_create_and_persist_pull_point_if_none_is_present() {
        String pullPointAddress = "http://pullpoint.com";
        when(nasClient.createPullPoint()).thenReturn(EndpointReferences.newEndpointReference(pullPointAddress));
        when(nasClient.subscribe(any(), any())).thenReturn(EndpointReferences.newEndpointReference("http://subscribe.com"));

        W3CEndpointReference pullPoint = persistedNasRepository.getPullPoint("myTopic");

        verify(nasClient).createPullPoint();
        verify(nasClient).subscribe(any(), any());

        verify(subscriptionWriter).write(any());

        assertEquals(pullPointAddress, EndpointReferences.extractAddress(pullPoint));
    }

    @Test
    void should_create_and_persist_pull_point_if_search_throws() {
        when(subscriptionReader.search(any())).thenThrow(new EmptyResultDataAccessException(1));

        String pullPointAddress = "http://pullpoint.com";
        when(nasClient.createPullPoint()).thenReturn(EndpointReferences.newEndpointReference(pullPointAddress));
        when(nasClient.subscribe(any(), any())).thenReturn(EndpointReferences.newEndpointReference("http://subscribe.com"));

        W3CEndpointReference pullPoint = persistedNasRepository.getPullPoint("myTopic");

        verify(nasClient).createPullPoint();
        verify(nasClient).subscribe(any(), any());

        verify(subscriptionWriter).write(any());

        assertEquals(pullPointAddress, EndpointReferences.extractAddress(pullPoint));
    }

    @Test
    void should_return_pull_point_if_present() {
        String pullPointAddress = "http://pullpoint.com";

        when(subscriptionReader.search("myTopic")).thenReturn(new NasSubscriptionImpl()
            .topic("myTopic")
            .pullPoint(pullPointAddress)
            .subscription("http://subscribe.com")
        );

        W3CEndpointReference pullPoint = persistedNasRepository.getPullPoint("myTopic");

        verifyNoInteractions(nasClient);
        verifyNoInteractions(subscriptionWriter);

        assertEquals(pullPointAddress, EndpointReferences.extractAddress(pullPoint));
    }

    @Test
    void should_create_and_persist_message_id_if_none_is_present() {
        persistedNasRepository.getNextMessageId("myTopic");

        verify(statusWriter).write(any());
    }

    @Test
    void should_create_and_persist_message_id_if_search_throws() {
        when(statusReader.search(any())).thenThrow(new EmptyResultDataAccessException(1));

        persistedNasRepository.getNextMessageId("myTopic");

        verify(statusWriter).write(any());
    }

    @Test
    void should_return_message_id_if_present() {
        UUID messageId = UUID.randomUUID();
        when(statusReader.search("myTopic")).thenReturn(new NasStatusImpl().topic("myTopic").messageId(messageId));

        UUID nextMessageId = persistedNasRepository.getNextMessageId("myTopic");

        verifyNoInteractions(statusWriter);
        assertEquals(messageId, nextMessageId);
    }

    @Test
    void should_delete_status_on_commit() {
        persistedNasRepository.commit("myTopic");
        verify(statusWriter).delete("myTopic");
    }
}
