/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datacache;

import com.fasterxml.jackson.databind.ObjectMapper;
import dk.sundhedsdatastyrelsen.cbs.dao.ChildRelation;
import dk.sundhedsdatastyrelsen.cbs.dao.LabResultRecord;
import dk.sundhedsdatastyrelsen.cbs.dao.NasLock;
import dk.sundhedsdatastyrelsen.cbs.dao.NasStatus;
import dk.sundhedsdatastyrelsen.cbs.dao.NasSubscription;
import dk.sundhedsdatastyrelsen.cbs.dao.PersonRecord;
import dk.sundhedsdatastyrelsen.cbs.dao.Record;
import dk.sundhedsdatastyrelsen.cbs.dao.ServiceClient;
import dk.sundhedsdatastyrelsen.cbs.dao.VaccinationRecord;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.ChildRelationImpl;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.LabResultRecordImpl;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.NasLockImpl;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.NasStatusImpl;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.NasSubscriptionImpl;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.PersonRecordImpl;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.RecordMappers;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.ServiceClientImpl;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.VaccinationRecordImpl;
import dk.sundhedsdatastyrelsen.cbs.datacache.impl.Readers;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.fluent.spring.jdbc.KeyMapper;
import dk.sundhedsdatastyrelsen.fluent.spring.jdbc.SearchMapper;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Lazy;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.lang.NonNull;

import javax.sql.DataSource;
import java.io.IOException;
import java.util.Collection;
import java.util.function.Function;
import java.util.function.Supplier;

import static dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.RecordMappers.mapper;
import static dk.sundhedsdatastyrelsen.cbs.datacache.impl.Readers.reader;
import static dk.sundhedsdatastyrelsen.cbs.datacache.impl.Readers.recordReader;
import static dk.sundhedsdatastyrelsen.cbs.datacache.impl.Writers.writer;
import static dk.sundhedsdatastyrelsen.cbs.datacache.impl.Writers.writerWithDelete;

@Configuration
public class DataCacheImplConfiguration {

    @Bean
    @Lazy
    @NonNull
    public JdbcTemplate jdbcTemplate(@NonNull final DataSource dataSource) {
        return new JdbcTemplate(dataSource);
    }

    @Bean
    @Lazy
    @NonNull
    public Reader<Vaccination, String> vaccinationReader(
        @NonNull final JdbcTemplate jdbcTemplate,
        @NonNull final KeyMapper<VaccinationRecordImpl<Vaccination>, String> mapper) {
        return recordReader(jdbcTemplate, mapper);
    }

    @Bean
    @NonNull
    public Writer<VaccinationRecord<Vaccination>> vaccinationWriter(
        @NonNull final JdbcTemplate jdbcTemplate,
        @NonNull final KeyMapper<VaccinationRecordImpl<Vaccination>, String> mapper) {
        return writer(jdbcTemplate, mapper);
    }

    @Bean
    @Lazy
    @NonNull
    public KeyMapper<VaccinationRecordImpl<Vaccination>, String> vaccinationMapper(
        @NonNull final Supplier<VaccinationRecordImpl<Vaccination>> constructor,
        @NonNull final ObjectMapper mapper) {
        return mapper(
            Record.CATALOG_NAME,
            Record.SCHEMA_NAME,
            VaccinationRecord.INSERT_TABLE_NAME,
            Record.CATALOG_NAME,
            Record.SCHEMA_NAME,
            VaccinationRecord.SELECT_TABLE_NAME,
            constructor,
            deserializer(mapper, Vaccination.class),
            serializer(mapper))
            .build();
    }

    @Bean
    @Lazy
    @NonNull
    public Reader<LabResults, String> labResultReader(
        @NonNull final JdbcTemplate jdbcTemplate,
        @NonNull final KeyMapper<LabResultRecordImpl<LabResults>, String> mapper) {
        return recordReader(jdbcTemplate, mapper);
    }


    @Bean
    @NonNull
    public Writer<LabResultRecord<LabResults>> labResultWriter(
        @NonNull final JdbcTemplate jdbcTemplate,
        @NonNull final KeyMapper<LabResultRecordImpl<LabResults>, String> mapper) {
        return writer(jdbcTemplate, mapper);
    }

    @Bean
    @Lazy
    @NonNull
    public KeyMapper<LabResultRecordImpl<LabResults>, String> labResultMapper(
        @NonNull final Supplier<LabResultRecordImpl<LabResults>> constructor,
        @NonNull final ObjectMapper mapper) {
        return mapper(
            Record.CATALOG_NAME,
            Record.SCHEMA_NAME,
            LabResultRecord.INSERT_TABLE_NAME,
            Record.CATALOG_NAME,
            Record.SCHEMA_NAME,
            LabResultRecord.SELECT_TABLE_NAME,
            constructor,
            deserializer(mapper, LabResults.class),
            serializer(mapper))
            .build();
    }

    @Bean
    @Lazy
    @NonNull
    public Supplier<VaccinationRecordImpl<Vaccination>> vaccinationConstructor() {
        return VaccinationRecordImpl::new;
    }

    @Bean
    @Lazy
    @NonNull
    public Supplier<LabResultRecordImpl<LabResults>> labResultConstructor() {
        return LabResultRecordImpl::new;
    }


    @Bean
    @NonNull
    public KeyMapper<NasLockImpl, String> nasLockMapper() {
        return RecordMappers.nasLockMapper(
            Record.CATALOG_NAME,
            Record.SCHEMA_NAME,
            NasLock.INSERT_TABLE_NAME,
            Record.CATALOG_NAME,
            Record.SCHEMA_NAME,
            NasLock.SELECT_TABLE_NAME,
            NasLockImpl::new)
            .build();
    }

    @Bean
    @NonNull
    public Reader<NasLock, String> nasLockReader(
        @NonNull final JdbcTemplate jdbcTemplate,
        @NonNull final KeyMapper<NasLockImpl, String> mapper) {
        return reader(jdbcTemplate, mapper);
    }

    @Bean
    @NonNull
    public WriterWithDelete<NasLock, String> nasLockWriter(
        @NonNull final JdbcTemplate jdbcTemplate,
        @NonNull final KeyMapper<NasLockImpl, String> mapper) {
        return writerWithDelete(jdbcTemplate, mapper);
    }

    @Bean
    @NonNull
    public KeyMapper<NasStatusImpl, String> nasStatusMapper() {
        return RecordMappers.nasStatusMapper(
            Record.CATALOG_NAME,
            Record.SCHEMA_NAME,
            NasStatus.INSERT_TABLE_NAME,
            Record.CATALOG_NAME,
            Record.SCHEMA_NAME,
            NasStatus.SELECT_TABLE_NAME,
            NasStatusImpl::new)
            .build();
    }

    @Bean
    @NonNull
    public Reader<NasStatus, String> nasStatusReader(
        @NonNull final JdbcTemplate jdbcTemplate,
        @NonNull final KeyMapper<NasStatusImpl, String> mapper) {
        return reader(jdbcTemplate, mapper);
    }

    @Bean
    @NonNull
    public WriterWithDelete<NasStatus, String> nasStatusWriter(
        @NonNull final JdbcTemplate jdbcTemplate,
        @NonNull final KeyMapper<NasStatusImpl, String> mapper) {
        return writerWithDelete(jdbcTemplate, mapper);
    }

    @Bean
    @NonNull
    public KeyMapper<NasSubscriptionImpl, String> nasSubscriptionMapper() {
        return RecordMappers.nasSubscriptionMapper(
            Record.CATALOG_NAME,
            Record.SCHEMA_NAME,
            NasSubscription.INSERT_TABLE_NAME,
            Record.CATALOG_NAME,
            Record.SCHEMA_NAME,
            NasSubscription.SELECT_TABLE_NAME,
            NasSubscriptionImpl::new)
            .build();
    }

    @Bean
    @NonNull
    public Reader<NasSubscription, String> nasSubscriptionReader(
        @NonNull final JdbcTemplate jdbcTemplate,
        @NonNull final KeyMapper<NasSubscriptionImpl, String> mapper) {
        return reader(jdbcTemplate, mapper);
    }

    @Bean
    @NonNull
    public Writer<NasSubscription> nasSubscriptionWriter(
        @NonNull final JdbcTemplate jdbcTemplate,
        @NonNull final KeyMapper<NasSubscriptionImpl, String> mapper) {
        return writer(jdbcTemplate, mapper);
    }

    @Bean
    @NonNull
    public KeyMapper<PersonRecordImpl, String> personRecordMapper() {
        return RecordMappers.personMapper(
            Record.CATALOG_NAME,
            Record.SCHEMA_NAME,
            "NA",
            Record.CATALOG_NAME,
            Record.SCHEMA_NAME,
            PersonRecord.SELECT_TABLE_NAME,
            PersonRecordImpl::new)
            .build();
    }

    @Bean
    @NonNull
    public Reader<PersonRecord, String> personReader(
        @NonNull final JdbcTemplate jdbcTemplate,
        @NonNull final KeyMapper<PersonRecordImpl, String> mapper) {
        return reader(jdbcTemplate, mapper);
    }

    @Bean
    @NonNull
    public SearchMapper<ChildRelationImpl, String> childRelationSearchMapper() {
        return RecordMappers.childRelationMapper(
            Record.CATALOG_NAME,
            Record.SCHEMA_NAME,
            "NA",
            Record.CATALOG_NAME,
            Record.SCHEMA_NAME,
            ChildRelation.SELECT_TABLE_NAME,
            ChildRelationImpl::new)
            .build();
    }

    @Bean
    @NonNull
    public CollectionReader<ChildRelation, String> childRelationReader(
        @NonNull final JdbcTemplate jdbcTemplate,
        @NonNull final SearchMapper<ChildRelationImpl, String> mapper
    ) {
        return Readers.reader(jdbcTemplate, mapper);
    }

    @Bean
    @NonNull
    public CollectionReader<ServiceClient, String> serviceClientReader(JdbcTemplate jdbcTemplate) {
        return Readers.reader(
            jdbcTemplate,
            RecordMappers.serviceClientMapper(
                Record.CATALOG_NAME,
                Record.SCHEMA_NAME,
                "NA",
                Record.CATALOG_NAME,
                Record.SCHEMA_NAME,
                ServiceClient.SELECT_TABLE_NAME,
                ServiceClientImpl::new).build()
        );
    }


    private static <T> Function<java.io.Reader, T> deserializer(@NonNull final ObjectMapper mapper, @NonNull final Class<T> clazz) {
        return reader -> {
            try {
                return reader == null ? null : mapper.readValue(reader, clazz);
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        };
    }

    private static <T> Function<T, String> serializer(@NonNull final ObjectMapper mapper) {
        return object -> {
            try {
                return object == null ? null : mapper.writeValueAsString(object);
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        };
    }

}
