/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datacache.impl;

import dk.sundhedsdatastyrelsen.cbs.dao.Record;
import dk.sundhedsdatastyrelsen.cbs.datacache.CollectionReader;
import dk.sundhedsdatastyrelsen.cbs.datacache.Reader;
import dk.sundhedsdatastyrelsen.fluent.spring.jdbc.KeyMapper;
import dk.sundhedsdatastyrelsen.fluent.spring.jdbc.SearchMapper;
import org.springframework.jdbc.core.ArgumentTypePreparedStatementSetter;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import java.util.Collection;

import static java.lang.String.format;

public final class Readers {

    @NonNull
    private static final String SEARCH_SQL = "select * from %s where %s = ?";

    private Readers() {
    }

    public static <T, K> Reader<T, K> reader(@NonNull final JdbcTemplate jdbcTemplate, @NonNull final KeyMapper<? extends T, K> mapper) {
        return new KeyReaderImpl<>(jdbcTemplate, mapper);
    }

    public static <T, K> CollectionReader<T, K> reader(@NonNull final JdbcTemplate jdbcTemplate, @NonNull final SearchMapper<? extends T, K> mapper) {
        return new SearchReaderImpl<>(jdbcTemplate, mapper);
    }

    public static <T, K> Reader<T, K> recordReader(@NonNull final JdbcTemplate jdbcTemplate, @NonNull final KeyMapper<? extends Record<T>, K> mapper) {
        return new RecordReaderImpl<>(jdbcTemplate, mapper);
    }

    private static final class SearchReaderImpl<T, S> implements CollectionReader<T, S> {

        @NonNull private final SearchMapper<T, S> mapper;
        @NonNull private final JdbcTemplate template;
        @NonNull private final String query;
        private final int[] sqlType;

        @SuppressWarnings("unchecked")
        private SearchReaderImpl(
            @NonNull final JdbcTemplate jdbcTemplate,
            @NonNull final SearchMapper<? extends T, S> mapper) {
            template = jdbcTemplate;
            query = format(SEARCH_SQL, mapper.selectTableMetaData().tableName(), mapper.searchColumnLabel());
            sqlType = new int[] { mapper.sqlType() };
            this.mapper = (SearchMapper<T, S>) mapper;
        }

        @Override
        @NonNull
        public Collection<T> search(@NonNull S value) {
            final Object[] serializedValue = new Object[] { mapper.serializeSearchValue(value) };
            final ArgumentTypePreparedStatementSetter setter = new ArgumentTypePreparedStatementSetter(serializedValue, sqlType);
            return template.query(query, setter, mapper);
        }

    }

    private static final class KeyReaderImpl<T, K> implements Reader<T, K> {
        @NonNull private final KeyMapper<? extends T, K> mapper;
        @NonNull private final JdbcTemplate template;
        @NonNull private final String query;
        private final int[] sqlType;

        private KeyReaderImpl(
            @NonNull final JdbcTemplate jdbcTemplate,
            @NonNull final KeyMapper<? extends T, K> mapper) {
            template = jdbcTemplate;
            query = format(SEARCH_SQL, mapper.selectTableMetaData().tableName(), mapper.keyColumnLabel());
            sqlType = new int[] { mapper.sqlType() };
            this.mapper = mapper;
        }

        @Override
        @Nullable
        public T search(@NonNull K key) {
            final Object[] serializedKey = new Object[] { mapper.serializeKey(key) };
            final ArgumentTypePreparedStatementSetter setter = new ArgumentTypePreparedStatementSetter(serializedKey, sqlType);
            return template.query(query, setter, mapper).stream().findFirst().orElse(null);
        }
    }

    private static final class RecordReaderImpl<T, K> implements Reader<T, K> {

        @NonNull private final JdbcTemplate template;
        @NonNull private final KeyMapper<? extends Record<T>, K> mapper;
        @NonNull private final String query;
        private final int[] sqlType;

        private RecordReaderImpl(@NonNull final JdbcTemplate jdbcTemplate, @NonNull final KeyMapper<? extends Record<T>, K> mapper) {
            template = jdbcTemplate;
            query = format(SEARCH_SQL, mapper.selectTableMetaData().tableName(), mapper.keyColumnLabel());
            sqlType = new int[] { mapper.sqlType() };
            this.mapper = mapper;
        }

        @Override
        public T search(@NonNull final K key) {
            final Object[] serializedKey = new Object[] { mapper.serializeKey(key) };
            final ArgumentTypePreparedStatementSetter setter = new ArgumentTypePreparedStatementSetter(serializedKey, sqlType);
            return template.query(query, setter, mapper).stream().findFirst()
                .map(Record::data)
                .orElse(null);
        }
    }
}
