/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datacache.impl;

import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.MyRecord;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.RecordMappers;
import dk.sundhedsdatastyrelsen.cbs.datacache.CollectionReader;
import dk.sundhedsdatastyrelsen.cbs.datacache.Reader;
import dk.sundhedsdatastyrelsen.fluent.spring.jdbc.KeyMapper;
import dk.sundhedsdatastyrelsen.fluent.spring.jdbc.SearchMapper;
import dk.sundhedsdatastyrelsen.fluent.spring.jdbc.TableMetaData;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import org.mockito.junit.jupiter.MockitoSettings;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.PreparedStatementSetter;
import org.springframework.jdbc.core.RowMapper;
import org.springframework.lang.NonNull;

import java.util.Collection;
import java.util.Collections;
import java.util.function.Function;

import static dk.sundhedsdatastyrelsen.fluent.spring.jdbc.Mappers.mapper;
import static dk.sundhedsdatastyrelsen.fluent.spring.jdbc.TableMetaDataBuilder.tableMetaData;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertSame;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.when;
import static org.mockito.quality.Strictness.LENIENT;

@ExtendWith(MockitoExtension.class)
@MockitoSettings(strictness = LENIENT)
class ReadersTest {

    @NonNull private static final String CATALOG_NAME = "catalog";
    @NonNull private static final String SCHEMA_NAME = "schema";
    @NonNull private static final String TABLE_NAME = "table";
    @NonNull private static final TableMetaData TABLE_META_DATA = tableMetaData(TABLE_NAME)
        .schemaName(SCHEMA_NAME)
        .catalogName(CATALOG_NAME)
        .build();
    @NonNull private static final String KEY_COLUMN_LABEL = "key";
    @NonNull private static final String DATA = "data";

    @NonNull private final JdbcTemplate template;
    @NonNull private final MyRecord record;

    @SuppressWarnings("unchecked")
    ReadersTest(@Mock final JdbcTemplate template) {
        this.template = template;
        record = new MyRecord().setData(DATA);
        when(template.query(anyString(), any(PreparedStatementSetter.class), any(RowMapper.class)))
            .thenReturn(Collections.singletonList(record));
    }

    @Test
    void keyReader() {
        final KeyMapper<MyRecord, String> mapper = mapper(
            String.class,
            TABLE_META_DATA,
            TABLE_META_DATA,
            MyRecord::new)
            .addKeyStringMapping(KEY_COLUMN_LABEL, MyRecord::key, MyRecord::key, Function.identity())
            .build();

        final Reader<MyRecord, String> reader = Readers.reader(template, mapper);
        assertSame(record, reader.search(""));
    }

    @Test
    void searchReader() {
        final SearchMapper<MyRecord, String> mapper = mapper(
            String.class,
            TABLE_META_DATA,
            TABLE_META_DATA,
            MyRecord::new)
            .addSearchStringMapping(KEY_COLUMN_LABEL, MyRecord::key, MyRecord::key, Function.identity())
            .build();

        final CollectionReader<MyRecord, String> reader = Readers.reader(template, mapper);
        final Collection<MyRecord> records = reader.search("");
        assertNotNull(records);
        assertTrue(records.contains(record));

    }

    @Test
    void recordReader() {
        final KeyMapper<MyRecord, String> mapper = RecordMappers.mapper(
            "catalog",
            "schema",
            "table",
            "catalog",
            "schema",
            "table",
            MyRecord::new,
            String::valueOf, String::toString)
            .build();
        final Reader<String, String> reader = Readers.recordReader(template, mapper);
        assertSame(DATA, reader.search(""));
    }


}
