/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datafetcher;

import dk.sundhedsdatastyrelsen.cbs.datafetcher.config.DataFetcherConfigProperties;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.exceptions.DataFetcherException;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.model.DataFetcherJobExecutionResult;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.model.DataFetcherJobExecutionResultBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.kafka.DataFetcherJob;
import dk.sundhedsdatastyrelsen.cbs.dto.kafka.FetchType;
import dk.sundhedsdatastyrelsen.cbs.kafka.pooling.CbsKafkaConsumerPool;
import dk.sundhedsdatastyrelsen.cbs.monitoring.IsAliveServlet;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.NasPublisher;
import dk.sundhedsdatastyrelsen.coronapass.ws.JobStatus;
import dk.sundhedsdatastyrelsen.coronapass.ws.JobStatusNotification;
import org.apache.commons.lang3.NotImplementedException;
import org.apache.log4j.Logger;
import org.apache.log4j.MDC;
import org.springframework.stereotype.Service;

import java.io.Closeable;
import java.time.Instant;
import java.util.*;
import java.util.concurrent.*;
import java.util.stream.Collectors;

import static dk.sundhedsdatastyrelsen.cbs.CbsLogging.*;
import static dk.sundhedsdatastyrelsen.cbs.datafetcher.DataFetcherConstants.NAS_ERROR_INTERNAL_SERVER_ERROR;

/**
 * Entry point for interacting with the datafetcher inside a Spring ApplicationContext
 */
@Service
public class DataFetcher implements Closeable {
    private static final Logger log = Logger.getLogger(DataFetcher.class);
    private static final String CLOSING_SERVER_JOB_FAILURE_MSG = "INTERNAL_CBS_ERROR: SERVER_SHUTDOWN";
    public static final String VACCINATION_ETL = "Vaccination ETL";
    public static final String LAB_RESULT_ETL = "LabResult ETL";
    private final CbsKafkaConsumerPool kafkaConsumerPool;
    private final EtlFunc vaccinationEtl;
    private final EtlFunc labResEtl;
    private final int processingEndTimeoutMillis;
    private final int processingBeginTimeoutMillis;
    private final long shutdownTimeoutMillis;
    private final Semaphore processingPermits;
    private Instant shutdownTime = null;
    private final ExecutorService executorService;
    private final NasPublisher nasPublisher;

    public DataFetcher(CbsKafkaConsumerPool kafkaConsumerPool,
                       ExecutorService executorService,
                       @VaccinationETL EtlFunc vaccinationEtl,
                       @LabResultETL EtlFunc labResEtl,
                       NasPublisher nasPublisher,
                       DataFetcherConfigProperties dataFetcherConfigProperties) {
        this.executorService = executorService;
        this.nasPublisher = nasPublisher;
        IsAliveServlet.setServiceName(DataFetcherConstants.COMPONENT_NAME);
        this.kafkaConsumerPool = kafkaConsumerPool;
        this.vaccinationEtl = vaccinationEtl;
        this.labResEtl = labResEtl;
        this.processingBeginTimeoutMillis = dataFetcherConfigProperties.getProcessingBeginTimeoutMillis();
        this.processingEndTimeoutMillis = dataFetcherConfigProperties.getProcessingEndTimeoutMillis();
        this.shutdownTimeoutMillis = dataFetcherConfigProperties.getShutdownTimeoutMillis();
        processingPermits = new Semaphore(dataFetcherConfigProperties.getMaxConcurrentProcessings());
    }

    /**
     * 1) Acquire a Kafka consumer from the consumer pool (blocking for x milliseconds if no consumer is available, throwing an exception on timeout). <br/>
     * 2) Use consumer to poll Kafka for up to x number of DataFetcherJobs (blocking for x milliseconds if no jobs available, returning 0 jobs on timeout) <br/>
     * 3) Use the provided ExecutorService to process the jobs concurrently <br/>
     * -- Each job fetches from DDV and/or MIBA, stores result in the database and sends NAS notification. <br/>
     * -- Any exception occurring during processing of an individual job causes the job to be marked as failed and a job-failure notification is sent on NAS <br/>
     */
    public List<DataFetcherJobExecutionResult> processJobs() throws Exception {
        if (shutdownTime != null) {
            log.warn("Not processing new tasks - server is shutting down");
            return Collections.emptyList();
        }
        try {
            if (!processingPermits.tryAcquire(processingBeginTimeoutMillis, TimeUnit.MILLISECONDS)) {
                log.info("No processor permit available within time limit of: " + processingBeginTimeoutMillis + " milliseconds. Returning empty result set to indicate nothing has been processed.");
                return Collections.emptyList();
            }
            return doProcessing();
        } finally {
            processingPermits.release();
        }
    }

    private List<DataFetcherJobExecutionResult> doProcessing() throws Exception {
        log.debug("Polling Kafka for jobs");
        final Set<DataFetcherJob> jobs = kafkaConsumerPool.pollDataFetcherJobs();
        if (jobs.isEmpty()) {
            log.info("No jobs found");
            return Collections.emptyList();
        }

        log.debug(String.format("Starting parallel processing of %s jobs", jobs.size()));

        Map<CompletableFuture<DataFetcherJobExecutionResult>, DataFetcherJob> jobTrackingMap = new HashMap<>();
        List<CompletableFuture<DataFetcherJobExecutionResult>> executions = jobs.stream().map(job -> {
            CompletableFuture<DataFetcherJobExecutionResult> future = startExecution(job, executorService);
            jobTrackingMap.put(future, job);
            return future;
        }).collect(Collectors.toList());

        try {
            CompletableFuture.allOf(executions.toArray(new CompletableFuture[0])).get(processingEndTimeoutMillis, TimeUnit.MILLISECONDS);
        } catch (TimeoutException timeoutException) {
            log.warn("Processing did not finish in time! Cancelling currently running job executions");
            executions.forEach(e -> e.cancel(true)); //cancelling causes the task to complete exceptionally
        } catch (ExecutionException | CancellationException exception) {
            if (shutdownTime != null && exception.getCause() instanceof InterruptedException) {
                throw new DataFetcherException("Processing interrupted due to server shutdown. Halting immediately. NAS messages are NOT sent!");
            }
        }
        handleExceptions(jobTrackingMap, executions);

        List<DataFetcherJobExecutionResult> results = getExecutionResults(jobTrackingMap, executions);
        int resultsSize = results.size();
        if (jobs.size() != resultsSize)
            log.error(String.format("INCONSISTENCY DETECTED: got %s jobs from Kafka, but returned %s execution results!", jobs.size(), resultsSize));

        long failed = results.stream().filter(DataFetcherJobExecutionResult::isFailed).count();
        log.info(String.format("PROCESSED=%s | SUCCESS=%s | FAILED=%s", resultsSize, resultsSize - failed, failed));
        return results;
    }

    private void handleExceptions(Map<CompletableFuture<DataFetcherJobExecutionResult>, DataFetcherJob> jobTrackingMap, List<CompletableFuture<DataFetcherJobExecutionResult>> executions) {
        List<CompletableFuture<DataFetcherJobExecutionResult>> exceptionalExecutions = executions.stream().filter(CompletableFuture::isCompletedExceptionally).collect(Collectors.toList());
        if (!exceptionalExecutions.isEmpty()) {
            log.error(String.format("%s jobs failed with unhandled exceptions! Will send failure notification for each job to NAS", exceptionalExecutions.size())); //note: If the job was interrupted after sending to NAS (UNLIKELY), we will have sent 2 notifications to NAS
            for (CompletableFuture<DataFetcherJobExecutionResult> exceptionalExecution : exceptionalExecutions) {
                //note: These NAS notifications are done sequentially and synchronously on this thread! (that's ok, no need to optimize for abnormal operation)
                exceptionalExecution.handle((nullRes, throwable) -> {
                    final Throwable failure = throwable instanceof CompletionException ? throwable.getCause() : throwable;
                    DataFetcherJob exceptionalJob = jobTrackingMap.get(exceptionalExecution);

                    return withMDC(JOB_ID_PROP, exceptionalJob.getJobId(),
                        supplyWithMDC(MESSAGE_ID_PROP, exceptionalJob.getMsgId(),
                            () -> {
                                if (failure instanceof CancellationException) {
                                    log.warn(String.format("Sending job failure to NAS due to timeout for Job='%s'", exceptionalJob)); //Ops do not want stacktraces for job timeouts
                                } else {
                                    log.warn(String.format("Sending failure to NAS for Job='%s'", exceptionalJob), failure);
                                }
                                DataFetcherJobExecutionResult resultAfterSending = sendNasFailureNotification(exceptionalJob, failure);
                                exceptionalExecution.obtrudeValue(resultAfterSending); //complete the job with the new result (that has been handled)
                                return null; //not used
                            }));
                });
            }
        }
    }

    private List<DataFetcherJobExecutionResult> getExecutionResults(Map<CompletableFuture<DataFetcherJobExecutionResult>, DataFetcherJob> jobTrackingMap, List<CompletableFuture<DataFetcherJobExecutionResult>> executions) {
        return executions.stream().map(e -> {
            try {
                DataFetcherJobExecutionResult valueIfAbsent = new DataFetcherJobExecutionResult(jobTrackingMap.get(e), true, "Absent execution result");
                return e.getNow(valueIfAbsent);
            } catch (Exception exception) {
                return new DataFetcherJobExecutionResult(jobTrackingMap.get(e), true, exception.toString());
            }
        }).collect(Collectors.toList());
    }

    private CompletableFuture<DataFetcherJobExecutionResult> startExecution(DataFetcherJob dataFetcherJob, ExecutorService executorService) {
        try {
            //this is the servlet thread
            String traceId = (String) MDC.get(TRACE_ID_PROP);
            MDC.put(MESSAGE_ID_PROP, dataFetcherJob.getMsgId());
            MDC.put(JOB_ID_PROP, dataFetcherJob.getJobId());
            if (shutdownTime != null && shutdownTime.plusMillis(shutdownTimeoutMillis).isBefore(Instant.now())) {
                return CompletableFuture.completedFuture(new DataFetcherJobExecutionResult(dataFetcherJob, true, CLOSING_SERVER_JOB_FAILURE_MSG));
            }

            if (log.isTraceEnabled())
                log.trace("Starting processing job: " + dataFetcherJob);
            if (dataFetcherJob.getFetchCprs().isEmpty()) {
                log.warn("Nothing to fetch! No 'fetchCprs' specified in " + dataFetcherJob);
                return CompletableFuture.completedFuture(sendNasNotification(new DataFetcherJobExecutionResult(dataFetcherJob)));
            }

            return CompletableFuture.supplyAsync(
                //this is a worker thread
                supplyWithMDCs(TRACE_ID_PROP, traceId, MESSAGE_ID_PROP, dataFetcherJob.getMsgId(), JOB_ID_PROP, dataFetcherJob.getJobId(),
                    () -> {
                        final DataFetcherJobExecutionResultBuilder result = DataFetcherJobExecutionResultBuilder.aDataFetcherJobExecutionResult().withDataFetcherJob(dataFetcherJob);
                        //fetch data for each CPR in the job sequentially, in a single thread (to not use up too many threads in case there are many fetches in this job)
                        for (Map.Entry<String, FetchType> fetchCpr : dataFetcherJob.getFetchCprs().entrySet()) {
                            final String cpr = fetchCpr.getKey();
                            final FetchType fetchType = fetchCpr.getValue();
                            switch (fetchType) {
                                case FETCH_VACCINATION_STATUS:
                                    result.append(performEtl(dataFetcherJob, cpr, vaccinationEtl, VACCINATION_ETL));
                                    break;
                                case FETCH_TEST_STATUS:
                                    result.append(performEtl(dataFetcherJob, cpr, labResEtl, LAB_RESULT_ETL));
                                    break;
                                case FETCH_ALL:
                                    result.append(performEtl(dataFetcherJob, cpr, vaccinationEtl, VACCINATION_ETL));
                                    result.append(performEtl(dataFetcherJob, cpr, labResEtl, LAB_RESULT_ETL));
                                    break;
                                default:
                                    throw new NotImplementedException("Unknown job type: " + fetchType);
                            }
                        }
                        return sendNasNotification(result.build());
                    }),
                executorService);
        } catch (Exception exception) {
            log.error("Exception starting job: " + dataFetcherJob, exception);
            return CompletableFuture.completedFuture(new DataFetcherJobExecutionResult(dataFetcherJob, true, "Exception starting job: " + exception));
        } finally {
            MDC.remove(MESSAGE_ID_PROP);
            MDC.remove(JOB_ID_PROP);
        }
    }

    private DataFetcherJobExecutionResult performEtl(DataFetcherJob dataFetcherJob, String cpr, EtlFunc etlFunction, String etlType) {
        if (log.isTraceEnabled()) {
            log.trace("Performing " + etlType + " on: " + dataFetcherJob);
        }

        boolean etlSuccess = etlFunction.apply(cpr, dataFetcherJob.getJobId());
        DataFetcherJobExecutionResult result = new DataFetcherJobExecutionResult(dataFetcherJob, !etlSuccess, etlSuccess ? null : etlType + " FAILED");
        if (log.isTraceEnabled()) {
            log.trace(etlType + " result: " + result);
        }
        return result;
    }

    private DataFetcherJobExecutionResult sendNasNotification(DataFetcherJobExecutionResult executionResult) {
        try {
            final JobStatusNotification nasNotification = new JobStatusNotification();
            nasNotification.setJobId(executionResult.getDataFetcherJob().getJobId());
            nasNotification.setJobStatus(executionResult.isFailed() ? JobStatus.FAILED : JobStatus.OK);
            if (executionResult.isFailed()) {
                log.info("Sending failed job to NAS: " + executionResult);
                nasNotification.setFailedReasonCode(NAS_ERROR_INTERNAL_SERVER_ERROR);
            }
            nasPublisher.publish(nasNotification, executionResult.getDataFetcherJob().getCpr());
            if (log.isTraceEnabled()) {
                log.trace("Successfully sent to NAS: " + executionResult);
            }
        } catch (Exception e) {
            log.error("Exception while sending NAS notification! Job has been processed, but no NAS notification has been sent. Job: " + executionResult.getDataFetcherJob(), e);
            executionResult.setFailed(true).setFailureMessage(e.getMessage());
        }
        return executionResult;
    }

    private DataFetcherJobExecutionResult sendNasFailureNotification(DataFetcherJob dataFetcherJob, Throwable exception) {
        DataFetcherJobExecutionResult executionResult = new DataFetcherJobExecutionResult(dataFetcherJob, true, exception.toString());
        return sendNasNotification(executionResult);
    }

    @Override
    public void close() {
        log.info("Closing " + DataFetcher.class.getSimpleName());
        shutdownTime = Instant.now();
        kafkaConsumerPool.close();
        executorService.shutdown();
        try {
            if (!executorService.awaitTermination(shutdownTimeoutMillis, TimeUnit.MILLISECONDS)) {
                executorService.shutdownNow(); //typical implementations will cancel via Thread.interrupt, so any task that fails to respond to interrupts may never terminate.
                log.error("Could not shutdown a running ExecutorService gracefully! Jobs that did not finish in time may have been consumed (committed) in Kafka, but failed job notifications may NOT have been sent to NAS!");
            }
        } catch (InterruptedException e) {
            //interrupted by another thread while waiting for timeout
            executorService.shutdownNow();
            log.error("Could not shutdown a running ExecutorService gracefully! Jobs that did not finish in time may have been consumed (committed) in Kafka, but failed job notifications have maybe not been sent to NAS!");
            Thread.currentThread().interrupt(); //re-interrupt (demanded by SonarLint rule 'java:S2142')
        }
    }
}
