/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datafetcher.config;

import dk.sundhedsdatastyrelsen.cbs.dao.VaccinationRecord;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.VaccinationRecordImpl;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.Extractor;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.Transformer;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.VaccinationETL;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.ddv.DDVVaccinePasService;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.ddv.DGWSHelper;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.ddv.WhitelistingConfig;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.exceptions.DDVException;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.exceptions.DataFetcherException;
import dk.sundhedsdatastyrelsen.cbs.dgws.DGWSRequestFactory;
import dk.sundhedsdatastyrelsen.cbs.dgws.STSConfig;
import dk.sundhedsdatastyrelsen.cbs.dgws.http.HttpClientService;
import dk.sundhedsdatastyrelsen.cbs.dgws.http.HttpClientServiceImpl;
import dk.sundhedsdatastyrelsen.cbs.dgws.http.HttpExecutor;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.monitoring.IsAliveServlet;
import dk.sundhedsdatastyrelsen.cbs.status.CertificateStatusChecker;
import dk.vaccinationsregister.schemas._2013._12._01.GetVaccinationPassportAsPDFResponseType;
import org.apache.commons.lang3.StringUtils;
import org.apache.log4j.MDC;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Lazy;
import org.springframework.context.annotation.PropertySource;
import org.springframework.core.convert.converter.Converter;
import org.springframework.core.io.Resource;
import org.springframework.lang.NonNull;

import javax.xml.bind.JAXBException;
import java.io.IOException;
import java.io.InputStream;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.time.LocalDateTime;
import java.util.function.Function;

import static dk.sundhedsdatastyrelsen.cbs.CbsLogging.TRACE_ID_PROP;

@Configuration
@PropertySource("classpath:application-datafetcher.properties")
public class DDVConfig {
    private final int vaccinationRecordsTimeToLiveDays;

    public DDVConfig(@Value("${datafetcher.vaccination.records.time-to-live-days:-1}") int vaccinationRecordsTimeToLiveDays) {
        this.vaccinationRecordsTimeToLiveDays = vaccinationRecordsTimeToLiveDays;
    }

    @Bean
    @Lazy
    public DGWSRequestFactory dgwsRequestFactory(STSConfig stsConfig, HttpClientService httpClientService) {
        return new DGWSRequestFactory(stsConfig, httpClientService);
    }

    @Bean
    @Lazy
    public DGWSHelper dgwsHelper(WhitelistingConfig whitelistingConfig, DGWSRequestFactory dgwsRequestFactory) {
        return new DGWSHelper(whitelistingConfig, dgwsRequestFactory);
    }

    @Bean
    @Lazy
    public HttpClientServiceImpl httpClientServiceImpl() {
        return new HttpClientServiceImpl(new HttpExecutor());
    }

    @Bean
    @Lazy
    public DDVVaccinePasService ddvVaccinePasService(DGWSHelper dgwsHelper, @Value("${ddv.service.endpoint}") String endpoint, HttpClientService httpClientService) {
        return new DDVVaccinePasService(dgwsHelper, endpoint, httpClientService);
    }

    @Bean
    @Lazy
    public WhitelistingConfig whitelistingConfig(
        @Value("${ddv.whitelisting.systemname}") String systemName,
        @Value("${ddv.whitelisting.systemownername}") String systemOwnerName,
        @Value("${ddv.whitelisting.systemversion}") String systemVersion,
        @Value("${ddv.whitelisting.orgusingidcvr}") String orgUsingIdCVR,
        @Value("${ddv.whitelisting.orgusingidname}") String orgUsingIdName,
        @Value("${ddv.whitelisting.requestedrole}") String requestedRole) {
        return new WhitelistingConfig(systemName, systemOwnerName, systemVersion, orgUsingIdCVR, orgUsingIdName, requestedRole);
    }

    @Autowired
    void certificateStatusChecker(@Value("${ddv.idcard.keystore.path}") Resource keystoreResource, @Value("${ddv.idcard.keystore.password}") String keystorePassword) throws IOException, KeyStoreException, CertificateException, NoSuchAlgorithmException {
        try (InputStream is = keystoreResource.getInputStream()) {
            final KeyStore keyStore = KeyStore.getInstance("JKS");
            keyStore.load(is, keystorePassword.toCharArray());
            IsAliveServlet.addStatusChecks(new CertificateStatusChecker((X509Certificate) keyStore.getCertificate("SOSI:ALIAS_SYSTEM"), "DDV-DGWS-CERT"));
        }
    }

    @Bean
    @Lazy
    public STSConfig stsConfig(
        @Value("${ddv.idcard.sts.endpoint}") String stsEndpoint,
        @Value("${ddv.idcard.name}") String organizationName,
        @Value("${ddv.idcard.cvr}") String organizationCode,
        @Value("${ddv.idcard.keystore.path}") String keystorePath,
        @Value("${ddv.idcard.keystore.password}") String keystorePassword,
        @Value("${ddv.idcard.systemName}") String systemName) {
        return new STSConfig(stsEndpoint, organizationName, organizationCode, keystorePath, keystorePassword, systemName);
    }

    /**
     * DDV extractor using "vaccinepas-service"
     *
     * @return vaccination status for provided cpr
     */
    @Bean
    @Lazy
    @VaccinationETL
    @NonNull
    public Extractor<GetVaccinationPassportAsPDFResponseType> ddvExtractor(DDVVaccinePasService service) {
        return (cpr, jobId) -> {
            if (StringUtils.isBlank(cpr))
                throw new DataFetcherException("CPR number is missing");
            try {
                return service.getVaccinationPassport(cpr, jobId);
            } catch (JAXBException e) {
                throw new DDVException("Exception while calling DDV:" + e.getMessage(), e);
            }
        };
    }

    @Bean
    @Lazy
    @VaccinationETL
    @NonNull
    public Transformer<GetVaccinationPassportAsPDFResponseType, VaccinationRecord<Vaccination>> ddvTransformer(
        @NonNull final Converter<GetVaccinationPassportAsPDFResponseType, Vaccination> xmlToDtoConverter,
        @NonNull final Function<Vaccination, VaccinationRecord<Vaccination>> dtoToDaoTransformer) {
        return (cpr, source) -> {
            if (StringUtils.isBlank(cpr))
                throw new DataFetcherException("CPR number is missing");
            if (source.getPersonInfo() == null) {
                throw new DDVException("DDV PersonInfo element is missing");
            }
            if (!cpr.equals(source.getPersonInfo().getPersonCivilRegistrationIdentifier()))
                throw new DataFetcherException("Input cpr does not match DDV PersonInfo cpr");
            Vaccination dto = xmlToDtoConverter.convert(source);
            return dtoToDaoTransformer.apply(dto);
        };
    }

    @Bean
    @Lazy
    @VaccinationETL
    @NonNull
    public Function<Vaccination, VaccinationRecord<Vaccination>> vaccinationRecordProducer() {
        return data -> new VaccinationRecordImpl<Vaccination>()
            .cpr(data.cpr())
            .creationTime(LocalDateTime.now())
            .traceId((String) MDC.get(TRACE_ID_PROP))
            .setData(data)
            .invalidationTime(vaccinationRecordsTimeToLiveDays > 0 ? LocalDateTime.now().plusDays(vaccinationRecordsTimeToLiveDays) : null);
    }

}
