/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datafetcher.config;

import dk.oio.rep.itst_dk.xml.schemas._2006._01._17.PersonNameStructureType;
import dk.sundhedsdatastyrelsen.cbs.audit.SystemAuditLogger;
import dk.sundhedsdatastyrelsen.cbs.dao.LabResultRecord;
import dk.sundhedsdatastyrelsen.cbs.dao.PersonRecord;
import dk.sundhedsdatastyrelsen.cbs.dao.VaccinationRecord;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.PersonRecordImpl;
import dk.sundhedsdatastyrelsen.cbs.datacache.Writer;
import dk.sundhedsdatastyrelsen.cbs.datacache.status.JdbcStatusChecker;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.*;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.kids.dao.KidsLabResult;
import dk.vaccinationsregister.schemas._2013._12._01.GetVaccinationPassportAsPDFResponseType;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.PropertySource;
import org.springframework.context.annotation.Scope;
import org.springframework.context.support.PropertySourcesPlaceholderConfigurer;
import org.springframework.dao.DuplicateKeyException;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.lang.NonNull;

import java.time.LocalDateTime;
import java.util.Collection;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.function.Function;

@Configuration
@PropertySource("classpath:application-datafetcher.properties")
public class DataFetcherConfig {

    /**
     * Bean that ensures that the application context fail if properties are missing
     */
    @Bean
    public static PropertySourcesPlaceholderConfigurer propertyPlaceholderConfigurer() {
        return new PropertySourcesPlaceholderConfigurer();
    }

    /**
     * ExecutorService for the concurrent processing in datafetcher.
     * Must be FIFO (e.g. by using an unbounded queue for tasks) to ensure fairness as the ExecutorService bean is shared between all concurrently invocations of the datafetcher
     * (each invocation creates a batch of jobs that are submitted to the executorservice)
     */
    @Bean
    @Scope(BeanDefinition.SCOPE_PROTOTYPE) //use Prototype scope to avoid sharing the executor service by mistake
    public ExecutorService executorService(@Value("${datafetcher.max-threads:500}") int maxThreads) {
        return Executors.newFixedThreadPool(maxThreads);
    }

    /**
     * Produces an ETL (extract transform load) function for vaccinations
     *
     * @param extractor   extracts vaccinations source data (xml) from DDV using a cpr number
     * @param transformer transforms the DDV xml to dao
     * @param writer      loads the dao, returns <code>true</code> when load succeeded else <code>false</code>
     * @return the ETL func
     */
    @Bean
    @NonNull
    @VaccinationETL
    public EtlFunc vaccinationEtl(
        @NonNull final @VaccinationETL Extractor<GetVaccinationPassportAsPDFResponseType> extractor,
        @NonNull final @VaccinationETL Transformer<GetVaccinationPassportAsPDFResponseType, VaccinationRecord<Vaccination>> transformer,
        @NonNull final Writer<VaccinationRecord<Vaccination>> writer) {
        return (cpr, jobId) -> {
            SystemAuditLogger.newAudit()
                .addEntry("operation", "GetVaccinationPassportAsPDF")
                .addEntry("personIdentifier", cpr)
                .log();
            final GetVaccinationPassportAsPDFResponseType card = extractor.extract(cpr, jobId);
            VaccinationRecord<Vaccination> dao = transformer.transform(cpr, card);
            return writer.write(dao);
        };
    }

    /**
     * Produces an ETL (extract transform load) function for test results
     *
     * @param extractor   extracts test results source data (KIDS daos) from KIDS using a cpr number
     * @param transformer transforms the KIDS daos to our dao
     * @param writer      loads the dao, returns <code>true</code> when load succeeded else <code>false</code>
     * @return the ETL func
     */
    @Bean
    @NonNull
    @LabResultETL
    public EtlFunc labResultsEtl(
        @NonNull final @LabResultETL Extractor<Collection<KidsLabResult>> extractor,
        @NonNull final @LabResultETL Transformer<Collection<KidsLabResult>, LabResultRecord<LabResults>> transformer,
        @NonNull final Writer<LabResultRecord<LabResults>> writer) {
        return (cpr, jobId) -> {
            SystemAuditLogger.newAudit()
                .addEntry("operation", "KidsLabResultLookup")
                .addEntry("personIdentifier", cpr)
                .log();
            final Collection<KidsLabResult> labResults = extractor.extract(cpr, jobId);
            LabResultRecord<LabResults> dao = transformer.transform(cpr, labResults);
            return writer.write(dao);
        };
    }

    @Bean
    @Qualifier("cbsJdbcStatusChecker")
    JdbcStatusChecker cbsJdbcStatusChecker(JdbcTemplate jdbcTemplate) {
        return new JdbcStatusChecker(jdbcTemplate, "CBS-DB");
    }
}
