/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datafetcher.ddv;

import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.sdsd.dgws._2012._06.WhiteListingHeader;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.exceptions.DDVException;
import dk.sundhedsdatastyrelsen.cbs.dgws.http.HttpClientService;
import dk.vaccinationsregister.schemas._2013._12._01.GetVaccinationPassportAsPDFRequestType;
import dk.vaccinationsregister.schemas._2013._12._01.GetVaccinationPassportAsPDFResponseType;
import dk.vaccinationsregister.schemas._2013._12._01.ObjectFactory;
import org.oasis_open.docs.wss._2004._01.oasis_200401_wss_wssecurity_secext_1_0.Security;
import org.xmlsoap.schemas.soap.envelope.Body;
import org.xmlsoap.schemas.soap.envelope.Envelope;
import org.xmlsoap.schemas.soap.envelope.Fault;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import java.io.StringReader;
import java.util.List;
import java.util.UUID;

import static dk.sundhedsdatastyrelsen.cbs.metrics.TimerMetric.timedCall;

public class DDVVaccinePasService {

    private static final org.xmlsoap.schemas.soap.envelope.ObjectFactory SOAP_FACTORY = new org.xmlsoap.schemas.soap.envelope.ObjectFactory();

    // Quotes are deliberate (see https://www.w3.org/TR/2000/NOTE-SOAP-20000508/#_Toc478383528)
    private static final String VACCINEPAS_SOAP_ACTION = "\"http://vaccinationsregister.dk/schemas/2013/12/01#GetVaccinationPassportAsPDF\"";

    private final String serviceEndpoint;
    private final DGWSHelper dgwsHelper;
    private final JAXBContext jaxbContext;
    private final HttpClientService httpClientService;

    public DDVVaccinePasService(DGWSHelper dgwsHelper, String serviceEndpoint, HttpClientService httpClientService) {
        this.serviceEndpoint = serviceEndpoint;
        this.dgwsHelper = dgwsHelper;
        this.httpClientService = httpClientService;
        try {
            this.jaxbContext = JAXBContext.newInstance(Envelope.class, Security.class, Header.class, WhiteListingHeader.class, GetVaccinationPassportAsPDFRequestType.class, GetVaccinationPassportAsPDFResponseType.class);
        } catch (JAXBException e) {
            throw new IllegalStateException(e);
        }
    }

    public GetVaccinationPassportAsPDFResponseType getVaccinationPassport(String cpr, String jobId) throws JAXBException {
        GetVaccinationPassportAsPDFRequestType request = new GetVaccinationPassportAsPDFRequestType();
        request.setPersonCivilRegistrationIdentifier(cpr);
        request.setIncludePdf(false);
        request.setIncludeVaccinationStatus(true);
        request.setIncludeVaccinesAndDrugs(true);
        request.setIncludeEffectuationCountryCode(true);
        request.setIncludeAdditionalEffectuations(true);
        request.setVaccinationPlanCategory("COVID_19");
        return timedCall("ddv_GetVaccinationPassportAsPDF", () -> callService(
            new ObjectFactory().createGetVaccinationPassportAsPDFRequest(request),
            serviceEndpoint, jobId));
    }

    private <R, T> T callService(R request, String endpoint, String jobId) throws JAXBException {
        Body soapBody = SOAP_FACTORY.createBody();
        soapBody.getAny().add(request);
        String newRequest = dgwsHelper.addHeaders(jaxbContext, soapBody, jobId + ":" + UUID.randomUUID(), jobId);

        String responseXml = httpClientService.soapRequest(endpoint, DDVVaccinePasService.VACCINEPAS_SOAP_ACTION, newRequest);
        Envelope responseEnvelope = (Envelope) jaxbContext.createUnmarshaller().unmarshal(new StringReader(responseXml));

        try {
            List<Object> any = responseEnvelope.getBody().getAny();
            if (any == null || any.isEmpty() || (!(any.get(0) instanceof JAXBElement))) {
                throw new DDVException("Unexpected DDV response");
            }
            JAXBElement<T> firstAny = (JAXBElement<T>) any.get(0);
            if (firstAny.getValue() instanceof Fault) {
                throw new DDVException(((Fault)firstAny.getValue()).getFaultstring());
            }
            return firstAny.getValue();
        } catch (Exception e) {
            throw new DDVException("Unable to parse body", e);
        }
    }

}
