/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datafetcher.servlet;

import com.google.common.collect.ImmutableMap;
import dk.sdsd.nsp.slalog.ws.SLALoggingServletFilter;
import dk.sundhedsdatastyrelsen.cbs.CbsSpringApplicationContext;
import dk.sundhedsdatastyrelsen.cbs.datacache.status.JdbcStatusChecker;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.DataFetcher;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.DataFetcherConstants;
import dk.sundhedsdatastyrelsen.cbs.kafka.status.KafkaStatusChecker;
import dk.sundhedsdatastyrelsen.cbs.logging.EnhancedSlaLoggingServletFilter;
import dk.sundhedsdatastyrelsen.cbs.logging.Log4JConfigurator;
import dk.sundhedsdatastyrelsen.cbs.monitoring.IsAliveServlet;
import org.apache.log4j.Logger;
import org.springframework.context.annotation.AnnotationConfigApplicationContext;

import javax.servlet.DispatcherType;
import javax.servlet.FilterRegistration;
import javax.servlet.ServletContextEvent;
import javax.servlet.ServletContextListener;
import javax.servlet.annotation.WebListener;
import java.util.EnumSet;

@WebListener
public class DataFetcherServletContextListener implements ServletContextListener {
    public static final String SPRING_APPLICATION_CONTEXT = "SPRING_APPLICATION_CONTEXT";

    // Configure Log4J
    static {
        Log4JConfigurator.configure(DataFetcherConstants.COMPONENT_NAME, DataFetcherServletContextListener.class.getClassLoader());
    }

    private final Logger logger = Logger.getLogger(DataFetcherServletContextListener.class);

    @Override
    public void contextInitialized(ServletContextEvent servletContextEvent) {
        // Configure SLA Logging filter
        SLALoggingServletFilter slaLoggingServletFilter = new EnhancedSlaLoggingServletFilter(ImmutableMap.of("/fetch", "fetch"));
        FilterRegistration.Dynamic dynamic = servletContextEvent.getServletContext().addFilter("slaLoggingFilter", slaLoggingServletFilter);
        dynamic.setInitParameter("shortAppName", "cbs"); // makes nsp-util look for "nspslalog-cbs.properties"
        EnumSet<DispatcherType> dispatcherTypes = EnumSet.of(DispatcherType.REQUEST, DispatcherType.FORWARD);
        dynamic.addMappingForUrlPatterns(dispatcherTypes, false, "/*");

        AnnotationConfigApplicationContext applicationContext = new AnnotationConfigApplicationContext(CbsSpringApplicationContext.class);
        servletContextEvent.getServletContext().setAttribute(SPRING_APPLICATION_CONTEXT, applicationContext);

        IsAliveServlet.setServiceName(DataFetcherConstants.COMPONENT_NAME);
        IsAliveServlet.addStatusChecks(
            applicationContext.getBean("cbsJdbcStatusChecker", JdbcStatusChecker.class),
            applicationContext.getBean("kidsJdbcStatusChecker", JdbcStatusChecker.class),
            applicationContext.getBean(KafkaStatusChecker.class)
        );

        logger.info(DataFetcherConstants.COMPONENT_NAME + " initialized");
    }

    @Override
    public void contextDestroyed(ServletContextEvent servletContextEvent) {
        logger.info("shutting down " + DataFetcherConstants.COMPONENT_NAME);
        AnnotationConfigApplicationContext applicationContext = (AnnotationConfigApplicationContext) servletContextEvent.getServletContext().getAttribute(SPRING_APPLICATION_CONTEXT);
        applicationContext.getBean(DataFetcher.class).close();
        applicationContext.close();
    }
}
