/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datafetcher;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Sets;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.config.DataFetcherConfig;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.config.DataFetcherConfigProperties;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.model.DataFetcherJobExecutionResult;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.kafka.DataFetcherJobBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.kafka.DataFetcherJob;
import dk.sundhedsdatastyrelsen.cbs.dto.kafka.FetchType;
import dk.sundhedsdatastyrelsen.cbs.exceptions.CbsKafkaException;
import dk.sundhedsdatastyrelsen.cbs.kafka.pooling.CbsKafkaConsumerPool;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.NasPublisher;
import dk.sundhedsdatastyrelsen.cbs.test.AssertionTestUtil;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import dk.sundhedsdatastyrelsen.cbs.test.KafkaTestUtils;
import dk.sundhedsdatastyrelsen.cbs.test.RandomData;
import dk.sundhedsdatastyrelsen.coronapass.ws.JobStatus;
import dk.sundhedsdatastyrelsen.coronapass.ws.JobStatusNotification;
import org.apache.log4j.MDC;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.Mock;

import java.util.*;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.ForkJoinPool;
import java.util.concurrent.ForkJoinTask;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.ReentrantLock;
import java.util.stream.Collectors;

import static com.google.common.truth.Truth.assertThat;
import static dk.sundhedsdatastyrelsen.cbs.CbsLogging.TRACE_ID_PROP;
import static dk.sundhedsdatastyrelsen.cbs.datafetcher.DataFetcherTest.FailureSource.*;
import static dk.sundhedsdatastyrelsen.cbs.test.KafkaTestUtils.createRandomJobsWithMultipleFetches;
import static dk.sundhedsdatastyrelsen.cbs.test.KafkaTestUtils.createRandomJobsWithSingleFetches;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

@CbsTest
class DataFetcherTest {
    @Mock
    private CbsKafkaConsumerPool cbsKafkaConsumerPool;
    @Mock
    private EtlFunc vaccinationEtl;
    @Mock
    private EtlFunc labResEtl;
    @Mock
    private NasPublisher nasPublisher;
    private DataFetcherConfigProperties dataFetcherConfigProperties;
    private final ReentrantLock lock = new ReentrantLock();
    private DataFetcher dataFetcher;
    private final ExecutorService executorService = new DataFetcherConfig().executorService(100);

    enum FailureSource {
        ETL_SELF_CHECK_FAILS,
        ETL_EXCEPTION,
        NAS_EXCEPTION,
    }

    @BeforeEach
    void setUp() {
        dataFetcherConfigProperties = new DataFetcherConfigProperties(
            5,
            10000,
            300000,
            20000);
        dataFetcher = new DataFetcher(cbsKafkaConsumerPool, executorService, vaccinationEtl, labResEtl, nasPublisher, dataFetcherConfigProperties);
        lenient().when(vaccinationEtl.apply(anyString(), anyString())).thenReturn(true);
        lenient().when(labResEtl.apply(anyString(), anyString())).thenReturn(true);
        MDC.put(TRACE_ID_PROP, UUID.randomUUID().toString());
    }

    @AfterEach
    void tearDown() {
        MDC.remove(TRACE_ID_PROP);
    }

    @Test
    void test_no_jobs() throws Exception {
        when(cbsKafkaConsumerPool.pollDataFetcherJobs()).thenReturn(Collections.emptySet());
        assertThat(dataFetcher.processJobs()).isEmpty();
        verify(cbsKafkaConsumerPool).pollDataFetcherJobs();
        verifyNoInteractions(vaccinationEtl, labResEtl, nasPublisher);
    }

    @ParameterizedTest(name = "multiple CPRs per job: {0}")
    @ValueSource(booleans = {true, false})
    void test_jobs_processed_successfully(boolean multiFetches) throws Exception {
        final int batchSize = 50;
        final Set<DataFetcherJob> jobs = multiFetches ? createRandomJobsWithMultipleFetches(batchSize) : createRandomJobsWithSingleFetches(batchSize);
        when(cbsKafkaConsumerPool.pollDataFetcherJobs()).thenReturn(jobs);
        final List<DataFetcherJobExecutionResult> results = dataFetcher.processJobs();
        assertEquals(batchSize, results.size());
        assertTrue(results.stream().noneMatch(DataFetcherJobExecutionResult::isFailed));
        verify(nasPublisher, times(batchSize)).publish(any(JobStatusNotification.class), anyString());

        final Set<Map.Entry<String, FetchType>> fetchCprs = jobs.stream()
            .map(DataFetcherJob::getFetchCprs)
            .flatMap(m -> m.entrySet().stream())
            .collect(Collectors.toSet());

        fetchCprs.forEach(f -> {
            if (f.getValue() == FetchType.FETCH_ALL || f.getValue() == FetchType.FETCH_VACCINATION_STATUS) {
                verify(vaccinationEtl).apply(eq(f.getKey()), anyString());
            }
            if (f.getValue() == FetchType.FETCH_ALL || f.getValue() == FetchType.FETCH_TEST_STATUS) {
                verify(labResEtl).apply(eq(f.getKey()), anyString());
            }
        });
    }

    /**
     * Large scope test <br/>
     * - that combines several failure sources in all possible ways: ETL returns false + ETL exception + NAS exception to verify error handling is correct <br/>
     * - that does so in a large job batch size to spot any problems with error handling while concurrently processing jobs <br/>
     * <p>
     * Note: do not parameterize this test. It is helpful to test all combinations of failures together in the same job batch
     */
    @SuppressWarnings({"rawtypes", "unchecked"})
    @Test
    void test_some_jobs_throwing_exceptions_while_others_finish_successfully() throws Exception {
        final String etlExceptionMsg = "ETL EXCEPTION";
        final String nasExceptionMsg = "NAS EXCEPTION";

        final ImmutableSet<FailureSource> failureSource = ImmutableSet.of(ETL_SELF_CHECK_FAILS, ETL_EXCEPTION, NAS_EXCEPTION);
        final Set<Set<FailureSource>> testCases = Sets.powerSet(failureSource) //test all possible combinations of failures
            .stream()
            .filter(tc -> !(tc.contains(ETL_SELF_CHECK_FAILS) && tc.contains(ETL_EXCEPTION))) //remove invalid test case: ETL cannot both throw exception and fail self-check
            .filter(tc -> !(tc.isEmpty())) //remove empty case
            .collect(Collectors.toSet());

        Iterator<Set<FailureSource>> testCasesIterator = testCases.iterator();

        final RandomData randomData = new RandomData(new Random().nextInt());
        final HashSet<DataFetcherJob> jobs = new HashSet<>();
        final int batchSize = 100; //use a high batch size to make concurrent processing occur
        final Map<String, Set<FailureSource>> failingCprs = new HashMap<>();

        //randomly pick the indexes of the jobs that will fail to make sure job order is not a factor in error handling
        final Set[] failJobIndexes = new Set[batchSize];
        randomData.ints(0, batchSize).distinct().limit(testCases.size()).forEach(i -> failJobIndexes[i] = testCasesIterator.next());
        //setup mocks
        for (int i = 0; i < batchSize; i++) {
            String cpr = randomData.cpr();
            String msgId = UUID.randomUUID().toString();
            String jobId = String.valueOf(i);
            if (failJobIndexes[i] != null) {
                Set<FailureSource> testCase = failJobIndexes[i];
                if (testCase.contains(ETL_EXCEPTION)) {
                    failingCprs.computeIfAbsent(cpr, s -> new HashSet<>()).add(ETL_EXCEPTION);
                    lenient().when(vaccinationEtl.apply(cpr, jobId)).thenThrow(new RuntimeException(etlExceptionMsg));
                    lenient().when(labResEtl.apply(cpr, jobId)).thenThrow(new RuntimeException(etlExceptionMsg));
                }
                if (testCase.contains(ETL_SELF_CHECK_FAILS)) {
                    failingCprs.computeIfAbsent(cpr, s -> new HashSet<>()).add(ETL_SELF_CHECK_FAILS);
                    lenient().when(vaccinationEtl.apply(cpr, jobId)).thenReturn(false);
                    lenient().when(labResEtl.apply(cpr, jobId)).thenReturn(false);
                }
                if (testCase.contains(NAS_EXCEPTION)) {
                    failingCprs.computeIfAbsent(cpr, s -> new HashSet<>()).add(NAS_EXCEPTION);
                    lenient().doThrow(new RuntimeException(nasExceptionMsg)).when(nasPublisher).publish(any(), eq(cpr));
                }
                msgId = testCase.toString();
            }
            DataFetcherJob job = DataFetcherJobBuilder.aDataFetcherJob()
                .withCpr(cpr)
                .withMsgId(msgId)
                .withJobId(jobId)
                .withFetchCprs(KafkaTestUtils.getRandomFetchCprs(cpr, 0, 10))
                .build();
            jobs.add(job);
        }
        when(cbsKafkaConsumerPool.pollDataFetcherJobs()).thenReturn(jobs);
        //execute
        final List<DataFetcherJobExecutionResult> results = dataFetcher.processJobs();

        //verify
        assertEquals(batchSize, results.size());
        final List<DataFetcherJobExecutionResult> failedJobs = results.stream().filter(DataFetcherJobExecutionResult::isFailed).collect(Collectors.toList());
        assertEquals(testCases.size(), failedJobs.size());
        assertEquals(batchSize - testCases.size(), results.stream().filter(res -> !res.isFailed()).count());

        List<String> failedCprs = failedJobs.stream()
            .map(DataFetcherJobExecutionResult::getDataFetcherJob)
            .map(DataFetcherJob::getCpr)
            .collect(Collectors.toList());
        assertThat(failingCprs.keySet()).containsExactlyElementsIn(failedCprs);

        for (DataFetcherJobExecutionResult failedJob : failedJobs) {
            Set<FailureSource> testCase = failingCprs.get(failedJob.getDataFetcherJob().getCpr());
            final String expectedFailureMsg = testCase.contains(NAS_EXCEPTION) ? nasExceptionMsg : testCase.contains(ETL_EXCEPTION) ? etlExceptionMsg : "ETL FAILED";
            assertThat(failedJob.getFailureMessage()).endsWith(expectedFailureMsg);
        }

        verify(nasPublisher, times(batchSize)).publish(any(JobStatusNotification.class), anyString());
    }

    @Test
    void test_kafka_exception() throws Exception {
        when(cbsKafkaConsumerPool.pollDataFetcherJobs()).thenThrow(CbsKafkaException.class);
        assertThrows(CbsKafkaException.class, () -> dataFetcher.processJobs());
    }


    enum FakeFetchType {
        fakeType
    }

    @Test
    void fails_on_unknown_job_type() throws Exception {
        final String cpr = "0101012233";
        DataFetcherJob mockJob = mock(DataFetcherJob.class);
        lenient().when(mockJob.getJobId()).thenReturn("jobId");
        lenient().when(mockJob.getCpr()).thenReturn(cpr);
        lenient().when(mockJob.getMsgId()).thenReturn("msgId");
        final ImmutableMap mockFetchMap = mock(ImmutableMap.class);
        when(mockFetchMap.entrySet()).thenReturn(ImmutableSet.of(new AbstractMap.SimpleEntry(cpr, FakeFetchType.fakeType)));
        lenient().when(mockFetchMap.get(anyString())).thenReturn(FakeFetchType.fakeType);
        when(mockJob.getFetchCprs()).thenReturn(mockFetchMap);

        when(cbsKafkaConsumerPool.pollDataFetcherJobs()).thenReturn(Collections.singleton(mockJob));

        List<DataFetcherJobExecutionResult> results = assertDoesNotThrow(() -> dataFetcher.processJobs());
        assertEquals(1, results.size());
        DataFetcherJobExecutionResult result = results.get(0);
        assertTrue(result.isFailed());
        assertThat(result.getFailureMessage()).contains("ClassCastException");
        verify(nasPublisher).publish(argThat(notification -> notification.getJobStatus() == JobStatus.FAILED), eq(cpr));
        verifyNoInteractions(vaccinationEtl, labResEtl);
    }

    @Test
    void test_processing_begin_timeout() throws Exception {
        dataFetcherConfigProperties
            .setMaxConcurrentProcessings(1)
            .setProcessingBeginTimeoutMillis(1);
        final DataFetcher impatientDataFetcher = new DataFetcher(cbsKafkaConsumerPool, executorService, vaccinationEtl, labResEtl, nasPublisher, dataFetcherConfigProperties);

        when(cbsKafkaConsumerPool.pollDataFetcherJobs()).thenAnswer(invocation -> {
            lock.lockInterruptibly(); //blocks if lock is already held
            lock.unlock();
            return createRandomJobsWithMultipleFetches(10);
        });

        lock.lockInterruptibly();
        final ForkJoinTask<List<DataFetcherJobExecutionResult>> task = ForkJoinPool.commonPool().submit(impatientDataFetcher::processJobs);//will hang
        AssertionTestUtil.waitForAsserted(() -> assertThat(lock.getQueueLength()).isGreaterThan(0));

        final List<DataFetcherJobExecutionResult> results = impatientDataFetcher.processJobs(); //should timeout
        assertEquals(0, results.size()); //times out and does not process anything
        verify(cbsKafkaConsumerPool).pollDataFetcherJobs();
        verifyNoInteractions(vaccinationEtl, labResEtl, nasPublisher);

        lock.unlock();
        task.cancel(true);
    }

    @Test
    void test_processing_end_timeout() throws Exception {
        final DataFetcherJob job = KafkaTestUtils.createRandomDataFetcherJobWithSingleFetchCpr(FetchType.FETCH_VACCINATION_STATUS);
        when(cbsKafkaConsumerPool.pollDataFetcherJobs()).thenReturn(Collections.singleton(job));
        when(vaccinationEtl.apply(job.getCpr(), job.getJobId())).thenAnswer(invocation -> {
            lock.lockInterruptibly(); //blocks if lock is already held
            lock.unlock();
            return true;
        });

        dataFetcherConfigProperties.setProcessingEndTimeoutMillis(100);
        final DataFetcher impatientDataFetcher = new DataFetcher(cbsKafkaConsumerPool, executorService, vaccinationEtl, labResEtl, nasPublisher, dataFetcherConfigProperties);
        lock.lockInterruptibly();
        final List<DataFetcherJobExecutionResult> results = impatientDataFetcher.processJobs();
        assertEquals(1, results.size());
        final DataFetcherJobExecutionResult result = results.iterator().next();
        assertTrue(result.isFailed());
        assertEquals(job, result.getDataFetcherJob());
        assertThat(result.getFailureMessage()).isNotEmpty();

        verify(vaccinationEtl).apply(job.getCpr(), job.getJobId());
        verify(nasPublisher).publish(any(JobStatusNotification.class), eq(result.getDataFetcherJob().getCpr()));
        lock.unlock();
    }

    @Test
    void test_shutdown_timeout_when_kafka_hangs() throws Exception {
        dataFetcherConfigProperties.setShutdownTimeoutMillis(1);
        final DataFetcher impatientDataFetcher = new DataFetcher(cbsKafkaConsumerPool, executorService, vaccinationEtl, labResEtl, nasPublisher, dataFetcherConfigProperties);

        when(cbsKafkaConsumerPool.pollDataFetcherJobs()).thenAnswer(invocation -> {
            lock.lockInterruptibly(); //blocks if lock is already held
            lock.unlock();
            return createRandomJobsWithMultipleFetches(10);
        });

        lock.lockInterruptibly();
        final ForkJoinTask<List<DataFetcherJobExecutionResult>> task = ForkJoinPool.commonPool().submit(impatientDataFetcher::processJobs);//will hang
        AssertionTestUtil.waitForAsserted(() -> assertThat(lock.getQueueLength()).isGreaterThan(0));

        impatientDataFetcher.close();
        verify(cbsKafkaConsumerPool).pollDataFetcherJobs();
        verifyNoInteractions(vaccinationEtl, labResEtl, nasPublisher);

        lock.unlock();
        task.cancel(true);
    }

    @Test
    void test_shutdown_timeout_when_etl_hangs() throws Exception {
        final DataFetcherJob job = KafkaTestUtils.createRandomDataFetcherJobWithSingleFetchCpr(FetchType.FETCH_VACCINATION_STATUS);
        when(cbsKafkaConsumerPool.pollDataFetcherJobs()).thenReturn(Collections.singleton(job));

        dataFetcherConfigProperties.setShutdownTimeoutMillis(1);
        final DataFetcher impatientDataFetcher = new DataFetcher(cbsKafkaConsumerPool, executorService, vaccinationEtl, labResEtl, nasPublisher, dataFetcherConfigProperties);

        when(vaccinationEtl.apply(anyString(), anyString())).thenAnswer(invocation -> {
            lock.lockInterruptibly(); //blocks if lock is already held
            lock.unlock();
            return true;
        });

        lock.lockInterruptibly();
        final ForkJoinTask<List<DataFetcherJobExecutionResult>> task = ForkJoinPool.commonPool().submit(impatientDataFetcher::processJobs);//will hang
        AssertionTestUtil.waitForAsserted(() -> verify(vaccinationEtl).apply(anyString(), anyString()));

        impatientDataFetcher.close();
        verify(cbsKafkaConsumerPool).pollDataFetcherJobs();
        verifyNoInteractions(labResEtl, nasPublisher);

        task.quietlyJoin();

        assertTrue(task.isCompletedAbnormally());
        assertThat(task.getException()).hasMessageThat().contains("Processing interrupted due to server shutdown. Halting immediately. NAS messages are NOT sent!");

        lock.unlock();
        task.cancel(true);
    }

    @Test
    void test_concurrent_invocation_overtaking() throws Exception {
        final DataFetcherJob blockingJob = KafkaTestUtils.createRandomDataFetcherJobWithSingleFetchCpr(FetchType.FETCH_ALL);
        final int batchSize = 10;
        //noinspection unchecked
        when(cbsKafkaConsumerPool.pollDataFetcherJobs()).thenReturn(
            Collections.singleton(blockingJob), //first invocation
            createRandomJobsWithMultipleFetches(batchSize) //second invocation
        );

        lenient().when(labResEtl.apply(blockingJob.getCpr(), blockingJob.getJobId())).thenAnswer(invocation -> {
            lock.lockInterruptibly(); //blocks if lock is already held
            lock.unlock();
            return true;
        });
        lenient().when(vaccinationEtl.apply(blockingJob.getCpr(), blockingJob.getJobId())).thenAnswer(invocation -> {
            lock.lockInterruptibly(); //blocks if lock is already held
            lock.unlock();
            return true;
        });

        lock.lockInterruptibly();
        final ForkJoinTask<List<DataFetcherJobExecutionResult>> task = ForkJoinPool.commonPool().submit(dataFetcher::processJobs);//will hang
        AssertionTestUtil.waitForAsserted(() -> assertThat(lock.getQueueLength()).isGreaterThan(0));

        final List<DataFetcherJobExecutionResult> results = dataFetcher.processJobs(); //should not hang
        assertEquals(batchSize, results.size());
        assertTrue(results.stream().noneMatch(DataFetcherJobExecutionResult::isFailed));
        verify(nasPublisher, times(batchSize)).publish(any(JobStatusNotification.class), anyString());
        verifyNoMoreInteractions(nasPublisher);

        lock.unlock();
        task.cancel(true);
    }

    @Test
    void test_max_threads_per_processing_is_respected() throws Exception {
        AtomicInteger messagesSentToNas = new AtomicInteger();

        when(cbsKafkaConsumerPool.pollDataFetcherJobs()).thenReturn(createRandomJobsWithMultipleFetches(10));
        doAnswer((invocation -> {
            messagesSentToNas.getAndIncrement();
            lock.lockInterruptibly(); //blocks the worker thread because lock is held by main test thread
            lock.unlock();
            return invocation.getArgument(0);
        })).when(nasPublisher).publish(any(JobStatusNotification.class), anyString());

        final ExecutorService singleThreadedExecutor = new DataFetcherConfig().executorService(1);
        final DataFetcher singleThreadedDataFetcher = new DataFetcher(cbsKafkaConsumerPool, singleThreadedExecutor, vaccinationEtl, labResEtl, nasPublisher, dataFetcherConfigProperties);
        lock.lockInterruptibly();
        final ForkJoinTask<List<DataFetcherJobExecutionResult>> task = ForkJoinPool.commonPool().submit(singleThreadedDataFetcher::processJobs);//will let 1 job through pipeline, and rest will hang
        AssertionTestUtil.ensureAssertedFor500Millis(() -> assertThat(messagesSentToNas.get()).isAtMost(1));

        lock.unlock();
        task.cancel(true);
    }

}
