/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datafetcher.config;

import dk.oio.rep.cpr_dk.xml.schemas.core._2006._01._17.SimpleCPRPersonType;
import dk.oio.rep.itst_dk.xml.schemas._2006._01._17.PersonNameStructureType;
import dk.sundhedsdatastyrelsen.cbs.dao.VaccinationRecord;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.Transformer;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.exceptions.DDVException;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.exceptions.DataFetcherException;
import dk.sundhedsdatastyrelsen.cbs.ddv.converter.DDVMapperConfiguration;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import dk.vaccinationsregister.schemas._2013._12._01.GetVaccinationPassportAsPDFResponseType;
import org.apache.log4j.MDC;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.context.junit.jupiter.SpringJUnitConfig;

import java.util.UUID;

import static com.google.common.truth.Truth.assertThat;
import static dk.sundhedsdatastyrelsen.cbs.CbsLogging.TRACE_ID_PROP;
import static org.junit.jupiter.api.Assertions.*;

@CbsTest
@SpringJUnitConfig({DDVConfig.class, DDVMapperConfiguration.class})
class DDVConfigTest {
    @Autowired
    private Transformer<GetVaccinationPassportAsPDFResponseType, VaccinationRecord<Vaccination>> ddvTransformer;
    private static final String cpr = "0101012233";

    @AfterEach
    void tearDown() {
        MDC.remove(TRACE_ID_PROP);
    }

    @Test
    void test_transform_minimal_ddv_response() {
        GetVaccinationPassportAsPDFResponseType ddvResponse = getMinimalDdvResponse();
        String traceId = UUID.randomUUID().toString();
        MDC.put(TRACE_ID_PROP, traceId);

        VaccinationRecord<Vaccination> dao = ddvTransformer.transform(cpr, ddvResponse);
        assertEquals(cpr, dao.cpr());
        assertEquals(traceId, dao.traceId());
        assertNotNull(dao.creationTime());
        assertNull(dao.invalidationTime());

        Vaccination vaccination = dao.data();
        assertNotNull(vaccination);
        assertEquals(cpr, vaccination.cpr());
        assertThat(vaccination.vaccinations()).isEmpty();
        assertThat(vaccination.mostRecentlyEffectuated()).isNull();
        assertThat(vaccination.completed()).isNull();
    }

    @Test
    void fail_transforming_invalid_ddv_response() {
        GetVaccinationPassportAsPDFResponseType ddvResponse = getMinimalDdvResponse();
        ddvResponse.setPersonInfo(null);

        DDVException exception = assertThrows(DDVException.class, () -> ddvTransformer.transform(cpr, ddvResponse));
        assertThat(exception).hasMessageThat().contains("PersonInfo element is missing");
    }

    @Test
    void fail_transforming_without_cpr() {
        GetVaccinationPassportAsPDFResponseType ddvResponse = new GetVaccinationPassportAsPDFResponseType();
        SimpleCPRPersonType personInfo = new SimpleCPRPersonType();
        personInfo.setPersonCivilRegistrationIdentifier(cpr);
        PersonNameStructureType personNameStructure = new PersonNameStructureType();
        personNameStructure.setPersonGivenName("Unit");
        personNameStructure.setPersonSurnameName("Test");
        personInfo.setPersonNameStructure(personNameStructure);
        ddvResponse.setPersonInfo(personInfo);

        //noinspection ConstantConditions
        DataFetcherException exception1 = assertThrows(DataFetcherException.class, () -> ddvTransformer.transform(null, ddvResponse));
        assertThat(exception1).hasMessageThat().contains("CPR number is missing");

        DataFetcherException exception2 = assertThrows(DataFetcherException.class, () -> ddvTransformer.transform("", ddvResponse));
        assertThat(exception2).hasMessageThat().contains("CPR number is missing");
    }

    @Test
    void fail_transforming_when_cpr_mismatch() {
        String cpr1 = "0101012233";
        String cpr2 = "0202023344";

        GetVaccinationPassportAsPDFResponseType ddvResponse = new GetVaccinationPassportAsPDFResponseType();
        SimpleCPRPersonType personInfo = new SimpleCPRPersonType();
        personInfo.setPersonCivilRegistrationIdentifier(cpr1);
        PersonNameStructureType personNameStructure = new PersonNameStructureType();
        personNameStructure.setPersonGivenName("Unit");
        personNameStructure.setPersonSurnameName("Test");
        personInfo.setPersonNameStructure(personNameStructure);
        ddvResponse.setPersonInfo(personInfo);

        DataFetcherException exception = assertThrows(DataFetcherException.class, () -> ddvTransformer.transform(cpr2, ddvResponse));
        assertThat(exception).hasMessageThat().contains("Input cpr does not match DDV PersonInfo cpr");
    }

    private GetVaccinationPassportAsPDFResponseType getMinimalDdvResponse() {
        GetVaccinationPassportAsPDFResponseType ddvResponse = new GetVaccinationPassportAsPDFResponseType();
        SimpleCPRPersonType personInfo = new SimpleCPRPersonType();
        personInfo.setPersonCivilRegistrationIdentifier(cpr);
        PersonNameStructureType personNameStructure = new PersonNameStructureType();
        personNameStructure.setPersonGivenName("Unit");
        personNameStructure.setPersonSurnameName("Test");
        personInfo.setPersonNameStructure(personNameStructure);
        ddvResponse.setPersonInfo(personInfo);
        return ddvResponse;
    }

}
