/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datafetcher.config;

import dk.sundhedsdatastyrelsen.cbs.dao.LabResultRecord;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.LabResultRecordImpl;
import dk.sundhedsdatastyrelsen.cbs.datacache.status.JdbcStatusChecker;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.Transformer;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.exceptions.DataFetcherException;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.exceptions.KidsDataException;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.kids.dao.KidsLabResult;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import dk.sundhedsdatastyrelsen.cbs.test.RandomData;
import org.apache.log4j.MDC;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Profile;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.junit.jupiter.SpringJUnitConfig;

import java.time.LocalDateTime;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.UUID;
import java.util.function.Function;
import java.util.function.Supplier;

import static com.google.common.truth.Truth.assertThat;
import static dk.sundhedsdatastyrelsen.cbs.CbsLogging.TRACE_ID_PROP;
import static dk.sundhedsdatastyrelsen.cbs.datafetcher.config.KIDSConfig.NEGATIV;
import static dk.sundhedsdatastyrelsen.cbs.datafetcher.config.KIDSConfig.POSITIV;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

@CbsTest
@SpringJUnitConfig({KIDSConfig.class, KIDSConfigTest.KidsConfigMocks.class})
@ActiveProfiles("KidsConfigMocks")
class KIDSConfigTest {
    private static final String cpr = "0101012233";
    private final RandomData randomData = new RandomData(System.currentTimeMillis());
    @Autowired
    private Transformer<Collection<KidsLabResult>, LabResultRecord<LabResults>> kidsTransformer;
    private String traceId;

    @BeforeEach
    void setUp() {
        traceId = UUID.randomUUID().toString();
        MDC.put(TRACE_ID_PROP, traceId);
    }

    @AfterEach
    void tearDown() {
        MDC.remove(TRACE_ID_PROP);
    }

    @Test
    void test_transform_no_kids_result() {
        LabResultRecord<LabResults> dao = kidsTransformer.transform(cpr, Collections.emptyList());
        assertNotNull(dao);
        assertEquals(cpr, dao.cpr());
        assertEquals(traceId, dao.traceId());
        assertNotNull(dao.creationTime());
        assertNull(dao.invalidationTime());

        LabResults labResults = dao.data();
        assertNotNull(labResults);
        assertEquals(cpr, labResults.cpr());
        assertEquals(0, labResults.results().size());
    }

    @Test
    void test_transform_minimal_result() {
        KidsLabResult kidsLabResult = mock(KidsLabResult.class);

        LocalDateTime creationTime = LocalDateTime.now().minusDays(1);
        LocalDateTime resultTime = LocalDateTime.now().minusDays(2);
        LocalDateTime sampleTime = LocalDateTime.now().minusDays(3);

        lenient().when(kidsLabResult.cpr()).thenReturn(cpr);
        lenient().when(kidsLabResult.id()).thenReturn(42L);
        lenient().when(kidsLabResult.result()).thenReturn(POSITIV);
        lenient().when(kidsLabResult.type()).thenReturn("SARS2");
        lenient().when(kidsLabResult.resultTime()).thenReturn(resultTime);
        lenient().when(kidsLabResult.sampleTime()).thenReturn(sampleTime);
        lenient().when(kidsLabResult.creationTime()).thenReturn(creationTime);
        lenient().when(kidsLabResult.center()).thenReturn("");
        lenient().when(kidsLabResult.manufacturer()).thenReturn("");
        lenient().when(kidsLabResult.name()).thenReturn("");

        LabResultRecord<LabResults> dao = kidsTransformer.transform(kidsLabResult.cpr(), Collections.singleton(kidsLabResult));
        assertEquals(kidsLabResult.cpr(), dao.cpr());
        assertEquals(traceId, dao.traceId());
        assertNotEquals(creationTime, dao.creationTime()); //KIDS record creation time should not be the same as datacache record creation time
        assertNull(dao.invalidationTime());

        LabResults labResults = dao.data();
        assertNotNull(labResults);
        assertEquals(kidsLabResult.cpr(), labResults.cpr());
        assertEquals(1, labResults.results().size());

        LabResult labResult = labResults.results().iterator().next();
        assertNotNull(labResult);
        assertEquals(kidsLabResult.id(), labResult.id());
        assertTrue(labResult.positive());
        assertEquals("SARS2", labResult.type());
        assertEquals(resultTime, labResult.resultDateTime());
        assertEquals(sampleTime, labResult.samplingDateTime());
        assertNull(labResult.center());
        assertNull(labResult.manufacturer());
        assertNull(labResult.name());
    }

    @Test
    void test_transform_full_result() {
        KidsLabResult kidsLabResult = getFullKidsLabResultMock();

        LabResultRecord<LabResults> dao = kidsTransformer.transform(cpr, Collections.singleton(kidsLabResult));
        assertEquals(cpr, dao.cpr());
        assertEquals(traceId, dao.traceId());
        assertNotEquals(kidsLabResult.creationTime(), dao.creationTime()); //KIDS record creation time should not be the same as datacache record creation time
        assertNull(dao.invalidationTime());

        LabResults labResults = dao.data();
        assertNotNull(labResults);
        assertEquals(cpr, labResults.cpr());
        assertEquals(1, labResults.results().size());

        LabResult labResult = labResults.results().iterator().next();
        assertNotNull(labResult);
        assertEquals(kidsLabResult.id(), labResult.id());
        assertEquals(kidsLabResult.result().equalsIgnoreCase(POSITIV), labResult.positive());
        assertEquals("SARS2", labResult.type());
        assertEquals(kidsLabResult.resultTime(), labResult.resultDateTime());
        assertEquals(kidsLabResult.sampleTime(), labResult.samplingDateTime());
        assertThat(labResult.center()).isNotEmpty();
        assertEquals(kidsLabResult.center(), labResult.center());
        assertThat(labResult.manufacturer()).isNotEmpty();
        assertEquals(kidsLabResult.manufacturer(), labResult.manufacturer());
        assertThat(labResult.name()).isNotEmpty();
        assertEquals(kidsLabResult.name(), labResult.name());
    }

    @Test
    void test_transform_multiple_kids_results() {
        KidsLabResult kidsLabResult1 = getMinimalKidsLabResultMock();
        KidsLabResult kidsLabResult2 = getFullKidsLabResultMock();
        when(kidsLabResult1.id()).thenReturn(1L);
        when(kidsLabResult2.id()).thenReturn(2L); //ensure ids are different

        LabResultRecord<LabResults> dao = kidsTransformer.transform(cpr, Arrays.asList(kidsLabResult1, kidsLabResult2));
        assertEquals(cpr, dao.cpr());
        assertEquals(traceId, dao.traceId());
        assertNotNull(dao.creationTime());
        assertNull(dao.invalidationTime());

        LabResults labResults = dao.data();
        assertNotNull(labResults);
        assertEquals(cpr, labResults.cpr());
        assertEquals(2, labResults.results().size());
    }

    @Test
    void fail_transforming_inconclusive_kids_record() {
        KidsLabResult kidsLabResult = getMinimalKidsLabResultMock();
        when(kidsLabResult.result()).thenReturn("gibberish");

        KidsDataException exception = assertThrows(KidsDataException.class, () -> kidsTransformer.transform(cpr, Collections.singleton(kidsLabResult)));
        assertThat(exception).hasMessageThat().contains("inconclusive");
    }


    @Test
    void fail_transforming_invalid_kids_record() {
        KidsLabResult kidsLabResult = getMinimalKidsLabResultMock();
        when(kidsLabResult.type()).thenReturn("");

        KidsDataException exception = assertThrows(KidsDataException.class, () -> kidsTransformer.transform(cpr, Collections.singleton(kidsLabResult)));
        assertThat(exception).hasMessageThat().contains("type missing");
    }

    @Test
    void fail_transforming_without_cpr() {
        DataFetcherException exception1 = assertThrows(DataFetcherException.class, () -> kidsTransformer.transform(null, Collections.emptyList()));
        assertThat(exception1).hasMessageThat().contains("CPR number is missing");

        DataFetcherException exception2 = assertThrows(DataFetcherException.class, () -> kidsTransformer.transform("", Collections.emptyList()));
        assertThat(exception2).hasMessageThat().contains("CPR number is missing");
    }

    @Test
    void fail_transforming_when_cpr_mismatch() {
        String cpr1 = "0101012233";
        String cpr2 = "0202023344";

        final KidsLabResult kidsLabResult1 = getMinimalKidsLabResultMock();
        when(kidsLabResult1.cpr()).thenReturn(cpr1);
        DataFetcherException exception = assertThrows(DataFetcherException.class, () -> kidsTransformer.transform(cpr2, Collections.singletonList(kidsLabResult1)));
        assertThat(exception).hasMessageThat().contains("Input cpr does not match KIDS record cpr");

        final KidsLabResult kidsLabResult2 = getMinimalKidsLabResultMock();
        when(kidsLabResult2.cpr()).thenReturn(cpr2);

        DataFetcherException exception2 = assertThrows(DataFetcherException.class, () -> kidsTransformer.transform(cpr2, Arrays.asList(kidsLabResult2, kidsLabResult1)));
        assertThat(exception2).hasMessageThat().contains("Input cpr does not match KIDS record cpr");
    }

    @Test
    void sets_future_invalidation_date_when_instructed_to() {
        KidsLabResult kidsLabResult = getMinimalKidsLabResultMock();

        int timeToLiveDays = 2;
        KIDSConfig kidsConfig = new KIDSConfig(timeToLiveDays);
        Function<LabResults, LabResultRecord<LabResults>> recordProducer = kidsConfig.labResultsRecordProducer(LabResultRecordImpl::new);

        LabResultRecord<LabResults> dao = kidsConfig
            .kidsTransformer(recordProducer)
            .transform(cpr, Collections.singleton(kidsLabResult));

        LocalDateTime invalidationTime = dao.invalidationTime();
        assertNotNull(invalidationTime);
        LocalDateTime notAfter = LocalDateTime.now().plusDays(timeToLiveDays).plusSeconds(5);
        LocalDateTime notBefore = LocalDateTime.now().plusDays(timeToLiveDays).minusSeconds(5);
        assertTrue(invalidationTime.isAfter(notBefore));
        assertTrue(invalidationTime.isBefore(notAfter));
    }

    private KidsLabResult getMinimalKidsLabResultMock() {
        KidsLabResult kidsLabResult = mock(KidsLabResult.class);

        lenient().when(kidsLabResult.cpr()).thenReturn(cpr);
        lenient().when(kidsLabResult.id()).thenReturn(randomData.nextLong());
        lenient().when(kidsLabResult.result()).thenReturn(POSITIV);
        lenient().when(kidsLabResult.type()).thenReturn("SARS2");
        lenient().when(kidsLabResult.resultTime()).thenReturn(LocalDateTime.now());
        lenient().when(kidsLabResult.sampleTime()).thenReturn(LocalDateTime.now());
        lenient().when(kidsLabResult.creationTime()).thenReturn(LocalDateTime.now());
        lenient().when(kidsLabResult.center()).thenReturn("");
        lenient().when(kidsLabResult.manufacturer()).thenReturn("");
        lenient().when(kidsLabResult.name()).thenReturn("");
        return kidsLabResult;
    }

    private KidsLabResult getFullKidsLabResultMock() {
        KidsLabResult kidsLabResult = mock(KidsLabResult.class);

        lenient().when(kidsLabResult.cpr()).thenReturn(cpr);
        lenient().when(kidsLabResult.id()).thenReturn(randomData.nextLong());
        lenient().when(kidsLabResult.result()).thenReturn(NEGATIV);
        lenient().when(kidsLabResult.type()).thenReturn("SARS2");
        lenient().when(kidsLabResult.resultTime()).thenReturn(LocalDateTime.now());
        lenient().when(kidsLabResult.sampleTime()).thenReturn(LocalDateTime.now());
        lenient().when(kidsLabResult.creationTime()).thenReturn(LocalDateTime.now());
        lenient().when(kidsLabResult.center()).thenReturn("center");
        lenient().when(kidsLabResult.manufacturer()).thenReturn("manufacturer");
        lenient().when(kidsLabResult.name()).thenReturn("name");
        return kidsLabResult;
    }

    @Profile("KidsConfigMocks")
    @Configuration
    static class KidsConfigMocks {
        @Bean
        public Supplier<LabResultRecordImpl<LabResults>> labResultRecordConstructor() {
            return LabResultRecordImpl::new;
        }
        @Bean
        @Qualifier("kidsJdbcStatusChecker")
        JdbcStatusChecker kidsJdbcStatusChecker() {
            return mock(JdbcStatusChecker.class);
        }
    }

}

