/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.ddv.converter;

import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.EffectuatedVaccination;
import dk.vaccinationsregister.schemas._2013._12._01.EffectuationOverviewType;
import dk.vaccinationsregister.schemas._2013._12._01.SSIDrugType;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.log4j.Logger;
import org.springframework.core.convert.converter.Converter;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import java.time.ZoneId;
import java.util.Locale;

import static dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.EffectuatedVaccinationBuilder.effectuatedVaccination;
import static java.util.Optional.ofNullable;

public class EffectuatedVaccinationConverter implements Converter<EffectuationOverviewType, EffectuatedVaccination> {
    private static final Logger log = Logger.getLogger(EffectuatedVaccinationConverter.class);

    @Override
    @Nullable
    public EffectuatedVaccination convert(@NonNull final EffectuationOverviewType source) {
        if (source.getEffectuatedDateTime() == null) {
            log.warn("Discarding EffectuationOverview element due to it having no EffectuatedDateTime.");
            return null; //irrelevant data
        }
        String countryCode = source.getCountryIdentificationCode();
        if (StringUtils.isBlank(countryCode)) {
            log.warn("No CountryIdentificationCode in response. Defaulting to DK");
            countryCode = "DK";
        }
        if (!ArrayUtils.contains(Locale.getISOCountries(), countryCode)) {
            log.warn(String.format("CountryIdentificationCode %s is not a known ISO 3166 two-letter country code. Passing code as-is", countryCode));
        }

        return effectuatedVaccination(
            ofNullable(source.getSSIDrug()).map(SSIDrugType::getDrugIdentifier).orElse(null),
            source.getEffectuatedDateTime().withZoneSameInstant(ZoneId.systemDefault()).toLocalDateTime(),
            countryCode
        ).build();
    }
}
