/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.ddv.converter;

import dk.sundhedsdatastyrelsen.cbs.datafetcher.exceptions.DDVException;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.VaccinationBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.EffectuatedVaccination;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.vaccinationsregister.schemas._2013._12._01.EffectuationOverviewType;
import dk.vaccinationsregister.schemas._2013._12._01.GetVaccinationPassportAsPDFResponseType;
import dk.vaccinationsregister.schemas._2013._12._01.VaccinationStatus;
import org.apache.log4j.Logger;
import org.springframework.core.convert.converter.Converter;
import org.springframework.lang.NonNull;

import java.util.Comparator;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;

import static dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.VaccinationBuilder.vaccination;

public class VaccinationConverter implements Converter<GetVaccinationPassportAsPDFResponseType, Vaccination> {
    private static final Logger log = Logger.getLogger(VaccinationConverter.class);
    public static final String UNKNOWN_CPR = "0000000000";
    public static final String INVALID_DDV_DATA_PREFIX = "INVALID_DDV_DATA: ";
    @NonNull
    private final Converter<EffectuationOverviewType, EffectuatedVaccination> effectuatedVaccinationConverter;

    public VaccinationConverter(
        @NonNull final Converter<EffectuationOverviewType, EffectuatedVaccination> effectuatedVaccinationConverter) {
        this.effectuatedVaccinationConverter = effectuatedVaccinationConverter;
    }

    @Override
    @NonNull
    public Vaccination convert(@NonNull final GetVaccinationPassportAsPDFResponseType source) {
        String cpr = source.getPersonInfo().getPersonCivilRegistrationIdentifier();
        //From CPR_PersonCivilRegistrationIdentifier.xsd: 'The civil registration number may also hold the value 0000000000.
        //This value is used where the civil registration number is required but unknown.
        if (cpr.equals(UNKNOWN_CPR))
            throw new DDVException(INVALID_DDV_DATA_PREFIX + "CPR was '" + UNKNOWN_CPR + "' - which means it is unknown for DDV");

        boolean completed = source.getVaccinationStatus() == VaccinationStatus.VACCINERET;
        final List<EffectuatedVaccination> vaccinations = source.getEffectuation().stream()
            .map(effectuatedVaccinationConverter::convert)
            .filter(Objects::nonNull)
            .collect(Collectors.toList());
        if (completed && vaccinations.isEmpty()) {
            log.warn("Invalid DDV data: VaccinationStatus == 'Vaccineret' but Effectuation list is empty! Marking vaccination as not completed");
            completed = false;
        }
        final List<EffectuatedVaccination> additionalVaccinations = source.getAdditionalEffectuation().stream()
            .map(effectuatedVaccinationConverter::convert)
            .filter(Objects::nonNull)
            .collect(Collectors.toList());
        if (!completed && !additionalVaccinations.isEmpty()) {
            log.warn("Invalid DDV data: Vaccination not completed, but found AdditionalEffectuation elements! Discarding the AdditionalEffectuation elements");
            additionalVaccinations.clear();
        }
        if (!vaccinations.isEmpty() && !additionalVaccinations.isEmpty()) {
            vaccinations.sort(Comparator.comparing(EffectuatedVaccination::effectuated).reversed()); //newest is first
            EffectuatedVaccination lastPlannedEffectuation = vaccinations.get(0);
            final boolean removedAny = additionalVaccinations.removeIf(additionalVac -> additionalVac.effectuated().isBefore(lastPlannedEffectuation.effectuated()));
            if (removedAny) log.warn("Invalid DDV data: There are 'AdditionalEffectuation' elements with 'EffectuatedDateTime' that is before the last 'EffectuatedDateTime' of the 'Effectuation' elements! Discarding the invalid 'AdditionalEffectuation' elements");
        }
        final VaccinationBuilder builder = vaccination(cpr, completed);
        vaccinations.forEach(builder::add);
        additionalVaccinations.forEach(builder::addAdditional);
        return builder.build();
    }

}
