/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.ddv.converter;

import com.google.common.truth.Correspondence;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.exceptions.DDVException;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.EffectuatedVaccination;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import dk.sundhedsdatastyrelsen.cbs.test.RandomData;
import dk.vaccinationsregister.schemas._2013._12._01.EffectuationOverviewType;
import dk.vaccinationsregister.schemas._2013._12._01.GetVaccinationPassportAsPDFResponseType;
import dk.vaccinationsregister.schemas._2013._12._01.VaccinationStatus;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.convert.converter.Converter;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;
import org.springframework.test.context.junit.jupiter.SpringJUnitConfig;

import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.ZoneOffset;
import java.time.ZonedDateTime;
import java.util.Objects;
import java.util.stream.Stream;

import static com.google.common.truth.Truth.assertThat;
import static dk.sundhedsdatastyrelsen.cbs.ddv.converter.VaccinationConverter.INVALID_DDV_DATA_PREFIX;
import static dk.sundhedsdatastyrelsen.cbs.ddv.converter.VaccinationConverter.UNKNOWN_CPR;
import static java.lang.String.format;
import static java.util.stream.Stream.generate;
import static org.junit.jupiter.api.Assertions.*;
import static org.junit.jupiter.params.provider.Arguments.arguments;

@CbsTest
@SpringJUnitConfig(DDVMapperConfiguration.class)
class VaccinationConverterTest {
    @NonNull
    private final Converter<GetVaccinationPassportAsPDFResponseType, Vaccination> converter;
    private static final RandomData RANDOM_DATA = new RandomData("CBS".hashCode());

    @Autowired
    VaccinationConverterTest(
        @NonNull final Converter<GetVaccinationPassportAsPDFResponseType, Vaccination> converter) {
        this.converter = converter;
    }

    @Test
    void effectuated_time_in_systemDefault_ZoneId() {
        ZoneId differentZone = ZoneId.systemDefault().equals(ZoneOffset.UTC) ? ZoneId.of("Europe/Copenhagen") : ZoneOffset.UTC;

        GetVaccinationPassportAsPDFResponseType ddvResponse = DDVTestUtils.newDdvResponse(RANDOM_DATA.cpr());
        ddvResponse.getEffectuation().clear();
        EffectuationOverviewType effectuationWihDate = DDVTestUtils.newEffectuation();
        Instant now = Instant.now();
        effectuationWihDate.setEffectuatedDateTime(now.atZone(differentZone));
        ddvResponse.getEffectuation().add(effectuationWihDate);
        Vaccination vaccination = converter.convert(ddvResponse);
        assertNotNull(vaccination);
        assertNotNull(vaccination.mostRecentlyEffectuated());
        LocalDateTime expectedEffectuationDate = now.atZone(ZoneId.systemDefault()).toLocalDateTime();
        assertEquals(expectedEffectuationDate, vaccination.mostRecentlyEffectuated().effectuated());

        //sanity check (test only makes sense if DDV uses one and we use another)
        assertNotEquals(vaccination.mostRecentlyEffectuated().effectuated(), effectuationWihDate.getEffectuatedDateTime().toLocalDateTime());
    }

    @Test
    void vaccination_complete() {
        GetVaccinationPassportAsPDFResponseType ddvResponse = DDVTestUtils.newDdvResponse(RANDOM_DATA.cpr());
        ddvResponse.setVaccinationStatus(VaccinationStatus.VACCINERET);
        ddvResponse.getEffectuation().clear();
        EffectuationOverviewType effectuationWihDate = DDVTestUtils.newEffectuation();
        ddvResponse.getEffectuation().add(effectuationWihDate);
        Vaccination vaccination = converter.convert(ddvResponse);
        assertNotNull(vaccination);
        LocalDateTime effectuationDate = effectuationWihDate.getEffectuatedDateTime().toLocalDateTime();
        assertEquals(effectuationDate, vaccination.completed());
    }

    @Test
    void planned_effectuations_without_date_are_discarded() {
        GetVaccinationPassportAsPDFResponseType ddvResponse = DDVTestUtils.newDdvResponse(RANDOM_DATA.cpr());
        EffectuationOverviewType plannedEffectuation = DDVTestUtils.newEffectuation();
        plannedEffectuation.setEffectuatedDateTime(null);
        ddvResponse.getEffectuation().add(plannedEffectuation);
        assertTrue(ddvResponse.getEffectuation().stream().anyMatch(e -> e.getEffectuatedDateTime() == null));

        Vaccination vaccination = converter.convert(ddvResponse);
        assertNotNull(vaccination);
        //noinspection ConstantConditions
        assertFalse(vaccination.vaccinations().stream().anyMatch(e -> e.effectuated() == null));
    }

    @Test
    void additional_effectuations_are_discarded_if_vaccination_is_not_completed() {
        GetVaccinationPassportAsPDFResponseType ddvResponse = DDVTestUtils.newDdvResponse(RANDOM_DATA.cpr());
        ddvResponse.setVaccinationStatus(VaccinationStatus.DELVIST_VACCINERET);
        ddvResponse.getAdditionalEffectuation().add(DDVTestUtils.newEffectuation());
        Vaccination vaccination = converter.convert(ddvResponse);
        assertNotNull(vaccination);
        assertThat(vaccination.additionalVaccinations()).isEmpty();
    }

    @Test
    void additional_effectuations_that_are_before_the_last_planned_effectuation_are_discarded() {
        GetVaccinationPassportAsPDFResponseType ddvResponse = DDVTestUtils.newDdvResponse(RANDOM_DATA.cpr());
        ddvResponse.setVaccinationStatus(VaccinationStatus.VACCINERET);
        final ZonedDateTime plannedEffectuationDate = RANDOM_DATA.zonedDateTime();
        final EffectuationOverviewType plannedEffectuation = DDVTestUtils.newEffectuation();
        plannedEffectuation.setEffectuatedDateTime(plannedEffectuationDate);
        ddvResponse.getEffectuation().clear();
        ddvResponse.getEffectuation().add(plannedEffectuation);

        ddvResponse.getAdditionalEffectuation().clear();
        final EffectuationOverviewType goodAdditionalEffectuation = DDVTestUtils.newEffectuation();
        goodAdditionalEffectuation.setEffectuatedDateTime(plannedEffectuationDate.plusDays(5));
        ddvResponse.getAdditionalEffectuation().add(goodAdditionalEffectuation);

        final EffectuationOverviewType badAdditionalEffectuation = DDVTestUtils.newEffectuation();
        badAdditionalEffectuation.setEffectuatedDateTime(plannedEffectuationDate.minusDays(5));
        ddvResponse.getAdditionalEffectuation().add(badAdditionalEffectuation);

        Vaccination vaccination = converter.convert(ddvResponse);
        assertNotNull(vaccination);
        assertThat(vaccination.vaccinations()).isNotEmpty();
        final Correspondence<EffectuatedVaccination, EffectuationOverviewType> correspondingEffectuationDate = Correspondence.from((EffectuatedVaccination cbsDto, EffectuationOverviewType ddvDto) -> cbsDto.effectuated().isEqual(ddvDto.getEffectuatedDateTime().toLocalDateTime()), "matching effectuation date");
        assertThat(vaccination.additionalVaccinations()).comparingElementsUsing(correspondingEffectuationDate).containsExactly(goodAdditionalEffectuation);
    }

    private static Stream<String> missingCountryCodes() {
        return Stream.of(null, "", " ");
    }

    @ParameterizedTest
    @MethodSource("missingCountryCodes")
    void missing_effectuation_country_code_defaults_to_DK(String missingCountryCode) {
        GetVaccinationPassportAsPDFResponseType ddvResponse = DDVTestUtils.newDdvResponse(RANDOM_DATA.cpr());
        ddvResponse.getEffectuation().clear();
        EffectuationOverviewType effectuation = DDVTestUtils.newEffectuation();
        effectuation.setCountryIdentificationCode(missingCountryCode);
        ddvResponse.getEffectuation().add(effectuation);
        Vaccination vaccination = converter.convert(ddvResponse);

        assertNotNull(vaccination);
        assertThat(vaccination.vaccinations()).hasSize(1);
        final EffectuatedVaccination effectuatedVaccination = vaccination.mostRecentlyEffectuated();
        assertNotNull(effectuatedVaccination);
        assertEquals("DK", effectuatedVaccination.countryCode());
        assertNotEquals("DK", missingCountryCode); //sanity check
    }

    @Test
    void invalid_effectuation_country_code_logs_warning_but_passed_as_is() {
        final String invalidCountryCode = "ASDF";
        GetVaccinationPassportAsPDFResponseType ddvResponse = DDVTestUtils.newDdvResponse(RANDOM_DATA.cpr());
        ddvResponse.getEffectuation().clear();
        EffectuationOverviewType effectuation = DDVTestUtils.newEffectuation();
        effectuation.setCountryIdentificationCode(invalidCountryCode);
        ddvResponse.getEffectuation().add(effectuation);
        Vaccination vaccination = converter.convert(ddvResponse);

        assertNotNull(vaccination);
        assertThat(vaccination.vaccinations()).hasSize(1);
        final EffectuatedVaccination effectuatedVaccination = vaccination.mostRecentlyEffectuated();
        assertNotNull(effectuatedVaccination);
        assertEquals(invalidCountryCode, effectuatedVaccination.countryCode());
        assertNotEquals("DK", invalidCountryCode); //sanity check
    }

    @Test
    void throws_exception_on_unknown_cpr() {
        GetVaccinationPassportAsPDFResponseType ddvResponse = DDVTestUtils.newDdvResponse(RANDOM_DATA.cpr());
        ddvResponse.getPersonInfo().setPersonCivilRegistrationIdentifier(UNKNOWN_CPR);
        DDVException ddvException = assertThrows(DDVException.class, () -> converter.convert(ddvResponse));
        assertThat(ddvException).hasMessageThat().contains(INVALID_DDV_DATA_PREFIX);
    }


    @Test
    void vaccination_not_completed_when_VaccinationStatus_is_missing() {
        GetVaccinationPassportAsPDFResponseType ddvResponse = DDVTestUtils.newDdvResponse(RANDOM_DATA.cpr());
        ddvResponse.setVaccinationStatus(null);
        Vaccination vaccination = converter.convert(ddvResponse);
        assertNotNull(vaccination);
        assertNull(vaccination.completed());
    }


    @Test
    void vaccination_not_completed_when_no_effectuations_even_if_VaccinationStatus_is_Vaccineret() {
        GetVaccinationPassportAsPDFResponseType ddvResponse = DDVTestUtils.newDdvResponse(RANDOM_DATA.cpr());
        ddvResponse.setVaccinationStatus(VaccinationStatus.VACCINERET);
        ddvResponse.getEffectuation().clear();
        Vaccination vaccination = converter.convert(ddvResponse);
        assertNotNull(vaccination);
        assertNull(vaccination.completed());
    }

    @Test
    void vaccination_not_completed_when_no_effectuation_with_date_even_if_status_is_complete() {
        GetVaccinationPassportAsPDFResponseType ddvResponse = DDVTestUtils.newDdvResponse(RANDOM_DATA.cpr());
        ddvResponse.setVaccinationStatus(VaccinationStatus.VACCINERET);
        ddvResponse.getEffectuation().clear();
        EffectuationOverviewType effectuationWihNoDate = DDVTestUtils.newEffectuation();
        effectuationWihNoDate.setEffectuatedDateTime(null);
        ddvResponse.getEffectuation().add(effectuationWihNoDate);
        Vaccination vaccination = converter.convert(ddvResponse);
        assertNotNull(vaccination);
        assertNull(vaccination.completed());
    }

    @Test
    void vaccination_not_completed_when_VaccinationStatus_is_not_Vaccineret() {
        GetVaccinationPassportAsPDFResponseType ddvResponse = DDVTestUtils.newDdvResponse(RANDOM_DATA.cpr());
        ddvResponse.setVaccinationStatus(VaccinationStatus.DELVIST_VACCINERET);
        Vaccination vaccination = converter.convert(ddvResponse);
        assertNotNull(vaccination);
        assertNull(vaccination.completed());
    }


    @SuppressWarnings("unused") // Idea does not pickup @ParameterizedTest using friendlyName
    @DisplayName("Conversion of Covid19VaccinationCardType to Vaccination")
    @ParameterizedTest(name = "{index}: {1}")
    @MethodSource("cards")
    void convert(@NonNull final GetVaccinationPassportAsPDFResponseType card, @Nullable final String friendly) {
        final Vaccination vaccination = converter.convert(card);
        assertNotNull(vaccination);
        assertEquals(card.getPersonInfo().getPersonCivilRegistrationIdentifier(), vaccination.cpr());
    }

    private static long countEffectuationsWithNonNullDateTime(@NonNull final GetVaccinationPassportAsPDFResponseType card) {
        return card.getEffectuation()
            .stream()
            .map(EffectuationOverviewType::getEffectuatedDateTime)
            .filter(Objects::nonNull)
            .count();
    }

    @SuppressWarnings("unused")
    @NonNull
    private static Stream<Arguments> cards() {
        return generate(RANDOM_DATA::cpr)
            .limit(100)
            .map(DDVTestUtils::newDdvResponse)
            .map(card -> arguments(card, friendly(card)));
    }

    @NonNull
    private static final String FORMAT_STRING = "{" +
        "\"cpr\"=\"%1$s\"," +
        "\"count(effectuation)\"=%2$d," +
        "\"completed\"=%3$b}";

    @NonNull
    private static String friendly(@NonNull final GetVaccinationPassportAsPDFResponseType card) {
        return format(
            FORMAT_STRING,
            card.getPersonInfo().getPersonCivilRegistrationIdentifier(),
            card.getEffectuation().size(),
            card.getVaccinationStatus() == VaccinationStatus.VACCINERET && countEffectuationsWithNonNullDateTime(card) > 0);
    }
}
