/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.dto.impl.kafka;

import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import com.fasterxml.jackson.annotation.JsonInclude;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.google.common.collect.ImmutableMap;
import dk.sundhedsdatastyrelsen.cbs.CPRFilter;
import dk.sundhedsdatastyrelsen.cbs.dto.kafka.DataFetcherJob;
import dk.sundhedsdatastyrelsen.cbs.dto.kafka.FetchType;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import static com.fasterxml.jackson.annotation.JsonInclude.Include.NON_NULL;

/**
 * POJO representing a job for the datafetcher created by the service
 */
@JsonIgnoreProperties(ignoreUnknown = true)
public class DataFetcherJobImpl implements DataFetcherJob {
    @NonNull
    private final String msgId;
    @NonNull
    private final String jobId;
    @NonNull
    private final String cpr;
    @NonNull
    private final ImmutableMap<String, FetchType> fetchCprs;

    public DataFetcherJobImpl(
        @NonNull @JsonProperty("msgId") final String msgId,
        @Nullable @JsonProperty("type") final FetchType type,
        @NonNull @JsonProperty("jobId") final String jobId,
        @NonNull @JsonProperty("cpr") final String cpr,
        @Nullable @JsonInclude(NON_NULL) @JsonProperty("fetchCprs") final ImmutableMap<String, FetchType> fetchCprs) {
        this.msgId = msgId;
        this.jobId = jobId;
        this.cpr = cpr;

        if (fetchCprs != null) {
            this.fetchCprs = fetchCprs; //1.2.0 behaviour
        } else if (type != null) {
            this.fetchCprs = ImmutableMap.of(cpr, type); //backwards compatible 1.1.0 behaviour (deprecated, to be removed in a later release)
        } else {
            throw new IllegalStateException("type and fetchCprs cannot both be null");
        }
    }

    @NonNull
    public String getMsgId() {
        return msgId;
    }

    @NonNull
    public String getJobId() {
        return jobId;
    }

    @NonNull
    public String getCpr() {
        return cpr;
    }

    @NonNull
    public ImmutableMap<String, FetchType> getFetchCprs() {
        return fetchCprs;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        DataFetcherJobImpl that = (DataFetcherJobImpl) o;

        return new EqualsBuilder()
            .append(msgId, that.msgId)
            .append(jobId, that.jobId)
            .append(cpr, that.cpr)
            .append(fetchCprs, that.fetchCprs).isEquals();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder(17, 37)
            .append(msgId)
            .append(jobId)
            .append(cpr)
            .append(fetchCprs)
            .toHashCode();
    }

    @Override
    public String toString() {
        return "DataFetcherJobImpl{" +
            "msgId='" + msgId + '\'' +
            ", jobId='" + jobId + '\'' +
            ", cpr='" + CPRFilter.PLACEHOLDER + '\'' +
            ", fetchCprs=" + CPRFilter.PLACEHOLDER +
            '}';
    }
}
