/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.dto.impl.lab;

import dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultImpl.AntigenResultImpl;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultImpl.PcrResultImpl;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult;
import org.apache.commons.lang3.NotImplementedException;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import java.time.LocalDateTime;

import static java.lang.String.format;

public class LabResultBuilder<T extends LabResult> {

    private final long id;
    private final LocalDateTime samplingDateTime;
    private final LocalDateTime resultDateTime;
    private final Class<T> clazz;
    private boolean positive;
    private String name;
    private String manufacturer;
    private String center;

    private LabResultBuilder(
        final long id,
        @NonNull final LocalDateTime samplingDateTime,
        @NonNull final LocalDateTime resultDateTime,
        @NonNull final Class<T> clazz) {
        this.id = id;
        this.samplingDateTime = samplingDateTime;
        this.resultDateTime = resultDateTime;
        this.clazz = clazz;
    }

    @NonNull
    public static LabResultBuilder<? extends LabResult> labResult(
        final long id,
        @NonNull final LocalDateTime samplingDateTime,
        @NonNull final LocalDateTime resultDateTime,
        @NonNull final String type) {
        switch (type) {
            case PcrResultImpl.TYPE:
                return pcrResult(id, samplingDateTime, resultDateTime);
            case AntigenResultImpl.TYPE:
                return antigenResult(id, samplingDateTime, resultDateTime);
            default:
                throw new NotImplementedException(format("No handler for lab result type %s implemented", type));
        }
    }

    @NonNull
    public static LabResultBuilder<LabResult.PcrResult> pcrResult(
        final long id,
        @NonNull final LocalDateTime samplingDateTime,
        @NonNull final LocalDateTime resultDateTime) {
        return new LabResultBuilder<>(id, samplingDateTime, resultDateTime, LabResult.PcrResult.class);
    }

    @NonNull
    public static LabResultBuilder<LabResult.AntigenResult> antigenResult(
        final long id,
        @NonNull final LocalDateTime samplingDateTime,
        @NonNull final LocalDateTime resultDateTime) {
        return new LabResultBuilder<>(id, samplingDateTime, resultDateTime, LabResult.AntigenResult.class);
    }

    @NonNull
    public LabResultBuilder<T> positive(final boolean positive) {
        this.positive = positive;
        return this;
    }

    @NonNull
    public LabResultBuilder<T> positive() {
        return positive(true);
    }

    @NonNull
    public LabResultBuilder<T> negative() {
        return positive(false);
    }

    @NonNull
    public LabResultBuilder<T> name(@Nullable final String name) {
        this.name = name;
        return this;
    }

    @NonNull
    public LabResultBuilder<T> manufacturer(@Nullable final String manufacturer) {
        this.manufacturer = manufacturer;
        return this;
    }

    @NonNull
    public LabResultBuilder<T> center(@Nullable final String center) {
        this.center = center;
        return this;
    }

    @SuppressWarnings("unchecked")
    @NonNull public T build() {
        if(clazz.isAssignableFrom(PcrResultImpl.class))
            return positive
                ? (T) new PcrResultImpl.PositivePcrResultImpl(id, name, manufacturer, center, samplingDateTime, resultDateTime)
                : (T) new PcrResultImpl.NegativePcrResultImpl(id, name, manufacturer, center, samplingDateTime, resultDateTime);
        if(clazz.isAssignableFrom(AntigenResultImpl.class))
            return positive
                ? (T)  new AntigenResultImpl.PositiveAntigenResultImpl(id, name, manufacturer, center, samplingDateTime, resultDateTime)
                : (T)  new AntigenResultImpl.NegativeAntigenResultImpl(id, name, manufacturer, center, samplingDateTime, resultDateTime);
        throw new NotImplementedException(format("No handler for lab result class %s implemented", clazz.getName()));

    }

}
