/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.dto.impl.lab;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonGetter;
import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import com.fasterxml.jackson.annotation.JsonProperty;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult.AntigenResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult.PcrResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;

import static dk.sundhedsdatastyrelsen.cbs.Objects.max;
import static java.util.Comparator.comparing;

@JsonIgnoreProperties(ignoreUnknown = true)
public class LabResultsImpl implements LabResults {

    private static final String PCR_RESULTS_PROPERTY = "pcrResults";
    private static final String ANTIGEN_RESULTS_PROPERTY = "antigenResults";
    private static final String CPR_PROPERTY = "cpr";

    @NonNull private final String cpr;
    @NonNull private final List<LabResult> results;
    @NonNull private final List<PcrResult> pcrResults;
    @NonNull private final List<AntigenResult> antigenResults;
    @Nullable private final LabResult mostRecentResult;
    @Nullable private final LabResult.PositiveResult mostRecentPositiveResult;
    @Nullable private final LabResult.NegativeResult mostRecentNegativeResult;
    @Nullable private final PcrResult mostRecentPcrResult;
    @Nullable private final LabResult.PositivePcrResult mostRecentPositivePcrResult;
    @Nullable private final LabResult.NegativePcrResult mostRecentNegativePcrResult;
    @Nullable private final AntigenResult mostRecentAntigenResult;
    @Nullable private final LabResult.PositiveAntigenResult mostRecentPositiveAntigenResult;
    @Nullable private final LabResult.NegativeAntigenResult mostRecentNegativeAntigenResult;

    @JsonCreator
    @SuppressWarnings({"java:S1117"})
    LabResultsImpl(
        @JsonProperty(CPR_PROPERTY) @NonNull String cpr,
        @JsonProperty(PCR_RESULTS_PROPERTY) @NonNull final Collection<? extends PcrResult> pcrResults,
        @JsonProperty(ANTIGEN_RESULTS_PROPERTY) @NonNull final Collection<? extends AntigenResult> antigenResults) {
        this.cpr = cpr;
        this.pcrResults = Collections.unmodifiableList(pcrResults.stream().sorted(comparing(LabResult::samplingDateTime).reversed()).collect(Collectors.toList()));
        this.antigenResults = Collections.unmodifiableList(antigenResults.stream().sorted(comparing(LabResult::samplingDateTime).reversed()).collect(Collectors.toList()));
        final List<LabResult> results = new ArrayList<>(pcrResults.size() + antigenResults.size());
        results.addAll(pcrResults);
        results.addAll(antigenResults);
        results.sort(comparing(LabResult::samplingDateTime));
        this.results = Collections.unmodifiableList(results);
        mostRecentPcrResult = pcrResults.isEmpty() ? null : this.pcrResults.get(0);
        mostRecentAntigenResult = antigenResults.isEmpty() ? null : this.antigenResults.get(0);
        mostRecentPositivePcrResult = (LabResult.PositivePcrResult) this.pcrResults.stream().filter(LabResult::positive).findFirst().orElse(null);
        mostRecentNegativePcrResult = (LabResult.NegativePcrResult) this.pcrResults.stream().filter(LabResult::negative).findFirst().orElse(null);
        mostRecentPositiveAntigenResult = (LabResult.PositiveAntigenResult) this.antigenResults.stream().filter(LabResult::positive).findFirst().orElse(null);
        mostRecentNegativeAntigenResult = (LabResult.NegativeAntigenResult) this.antigenResults.stream().filter(LabResult::negative).findFirst().orElse(null);
        mostRecentResult = max(
            comparing(LabResult::samplingDateTime),
            mostRecentPcrResult,
            mostRecentAntigenResult);
        mostRecentPositiveResult = max(
            comparing(LabResult::samplingDateTime),
            mostRecentPositivePcrResult,
            mostRecentPositiveAntigenResult);
        mostRecentNegativeResult = max(
            comparing(LabResult::samplingDateTime),
            mostRecentNegativePcrResult,
            mostRecentNegativeAntigenResult);
    }

    @Override
    @NonNull
    @JsonGetter
    @JsonProperty(CPR_PROPERTY)
    public String cpr() {
        return cpr;
    }

    @Override
    @NonNull
    public List<LabResult> results() {
        return results;
    }

    @Override
    @NonNull
    @JsonGetter
    @JsonProperty(PCR_RESULTS_PROPERTY)
    public List<PcrResult> pcrResults() {
        return pcrResults;
    }

    @NonNull
    @JsonGetter
    @JsonProperty(ANTIGEN_RESULTS_PROPERTY)
    public List<AntigenResult> antigenResults() {
        return antigenResults;
    }

    @Override
    @Nullable
    public LabResult mostRecentResult() {
        return mostRecentResult;
    }

    @Override
    @Nullable
    public LabResult.PositiveResult mostRecentPositiveResult() {
        return mostRecentPositiveResult;
    }

    @Override
    @Nullable
    public LabResult.NegativeResult mostRecentNegativeResult() {
        return mostRecentNegativeResult;
    }

    @Override
    @Nullable
    public LabResult.PositivePcrResult mostRecentPositivePcrResult() {
        return mostRecentPositivePcrResult;
    }

    @Override
    @Nullable
    public LabResult.NegativePcrResult mostRecentNegativePcrResult() {
        return mostRecentNegativePcrResult;
    }

    @Override
    @Nullable
    public LabResult.PositiveAntigenResult mostRecentPositiveAntigenResult() {
        return mostRecentPositiveAntigenResult;
    }

    @Override
    @Nullable
    public LabResult.NegativeAntigenResult mostRecentNegativeAntigenResult() {
        return mostRecentNegativeAntigenResult;
    }

    @Override
    @Nullable
    public PcrResult mostRecentPcrResult() {
        return mostRecentPcrResult;
    }

    @Override
    @Nullable
    public AntigenResult mostRecentAntigenResult() {
        return mostRecentAntigenResult;
    }
}
