/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonGetter;
import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.databind.SerializerProvider;
import com.fasterxml.jackson.databind.annotation.JsonSerialize;
import com.fasterxml.jackson.databind.ser.std.StdSerializer;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.EffectuatedVaccination;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.VaccinationScheduleType;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import java.io.IOException;
import java.time.LocalDateTime;
import java.util.Collection;
import java.util.Collections;
import java.util.stream.Collectors;

import static dk.sundhedsdatastyrelsen.cbs.dto.vaccination.VaccinationScheduleType.MULTI_DOSE_VACCINATION_SCHEDULE;
import static dk.sundhedsdatastyrelsen.cbs.dto.vaccination.VaccinationScheduleType.SINGLE_DOSE_VACCINATION_SCHEDULE;
import static java.util.Comparator.comparing;
import static java.util.Objects.requireNonNull;

@JsonIgnoreProperties(ignoreUnknown = true)
public class VaccinationImpl implements Vaccination {

    private static final String CPR_NAME_PROPERTY = "cpr";
    private static final String VACCINATIONS_PROPERTY = "vaccinations";
    private static final String ADDITIONAL_VACCINATIONS_PROPERTY = "additionalVaccinations";
    private static final String COMPLETED_PROPERTY = "completed";

    @NonNull
    private final String cpr;
    @NonNull
    private final Collection<EffectuatedVaccination> vaccinations;
    @NonNull
    private final Collection<EffectuatedVaccination> additionalVaccinations;
    @Nullable
    private final LocalDateTime completed;
    @Nullable
    private final EffectuatedVaccination mostRecentlyEffectuated;
    @Nullable
    private final VaccinationScheduleType scheduleType;

    @JsonCreator
    VaccinationImpl(
        @NonNull @JsonProperty(CPR_NAME_PROPERTY) final String cpr,
        @JsonProperty(COMPLETED_PROPERTY) boolean completed,
        @NonNull @JsonProperty(VACCINATIONS_PROPERTY) final Collection<EffectuatedVaccination> vaccinations,
        @NonNull @JsonProperty(ADDITIONAL_VACCINATIONS_PROPERTY) Collection<EffectuatedVaccination> additionalVaccinations) {
        this.cpr = cpr;
        this.vaccinations = Collections.unmodifiableList(vaccinations.stream()
            .sorted(comparing(EffectuatedVaccination::effectuated).reversed())
            .collect(Collectors.toList()));
        this.additionalVaccinations = Collections.unmodifiableList(additionalVaccinations.stream()
            .sorted(comparing(EffectuatedVaccination::effectuated).reversed())
            .collect(Collectors.toList()));
        mostRecentlyEffectuated = this.vaccinations.stream().findFirst().orElse(null);
        this.completed = completed ? requireNonNull(mostRecentlyEffectuated, "Vaccination marked as 'completed' did not contain any effectuations").effectuated() : null;
        if (this.completed != null) scheduleType = vaccinations.size() == 1 ? SINGLE_DOSE_VACCINATION_SCHEDULE : MULTI_DOSE_VACCINATION_SCHEDULE;
        else scheduleType = vaccinations.isEmpty() ? null : MULTI_DOSE_VACCINATION_SCHEDULE;
    }


    @Nullable
    @Override
    public EffectuatedVaccination mostRecentlyEffectuated() {
        return mostRecentlyEffectuated;
    }

    @Nullable
    @Override
    public VaccinationScheduleType scheduleType() {
        return scheduleType;
    }


    @JsonGetter
    @JsonProperty(CPR_NAME_PROPERTY)
    @Override
    @NonNull
    public String cpr() {
        return cpr;
    }

    @Override
    @Nullable
    @JsonGetter
    @JsonProperty(COMPLETED_PROPERTY)
    @JsonSerialize(nullsUsing = CompletedSerializer.class, using = CompletedSerializer.class)
    public LocalDateTime completed() {
        return completed;
    }

    @JsonGetter
    @JsonProperty(VACCINATIONS_PROPERTY)
    @Override
    @NonNull
    public Collection<EffectuatedVaccination> vaccinations() {
        return vaccinations;
    }

    @JsonGetter
    @JsonProperty(ADDITIONAL_VACCINATIONS_PROPERTY)
    @Override
    @NonNull
    public Collection<EffectuatedVaccination> additionalVaccinations() {
        return additionalVaccinations;
    }

    public static class CompletedSerializer extends StdSerializer<LocalDateTime> {

        protected CompletedSerializer() {
            super(LocalDateTime.class);
        }

        @Override
        public void serialize(LocalDateTime value, JsonGenerator gen, SerializerProvider provider) throws IOException {
            gen.writeBoolean(value != null);
        }
    }

}
