/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.dto.impl.kafka;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.google.common.collect.ImmutableMap;
import dk.sundhedsdatastyrelsen.cbs.dto.kafka.DataFetcherJob;
import dk.sundhedsdatastyrelsen.cbs.dto.kafka.FetchType;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.springframework.lang.NonNull;

import java.util.HashMap;
import java.util.UUID;
import java.util.stream.Stream;

import static dk.sundhedsdatastyrelsen.cbs.dto.impl.Harness.OBJECT_MAPPER;
import static dk.sundhedsdatastyrelsen.cbs.dto.impl.Harness.RANDOM_DATA;
import static java.util.stream.Stream.generate;
import static org.junit.jupiter.api.Assertions.*;
import static org.junit.jupiter.api.Assumptions.assumeFalse;

@CbsTest
class DataFetcherJobImplTest {

    @Test
    void test_equals() {
        final String cpr = "0101012233";
        final String msgId = UUID.randomUUID().toString();
        final String jobId = UUID.randomUUID().toString();
        final FetchType fetchType = FetchType.FETCH_ALL;

        DataFetcherJobImpl job1 = new DataFetcherJobImpl(msgId, fetchType, jobId, cpr, null);
        DataFetcherJobImpl job2 = new DataFetcherJobImpl(msgId, fetchType, jobId, cpr, null);

        assertEquals(job1, job2);
        assertEquals(job1.hashCode(), job2.hashCode());
        assertEquals(job1.toString(), job2.toString());

        job1 = new DataFetcherJobImpl(msgId, null, jobId, cpr, ImmutableMap.of("0303033333", fetchType));
        job2 = new DataFetcherJobImpl(msgId, null, jobId, cpr, ImmutableMap.of("0303033333", fetchType));

        assertEquals(job1, job2);
        assertEquals(job1.hashCode(), job2.hashCode());
        assertEquals(job1.toString(), job2.toString());
    }

    @Test
    void test_cpr_not_in_toString() {
        final String msgId = UUID.randomUUID().toString();
        final String jobId = UUID.randomUUID().toString();
        final String cpr1 = "0101012233";
        final String cpr2 = "0202023344";
        DataFetcherJobImpl job1 = new DataFetcherJobImpl(msgId, null, jobId, cpr1, ImmutableMap.of(cpr1, FetchType.FETCH_ALL));
        DataFetcherJobImpl job2 = new DataFetcherJobImpl(msgId, null, jobId, cpr2, ImmutableMap.of(cpr2, FetchType.FETCH_ALL));

        assertNotEquals(job1, job2);
        assertNotEquals(job1.hashCode(), job2.hashCode());
        assertEquals(job1.toString(), job2.toString());

        final ImmutableMap<String, FetchType> fetchCprs1 = ImmutableMap.<String, FetchType>builder()
            .put(cpr1, FetchType.FETCH_ALL)
            .put("0303033333", FetchType.FETCH_TEST_STATUS)
            .build();

        final ImmutableMap<String, FetchType> fetchCprs2 = ImmutableMap.<String, FetchType>builder()
            .put(cpr2, FetchType.FETCH_ALL)
            .put("0404040444", FetchType.FETCH_TEST_STATUS)
            .build();
        job1 = new DataFetcherJobImpl(msgId, null, jobId, cpr1, fetchCprs1);
        job2 = new DataFetcherJobImpl(msgId, null, jobId, cpr2, fetchCprs2);

        assertNotEquals(job1, job2);
        assertNotEquals(job1.hashCode(), job2.hashCode());
        assertEquals(job1.toString(), job2.toString());
    }

    @Test
    void cannot_insert_nulls_in_fetchCprs_map() {
        final HashMap<String, FetchType> nullKeyMap = new HashMap<>();
        nullKeyMap.put(null, FetchType.FETCH_ALL);

        assertThrows(RuntimeException.class, () ->
            new DataFetcherJobImpl(UUID.randomUUID().toString(), null, UUID.randomUUID().toString(), RANDOM_DATA.cpr(), ImmutableMap.copyOf(nullKeyMap))
        );

        final HashMap<String, FetchType> nullValueMap = new HashMap<>();
        nullValueMap.put("0101011111", null);
        assertThrows(RuntimeException.class, () ->
            new DataFetcherJobImpl(UUID.randomUUID().toString(), null, UUID.randomUUID().toString(), RANDOM_DATA.cpr(), ImmutableMap.copyOf(nullValueMap))
        );
    }

    @MethodSource("source")
    @ParameterizedTest
    void serializeDeserialize(@NonNull final DataFetcherJob source) throws JsonProcessingException {
        final String serialized = OBJECT_MAPPER.writeValueAsString(source);
        assertNotNull(serialized);

        final DataFetcherJob deserialized = OBJECT_MAPPER.readValue(serialized, DataFetcherJob.class);
        assertNotNull(deserialized);
        assertEquals(source, deserialized);
    }

    @NonNull
    private static Stream<Arguments> source() {
        return generate(DataFetcherJobImplTest::newJob)
            .limit(100)
            .map(Arguments::arguments);
    }

    @NonNull
    private static DataFetcherJob newJob() {
        final FetchType fetchType = RANDOM_DATA.choose(FetchType.class);
        assumeFalse(fetchType == null, String.format("No %s found", FetchType.class.getSimpleName()));
        final String cpr = RANDOM_DATA.cpr();
        return DataFetcherJobBuilder.aDataFetcherJob()
            .withCpr(cpr)
            .withMsgId(RANDOM_DATA.string(10))
            .withJobId(RANDOM_DATA.string(10))
            .withFetchCprs(ImmutableMap.of(cpr, fetchType))
            .build();
    }

}
