/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.dto.impl.lab;

import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.core.JsonProcessingException;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.Harness;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.springframework.lang.NonNull;

import java.util.Collection;
import java.util.Iterator;
import java.util.stream.Stream;

import static dk.sundhedsdatastyrelsen.cbs.dto.impl.Harness.OBJECT_MAPPER;
import static dk.sundhedsdatastyrelsen.cbs.dto.impl.Harness.RANDOM_DATA;
import static dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultsBuilder.cpr;
import static java.util.stream.Stream.generate;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assumptions.assumeTrue;

class LabResultsImplTest {

    @MethodSource("results")
    @ParameterizedTest
    void serializeDeserialize(@NonNull final LabResults source) throws JsonProcessingException {
        final String serialized = OBJECT_MAPPER.writeValueAsString(source);
        assertNotNull(serialized);

        final LabResults deserialized = OBJECT_MAPPER.readValue(serialized, LabResults.class);
        assertNotNull(deserialized);
        assertEquals(source.cpr(), deserialized.cpr());
        assertEquals(source.mostRecentResult(), deserialized.mostRecentResult());
        assertEquals(source.mostRecentPositiveResult(), deserialized.mostRecentPositiveResult());
        assertEquals(source.mostRecentNegativeResult(), deserialized.mostRecentNegativeResult());
        assertEquals(source.mostRecentPcrResult(), deserialized.mostRecentPcrResult());
        assertEquals(source.mostRecentPositivePcrResult(), deserialized.mostRecentPositivePcrResult());
        assertEquals(source.mostRecentNegativePcrResult(), deserialized.mostRecentNegativePcrResult());
        assertEquals(source.mostRecentAntigenResult(), deserialized.mostRecentAntigenResult());
        assertEquals(source.mostRecentPositiveAntigenResult(), deserialized.mostRecentPositiveAntigenResult());
        assertEquals(source.mostRecentNegativeAntigenResult(), deserialized.mostRecentNegativeAntigenResult());
        assertEquals(source.getClass(), deserialized.getClass());

        equals(source.results(), deserialized.results());
        equals(source.pcrResults(), deserialized.pcrResults());
        equals(source.antigenResults(), deserialized.antigenResults());
    }

    @MethodSource("results")
    @ParameterizedTest
    @DisplayName("Unknown property in serialized form")
    void unknownProperty(@NonNull final LabResults source) throws JsonProcessingException {
        final String serialized = OBJECT_MAPPER.writeValueAsString(new UnknownPropertyLabResults(source));
        assertNotNull(serialized);
        assumeTrue(serialized.contains(UnknownPropertyLabResults.UNKNOWN));

        final LabResults deserialized = OBJECT_MAPPER.readValue(serialized, LabResults.class);
        assertNotNull(deserialized);
    }

    private static void equals(
        @NonNull final Collection<? extends LabResult> sourceCollection,
        @NonNull final Collection<? extends LabResult> deserializedCollection) {
        final Iterator<? extends LabResult> sourceIterator, deserializedIterator;
        for (sourceIterator = sourceCollection.iterator(),
             deserializedIterator = deserializedCollection.iterator();
             sourceIterator.hasNext() && deserializedIterator.hasNext(); ) {
            final LabResult source = sourceIterator.next();
            final LabResult deserialized = deserializedIterator.next();
            Harness.equals(source, deserialized);
        }
        assertEquals(sourceIterator.hasNext(), deserializedIterator.hasNext());
    }


    @NonNull
    private static Stream<Arguments> results() {
        return generate(() -> cpr(RANDOM_DATA.cpr()))
            .flatMap(builder -> RANDOM_DATA.stream(10, Harness::newPcrResult).map(builder::add))
            .flatMap(builder -> RANDOM_DATA.stream(10, Harness::newAntigenResult).map(builder::add))
            .map(LabResultsBuilder::build)
            .limit(100)
            .map(Arguments::arguments);
    }

    private static final class UnknownPropertyLabResults extends LabResultsImpl {
        private static final String UNKNOWN = "unknown";
        @SuppressWarnings({"unused", "FieldCanBeLocal"})
        @JsonProperty("unknown")
        @NonNull
        private final String unknown;
        private UnknownPropertyLabResults(@NonNull final LabResults results) {
            super(results.cpr(), results.pcrResults(), results.antigenResults());
            unknown = UNKNOWN;
        }
    }
}
