/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination;

import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.core.JsonProcessingException;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.EffectuatedVaccination;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.springframework.lang.NonNull;

import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.stream.Stream;

import static dk.sundhedsdatastyrelsen.cbs.dto.impl.Harness.OBJECT_MAPPER;
import static dk.sundhedsdatastyrelsen.cbs.dto.impl.Harness.RANDOM_DATA;
import static dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.EffectuatedVaccinationBuilder.effectuatedVaccination;
import static dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.VaccinationBuilder.vaccination;
import static dk.sundhedsdatastyrelsen.cbs.dto.vaccination.VaccinationScheduleType.MULTI_DOSE_VACCINATION_SCHEDULE;
import static dk.sundhedsdatastyrelsen.cbs.dto.vaccination.VaccinationScheduleType.SINGLE_DOSE_VACCINATION_SCHEDULE;
import static java.lang.String.format;
import static java.util.Comparator.naturalOrder;
import static java.util.stream.Stream.generate;
import static org.junit.jupiter.api.Assertions.*;
import static org.junit.jupiter.api.Assumptions.assumeTrue;

class VaccinationImplTest {

    private static final DateTimeFormatter DATE_FORMAT = DateTimeFormatter.ISO_DATE_TIME;

    @SuppressWarnings("unused")
    @ParameterizedTest(name = "[{index}] {0}")
    @MethodSource("sources")
    @DisplayName("Invariant: If there is a completed date and time, it is the date and time of the most recently effectuated vaccination.")
    void completedAndVaccinations(@NonNull final String friendly, @NonNull final Vaccination vaccination) {
        if (completed(vaccination)) assertTrue(thereAreVaccinations(vaccination));
        if (completed(vaccination)) assertEquals(vaccination.completed(), mostRecentVaccinationDate(vaccination));
        if (!vaccination.vaccinations().isEmpty()) {
            final EffectuatedVaccination mostRecentlyEffectuated = vaccination.mostRecentlyEffectuated();
            assertNotNull(mostRecentlyEffectuated);
            assertEquals(mostRecentVaccinationDate(vaccination), mostRecentlyEffectuated.effectuated());
        }
    }

    @SuppressWarnings("unused")
    @ParameterizedTest(name = "[{index}] {0}")
    @MethodSource("sources")
    @DisplayName("Invariant: If and only if there is a completed date and time and there is only one effectuated vaccination, it is a single dose vaccination schedule")
    void singleDoseVaccinationSchedule(@NonNull final String friendly, @NonNull final Vaccination vaccination) {
        final int numberOfEffectuatedVaccinations = vaccination.vaccinations().size();

        if (completed(vaccination) && numberOfEffectuatedVaccinations == 1) assertEquals(SINGLE_DOSE_VACCINATION_SCHEDULE, vaccination.scheduleType());
        if (vaccination.scheduleType() == SINGLE_DOSE_VACCINATION_SCHEDULE) {
            assertTrue(completed(vaccination));
            assertEquals(1, numberOfEffectuatedVaccinations);
        }
    }

    @SuppressWarnings("unused")
    @ParameterizedTest(name = "[{index}] {0}")
    @MethodSource("sources")
    @DisplayName(
        "Invariant: If and only if the vaccination is a multi dose vaccination, " +
            "it is not completed and has at least one effectuated vaccination or " +
            "it has at least two effectuated vaccinations")
    void multiDoseVaccinationSchedule(@NonNull final String friendly, @NonNull final Vaccination vaccination) {
        final int numberOfEffectuatedVaccinations = vaccination.vaccinations().size();
        if (vaccination.scheduleType() == MULTI_DOSE_VACCINATION_SCHEDULE) {
            assertTrue(
                notCompleted(vaccination) && numberOfEffectuatedVaccinations >= 1 ||
                    numberOfEffectuatedVaccinations >= 2);
        }
        if (notCompleted(vaccination) && numberOfEffectuatedVaccinations >= 1 || numberOfEffectuatedVaccinations >= 2) {
            assertEquals(MULTI_DOSE_VACCINATION_SCHEDULE, vaccination.scheduleType());
        }
    }

    @SuppressWarnings("unused")
    @ParameterizedTest(name = "[{index}] {0}")
    @MethodSource("sources")
    @DisplayName("Invariant: If there are additional effectuations, the vaccination schedule is completed")
    void additionalVaccinations(@NonNull final String friendly, @NonNull final Vaccination vaccination) {
        if (thereAreAdditionalVaccinations(vaccination)) assertTrue(completed(vaccination));
        if (thereAreAdditionalVaccinations(vaccination)) assertTrue(thereAreVaccinations(vaccination));

        if (notCompleted(vaccination)) assertFalse(thereAreAdditionalVaccinations(vaccination));
        if (!thereAreVaccinations(vaccination)) assertFalse(thereAreAdditionalVaccinations(vaccination));
    }

    @SuppressWarnings("unused")
    @ParameterizedTest(name = "[{index}] {0}")
    @MethodSource("sources")
    @DisplayName("Invariant: If and only if there are no effectuated vaccinations, it has no schedule type")
    void noSchedule(@NonNull final String friendly, @NonNull final Vaccination vaccination) {
        final boolean noEffectuatedVacciantions = vaccination.vaccinations().isEmpty();
        if (vaccination.scheduleType() == null) assertTrue(noEffectuatedVacciantions);
        if (noEffectuatedVacciantions) assertNull(vaccination.scheduleType());
    }

    @SuppressWarnings("unused")
    @ParameterizedTest(name = "[{index}] {0}")
    @MethodSource("sources")
    @DisplayName("Serialize and deserialze")
    void serializeDeserialize(@NonNull final String friendly, @NonNull final Vaccination source) throws JsonProcessingException {

        final String serialized = OBJECT_MAPPER.writeValueAsString(source);
        assertNotNull(serialized);

        final Vaccination deserialized = OBJECT_MAPPER.readValue(serialized, Vaccination.class);
        assertNotNull(deserialized);
        assertEquals(source.cpr(), deserialized.cpr());
        assertEquals(source.completed(), deserialized.completed());
        assertEquals(source.vaccinations().size(), deserialized.vaccinations().size());
        assertEquals(source.additionalVaccinations().size(), deserialized.additionalVaccinations().size());
        assertEquals(source.scheduleType(), deserialized.scheduleType());
        assertEquals(source.getClass(), deserialized.getClass());
    }

    @SuppressWarnings("unused")
    @ParameterizedTest(name = "[{index}] {0}")
    @MethodSource("sources")
    @DisplayName("Unknown property in serialized form")
    void unknownProperty(@NonNull final String friendly, @NonNull final Vaccination source) throws JsonProcessingException {

        final String serialized = OBJECT_MAPPER.writeValueAsString(new UnknownPropertyVaccination(source));
        assertNotNull(serialized);
        assumeTrue(serialized.contains(UnknownPropertyVaccination.UNKNOWN));

        final Vaccination deserialized = OBJECT_MAPPER.readValue(serialized, Vaccination.class);
        assertNotNull(deserialized);
    }

    private static boolean completed(@NonNull final Vaccination vaccination) {
        return vaccination.completed() != null;
    }

    private static boolean notCompleted(@NonNull final Vaccination vaccination) {
        return !completed(vaccination);
    }

    private LocalDateTime mostRecentVaccinationDate(Vaccination vaccination) {
        return vaccination.vaccinations().stream()
            .map(EffectuatedVaccination::effectuated)
            .max(naturalOrder())
            .orElse(null);
    }

    private static boolean thereAreVaccinations(Vaccination vaccination) {
        return !vaccination.vaccinations().isEmpty();
    }

    private static boolean thereAreAdditionalVaccinations(Vaccination vaccination) {
        return !vaccination.additionalVaccinations().isEmpty();
    }

    @NonNull
    private static Stream<Arguments> sources() {
        return generate(VaccinationImplTest::newVaccination)
            .limit(100)
            .map(v -> Arguments.arguments(friendly(v), v));
    }

    @NonNull
    private static Vaccination newVaccination() {
        final boolean completed = RANDOM_DATA.nextBoolean();
        final VaccinationBuilder builder = vaccination(RANDOM_DATA.cpr(), completed);
        generate(VaccinationImplTest::newEffectuatedVaccination)
            .limit(RANDOM_DATA.nextInt(completed ? 1 : 0, 5))
            .forEach(builder::add);
        if (completed) {
            generate(VaccinationImplTest::newEffectuatedVaccination)
                .limit(RANDOM_DATA.nextInt(2))
                .forEach(builder::addAdditional);
        }
        return builder.build();
    }

    @NonNull
    private static EffectuatedVaccination newEffectuatedVaccination() {
        return effectuatedVaccination(
            RANDOM_DATA.nextBoolean() ? RANDOM_DATA.nextLong() : null,
            RANDOM_DATA.localDateTime(),
            RANDOM_DATA.countryCode())
            .build();
    }

    @NonNull
    public static String friendly(@NonNull final Vaccination vaccination) {
        final LocalDateTime completed = vaccination.completed();
        return format(
            "{ completed = %s, size = %d, type = %s }",
            completed == null ? null : DATE_FORMAT.format(completed),
            vaccination.vaccinations().size(),
            vaccination.scheduleType());
    }


    private static final class UnknownPropertyVaccination extends VaccinationImpl {
        private static final String UNKNOWN = "unknown";
        @SuppressWarnings({"unused", "FieldCanBeLocal"})
        @JsonProperty("unknown")
        @NonNull
        private final String unknown;

        private UnknownPropertyVaccination(@NonNull final Vaccination vaccination) {
            super(
                vaccination.cpr(),
                vaccination.completed() != null,
                vaccination.vaccinations(),
                vaccination.additionalVaccinations());
            unknown = UNKNOWN;
        }
    }

}
