/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.test;

import dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.VaccinationBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.EffectuatedVaccination;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import org.apache.commons.lang3.StringUtils;
import org.springframework.lang.NonNull;

import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Arrays;
import java.util.concurrent.atomic.AtomicLong;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultBuilder.antigenResult;
import static dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultBuilder.pcrResult;
import static dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultsBuilder.cpr;
import static dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.EffectuatedVaccinationBuilder.effectuatedVaccination;
import static java.lang.String.format;

public final class Data {
    private static final DateTimeFormatter DATE_FORMAT = DateTimeFormatter.ISO_DATE_TIME;
    public static final String EU_APPROVED_MANUFACTURER = "1833";
    public static final String EU_UNAPPROVED_MANUFACTURER = "42";
    public static final long MODERNA_DRUG_IDENTIFIER = 32138956360L;
    public static final long PHIZER_DRUG_IDENTIFIER = 32138956351L;
    public static final long ASTRA_ZENICA_DRUG_IDENTIFIER = 32139944147L;
    private static final LocalDateTime[] NO_REVACCINATIONS = new LocalDateTime[0];

    private Data() {
    }

    private static final RandomData RANDOM_DATA = new RandomData("CBS".hashCode());
    private static final String DEFAULT_CPR = RANDOM_DATA.cpr();

    @NonNull
    public static String friendly(@NonNull final LabResult labResult) {
        return format(
            "{ time = %s, type = %s, result = %s }",
            DATE_FORMAT.format(labResult.samplingDateTime()),
            labResult instanceof LabResult.PcrResult ? "pcr" : "ag",
            labResult.positive() ? "pos" : "neg");
    }

    @NonNull
    public static String friendly(@NonNull final Vaccination vaccination) {
        final LocalDateTime completed = vaccination.completed();
        return format(
            "{ completed = %s, type = %s, [ %s ] }",
            completed == null ? null : DATE_FORMAT.format(completed),
            vaccination.scheduleType(),
            join(vaccination.vaccinations().stream().map(Data::friendly), ","));
    }

    @NonNull
    public static String friendly(@NonNull final EffectuatedVaccination vaccination) {
        return format(
            "{ date = %s }",
            DATE_FORMAT.format(vaccination.effectuated()));
    }

    @NonNull
    public static String join(@NonNull final Stream<?> items, @NonNull final CharSequence separator) {
        return StringUtils.join(items.collect(Collectors.toList()), separator);
    }

    public static final class Vaccinations {
        private Vaccinations() {
        }

        public static final Vaccination EMPTY = VaccinationBuilder.vaccination(DEFAULT_CPR, false).build();

        @NonNull
        public static Vaccination completedMultiDose(@NonNull final LocalDateTime completedDate) {
            return completed(completedDate, true);
        }

        @NonNull
        public static Vaccination completedMultiDose(@NonNull final String cpr, @NonNull final LocalDateTime completedDate) {
            return completed(cpr, completedDate, true);
        }

        @NonNull
        public static Vaccination completedMultiDoseWithAdditionalVaccinations(@NonNull final LocalDateTime completedDate, final int additionalVaccinations) {
            return completed(completedDate, true, additionalVaccinations);
        }

        @NonNull
        public static Vaccination completedMultiDoseWithAdditionalVaccinations(@NonNull final String cpr, @NonNull final LocalDateTime completedDate, final int additionalVaccinations) {
            return completed(cpr, completedDate, true, additionalVaccinations);
        }

        @NonNull
        public static Vaccination completedSingleDose(@NonNull final LocalDateTime completedDate) {
            return completed(completedDate, false);
        }

        @NonNull
        public static Vaccination completedSingleDose(@NonNull final String cpr, @NonNull final LocalDateTime completedDate) {
            return completed(cpr, completedDate, false);
        }

        @NonNull
        public static Vaccination completedSingleDoseWithAdditionalVaccinations(@NonNull final LocalDateTime completedDate, final int additionalVaccinations) {
            return completed(completedDate, false, additionalVaccinations);
        }

        @NonNull
        public static Vaccination completedSingleDoseWithAdditionalVaccinations(@NonNull final String cpr, @NonNull final LocalDateTime completedDate, final int additionalVaccinations) {
            return completed(cpr, completedDate, false, additionalVaccinations);
        }

        @NonNull
        public static Vaccination completed(@NonNull final LocalDateTime completedDate, final boolean multiDose, final int additionalVaccinations) {
            return multiDose
                ? vaccination(true, additionalVaccinations, completedDate, completedDate.minusDays(40))
                : vaccination(true, additionalVaccinations, completedDate);
        }

        @NonNull
        public static Vaccination completed(@NonNull final String cpr, @NonNull final LocalDateTime completedDate, final boolean multiDose, final int additionalVaccinations) {
            return multiDose
                ? vaccination(cpr, true, additionalVaccinations, completedDate, completedDate.minusDays(40))
                : vaccination(cpr, true, additionalVaccinations, completedDate);
        }

        @NonNull
        public static Vaccination partial(@NonNull final LocalDateTime... effectuatedDates) {
            return vaccination(false, NO_REVACCINATIONS, effectuatedDates);
        }

        @NonNull
        public static Vaccination partial(@NonNull final String cpr, @NonNull final LocalDateTime... effectuatedDates) {
            return vaccination(cpr, false, NO_REVACCINATIONS, effectuatedDates);
        }

        @NonNull
        public static Vaccination vaccination(final boolean completed, final int additionalVaccinations, @NonNull final LocalDateTime... effectuatedDates) {
            return vaccination(DEFAULT_CPR, completed, additionalVaccinations, effectuatedDates);
        }

        @NonNull
        public static Vaccination vaccination(@NonNull final String cpr, final boolean completed, final int additionalVaccinations, @NonNull final LocalDateTime... effectuatedDates) {
            final VaccinationBuilder builder = VaccinationBuilder.vaccination(cpr, completed);
            Arrays.stream(effectuatedDates)
                .map(e -> effectuatedVaccination(
                    RANDOM_DATA.choose(MODERNA_DRUG_IDENTIFIER, ASTRA_ZENICA_DRUG_IDENTIFIER, PHIZER_DRUG_IDENTIFIER),
                    e,
                    RANDOM_DATA.countryCode()).build())
                .forEach(builder::add);
            if (completed) {
                LocalDateTime additionalVaccinationDate = effectuatedDates[effectuatedDates.length - 1].plusMonths(6);
                for (int i = 0; i < additionalVaccinations; i++) {
                    builder.addAdditional(effectuatedVaccination(
                        RANDOM_DATA.choose(MODERNA_DRUG_IDENTIFIER, ASTRA_ZENICA_DRUG_IDENTIFIER, PHIZER_DRUG_IDENTIFIER),
                        additionalVaccinationDate,
                        RANDOM_DATA.countryCode()).build());
                    additionalVaccinationDate = additionalVaccinationDate.plusMonths(6);
                }
            }
            return builder.build();
        }

        @NonNull
        public static Vaccination completedMultiDoseWithAdditionalVaccinations(@NonNull final LocalDateTime completedDate, @NonNull final LocalDateTime... revaccinationDates) {
            return completed(completedDate, true, revaccinationDates);
        }

        @NonNull
        public static Vaccination completedMultiDoseWithAdditionalVaccinations(@NonNull final String cpr, @NonNull final LocalDateTime completedDate, @NonNull final LocalDateTime... revaccinationDates) {
            return completed(cpr, completedDate, true, revaccinationDates);
        }

        @NonNull
        public static Vaccination completedSingleDoseWithAdditionalVaccinations(@NonNull final LocalDateTime completedDate, @NonNull final LocalDateTime... revaccinationDates) {
            return completed(completedDate, false, revaccinationDates);
        }

        @NonNull
        public static Vaccination completedSingleDoseWithAdditionalVaccinations(@NonNull final String cpr, @NonNull final LocalDateTime completedDate, @NonNull final LocalDateTime... revaccinationDates) {
            return completed(cpr, completedDate, false, revaccinationDates);
        }

        @NonNull
        public static Vaccination completed(@NonNull final LocalDateTime completedDate, final boolean multiDose, @NonNull final LocalDateTime... revaccinationDates) {
            return multiDose
                ? vaccination(true, revaccinationDates, completedDate, completedDate.minusDays(40))
                : vaccination(true, revaccinationDates, completedDate);
        }

        @NonNull
        public static Vaccination completed(@NonNull final String cpr, @NonNull final LocalDateTime completedDate, final boolean multiDose, @NonNull final LocalDateTime... revaccinationDates) {
            return multiDose
                ? vaccination(cpr, true, revaccinationDates, completedDate, completedDate.minusDays(40))
                : vaccination(cpr, true, revaccinationDates, completedDate);
        }

        @NonNull
        public static Vaccination vaccination(final boolean completed, @NonNull final LocalDateTime[] revaccinationDates, @NonNull final LocalDateTime... effectuatedDates) {
            return vaccination(DEFAULT_CPR, completed, revaccinationDates, effectuatedDates);
        }

        @NonNull
        public static Vaccination vaccination(@NonNull final String cpr, final boolean completed, @NonNull final LocalDateTime[] revaccinationDates, @NonNull final LocalDateTime... effectuatedDates) {
            final VaccinationBuilder builder = VaccinationBuilder.vaccination(cpr, completed);
            Arrays.stream(effectuatedDates)
                .map(e -> effectuatedVaccination(
                    RANDOM_DATA.choose(MODERNA_DRUG_IDENTIFIER, ASTRA_ZENICA_DRUG_IDENTIFIER, PHIZER_DRUG_IDENTIFIER),
                    e,
                    RANDOM_DATA.countryCode()).build())
                .forEach(builder::add);
            if (completed) {
                for (LocalDateTime reVaccinationDate : revaccinationDates) {
                    builder.addAdditional(effectuatedVaccination(
                        RANDOM_DATA.choose(MODERNA_DRUG_IDENTIFIER, ASTRA_ZENICA_DRUG_IDENTIFIER, PHIZER_DRUG_IDENTIFIER),
                        reVaccinationDate,
                        RANDOM_DATA.countryCode()).build());
                }
            }
            return builder.build();
        }
    }

    public static final class Results {
        private Results() {
        }

        public static final LabResults EMPTY = cpr(DEFAULT_CPR).build();
        private static final AtomicLong NEXT_IDENTIFIER = new AtomicLong();

        @NonNull
        public static LabResults positivePcr(@NonNull final LocalDateTime... samplingTimes) {
            return positivePcr(DEFAULT_CPR, samplingTimes);
        }

        @NonNull
        public static LabResults positivePcr(@NonNull String cpr, @NonNull final LocalDateTime... samplingTimes) {
            return Arrays.stream(samplingTimes).sequential()
                .reduce(
                    cpr(cpr), (builder, samplingTime) -> builder.add(pcrResult(NEXT_IDENTIFIER.incrementAndGet(), samplingTime, samplingTime.plusHours(RANDOM_DATA.nextInt(4, 42))).positive().build()), (a, b) -> null)
                .build();
        }

        @NonNull
        public static LabResults positivePcrBeforeNegativePcr(@NonNull final LocalDateTime negativePcr) {
            return cpr(DEFAULT_CPR)
                .add(pcrResult(NEXT_IDENTIFIER.incrementAndGet(), negativePcr.minusNanos(1), negativePcr.plusHours(RANDOM_DATA.nextInt(4, 42))).positive().build())
                .add(pcrResult(NEXT_IDENTIFIER.incrementAndGet(), negativePcr, negativePcr.plusHours(RANDOM_DATA.nextInt(4, 42))).negative().build())
                .build();
        }

        @NonNull
        public static LabResults negativePcrAndNegativeAntigen(@NonNull final LocalDateTime negativePcr, @NonNull final LocalDateTime negativeAntigen) {
            return cpr(DEFAULT_CPR)
                .add(pcrResult(NEXT_IDENTIFIER.incrementAndGet(), negativePcr, negativePcr.plusHours(RANDOM_DATA.nextInt(4, 42))).negative().build())
                .add(antigenResult(NEXT_IDENTIFIER.incrementAndGet(), negativeAntigen, negativeAntigen.plusHours(RANDOM_DATA.nextInt(4, 42))).negative().build())
                .build();
        }

        @NonNull
        public static LabResults negativePcr(@NonNull final LocalDateTime... samplingTimes) {
            return negativePcr(DEFAULT_CPR, samplingTimes);
        }

        @NonNull
        public static LabResults negativePcr(@NonNull String cpr, @NonNull final LocalDateTime... samplingTimes) {
            return Arrays.stream(samplingTimes).sequential()
                .reduce(
                    cpr(cpr), (builder, samplingTime) -> builder.add(pcrResult(NEXT_IDENTIFIER.incrementAndGet(), samplingTime, samplingTime.plusHours(RANDOM_DATA.nextInt(4, 42))).negative().build()), (a, b) -> null)
                .build();
        }

        @NonNull
        public static LabResults negativeAntigen(@NonNull final LocalDateTime... samplingTimes) {
            return negativeAntigen(DEFAULT_CPR, samplingTimes);
        }

        public static LabResults negativeAntigen(@NonNull String cpr, @NonNull final LocalDateTime... samplingTimes) {
            return Arrays.stream(samplingTimes).sequential()
                .reduce(
                    cpr(cpr), (builder, samplingTime) -> builder.add(antigenResult(NEXT_IDENTIFIER.incrementAndGet(), samplingTime, samplingTime.plusHours(RANDOM_DATA.nextInt(4, 42))).negative().manufacturer(EU_APPROVED_MANUFACTURER).build()), (a, b) -> null)
                .build();
        }

        @NonNull
        public static LabResults negativeEuApprovedAntigen(@NonNull final LocalDateTime samplingTime) {
            return cpr(DEFAULT_CPR)
                .add(antigenResult(NEXT_IDENTIFIER.incrementAndGet(), samplingTime, samplingTime.plusHours(RANDOM_DATA.nextInt(4, 42))).negative().manufacturer(EU_APPROVED_MANUFACTURER).build())
                .build();
        }

        @NonNull
        public static LabResults negativeEuUnapprovedAntigen(@NonNull final LocalDateTime samplingTime) {
            return cpr(DEFAULT_CPR)
                .add(antigenResult(NEXT_IDENTIFIER.incrementAndGet(), samplingTime, samplingTime.plusHours(RANDOM_DATA.nextInt(4, 42))).negative().manufacturer(EU_UNAPPROVED_MANUFACTURER).build())
                .build();
        }

        @NonNull
        public static LabResults negativeAntigenPositivePcr(@NonNull final LocalDateTime antigenTime, @NonNull final LocalDateTime pcrTime) {
            return cpr(DEFAULT_CPR)
                .add(antigenResult(NEXT_IDENTIFIER.incrementAndGet(), antigenTime, antigenTime.plusHours(RANDOM_DATA.nextInt(4, 42))).negative().build())
                .add(pcrResult(NEXT_IDENTIFIER.incrementAndGet(), pcrTime, pcrTime.plusHours(RANDOM_DATA.nextInt(4, 42))).positive().build())
                .build();
        }

        @NonNull
        public static LabResults positiveAntigen(@NonNull final LocalDateTime samplingTime) {
            return positiveAntigen(DEFAULT_CPR, samplingTime);
        }

        @NonNull
        public static LabResults positiveAntigen(@NonNull String cpr, @NonNull final LocalDateTime samplingTime) {
            return cpr(cpr)
                .add(antigenResult(NEXT_IDENTIFIER.incrementAndGet(), samplingTime, samplingTime.plusHours(RANDOM_DATA.nextInt(4, 42))).positive().manufacturer(EU_APPROVED_MANUFACTURER).build())
                .build();
        }

        @NonNull
        public static LabResults positiveEuApprovedAntigen(@NonNull final LocalDateTime samplingTime) {
            return cpr(DEFAULT_CPR)
                .add(antigenResult(NEXT_IDENTIFIER.incrementAndGet(), samplingTime, samplingTime.plusHours(RANDOM_DATA.nextInt(4, 42))).positive().manufacturer(EU_APPROVED_MANUFACTURER).build())
                .build();
        }

        @NonNull
        public static LabResults positiveEuUnapprovedAntigen(@NonNull final LocalDateTime samplingTime) {
            return cpr(DEFAULT_CPR)
                .add(antigenResult(NEXT_IDENTIFIER.incrementAndGet(), samplingTime, samplingTime.plusHours(RANDOM_DATA.nextInt(4, 42))).positive().manufacturer(EU_UNAPPROVED_MANUFACTURER).build())
                .build();
        }

        @NonNull
        public static LabResults positiveAntigenBeforeNegativePcr(@NonNull final LocalDateTime samplingTime) {
            return cpr(DEFAULT_CPR)
                .add(antigenResult(NEXT_IDENTIFIER.incrementAndGet(), samplingTime, samplingTime.plusHours(RANDOM_DATA.nextInt(4, 42))).positive().build())
                .add(pcrResult(NEXT_IDENTIFIER.incrementAndGet(), samplingTime.plusNanos(1), samplingTime.plusHours(RANDOM_DATA.nextInt(4, 42))).negative().build())
                .build();
        }

    }

    public static final class Result {
        private Result() {
        }

        private static final AtomicLong NEXT_IDENTIFIER = new AtomicLong();

        @NonNull
        public static LabResult positivePcr(@NonNull final LocalDateTime samplingTime) {
            return LabResultBuilder.pcrResult(NEXT_IDENTIFIER.incrementAndGet(), samplingTime, samplingTime.plusHours(RANDOM_DATA.nextInt(4, 42))).positive().build();
        }

        @NonNull
        public static LabResult negativePcr(@NonNull final LocalDateTime samplingTime) {
            return LabResultBuilder.pcrResult(NEXT_IDENTIFIER.incrementAndGet(), samplingTime, samplingTime.plusHours(RANDOM_DATA.nextInt(4, 42))).negative().build();
        }

        @NonNull
        public static LabResult negativeAntigen(@NonNull final LocalDateTime samplingTime) {
            return negativeEuApprovedAntigen(samplingTime);
        }

        @NonNull
        public static LabResult negativeEuApprovedAntigen(@NonNull final LocalDateTime samplingTime) {
            return LabResultBuilder.antigenResult(NEXT_IDENTIFIER.incrementAndGet(), samplingTime, samplingTime.plusHours(RANDOM_DATA.nextInt(4, 42))).negative().manufacturer(EU_APPROVED_MANUFACTURER).build();
        }

        @NonNull
        public static LabResult negativeEuUnapprovedAntigen(@NonNull final LocalDateTime samplingTime) {
            return LabResultBuilder.antigenResult(NEXT_IDENTIFIER.incrementAndGet(), samplingTime, samplingTime.plusHours(RANDOM_DATA.nextInt(4, 42))).negative().manufacturer(EU_UNAPPROVED_MANUFACTURER).build();
        }

        @NonNull
        public static LabResult positiveAntigen(@NonNull final LocalDateTime samplingTime) {
            return positiveEuApprovedAntigen(samplingTime);
        }

        @NonNull
        public static LabResult positiveEuApprovedAntigen(@NonNull final LocalDateTime samplingTime) {
            return LabResultBuilder.antigenResult(NEXT_IDENTIFIER.incrementAndGet(), samplingTime, samplingTime.plusHours(RANDOM_DATA.nextInt(4, 42))).positive().manufacturer(EU_APPROVED_MANUFACTURER).build();
        }

        @NonNull
        public static LabResult positiveEuUnapprovedAntigen(@NonNull final LocalDateTime samplingTime) {
            return LabResultBuilder.antigenResult(NEXT_IDENTIFIER.incrementAndGet(), samplingTime, samplingTime.plusHours(RANDOM_DATA.nextInt(4, 42))).positive().manufacturer(EU_UNAPPROVED_MANUFACTURER).build();
        }
    }
}
