/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.dto.vaccination;

import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import java.time.LocalDateTime;
import java.util.Collection;

/**
 * Covid 19 Vaccination data
 */
public interface Vaccination {

    /**
     * Gets the cpr-number of the citizen whom this vaccination data belongs to
     *
     * @return the cpr-number
     */
    @NonNull
    String cpr();

    /**
     * Get the date and time of the most recent effectuated vaccination, that is part of an completed vaccination plan.
     * @return the date and time, <code>null</code> if no completed vaccination plan exists.
     */
    @Nullable
    LocalDateTime completed();

    /**
     * Gets a collection of effectuated (administered) vaccinations in a COVID-19 vaccination program, ordered by effectuation time descending
     *
     * @return the effectuated vaccinations in the program
     */
    @NonNull
    Collection<EffectuatedVaccination> vaccinations();

    /**
     * Gets a collection of effectuated (administered) vaccinations given after a completed COVID-19 vaccination plan, ordered by effectuation time descending
     */
    @NonNull
    Collection<EffectuatedVaccination> additionalVaccinations();

    /**
     * Gets the most recently effectuated vaccination part of a vaccination plan, if any
     */
    @Nullable
    EffectuatedVaccination mostRecentlyEffectuated();

    @Nullable
    VaccinationScheduleType scheduleType();
}
