/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.rules;

import dk.sundhedsdatastyrelsen.cbs.Objects;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.RecoveryLabResult;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.RecoveryPassportEvaluation;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import java.time.LocalDateTime;
import java.time.temporal.TemporalAmount;

import static dk.sundhedsdatastyrelsen.cbs.rules.StableLocalDateTime.*;

public class EuRecoveryPassportEvaluator implements PassportEvaluator<RecoveryPassportEvaluation> {
    @NonNull
    private final TemporalAmount minimumPeriod;
    @NonNull
    private final TemporalAmount maximumPeriod;

    public EuRecoveryPassportEvaluator(
        @NonNull final TemporalAmount minimumPeriod,
        @NonNull final TemporalAmount maximumPeriod) {
        this.minimumPeriod = minimumPeriod;
        this.maximumPeriod = maximumPeriod;
    }

    @Override
    @NonNull
    public RecoveryPassportEvaluation evaluate(
        @NonNull final Vaccination vaccination,
        @NonNull final LabResults results,
        @NonNull final LocalDateTime now) {
        try {
            setNow(now);
            final LabResult.PositivePcrResult newestPositivePcrNotAfterMinimumPeriod = findNewestPositivePcrNotAfterMinimumPeriod(results);
            return Objects.firstNonNullOrDefaultSupplier(
                () -> grant(newestPositivePcrNotAfterMinimumPeriod),
                () -> denyForNoPositivePcrResult(results.mostRecentPositivePcrResult()),
                () -> denyForRecoveryAfterEndOfMaximumPeriod(newestPositivePcrNotAfterMinimumPeriod),
                () -> denyForRecoveryBeforeEndOfMinimumPeriod(newestPositivePcrNotAfterMinimumPeriod, results)
            );
        } finally {
            clearNow();
        }
    }

    @Nullable
    private RecoveryPassportEvaluation denyForRecoveryAfterEndOfMaximumPeriod(@Nullable final LabResult.PositivePcrResult result) {
        if (result == null) return null;
        final LocalDateTime endOfRecovery = endOfRecovery(result);
        return now().isBefore(endOfRecovery)
            ? null
            : RecoveryPassportEvaluation.denyIndefinitely("maximum period after positive pcr has passed");
    }

    @Nullable
    @SuppressWarnings("java:S135") //SonarLint accepts at most 1 continue or break statement in a for loop, but we ignore this here for purpose of better clarity and expressiveness
    private RecoveryPassportEvaluation denyForRecoveryBeforeEndOfMinimumPeriod(@Nullable final LabResult.PositivePcrResult newestPositivePcrNotAfterMinimumPeriod, @NonNull final LabResults results) {
        if (newestPositivePcrNotAfterMinimumPeriod != null) return null;

        //find oldest positive PCR to base the 'expiresAfter' of the denied evaluation on
        LabResult.PositivePcrResult oldestPositivePcrBeforeMinimumPeriod = null;
        for (LabResult.PcrResult pcrResult : results.pcrResults()) { //iterates in descending order (newest first)
            if (pcrResult.negative()) continue;
            if (pcrResult.samplingDateTime().isBefore(now().minus(minimumPeriod))) break; //no more results from active infection
            oldestPositivePcrBeforeMinimumPeriod = (LabResult.PositivePcrResult) pcrResult;
        }
        if (oldestPositivePcrBeforeMinimumPeriod == null) throw new IllegalStateException("No positive PCR result found that is before end of minimum period");
        return RecoveryPassportEvaluation.deny(startOfRecovery(oldestPositivePcrBeforeMinimumPeriod), "minimum period after positive pcr has not passed");
    }

    @Nullable
    private RecoveryPassportEvaluation denyForNoPositivePcrResult(@Nullable final LabResult.PositivePcrResult result) {
        return result == null
            ? RecoveryPassportEvaluation.denyIndefinitely("No positive pcr results")
            : null;
    }

    @NonNull
    private RecoveryPassportEvaluation grant(final LabResult.PositivePcrResult result) {
        final LocalDateTime startOfRecovery = startOfRecovery(result);
        final LocalDateTime endOfRecovery = endOfRecovery(result);
        return RecoveryPassportEvaluation.grant(
            endOfRecovery,
            RecoveryLabResult.of(
                result,
                startOfRecovery,
                endOfRecovery));
    }

    @NonNull
    private LocalDateTime startOfRecovery(@NonNull LabResult.PositivePcrResult result) {
        return result.samplingDateTime().plus(minimumPeriod);
    }

    @NonNull
    private LocalDateTime endOfRecovery(@NonNull LabResult.PositivePcrResult result) {
        return result.samplingDateTime().plus(maximumPeriod);
    }

    @Nullable
    private LabResult.PositivePcrResult findNewestPositivePcrNotAfterMinimumPeriod(@NonNull LabResults labResults) {
        for (LabResult.PcrResult pcrResult : labResults.pcrResults()) { //iterates in descending order (newest first)
            if (pcrResult.positive() && notAfterMinimumPeriod(pcrResult))
                return (LabResult.PositivePcrResult) pcrResult;
        }
        return null;
    }

    private boolean notAfterMinimumPeriod(@NonNull LabResult.PcrResult pcrResult) {
        return !pcrResult.samplingDateTime().isAfter(now().minus(minimumPeriod));
    }

}
