/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.rules;

import dk.sundhedsdatastyrelsen.cbs.Objects;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.RecoveryLabResult;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.RecoveryPassportEvaluation;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import java.time.LocalDateTime;
import java.time.temporal.TemporalAmount;

import static dk.sundhedsdatastyrelsen.cbs.rules.StableLocalDateTime.*;
import static java.util.Optional.ofNullable;

public class RecoveryPassportEvaluator implements PassportEvaluator<RecoveryPassportEvaluation> {

    @NonNull private final TemporalAmount minimumPeriod;
    @NonNull private final TemporalAmount maximumPeriod;
    @NonNull private final TemporalAmount inactiveInfectionPeriod;

    public RecoveryPassportEvaluator(
        @NonNull final TemporalAmount minimumPeriod,
        final boolean keepTimeOnMinimumPeriod,
        @NonNull final TemporalAmount maximumPeriod,
        final boolean keepTimeOnMaximumPeriod,
        @NonNull final TemporalAmount inactiveInfectionPeriod,
        final boolean keepTimeOnInactiveInfectionPeriod) {
        this.minimumPeriod = minimumPeriod;
        this.maximumPeriod = maximumPeriod;
        this.inactiveInfectionPeriod = inactiveInfectionPeriod;
    }

    @Override
    @NonNull
    public RecoveryPassportEvaluation evaluate(
        @NonNull final Vaccination vaccination,
        @NonNull final LabResults results,
        @NonNull final LocalDateTime now) {
        try {
            setNow(now);
            final LabResult.PositivePcrResult firstPositivePcrResultOfLatestEpisodeOfInfection = findFirstPositivePcrResultOfLatestEpisodeOfInfection(results);
            return Objects.firstNonNullOrDefaultSupplier(
                () -> grant(firstPositivePcrResultOfLatestEpisodeOfInfection),
                // Reasons to grant or deny a recovery passport follow below,
                // if none are met return above
                () -> denyForNoPositivePcrResult(firstPositivePcrResultOfLatestEpisodeOfInfection),
                () -> denyForPositiveAntigenAfterInactiveInfectionPeriod(firstPositivePcrResultOfLatestEpisodeOfInfection, results),
                () -> denyForRecoveryBeforeEndOfMinimumPeriod(firstPositivePcrResultOfLatestEpisodeOfInfection),
                () -> denyForRecoveryAfterEndOfMaximumPeriod(firstPositivePcrResultOfLatestEpisodeOfInfection));
        } finally {
            clearNow();
        }
    }

    @Nullable
    private RecoveryPassportEvaluation denyForRecoveryAfterEndOfMaximumPeriod(@NonNull final LabResult.PositivePcrResult result) {
        final LocalDateTime endOfRecovery = endOfRecovery(result);
        return now().isBefore(endOfRecovery)
            ? null
            : RecoveryPassportEvaluation.denyIndefinitely("maximum period after positive pcr has passed");
    }

    @Nullable
    private RecoveryPassportEvaluation denyForRecoveryBeforeEndOfMinimumPeriod(@NonNull final LabResult.PositivePcrResult result) {
        final LocalDateTime startOfRecovery = startOfRecovery(result);
        return now().isBefore(startOfRecovery)
            ? RecoveryPassportEvaluation.deny(startOfRecovery, "minimum period after positive pcr has not passed")
            : null;
    }

    @Nullable
    private RecoveryPassportEvaluation denyForNoPositivePcrResult(@Nullable final LabResult.PositivePcrResult result) {
        return result == null
            ? RecoveryPassportEvaluation.denyIndefinitely("No positive pcr results")
            : null;
    }

    @NonNull
    private RecoveryPassportEvaluation grant(final LabResult.PositivePcrResult result) {
        final LocalDateTime startOfRecovery = startOfRecovery(result);
        final LocalDateTime endOfRecovery = endOfRecovery(result);
        return RecoveryPassportEvaluation.grant(
            endOfRecovery,
            RecoveryLabResult.of(
                result,
                startOfRecovery,
                endOfRecovery));
    }

    @NonNull
    private LocalDateTime startOfRecovery(LabResult.PositivePcrResult result) {
        return result.samplingDateTime().plus(minimumPeriod);
    }

    @NonNull
    private LocalDateTime endOfRecovery(LabResult.PositivePcrResult result) {
        return result.samplingDateTime().plus(maximumPeriod);
    }

    @Nullable
    private RecoveryPassportEvaluation denyForPositiveAntigenAfterInactiveInfectionPeriod(
        @NonNull final LabResult.PositivePcrResult result,
        @NonNull LabResults results) {
        final LabResult.PositiveAntigenResult mostRecentPositiveAntigenResult = results.mostRecentPositiveAntigenResult();
        final LocalDateTime timeOfMostRecentPositiveAntigenResult = ofNullable(mostRecentPositiveAntigenResult).map(LabResult::samplingDateTime).orElse(LocalDateTime.MIN);
        final LocalDateTime endOfInactiveInfectionPeriod = endOfInactiveInfectionPeriod(result);
        final LabResult.NegativePcrResult mostRecentNegativePcrResult = results.mostRecentNegativePcrResult();
        final LocalDateTime timeOfMostRecentNegativePcrResult = ofNullable(mostRecentNegativePcrResult).map(LabResult::samplingDateTime).orElse(LocalDateTime.MIN);

        if(timeOfMostRecentPositiveAntigenResult.isBefore(endOfInactiveInfectionPeriod)) return null;

        return timeOfMostRecentPositiveAntigenResult.isAfter(timeOfMostRecentNegativePcrResult)
            ? RecoveryPassportEvaluation.denyIndefinitely("No negative pcr result after positive antigen result")
            : null;

    }

    @Nullable
    @SuppressWarnings("java:S135")
    private LabResult.PositivePcrResult findFirstPositivePcrResultOfLatestEpisodeOfInfection(LabResults labResults) {
        LabResult.PositivePcrResult firstPositivePcrResultOfLatestEpisodeOfInfection = null;
        LocalDateTime endOfCurrentInactiveInfectionPeriod = LocalDateTime.MIN;
        for (final LabResult result : labResults.results()) {
            if(notPositivePcrResult(result)) continue;
            if(result.samplingDateTime().isBefore(endOfCurrentInactiveInfectionPeriod)) continue;

            endOfCurrentInactiveInfectionPeriod = endOfInactiveInfectionPeriod(result);
            firstPositivePcrResultOfLatestEpisodeOfInfection = (LabResult.PositivePcrResult) result;
        }
        return firstPositivePcrResultOfLatestEpisodeOfInfection;
    }

    @NonNull
    private LocalDateTime endOfInactiveInfectionPeriod(@NonNull final LabResult result) {
        return result.samplingDateTime().plus(inactiveInfectionPeriod);
    }

    private static boolean notPositivePcrResult(LabResult result) {
        return !(result instanceof LabResult.PositivePcrResult);
    }
}
