/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.rules;

import dk.sundhedsdatastyrelsen.cbs.Objects;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult.AntigenResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult.PcrResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.EffectuatedVaccination;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportDeniedEvaluation;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluation;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportGrantedEvaluation;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import javax.validation.constraints.NotEmpty;
import java.time.LocalDateTime;
import java.time.temporal.TemporalAmount;
import java.util.Collection;
import java.util.Comparator;
import java.util.Optional;

import static dk.sundhedsdatastyrelsen.cbs.Objects.firstNonNullOrDefaultSupplier;
import static dk.sundhedsdatastyrelsen.cbs.rules.LocalDateTimes.resolutionEndOfDay;
import static dk.sundhedsdatastyrelsen.cbs.rules.LocalDateTimes.resolutionStartOfDay;
import static dk.sundhedsdatastyrelsen.cbs.rules.StableLocalDateTime.*;
import static dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluation.*;

public final class RevaccinationPassportEvaluator implements PassportEvaluator<PassportEvaluation> {
    @NonNull
    private static final String MINIMUM_DURATION_NOT_PASSED = "minimum duration not passed";
    @NonNull
    private static final String REVACCINATION_EXPIRED = "re-vaccination expired";
    @NonNull
    private static final String POSITIVE_ANTIGEN_RESULT_WITHIN_SUSPENSION_PERIOD_WITHOUT_SUBSEQUENT_NEGATIVE_PCR_RESULT = "positive antigen result within suspension period without subsequent negative pcr result";
    @NonNull
    private static final String POSITIVE_PCR_RESULT_WITHIN_SUSPENSION_PERIOD = "positive pcr result within suspension period";
    @NonNull
    private static final String NO_REVACCINATIONS = "no re-vaccinations";
    @NonNull
    private final TemporalAmount minimumPeriodAfterRevaccination;
    private final boolean keepTimeOnMinimumPeriodAfterRevaccination;
    @NonNull
    private final TemporalAmount maximumPeriodAfterRevaccination;
    private final boolean keepTimeOnMaximumPeriodAfterRevaccination;
    @NonNull
    private final TemporalAmount suspensionPeriodAfterRevaccination;
    private final boolean keepTimeOnSuspensionPeriodAfterRevaccination;


    public RevaccinationPassportEvaluator(@NonNull TemporalAmount minimumPeriodAfterRevaccination,
                                          boolean keepTimeOnMinimumPeriodAfterRevaccination,
                                          @NonNull TemporalAmount maximumPeriodAfterRevaccination,
                                          boolean keepTimeOnMaximumPeriodAfterRevaccination,
                                          @NonNull TemporalAmount suspensionPeriodAfterRevaccination,
                                          boolean keepTimeOnSuspensionPeriodAfterRevaccination) {
        this.minimumPeriodAfterRevaccination = minimumPeriodAfterRevaccination;
        this.keepTimeOnMinimumPeriodAfterRevaccination = keepTimeOnMinimumPeriodAfterRevaccination;
        this.maximumPeriodAfterRevaccination = maximumPeriodAfterRevaccination;
        this.keepTimeOnMaximumPeriodAfterRevaccination = keepTimeOnMaximumPeriodAfterRevaccination;
        this.suspensionPeriodAfterRevaccination = suspensionPeriodAfterRevaccination;
        this.keepTimeOnSuspensionPeriodAfterRevaccination = keepTimeOnSuspensionPeriodAfterRevaccination;
    }

    /**
     * Evaluates re-vaccinations and lab results in the context of re-vaccination passport rules.
     *
     * @return the result of the evaluation.
     */
    @NonNull
    @Override
    public PassportEvaluation evaluate(@NonNull final Vaccination vaccination, @NonNull final LabResults results, @NonNull final LocalDateTime now) {
        try {
            setNow(now);
            return firstNonNullOrDefaultSupplier(
                () -> grantForMostRecentValidRevaccination(vaccination.additionalVaccinations()),
                // Reasons not to grant a re-vaccination passport follow below,
                // if none are met return above
                () -> denyForNoRevaccinations(vaccination),
                () -> denyForNoValidRevaccinations(vaccination.additionalVaccinations()),
                () -> denyForSuspendedDueToPositiveLabResult(results));
        } finally {
            clearNow();
        }
    }

    @NonNull
    private PassportGrantedEvaluation grantForMostRecentValidRevaccination(@NotEmpty final Collection<EffectuatedVaccination> reVaccinations) {
        final EffectuatedVaccination mostRecentValidRevaccinations = reVaccinations.stream()
            .filter(v -> !now().isBefore(reVaccinationValidFrom(v.effectuated())) && now().isBefore(reVaccinationValidUntil(v.effectuated())))
            .findFirst()
            .orElseThrow(() -> new IllegalArgumentException("cannot grant passport because there are no valid re-vaccinations (this should not happen because passport should already have been denied at this point)"));
        return grant(reVaccinationValidUntil(mostRecentValidRevaccinations.effectuated()));
    }

    /**
     * Evaluates if a passport should be denied due to no effectuated re-vaccinations
     *
     * @param vaccination the vaccinations
     * @return the denied evaluation if the passport was denied, else <code>null</code>.
     */
    @Nullable
    private PassportDeniedEvaluation denyForNoRevaccinations(@NonNull final Vaccination vaccination) {
        return vaccination.additionalVaccinations().isEmpty() ? denyIndefinitely(NO_REVACCINATIONS) : null;
    }

    @Nullable
    private PassportDeniedEvaluation denyForNoValidRevaccinations(@NotEmpty final Collection<EffectuatedVaccination> reVaccinations) {
        final boolean anyValidRevaccinations = reVaccinations.stream().anyMatch(v -> !now().isBefore(reVaccinationValidFrom(v.effectuated())) && now().isBefore(reVaccinationValidUntil(v.effectuated())));
        if (anyValidRevaccinations) return null;

        return findEffectuationTimeOfNextValidRevaccination(reVaccinations)
            .map(effectuated -> deny(reVaccinationValidFrom(effectuated), MINIMUM_DURATION_NOT_PASSED))
            .orElse(denyIndefinitely(REVACCINATION_EXPIRED));
    }

    @NonNull
    private Optional<LocalDateTime> findEffectuationTimeOfNextValidRevaccination(@NotEmpty final Collection<EffectuatedVaccination> reVaccinations) {
        return reVaccinations.stream()
            .map(EffectuatedVaccination::effectuated)
            .filter(effectuated -> !now().isAfter(reVaccinationValidFrom(effectuated)))
            .min(Comparator.naturalOrder());
    }

    /**
     * Evaluates if a passport should be denied temporarily due to positive lab results.
     *
     * @param results the lab results to look among.
     * @return the denied evaluation if the passport was denied, else <code>null</code>.
     */
    @Nullable
    private PassportDeniedEvaluation denyForSuspendedDueToPositiveLabResult(@NonNull final LabResults results) {
        return Objects.firstNonNullSupplier(
            () -> denyForPositivePcrResultWithinSuspensionPeriod(results),
            () -> denyForPositiveAntigenResultWithinSuspensionPeriodThatHasNotBeenFollowedByANegativePcrResult(results));
    }

    /**
     * Evaluates if a passport should be denied temporarily due to positive antigen result within the suspension period
     * that has not been set aside by a subsequent negative pcr result.
     *
     * @param results the lab results to look among.
     * @return the denied evaluation if the passport was denied, else <code>null</code>.
     */
    @Nullable
    private PassportDeniedEvaluation denyForPositiveAntigenResultWithinSuspensionPeriodThatHasNotBeenFollowedByANegativePcrResult(@NonNull final LabResults results) {
        final LabResult.PositiveAntigenResult mostRecentPositiveAntigenResult = results.mostRecentPositiveAntigenResult();
        final LabResult.NegativePcrResult mostRecentNegativePcrResult = results.mostRecentNegativePcrResult();

        if (mostRecentPositiveAntigenResult == null) return null;
        if (resultOutsideSuspensionPeriod(mostRecentPositiveAntigenResult)) return null;
        if (pcrResultAfterAntigenResult(mostRecentNegativePcrResult, mostRecentPositiveAntigenResult)) return null;

        return deny(positiveResultExpiresSuspensionPeriod(mostRecentPositiveAntigenResult), POSITIVE_ANTIGEN_RESULT_WITHIN_SUSPENSION_PERIOD_WITHOUT_SUBSEQUENT_NEGATIVE_PCR_RESULT);
    }

    /**
     * Evaluates if a passport should be denied because there is a positive pcr result within the suspension period.
     *
     * @param results the lab results to look among.
     * @return the denied evaluation if the passport was denied, else <code>null</code>.
     */
    @Nullable
    private PassportDeniedEvaluation denyForPositivePcrResultWithinSuspensionPeriod(@NonNull final LabResults results) {
        final LabResult.PositivePcrResult mostRecentPositivePcrResult = results.mostRecentPositivePcrResult();

        if (mostRecentPositivePcrResult == null) return null;
        if (resultOutsideSuspensionPeriod(mostRecentPositivePcrResult)) return null;

        return deny(positiveResultExpiresSuspensionPeriod(mostRecentPositivePcrResult), POSITIVE_PCR_RESULT_WITHIN_SUSPENSION_PERIOD);
    }

    /**
     * Evaluates whether a result is outside the suspension period.
     *
     * @param result the lab result to evaluate.
     * @return <code>false</code> if the result is within the suspension period, else <code>true</code>.
     */
    private boolean resultOutsideSuspensionPeriod(@NonNull final LabResult result) {
        final LocalDateTime samplingDateTime = result.samplingDateTime();
        return !samplingDateTime.isAfter(startOfSuspensionPeriod());
    }

    /**
     * Evaluates whether an antigen result is before a pcr result.
     *
     * @param pcrResult     the most recent pcr result, can be <code>null</code>.
     * @param antigenResult the most recent positive antigen result.
     * @return <code>true</code> if the pcr result is not <code>null</code> and after the antigen result.
     */
    private static boolean pcrResultAfterAntigenResult(
        @Nullable final PcrResult pcrResult,
        @NonNull final AntigenResult antigenResult) {
        if (pcrResult == null) return false;

        final LocalDateTime dateTimeOfMostRecentPositiveAntigenResult = antigenResult.samplingDateTime();
        final LocalDateTime dateTimeOfMostRecentNegativePcrResult = pcrResult.samplingDateTime();
        return dateTimeOfMostRecentNegativePcrResult.isAfter(dateTimeOfMostRecentPositiveAntigenResult);
    }

    /**
     * Calculates the earliest date from which an effectuated re-vaccination can grant a passport
     *
     * @param effectuated the date and time of the effectuated vaccination
     * @return the date
     */
    @NonNull
    private LocalDateTime reVaccinationValidFrom(@NonNull final LocalDateTime effectuated) {
        return resolutionStartOfDay(effectuated.plus(minimumPeriodAfterRevaccination), keepTimeOnMinimumPeriodAfterRevaccination);
    }

    /**
     * Calculates the latest date and time from which an effectuated re-vaccination can grant a passport
     *
     * @param mostRecentlyEffectuated the date and time of the most recently effectuated vaccination
     * @return the date and time
     */
    @NonNull
    private LocalDateTime reVaccinationValidUntil(@NonNull final LocalDateTime mostRecentlyEffectuated) {
        return resolutionEndOfDay(mostRecentlyEffectuated.plus(maximumPeriodAfterRevaccination), keepTimeOnMaximumPeriodAfterRevaccination);
    }

    /**
     * Calculates the earliest day of a possible suspension of an effectuated re-vaccination due to positive results.
     *
     * @return the date
     */
    @NonNull
    private LocalDateTime startOfSuspensionPeriod() {
        return resolutionStartOfDay(now().minus(suspensionPeriodAfterRevaccination), keepTimeOnSuspensionPeriodAfterRevaccination);
    }

    /**
     * Calculates the last day of a suspension of an effectuated re-vaccination due to positive results.
     *
     * @param positiveResult the positive result
     * @return the date
     */
    @NonNull
    private LocalDateTime positiveResultExpiresSuspensionPeriod(@NonNull final LabResult.PositiveResult positiveResult) {
        return resolutionEndOfDay(positiveResult.samplingDateTime().plus(suspensionPeriodAfterRevaccination), keepTimeOnSuspensionPeriodAfterRevaccination);
    }

}
