/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.rules;

import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult.PcrResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluation;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import java.time.Duration;
import java.time.LocalDateTime;
import java.time.temporal.TemporalAmount;
import java.util.Comparator;
import java.util.Optional;

import static dk.sundhedsdatastyrelsen.cbs.Objects.firstNonNullOrDefaultSupplier;
import static dk.sundhedsdatastyrelsen.cbs.rules.LocalDateTimes.resolutionEndOfDay;
import static dk.sundhedsdatastyrelsen.cbs.rules.LocalDateTimes.resolutionStartOfDay;
import static dk.sundhedsdatastyrelsen.cbs.rules.StableLocalDateTime.clearNow;
import static dk.sundhedsdatastyrelsen.cbs.rules.StableLocalDateTime.now;
import static dk.sundhedsdatastyrelsen.cbs.rules.StableLocalDateTime.setNow;
import static dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluation.deny;
import static dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluation.denyIndefinitely;
import static dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluation.grant;

@SuppressWarnings("unused")
public final class TestPassportEvaluator implements PassportEvaluator<PassportEvaluation> {

    private static final String NO_NEGATIVE_TEST_RESULTS = "no negative test results";
    private static final String NEGATIVE_TEST_EXPIRED = "negative test expired";
    private static final String POSITIVE_PCR_TEST_WITHIN_SUSPENSION_PERIOD = "positive pcr result within suspension period";
    private static final String MOST_RECENT_RESULT_IS_POSITIVE = "most recent result is positive";
    private static final String NEGATIVE_TEST_PENDING = "negative test not yet valid";

    @NonNull
    private final TemporalAmount minimumPeriod;
    @NonNull
    private final TemporalAmount maximumPeriodPcr;
    @NonNull
    private final TemporalAmount maximumPeriodAg;
    @NonNull
    private final TemporalAmount suspensionPeriod;
    private final boolean keepTimeOnMinimumPeriod;
    private final boolean keepTimeOnMaximumPeriod;
    private final boolean keepTimeOnSuspensionPeriod;

    public TestPassportEvaluator(
        @NonNull final TemporalAmount minimumPeriod, final boolean keepTimeOnMinimumPeriod,
        @NonNull final TemporalAmount maximumPeriodPcr, @NonNull final TemporalAmount maximumPeriodAg, final boolean keepTimeOnMaximumPeriod,
        @NonNull final TemporalAmount suspensionPeriod, final boolean keepTimeOnSuspensionPeriod) {
        this.minimumPeriod = minimumPeriod;
        this.maximumPeriodPcr = maximumPeriodPcr;
        this.maximumPeriodAg = maximumPeriodAg;
        this.suspensionPeriod = suspensionPeriod;
        this.keepTimeOnMinimumPeriod = keepTimeOnMinimumPeriod;
        this.keepTimeOnMaximumPeriod = keepTimeOnMaximumPeriod;
        this.keepTimeOnSuspensionPeriod = keepTimeOnSuspensionPeriod;
    }

    /**
     * Evaluates lab results in the context of test passport rules.
     *
     * @return the result of the evaluation.
     */
    @NonNull
    @Override
    public PassportEvaluation evaluate(@NonNull final Vaccination vaccination, @NonNull final LabResults results, @NonNull final LocalDateTime now) {
        try {
            setNow(now);
            return firstNonNullOrDefaultSupplier(
                () -> denyIndefinitely(NO_NEGATIVE_TEST_RESULTS),
                // Reasons to grant or deny a test passport follow below,
                // if none are met return above
                () -> denyForMostRecentTestIsPositive(results),
                () -> grantOrDenyForNegativeTestWithinValidityPeriod(results));
        } finally {
            clearNow();
        }
    }

    @Nullable
    private PassportEvaluation denyForMostRecentTestIsPositive(@NonNull final LabResults results) {
        final LabResult mostRecentResult = results.mostRecentResult();
        if (mostRecentResult == null) return null;
        if (mostRecentResult.negative()) return null;

        return denyIndefinitely(MOST_RECENT_RESULT_IS_POSITIVE);
    }

    @Nullable
    private PassportEvaluation grantOrDenyForNegativeTestWithinValidityPeriod(@NonNull final LabResults results) {
        final LabResult.NegativePcrResult mostRecentValidNegativePcrResult = getMostRecentNegativePcrResultInValidityPeriod(results);
        final LabResult.NegativeAntigenResult mostRecentValidNegativeAntigenResult = getMostRecentNegativeAntigenResultInValidityPeriod(results);

        if (mostRecentValidNegativePcrResult == null && mostRecentValidNegativeAntigenResult == null) {
            if (results.results().stream().noneMatch(LabResult::negative)) {
                return null;
            }

            return findSamplingTimeOfNextValidLabResult(results)
                .map(date -> deny(expirationOfPendingTestResult(date), NEGATIVE_TEST_PENDING))
                .orElse(denyIndefinitely(NEGATIVE_TEST_EXPIRED));
        }

        if (mostRecentValidNegativeAntigenResult != null && antigenIsValidLongerThanPcr(mostRecentValidNegativeAntigenResult, mostRecentValidNegativePcrResult)) {
            return denyOrGrantForNegativeAntigenResultSuspendedByPositivePcrResult(mostRecentValidNegativeAntigenResult, results.mostRecentPositivePcrResult());
        } else {
            return grant(validUntil(mostRecentValidNegativePcrResult));
        }
    }

    @NonNull
    private Optional<LocalDateTime> findSamplingTimeOfNextValidLabResult(@NonNull final LabResults results) {
        return results.results().stream()
            .filter(LabResult::negative)
            .map(LabResult.NegativeResult.class::cast)
            .filter(this::isPending)
            .map(LabResult::samplingDateTime)
            .min(Comparator.naturalOrder());
    }

    private LocalDateTime expirationOfPendingTestResult(@NonNull final LocalDateTime sampleTime) {
        final Duration durationUntilValid = Duration.between(resolutionEndOfDay(now().minus(minimumPeriod), keepTimeOnMinimumPeriod), sampleTime);
        return now().plus(durationUntilValid);
    }

    private LabResult.NegativeAntigenResult getMostRecentNegativeAntigenResultInValidityPeriod(@NonNull final LabResults results) {
        return results.antigenResults().stream()
            .filter(LabResult::negative)
            .map(LabResult.NegativeAntigenResult.class::cast)
            .filter(this::negativeResultIsWithinValidityPeriod)
            .findFirst()
            .orElse(null);
    }

    private LabResult.NegativePcrResult getMostRecentNegativePcrResultInValidityPeriod(@NonNull final LabResults results) {
        return results.pcrResults().stream()
            .filter(LabResult::negative)
            .map(LabResult.NegativePcrResult.class::cast)
            .filter(this::negativeResultIsWithinValidityPeriod)
            .findFirst()
            .orElse(null);
    }

    private boolean antigenIsValidLongerThanPcr(@NonNull final LabResult.NegativeAntigenResult negativeAntigenResult, @Nullable final LabResult.NegativePcrResult negativePcrResult) {
        return negativePcrResult == null || validUntil(negativeAntigenResult).isAfter(validUntil(negativePcrResult));
    }

    @NonNull
    private PassportEvaluation denyOrGrantForNegativeAntigenResultSuspendedByPositivePcrResult(
        @NonNull final LabResult.NegativeAntigenResult mostRecentAntigenResult,
        @Nullable final LabResult.PositivePcrResult mostRecentPositivePcrResult) {

        if (mostRecentPositivePcrResult == null) return grant(validUntil(mostRecentAntigenResult));
        if (positivePcrResultWithinSuspensionPeriod(mostRecentPositivePcrResult)) {
            return deny(
                endOfSuspensionPeriod(mostRecentPositivePcrResult),
                POSITIVE_PCR_TEST_WITHIN_SUSPENSION_PERIOD);
        }

        return grant(validUntil(mostRecentAntigenResult));
    }

    private boolean negativeResultIsWithinValidityPeriod(@NonNull final LabResult.NegativePcrResult negativeLabResult) {
        final boolean pending = isPending(negativeLabResult);
        final boolean expired = isExpired(negativeLabResult);
        return !expired && !pending;
    }

    private boolean isExpired(@NonNull final LabResult.NegativePcrResult negativeLabResult) {
        final LocalDateTime startOfValidityPeriod = resolutionStartOfDay(now().minus(maximumPeriodPcr), keepTimeOnMaximumPeriod);
        return !startOfValidityPeriod.isBefore(negativeLabResult.samplingDateTime());
    }

    private boolean negativeResultIsWithinValidityPeriod(@NonNull final LabResult.NegativeAntigenResult negativeLabResult) {
        final boolean pending = isPending(negativeLabResult);
        final boolean expired = isExpired(negativeLabResult);
        return !expired && !pending;
    }

    private boolean isExpired(@NonNull final LabResult.NegativeAntigenResult negativeAntigenResult) {
        final LocalDateTime startOfValidityPeriod = resolutionStartOfDay(now().minus(maximumPeriodAg), keepTimeOnMaximumPeriod);
        return !startOfValidityPeriod.isBefore(negativeAntigenResult.samplingDateTime());
    }

    private boolean isPending(@NonNull final LabResult.NegativeResult negativeLabResult) {
        final LocalDateTime endOfValidityPeriod = resolutionEndOfDay(now().minus(minimumPeriod), keepTimeOnMinimumPeriod);
        return negativeLabResult.samplingDateTime().isAfter(endOfValidityPeriod);
    }

    private boolean negativeResultIsNotWithinValidityPeriod(@NonNull final LabResult.NegativePcrResult negativeLabResult) {
        return !negativeResultIsWithinValidityPeriod(negativeLabResult);
    }

    private boolean negativeResultIsNotWithinValidityPeriod(@NonNull final LabResult.NegativeAntigenResult negativeLabResult) {
        return !negativeResultIsWithinValidityPeriod(negativeLabResult);
    }

    private boolean positivePcrResultWithinSuspensionPeriod(@NonNull final LabResult.PositivePcrResult pcrResult) {
        final LocalDateTime startOfSuspensionPeriod = resolutionStartOfDay(now().minus(suspensionPeriod), keepTimeOnSuspensionPeriod);

        return pcrResult.samplingDateTime().isAfter(startOfSuspensionPeriod);
    }

    @NonNull
    private LocalDateTime endOfSuspensionPeriod(@NonNull final PcrResult positivePcrResult) {
        return resolutionEndOfDay(positivePcrResult.samplingDateTime().plus(suspensionPeriod), keepTimeOnSuspensionPeriod);
    }

    @NonNull
    private LocalDateTime validUntil(@NonNull final LabResult.NegativePcrResult negativePcrResult) {
        return resolutionEndOfDay(negativePcrResult.samplingDateTime().plus(maximumPeriodPcr), keepTimeOnMaximumPeriod);
    }

    @NonNull
    private LocalDateTime validUntil(@NonNull final LabResult.NegativeAntigenResult negativeAntigenResult) {
        return resolutionEndOfDay(negativeAntigenResult.samplingDateTime().plus(maximumPeriodAg), keepTimeOnMaximumPeriod);
    }
}
