/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.rules;

import com.google.common.collect.Range;
import com.google.common.collect.RangeSet;
import com.google.common.collect.TreeRangeSet;
import dk.sundhedsdatastyrelsen.cbs.Objects;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult.AntigenResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult.PcrResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.EffectuatedVaccination;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.VaccinationScheduleType;
import dk.sundhedsdatastyrelsen.cbs.rules.configuration.VaccinationPassportConfiguration;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportDeniedEvaluation;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluation;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import java.time.Duration;
import java.time.LocalDateTime;
import java.time.temporal.TemporalAmount;
import java.util.Collection;

import static com.google.common.collect.Range.closedOpen;
import static dk.sundhedsdatastyrelsen.cbs.Objects.firstNonNullOrDefaultSupplier;
import static dk.sundhedsdatastyrelsen.cbs.dto.vaccination.VaccinationScheduleType.MULTI_DOSE_VACCINATION_SCHEDULE;
import static dk.sundhedsdatastyrelsen.cbs.dto.vaccination.VaccinationScheduleType.SINGLE_DOSE_VACCINATION_SCHEDULE;
import static dk.sundhedsdatastyrelsen.cbs.rules.LocalDateTimes.resolutionEndOfDay;
import static dk.sundhedsdatastyrelsen.cbs.rules.LocalDateTimes.resolutionStartOfDay;
import static dk.sundhedsdatastyrelsen.cbs.rules.StableLocalDateTime.*;
import static dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluation.*;
import static java.util.Optional.ofNullable;

public final class VaccinationPassportEvaluator implements PassportEvaluator<PassportEvaluation> {

    @NonNull
    private static final String MINIMUM_DURATION_NOT_PASSED = "minimum duration not passed";
    @NonNull
    private static final String VACCINATION_EXPIRED = "vaccination expired";
    @NonNull
    private static final String POSITIVE_ANTIGEN_RESULT_WITHIN_SUSPENSION_PERIOD_WITHOUT_SUBSEQUENT_NEGATIVE_PCR_RESULT = "postive antigen result within suspension period without subsequent negative pcr result";
    @NonNull
    private static final String POSITIVE_PCR_RESULT_WITHIN_SUSPENSION_PERIOD = "positive pcr result within suspension period";
    @NonNull
    private static final String NO_VACCINATIONS = "no effectuated vaccinations";

    private final CprFormatBasedAgeCalculator ageCalculator;

    @NonNull
    private final TemporalAmount minimumPeriodAfterCompletedSingleVaccinationSchedule;
    @NonNull
    private final boolean keepTimeOnMinimumPeriodAfterCompletedSingleVaccinationSchedule;
    @NonNull
    private final TemporalAmount minimumPeriodAfterCompletedMultiVaccinationSchedule;
    @NonNull
    private final boolean keepTimeOnMinimumPeriodAfterCompletedMultiVaccinationSchedule;
    @NonNull
    private final TemporalAmount maximumPeriodAfterCompletedVaccinationSchedule;
    @NonNull
    private final boolean keepTimeOnMaximumPeriodAfterCompletedVaccinationSchedule;
    @NonNull
    private final TemporalAmount suspensionPeriodAfterCompletedVaccinationSchedule;
    @NonNull
    private final boolean keepTimeOnSuspensionPeriodAfterCompletedVaccinationSchedule;
    @NonNull
    private final TemporalAmount minimumPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule;
    @NonNull
    private final boolean keepTimeOnMinimumPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule;
    @NonNull
    private final TemporalAmount maximumPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule;
    @NonNull
    private final boolean keepTimeOnMaximumPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule;
    @NonNull
    private final TemporalAmount suspensionPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule;
    @NonNull
    private final boolean keepTimeOnSuspensionPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule;

    @NonNull
    private final TemporalAmount maximumPeriodAfterCompletedVaccinationScheduleUnder18;
    @NonNull
    private final boolean keepTimeOnMaximumPeriodAfterCompletedVaccinationScheduleUnder18;


    public VaccinationPassportEvaluator(
        @NonNull final VaccinationPassportConfiguration configuration, CprFormatBasedAgeCalculator ageCalculator) {
        minimumPeriodAfterCompletedSingleVaccinationSchedule = configuration.minimumPeriodAfterCompletedSingleVaccinationSchedule();
        keepTimeOnMinimumPeriodAfterCompletedSingleVaccinationSchedule = configuration.keepTimeOnMinimumPeriodAfterCompletedSingleVaccinationSchedule();
        minimumPeriodAfterCompletedMultiVaccinationSchedule = configuration.minimumPeriodAfterCompletedMultiVaccinationSchedule();
        keepTimeOnMinimumPeriodAfterCompletedMultiVaccinationSchedule = configuration.keepTimeOnMinimumPeriodAfterCompletedMultiVaccinationSchedule();
        maximumPeriodAfterCompletedVaccinationSchedule = configuration.maximumPeriodAfterCompletedVaccinationSchedule();
        keepTimeOnMaximumPeriodAfterCompletedVaccinationSchedule = configuration.keepTimeOnMaximumPeriodAfterCompletedVaccinationSchedule();
        suspensionPeriodAfterCompletedVaccinationSchedule = configuration.suspensionPeriodAfterCompletedVaccinationSchedule();
        keepTimeOnSuspensionPeriodAfterCompletedVaccinationSchedule = configuration.keepTimeOnSuspensionPeriodAfterCompletedVaccinationSchedule();
        minimumPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule = configuration.minimumPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule();
        keepTimeOnMinimumPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule = configuration.keepTimeOnMinimumPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule();
        maximumPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule = configuration.maximumPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule();
        keepTimeOnMaximumPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule = configuration.keepTimeOnMaximumPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule();
        suspensionPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule = configuration.suspensionPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule();
        keepTimeOnSuspensionPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule = configuration.keepTimeOnSuspensionPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule();
        maximumPeriodAfterCompletedVaccinationScheduleUnder18 = configuration.maximumPeriodAfterCompletedVaccinationScheduleUnder18();
        keepTimeOnMaximumPeriodAfterCompletedVaccinationScheduleUnder18 = configuration.keepTimeOnMaximumPeriodAfterCompletedVaccinationScheduleUnder18();
        this.ageCalculator = ageCalculator;
    }

    /**
     * Evaluates vaccinations and lab results in the context of vaccination passport rules.
     *
     * @return the result of the evaluation.
     */
    @NonNull
    @Override
    @SuppressWarnings({"java:S2637", "ConstantConditions"})
    public final PassportEvaluation evaluate(@NonNull final Vaccination vaccination, @NonNull final LabResults results, @NonNull final LocalDateTime now) {
        try {
            setNow(now);
            final boolean completed = vaccination.completed() != null;
            final VaccinationScheduleType scheduleType = vaccination.scheduleType();
            final LocalDateTime mostRecentEffectuation =
                ofNullable(vaccination.mostRecentlyEffectuated())
                    .map(EffectuatedVaccination::effectuated)
                    .orElse(null);

            final boolean isUnder18 = ageCalculator.isUnder18(vaccination.cpr(), now);
            final TemporalAmount timeUntil18 = ageCalculator.timeUntil18(vaccination.cpr(), now);

            return firstNonNullOrDefaultSupplier(
                () -> grant(validUntil(vaccination, completed, mostRecentEffectuation, isUnder18, timeUntil18)),
                // Reasons not to grant a vaccination passport follow below,
                // if none are met return above
                () -> denyForNoVaccinations(vaccination),
                () -> denyForMinimumDurationAfterCompletedVaccinationHasNotPassed(mostRecentEffectuation, scheduleType, completed),
                () -> denyForMinimumDurationAfterPartialVaccinationHasNotPassed(vaccination, completed),
                () -> denyForMaximumDurationAfterVaccinationHasPassed(mostRecentEffectuation, completed, isUnder18, timeUntil18),
                () -> denyForSuspendedDueToPositiveLabResult(results, completed));
        } finally {
            clearNow();
        }
    }


    /**
     * Calculates the valid until date time of a vaccination
     *
     * @param vaccination
     * @param completed
     * @param mostRecentEffectuation
     * @param isUnder18
     * @param timeUntil18
     * @return
     */
    private LocalDateTime validUntil(
        @NonNull final Vaccination vaccination,
        final boolean completed,
        @NonNull final LocalDateTime mostRecentEffectuation, boolean isUnder18, TemporalAmount timeUntil18) {
        return completed
            ? vaccinationValidUntil(mostRecentEffectuation, completed, isUnder18, timeUntil18)
            : vaccinationValidUntil(vaccination);
    }

    @SuppressWarnings("UnstableApiUsage")
    private LocalDateTime vaccinationValidUntil(@NonNull final Vaccination vaccination) {
        final RangeSet<LocalDateTime> validities = calculateValidities(vaccination.vaccinations());
        final Range<LocalDateTime> validity = validities.rangeContaining(now());
        return validity.upperEndpoint();
    }

    /**
     * Evaluates if a passport should be denied due to no effectuated vaccinations
     *
     * @param vaccination the vaccination schedule to look at
     * @return the denied evaluation if the passport was denied, else <code>null</code>.
     */
    @Nullable
    private PassportDeniedEvaluation denyForNoVaccinations(@NonNull final Vaccination vaccination) {
        return vaccination.vaccinations().isEmpty() ? denyIndefinitely(NO_VACCINATIONS) : null;
    }

    /**
     * Evaluates if a passport should be denied temporarily due to positive lab results.
     *
     * @param results   the lab results to look among.
     * @param completed whether or not the vaccination schedule is completed
     * @return the denied evaluation if the passport was denied, else <code>null</code>.
     */
    @Nullable
    private PassportDeniedEvaluation denyForSuspendedDueToPositiveLabResult(
        @NonNull final LabResults results,
        @NonNull final boolean completed) {
        return Objects.firstNonNullSupplier(
            () -> denyForPositivePcrResultWithinSuspensionPeriod(results, completed),
            () -> denyForPositiveAntigenResultWithinSuspensionPeriodThatHasNotBeenFollowedByANegativePcrResult(results, completed));
    }

    /**
     * Evaluates if a passport should be denied temporarily due to positive antigen result within the suspension period
     * that has not be set aside by a subsequent negative pcr result.
     *
     * @param results   the lab results to look among.
     * @param completed whether or not the vaccination schedule is completed
     * @return the denied evaluation if the passport was denied, else <code>null</code>.
     */
    @Nullable
    private PassportDeniedEvaluation denyForPositiveAntigenResultWithinSuspensionPeriodThatHasNotBeenFollowedByANegativePcrResult(
        @NonNull final LabResults results,
        final boolean completed) {
        final LabResult.PositiveAntigenResult mostRecentPositiveAntigenResult = results.mostRecentPositiveAntigenResult();
        final LabResult.NegativePcrResult mostRecentNegativePcrResult = results.mostRecentNegativePcrResult();

        if (mostRecentPositiveAntigenResult == null) return null;
        if (resultOutsideSuspensionPeriod(mostRecentPositiveAntigenResult, completed)) return null;
        if (pcrResultAfterAntigenResult(mostRecentNegativePcrResult, mostRecentPositiveAntigenResult)) return null;

        return deny(positiveResultExpiresSuspensionPeriod(mostRecentPositiveAntigenResult, completed), POSITIVE_ANTIGEN_RESULT_WITHIN_SUSPENSION_PERIOD_WITHOUT_SUBSEQUENT_NEGATIVE_PCR_RESULT);
    }

    /**
     * Evaluates if a passport should be denied because the minimum duration after a completed vaccination has been effectuated
     * has not passed.
     *
     * @param mostRecentlyEffectuated the date and time of the most recently effectuated vaccination
     * @param scheduleType            the type of the vaccination schedule
     * @param completed               whether or not the vaccination schedule is completed
     * @return the denied evaluation if the passport was denied, else <code>null</code>.
     */
    @Nullable
    private PassportDeniedEvaluation denyForMinimumDurationAfterCompletedVaccinationHasNotPassed(
        @NonNull final LocalDateTime mostRecentlyEffectuated,
        @NonNull final VaccinationScheduleType scheduleType,
        final boolean completed) {
        if (!completed) return null;

        final LocalDateTime vaccinationValidFrom = vaccinationValidFrom(mostRecentlyEffectuated, scheduleType, completed);
        return now().isBefore(vaccinationValidFrom) ? deny(vaccinationValidFrom, MINIMUM_DURATION_NOT_PASSED) : null;
    }

    /**
     * Evaluates if a passport should be denied because the minimum duration after partial vaccinations have been effectuated has not passed.
     *
     * @param vaccination the effectuated vaccinations to look among
     * @param completed   whether or not the vaccination schedule is completed
     * @return the denied evaluation if the passport was denied, else <code>null</code>.
     */
    @SuppressWarnings("UnstableApiUsage")
    @Nullable
    private PassportDeniedEvaluation denyForMinimumDurationAfterPartialVaccinationHasNotPassed(
        @NonNull final Vaccination vaccination,
        @NonNull final boolean completed) {
        if (completed) return null;

        final LocalDateTime now = now();
        final RangeSet<LocalDateTime> validities = calculateValidities(vaccination.vaccinations());

        if (validities.contains(now)) return null;

        final LocalDateTime earliestValidity = validities.asRanges().stream()
            .map(Range::lowerEndpoint)
            .filter(now::isBefore)
            .findFirst()
            .orElse(null);

        return earliestValidity == null
            ? denyIndefinitely(VACCINATION_EXPIRED)
            : deny(earliestValidity, MINIMUM_DURATION_NOT_PASSED);
    }

    /**
     * Calculates a range set of validities
     *
     * @param vaccinations the effectuated vaccinations
     * @return the range set
     */
    @SuppressWarnings("UnstableApiUsage")
    private RangeSet<LocalDateTime> calculateValidities(@NonNull final Collection<EffectuatedVaccination> vaccinations) {
        final RangeSet<LocalDateTime> rangeSet = TreeRangeSet.create();
        vaccinations.stream()
            .map(EffectuatedVaccination::effectuated)
            .filter(this::validityOfUncompletedMultiDoseVaccinationHasNotExpired)
            .map(this::validityOfUncompletedMultiDoseVaccination)
            .forEach(rangeSet::add);
        return rangeSet;
    }

    private boolean validityOfUncompletedMultiDoseVaccinationHasNotExpired(
        @NonNull final LocalDateTime effectuated) {
        final LocalDateTime validUntil = uncompletedVaccinationValidUntil(effectuated);
        return now().isBefore(validUntil);
    }

    private Range<LocalDateTime> validityOfUncompletedMultiDoseVaccination(@NonNull final LocalDateTime effectuated) {
        return closedOpen(
            vaccinationValidFrom(effectuated, MULTI_DOSE_VACCINATION_SCHEDULE, false),
            uncompletedVaccinationValidUntil(effectuated));
    }


    /**
     * Evaluates if a passport should be denied because the maximum duration after a vaccination has been effectuated
     * has passed.
     *
     * @param mostRecentlyEffectuated the date and time of the most recently effectuated vaccination
     * @param completed               whether the vaccination schedule is completed
     * @param isUnder18               whether the person is under 18 years old
     * @param timeUntil18
     * @return the denied evaluation if the passport was denied, else <code>null</code>.
     */
    @Nullable
    private PassportDeniedEvaluation denyForMaximumDurationAfterVaccinationHasPassed(
        @NonNull final LocalDateTime mostRecentlyEffectuated,
        final boolean completed, boolean isUnder18, TemporalAmount timeUntil18) {
        final LocalDateTime validUntil = vaccinationValidUntil(mostRecentlyEffectuated, completed, isUnder18, timeUntil18);
        return now().isBefore(validUntil) ? null : denyIndefinitely(VACCINATION_EXPIRED);
    }

    /**
     * Evaluates if a passport should be denied because there is a positve pcr result within the suspension period.
     *
     * @param results   the lab results to look among.
     * @param completed whether or not the vaccination schedule is completed
     * @return the denied evaluation if the passport was denied, else <code>null</code>.
     */
    @Nullable
    private PassportDeniedEvaluation denyForPositivePcrResultWithinSuspensionPeriod(
        @NonNull final LabResults results,
        final boolean completed) {
        final LabResult.PositivePcrResult mostRecentPositivePcrResult = results.mostRecentPositivePcrResult();

        if (mostRecentPositivePcrResult == null) return null;
        if (resultOutsideSuspensionPeriod(mostRecentPositivePcrResult, completed)) return null;

        return deny(positiveResultExpiresSuspensionPeriod(mostRecentPositivePcrResult, completed), POSITIVE_PCR_RESULT_WITHIN_SUSPENSION_PERIOD);
    }

    /**
     * Evaluates whether or not there is a result within the suspension period.
     *
     * @param result    the lab result to evaluate.
     * @param completed whether or not the vaccination schedule is completed
     * @return <code>false</code> if the result is within the suspension period, else <code>true</code>.
     */
    private boolean resultOutsideSuspensionPeriod(@NonNull final LabResult result, final boolean completed) {
        final LocalDateTime samplingDateTime = result.samplingDateTime();
        return !samplingDateTime.isAfter(startOfSuspensionPeriod(completed));
    }

    /**
     * Evaluates whether or not an antigen result is before a pcr result.
     *
     * @param pcrResult     the most recent pcr result, can be <code>null</code>.
     * @param antigenResult the most recent positive antigen result.
     * @return <code>true</code> if the pcr result is not <code>null</code> and after the antigen result.
     */
    private static boolean pcrResultAfterAntigenResult(
        @Nullable final PcrResult pcrResult,
        @NonNull final AntigenResult antigenResult) {
        if (pcrResult == null) return false;

        final LocalDateTime dateTimeOfMostRecentPositiveAntigenResult = antigenResult.samplingDateTime();
        final LocalDateTime dateTimeOfMostRecentNegativePcrResult = pcrResult.samplingDateTime();
        return dateTimeOfMostRecentNegativePcrResult.isAfter(dateTimeOfMostRecentPositiveAntigenResult);
    }

    /**
     * Calculates the earliest date from which an effectuated vaccination can grant a passport
     *
     * @param effectuated  the date and time of the effectuated vaccination
     * @param scheduleType the type of the vaccination schedule
     * @param completed    whether or not the vaccination schedule is completed
     * @return the date
     */
    @NonNull
    private LocalDateTime vaccinationValidFrom(
        @NonNull final LocalDateTime effectuated,
        @NonNull final VaccinationScheduleType scheduleType,
        final boolean completed) {

        if (completed) {
            return scheduleType == SINGLE_DOSE_VACCINATION_SCHEDULE
                ? resolutionStartOfDay(effectuated.plus(minimumPeriodAfterCompletedSingleVaccinationSchedule), keepTimeOnMinimumPeriodAfterCompletedSingleVaccinationSchedule)
                : resolutionStartOfDay(effectuated.plus(minimumPeriodAfterCompletedMultiVaccinationSchedule), keepTimeOnMinimumPeriodAfterCompletedMultiVaccinationSchedule);
        }

        return resolutionStartOfDay(effectuated.plus(minimumPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule), keepTimeOnMinimumPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule);

    }

    /**
     * Calculates the latest date and time from which an effectuated vaccination can grant a passport
     *
     * @param mostRecentlyEffectuated the date and time of the most recently effectuated vaccination
     * @param completed               whether or not the vaccination schedule is completed
     * @param isUnder18               whether the person is under 18 yeas old
     * @param timeUntil18             time until the person is 18 years old
     * @return the date and time
     */
    @NonNull
    private LocalDateTime vaccinationValidUntil(
        @NonNull final LocalDateTime mostRecentlyEffectuated,
        final boolean completed, boolean isUnder18, TemporalAmount timeUntil18) {
        if (completed && isUnder18) {
            if (now().plus(timeUntil18).isBefore(mostRecentlyEffectuated.plus(maximumPeriodAfterCompletedVaccinationSchedule))) {
                // person will be 18 before the vaccination passport expires, thus the maximum period still applies
                return resolutionEndOfDay(mostRecentlyEffectuated.plus(maximumPeriodAfterCompletedVaccinationSchedule), keepTimeOnMaximumPeriodAfterCompletedVaccinationSchedule);
            } else if (now().plus(timeUntil18).isBefore(mostRecentlyEffectuated.plus(maximumPeriodAfterCompletedVaccinationScheduleUnder18))) {
                // person will be 18 before the exemption period has passed, thus they are exempt until they are 18 in stead
                return resolutionStartOfDay(now().plus(timeUntil18), false);
            } else {
                return resolutionEndOfDay(mostRecentlyEffectuated.plus(maximumPeriodAfterCompletedVaccinationScheduleUnder18), keepTimeOnMaximumPeriodAfterCompletedVaccinationScheduleUnder18);
            }
        } else if (completed) {
            return resolutionEndOfDay(mostRecentlyEffectuated.plus(maximumPeriodAfterCompletedVaccinationSchedule), keepTimeOnMaximumPeriodAfterCompletedVaccinationSchedule);
        } else {
            return resolutionEndOfDay(mostRecentlyEffectuated.plus(maximumPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule), keepTimeOnMaximumPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule);
        }
    }

    @NonNull
    LocalDateTime uncompletedVaccinationValidUntil(@NonNull final LocalDateTime mostRecentlyEffectuated) {
        return vaccinationValidUntil(mostRecentlyEffectuated, false, false, Duration.ofHours(0));
    }

    /**
     * Calculates the earliest day of a possible suspension of an effectuated vaccination due to positive results.
     *
     * @param completed whether or not the vaccination schedule is completed
     * @return the date
     */
    @NonNull
    private LocalDateTime startOfSuspensionPeriod(final boolean completed) {
        return completed
            ? resolutionStartOfDay(now().minus(suspensionPeriodAfterCompletedVaccinationSchedule), keepTimeOnSuspensionPeriodAfterCompletedVaccinationSchedule)
            : resolutionStartOfDay(now().minus(suspensionPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule), keepTimeOnSuspensionPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule);
    }

    /**
     * Calculates the last day of a suspension of an effectuated vaccination due to positive results.
     *
     * @param positiveResult the positive result
     * @param completed      whether or not the vaccination schedule is completed
     * @return the date
     */
    @NonNull
    private LocalDateTime positiveResultExpiresSuspensionPeriod(
        @NonNull final LabResult.PositiveResult positiveResult,
        final boolean completed) {
        return completed
            ? resolutionEndOfDay(positiveResult.samplingDateTime().plus(suspensionPeriodAfterCompletedVaccinationSchedule), keepTimeOnSuspensionPeriodAfterCompletedVaccinationSchedule)
            : resolutionEndOfDay(positiveResult.samplingDateTime().plus(suspensionPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule), keepTimeOnSuspensionPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule);
    }

}
