/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.rules.configuration;

import org.springframework.lang.NonNull;

import java.time.temporal.TemporalAmount;

import static dk.sundhedsdatastyrelsen.cbs.rules.configuration.TemporalAmountWithTimeInstruction.of;

public final class VaccinationPassportConfigurationBuilder {
    private TemporalAmountWithTimeInstruction minimumAmountAfterCompletedSingleVaccinationSchedule;
    private TemporalAmountWithTimeInstruction minimumAmountAfterCompletedMultiVaccinationSchedule;
    private TemporalAmountWithTimeInstruction maximumAmountAfterCompletedVaccinationSchedule;
    private TemporalAmountWithTimeInstruction suspensionAmountAfterCompletedVaccinationSchedule;
    private TemporalAmountWithTimeInstruction minimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule;
    private TemporalAmountWithTimeInstruction maximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule;
    private TemporalAmountWithTimeInstruction suspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule;
    private TemporalAmountWithTimeInstruction maximumPeriodAfterCompletedVaccinationScheduleUnder18;

    private VaccinationPassportConfigurationBuilder() {}

    @NonNull
    public static VaccinationPassportConfigurationBuilder configuration() {
        return new VaccinationPassportConfigurationBuilder();
    }

    @NonNull
    public VaccinationPassportConfigurationBuilder minimumAmountAfterCompletedSingleVaccinationSchedule(
        @NonNull final TemporalAmount amount, final boolean keepTime) {
        return minimumAmountAfterCompletedSingleVaccinationSchedule(of(amount, keepTime));
    }

    @NonNull
    public VaccinationPassportConfigurationBuilder minimumAmountAfterCompletedMultiVaccinationSchedule(
        @NonNull final TemporalAmount amount, final boolean keepTime) {
        return minimumAmountAfterCompletedMultiVaccinationSchedule(of(amount, keepTime));
    }

    @NonNull
    public VaccinationPassportConfigurationBuilder maximumAmountAfterCompletedVaccinationSchedule(
        @NonNull final TemporalAmount amount, final boolean keepTime) {
        return maximumAmountAfterCompletedVaccinationSchedule(of(amount, keepTime));
    }

    @NonNull
    public VaccinationPassportConfigurationBuilder suspensionAmountAfterCompletedVaccinationSchedule(
        @NonNull final TemporalAmount amount, final boolean keepTime) {
        return suspensionAmountAfterCompletedVaccinationSchedule(of(amount, keepTime));
    }

    @NonNull
    public VaccinationPassportConfigurationBuilder minimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(
        @NonNull final TemporalAmount amount, final boolean keepTime) {
        return minimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(of(amount, keepTime));
    }

    @NonNull
    public VaccinationPassportConfigurationBuilder maximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(
        @NonNull final TemporalAmount amount, final boolean keepTime) {
        return maximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(of(amount, keepTime));
    }

    @NonNull
    public VaccinationPassportConfigurationBuilder suspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(
        @NonNull final TemporalAmount amount, final boolean keepTime) {
        return suspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(of(amount, keepTime));
    }

    @NonNull
    public VaccinationPassportConfigurationBuilder maximumPeriodAfterCompletedVaccinationScheduleUnder18(
        @NonNull final TemporalAmount amount, final boolean keepTime) {
        return maximumPeriodAfterCompletedVaccinationScheduleUnder18(of(amount, keepTime));
    }

    @NonNull
    public VaccinationPassportConfigurationBuilder minimumAmountAfterCompletedSingleVaccinationSchedule(
        @NonNull final TemporalAmountWithTimeInstruction amount) {
        minimumAmountAfterCompletedSingleVaccinationSchedule = amount;
        return this;
    }

    @NonNull
    public VaccinationPassportConfigurationBuilder minimumAmountAfterCompletedMultiVaccinationSchedule(
        @NonNull final TemporalAmountWithTimeInstruction amount) {
        minimumAmountAfterCompletedMultiVaccinationSchedule = amount;
        return this;
    }

    @NonNull
    public VaccinationPassportConfigurationBuilder maximumAmountAfterCompletedVaccinationSchedule(
        @NonNull final TemporalAmountWithTimeInstruction amount) {
        maximumAmountAfterCompletedVaccinationSchedule = amount;
        return this;
    }

    @NonNull
    public VaccinationPassportConfigurationBuilder suspensionAmountAfterCompletedVaccinationSchedule(
        @NonNull final TemporalAmountWithTimeInstruction amount) {
        suspensionAmountAfterCompletedVaccinationSchedule = amount;
        return this;
    }

    @NonNull
    public VaccinationPassportConfigurationBuilder minimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(
    @NonNull final TemporalAmountWithTimeInstruction amount) {
        minimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule = amount;
        return this;
    }

    @NonNull
    public VaccinationPassportConfigurationBuilder maximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(
    @NonNull final TemporalAmountWithTimeInstruction amount) {
        maximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule = amount;
        return this;
    }

    @NonNull
    public VaccinationPassportConfigurationBuilder suspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(
    @NonNull final TemporalAmountWithTimeInstruction amount) {
        suspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule = amount;
        return this;
    }

    @NonNull
    public VaccinationPassportConfigurationBuilder maximumPeriodAfterCompletedVaccinationScheduleUnder18(
        @NonNull final TemporalAmountWithTimeInstruction amount) {
        maximumPeriodAfterCompletedVaccinationScheduleUnder18 = amount;
        return this;
    }

    @NonNull
    public VaccinationPassportConfiguration build() {
        return new VaccinationPassportConfiguration(
            minimumAmountAfterCompletedSingleVaccinationSchedule,
            minimumAmountAfterCompletedMultiVaccinationSchedule,
            maximumAmountAfterCompletedVaccinationSchedule,
            suspensionAmountAfterCompletedVaccinationSchedule,
            minimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule,
            maximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule,
            suspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule,
            maximumPeriodAfterCompletedVaccinationScheduleUnder18);
    }
}
