/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.rules.evaluation;

import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import java.time.LocalDateTime;

public interface PassportEvaluation {

    /**
     * Gets if this passport evaluation resulted in a passport.
     * @return <code>true</code> if the passport is granted, else <code>false</code>.
     */
    boolean passportGranted();

    /**
     * Gets the date after which this passport evaluation must be evaluated again,
     * e.g. date the passport evaluation can be cached until.
     * If the passport evaluation can be cached indefinitely, <code>null</code> is returned.
     * @return the date, possibly <code>null</code>.
     */
    @Nullable
    LocalDateTime expiresAfter();

    /**
     * Gets the rejection reason. If this passport evaluation did not resulted in a passport,
     * a rejection reason is set.
     * @return when {@link #passportGranted()} is <code>false</code> the rejection reason, else <code>null</code>.
     */
    @Nullable
    String rejectionReason();

    @NonNull
    static PassportDeniedEvaluation denyIndefinitely(@NonNull final String rejectionReason) {
        return new PassportDeniedEvaluation(null, rejectionReason);
    }

    @NonNull
    static PassportDeniedEvaluation deny(@NonNull final LocalDateTime expireAfter, @NonNull final String rejectionReason) {
        return new PassportDeniedEvaluation(expireAfter, rejectionReason);
    }

    @NonNull
    static PassportGrantedEvaluation grant(@NonNull final LocalDateTime expireAfter) {
        return new PassportGrantedEvaluation(expireAfter);
    }

}
