/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.rules;

import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.rules.configuration.VaccinationPassportConfiguration;
import dk.sundhedsdatastyrelsen.cbs.rules.configuration.VaccinationPassportConfigurationBuilder;
import dk.sundhedsdatastyrelsen.cbs.test.Data;
import dk.sundhedsdatastyrelsen.cbs.test.Data.Results;
import dk.sundhedsdatastyrelsen.cbs.test.Data.Vaccinations;
import org.junit.jupiter.params.provider.Arguments;
import org.springframework.lang.NonNull;

import java.time.Duration;
import java.time.LocalDateTime;
import java.time.Period;
import java.time.format.DateTimeFormatter;
import java.time.temporal.TemporalAmount;

import static java.lang.String.format;

public abstract class ArgumentBuilder<T extends ArgumentBuilder<T>> {

    private static final DateTimeFormatter DATE_FORMAT = DateTimeFormatter.ISO_DATE_TIME;

    public static final class Vaccine {
        @NonNull
        public static final TemporalAmount DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_SINGLE_VACCINATION_SCHEDULE = Period.ofDays(14);
        @NonNull
        public static final TemporalAmount DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE = Period.ZERO;
        @NonNull
        public static final TemporalAmount DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE = Period.ofMonths(8);
        @NonNull
        public static final TemporalAmount DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE_UNDER_18 = Period.ofYears(5);
        @NonNull
        public static final TemporalAmount DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE = Period.ofDays(14);
        @NonNull
        public static final TemporalAmount DEFAULT_MINIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE = Period.ofDays(14);
        @NonNull
        public static final TemporalAmount DEFAULT_MAXIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE = Period.ofDays(42);
        @NonNull
        public static final TemporalAmount DEFAULT_SUSPENSION_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE = Period.ofDays(14);

        public static final TemporalAmount DEFAULT_MINIMUM_AMOUNT_AFTER_REVACCINATION = Period.ofDays(5); //note: this is not the prod value - different here to differentiate from DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE
        @NonNull
        public static final TemporalAmount DEFAULT_MAXIMUM_AMOUNT_AFTER_REVACCINATION = Period.ofYears(3);
        @NonNull
        public static final TemporalAmount DEFAULT_SUSPENSION_AMOUNT_AFTER_REVACCINATION = Period.ofDays(33); //note: this is not the prod value - different here to differentiate from DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE
    }

    public static final class Test {
        @NonNull
        public static final TemporalAmount DEFAULT_MINIMUM_AMOUNT = Period.ZERO;
        @NonNull
        public static final TemporalAmount DEFAULT_MAXIMUM_AMOUNT_PCR = Duration.ofHours(96);
        @NonNull
        public static final TemporalAmount DEFAULT_MAXIMUM_AMOUNT_AG = Duration.ofHours(72);
        @NonNull
        public static final TemporalAmount DEFAULT_SUSPENSION_AMOUNT = Period.ofDays(14);
    }

    public static final class Recovery {
        @NonNull
        public static final TemporalAmount DEFAULT_MINIMUM_AMOUNT = Period.ofDays(14);
        @NonNull
        public static final TemporalAmount DEFAULT_MAXIMUM_AMOUNT = Period.ofMonths(8);
        @NonNull
        public static final TemporalAmount DEFAULT_INACTIVE_INFECTION_AMOUNT = Period.ofWeeks(12);
    }

    public static final class EuRecovery {
        @NonNull
        public static final TemporalAmount DEFAULT_MINIMUM_AMOUNT = Period.ofDays(11);
        @NonNull
        public static final TemporalAmount DEFAULT_MAXIMUM_AMOUNT = Period.ofDays(180);
    }


    public static final class TestArgumentsBuilder extends ArgumentBuilder<TestArgumentsBuilder> {

        private TemporalAmount minimumAmount;
        private boolean keepTimeOnMinimumAmount;
        private TemporalAmount maximumAmountPcr;
        private TemporalAmount maximumAmountAg;
        private boolean keepTimeOnMaximumAmount;
        private TemporalAmount suspensionAmount;
        private boolean keepTimeOnSuspensionAmount;

        private TestArgumentsBuilder(@NonNull final String name) {
            super(name);
        }

        @NonNull
        public TestArgumentsBuilder minimumAmount(@NonNull final TemporalAmount minimumAmount) {
            this.minimumAmount = minimumAmount;
            return this;
        }

        @NonNull
        public TestArgumentsBuilder maximumAmountPcr(@NonNull final TemporalAmount maximumAmountPcr) {
            this.maximumAmountPcr = maximumAmountPcr;
            return this;
        }

        @NonNull
        public TestArgumentsBuilder maximumAmountAg(@NonNull final TemporalAmount maximumAmountAg) {
            this.maximumAmountAg = maximumAmountAg;
            return this;
        }

        @NonNull
        public TestArgumentsBuilder suspensionOrInactiveInfectionAmount(@NonNull final TemporalAmount suspensionAmount) {
            this.suspensionAmount = suspensionAmount;
            return this;
        }

        @NonNull
        public TestArgumentsBuilder keepTimeOnMinimumAmount(final boolean keepTimeOnMinimumAmount) {
            this.keepTimeOnMinimumAmount = keepTimeOnMinimumAmount;
            return this;
        }

        @NonNull
        public TestArgumentsBuilder keepTimeOnMaximumAmount(final boolean keepTimeOnMaximumAmount) {
            this.keepTimeOnMaximumAmount = keepTimeOnMaximumAmount;
            return this;
        }

        @NonNull
        public TestArgumentsBuilder keepTimeOnSuspensionAmount(final boolean keepTimeOnSuspensionAmount) {
            this.keepTimeOnSuspensionAmount = keepTimeOnSuspensionAmount;
            return this;
        }

        @NonNull
        public TestArgumentsBuilder keepTimeOnMinimumAmount() {
            return keepTimeOnMinimumAmount(true);
        }

        @NonNull
        public TestArgumentsBuilder keepTimeOnMaximumAmount() {
            return keepTimeOnMaximumAmount(true);
        }

        @NonNull
        public TestArgumentsBuilder keepTimeOnSuspensionAmount() {
            return keepTimeOnSuspensionAmount(true);
        }

        @NonNull
        public Arguments build() {
            return Arguments.arguments(
                granted ? "GRANT" : "DENY",
                name,
                friendly(),
                minimumAmount,
                keepTimeOnMinimumAmount,
                maximumAmountPcr,
                maximumAmountAg,
                keepTimeOnMaximumAmount,
                suspensionAmount,
                keepTimeOnSuspensionAmount,
                vaccination,
                labResults,
                now,
                granted,
                expire);
        }

        @NonNull
        private String friendly() {
            return format(
                "{ min = %s, keepMin = %s, maxPcr = %s, maxAg = %s, keepMax = %s, susp = %s, keep = %s, vac = %s, lab = %s, now = %s, granted = %s, expire = %s }",
                minimumAmount,
                keepTimeOnMinimumAmount,
                maximumAmountPcr,
                maximumAmountAg,
                keepTimeOnMaximumAmount,
                suspensionAmount,
                keepTimeOnSuspensionAmount,
                Data.friendly(vaccination),
                friendly(labResults),
                DATE_FORMAT.format(now),
                granted,
                expire == null ? null : DATE_FORMAT.format(expire));
        }
    }

    public static final class RecoveryArgumentsBuilder extends ArgumentBuilder<RecoveryArgumentsBuilder> {

        private TemporalAmount minimumAmount;
        private boolean keepTimeOnMinimumAmount;
        private TemporalAmount maximumAmount;
        private boolean keepTimeOnMaximumAmount;
        private TemporalAmount inactiveInfectionAmount;
        private boolean keepTimeOnInactiveInfectionAmount;

        private RecoveryArgumentsBuilder(@NonNull final String name) {
            super(name);
        }

        @NonNull
        public RecoveryArgumentsBuilder minimumAmount(@NonNull final TemporalAmount minimumAmount) {
            this.minimumAmount = minimumAmount;
            return this;
        }

        @NonNull
        public RecoveryArgumentsBuilder maximumAmount(@NonNull final TemporalAmount maximumAmount) {
            this.maximumAmount = maximumAmount;
            return this;
        }

        @NonNull
        public RecoveryArgumentsBuilder inactiveInfectionAmount(@NonNull final TemporalAmount suspensionAmount) {
            this.inactiveInfectionAmount = suspensionAmount;
            return this;
        }

        @NonNull
        public RecoveryArgumentsBuilder keepTimeOnMinimumAmount(final boolean keepTimeOnMinimumAmount) {
            this.keepTimeOnMinimumAmount = keepTimeOnMinimumAmount;
            return this;
        }

        @NonNull
        public RecoveryArgumentsBuilder keepTimeOnMaximumAmount(final boolean keepTimeOnMaximumAmount) {
            this.keepTimeOnMaximumAmount = keepTimeOnMaximumAmount;
            return this;
        }

        @NonNull
        public RecoveryArgumentsBuilder keepTimeOnSuspensionAmount(final boolean keepTimeOnSuspensionAmount) {
            this.keepTimeOnInactiveInfectionAmount = keepTimeOnSuspensionAmount;
            return this;
        }

        @NonNull
        public RecoveryArgumentsBuilder keepTimeOnMinimumAmount() {
            return keepTimeOnMinimumAmount(true);
        }

        @NonNull
        public RecoveryArgumentsBuilder keepTimeOnMaximumAmount() {
            return keepTimeOnMaximumAmount(true);
        }

        @NonNull
        public RecoveryArgumentsBuilder keepTimeOnSuspensionAmount() {
            return keepTimeOnSuspensionAmount(true);
        }

        @NonNull
        public Arguments build() {
            return Arguments.arguments(
                granted ? "GRANT" : "DENY",
                name,
                friendly(),
                minimumAmount,
                keepTimeOnMinimumAmount,
                maximumAmount,
                keepTimeOnMaximumAmount,
                inactiveInfectionAmount,
                keepTimeOnInactiveInfectionAmount,
                vaccination,
                labResults,
                now,
                granted,
                expire);
        }

        @NonNull
        private String friendly() {
            return format(
                "{ min = %s, keep = %s, max = %s, keep = %s, inactive = %s, keep = %s, vac = %s, lab = %s, now = %s, granted = %s, expire = %s }",
                minimumAmount,
                keepTimeOnMinimumAmount,
                maximumAmount,
                keepTimeOnMaximumAmount,
                inactiveInfectionAmount,
                keepTimeOnInactiveInfectionAmount,
                Data.friendly(vaccination),
                friendly(labResults),
                DATE_FORMAT.format(now),
                granted,
                expire == null ? null : DATE_FORMAT.format(expire));
        }
    }


    public static final class EuRecoveryArgumentsBuilder extends ArgumentBuilder<EuRecoveryArgumentsBuilder> {
        private TemporalAmount minimumAmount;
        private TemporalAmount maximumAmount;

        private EuRecoveryArgumentsBuilder(@NonNull final String name) {
            super(name);
        }

        @NonNull
        public EuRecoveryArgumentsBuilder minimumAmount(@NonNull final TemporalAmount minimumAmount) {
            this.minimumAmount = minimumAmount;
            return this;
        }

        @NonNull
        public EuRecoveryArgumentsBuilder maximumAmount(@NonNull final TemporalAmount maximumAmount) {
            this.maximumAmount = maximumAmount;
            return this;
        }

        @NonNull
        public Arguments build() {
            return Arguments.arguments(
                granted ? "GRANT" : "DENY",
                name,
                friendly(),
                minimumAmount,
                maximumAmount,
                vaccination,
                labResults,
                now,
                granted,
                expire);
        }

        @NonNull
        private String friendly() {
            return format(
                "{ min = %s, max = %s, vac = %s, lab = %s, now = %s, granted = %s, expire = %s }",
                minimumAmount,
                maximumAmount,
                Data.friendly(vaccination),
                friendly(labResults),
                DATE_FORMAT.format(now),
                granted,
                expire == null ? null : DATE_FORMAT.format(expire));
        }
    }

    public static final class VaccinationArgumentsBuilder extends ArgumentBuilder<VaccinationArgumentsBuilder> {

        private TemporalAmount minimumAmountAfterCompletedSingleVaccinationSchedule;
        private boolean keepTimeOnMinimumAmountAfterCompletedSingleVaccinationSchedule;
        private TemporalAmount minimumAmountAfterCompletedMultiVaccinationSchedule;
        private boolean keepTimeOnMinimumAmountAfterCompletedMultiVaccinationSchedule;
        private TemporalAmount maximumAmountAfterCompletedVaccinationSchedule;
        private boolean keepTimeOnMaximumAmountAfterCompletedVaccinationSchedule;
        private TemporalAmount suspensionAmountAfterCompletedVaccinationSchedule;
        private boolean keepTimeOnSuspensionAmountAfterCompletedVaccinationSchedule;

        private TemporalAmount maximumAmountAfterCompletedVaccinationScheduleUnder18;
        private boolean keepTimeOnMaximumAmountAfterCompletedVaccinationScheduleUnder18;

        private TemporalAmount minimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule;
        private boolean keepTimeOnMinimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule;
        private TemporalAmount maximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule;
        private boolean keepTimeOnMaximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule;
        private TemporalAmount suspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule;
        private boolean keepTimeOnSuspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule;

        private VaccinationArgumentsBuilder(@NonNull final String name) {
            super(name);
        }

        @NonNull
        public VaccinationArgumentsBuilder minimumAmountAfterCompletedMultiVaccinationSchedule(@NonNull final TemporalAmount amount) {
            minimumAmountAfterCompletedMultiVaccinationSchedule = amount;
            return this;
        }

        @NonNull
        public VaccinationArgumentsBuilder minimumAmountAfterCompletedSingleVaccinationSchedule(@NonNull final TemporalAmount amount) {
            minimumAmountAfterCompletedSingleVaccinationSchedule = amount;
            return this;
        }

        @NonNull
        public VaccinationArgumentsBuilder maximumAmountAfterCompletedVaccinationSchedule(@NonNull final TemporalAmount amount) {
            maximumAmountAfterCompletedVaccinationSchedule = amount;
            return this;
        }

        @NonNull
        public VaccinationArgumentsBuilder suspensionAmountAfterCompletedVaccinationSchedule(@NonNull final TemporalAmount amount) {
            suspensionAmountAfterCompletedVaccinationSchedule = amount;
            return this;
        }

        @NonNull
        public VaccinationArgumentsBuilder maximumAmountAfterCompletedVaccinationScheduleUnder18(@NonNull final TemporalAmount amount) {
            maximumAmountAfterCompletedVaccinationScheduleUnder18 = amount;
            return this;
        }

        @NonNull
        public VaccinationArgumentsBuilder keepTimeOnMinimumAmountAfterCompletedSingleVaccinationSchedule(final boolean keepTime) {
            keepTimeOnMinimumAmountAfterCompletedSingleVaccinationSchedule = keepTime;
            return this;
        }

        @NonNull
        public VaccinationArgumentsBuilder keepTimeOnMinimumAmountAfterCompletedMultiVaccinationSchedule(final boolean keepTime) {
            keepTimeOnMinimumAmountAfterCompletedMultiVaccinationSchedule = keepTime;
            return this;
        }

        @NonNull
        public VaccinationArgumentsBuilder keepTimeOnMaximumAmountAfterCompletedVaccinationSchedule(final boolean keepTime) {
            keepTimeOnMaximumAmountAfterCompletedVaccinationSchedule = keepTime;
            return this;
        }

        @NonNull
        public VaccinationArgumentsBuilder keepTimeOnSuspensionAmountAfterCompletedVaccinationSchedule(final boolean keepTime) {
            keepTimeOnSuspensionAmountAfterCompletedVaccinationSchedule = keepTime;
            return this;
        }

        @NonNull
        public VaccinationArgumentsBuilder keepTimeOnMaximumAmountAfterCompletedVaccinationScheduleUnder18(final boolean keepTime) {
            keepTimeOnMaximumAmountAfterCompletedVaccinationScheduleUnder18 = keepTime;
            return this;
        }

        @NonNull
        public VaccinationArgumentsBuilder keepTimeOnMinimumAmountAfterCompletedSingleVaccinationSchedule() {
            return keepTimeOnMinimumAmountAfterCompletedSingleVaccinationSchedule(true);
        }

        @NonNull
        public VaccinationArgumentsBuilder keepTimeOnMinimumAmountAfterCompletedMultiVaccinationSchedule() {
            return keepTimeOnMinimumAmountAfterCompletedMultiVaccinationSchedule(true);
        }

        @NonNull
        public VaccinationArgumentsBuilder keepTimeOnMaximumAmountAfterCompletedVaccinationSchedule() {
            return keepTimeOnMaximumAmountAfterCompletedVaccinationSchedule(true);
        }

        @NonNull
        public VaccinationArgumentsBuilder keepTimeOnSuspensionAmountAfterCompletedVaccinationSchedule() {
            return keepTimeOnSuspensionAmountAfterCompletedVaccinationSchedule(true);
        }

        @NonNull
        public VaccinationArgumentsBuilder keepTimeOnMaximumAmountAfterCompletedVaccinationScheduleUnder18() {
            return keepTimeOnMaximumAmountAfterCompletedVaccinationScheduleUnder18(true);
        }

        @NonNull
        public VaccinationArgumentsBuilder minimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(@NonNull final TemporalAmount amount) {
            minimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule = amount;
            return this;
        }

        @NonNull
        public VaccinationArgumentsBuilder maximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(@NonNull final TemporalAmount amount) {
            maximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule = amount;
            return this;
        }

        @NonNull
        public VaccinationArgumentsBuilder suspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(@NonNull final TemporalAmount amount) {
            suspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule = amount;
            return this;
        }

        @NonNull
        public VaccinationArgumentsBuilder keepTimeOnMinimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(final boolean keepTime) {
            keepTimeOnMinimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule = keepTime;
            return this;
        }

        @NonNull
        public VaccinationArgumentsBuilder keepTimeOnMaximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(final boolean keepTime) {
            keepTimeOnMaximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule = keepTime;
            return this;
        }

        @NonNull
        public VaccinationArgumentsBuilder keepTimeOnSuspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(final boolean keepTime) {
            keepTimeOnSuspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule = keepTime;
            return this;
        }

        @NonNull
        public VaccinationArgumentsBuilder keepTimeOnMinimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule() {
            return keepTimeOnMinimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(true);
        }

        @NonNull
        public VaccinationArgumentsBuilder keepTimeOnMaximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule() {
            return keepTimeOnMaximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(true);
        }

        @NonNull
        public VaccinationArgumentsBuilder keepTimeOnSuspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule() {
            return keepTimeOnSuspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(true);
        }

        @NonNull
        public Arguments build() {
            return Arguments.arguments(
                granted ? "GRANT" : "DENY",
                name,
                friendly(),
                configuration(),
                vaccination,
                labResults,
                now,
                granted,
                expire);
        }

        @NonNull
        private String friendly() {
            return format(
                "{ minCompSingle = %s, keep = %s, minCompMulti = %s, keep = %s, maxComp = %s, keep = %s, suspComp = %s, keep = %s, minPart= %s, keep = %s, maxPart= %s, keep = %s, suspPart= %s, keep = %s, vac = %s, lab = %s, now = %s, granted = %s, expire = %s }",
                minimumAmountAfterCompletedSingleVaccinationSchedule,
                keepTimeOnMinimumAmountAfterCompletedSingleVaccinationSchedule,
                minimumAmountAfterCompletedMultiVaccinationSchedule,
                keepTimeOnMinimumAmountAfterCompletedMultiVaccinationSchedule,
                maximumAmountAfterCompletedVaccinationSchedule,
                keepTimeOnMaximumAmountAfterCompletedVaccinationSchedule,
                suspensionAmountAfterCompletedVaccinationSchedule,
                keepTimeOnSuspensionAmountAfterCompletedVaccinationSchedule,
                minimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule,
                keepTimeOnMinimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule,
                maximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule,
                keepTimeOnMaximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule,
                suspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule,
                keepTimeOnSuspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule,
                Data.friendly(vaccination),
                friendly(labResults),
                DATE_FORMAT.format(now),
                granted,
                expire == null ? null : DATE_FORMAT.format(expire));
        }

        public VaccinationPassportConfiguration configuration() {
            return VaccinationPassportConfigurationBuilder.configuration()
                .minimumAmountAfterCompletedSingleVaccinationSchedule(minimumAmountAfterCompletedSingleVaccinationSchedule, keepTimeOnMinimumAmountAfterCompletedSingleVaccinationSchedule)
                .minimumAmountAfterCompletedMultiVaccinationSchedule(minimumAmountAfterCompletedMultiVaccinationSchedule, keepTimeOnMinimumAmountAfterCompletedMultiVaccinationSchedule)
                .maximumAmountAfterCompletedVaccinationSchedule(maximumAmountAfterCompletedVaccinationSchedule, keepTimeOnMaximumAmountAfterCompletedVaccinationSchedule)
                .suspensionAmountAfterCompletedVaccinationSchedule(suspensionAmountAfterCompletedVaccinationSchedule, keepTimeOnSuspensionAmountAfterCompletedVaccinationSchedule)
                .minimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(minimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule, keepTimeOnMinimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule)
                .maximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(maximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule, keepTimeOnMaximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule)
                .suspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(suspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule, keepTimeOnSuspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule)
                .maximumPeriodAfterCompletedVaccinationScheduleUnder18(maximumAmountAfterCompletedVaccinationScheduleUnder18, keepTimeOnMaximumAmountAfterCompletedVaccinationScheduleUnder18)
                .build();
        }
    }

    public static final class RevaccinationArgumentsBuilder extends ArgumentBuilder<RevaccinationArgumentsBuilder> {
        private TemporalAmount minimumPeriodAfterRevaccination;
        private boolean keepTimeOnMinimumPeriodAfterRevaccination;
        private TemporalAmount maximumPeriodAfterRevaccination;
        private boolean keepTimeOnMaximumPeriodAfterRevaccination;
        private TemporalAmount suspensionPeriodAfterRevaccination;
        private boolean keepTimeOnSuspensionPeriodAfterRevaccination;

        private RevaccinationArgumentsBuilder(@NonNull final String name) {
            super(name);
        }

        public RevaccinationArgumentsBuilder minimumPeriodAfterRevaccination(TemporalAmount minimumPeriodAfterRevaccination) {
            this.minimumPeriodAfterRevaccination = minimumPeriodAfterRevaccination;
            return this;
        }

        public RevaccinationArgumentsBuilder keepTimeOnMinimumPeriodAfterRevaccination(boolean keepTimeOnMinimumPeriodAfterRevaccination) {
            this.keepTimeOnMinimumPeriodAfterRevaccination = keepTimeOnMinimumPeriodAfterRevaccination;
            return this;
        }

        public RevaccinationArgumentsBuilder maximumPeriodAfterRevaccination(TemporalAmount maximumPeriodAfterRevaccination) {
            this.maximumPeriodAfterRevaccination = maximumPeriodAfterRevaccination;
            return this;
        }

        public RevaccinationArgumentsBuilder keepTimeOnMaximumPeriodAfterRevaccination(boolean keepTimeOnMaximumPeriodAfterRevaccination) {
            this.keepTimeOnMaximumPeriodAfterRevaccination = keepTimeOnMaximumPeriodAfterRevaccination;
            return this;
        }

        public RevaccinationArgumentsBuilder suspensionPeriodAfterRevaccination(TemporalAmount suspensionPeriodAfterRevaccination) {
            this.suspensionPeriodAfterRevaccination = suspensionPeriodAfterRevaccination;
            return this;
        }

        public RevaccinationArgumentsBuilder keepTimeOnSuspensionPeriodAfterRevaccination(boolean keepTimeOnSuspensionPeriodAfterRevaccination) {
            this.keepTimeOnSuspensionPeriodAfterRevaccination = keepTimeOnSuspensionPeriodAfterRevaccination;
            return this;
        }

        @NonNull
        public Arguments build() {
            return Arguments.arguments(
                granted ? "GRANT" : "DENY",
                name,
                friendly(),
                minimumPeriodAfterRevaccination,
                keepTimeOnMinimumPeriodAfterRevaccination,
                maximumPeriodAfterRevaccination,
                keepTimeOnMaximumPeriodAfterRevaccination,
                suspensionPeriodAfterRevaccination,
                keepTimeOnSuspensionPeriodAfterRevaccination,
                vaccination,
                labResults,
                now,
                granted,
                expire);
        }

        @NonNull
        private String friendly() {
            return format(
                "{ min = %s (keep=%s), max = %s (keep=%s), susp = %s (keep=%s), vac = %s, lab = %s, now = %s, granted = %s, expire = %s }",
                minimumPeriodAfterRevaccination,
                keepTimeOnMinimumPeriodAfterRevaccination,
                maximumPeriodAfterRevaccination,
                keepTimeOnMaximumPeriodAfterRevaccination,
                suspensionPeriodAfterRevaccination,
                keepTimeOnSuspensionPeriodAfterRevaccination,
                Data.friendly(vaccination),
                friendly(labResults),
                DATE_FORMAT.format(now),
                granted,
                expire == null ? null : DATE_FORMAT.format(expire));
        }
    }

    public static final LocalDateTime DEFAULT_NOW = LocalDateTime.of(2000, 1, 15, 0, 0, 0);

    @NonNull
    protected final String name;
    @NonNull
    protected final T actual;
    protected LocalDateTime now = DEFAULT_NOW;
    protected Vaccination vaccination = Vaccinations.EMPTY;
    protected LabResults labResults = Results.EMPTY;
    protected boolean granted;
    protected LocalDateTime expire;

    @SuppressWarnings("unchecked")
    private ArgumentBuilder(@NonNull final String name) {
        this.name = name;
        actual = (T) this;
    }

    @NonNull
    public static VaccinationArgumentsBuilder defaultForVaccinationPass(@NonNull final String name) {
        return new VaccinationArgumentsBuilder(name)
            .minimumAmountAfterCompletedSingleVaccinationSchedule(Vaccine.DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_SINGLE_VACCINATION_SCHEDULE)
            .minimumAmountAfterCompletedMultiVaccinationSchedule(Vaccine.DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE)
            .maximumAmountAfterCompletedVaccinationSchedule(Vaccine.DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE)
            .maximumAmountAfterCompletedVaccinationSchedule(Vaccine.DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE)
            .suspensionAmountAfterCompletedVaccinationSchedule(Vaccine.DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE)
            .keepTimeOnMinimumAmountAfterCompletedSingleVaccinationSchedule()
            .keepTimeOnMinimumAmountAfterCompletedMultiVaccinationSchedule()
            .keepTimeOnMaximumAmountAfterCompletedVaccinationSchedule()
            .keepTimeOnSuspensionAmountAfterCompletedVaccinationSchedule()
            .minimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(Vaccine.DEFAULT_MINIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE)
            .maximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(Vaccine.DEFAULT_MAXIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE)
            .suspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(Vaccine.DEFAULT_SUSPENSION_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE)
            .keepTimeOnMinimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule()
            .keepTimeOnMaximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule()
            .keepTimeOnSuspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule()
            .maximumAmountAfterCompletedVaccinationScheduleUnder18(Vaccine.DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE_UNDER_18)
            .keepTimeOnMaximumAmountAfterCompletedVaccinationScheduleUnder18()
            ;
    }

    @NonNull
    public static RevaccinationArgumentsBuilder defaultForRevaccinationPass(@NonNull final String name) {
        return new RevaccinationArgumentsBuilder(name)
            .minimumPeriodAfterRevaccination(Vaccine.DEFAULT_MINIMUM_AMOUNT_AFTER_REVACCINATION)
            .maximumPeriodAfterRevaccination(Vaccine.DEFAULT_MAXIMUM_AMOUNT_AFTER_REVACCINATION)
            .suspensionPeriodAfterRevaccination(Vaccine.DEFAULT_SUSPENSION_AMOUNT_AFTER_REVACCINATION)
            .keepTimeOnMinimumPeriodAfterRevaccination(true)
            .keepTimeOnMaximumPeriodAfterRevaccination(true)
            .keepTimeOnSuspensionPeriodAfterRevaccination(true);
    }


    @NonNull
    public static TestArgumentsBuilder defaultForTestPass(@NonNull final String name) {
        return new TestArgumentsBuilder(name)
            .minimumAmount(Test.DEFAULT_MINIMUM_AMOUNT)
            .maximumAmountPcr(Test.DEFAULT_MAXIMUM_AMOUNT_PCR)
            .maximumAmountAg(Test.DEFAULT_MAXIMUM_AMOUNT_AG)
            .suspensionOrInactiveInfectionAmount(Test.DEFAULT_SUSPENSION_AMOUNT)
            .keepTimeOnMinimumAmount()
            .keepTimeOnMaximumAmount()
            .keepTimeOnSuspensionAmount();
    }

    @NonNull
    public static RecoveryArgumentsBuilder defaultForRecoveryPass(@NonNull final String name) {
        return new RecoveryArgumentsBuilder(name)
            .minimumAmount(Recovery.DEFAULT_MINIMUM_AMOUNT)
            .maximumAmount(Recovery.DEFAULT_MAXIMUM_AMOUNT)
            .inactiveInfectionAmount(Recovery.DEFAULT_INACTIVE_INFECTION_AMOUNT)
            .keepTimeOnMinimumAmount()
            .keepTimeOnMaximumAmount()
            .keepTimeOnSuspensionAmount();
    }

    @NonNull
    public static EuRecoveryArgumentsBuilder defaultForEuRecoveryPass(@NonNull final String name) {
        return new EuRecoveryArgumentsBuilder(name)
            .minimumAmount(EuRecovery.DEFAULT_MINIMUM_AMOUNT)
            .maximumAmount(EuRecovery.DEFAULT_MAXIMUM_AMOUNT);
    }

    @NonNull
    public T vaccination(@NonNull final Vaccination vaccination) {
        this.vaccination = vaccination;
        return actual;
    }

    @NonNull
    public T labResults(@NonNull final LabResults labResults) {
        this.labResults = labResults;
        return actual;
    }

    @NonNull
    public T now(@NonNull final LocalDateTime now) {
        this.now = now;
        return actual;
    }

    @NonNull
    private T granted(final boolean granted) {
        this.granted = granted;
        return actual;
    }

    @NonNull
    public T granted() {
        return granted(true);
    }

    @NonNull
    public T denied() {
        return granted(false);
    }

    @NonNull
    public T expire(@NonNull final LocalDateTime expire) {
        this.expire = expire;
        return actual;
    }

    @NonNull
    public T expireNever() {
        this.expire = null;
        return actual;
    }

    @NonNull
    protected final String friendly(@NonNull final LabResults labResults) {
        return "{ " +
            labResults.results().stream().reduce(
                "",
                (a, l) -> "".equals(a) ? Data.friendly(l) : a + "," + Data.friendly(l),
                (x, y) -> x + "," + y
            ) +
            " }";
    }

}
