/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.rules;

import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.RecoveryPassportEvaluation;
import org.apache.log4j.Logger;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.springframework.lang.NonNull;

import java.time.LocalDateTime;
import java.time.temporal.TemporalAmount;
import java.util.stream.Stream;

import static dk.sundhedsdatastyrelsen.cbs.rules.ArgumentBuilder.DEFAULT_NOW;
import static dk.sundhedsdatastyrelsen.cbs.rules.ArgumentBuilder.EuRecovery.DEFAULT_MAXIMUM_AMOUNT;
import static dk.sundhedsdatastyrelsen.cbs.rules.ArgumentBuilder.EuRecovery.DEFAULT_MINIMUM_AMOUNT;
import static dk.sundhedsdatastyrelsen.cbs.rules.ArgumentBuilder.defaultForEuRecoveryPass;
import static dk.sundhedsdatastyrelsen.cbs.rules.harness.LocalDateTimes.afterMin;
import static dk.sundhedsdatastyrelsen.cbs.rules.harness.LocalDateTimes.beforeMin;
import static dk.sundhedsdatastyrelsen.cbs.rules.harness.LocalDateTimes.expireDenied;
import static dk.sundhedsdatastyrelsen.cbs.rules.harness.LocalDateTimes.expireGranted;
import static dk.sundhedsdatastyrelsen.cbs.test.Data.Results.positivePcr;
import static org.junit.jupiter.api.Assertions.assertEquals;

class EuRecoveryPassportEvaluatorTest {
    private static final Logger logger = Logger.getLogger(EuRecoveryPassportEvaluatorTest.class);

    @SuppressWarnings("unused")
    @DisplayName("evaluation")
    @ParameterizedTest(name = "[{0}] - {1} : {2}")
    @MethodSource("evaluateSource")
    void evaluate(
        @NonNull final String outcome,
        @NonNull final String name,
        @NonNull final String friendly,
        @NonNull final TemporalAmount minimumPeriod,
        @NonNull final TemporalAmount maximumPeriod,
        @NonNull final Vaccination vaccination,
        @NonNull final LabResults labResults,
        @NonNull final LocalDateTime now,
        @NonNull final boolean granted,
        @NonNull final LocalDateTime validUntil) {
        final EuRecoveryPassportEvaluator evaluator = new EuRecoveryPassportEvaluator(minimumPeriod, maximumPeriod);
        final RecoveryPassportEvaluation evaluation = evaluator.evaluate(vaccination, labResults, now);
        logger.info(evaluation.rejectionReason());
        assertEquals(granted, evaluation.passportGranted());
        assertEquals(validUntil, evaluation.expiresAfter());
    }

    @NonNull
    private static Stream<Arguments> evaluateSource() {
        return Stream.of(
            defaultForEuRecoveryPass("No results").denied().expireNever().build(),
            defaultForEuRecoveryPass("Positive pcr test more than minimum period has passed").labResults(positivePcr(beforeMin(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT, true))).granted().expire(expireGranted(beforeMin(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT, true), DEFAULT_MAXIMUM_AMOUNT, true)).build(),
            defaultForEuRecoveryPass("Positive pcr test less than minimum period has passed").labResults(positivePcr(afterMin(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT, true))).denied().expire(expireDenied(afterMin(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT, true), DEFAULT_MINIMUM_AMOUNT, true)).build(),
            defaultForEuRecoveryPass("Overlapping inactive infections").labResults(positivePcr(afterMin(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT, true), beforeMin(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT, true))).granted().expire(expireGranted(beforeMin(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT, true), DEFAULT_MAXIMUM_AMOUNT, true)).build());
    }
}
