/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.rules;

import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluation;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.RecoveryLabResult;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.RecoveryPassportEvaluation;
import org.junit.jupiter.api.Test;

import java.time.LocalDateTime;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.mock;


class PassportEvaluationsImplTest {
    private final LocalDateTime expiration = LocalDateTime.now().plusDays(3);

    private final PassportEvaluationsImpl deniedEvaluation = new PassportEvaluationsImpl(
        PassportEvaluation.deny(expiration, "vaccination passport denied"),
        PassportEvaluation.deny(expiration, "revaccination passport denied"),
        PassportEvaluation.deny(expiration, "test passport denied"),
        RecoveryPassportEvaluation.deny(expiration, "recovery passport denied"));

    private final PassportEvaluationsImpl partialDeniedEvaluation = new PassportEvaluationsImpl(
        PassportEvaluation.deny(expiration, "vaccination passport denied"),
        PassportEvaluation.deny(expiration, "revaccination passport denied"),
        PassportEvaluation.grant(expiration),
        RecoveryPassportEvaluation.deny(expiration, "recovery passport denied"));

    private final PassportEvaluationsImpl grantedEvaluation = new PassportEvaluationsImpl(
        PassportEvaluation.grant(expiration),
        PassportEvaluation.grant(expiration),
        PassportEvaluation.grant(expiration),
        RecoveryPassportEvaluation.grant(expiration, mock(RecoveryLabResult.class)));


    @Test
    void passportGranted() {
        assertTrue(grantedEvaluation.passportGranted());
    }

    @Test
    void passportDenied() {
        assertFalse(deniedEvaluation.passportGranted());
    }

    @Test
    void expiresNow() {
        assertEquals(expiration, grantedEvaluation.expiresAfter());
    }

    @Test
    void rejectionReason() {
        assertEquals("vaccination passport denied,revaccination passport denied,test passport denied,recovery passport denied", deniedEvaluation.rejectionReason());
    }

    @Test
    void deniedByPassports() {
        assertFalse(partialDeniedEvaluation.vaccinationPassport().passportGranted());
        assertFalse(partialDeniedEvaluation.revaccinationPassport().passportGranted());
        assertTrue(partialDeniedEvaluation.testPassport().passportGranted());
        assertFalse(partialDeniedEvaluation.recoveryPassport().passportGranted());
    }

    @Test
    void expirationOfDenialIsBasedOnLongestExpiration() {
        final LocalDateTime now = LocalDateTime.now();

        final PassportEvaluationsImpl passportEvaluations = new PassportEvaluationsImpl(
            PassportEvaluation.deny(now.plusHours(1L), ""),
            PassportEvaluation.deny(now.plusHours(2L), ""),
            PassportEvaluation.deny(now.plusHours(3L), ""),
            RecoveryPassportEvaluation.deny(now.plusHours(4L), "")
        );

        assertEquals(now.plusHours(4L), passportEvaluations.expiresAfter());
    }

    @Test
    void expirationOfGrantIsBasedOnLongestGrantExpiration() {
        final LocalDateTime now = LocalDateTime.now();

        final PassportEvaluationsImpl passportEvaluations = new PassportEvaluationsImpl(
            PassportEvaluation.grant(now.plusHours(3L)),
            PassportEvaluation.grant(now.plusHours(4L)),
            PassportEvaluation.deny(now.plusHours(10L), ""),
            RecoveryPassportEvaluation.grant(now.plusHours(5L), mock(RecoveryLabResult.class))
        );

        assertEquals(now.plusHours(5L), passportEvaluations.expiresAfter());
    }
}
