/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.rules;

import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluation;
import dk.sundhedsdatastyrelsen.cbs.test.RandomData;
import org.apache.log4j.Logger;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.springframework.lang.NonNull;

import java.time.LocalDateTime;
import java.time.temporal.TemporalAmount;
import java.util.stream.Stream;

import static dk.sundhedsdatastyrelsen.cbs.rules.ArgumentBuilder.DEFAULT_NOW;
import static dk.sundhedsdatastyrelsen.cbs.rules.ArgumentBuilder.Vaccine.*;
import static dk.sundhedsdatastyrelsen.cbs.rules.ArgumentBuilder.defaultForRevaccinationPass;
import static dk.sundhedsdatastyrelsen.cbs.rules.harness.LocalDateTimes.*;
import static dk.sundhedsdatastyrelsen.cbs.test.Data.Results.*;
import static dk.sundhedsdatastyrelsen.cbs.test.Data.Vaccinations.completedMultiDoseWithAdditionalVaccinations;
import static dk.sundhedsdatastyrelsen.cbs.test.Data.Vaccinations.completedSingleDoseWithAdditionalVaccinations;
import static org.junit.jupiter.api.Assertions.assertEquals;

class RevaccinationPassportEvaluatorTest {
    private static final Logger logger = Logger.getLogger(VaccinationPassportEvaluatorTest.class);
    private static final RandomData randomData = new RandomData();

    @SuppressWarnings("unused")
    @DisplayName("evaluation")
    @ParameterizedTest(name = "[{0}] - {1} : {2}")
    @MethodSource("evaluateSource")
    void evaluate(@NonNull final String outcome,
                  @NonNull final String name,
                  @NonNull final String friendly,
                  @NonNull final TemporalAmount minimumPeriodAfterRevaccination,
                  @NonNull final boolean keepTimeOnMinimumPeriodAfterRevaccination,
                  @NonNull final TemporalAmount maximumPeriodAfterRevaccination,
                  @NonNull final boolean keepTimeOnMaximumPeriodAfterRevaccination,
                  @NonNull final TemporalAmount suspensionPeriodAfterRevaccination,
                  @NonNull final boolean keepTimeOnSuspensionPeriodAfterRevaccination,
                  @NonNull final Vaccination vaccination,
                  @NonNull final LabResults labResults,
                  @NonNull final LocalDateTime now,
                  @NonNull final boolean granted,
                  @NonNull final LocalDateTime validUntil) {
        final RevaccinationPassportEvaluator evaluator = new RevaccinationPassportEvaluator(minimumPeriodAfterRevaccination, keepTimeOnMinimumPeriodAfterRevaccination, maximumPeriodAfterRevaccination, keepTimeOnMaximumPeriodAfterRevaccination, suspensionPeriodAfterRevaccination, keepTimeOnSuspensionPeriodAfterRevaccination);
        final PassportEvaluation evaluation = evaluator.evaluate(vaccination, labResults, now);
        logger.info(evaluation.rejectionReason());
        assertEquals(granted, evaluation.passportGranted());
        assertEquals(validUntil, evaluation.expiresAfter());
    }

    @NonNull
    private static Stream<Arguments> evaluateSource() {
        LocalDateTime onMinAfterCompletedMultiVaccination = on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true);
        LocalDateTime onMinAfterCompletedSingleVaccination = on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_SINGLE_VACCINATION_SCHEDULE, true);
        LocalDateTime afterMinAfterRevaccination = afterMin(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_REVACCINATION, true);
        LocalDateTime onMinAfterRevaccination = on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_REVACCINATION, true);
        LocalDateTime beforeMaxAfterRevaccination = beforeMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AFTER_REVACCINATION, true);
        LocalDateTime onMaxAfterRevaccination = on(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AFTER_REVACCINATION, true);
        LocalDateTime afterMaxAfterRevaccination = afterMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AFTER_REVACCINATION, true);
        return Stream.of(
            defaultForRevaccinationPass("No re-vaccinations")
                .denied()
                .expireNever()
                .build()
            ,
            defaultForRevaccinationPass("re-vaccination where less than minimum period has passed")
                .vaccination(completedMultiDoseWithAdditionalVaccinations(onMinAfterCompletedMultiVaccination, afterMinAfterRevaccination))
                .denied()
                .expire(expireDenied(afterMinAfterRevaccination, DEFAULT_MINIMUM_AMOUNT_AFTER_REVACCINATION, true))
                .build()
            ,
            defaultForRevaccinationPass("re-vaccination where exactly minimum period has passed")
                .vaccination(completedMultiDoseWithAdditionalVaccinations(onMinAfterCompletedMultiVaccination, onMinAfterRevaccination))
                .granted()
                .expire(expireGranted(onMinAfterRevaccination, DEFAULT_MAXIMUM_AMOUNT_AFTER_REVACCINATION, true))
                .build()
            ,
            defaultForRevaccinationPass("re-vaccination where less than maximum period has passed")
                .vaccination(completedMultiDoseWithAdditionalVaccinations(onMinAfterCompletedMultiVaccination, beforeMaxAfterRevaccination))
                .granted()
                .expire(expireGranted(beforeMaxAfterRevaccination, DEFAULT_MAXIMUM_AMOUNT_AFTER_REVACCINATION, true))
                .build()
            ,
            defaultForRevaccinationPass("re-vaccination where exactly maximum period has passed")
                .vaccination(completedMultiDoseWithAdditionalVaccinations(onMinAfterCompletedMultiVaccination, onMaxAfterRevaccination))
                .denied()
                .expireNever()
                .build()
            ,
            defaultForRevaccinationPass("re-vaccination where more than maximum period has passed")
                .vaccination(completedMultiDoseWithAdditionalVaccinations(onMinAfterCompletedMultiVaccination, afterMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AFTER_REVACCINATION, true)))
                .denied()
                .expireNever()
                .build()
            ,
            defaultForRevaccinationPass("re-vaccination where exactly minimum period has passed and positive pcr where more than suspension period has passed")
                .vaccination(completedMultiDoseWithAdditionalVaccinations(onMinAfterCompletedMultiVaccination, onMinAfterRevaccination))
                .labResults(positivePcr(beforeMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_REVACCINATION, true)))
                .granted()
                .expire(expireGranted(onMinAfterRevaccination, DEFAULT_MAXIMUM_AMOUNT_AFTER_REVACCINATION, true))
                .build()
            ,
            defaultForRevaccinationPass("re-vaccination where exactly minimum period has passed and positive pcr where exactly suspension period has passed")
                .vaccination(completedMultiDoseWithAdditionalVaccinations(onMinAfterCompletedMultiVaccination, onMinAfterRevaccination))
                .labResults(positivePcr(on(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_REVACCINATION, true)))
                .granted()
                .expire(expireGranted(onMinAfterRevaccination, DEFAULT_MAXIMUM_AMOUNT_AFTER_REVACCINATION, true))
                .build()
            ,
            defaultForRevaccinationPass("re-vaccination where exactly minimum period has passed and positive pcr is now")
                .vaccination(completedMultiDoseWithAdditionalVaccinations(onMinAfterCompletedMultiVaccination, onMinAfterRevaccination))
                .labResults(positivePcr(DEFAULT_NOW))
                .denied()
                .expire(expireDenied(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_REVACCINATION, true))
                .build()
            ,
            defaultForRevaccinationPass("re-vaccination where exactly minimum period has passed and positive pcr is in future")
                .vaccination(completedMultiDoseWithAdditionalVaccinations(onMinAfterCompletedMultiVaccination, onMinAfterRevaccination))
                .labResults(positivePcr(DEFAULT_NOW.plusNanos(1)))
                .denied()
                .expire(expireDenied(DEFAULT_NOW.plusNanos(1), DEFAULT_SUSPENSION_AMOUNT_AFTER_REVACCINATION, true))
                .build()
            ,
            defaultForRevaccinationPass("re-vaccination where exactly minimum period has passed and positive pcr where less than suspension period has passed")
                .vaccination(completedMultiDoseWithAdditionalVaccinations(onMinAfterCompletedMultiVaccination, onMinAfterRevaccination))
                .labResults(positivePcr(afterMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_REVACCINATION, true)))
                .denied()
                .expire(expireDenied(afterMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_REVACCINATION, true), DEFAULT_SUSPENSION_AMOUNT_AFTER_REVACCINATION, true))
                .build()
            ,
            defaultForRevaccinationPass("re-vaccination where exactly minimum period has passed and positive antigen where more than suspension period has passed")
                .vaccination(completedMultiDoseWithAdditionalVaccinations(onMinAfterCompletedMultiVaccination, onMinAfterRevaccination))
                .labResults(positiveAntigen(beforeMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_REVACCINATION, true)))
                .granted()
                .expire(expireGranted(onMinAfterRevaccination, DEFAULT_MAXIMUM_AMOUNT_AFTER_REVACCINATION, true))
                .build()
            ,
            defaultForRevaccinationPass("re-vaccination where exactly minimum period has passed and positive antigen where exactly suspension period has passed")
                .vaccination(completedMultiDoseWithAdditionalVaccinations(onMinAfterCompletedMultiVaccination, onMinAfterRevaccination))
                .labResults(positiveAntigen(on(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_REVACCINATION, true)))
                .granted().expire(expireGranted(onMinAfterRevaccination, DEFAULT_MAXIMUM_AMOUNT_AFTER_REVACCINATION, true))
                .build()
            ,
            defaultForRevaccinationPass("re-vaccination where less than maximum period has passed and positive antigen where less than suspension period has passed")
                .vaccination(completedMultiDoseWithAdditionalVaccinations(beforeMaxAfterRevaccination, onMinAfterRevaccination))
                .labResults(positiveAntigen(afterMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_REVACCINATION, true)))
                .denied()
                .expire(expireDenied(afterMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_REVACCINATION, true), DEFAULT_SUSPENSION_AMOUNT_AFTER_REVACCINATION, true))
                .build()
            ,
            defaultForRevaccinationPass("re-vaccination where exactly minimum period has passed and positive antigen where less than suspension period has passed and negative pcr test after positive antigen")
                .vaccination(completedMultiDoseWithAdditionalVaccinations(onMinAfterCompletedMultiVaccination, onMinAfterRevaccination))
                .labResults(positiveAntigenBeforeNegativePcr(afterMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_REVACCINATION, true)))
                .granted()
                .expire(expireGranted(onMinAfterRevaccination, DEFAULT_MAXIMUM_AMOUNT_AFTER_REVACCINATION, true))
                .build()
            ,

            //MULTIPLE REVACCINATIONS
            defaultForRevaccinationPass("two re-vaccinations where both are valid")
                .vaccination(completedSingleDoseWithAdditionalVaccinations(onMinAfterCompletedSingleVaccination,
                    beforeMaxAfterRevaccination,
                    onMinAfterRevaccination))
                .granted()
                .expire(expireGranted(onMinAfterRevaccination, DEFAULT_MAXIMUM_AMOUNT_AFTER_REVACCINATION, true))
                .build()
            ,
            defaultForRevaccinationPass("two re-vaccinations where oldest has less than maximum period passed and most recent has less than minimum period passed")
                .vaccination(completedSingleDoseWithAdditionalVaccinations(onMinAfterCompletedSingleVaccination,
                    beforeMaxAfterRevaccination,
                    afterMinAfterRevaccination))
                .granted()
                .expire(expireGranted(beforeMaxAfterRevaccination, DEFAULT_MAXIMUM_AMOUNT_AFTER_REVACCINATION, true))
                .build()
            ,
            defaultForRevaccinationPass("two re-vaccinations where oldest has less than maximum period passed and most recent has exactly minimum period passed")
                .vaccination(completedSingleDoseWithAdditionalVaccinations(onMinAfterCompletedSingleVaccination,
                    beforeMaxAfterRevaccination,
                    onMinAfterRevaccination))
                .granted()
                .expire(expireGranted(onMinAfterRevaccination, DEFAULT_MAXIMUM_AMOUNT_AFTER_REVACCINATION, true))
                .build()
            ,
            defaultForRevaccinationPass("two re-vaccinations where oldest has exactly maximum period passed and most recent has less than minimum period passed")
                .vaccination(completedSingleDoseWithAdditionalVaccinations(onMinAfterCompletedSingleVaccination,
                    onMaxAfterRevaccination,
                    afterMinAfterRevaccination))
                .denied()
                .expire(expireDenied(afterMinAfterRevaccination, DEFAULT_MINIMUM_AMOUNT_AFTER_REVACCINATION, true))
                .build()
            ,
            defaultForRevaccinationPass("two re-vaccinations where both are valid but there is a positive test within suspension period")
                .vaccination(completedSingleDoseWithAdditionalVaccinations(onMinAfterCompletedSingleVaccination,
                    beforeMaxAfterRevaccination,
                    onMinAfterRevaccination))
                .labResults(positivePcr(DEFAULT_NOW))
                .denied()
                .expire(expireDenied(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_REVACCINATION, true))
                .build()
            ,
            defaultForRevaccinationPass("three re-vaccinations where oldest has expired, middle is valid, and most recent is not yet valid")
                .vaccination(completedSingleDoseWithAdditionalVaccinations(onMinAfterCompletedSingleVaccination,
                    afterMaxAfterRevaccination,
                    beforeMaxAfterRevaccination,
                    afterMinAfterRevaccination))
                .granted()
                .expire(expireGranted(beforeMaxAfterRevaccination, DEFAULT_MAXIMUM_AMOUNT_AFTER_REVACCINATION, true))
                .build()

            );


    }
}
