/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.rules;

import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluation;
import org.apache.log4j.Logger;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.springframework.lang.NonNull;

import java.time.Duration;
import java.time.LocalDateTime;
import java.time.temporal.TemporalAmount;
import java.util.stream.Stream;

import static dk.sundhedsdatastyrelsen.cbs.rules.ArgumentBuilder.DEFAULT_NOW;
import static dk.sundhedsdatastyrelsen.cbs.rules.ArgumentBuilder.Test.*;
import static dk.sundhedsdatastyrelsen.cbs.rules.ArgumentBuilder.defaultForTestPass;
import static dk.sundhedsdatastyrelsen.cbs.rules.harness.LocalDateTimes.*;
import static dk.sundhedsdatastyrelsen.cbs.test.Data.Results.*;
import static org.junit.jupiter.api.Assertions.assertEquals;

class TestPassportEvaluatorTest {
    private static final Logger LOGGER = Logger.getLogger(TestPassportEvaluatorTest.class);

    @SuppressWarnings("unused")
    @DisplayName("evaluation")
    @ParameterizedTest(name = "[{0}] - {1} : {2}")
    @MethodSource("evaluateSource")
    void evaluate(
        @NonNull final String outcome,
        @NonNull final String name,
        @NonNull final String friendly,
        @NonNull final TemporalAmount minimumPeriod,
        final boolean keepTimeOnMinimumPeriod,
        @NonNull final TemporalAmount maximumPeriodPcr,
        @NonNull final TemporalAmount maximumPeriodAg,
        final boolean keepTimeOnMaximumPeriod,
        @NonNull final TemporalAmount suspensionPeriod,
        final boolean keepTimeOnSuspensionPeriod,
        @NonNull final Vaccination vaccination,
        @NonNull final LabResults labResults,
        @NonNull final LocalDateTime now,
        @NonNull final boolean granted,
        @NonNull final LocalDateTime validUntil) {
        final TestPassportEvaluator evaluator = new TestPassportEvaluator(minimumPeriod, keepTimeOnMinimumPeriod, maximumPeriodPcr, maximumPeriodAg, keepTimeOnMaximumPeriod, suspensionPeriod, keepTimeOnSuspensionPeriod);
        final PassportEvaluation evaluation = evaluator.evaluate(vaccination, labResults, now);
        if (!evaluation.passportGranted()) {
            LOGGER.info("Rejection reason: " + evaluation.rejectionReason());
        }
        assertEquals(granted, evaluation.passportGranted());
        assertEquals(validUntil, evaluation.expiresAfter());
    }

    @NonNull
    private static Stream<Arguments> evaluateSource() {
        return Stream.of(
            defaultForTestPass("No results").denied().expireNever().build(),
            defaultForTestPass("Positive pcr test more than minimum period has passed before negative pcr test").labResults(positivePcrBeforeNegativePcr(beforeMin(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT, true))).granted().expire(expireGranted(beforeMin(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT, true), DEFAULT_MAXIMUM_AMOUNT_PCR, true)).build(),
            defaultForTestPass("Positive pcr test less than maximum period has passed").labResults(positivePcr(beforeMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_PCR, true))).denied().expireNever().build(),
            defaultForTestPass("Positive pcr test more than maximum period has passed").labResults(positivePcr(afterMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_PCR, true))).denied().expireNever().build(),
            defaultForTestPass("Negative pcr test exactly minimum period has passed").labResults(negativePcr(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT, true))).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT, true), DEFAULT_MAXIMUM_AMOUNT_PCR, true)).build(),
            defaultForTestPass("Negative pcr test less than maximum period has passed").labResults(negativePcr(beforeMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_PCR, true))).granted().expire(expireGranted(beforeMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_PCR, true), DEFAULT_MAXIMUM_AMOUNT_PCR, true)).build(),
            defaultForTestPass("Negative pcr test exactly maximum period has passed").labResults(negativePcr(on(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_PCR, true))).denied().build(),
            defaultForTestPass("Negative pcr test more than maximum period has passed").labResults(negativePcr(afterMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_PCR, true))).denied().build(),
            defaultForTestPass("Negative antigen test less than maximum period has passed").labResults(negativeAntigen(beforeMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AG, true))).granted().expire(expireGranted(beforeMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AG, true), DEFAULT_MAXIMUM_AMOUNT_AG, true)).build(),
            defaultForTestPass("Negative antigen more less than maximum period has passed").labResults(negativeAntigen(afterMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AG, true))).denied().build(),
            defaultForTestPass("Negative antigen test less than maximum period has passed and positive pcr test less than suspension period has passed").labResults(negativeAntigenPositivePcr(beforeMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AG, true), beforeMax(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT, true))).denied().expire(expireDenied(beforeMax(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT, true), DEFAULT_SUSPENSION_AMOUNT, true)).build(),
            defaultForTestPass("Negative antigen test less than maximum period has passed and positive pcr test more than suspension period has passed").labResults(negativeAntigenPositivePcr(beforeMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AG, true), afterMax(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT, true))).granted().expire(expireGranted(beforeMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AG, true), DEFAULT_MAXIMUM_AMOUNT_AG, true)).build(),
            defaultForTestPass("Negative pcr test less than minimum period has passed").minimumAmount(Duration.ofHours(3)).labResults(negativePcr(beforeMin(DEFAULT_NOW, Duration.ofHours(2), true))).denied().expire(expireDenied(DEFAULT_NOW, Duration.ofHours(1).minusNanos(1), true)).build(),
            defaultForTestPass("Negative ag test less than minimum period has passed").minimumAmount(Duration.ofHours(3)).labResults(negativeAntigen(beforeMin(DEFAULT_NOW, Duration.ofHours(2), true))).denied().expire(expireDenied(DEFAULT_NOW, Duration.ofHours(1).minusNanos(1), true)).build(),
            defaultForTestPass("Negative ag test less than minimum period has passed and negative ag test in validity period").minimumAmount(Duration.ofHours(3)).labResults(negativeAntigen(beforeMin(DEFAULT_NOW, Duration.ofHours(2), true), beforeMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AG, true))).granted().expire(expireGranted(beforeMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AG, true), DEFAULT_MAXIMUM_AMOUNT_AG, true)).build(),
            defaultForTestPass("Negative pcr test less than minimum period has passed and negative pcr test in validity period").minimumAmount(Duration.ofHours(3)).labResults(negativePcr(beforeMin(DEFAULT_NOW, Duration.ofHours(2), true), beforeMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_PCR, true))).granted().expire(expireGranted(beforeMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_PCR, true), DEFAULT_MAXIMUM_AMOUNT_PCR, true)).build(),
            defaultForTestPass("Negative pcr test less than minimum period has passed and negative ag test in validity period").minimumAmount(Duration.ofHours(3)).labResults(negativePcrAndNegativeAntigen(beforeMin(DEFAULT_NOW, Duration.ofHours(2), true), beforeMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AG, true))).granted().expire(expireGranted(beforeMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AG, true), DEFAULT_MAXIMUM_AMOUNT_AG, true)).build(),
            defaultForTestPass("Negative ag test less than minimum period has passed and negative pcr test in validity period").minimumAmount(Duration.ofHours(3)).labResults(negativePcrAndNegativeAntigen(beforeMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_PCR, true), beforeMin(DEFAULT_NOW, Duration.ofHours(2), true))).granted().expire(expireGranted(beforeMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_PCR, true), DEFAULT_MAXIMUM_AMOUNT_PCR, true)).build()
        );
    }
}
