/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.rules;

import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.rules.configuration.VaccinationPassportConfiguration;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluation;
import dk.sundhedsdatastyrelsen.cbs.test.Data;
import org.apache.log4j.Logger;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import org.springframework.lang.NonNull;

import java.time.Duration;
import java.time.LocalDateTime;
import java.time.Period;
import java.time.temporal.TemporalAmount;
import java.util.stream.Stream;

import static dk.sundhedsdatastyrelsen.cbs.rules.ArgumentBuilder.DEFAULT_NOW;
import static dk.sundhedsdatastyrelsen.cbs.rules.ArgumentBuilder.Vaccine.*;
import static dk.sundhedsdatastyrelsen.cbs.rules.ArgumentBuilder.defaultForVaccinationPass;
import static dk.sundhedsdatastyrelsen.cbs.rules.harness.LocalDateTimes.*;
import static dk.sundhedsdatastyrelsen.cbs.test.Data.Results.*;
import static dk.sundhedsdatastyrelsen.cbs.test.Data.Vaccinations.*;
import static java.util.stream.Stream.of;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.when;

class VaccinationPassportEvaluatorTest {
    private static final Logger logger = Logger.getLogger(VaccinationPassportEvaluatorTest.class);

    @SuppressWarnings("unused")
    @DisplayName("evaluation over 18")
    @ParameterizedTest(name = "[{0}] - {1} : {2}")
    @MethodSource("evaluateOver18Source")
    void evaluateOver18(
        @NonNull final String outcome,
        @NonNull final String name,
        @NonNull final String friendly,
        @NonNull final VaccinationPassportConfiguration configuration,
        @NonNull final Vaccination vaccination,
        @NonNull final LabResults labResults,
        @NonNull final LocalDateTime now,
        @NonNull final boolean granted,
        @NonNull final LocalDateTime validUntil) {
        final VaccinationPassportEvaluator evaluator = new VaccinationPassportEvaluator(configuration, new CprFormatBasedAgeCalculator() {
        });
        final PassportEvaluation evaluation = evaluator.evaluate(vaccination, labResults, now);
        logger.info(evaluation.rejectionReason());
        assertEquals(granted, evaluation.passportGranted());
        assertEquals(validUntil, evaluation.expiresAfter());
    }

    @NonNull
    private static Stream<Arguments> evaluateOver18Source() {
        return of(
            defaultForVaccinationPass("No vaccinations").denied().build(),
            defaultForVaccinationPass("Completed vaccination where less than minimum period has passed").vaccination(completedMultiDoseWithAdditionalVaccinations(afterMin(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), 2)).denied().expire(expireDenied(afterMin(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed").vaccination(completedMultiDoseWithAdditionalVaccinations(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), 1)).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where less than maximum period has passed").vaccination(completedMultiDose(beforeMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true))).granted().expire(expireGranted(beforeMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly maximum period has passed").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true))).denied().expireNever().build(),
            defaultForVaccinationPass("Completed vaccination where more than maximum period has passed").vaccination(completedMultiDoseWithAdditionalVaccinations(afterMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true), 1)).denied().expireNever().build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive pcr where more than suspension period has passed").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true))).labResults(positivePcr(beforeMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true))).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive pcr where exactly suspension period has passed").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true))).labResults(positivePcr(on(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true))).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive pcr is now").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true))).labResults(positivePcr(DEFAULT_NOW)).denied().expire(expireDenied(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive pcr is in future").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true))).labResults(positivePcr(DEFAULT_NOW.plusNanos(1))).denied().expire(expireDenied(DEFAULT_NOW.plusNanos(1), DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive pcr where less suspension period has passed").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true))).labResults(positivePcr(afterMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true))).denied().expire(expireDenied(afterMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true), DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive antigen where more than suspension period has passed").vaccination(completedMultiDoseWithAdditionalVaccinations(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), 2)).labResults(positiveAntigen(beforeMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true))).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive antigen where exactly suspension period has passed").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true))).labResults(positiveAntigen(on(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true))).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive antigen where less suspension period has passed").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true))).labResults(positiveAntigen(afterMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true))).denied().expire(expireDenied(afterMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true), DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive antigen where less suspension period has passed and negative pcr test after positive antigen").vaccination(completedMultiDoseWithAdditionalVaccinations(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), 2)).labResults(positiveAntigenBeforeNegativePcr(afterMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true))).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Partial vaccination with one vaccination where exactly minimum period has passed").vaccination(partial(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true))).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Partial vaccination with 1st vaccination where exactly minimum period has passed and 2nd vaccination where less than minimum period has passed").vaccination(partial(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true), afterMin(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true))).granted().expire(expireGranted(afterMin(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Partial vaccination with 1st vaccination where exactly maximum minus minumum period has passed and 2nd vaccination administered now").vaccination(partial(on(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true), DEFAULT_NOW)).granted().expire(expireGranted(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true)).build()

        );
    }


    @SuppressWarnings("unused")
    @DisplayName("evaluation under 18")
    @ParameterizedTest(name = "[{0}] - {1} : {2}")
    @MethodSource("evaluateUnder18Source")
    void evaluateUnder18(
        @NonNull final String outcome,
        @NonNull final String name,
        @NonNull final String friendly,
        @NonNull final VaccinationPassportConfiguration configuration,
        @NonNull final Vaccination vaccination,
        @NonNull final LabResults labResults,
        @NonNull final LocalDateTime now,
        @NonNull final boolean granted,
        @NonNull final LocalDateTime validUntil) {
        final VaccinationPassportEvaluator evaluator = new VaccinationPassportEvaluator(configuration, new CprFormatBasedAgeCalculator() {
            @Override
            public boolean isUnder18(String cpr, LocalDateTime now) {
                return true;
            }

            @Override
            public TemporalAmount timeUntil18(String cpr, LocalDateTime now) {
                return Period.from(DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE_UNDER_18).plusDays(10);
            }
        });
        final PassportEvaluation evaluation = evaluator.evaluate(vaccination, labResults, now);
        logger.info(evaluation.rejectionReason());
        assertEquals(granted, evaluation.passportGranted());
        assertEquals(validUntil, evaluation.expiresAfter());
    }

    @NonNull
    private static Stream<Arguments> evaluateUnder18Source() {
        return of(
            defaultForVaccinationPass("No vaccinations").denied().build(),
            defaultForVaccinationPass("Completed vaccination where less than minimum period has passed").vaccination(completedMultiDoseWithAdditionalVaccinations(afterMin(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), 2)).denied().expire(expireDenied(afterMin(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed").vaccination(completedMultiDoseWithAdditionalVaccinations(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), 1)).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE_UNDER_18, true)).build(),
            defaultForVaccinationPass("Completed vaccination where less than maximum period has passed").vaccination(completedMultiDose(beforeMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE_UNDER_18, true))).granted().expire(expireGranted(beforeMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE_UNDER_18, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE_UNDER_18, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly maximum period has passed").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE_UNDER_18, true))).denied().expireNever().build(),
            defaultForVaccinationPass("Completed vaccination where more than maximum period has passed").vaccination(completedMultiDoseWithAdditionalVaccinations(afterMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE_UNDER_18, true), 1)).denied().expireNever().build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive pcr where more than suspension period has passed").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true))).labResults(positivePcr(beforeMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true))).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE_UNDER_18, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive pcr where exactly suspension period has passed").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true))).labResults(positivePcr(on(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true))).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE_UNDER_18, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive pcr is now").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true))).labResults(positivePcr(DEFAULT_NOW)).denied().expire(expireDenied(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive pcr is in future").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true))).labResults(positivePcr(DEFAULT_NOW.plusNanos(1))).denied().expire(expireDenied(DEFAULT_NOW.plusNanos(1), DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive pcr where less suspension period has passed").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true))).labResults(positivePcr(afterMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true))).denied().expire(expireDenied(afterMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true), DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive antigen where more than suspension period has passed").vaccination(completedMultiDoseWithAdditionalVaccinations(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), 2)).labResults(positiveAntigen(beforeMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true))).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE_UNDER_18, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive antigen where exactly suspension period has passed").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true))).labResults(positiveAntigen(on(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true))).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE_UNDER_18, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive antigen where less suspension period has passed").vaccination(completedMultiDose(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true))).labResults(positiveAntigen(afterMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true))).denied().expire(expireDenied(afterMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true), DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Completed vaccination where exactly minimum period has passed and positive antigen where less suspension period has passed and negative pcr test after positive antigen").vaccination(completedMultiDoseWithAdditionalVaccinations(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), 2)).labResults(positiveAntigenBeforeNegativePcr(afterMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE, true))).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_COMPLETED_MULTI_VACCINATION_SCHEDULE, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE_UNDER_18, true)).build(),
            defaultForVaccinationPass("Partial vaccination with one vaccination where exactly minimum period has passed").vaccination(partial(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true))).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Partial vaccination with 1st vaccination where exactly minimum period has passed and 2nd vaccination where less than minimum period has passed").vaccination(partial(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true), afterMin(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true))).granted().expire(expireGranted(afterMin(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true), DEFAULT_MAXIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true)).build(),
            defaultForVaccinationPass("Partial vaccination with 1st vaccination where exactly maximum minus minumum period has passed and 2nd vaccination administered now").vaccination(partial(on(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true), DEFAULT_NOW)).granted().expire(expireGranted(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT_AFTER_FIRST_VACCINATION_OF_PARTIALLY_COMPLETED_VACCINATION_SCHEDULE, true)).build()

        );
    }

    @Nested
    @ExtendWith(MockitoExtension.class)
    class EdgeCasesUnder18 {

        @Mock
        private CprFormatBasedAgeCalculator ageCalculator;

        private VaccinationPassportEvaluator vaccinationPassportEvaluator;

        @BeforeEach
        void setUp() {
            final VaccinationPassportConfiguration configuration = defaultForVaccinationPass("").configuration();
            vaccinationPassportEvaluator = new VaccinationPassportEvaluator(configuration, ageCalculator);
            when(ageCalculator.isUnder18(anyString(), any())).thenReturn(true);
        }

        @Test
        void vaccinationPassportValidUntil18thBirthday() {
            final LocalDateTime now = LocalDateTime.now();

            // Birthday in 20 days
            final LocalDateTime dateOf18thBirthday = now.plusDays(20).toLocalDate().atStartOfDay();

            when(ageCalculator.timeUntil18(anyString(), any())).thenReturn(Duration.between(now, dateOf18thBirthday));

            // vaccination is expired if over 18
            final Vaccination vaccination = completedMultiDose(now.minus(DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE).minusDays(10));

            final PassportEvaluation evaluation = vaccinationPassportEvaluator.evaluate(vaccination, Data.Results.EMPTY, now);

            assertTrue(evaluation.passportGranted());
            assertEquals(dateOf18thBirthday, evaluation.expiresAfter());
        }

        @Test
        void vaccinationPassportValidUntilDefaultMaxPeriod() {
            final LocalDateTime now = LocalDateTime.now();

            // Birthday in 20 days
            final LocalDateTime dateOf18thBirthday = now.plusDays(20).toLocalDate().atStartOfDay();

            when(ageCalculator.timeUntil18(anyString(), any())).thenReturn(Duration.between(now, dateOf18thBirthday));

            // vaccination will expire in 30 days
            final LocalDateTime effectuationTime = now.minus(DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE).plusDays(30);
            final Vaccination vaccination = completedMultiDose(effectuationTime);

            final PassportEvaluation evaluation = vaccinationPassportEvaluator.evaluate(vaccination, Data.Results.EMPTY, now);

            // passport should be valid for 30 days
            assertTrue(evaluation.passportGranted());
            assertEquals(effectuationTime.plus(DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE), evaluation.expiresAfter());
        }

        @Test
        void vaccinationPassportValidUntilExemptionPeriod() {
            final LocalDateTime now = LocalDateTime.now();

            // Birthday after exemption period
            final LocalDateTime dateOf18thBirthday = now.plus(DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE_UNDER_18).plusDays(20).toLocalDate().atStartOfDay();

            when(ageCalculator.timeUntil18(anyString(), any())).thenReturn(Duration.between(now, dateOf18thBirthday));

            // vaccination will expire in 30 days
            final LocalDateTime effectuationTime = now.minus(DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE).plusDays(30);
            final Vaccination vaccination = completedMultiDose(effectuationTime);

            final PassportEvaluation evaluation = vaccinationPassportEvaluator.evaluate(vaccination, Data.Results.EMPTY, now);

            // passport should be valid for 30 days
            assertTrue(evaluation.passportGranted());
            assertEquals(effectuationTime.plus(DEFAULT_MAXIMUM_AMOUNT_AFTER_COMPLETED_VACCINATION_SCHEDULE_UNDER_18), evaluation.expiresAfter());
        }
    }
}
