/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.rules.cucumber;

import dk.sundhedsdatastyrelsen.cbs.Objects;
import dk.sundhedsdatastyrelsen.cbs.rules.*;
import dk.sundhedsdatastyrelsen.cbs.rules.configuration.TemporalAmountWithTimeInstruction;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import java.time.Duration;
import java.time.LocalDateTime;
import java.time.Period;
import java.time.format.DateTimeParseException;
import java.time.temporal.TemporalAmount;
import java.util.HashMap;
import java.util.Map;

import static dk.sundhedsdatastyrelsen.cbs.rules.configuration.TemporalAmountWithTimeInstruction.of;
import static dk.sundhedsdatastyrelsen.cbs.rules.configuration.VaccinationPassportConfigurationBuilder.configuration;
import static java.util.Objects.requireNonNull;


public class RuleConfiguration {

    private static final String NOW_VARIABLE = "DD";
    private static final String NOW_VALUE = "NOW";
    private static final LocalDateTime NOW = LocalDateTime.now();
    private static final String VAC_COMP_SINGLE_MIN = "VacGyldigStartEnkel";
    private static final String VAC_COMP_MULTI_MIN = "VacGyldigStartMultiFærdig";
    private static final String VAC_COMP_MAX = "VacGyldigSlutFærdig";
    private static final String VAC_COMP_SUSP = "VacSuspFærdig";
    private static final String VAC_PART_MIN = "VacGyldigStartMultiBegyndt";
    private static final String VAC_PART_MAX = "VacGyldigSlutMultiBegyndt";
    private static final String VAC_PART_SUSP = "VacSuspMultiBegyndt";
    private static final String TEST_MIN = "TestGyldigStart";
    private static final String TEST_MAX_PCR = "TestGyldigSlutPcr";
    private static final String TEST_MAX_AG = "TestGyldigSlutAg";
    private static final String TEST_SUSP = "TestSusp";
    private static final String RECOVERY_MIN = "ImmunitetGyldigStart";
    private static final String RECOVERY_MAX = "ImmunitetGyldigSlut";
    private static final String RECOVERY_SUSP = "SygdomsForløbLængde";


    private TemporalAmountWithTimeInstruction minimumAmountAfterCompletedSingleVaccinationSchedule;
    private TemporalAmountWithTimeInstruction minimumAmountAfterCompletedMultiVaccinationSchedule;
    private TemporalAmountWithTimeInstruction maximumAmountAfterCompletedVaccinationSchedule;
    private TemporalAmountWithTimeInstruction suspensionAmountAfterCompletedVaccinationSchedule;
    private TemporalAmountWithTimeInstruction minimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule;
    private TemporalAmountWithTimeInstruction maximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule;
    private TemporalAmountWithTimeInstruction suspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule;
    private TemporalAmountWithTimeInstruction maximumAmountAfterCompletedVaccinationScheduleUnder18;


    private TemporalAmount testMinimum;
    private boolean testKeepTimeOnMinimum;
    private TemporalAmount testMaximumPcr;
    private TemporalAmount testMaximumAg;
    private boolean testKeepTimeOnMaximum;
    private TemporalAmount testSuspension;
    private boolean testKeepTimeOnSuspension;
    private TemporalAmount recoveryMinimum;
    private boolean recoveryKeepTimeOnMinimum;
    private TemporalAmount recoveryMaximum;
    private boolean recoveryKeepTimeOnMaximum;
    private TemporalAmount recoverySuspension;
    private boolean recoveryKeepTimeOnSuspension;
    private LocalDateTime now;

    private final Map<String, TemporalAmount> variableToAmount = new HashMap<>();
    private final Map<String, LocalDateTime> variableToDateTime = new HashMap<>();

    public void add(@NonNull final String variable, final String value, final String keepTimeString) {
        final LocalDateTime localDateTime = parseDateTime(value);
        if (localDateTime != null) variableToDateTime.putIfAbsent(variable, localDateTime);

        final TemporalAmount amount = parseAmount(value);
        if (amount != null) variableToAmount.putIfAbsent(variable, amount);

        final boolean keepTime = Boolean.parseBoolean(keepTimeString);
        switch (variable) {
            case VAC_COMP_SINGLE_MIN:
                minimumAmountAfterCompletedSingleVaccinationSchedule = of(amount, keepTime);
                return;
            case VAC_COMP_MULTI_MIN:
                minimumAmountAfterCompletedMultiVaccinationSchedule = of(amount, keepTime);
                return;
            case VAC_COMP_MAX:
                maximumAmountAfterCompletedVaccinationSchedule = of(amount, keepTime);
                // TODO: 09-12-2021 workaround!. make test for this in stead
                maximumAmountAfterCompletedVaccinationScheduleUnder18 = of(amount, keepTime);
                return;
            case VAC_COMP_SUSP:
                suspensionAmountAfterCompletedVaccinationSchedule = of(amount, keepTime);
                return;
            case VAC_PART_MIN:
                minimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule = of(amount, keepTime);
                return;
            case VAC_PART_MAX:
                maximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule = of(amount, keepTime);
                return;
            case VAC_PART_SUSP:
                suspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule = of(amount, keepTime);
                return;
            case TEST_MIN:
                testMinimum = amount;
                this.testKeepTimeOnMinimum = keepTime;
                return;
            case TEST_MAX_PCR:
                testMaximumPcr = amount;
                testKeepTimeOnMaximum = keepTime;
                return;
            case TEST_MAX_AG:
                testMaximumAg = amount;
                testKeepTimeOnMaximum = keepTime;
                return;
            case TEST_SUSP:
                testSuspension = amount;
                testKeepTimeOnSuspension = keepTime;
                return;
            case RECOVERY_MIN:
                recoveryMinimum = amount;
                recoveryKeepTimeOnMinimum = keepTime;
                return;
            case RECOVERY_MAX:
                recoveryMaximum = amount;
                recoveryKeepTimeOnMaximum = keepTime;
                return;
            case RECOVERY_SUSP:
                recoverySuspension = amount;
                recoveryKeepTimeOnSuspension = keepTime;
                return;
            case NOW_VARIABLE:
                now = localDateTime;
        }
    }

    @NonNull
    public VaccinationPassportEvaluator vaccinationPassportEvaluator() {
        return new VaccinationPassportEvaluator(
            configuration()
                .minimumAmountAfterCompletedSingleVaccinationSchedule(minimumAmountAfterCompletedSingleVaccinationSchedule)
                .minimumAmountAfterCompletedMultiVaccinationSchedule(minimumAmountAfterCompletedMultiVaccinationSchedule)
                .maximumAmountAfterCompletedVaccinationSchedule(maximumAmountAfterCompletedVaccinationSchedule)
                .suspensionAmountAfterCompletedVaccinationSchedule(suspensionAmountAfterCompletedVaccinationSchedule)
                .minimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(minimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule)
                .maximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(maximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule)
                .suspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(suspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule)
                .maximumPeriodAfterCompletedVaccinationScheduleUnder18(maximumAmountAfterCompletedVaccinationScheduleUnder18)
                .build(), new CprFormatBasedAgeCalculator() {});
    }

    @NonNull
    public TestPassportEvaluator testPassportEvaluator() {
        return new TestPassportEvaluator(
            testMinimum,
            testKeepTimeOnMinimum,
            testMaximumPcr,
            testMaximumAg,
            testKeepTimeOnMaximum,
            testSuspension,
            testKeepTimeOnSuspension);
    }

    @NonNull
    public RecoveryPassportEvaluator recoveryPassportEvaluator() {
        return new RecoveryPassportEvaluator(
            recoveryMinimum,
            recoveryKeepTimeOnMinimum,
            recoveryMaximum,
            recoveryKeepTimeOnMaximum,
            recoverySuspension,
            recoveryKeepTimeOnSuspension);
    }

    @NonNull
    public EuRecoveryPassportEvaluator euRecoveryPassportEvaluator() {
        return new EuRecoveryPassportEvaluator(recoveryMinimum, recoveryMaximum);
    }

    @NonNull
    public TemporalAmount resolveAmount(@NonNull final String value) {
        final TemporalAmount variableAmount = variableToAmount.get(value);
        return variableAmount != null ? variableAmount : parseAmount(requireNonNull(value));
    }

    @Nullable
    public LocalDateTime resolveDateTime(@NonNull final String variable) {
        return Objects.firstNonNull(
            variableToDateTime.get(variable),
            parseDateTime(variable));
    }

    @Nullable
    private static LocalDateTime parseDateTime(@NonNull final String value) {
        if (NOW_VALUE.equalsIgnoreCase(value)) return NOW;
        try {
            return LocalDateTime.parse(value);
        } catch (DateTimeParseException e) {
            return null;
        }
    }

    @Nullable
    private static TemporalAmount parseAmount(@NonNull final String value) {
        if (value.startsWith("PT")) return Duration.parse(value);
        if (value.startsWith("P")) return Period.parse(value);
        return null;
    }

    public LocalDateTime now() {
        return now;
    }
}
