/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.rules.cucumber;

import com.google.common.base.Strings;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultsBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.EffectuatedVaccinationBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.VaccinationBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.rules.cucumber.expression.ExpressionEvaluator;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluation;
import io.cucumber.datatable.DataTable;
import io.cucumber.java8.En;
import io.cucumber.junit.platform.engine.Cucumber;
import org.apache.commons.lang3.StringUtils;

import java.time.LocalDateTime;

import static dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultBuilder.antigenResult;
import static dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultBuilder.pcrResult;
import static dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultsBuilder.cpr;
import static dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.VaccinationBuilder.vaccination;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assumptions.assumeFalse;

@Cucumber
public class Steps implements En {

    private static final String COMPLETED = "vaccinated";
    private static final String GRANTED = "gyldig";
    private static final String POSITIVE = "positiv";
    private static final String MULTI_DOSE = "multi-dose";

    private final RuleConfiguration configuration;
    private boolean isEuScenario = false;
    private PassportEvaluation testPassport;
    private PassportEvaluation vaccinationPassport;
    private PassportEvaluation recoveryPassport;
    private ExpressionEvaluator evaluator;
    private boolean vaccinationCompleted;
    private VaccinationBuilder vaccinationBuilder;
    private final String cpr = "010101010101";
    private Vaccination vaccination;
    private LabResults results;
    private final LabResultsBuilder labResultsBuilder;
    private boolean multiDose;

    public Steps() {
        configuration = new RuleConfiguration();
        labResultsBuilder = cpr(cpr);
        vaccinationBuilder = vaccination(cpr, false);

        Given("the following variables:", (DataTable table) -> {
            table.rows(1).cells().forEach(cell -> configuration.add(cell.get(0), cell.get(1), cell.get(2)));
            evaluator = new ExpressionEvaluator(configuration::resolveDateTime, configuration::resolveAmount);
        });

        Given("the user is {} with a {}", (String status, String scheduleType) -> {
            vaccinationCompleted = COMPLETED.equals(status);
            multiDose = vaccinationCompleted && MULTI_DOSE.equals(scheduleType);
            vaccinationBuilder = vaccination(cpr, vaccinationCompleted);
        });
        And("the vaccination date is {} and the user has {} additional vaccinations after that date", (String vaccinationDate, String additionalVaccinations) -> {
            if ("".equals(vaccinationDate)) return;

            final LocalDateTime completedDate = evaluator.calculate(vaccinationDate);
            assumeFalse(completedDate == null, "completion date is null");
            if (multiDose) vaccinationBuilder.add(EffectuatedVaccinationBuilder.effectuatedVaccination(1L, completedDate.minusDays(40), "DK").build());
            vaccinationBuilder.add(EffectuatedVaccinationBuilder.effectuatedVaccination(2L, completedDate, "DK").build());
            if (StringUtils.isNotBlank(additionalVaccinations)) {
                if (!vaccinationCompleted) throw new IllegalArgumentException("Cannot add additional vaccinations when vaccination status is not " + COMPLETED);
                LocalDateTime additionalVaccinationCompletedDate = completedDate.plusMonths(6);
                for (int i = 0; i < Integer.parseInt(additionalVaccinations); i++) {
                    vaccinationBuilder.addAdditional(EffectuatedVaccinationBuilder.effectuatedVaccination(3L, additionalVaccinationCompletedDate, "DK").build());
                    additionalVaccinationCompletedDate = additionalVaccinationCompletedDate.plusMonths(6);
                }
            }
        });

        And("the user has a {} pcr test taken {}", (String result, String expression) -> {
            if ("".equals(result)) return;

            final boolean positive = POSITIVE.equalsIgnoreCase(result);
            final LocalDateTime samplingTime = evaluator.calculate(expression);
            assumeFalse(samplingTime == null, "completion date is null");

            labResultsBuilder.add(pcrResult(1, samplingTime, samplingTime.plusHours(26)).positive(positive).build());
        });

        And("the user has a {} antigen test taken {}", (String result, String expression) -> {
            if ("".equals(result)) return;

            final boolean positive = POSITIVE.equalsIgnoreCase(result);
            final LocalDateTime samplingTime = evaluator.calculate(expression);
            assumeFalse(samplingTime == null, "completion date is null");

            labResultsBuilder.add(antigenResult(1, samplingTime, samplingTime.plusMinutes(15)).positive(positive).build());
        });

        When("the passport is evaluated", () -> {
            vaccination = vaccinationBuilder.build();
            results = labResultsBuilder.build();
            if (!isEuScenario) {
                vaccinationPassport = configuration.vaccinationPassportEvaluator().evaluate(vaccination, results, configuration.now());
            }
            if (!isEuScenario) {
                testPassport = configuration.testPassportEvaluator().evaluate(vaccination, results, configuration.now());
            }
            if (isEuScenario) {
                recoveryPassport = configuration.euRecoveryPassportEvaluator().evaluate(vaccination, results, configuration.now());
            } else {
                recoveryPassport = configuration.recoveryPassportEvaluator().evaluate(vaccination, results, configuration.now());
            }
        });
        Then("the vaccination passport is {} with the expiration {} and rejection reason {}", (String state, String expression, String rejectionReason) -> {
            final boolean expectedPassGranted = GRANTED.equalsIgnoreCase(state);
            final boolean actualPassGranted = vaccinationPassport.passportGranted();
            final LocalDateTime expectedExpiration = evaluator.calculate(expression);

            assertEquals(Strings.emptyToNull(rejectionReason), vaccinationPassport.rejectionReason());
            assertEquals(expectedPassGranted, actualPassGranted);
            if (actualPassGranted) {
                assertEquals(expectedExpiration, vaccinationPassport.expiresAfter());
            }
        });
        And("the test passport is {} with the expiration {} and rejection reason {}", (String state, String expression, String rejectionReason) -> {
            final boolean expectedPassGranted = GRANTED.equalsIgnoreCase(state);
            final boolean actualPassGranted = testPassport.passportGranted();
            final LocalDateTime expectedExpiration = evaluator.calculate(expression);

            assertEquals(Strings.emptyToNull(rejectionReason), testPassport.rejectionReason());
            assertEquals(expectedPassGranted, actualPassGranted);
            if (actualPassGranted) {
                assertEquals(expectedExpiration, testPassport.expiresAfter());
            }
        });

        And("the recovery passport is {} with the expiration {} and rejection reason {}", (String state, String expression, String rejectionReason) -> {
            final boolean expectedPassGranted = GRANTED.equalsIgnoreCase(state);
            final boolean actualPassGranted = recoveryPassport.passportGranted();
            final LocalDateTime expectedExpiration = evaluator.calculate(expression);

            assertEquals(Strings.emptyToNull(rejectionReason), recoveryPassport.rejectionReason());
            assertEquals(expectedPassGranted, actualPassGranted);
            if (actualPassGranted) {
                assertEquals(expectedExpiration, recoveryPassport.expiresAfter());
            }
        });

        Given("This is a {} passport evaluation", (String passportType) -> {
            if ("EU".equalsIgnoreCase(passportType)) {
                isEuScenario = true;
            }
        });
    }
}
