/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.rules.cucumber.expression;

import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import java.time.LocalDateTime;
import java.time.temporal.TemporalAmount;

import static java.lang.String.format;

public interface State {

    boolean valid();

    @Nullable
    default String error() { return null; }

    @Nullable
    LocalDateTime value();

    @NonNull
    State next(@NonNull final String token);

    static State initial(
        @NonNull final Resolver<String, LocalDateTime> dateTimeResolver,
        @NonNull final Resolver<String, TemporalAmount> amountResolver) {
        return new AbstractState.InitialState(dateTimeResolver, amountResolver);
    }

    abstract class AbstractState implements State {

        @NonNull protected final Resolver<String, LocalDateTime> dateTimeResolver;
        @NonNull protected final Resolver<String, TemporalAmount> amountResolver;
        @Nullable protected final LocalDateTime value;
        private final boolean failed;
        protected final int index;


        protected AbstractState(
            @NonNull final Resolver<String, LocalDateTime> dateTimeResolver,
            @NonNull final Resolver<String, TemporalAmount> amountResolver,
            @Nullable final LocalDateTime value,
            final boolean failed,
            final int index) {
            this.dateTimeResolver = dateTimeResolver;
            this.amountResolver = amountResolver;
            this.value = value;
            this.failed = failed;
            this.index = index;
        }

        @Override
        @Nullable
        public LocalDateTime value() {
            return value;
        }

        @Override
        public boolean valid() {
            return !failed;
        }

        private static final class InitialState extends AbstractState {

            private InitialState(
                @NonNull final Resolver<String, LocalDateTime> dateTimeResolver,
                @NonNull final Resolver<String, TemporalAmount> amountResolver) {
                super(dateTimeResolver, amountResolver, null, false, 0);
            }

            @Override
            @NonNull
            public State next(@NonNull final String variable) {
                final LocalDateTime value = dateTimeResolver.resolve(variable);
                if(value == null) return new FailedState(0, "Illegal datetime variable '%s'", variable);
                return new CalculationCompletedState(dateTimeResolver, amountResolver, value, variable.length());
            }
        }

        private static final class FailedState implements State {
            @NonNull private final String error;

            private FailedState(
                final int index,
                @NonNull final String errorFormat,
                @NonNull final Object... args) {
                this.error = format("%s at index %d", format(errorFormat, args), index);
            }

            @Override
            public boolean valid() {
                return false;
            }

            @Override
            @Nullable
            public LocalDateTime value() {
                return null;
            }

            @Override
            @NonNull
            public String error() {
                return error;
            }

            @Override
            @NonNull
            public State next(@NonNull final String token) {
                return this;
            }
        }

        private static final class CalculationPendingState extends AbstractState {
            @NonNull private static final String PLUS = "+";
            @NonNull private static final String MINUS = "-";
            @NonNull private final String operator;
            @NonNull private final LocalDateTime value;

            private CalculationPendingState(
                @NonNull final Resolver<String, LocalDateTime> dateTimeResolver,
                @NonNull final Resolver<String, TemporalAmount> amountResolver,
                @NonNull final LocalDateTime value,
                @NonNull final String operator,
                final int index) {
                super(dateTimeResolver, amountResolver, value, true, index);
                this.operator = operator;
                this.value = value;
            }

            @NonNull
            private static State operator(
                @NonNull final CalculationCompletedState current,
                @NonNull final String operator,
                final int index) {
                    return new CalculationPendingState(current.dateTimeResolver, current.amountResolver, current.value , operator, index + operator.length());
            }

            @Override
            @NonNull
            public State next(@NonNull final String variable) {
                final TemporalAmount amount = amountResolver.resolve(variable);
                if(amount == null) return new FailedState(index, "Illegal amount variable %s", variable);

                switch (operator) {
                    case PLUS:
                        return CalculationCompletedState.complete(this, value.plus(amount), index + variable.length());
                    case MINUS:
                        return CalculationCompletedState.complete(this, value.minus(amount), index + variable.length());
                    default:
                        return new FailedState(index, "Illegal operator '%s'", operator);
                }
            }
        }

        private static final class CalculationCompletedState extends AbstractState {

            @NonNull private final LocalDateTime value;

            protected CalculationCompletedState(
                @NonNull final Resolver<String, LocalDateTime> dateTimeResolver,
                @NonNull final Resolver<String, TemporalAmount> amountResolver,
                @NonNull final LocalDateTime value,
                int index) {
                super(dateTimeResolver, amountResolver, value, false, index);
                this.value = value;
            }

            @NonNull
            private static CalculationCompletedState complete(
                @NonNull final CalculationPendingState current,
                @NonNull final LocalDateTime value,
                @NonNull final int index) {
                return new CalculationCompletedState(current.dateTimeResolver, current.amountResolver, value, index);
            }

            @Override
            @NonNull
            public State next(@NonNull final String operator) {
                return CalculationPendingState.operator(this, operator, index);
            }
        }
    }
}
