/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service;

import com.google.common.collect.ImmutableMap;
import dk.sundhedsdatastyrelsen.cbs.dto.kafka.FetchType;
import dk.sundhedsdatastyrelsen.cbs.kafka.CbsKafkaProducer;
import dk.sundhedsdatastyrelsen.cbs.logging.AuditData;
import dk.sundhedsdatastyrelsen.cbs.logging.AuditLogger;
import dk.sundhedsdatastyrelsen.cbs.monitoring.IsAliveServlet;
import dk.sundhedsdatastyrelsen.cbs.service.cache.Data;
import dk.sundhedsdatastyrelsen.cbs.service.cache.DataReader;
import dk.sundhedsdatastyrelsen.cbs.service.model.CoronaPass;
import dk.sundhedsdatastyrelsen.cbs.service.model.CoronaPassResult;
import dk.sundhedsdatastyrelsen.cbs.service.passports.CombinedPassportFactory;
import dk.sundhedsdatastyrelsen.cbs.service.passports.model.Custody;
import org.apache.log4j.Logger;
import org.apache.log4j.MDC;
import org.springframework.lang.NonNull;
import org.springframework.stereotype.Service;

import java.util.Collection;
import java.util.List;
import java.util.Set;
import java.util.UUID;
import java.util.stream.Collectors;

import static dk.sundhedsdatastyrelsen.cbs.CbsLogging.JOB_ID_PROP;
import static dk.sundhedsdatastyrelsen.cbs.Objects.nonNullOrDefault;
import static dk.sundhedsdatastyrelsen.cbs.dto.kafka.FetchType.*;

/**
 * Entry point for interacting with the service inside a Spring ApplicationContext
 */
@Service
public class CoronaPassService {
    private static final Logger LOGGER = Logger.getLogger(CoronaPassService.class);
    private final CbsKafkaProducer cbsKafkaProducer;
    private final CombinedPassportFactory passportFactory;
    private final AuditLogger auditLogger;

    private final DataReader dataReader;

    public CoronaPassService(
        CbsKafkaProducer cbsKafkaProducer,
        CombinedPassportFactory passportFactory,
        AuditLogger auditLogger,
        DataReader dataReader
    ) {
        this.cbsKafkaProducer = cbsKafkaProducer;
        this.passportFactory = passportFactory;
        this.auditLogger = auditLogger;
        this.dataReader = dataReader;
        IsAliveServlet.setServiceName(ServiceConstants.COMPONENT_NAME);
    }

    @NonNull
    public CoronaPassResult getCoronaPass(String cpr, String idwsMessageId) {
        final Collection<Data> data = dataReader.getData(cpr);

        final AuditData auditData = new AuditData();

        data.forEach(d -> auditData
            .withSensitivePersonalInformation("personIdentifier" + Custody.toSuffix(d.custody()), d.personRecord().personIdentifier())
            .withNonPersonalInformation("cacheHit" + Custody.toSuffix(d.custody()), !d.isCacheMiss())
        );

        if (data.stream().anyMatch(Data::isCacheMiss)) {
            String jobId = createJob(data, cpr, idwsMessageId, auditData);
            return new CoronaPassResult(jobId);
        }

        LOGGER.info("Cache hit");

        final Set<CoronaPass> coronaPasses = data.stream()
            .flatMap(d -> passportFactory.createPassportFrom(d.vaccination(), d.labResults(), d.personRecord(), auditData, d.custody()).stream())
            .collect(Collectors.toSet());

        auditLogger.log(auditData, "GetCoronaPass");
        return new CoronaPassResult(coronaPasses);
    }

    @NonNull
    private String createJob(
        Collection<Data> data,
        @NonNull final String cpr,
        @NonNull final String idwsMessageId,
        AuditData auditData) {

        final ImmutableMap.Builder<String, FetchType> builder = ImmutableMap.builder();

        data.stream()
            .filter(Data::isCacheMiss)
            .forEach(d -> builder.put(d.personRecord().personIdentifier(), getFetchType(d)));

        final ImmutableMap<String, FetchType> fetchJobs = builder.build();

        final List<String> fetchTypes = fetchJobs.values().stream().map(FetchType::name).collect(Collectors.toList());

        LOGGER.info("Cache miss: Creating jobs: " + fetchTypes);

        final String jobId = createDataFetcherJob(cpr, idwsMessageId, fetchJobs);

        auditLogger.log(auditData
            .withNonPersonalInformation("jobId", jobId)
            .withNonPersonalInformation("fetchTypes", fetchTypes), "GetCoronaPass");

        return jobId;
    }

    private FetchType getFetchType(Data data) {
        final FetchType vaccinationFetchType = data.vaccination() == null ? FETCH_VACCINATION_STATUS : null;
        final FetchType labResultFetchType = data.labResults() == null ? FETCH_TEST_STATUS : null;
        // FetchType.of can potentially return null, if the code is changed.
        // In that case it is safest to create a job fetching everything.
        return nonNullOrDefault(FetchType.of(vaccinationFetchType, labResultFetchType), FETCH_ALL);
    }

    /**
     * Creates a DataFetcher job, async (and adds jobId to the logging MDC, removed by LoggingSoapHandler#close)
     *
     * @return jobId to be returned in IDWS response, which is the id DataFetcher will use when notifying the job has finished to app-backend via NAS
     */
    public String createDataFetcherJob(String cpr, String idwsMessageId, ImmutableMap<String, FetchType> fetchJobs) {
        String jobId = UUID.randomUUID().toString();
        MDC.put(JOB_ID_PROP, jobId);

        cbsKafkaProducer.sendDataFetcherJob(cpr, jobId, idwsMessageId, fetchJobs);

        return jobId;
    }

}
