/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.cache;

import com.fasterxml.jackson.databind.ObjectMapper;
import dk.sundhedsdatastyrelsen.cbs.dao.ChildRelation;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.PersonRecordImpl;
import dk.sundhedsdatastyrelsen.cbs.datacache.CollectionReader;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.service.exceptions.MissingInformationException;
import dk.sundhedsdatastyrelsen.cbs.service.passports.model.Custody;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;

import java.io.IOException;
import java.io.Reader;
import java.io.UncheckedIOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Function;
import java.util.stream.Collectors;

public class DataReader {


    private final NamedParameterJdbcTemplate jdbcTemplate;
    private final ObjectMapper objectMapper;
    private final boolean includeChildren;

    private final CollectionReader<ChildRelation, String> childRelationReader;

    public DataReader(NamedParameterJdbcTemplate jdbcTemplate, ObjectMapper objectMapper, CollectionReader<ChildRelation, String> childRelationReader, boolean includeChildren) {
        this.jdbcTemplate = jdbcTemplate;
        this.objectMapper = objectMapper;
        this.childRelationReader = childRelationReader;
        this.includeChildren = includeChildren;
    }

    public Collection<Data> getData(String cpr) {

        final List<String> personIdentifiers = includeChildren
            ? childRelationReader.search(cpr).stream()
            .map(ChildRelation::childIdentifier)
            .collect(Collectors.toList())
            : new ArrayList<>();

        personIdentifiers.add(cpr);

        final MapSqlParameterSource paramSource = new MapSqlParameterSource("cpr", personIdentifiers);

        final Map<String, Vaccination> vaccinationMap = jdbcTemplate.query(
            "select data from vaccination_view where cpr in (:cpr)",
            paramSource,
            (rs, rn) -> map(rs.getCharacterStream("data"), Vaccination.class)).stream()
            .collect(Collectors.toMap(Vaccination::cpr, Function.identity()));

        final Map<String, LabResults> labResultsMap = jdbcTemplate.query(
            "select data from lab_result_view where cpr in (:cpr)",
            paramSource,
            (rs, rn) -> map(rs.getCharacterStream("data"), LabResults.class)).stream()
            .collect(Collectors.toMap(LabResults::cpr, Function.identity()));

        final Map<String, PersonRecordImpl> personRecordMap = jdbcTemplate.query(
            "select person_identifier, givenName, familyName from stamdata_person where person_identifier in (:cpr)",
            paramSource,
            (rs, rn) -> new PersonRecordImpl()
                .personIdentifier(rs.getString("person_identifier"))
                .givenName(rs.getString("givenName"))
                .familyName(rs.getString("familyName"))).stream()
            .collect(Collectors.toMap(PersonRecordImpl::personIdentifier, Function.identity()));

        final List<Data> data = personIdentifiers.stream()
            .map(pid -> new Data(
                vaccinationMap.get(pid),
                labResultsMap.get(pid),
                personRecordMap.get(pid)
            )).collect(Collectors.toList());

        if (data.stream().map(Data::personRecord).anyMatch(Objects::isNull)) {
            throw new MissingInformationException("Person data is missing");
        }

        final Function<Data, Custody> custodyFunction = custodyFunction(cpr);

        data.forEach(d -> d.custody(custodyFunction.apply(d)));

        return data;
    }

    private Function<Data, Custody> custodyFunction(String parentCpr) {
        AtomicInteger counter = new AtomicInteger(0);

        if (!includeChildren) {
            return data -> null;
        }

        return data -> parentCpr.equals(data.personRecord().personIdentifier())
            ? Custody.parent()
            : Custody.child(counter.incrementAndGet());
    }


    private <T> T map(Reader reader, Class<T> valueType) {
        if (reader == null) {
            return null;
        }

        try {
            return objectMapper.readValue(reader, valueType);
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        }
    }


}
