/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.config;

import dk.sundhedsdatastyrelsen.cbs.rules.*;
import dk.sundhedsdatastyrelsen.cbs.rules.configuration.VaccinationPassportConfiguration;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.CombinedPassportEvaluator;
import dk.sundhedsdatastyrelsen.cbs.service.passports.DkInfoPassportFactory;
import dk.sundhedsdatastyrelsen.cbs.service.passports.VaccineMapper;
import dk.sundhedsdatastyrelsen.cbs.service.passports.model.DkRules;
import dk.sundhedsdatastyrelsen.cbs.service.util.CprFormatBasedAgeCalculatorImpl;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.PropertySource;
import org.springframework.lang.NonNull;
import se.digg.dgc.service.DGCEncoder;

import java.time.temporal.TemporalAmount;

import static dk.sundhedsdatastyrelsen.cbs.rules.configuration.VaccinationPassportConfigurationBuilder.configuration;

@Configuration
@PropertySource("classpath:application-service.properties")
public class DkPassportConfiguration {

    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${cwt.dk.maximum-validity:PT1H}')}")
    private TemporalAmount cwtDkMaximumValidity;

    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${cwt.dkinfo.lab-result.maximum-age:P7D}')}")
    private TemporalAmount cwtDkInfoLabResultMaxAge;


    //---------- Vaccination rule parameters ----------\\
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.vaccination.validator.minimum-amount-after-completed-single-vaccination-schedule:P14D}')}")
    private TemporalAmount minimumAmountAfterCompletedSingleVaccinationSchedule;
    @Value("${dk.vaccination.validator.keep-time-on-minimum-amount-after-completed-single-vaccination-schedule:true}")
    private boolean keepTimeOnMinimumAmountAfterCompletedSingleVaccinationSchedule;

    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.vaccination.validator.minimum-amount-after-completed-multi-vaccination-schedule:P0D}')}")
    private TemporalAmount minimumAmountAfterCompletedMultiVaccinationSchedule;
    @Value("${dk.vaccination.validator.keep-time-on-minimum-amount-after-completed-multi-vaccination-schedule:true}")
    private boolean keepTimeOnMinimumAmountAfterCompletedMultiVaccinationSchedule;

    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.vaccination.validator.maximum-amount-after-completed-vaccination-schedule:P7M}')}")
    private TemporalAmount maximumAmountAfterCompletedVaccinationSchedule;
    @Value("${dk.vaccination.validator.keep-time-on-maximum-amount-after-completed-vaccination-schedule:true}")
    private boolean keepTimeOnMaximumAmountAfterCompletedVaccinationSchedule;

    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.vaccination.validator.suspension-amount-after-completed-vaccination-schedule:P14D}')}")
    private TemporalAmount suspensionAmountAfterCompletedVaccinationSchedule;
    @Value("${dk.vaccination.validator.keep-time-on-suspension-amount-after-completed-vaccination-schedule:true}")
    private boolean keepTimeOnSuspensionAmountAfterCompletedVaccinationSchedule;

    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.vaccination.validator.minimum-amount-after-first-vaccination-of-partially-completed-vaccination-schedule:P14D}')}")
    private TemporalAmount minimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule;
    @Value("${dk.vaccination.validator.keep-time-on-minimum-amount-after-first-vaccination-of-partially-completed-vaccination-schedule:true}")
    private boolean keepTimeOnMinimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule;

    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.vaccination.validator.maximum-amount-after-first-vaccination-of-partially-completed-vaccination-schedule:P42D}')}")
    private TemporalAmount maximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule;
    @Value("${dk.vaccination.validator.keep-time-on-maximum-amount-after-first-vaccination-of-partially-completed-vaccination-schedule:true}")
    private boolean keepTimeOnMaximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule;

    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.vaccination.validator.suspension-amount-after-first-vaccination-of-partially-completed-vaccination-schedule:P14D}')}")
    private TemporalAmount suspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule;
    @Value("${dk.vaccination.validator.keep-time-on-suspension-amount-after-first-vaccination-of-partially-completed-vaccination-schedule:true}")
    private boolean keepTimeOnSuspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule;

    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.vaccination.validator.maximum-amount-after-completed-vaccination-schedule-under-18:P5Y}')}")
    private TemporalAmount maximumAmountAfterCompletedVaccinationScheduleUnder18;
    @Value("${dk.vaccination.validator.keep-time-on-maximum-amount-after-completed-vaccination-schedule-under-18:true}")
    private boolean keepTimeOnMaximumAmountAfterCompletedVaccinationScheduleUnder18;


    //---------- Re-vaccination rule parameters ----------\\
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.revaccination.validator.minimum-period-after-revaccination:P0D}')}")
    private TemporalAmount minimumPeriodAfterRevaccination;
    @Value("${dk.revaccination.validator.keep-time-on-minimum-period-after-revaccination:true}")
    private boolean keepTimeOnMinimumPeriodAfterRevaccination;
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.revaccination.validator.maximum-period-after-revaccination:P3Y}')}")
    private TemporalAmount maximumPeriodAfterRevaccination;
    @Value("${dk.revaccination.validator.keep-time-on-maximum-period-after-revaccination:true}")
    private boolean keepTimeOnMaximumPeriodAfterRevaccination;
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.revaccination.validator.suspension-period-after-revaccination:P14D}')}")
    private TemporalAmount suspensionPeriodAfterRevaccination;
    @Value("${dk.revaccination.validator.keep-time-on-suspension-period-after-revaccination:true}")
    private boolean keepTimeOnSuspensionPeriodAfterRevaccination;


    //---------- Test rule parameters ----------\\
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.test.validator.minimum-period-after-test:PT0H}')}")
    private TemporalAmount minimumPeriodAfterTest;
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.test.validator.maximum-period-after-test.pcr:PT96H}')}")
    private TemporalAmount maximumPeriodAfterTestPcr;
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.test.validator.maximum-period-after-test.ag:PT72H}')}")
    private TemporalAmount maximumPeriodAfterTestAg;
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.test.validator.suspension-period-after-test:P14D}')}")
    private TemporalAmount suspensionPeriodAfterTest;
    @Value("${dk.test.validator.keep-time-on-minimum-period:true}")
    private boolean keepTimeOnMinimumPeriodAfterTest;
    @Value("${dk.test.validator.keep-time-on-maximum-period:true}")
    private boolean keepTimeOnMaximumPeriodAfterTest;
    @Value("${dk.test.validator.keep-time-on-suspension-period:true}")
    private boolean keepTimeOnSuspensionPeriodAfterTest;


    //---------- Recovery rule parameters ----------\\
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.recovery.validator.minimum-period-after-recovery:P14D}')}")
    private TemporalAmount minimumPeriodAfterRecovery;
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.recovery.validator.maximum-period-after-recovery:P12M}')}")
    private TemporalAmount maximumPeriodAfterRecovery;
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.recovery.validator.suspension-period-after-recovery:P84D}')}")
    private TemporalAmount suspensionPeriodAfterRecovery;
    @Value("${dk.recovery.validator.keep-time-on-minimum-period:true}")
    private boolean keepTimeOnMinimumPeriodAfterRecovery;
    @Value("${dk.recovery.validator.keep-time-on-maximum-period:true}")
    private boolean keepTimeOnMaximumPeriodAfterRecovery;
    @Value("${dk.recovery.validator.keep-time-on-suspension-period:true}")
    private boolean keepTimeOnSuspensionPeriodAfterRecovery;

    @Value("${cwt.children.include-in-response.dk-passes:false}")
    private boolean makeChildrenDkPasses;

    public TemporalAmount getCwtDkMaximumValidity() {
        return cwtDkMaximumValidity;
    }

    public boolean makeChildrenDkPasses() {
        return makeChildrenDkPasses;
    }

    @Bean
    @DkRules
    @NonNull
    public VaccinationPassportConfiguration dkVaccinationPassportConfiguration() {
        return configuration()
            .minimumAmountAfterCompletedSingleVaccinationSchedule(minimumAmountAfterCompletedSingleVaccinationSchedule, keepTimeOnMinimumAmountAfterCompletedSingleVaccinationSchedule)
            .minimumAmountAfterCompletedMultiVaccinationSchedule(minimumAmountAfterCompletedMultiVaccinationSchedule, keepTimeOnMinimumAmountAfterCompletedMultiVaccinationSchedule)
            .maximumAmountAfterCompletedVaccinationSchedule(maximumAmountAfterCompletedVaccinationSchedule, keepTimeOnMaximumAmountAfterCompletedVaccinationSchedule)
            .suspensionAmountAfterCompletedVaccinationSchedule(suspensionAmountAfterCompletedVaccinationSchedule, keepTimeOnSuspensionAmountAfterCompletedVaccinationSchedule)
            .minimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(minimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule, keepTimeOnMinimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule)
            .maximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(maximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule, keepTimeOnMaximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule)
            .suspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule(suspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule, keepTimeOnSuspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule)
            .maximumPeriodAfterCompletedVaccinationScheduleUnder18(maximumAmountAfterCompletedVaccinationScheduleUnder18, keepTimeOnMaximumAmountAfterCompletedVaccinationScheduleUnder18)
            .build();
    }

    @Bean
    @DkRules
    @NonNull
    public VaccinationPassportEvaluator dkVaccinationPassportEvaluator(
        @DkRules @NonNull final VaccinationPassportConfiguration configuration) {
        return new VaccinationPassportEvaluator(configuration, new CprFormatBasedAgeCalculatorImpl());
    }

    @Bean
    @DkRules
    @NonNull
    public RevaccinationPassportEvaluator dkRevaccinationPassportEvaluator() {
        return new RevaccinationPassportEvaluator(
            minimumPeriodAfterRevaccination,
            keepTimeOnMinimumPeriodAfterRevaccination,
            maximumPeriodAfterRevaccination,
            keepTimeOnMaximumPeriodAfterRevaccination,
            suspensionPeriodAfterRevaccination,
            keepTimeOnSuspensionPeriodAfterRevaccination);
    }

    @Bean
    @DkRules
    @NonNull
    public TestPassportEvaluator dkTestPassportEvaluator() {
        return new TestPassportEvaluator(
            minimumPeriodAfterTest, keepTimeOnMinimumPeriodAfterTest,
            maximumPeriodAfterTestPcr, maximumPeriodAfterTestAg, keepTimeOnMaximumPeriodAfterTest,
            suspensionPeriodAfterTest, keepTimeOnSuspensionPeriodAfterTest);
    }

    @Bean
    @DkRules
    @NonNull
    public RecoveryPassportEvaluator dkRecoveryPassportEvaluator() {
        return new RecoveryPassportEvaluator(
            minimumPeriodAfterRecovery, keepTimeOnMinimumPeriodAfterRecovery,
            maximumPeriodAfterRecovery, keepTimeOnMaximumPeriodAfterRecovery,
            suspensionPeriodAfterRecovery, keepTimeOnSuspensionPeriodAfterRecovery);
    }

    @Bean
    @DkRules
    @NonNull
    public CombinedPassportEvaluator dkCombinedPassportEvaluator(
        @DkRules @NonNull final VaccinationPassportEvaluator dkVaccinationPassportEvaluator,
        @DkRules @NonNull final RevaccinationPassportEvaluator dkRevaccinationPassportEvaluator,
        @DkRules @NonNull final TestPassportEvaluator dkTestPassportEvaluator,
        @DkRules @NonNull final RecoveryPassportEvaluator dkRecoveryPassportEvaluator) {
        return new CombinedPassportEvaluatorImpl(dkVaccinationPassportEvaluator, dkRevaccinationPassportEvaluator, dkTestPassportEvaluator, dkRecoveryPassportEvaluator);
    }

    @Bean
    DkInfoPassportFactory dkInfoPassportFactory(
        VaccineMapper vaccineMapper,
        DGCEncoder encoder,
        @DkRules VaccinationPassportConfiguration vaccinationPassportConfiguration,
        @DkRules RecoveryPassportEvaluator recoveryPassportEvaluator
    ) {
        return new DkInfoPassportFactory(
            vaccineMapper,
            encoder,
            recoveryPassportEvaluator,
            vaccinationPassportConfiguration,
            maximumPeriodAfterRevaccination,
            maximumPeriodAfterTestPcr,
            maximumPeriodAfterTestAg,
            cwtDkInfoLabResultMaxAge);
    }

    @Override
    public String toString() {
        return "DkPassportConfiguration{" +
            "cwtDkMaximumValidity=" + cwtDkMaximumValidity +
            ", cwtDkInfoLabResultMaxAge=" + cwtDkInfoLabResultMaxAge +
            ", makeChildrenDkPasses=" + makeChildrenDkPasses +
            ", minimumAmountAfterCompletedSingleVaccinationSchedule=" + minimumAmountAfterCompletedSingleVaccinationSchedule +
            ", keepTimeOnMinimumAmountAfterCompletedSingleVaccinationSchedule=" + keepTimeOnMinimumAmountAfterCompletedSingleVaccinationSchedule +
            ", minimumAmountAfterCompletedMultiVaccinationSchedule=" + minimumAmountAfterCompletedMultiVaccinationSchedule +
            ", keepTimeOnMinimumAmountAfterCompletedMultiVaccinationSchedule=" + keepTimeOnMinimumAmountAfterCompletedMultiVaccinationSchedule +
            ", maximumAmountAfterCompletedVaccinationSchedule=" + maximumAmountAfterCompletedVaccinationSchedule +
            ", keepTimeOnMaximumAmountAfterCompletedVaccinationSchedule=" + keepTimeOnMaximumAmountAfterCompletedVaccinationSchedule +
            ", suspensionAmountAfterCompletedVaccinationSchedule=" + suspensionAmountAfterCompletedVaccinationSchedule +
            ", keepTimeOnSuspensionAmountAfterCompletedVaccinationSchedule=" + keepTimeOnSuspensionAmountAfterCompletedVaccinationSchedule +
            ", minimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule=" + minimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule +
            ", keepTimeOnMinimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule=" + keepTimeOnMinimumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule +
            ", maximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule=" + maximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule +
            ", keepTimeOnMaximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule=" + keepTimeOnMaximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule +
            ", maximumAmountAfterCompletedVaccinationScheduleUnder18=" + maximumAmountAfterCompletedVaccinationScheduleUnder18 +
            ", keepTimeOnMaximumAmountAfterCompletedVaccinationScheduleUnder18=" + keepTimeOnMaximumAmountAfterCompletedVaccinationScheduleUnder18 +
            ", suspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule=" + suspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule +
            ", keepTimeOnSuspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule=" + keepTimeOnSuspensionAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule +
            ", minimumPeriodAfterRevaccination=" + minimumPeriodAfterRevaccination +
            ", keepTimeOnMinimumPeriodAfterRevaccination=" + keepTimeOnMinimumPeriodAfterRevaccination +
            ", maximumPeriodAfterRevaccination=" + maximumPeriodAfterRevaccination +
            ", keepTimeOnMaximumPeriodAfterRevaccination=" + keepTimeOnMaximumPeriodAfterRevaccination +
            ", suspensionPeriodAfterRevaccination=" + suspensionPeriodAfterRevaccination +
            ", keepTimeOnSuspensionPeriodAfterRevaccination=" + keepTimeOnSuspensionPeriodAfterRevaccination +
            ", minimumPeriodAfterTest=" + minimumPeriodAfterTest +
            ", maximumPeriodAfterTestPcr=" + maximumPeriodAfterTestPcr +
            ", maximumPeriodAfterTestAg=" + maximumPeriodAfterTestAg +
            ", suspensionPeriodAfterTest=" + suspensionPeriodAfterTest +
            ", keepTimeOnMinimumPeriodAfterTest=" + keepTimeOnMinimumPeriodAfterTest +
            ", keepTimeOnMaximumPeriodAfterTest=" + keepTimeOnMaximumPeriodAfterTest +
            ", keepTimeOnSuspensionPeriodAfterTest=" + keepTimeOnSuspensionPeriodAfterTest +
            ", minimumPeriodAfterRecovery=" + minimumPeriodAfterRecovery +
            ", maximumPeriodAfterRecovery=" + maximumPeriodAfterRecovery +
            ", suspensionPeriodAfterRecovery=" + suspensionPeriodAfterRecovery +
            ", keepTimeOnMinimumPeriodAfterRecovery=" + keepTimeOnMinimumPeriodAfterRecovery +
            ", keepTimeOnMaximumPeriodAfterRecovery=" + keepTimeOnMaximumPeriodAfterRecovery +
            ", keepTimeOnSuspensionPeriodAfterRecovery=" + keepTimeOnSuspensionPeriodAfterRecovery +
            '}';
    }
}
