/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.config;

import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import dk.sundhedsdatastyrelsen.cbs.dao.EUAntigenSpec;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.EUAntigenSpecImpl;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.RecordMappers;
import dk.sundhedsdatastyrelsen.cbs.datacache.CollectionReader;
import dk.sundhedsdatastyrelsen.cbs.datacache.impl.Readers;
import dk.sundhedsdatastyrelsen.cbs.service.util.EUAntigenFiltration;
import dk.sundhedsdatastyrelsen.fluent.spring.jdbc.SearchMapper;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.datasource.lookup.JndiDataSourceLookup;
import org.springframework.lang.NonNull;

import javax.sql.DataSource;
import java.util.Collection;
import java.util.concurrent.TimeUnit;

@Configuration
public class EUAntigenSpecConfiguration {

    @Value("${euantigenspec.use-cache:false}")
    private boolean euAntigenSpecUseCache;

    @Value("${euantigenspec.cache-time-to-live-seconds:300}")
    private int euAntigenSpecCacheSeconds;

    @Value("${euantigenspec.datasource.jndi-name:java:jboss/datasources/euantigenspec}")
    private String dataSourceJndiName;

    @Bean
    @NonNull
    public CollectionReader<EUAntigenSpec, String> euAntigenSpecReader(
        @NonNull final JdbcTemplate euAntigenSpecJdbcTemplate,
        @NonNull final SearchMapper<EUAntigenSpecImpl, String> mapper) {
        return Readers.reader(euAntigenSpecJdbcTemplate, mapper);
    }

    @Bean
    @NonNull
    public JdbcTemplate euAntigenSpecJdbcTemplate(@NonNull final DataSource euAntigenSpecDataSource) {
        return new JdbcTemplate(euAntigenSpecDataSource);
    }

    @Bean
    @NonNull
    public DataSource euAntigenSpecDataSource() {
        return new JndiDataSourceLookup().getDataSource(dataSourceJndiName);
    }

    @Bean
    @NonNull
    public SearchMapper<EUAntigenSpecImpl, String> euAntigenSpecMapper() {
        return RecordMappers.euAntigenSpecMapper(
            EUAntigenSpec.CATALOG_NAME,
            EUAntigenSpec.SCHEMA_NAME,
            EUAntigenSpec.INSERT_TABLE_NAME,
            EUAntigenSpec.CATALOG_NAME,
            EUAntigenSpec.SCHEMA_NAME,
            EUAntigenSpec.SELECT_TABLE_NAME,
            EUAntigenSpecImpl::new)
            .build();
    }

    @Bean
    public EUAntigenFiltration euAntigenFiltration(LoadingCache<String, Collection<EUAntigenSpec>> euAntigenSpecCache,
                                                   CollectionReader<EUAntigenSpec, String> euAntigenSpecReader) {
        return new EUAntigenFiltration(euAntigenSpecCache, euAntigenSpecReader, euAntigenSpecUseCache);
    }

    @Bean
    @NonNull
    public LoadingCache<String, Collection<EUAntigenSpec>> euAntigenSpecCache(CollectionReader<EUAntigenSpec, String> euAntigenSpecReader) {
        return CacheBuilder.newBuilder()
            .expireAfterAccess(euAntigenSpecCacheSeconds, TimeUnit.SECONDS)
            .build(CacheLoader.from(euAntigenSpecReader::search));
    }

}
