/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.config;

import com.fasterxml.jackson.databind.ObjectMapper;
import dk.sundhedsdatastyrelsen.cbs.audit.SystemAuditLogger;
import dk.sundhedsdatastyrelsen.cbs.dao.ChildRelation;
import dk.sundhedsdatastyrelsen.cbs.datacache.CollectionReader;
import dk.sundhedsdatastyrelsen.cbs.datacache.status.JdbcStatusChecker;
import dk.sundhedsdatastyrelsen.cbs.logging.AuditLogger;
import dk.sundhedsdatastyrelsen.cbs.service.ServiceConstants;
import dk.sundhedsdatastyrelsen.cbs.service.cache.DataReader;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.PropertySource;
import org.springframework.context.support.PropertySourcesPlaceholderConfigurer;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;

import java.nio.charset.StandardCharsets;
import java.util.Collections;
import java.util.zip.CRC32;

@Configuration
@PropertySource("classpath:application-service.properties")
public class ServiceConfig {
    /**
     * Bean that ensures that the application context fail if properties are missing
     */
    @Bean
    public static PropertySourcesPlaceholderConfigurer propertyPlaceholderConfigurer() {
        return new PropertySourcesPlaceholderConfigurer();
    }

    @Bean
    AuditLogger auditLogger(DkPassportConfiguration dkPassportConfiguration, EuPassportConfiguration euPassportConfiguration, @Value("${cwt.children.include-in-response:true}") boolean fetchChildPasses) {

        final String dkPassportConfigurationText = dkPassportConfiguration.toString();
        final String euPassportConfigurationText = euPassportConfiguration.toString();

        final CRC32 crc32 = new CRC32();
        crc32.update((dkPassportConfigurationText + euPassportConfigurationText + fetchChildPasses).getBytes(StandardCharsets.UTF_8));
        final long checksum = crc32.getValue();

        SystemAuditLogger.newAudit()
            .addEntry("DkPassportConfiguration", dkPassportConfigurationText)
            .addEntry("EuPassportConfiguration", euPassportConfigurationText)
            .addEntry("cwt.children.include-in-response", fetchChildPasses)
            .addEntry("PassportConfigurationChecksum", checksum)
            .log();

        return new AuditLogger(ServiceConstants.COMPONENT_NAME, Collections.singletonMap("PassportConfigurationChecksum", checksum));
    }

    @Bean
    JdbcStatusChecker jdbcStatusChecker(JdbcTemplate jdbcTemplate) {
        return new JdbcStatusChecker(jdbcTemplate, "CBS-DB");
    }

    @Bean
    DataReader dataReader(JdbcTemplate jdbcTemplate, ObjectMapper objectMapper, CollectionReader<ChildRelation, String> childRelationReader, @Value("${cwt.children.include-in-response:true}") boolean fetchChildPasses) {
        return new DataReader(new NamedParameterJdbcTemplate(jdbcTemplate), objectMapper, childRelationReader, fetchChildPasses);
    }

}
