/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.handler;

import dk.sundhedsdatastyrelsen.cbs.service.ErrorMessages;
import dk.sundhedsdatastyrelsen.cbs.service.config.EUCovidTestCertificateSigningConfiguration;
import dk.sundhedsdatastyrelsen.cbs.service.exceptions.CertificateSigningException;
import dk.sundhedsdatastyrelsen.cbs.service.model.SignedCertificateCWT;
import dk.sundhedsdatastyrelsen.cbs.service.passports.ChecksumUtils;
import dk.sundhedsdatastyrelsen.cbs.service.passports.VaccineMapper;
import dk.sundhedsdatastyrelsen.cbs.service.util.EUAntigenFiltration;
import dk.sundhedsdatastyrelsen.cbs.service.util.PassportUtils;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.GetSignedEUCovidTestCertificateRequest;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.TestResult;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.TestSample;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.TestType;
import org.springframework.stereotype.Component;
import se.digg.dgc.payload.v1.DGCSchemaException;
import se.digg.dgc.payload.v1.DigitalGreenCertificate;
import se.digg.dgc.payload.v1.RecoveryEntry;
import se.digg.dgc.payload.v1.TestEntry;
import se.digg.dgc.service.DGCEncoder;

import java.io.IOException;
import java.security.SignatureException;
import java.time.Instant;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.Collections;
import java.util.UUID;

@Component
public class EUCovidTestCertificateSigningHandler {

    private final DGCEncoder encoder;
    private final EUAntigenFiltration euAntigenFiltration;
    private final EUCovidTestCertificateSigningConfiguration configuration;

    public EUCovidTestCertificateSigningHandler(
        DGCEncoder encoder,
        EUAntigenFiltration euAntigenFiltration,
        EUCovidTestCertificateSigningConfiguration configuration) {
        this.encoder = encoder;
        this.euAntigenFiltration = euAntigenFiltration;
        this.configuration = configuration;
    }

    /**
     *
     * @param request The request containing required data.
     *                The field values are never null, and null checks are therefore unnecessary.
     * @return Signed and encoded DGC with Certificate identifier
     */
    public SignedCertificateCWT getSignedCertificateCWT(GetSignedEUCovidTestCertificateRequest request) {
        final TestSample payload = request.getPayload();

        // check that sampleTime is not in the future, allowing for a configurable clock skew
        if(payload.getSampleCollected().toInstant().isAfter(Instant.now().plus(configuration.getAllowedClockSkew()))) {
            throw new CertificateSigningException(ErrorMessages.INVALID_FUTURE_SAMPLING_TIME);
        }

        if (payload.getTestResult().equals(TestResult.POSITIVE) && payload.getTypeOfTest().equals(TestType.PCR)) {
            return issueRecoveryCertificate(payload);
        } else {
            return issueTestCertificate(payload);
        }
    }

    private SignedCertificateCWT issueTestCertificate(TestSample payload) {
        if (payload.getTestResult().equals(TestResult.POSITIVE)) {
            throw new CertificateSigningException(ErrorMessages.POSITIVE_RESULT);
        }

        if (payload.getTypeOfTest().equals(TestType.RAT) && isNotApproved(payload)) {
            throw new CertificateSigningException(ErrorMessages.NOT_HSC_APPROVED);
        }

        final Instant sampleCollected = payload.getSampleCollected().toInstant();
        final Instant testExpiration = sampleCollected.plus(payload.getTypeOfTest().equals(TestType.PCR) ? configuration.getPcrResultsMaxAge() : configuration.getAntiGenResultsMaxAge());

        if (testExpiration.isBefore(Instant.now())) {
            throw new CertificateSigningException(ErrorMessages.TEST_IS_EXPIRED);
        }

        final DigitalGreenCertificate certificate = PassportUtils.buildDigitalGreenCertificate(payload.getGivenName(), payload.getFamilyName(), payload.getDateOfBirth());

        final String ci = ChecksumUtils.generateUVCI(UUID.randomUUID().toString());

        certificate.setT(Collections.singletonList(
            new TestEntry()
                .withIs(VaccineMapper.ISSUER)
                .withCo(payload.getCountryOfTest().value())
                .withCi(ci)
                .withTg(payload.getDiseaseAgentTargeted().value())
                .withSc(sampleCollected)
                .withMa(payload.getManufacturer())
                .withTt(payload.getTypeOfTest().value())
                .withTc(payload.getTestingCenter())
                .withTr(payload.getTestResult().value())
        ));

        final Instant maxExpiration = Instant.now().plus(configuration.getCwtTestResultMaxValidity());

        final String cwt = encode(certificate, PassportUtils.earliestInstant(maxExpiration, testExpiration));

        return new SignedCertificateCWT(cwt, ci);
    }

    private SignedCertificateCWT issueRecoveryCertificate(TestSample payload) {
        final DigitalGreenCertificate certificate = PassportUtils.buildDigitalGreenCertificate(payload.getGivenName(), payload.getFamilyName(), payload.getDateOfBirth());

        final ZonedDateTime firstPositiveResult = payload.getSampleCollected();

        final ZonedDateTime validFrom = firstPositiveResult.plus(configuration.getRecoveryMinimumPeriod());
        final ZonedDateTime validTo = firstPositiveResult.plus(configuration.getRecoveryMaximumPeriod());

        if (validTo.toInstant().isBefore(Instant.now())) {
            throw new CertificateSigningException(ErrorMessages.TEST_IS_EXPIRED);
        }

        final String ci = ChecksumUtils.generateUVCI(UUID.randomUUID().toString());

        certificate.setR(Collections.singletonList(
            new RecoveryEntry()
                .withCo(payload.getCountryOfTest().value())
                .withCi(ci)
                .withTg(payload.getDiseaseAgentTargeted().value())
                .withIs(VaccineMapper.ISSUER)
                .withFr(firstPositiveResult.toLocalDate())
                .withDf(validFrom.toLocalDate())
                .withDu(validTo.toLocalDate())
        ));

        final Instant maxExpiration = Instant.now().plus(configuration.getCwtRecoveryMaxValidity());

        final String cwt = encode(certificate, PassportUtils.earliestInstant(maxExpiration, validTo.toInstant()));

        return new SignedCertificateCWT(cwt, ci);
    }

    private boolean isNotApproved(TestSample payload) {
        return !euAntigenFiltration.isIdDeviceLegalAtTime(
            payload.getManufacturer(),
            payload.getSampleCollected().withZoneSameInstant(ZoneId.systemDefault()).toLocalDateTime()
        );
    }

    private String encode(DigitalGreenCertificate certificate, Instant expiration) {
        try {
            return encoder.encode(certificate, expiration);
        } catch (DGCSchemaException | IOException | SignatureException e) {
            throw new CertificateSigningException(ErrorMessages.ENCODING_ERROR, e);
        }
    }

}
