/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.passports;

import com.google.common.collect.Ordering;
import dk.sundhedsdatastyrelsen.cbs.dao.PersonRecord;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.logging.AuditData;
import dk.sundhedsdatastyrelsen.cbs.rules.CprFormatBasedAgeCalculator;
import dk.sundhedsdatastyrelsen.cbs.rules.RecoveryPassportEvaluator;
import dk.sundhedsdatastyrelsen.cbs.rules.configuration.VaccinationPassportConfiguration;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.RecoveryLabResult;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.RecoveryPassportEvaluation;
import dk.sundhedsdatastyrelsen.cbs.service.exceptions.CwtException;
import dk.sundhedsdatastyrelsen.cbs.service.exceptions.VaccineMapperException;
import dk.sundhedsdatastyrelsen.cbs.service.model.CoronaPass;
import dk.sundhedsdatastyrelsen.cbs.service.model.PassType;
import dk.sundhedsdatastyrelsen.cbs.service.passports.model.Custody;
import dk.sundhedsdatastyrelsen.cbs.service.passports.model.DkRules;
import dk.sundhedsdatastyrelsen.cbs.service.passports.model.EnhancedTestEntry;
import dk.sundhedsdatastyrelsen.cbs.service.passports.model.EnhancedVaccinationEntry;
import dk.sundhedsdatastyrelsen.cbs.service.util.AuditUtils;
import dk.sundhedsdatastyrelsen.cbs.service.util.CprFormatBasedAgeCalculatorImpl;
import dk.sundhedsdatastyrelsen.cbs.service.util.PassportUtils;
import org.springframework.lang.NonNull;
import se.digg.dgc.encoding.DGCConstants;
import se.digg.dgc.payload.v1.DigitalGreenCertificate;
import se.digg.dgc.payload.v1.RecoveryEntry;
import se.digg.dgc.payload.v1.TestEntry;
import se.digg.dgc.payload.v1.VaccinationEntry;
import se.digg.dgc.service.DGCEncoder;

import java.time.*;
import java.time.temporal.TemporalAmount;
import java.util.*;
import java.util.stream.Collectors;

import static dk.sundhedsdatastyrelsen.cbs.service.passports.model.Custody.getLabel;

public class DkInfoPassportFactory implements PassportFactory {
    public static final String DKINFO_HEADER = "DK3:";
    private final VaccineMapper vaccineMapper;
    private final CprFormatBasedAgeCalculator ageCalculator = new CprFormatBasedAgeCalculatorImpl();
    private final DGCEncoder encoder;
    private final RecoveryPassportEvaluator dkRecoveryPassportEvaluator;
    private final TemporalAmount partialVaccinationValidityAmount;
    private final TemporalAmount completedVaccinationValidityAmount;
    private final TemporalAmount completedVaccinationValidityAmountWhenUnder18;
    private final TemporalAmount revaccinationValidityAmount;
    private final TemporalAmount pcrTestValidityAmount;
    private final TemporalAmount agTestValidityAmount;
    private final TemporalAmount labResultMaxAge;

    public DkInfoPassportFactory(VaccineMapper vaccineMapper,
                                 DGCEncoder encoder,
                                 @DkRules RecoveryPassportEvaluator dkRecoveryPassportEvaluator,
                                 @DkRules VaccinationPassportConfiguration vaccinationPassportConfiguration,
                                 TemporalAmount revaccinationValidityAmount,
                                 TemporalAmount pcrTestValidityAmount,
                                 TemporalAmount agTestValidityAmount,
                                 TemporalAmount labResultMaxAge) {
        this.vaccineMapper = vaccineMapper;
        this.encoder = encoder;
        this.dkRecoveryPassportEvaluator = dkRecoveryPassportEvaluator;
        this.partialVaccinationValidityAmount = vaccinationPassportConfiguration.maximumPeriodAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule();
        this.completedVaccinationValidityAmount = vaccinationPassportConfiguration.maximumPeriodAfterCompletedVaccinationSchedule();
        this.completedVaccinationValidityAmountWhenUnder18 = vaccinationPassportConfiguration.maximumPeriodAfterCompletedVaccinationScheduleUnder18();
        this.revaccinationValidityAmount = revaccinationValidityAmount;
        this.pcrTestValidityAmount = pcrTestValidityAmount;
        this.agTestValidityAmount = agTestValidityAmount;
        this.labResultMaxAge = labResultMaxAge;
    }

    @Override
    @NonNull
    public Set<CoronaPass> createPassportFrom(@NonNull Vaccination vaccination, @NonNull LabResults labResults, PersonRecord person, AuditData auditData, Custody custody) {
        final LocalDateTime now = LocalDateTime.now();
        final DigitalGreenCertificate certificate = PassportUtils.buildCertificateFrom(person);

        final List<TestEntry> testEntries = getTestEntries(labResults);
        if (!testEntries.isEmpty()) {
            certificate.setT(testEntries);
        }

        final List<VaccinationEntry> vaccinationEntries = getVaccinationEntries(vaccination, ageCalculator.isUnder18(person.personIdentifier(), now));
        if (!vaccinationEntries.isEmpty()) {
            certificate.setV(vaccinationEntries);
        }

        //todo refactor and somehow use the RecoveryPassportEvaluation object received in DkPassportFactory instead of re-running evaluation here
        final List<RecoveryEntry> recoveryEntries = getRecoveryEntries(dkRecoveryPassportEvaluator.evaluate(vaccination, labResults, now));
        if (!recoveryEntries.isEmpty()) {
            certificate.setR(recoveryEntries);
        }

        final CoronaPass coronaPass = new CoronaPass(PassType.DKINFO, encode(certificate));
        if (custody != null) {
            coronaPass.setProperty("Custody", custody.getLabel());
        }

        final Set<CoronaPass> coronaPasses = Collections.singleton(coronaPass);
        AuditUtils.addDkInfoPassportAuditData(coronaPasses, auditData, getLabel(custody));
        return coronaPasses;
    }

    @NonNull
    private List<VaccinationEntry> getVaccinationEntries(Vaccination vaccination, boolean isUnder18) {
        List<VaccinationEntry> allVaccinationEntries = new ArrayList<>();

        for (VaccinationEntry programVaccinationEntry : vaccineMapper.getProgramVaccinationEntries(vaccination)) {
            allVaccinationEntries.add(withValidity(programVaccinationEntry, false, isUnder18));
        }
        for (VaccinationEntry additionalVaccinationEntry : vaccineMapper.getAdditionalVaccinationEntries(vaccination)) {
            allVaccinationEntries.add(withValidity(additionalVaccinationEntry, true, isUnder18));
        }

        final boolean orderedByEffectuationDate = Ordering.from(Comparator.comparing(VaccinationEntry::getDt)).isOrdered(allVaccinationEntries);
        if (!orderedByEffectuationDate) //sanity check
            throw new VaccineMapperException("Invalid vaccination data: Vaccination has 'additional vaccinations' that are effectuated before the last effectuation in the vaccination plan");
        return allVaccinationEntries;
    }

    private EnhancedVaccinationEntry withValidity(VaccinationEntry entry, boolean isRevaccination, boolean isUnder18) {
        final LocalDateTime effectuationTime = entry.getDt().atStartOfDay();
        final LocalDateTime validTo;
        if (isRevaccination) {
            validTo = effectuationTime.plus(revaccinationValidityAmount);
        } else {
            final TemporalAmount ageAwareCompletedVaccinationValidityAmount = isUnder18 ? completedVaccinationValidityAmountWhenUnder18 : completedVaccinationValidityAmount;
            final TemporalAmount validityAmount = Objects.equals(entry.getDn(), entry.getSd()) ? ageAwareCompletedVaccinationValidityAmount : partialVaccinationValidityAmount;
            validTo = effectuationTime.plus(validityAmount);
        }
        final Instant validToInstant = validTo.toInstant(ZoneId.systemDefault().getRules().getOffset(validTo));
        return new EnhancedVaccinationEntry(entry, validToInstant);
    }

    @NonNull
    private List<TestEntry> getTestEntries(LabResults labResults) {
        final ZonedDateTime now = ZonedDateTime.now();
        return labResults.results().stream()
            .filter(LabResult::negative)
            .filter(result -> result.samplingDateTime().isAfter(now.minus(labResultMaxAge).toLocalDateTime()))
            .map(this::toTestEntry)
            .collect(Collectors.toList());
    }

    @NonNull
    private List<RecoveryEntry> getRecoveryEntries(RecoveryPassportEvaluation recoveryEvaluation) {
        final RecoveryLabResult recoveryLabResult = recoveryEvaluation.result();
        if (!recoveryEvaluation.passportGranted() || recoveryLabResult == null) {
            return Collections.emptyList();
        }
        final RecoveryEntry recoveryEntry = RecoveryMapper.toRecovery(
            recoveryLabResult.samplingDateTime().toLocalDate(),
            recoveryLabResult.recoveryValidFrom().toLocalDate(),
            recoveryLabResult.recoveryValidUntil().toLocalDate()
        );
        return Collections.singletonList(recoveryEntry);
    }

    private EnhancedTestEntry toTestEntry(LabResult result) {
        final TestEntry entry = LabResultMapper.toTestEntry(result);
        final TemporalAmount validityAmount;
        if (result instanceof LabResult.PcrResult) {
            validityAmount = pcrTestValidityAmount;
        } else if (result instanceof LabResult.AntigenResult) {
            validityAmount = agTestValidityAmount;
        } else {
            throw new UnsupportedOperationException("Unknow test type: " + result.type());
        }
        final ZoneOffset offset = ZoneId.systemDefault().getRules().getOffset(result.samplingDateTime());
        return new EnhancedTestEntry(entry, result.samplingDateTime().plus(validityAmount).toInstant(offset));
    }

    private String encode(DigitalGreenCertificate certificate) {
        try {
            return encoder.encode(certificate, Instant.now().plus(Duration.ofHours(72L))).replaceFirst(DGCConstants.DGC_V1_HEADER, DKINFO_HEADER);
        } catch (Exception e) {
            throw new CwtException("CWT Encoding Error", e);
        }
    }
}
