/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.passports;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.upokecenter.cbor.CBORObject;
import dk.sundhedsdatastyrelsen.cbs.dao.PersonRecord;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.logging.AuditData;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.CombinedPassportEvaluator;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluations;
import dk.sundhedsdatastyrelsen.cbs.service.config.DkPassportConfiguration;
import dk.sundhedsdatastyrelsen.cbs.service.exceptions.CwtException;
import dk.sundhedsdatastyrelsen.cbs.service.model.CoronaPass;
import dk.sundhedsdatastyrelsen.cbs.service.model.PassType;
import dk.sundhedsdatastyrelsen.cbs.service.passports.model.Custody;
import dk.sundhedsdatastyrelsen.cbs.service.passports.model.DKPlus;
import dk.sundhedsdatastyrelsen.cbs.service.passports.model.DkRules;
import dk.sundhedsdatastyrelsen.cbs.service.util.AuditUtils;
import dk.sundhedsdatastyrelsen.cbs.service.util.PassportUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.lang.NonNull;
import org.springframework.stereotype.Service;
import se.digg.dgc.service.impl.DefaultDKEncoder;

import java.time.Instant;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.time.temporal.TemporalAmount;
import java.util.*;

import static dk.sundhedsdatastyrelsen.cbs.service.passports.model.Custody.getLabel;
import static dk.sundhedsdatastyrelsen.cbs.service.passports.model.Custody.isChild;

@Service
public class DkPassportFactory implements PassportFactory {
    private final DefaultDKEncoder dkEncoder;
    private final CombinedPassportEvaluator dkPassportEvaluator;
    private final TemporalAmount dkMaximumValidity;
    private final boolean makeChildrenDkPasses;

    @Autowired
    public DkPassportFactory(DefaultDKEncoder dkEncoder,
                             @DkRules CombinedPassportEvaluator dkPassportEvaluator,
                             DkPassportConfiguration dkPassportConfiguration
    ) {
        this.dkEncoder = dkEncoder;
        this.dkPassportEvaluator = dkPassportEvaluator;
        this.dkMaximumValidity = dkPassportConfiguration.getCwtDkMaximumValidity();
        this.makeChildrenDkPasses = dkPassportConfiguration.makeChildrenDkPasses();
    }

    @NonNull
    public Set<CoronaPass> createPassportFrom(@NonNull final Vaccination vaccination, @NonNull final LabResults labResults, PersonRecord person, AuditData auditData, Custody custody) {
        if (!makeChildrenDkPasses && isChild(custody)) {
            return Collections.emptySet();
        }
        PassportEvaluations evaluations = dkPassportEvaluator.evaluate(vaccination, labResults, LocalDateTime.now());
        final ZonedDateTime tokenExpiresAfter = ZonedDateTime.now().plus(dkMaximumValidity);
        final ZonedDateTime evaluationExpiresAfter = Optional.ofNullable(evaluations.expiresAfter()).map(local -> local.atZone(ZoneId.systemDefault())).orElse(null);
        Instant tokenExpiry = PassportUtils.earliestDate(tokenExpiresAfter, evaluationExpiresAfter).toInstant();

        Set<CoronaPass> coronaPasses = new HashSet<>();
        if (evaluations.passportGranted()) {
            coronaPasses.addAll(buildDKpassports(person, tokenExpiry));
        }
        AuditUtils.addDkPassportAuditData(coronaPasses, evaluations, auditData, getLabel(custody));

        if (custody != null) {
            coronaPasses.forEach(pass -> pass.setProperty("Custody", custody.getLabel()));
        }

        return coronaPasses;
    }

    private CoronaPass getDKPlusPass(PersonRecord person, Instant expiry) {
        LocalDate date = PassportUtils.cprToDate(person.personIdentifier());
        String name = person.givenName() + " " + person.familyName();

        DKPlus dkPlus = new DKPlus();
        dkPlus.setLegalName(name);
        dkPlus.setDateOfBirth(DateTimeFormatter.ISO_DATE.format(date));

        try {
            String json = new ObjectMapper().writeValueAsString(dkPlus);
            String token = dkEncoder.encodeDkPlus(CBORObject.FromJSONString(json), expiry, UUID.randomUUID().toString());
            return new CoronaPass(PassType.DKPLUS, token);
        } catch (Exception e) {
            throw new CwtException("DKPLUS CWT Encoding Error", e);
        }
    }

    private CoronaPass getDKPass(Instant expiry) {
        try {
            String token = dkEncoder.encodeDk(CBORObject.NewMap(), expiry, UUID.randomUUID().toString());
            return new CoronaPass(PassType.DK, token);
        } catch (Exception e) {
            throw new CwtException("DK CWT Encoding Error", e);
        }
    }

    private List<CoronaPass> buildDKpassports(PersonRecord person, Instant expiry) {
        return Arrays.asList(getDKPass(expiry), getDKPlusPass(person, expiry));
    }


}
