/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.passports;

import dk.sundhedsdatastyrelsen.cbs.dao.PersonRecord;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultsBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.logging.AuditData;
import dk.sundhedsdatastyrelsen.cbs.rules.EuRecoveryPassportEvaluator;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.RecoveryPassportEvaluation;
import dk.sundhedsdatastyrelsen.cbs.service.config.EuPassportConfiguration;
import dk.sundhedsdatastyrelsen.cbs.service.exceptions.CwtException;
import dk.sundhedsdatastyrelsen.cbs.service.exceptions.PassportCreationException;
import dk.sundhedsdatastyrelsen.cbs.service.model.CoronaPass;
import dk.sundhedsdatastyrelsen.cbs.service.model.PassType;
import dk.sundhedsdatastyrelsen.cbs.service.passports.model.Custody;
import dk.sundhedsdatastyrelsen.cbs.service.passports.model.EuRules;
import dk.sundhedsdatastyrelsen.cbs.service.util.AuditUtils;
import dk.sundhedsdatastyrelsen.cbs.service.util.EUAntigenFiltration;
import dk.sundhedsdatastyrelsen.cbs.service.util.PassportUtils;
import org.apache.commons.lang3.NotImplementedException;
import org.springframework.lang.NonNull;
import org.springframework.stereotype.Service;
import se.digg.dgc.payload.v1.DigitalGreenCertificate;
import se.digg.dgc.payload.v1.RecoveryEntry;
import se.digg.dgc.payload.v1.TestEntry;
import se.digg.dgc.payload.v1.VaccinationEntry;
import se.digg.dgc.service.DGCEncoder;

import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.temporal.TemporalAmount;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Stream;

import static dk.sundhedsdatastyrelsen.cbs.service.passports.LabResultFilter.filterLabResults;
import static dk.sundhedsdatastyrelsen.cbs.service.passports.LabResultMapper.LOINC_ANTIGEN;
import static dk.sundhedsdatastyrelsen.cbs.service.passports.LabResultMapper.LOINC_PCR;
import static dk.sundhedsdatastyrelsen.cbs.service.passports.model.Custody.getLabel;
import static java.util.stream.Collectors.toList;

@Service
public class EuPassportFactory implements PassportFactory {
    private final DGCEncoder euEncoder;
    private final VaccineMapper vaccineMapper;
    private final EuRecoveryPassportEvaluator euRecoveryPassportEvaluator;
    private final TemporalAmount euTokenMaxValidity;
    private final TemporalAmount antiGenResultsMaxAge;
    private final TemporalAmount pcrResultsMaxAge;
    private final boolean includePositiveResults;
    private final EUAntigenFiltration euAntigenFiltration;
    private final int antiGenResultLimit;
    private final int pcrResultLimit;

    public EuPassportFactory(DGCEncoder euEncoder,
                             VaccineMapper vaccineMapper,
                             @EuRules EuRecoveryPassportEvaluator euRecoveryPassportEvaluator,
                             EUAntigenFiltration euAntigenFiltration,
                             EuPassportConfiguration euPassportConfiguration
    ) {
        this.antiGenResultsMaxAge = euPassportConfiguration.getAntiGenResultsMaxAge();
        this.pcrResultsMaxAge = euPassportConfiguration.getPcrResultsMaxAge();
        this.includePositiveResults = euPassportConfiguration.isIncludePositiveResults();
        this.euEncoder = euEncoder;
        this.vaccineMapper = vaccineMapper;
        this.euRecoveryPassportEvaluator = euRecoveryPassportEvaluator;
        this.euAntigenFiltration = euAntigenFiltration;
        this.euTokenMaxValidity = euPassportConfiguration.getCwtEuMaximumValidity();
        this.antiGenResultLimit = euPassportConfiguration.getAntiGenResultLimit();
        this.pcrResultLimit = euPassportConfiguration.getPcrResultLimit();
    }

    @NonNull
    public Set<CoronaPass> createPassportFrom(@NonNull final Vaccination vaccination, @NonNull final LabResults labResults, PersonRecord person, AuditData auditData, Custody custody) {
        final ZonedDateTime tokenExpiresAfter = ZonedDateTime.now().plus(euTokenMaxValidity);

        Set<CoronaPass> coronaPasses = new HashSet<>();
        constructVaccinationDGC(vaccination, person, tokenExpiresAfter.toInstant(), auditData, custody) //eu vaccination cwt's have no expiry, besides maximum token validity
            .forEach(coronaPasses::add);

        LabResults filteredLabResults = filterEUAntigenTests(labResults);

        constructTestDGC(filteredLabResults, person, tokenExpiresAfter.toInstant(), auditData, custody)
            .forEach(coronaPasses::add);

        RecoveryPassportEvaluation evaluation = euRecoveryPassportEvaluator.evaluate(vaccination, filteredLabResults, LocalDateTime.now());
        final ZonedDateTime evaluationExpiresAfter = Optional.ofNullable(evaluation.expiresAfter()).map(local -> local.atZone(ZoneId.systemDefault())).orElse(null);
        Instant recoveryExpiry = PassportUtils.earliestDate(tokenExpiresAfter, evaluationExpiresAfter).toInstant();
        constructImmunizationDGC(person, recoveryExpiry, evaluation, auditData, custody)
            .ifPresent(coronaPasses::add);

        AuditUtils.addEuPassportAuditData(coronaPasses, evaluation, auditData, getLabel(custody));

        if (custody != null) {
            coronaPasses.forEach(pass -> pass.setProperty("Custody", custody.getLabel()));
        }

        return coronaPasses;
    }

    private LabResults filterEUAntigenTests(LabResults labResults) {
        // Only include AG tests that are EU valid
        Collection<LabResult> results = labResults.results();
        List<LabResult> euValidTests = results.stream()
            .filter(test -> test instanceof LabResult.AntigenResult)
            .filter(test -> euAntigenFiltration.isLegalForEU((LabResult.AntigenResult) test))
            .collect(toList());
        results.stream().filter(test -> test instanceof LabResult.PcrResult).forEach(euValidTests::add);

        return LabResultsBuilder.cpr(labResults.cpr())
            .addAll(euValidTests)
            .build();
    }


    private Stream<CoronaPass> constructTestDGC(LabResults labResults, PersonRecord person, Instant tokenMaxAge, AuditData auditData, Custody custody) {
        final ZonedDateTime now = ZonedDateTime.now();

        final List<TestEntry> testEntries = Stream
            .concat(
                filterLabResults(labResults.antigenResults(), now.minus(antiGenResultsMaxAge), antiGenResultLimit, includePositiveResults),
                filterLabResults(labResults.pcrResults(), now.minus(pcrResultsMaxAge), pcrResultLimit, includePositiveResults)
            )
            .map(LabResultMapper::toTestEntry)
            .collect(toList());

        if (!testEntries.isEmpty()) {
            auditData.withSensitivePersonalInformation("CI_EU-lab" + Custody.toSuffix(custody), testEntries.stream().map(TestEntry::getCi).collect(toList()));
            auditData.withSensitivePersonalInformation("STAT_EU-lab.pcr" + Custody.toSuffix(custody), testEntries.stream().map(TestEntry::getTt).filter(LOINC_PCR::equals).count());
            auditData.withSensitivePersonalInformation("STAT_EU-lab.rat" + Custody.toSuffix(custody), testEntries.stream().map(TestEntry::getTt).filter(LOINC_ANTIGEN::equals).count());
        }

        return testEntries.stream()
            .map(testEntry -> addLabResultsAndEncode(person, tokenMaxAge, testEntry));
    }

    private Optional<CoronaPass> constructImmunizationDGC(PersonRecord person, Instant expiry, RecoveryPassportEvaluation recoveryState, AuditData auditData, Custody custody) {
        return Optional.of(recoveryState)
            .map(RecoveryPassportEvaluation::result)
            .map(result -> addRecoveryAndEncode(PassportUtils.buildCertificateFrom(person),
                result.samplingDateTime(),
                result.recoveryValidFrom(),
                result.recoveryValidUntil(),
                expiry,
                auditData, custody));
    }

    private Stream<CoronaPass> constructVaccinationDGC(Vaccination vaccination, PersonRecord person, Instant expiry, AuditData auditData, Custody custody) {
        final List<VaccinationEntry> vaccinationEntries = vaccineMapper.getAllVaccinationEntries(vaccination);
        if (!vaccinationEntries.isEmpty()) {
            auditData.withSensitivePersonalInformation("CI_EU-vac" + Custody.toSuffix(custody), vaccinationEntries.stream().map(VaccinationEntry::getCi).collect(toList()));
            auditData.withSensitivePersonalInformation("STAT_EU-vac" + Custody.toSuffix(custody), vaccinationEntries.size());
        }

        return vaccinationEntries.stream()
            .map(entry -> addVaccineAndEncode(person, expiry, entry));
    }

    private CoronaPass addVaccineAndEncode(PersonRecord person, Instant expiry, VaccinationEntry entry) {
        final DigitalGreenCertificate certificate = PassportUtils.buildCertificateFrom(person);
        certificate.setV(Collections.singletonList(entry));
        return encode(certificate, "vac", expiry);
    }

    private CoronaPass addRecoveryAndEncode(DigitalGreenCertificate certificate, LocalDateTime firstPositiveTestResult, LocalDateTime validFrom, LocalDateTime validTo, Instant expiry, AuditData auditData, Custody custody) {
        final RecoveryEntry recoveryEntry = RecoveryMapper.toRecovery(firstPositiveTestResult.toLocalDate(), validFrom.toLocalDate(), validTo.toLocalDate());
        auditData.withSensitivePersonalInformation("CI_EU-rec" + Custody.toSuffix(custody), recoveryEntry.getCi());
        auditData.withSensitivePersonalInformation("STAT_EU-rec" + Custody.toSuffix(custody), 1);
        certificate.setR(Collections.singletonList(recoveryEntry));
        return encode(certificate, "rec", expiry);
    }


    private CoronaPass addLabResultsAndEncode(PersonRecord person, Instant tokenExpiration, TestEntry testEntry) {
        DigitalGreenCertificate certificate = PassportUtils.buildCertificateFrom(person);
        certificate.setT(Collections.singletonList(testEntry));
        TemporalAmount maxAge = getMaxAge(testEntry);
        Instant testExpiration = testEntry.getSc().plus(maxAge);
        if (testExpiration.isBefore(Instant.now())) {
            throw new PassportCreationException("Attempting to create a passport that already has expired");
        }
        Instant expiry = PassportUtils.earliestInstant(testExpiration, tokenExpiration);
        return encode(certificate, "lab", expiry);
    }


    private CoronaPass encode(DigitalGreenCertificate certificate, String value, Instant defaultExpiry) {
        String token;
        try {
            token = euEncoder.encode(certificate, defaultExpiry);
        } catch (Exception e) {
            throw new CwtException("CWT Encoding Error", e);
        }
        CoronaPass coronaPass = new CoronaPass(PassType.EU, token);
        coronaPass.setProperty("eu:keytype", value);
        return coronaPass;
    }


    private TemporalAmount getMaxAge(TestEntry testEntry) {
        switch (testEntry.getTt()) {
            case LOINC_ANTIGEN:
                return antiGenResultsMaxAge;
            case LOINC_PCR:
                return pcrResultsMaxAge;
            default:
                throw new NotImplementedException("Unknown test type");
        }
    }

}
