/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.passports;

import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult;
import dk.sundhedsdatastyrelsen.cbs.service.exceptions.LabResultMapperException;
import se.digg.dgc.payload.v1.TestEntry;

import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.ZoneOffset;

public class LabResultMapper {

    public static final String SNOMED_CT_DETECTED_POSITIVE = "260373001";
    public static final String SNOMED_CT_DETECTED_NEGATIVE = "260415000";

    public static final String LOINC_ANTIGEN = "LP217198-3";
    public static final String LOINC_PCR = "LP6464-4";

    private LabResultMapper() {
    }

    public static TestEntry toTestEntry(LabResult labResult) {
        LocalDateTime samplingDateTime = labResult.samplingDateTime();

        //CBS uses LocalDateTime in the systemDefault zone.
        //java.sql.Timestamp#toLocalDateTime creates LocalDateTime based on the system time zone so we need to convert LocalDateTime to Instant also using the systemDefault zone rules.
        //See this review comment for an explanation: https://git.nspop.dk/projects/COM/repos/covid-19-borgerstatus/pull-requests/103/overview?commentId=1561 (PR for https://jira.nspop.dk/browse/CR-162)
        ZoneOffset samplingZoneOffset = ZoneId.systemDefault().getRules().getOffset(samplingDateTime);

        return new TestEntry()
            .withCo("DK")
            .withTg(VaccineMapper.SNOMED_CT_COVID_CONCEPT_ID)
            .withSc(samplingDateTime.toInstant(samplingZoneOffset))
            .withMa(labResult.manufacturer())
            .withTc(labResult.center())
            .withTr(labResult.positive() ? SNOMED_CT_DETECTED_POSITIVE : SNOMED_CT_DETECTED_NEGATIVE)
            .withTt(toLOINC(labResult.type()))
            .withNm(labResult.name())
            .withIs(VaccineMapper.ISSUER)
            .withCi(ChecksumUtils.generateUVCI(String.valueOf(labResult.id())));
    }

    public static String toLOINC(String labResultType) {
        if ("SARSG".equalsIgnoreCase(labResultType)) {
            return LOINC_ANTIGEN;
        } else if ("SARS2".equalsIgnoreCase(labResultType)) {
            return LOINC_PCR;
        } else {
            throw new LabResultMapperException(String.format("Unknown LabResult.type: '%s' - labresult cannot be mapped", labResultType));
        }
    }
}
