/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.passports;

import com.google.common.collect.Ordering;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.EffectuatedVaccination;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.VaccinationScheduleType;
import dk.sundhedsdatastyrelsen.cbs.service.exceptions.VaccineMapperException;
import dk.sundhedsdatastyrelsen.cbs.service.passports.VaccineMappingConfiguration.VaccineInformation;
import se.digg.dgc.payload.v1.VaccinationEntry;

import java.util.Comparator;
import java.util.List;
import java.util.UUID;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.IntSupplier;
import java.util.stream.Collectors;

public class VaccineMapper {

    public static final String SNOMED_CT_COVID_CONCEPT_ID = "840539006";
    public static final String ISSUER = "Danish Health Data Authority";


    public VaccineMapper(VaccineMappingConfiguration vaccineMappingConfiguration) {
        this.vaccineMappingConfiguration = vaccineMappingConfiguration;
    }

    private final VaccineMappingConfiguration vaccineMappingConfiguration;

    public List<VaccinationEntry> getProgramVaccinationEntries(Vaccination vaccination) {
        final int size = vaccination.vaccinations().size();
        final boolean completed = vaccination.completed() != null;

        //CR-242: If vaccination is not completed then number all effectuations as 1/2
        final int dosesInSeries = completed ? size : 2;
        final AtomicInteger doseCounter = new AtomicInteger(0);
        final IntSupplier doseNumber = () -> completed ? doseCounter.incrementAndGet() : 1;

        return vaccination.vaccinations().stream()
            .sorted(Comparator.comparing(EffectuatedVaccination::effectuated))
            .map(effectuatedVaccination -> toVaccinationEntry(effectuatedVaccination, doseNumber.getAsInt(), dosesInSeries))
            .collect(Collectors.toList());
    }

    public List<VaccinationEntry> getAdditionalVaccinationEntries(Vaccination vaccination) {
        if (vaccination.completed() == null && !vaccination.additionalVaccinations().isEmpty()) //sanity check
            throw new VaccineMapperException("Invalid vaccination data: Vaccination was not marked as completed, but contained 'additional vaccinations'");

        final AtomicInteger doseCounter = new AtomicInteger(vaccination.vaccinations().size());

        final boolean isSingleDoseProgram = vaccination.scheduleType() == VaccinationScheduleType.SINGLE_DOSE_VACCINATION_SCHEDULE;

        return vaccination.additionalVaccinations().stream()
            .sorted(Comparator.comparing(EffectuatedVaccination::effectuated))
            .map(effectuatedVaccination -> {
                final int additionalDoseNumber = doseCounter.incrementAndGet();
                return toVaccinationEntry(effectuatedVaccination, additionalDoseNumber, isSingleDoseProgram ? 1 : additionalDoseNumber);
            }).collect(Collectors.toList());
    }

    public List<VaccinationEntry> getAllVaccinationEntries(Vaccination vaccination) {
        final List<VaccinationEntry> allVaccinationEntries = getProgramVaccinationEntries(vaccination);
        allVaccinationEntries.addAll(getAdditionalVaccinationEntries(vaccination));

        final boolean orderedByEffectuationDate = Ordering.from(Comparator.comparing(VaccinationEntry::getDt)).isOrdered(allVaccinationEntries);
        if (!orderedByEffectuationDate) //sanity check
            throw new VaccineMapperException("Invalid vaccination data: Vaccination has 'additional vaccinations' that are effectuated before the last effectuation in the vaccination plan");
        return allVaccinationEntries;
    }

    private VaccinationEntry toVaccinationEntry(EffectuatedVaccination effectuatedVaccination, int doseNumber, int dosesInSeries) {
        VaccineInformation values = vaccineMappingConfiguration.getVaccineIdValues(effectuatedVaccination.drugIdentifier());

        return new VaccinationEntry()
            .withDt(effectuatedVaccination.effectuated().toLocalDate())
            .withTg(SNOMED_CT_COVID_CONCEPT_ID)
            .withVp(values.vp)
            .withMp(values.mp)
            .withMa(values.ma)
            .withCo(effectuatedVaccination.countryCode())
            .withDn(doseNumber)
            .withIs(ISSUER)
            .withCi(ChecksumUtils.generateUVCI(UUID.randomUUID().toString()))
            .withSd(dosesInSeries);
    }

}
