/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.security;

import org.apache.log4j.Logger;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;
import se.digg.dgc.signatures.DGCSigner;
import se.digg.dgc.signatures.impl.DefaultDGCSigner;
import se.digg.dgc.signatures.impl.DefaultDKSigner;

import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.Provider;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;

public class SignerFactory {

    private static final Logger LOGGER = Logger.getLogger(SignerFactory.class);

    private final PrivateKey key;
    private final X509Certificate certificate;
    private final Provider provider;

    public SignerFactory(@NonNull KeyStore keyStore, @NonNull String alias, @NonNull char[] password, @Nullable Provider provider) throws UnrecoverableKeyException, KeyStoreException, NoSuchAlgorithmException {
        LOGGER.debug("Getting Key and Certificate");

        this.provider = provider;
        key = (PrivateKey) keyStore.getKey(alias, password);
        certificate = (X509Certificate) keyStore.getCertificate(alias);

        LOGGER.debug("Key loaded = " + key.getClass().getName());
        LOGGER.debug("Certificate loaded = " + certificate.getClass().getName());

        LOGGER.debug("Security Provider = " + (provider != null ? provider.getName() : "none"));
    }

    public DGCSigner dgcSigner() throws CertificateException {
        DefaultDGCSigner defaultDGCSigner = new DefaultDGCSigner(key, certificate);
        if (provider != null) {
            defaultDGCSigner.setSecurityProvider(provider);
        }
        return defaultDGCSigner;
    }

    public DefaultDKSigner dkSigner() throws CertificateException {
        DefaultDKSigner defaultDKSigner = new DefaultDKSigner(key, certificate);

        if (provider != null) {
            defaultDKSigner.setSecurityProvider(provider);
        }
        return defaultDKSigner;
    }

}
