/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.security;

import dk.sds.nsp.security.SecurityContext;
import dk.sds.nsp.security.SecurityContext.User;
import dk.sundhedsdatastyrelsen.cbs.dao.ChildRelation;
import dk.sundhedsdatastyrelsen.cbs.dao.ServiceClient;
import dk.sundhedsdatastyrelsen.cbs.datacache.CollectionReader;
import dk.sundhedsdatastyrelsen.cbs.service.ErrorMessages;
import dk.sundhedsdatastyrelsen.cbs.service.ServiceConstants;
import dk.sundhedsdatastyrelsen.cbs.service.exceptions.SecurityRuntimeException;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.lang.NonNull;
import org.springframework.stereotype.Component;

import java.util.Objects;
import java.util.Optional;
import java.util.function.Predicate;

import static java.util.function.Predicate.isEqual;

@Component
public class WsSecurity {

    private final CollectionReader<ChildRelation, String> childRelationReader;
    private final CollectionReader<ServiceClient, String> serviceClientReader;
    private final String cbsPowerOfAttorneyPrivilege;

    public WsSecurity(CollectionReader<ChildRelation, String> childRelationReader,
                      CollectionReader<ServiceClient, String> serviceClientReader,
                      @Value("${security.power-of-attorney.privilege-name:urn:dk:nspop:sts:cbs:read}") String cbsPowerOfAttorneyPrivilege) {
        this.childRelationReader = childRelationReader;
        this.serviceClientReader = serviceClientReader;
        this.cbsPowerOfAttorneyPrivilege = cbsPowerOfAttorneyPrivilege;
    }


    public void checkSecurityTicket(
        @NonNull final SecurityContext securityContext,
        @NonNull String personIdentifier,
        @NonNull String serviceIdentifier) {
        Optional<SecurityContext.Ticket> ticketOptional = securityContext.getTicket();
        if (!ticketOptional.isPresent()) {
            throw new SecurityRuntimeException(ErrorMessages.NO_SECURITY_TICKET);
        }
        SecurityContext.Ticket ticket = ticketOptional.get();
        if (!ticket.isValid()) {
            throw new SecurityRuntimeException(ErrorMessages.INVALID_SECURITY_TICKET);
        }

        if (securityContext.getActingUser().isPresent()) {
            checkIdws(securityContext, personIdentifier);
        } else {
            checkDgws(securityContext, serviceIdentifier);
        }
    }

    private void checkIdws(@NonNull SecurityContext securityContext, @NonNull String personIdentifier) {
        checkAudience(securityContext);

        final String cprFromTicket = securityContext
            .getActingUser()
            .map(User::getIdentifier)
            .orElseThrow(() -> new SecurityRuntimeException("No user present"));

        if (!isCitizenUser(securityContext)) {
            throw new SecurityRuntimeException(ErrorMessages.USER_TYPE_NOT_ALLOWED);
        }

        if (!cprFromTicket.equals(personIdentifier)) {
            final boolean hasPowerOfAttorneyPrivilege = securityContext.getActingUser().flatMap(User::getCredentials).map(User.Credentials::getPowerOfAttorneyPrivileges).filter(privileges -> privileges.contains(cbsPowerOfAttorneyPrivilege)).isPresent();
            final boolean principalUserIdentifierMatches = securityContext.getPrincipalUser().map(User::getIdentifier).equals(Optional.of(personIdentifier));
            if ((!hasPowerOfAttorneyPrivilege || !principalUserIdentifierMatches) && !hasChildRelation(cprFromTicket, personIdentifier)) {
                throw new SecurityRuntimeException(ErrorMessages.NOT_AUTHORIZED_CPR_MISMATCH);
            }
        }
    }

    private void checkDgws(@NonNull SecurityContext securityContext, @NonNull String serviceIdentifier) {
        if (!securityContext.getOrganisation().isPresent()) {
            throw new SecurityRuntimeException(ErrorMessages.ORGANIZATION_NOT_PRESENT);
        }

        final SecurityContext.PersistentUniqueKey persistentUniqueKey = securityContext
            .getClient()
            .orElseThrow(() -> new SecurityRuntimeException(ErrorMessages.CLIENT_NOT_PRESENT))
            .getPersistentUniqueKey()
            .orElseThrow(() -> new SecurityRuntimeException(ErrorMessages.PERSISTENT_KEY_NOT_PRESENT));

        final boolean isWhitelisted = isClientWhitelisted(serviceIdentifier, persistentUniqueKey);

        if (!isWhitelisted) {
            throw new SecurityRuntimeException(ErrorMessages.CLIENT_NOT_WHITELISTED);
        }
    }

    private boolean isClientWhitelisted(String serviceIdentifier, SecurityContext.PersistentUniqueKey persistentUniqueKey) {
        return serviceClientReader.search(persistentUniqueKey.value())
            .stream()
            .map(ServiceClient::serviceIdentifier)
            .anyMatch(Predicate.isEqual(serviceIdentifier));
    }

    private boolean hasChildRelation(String cprFromTicket, String cpr) {
        return childRelationReader.search(cprFromTicket).stream()
            .map(ChildRelation::childIdentifier)
            .anyMatch(isEqual(cpr));
    }

    private void checkAudience(SecurityContext securityContext) {
        String audience = securityContext.getTicket()
            .flatMap(SecurityContext.Ticket::getAudience)
            .orElseThrow(() -> new SecurityRuntimeException(ErrorMessages.MISSING_AUDIENCE));

        if (!Objects.equals(audience, ServiceConstants.AUDIENCE)) {
            throw new SecurityRuntimeException(ErrorMessages.INVALID_AUDIENCE);
        }
    }

    private boolean isCitizenUser(SecurityContext securityContext) {
        return securityContext.getActingUser()
            .map(User::getUserType)
            .filter(isEqual(User.UserType.Citizen))
            .isPresent();
    }


}
