/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.util;

import dk.sundhedsdatastyrelsen.cbs.logging.AuditData;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluations;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.RecoveryPassportEvaluation;
import dk.sundhedsdatastyrelsen.cbs.service.model.CoronaPass;

import java.util.Collection;
import java.util.Map;

import static java.util.stream.Collectors.*;

public class AuditUtils {
    private AuditUtils() {
    }

    public static void addDkPassportAuditData(Collection<CoronaPass> coronaPasses, PassportEvaluations evaluations, AuditData auditData, String custodyLabel) {
        coronaPasses.stream()
            .collect(groupingBy(coronaPass -> getCWTTypeAuditKey(coronaPass, custodyLabel), mapping(CoronaPass::getToken, toList())))
            .forEach(auditData::withSensitivePersonalInformation);

        String suffix = custodyLabel != null ? '_' + custodyLabel : "";

        auditData
            .withSensitivePersonalInformation("dkRejectionReason" + suffix, evaluations.rejectionReason())
            .withSensitivePersonalInformation("dkRejectionReasonRecovery" + suffix, evaluations.recoveryPassport().rejectionReason())
            .withSensitivePersonalInformation("dkRejectionReasonTest" + suffix, evaluations.testPassport().rejectionReason())
            .withSensitivePersonalInformation("dkRejectionReasonVaccination" + suffix, evaluations.vaccinationPassport().rejectionReason())
            .withSensitivePersonalInformation("dkRejectionReasonRevaccination" + suffix, evaluations.revaccinationPassport().rejectionReason());
    }

    public static void addEuPassportAuditData(Collection<CoronaPass> coronaPasses, RecoveryPassportEvaluation recoveryPassportEvaluation, AuditData auditData, String custodyLabel) {
        coronaPasses.stream()
            .collect(groupingBy(coronaPass -> getCWTTypeAuditKey(coronaPass, custodyLabel), mapping(CoronaPass::getToken, toList())))
            .forEach(auditData::withSensitivePersonalInformation);

        String suffix = custodyLabel != null ? '_' + custodyLabel : "";

        auditData.withSensitivePersonalInformation("euRejectionReasonRecovery" + suffix, recoveryPassportEvaluation.rejectionReason());
    }

    public static void addDkInfoPassportAuditData(Collection<CoronaPass> coronaPasses, AuditData auditData, String custodyLabel) {
        coronaPasses.stream()
            .collect(groupingBy(coronaPass -> getCWTTypeAuditKey(coronaPass, custodyLabel), mapping(CoronaPass::getToken, toList())))
            .forEach(auditData::withSensitivePersonalInformation);
    }

    /**
     * Creates key identifying the Pass Type
     * <p>
     * Type examples:
     * <ul>
     *   <li> CWT_DK
     *   <li> CWT_DKPLUS
     *   <li> CWT_DKINFO
     *   <li> CWT_EU-vac
     *   <li> CWT_EU-lab
     *   <li> CWT_EU-rec
     * </ul>
     *
     * @param coronaPass   the Pass
     * @param custodyLabel
     * @return key identifying Pass Type
     */
    private static String getCWTTypeAuditKey(CoronaPass coronaPass, String custodyLabel) {
        StringBuilder type = new StringBuilder("CWT")
            .append('_')
            .append(coronaPass.getPassType());


        coronaPass.getProperties().entrySet().stream()
            .filter(p -> p.getKey().equals("eu:keytype"))
            .map(Map.Entry::getValue)
            .findFirst()
            .ifPresent(keyType -> type.append('-').append(keyType));

        if (custodyLabel != null) {
            type.append('_').append(custodyLabel);
        }

        return type.toString();
    }
}
