/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.util;

import dk.sundhedsdatastyrelsen.cbs.rules.CprFormatBasedAgeCalculator;

import java.time.Duration;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.Period;
import java.time.temporal.TemporalAmount;
import java.util.Optional;

public class CprFormatBasedAgeCalculatorImpl implements CprFormatBasedAgeCalculator {

    @Override
    public boolean isUnder18(String cpr, LocalDateTime now) {
        final LocalDate today = now.toLocalDate();

        return cprToDate(cpr)
            .map(dateOfBirth -> isUnder18(today, dateOfBirth))
            .orElse(false);
    }

    @Override
    public TemporalAmount timeUntil18(String cpr, LocalDateTime now) {
        final LocalDate today = now.toLocalDate();

        return cprToDate(cpr)
            .filter(dateOfBirth -> isUnder18(today, dateOfBirth))
            .map(dateOfBirth -> dateOfBirth.plusYears(18).atStartOfDay())
            .map(dateOf18thBirthday -> Duration.between(now, dateOf18thBirthday))
            .orElse(Duration.ZERO);
    }

    private boolean isUnder18(LocalDate today, LocalDate dateOfBirth) {
        return Period.between(dateOfBirth, today).getYears() < 18;
    }

    private Optional<LocalDate> cprToDate(String cpr) {
        try {
            return Optional.of(PassportUtils.cprToDate(cpr));
        } catch (IllegalArgumentException e) {
            return Optional.empty();
        }
    }
}
