/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.util;

import com.google.common.cache.LoadingCache;
import dk.sundhedsdatastyrelsen.cbs.dao.EUAntigenSpec;
import dk.sundhedsdatastyrelsen.cbs.datacache.CollectionReader;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult;
import dk.sundhedsdatastyrelsen.cbs.service.exceptions.CacheException;
import org.apache.commons.lang3.StringUtils;
import org.apache.log4j.Logger;

import java.time.LocalDateTime;
import java.util.Collection;
import java.util.concurrent.ExecutionException;

public class EUAntigenFiltration {
    private static final Logger LOGGER = Logger.getLogger(EUAntigenFiltration.class);
    // In-memory cache of legal EU antigen tests
    private final LoadingCache<String, Collection<EUAntigenSpec>> euAntigenSpecCache;
    private final CollectionReader<EUAntigenSpec, String> euAntigenSpecReader;
    private final boolean useCache;

    public EUAntigenFiltration(LoadingCache<String, Collection<EUAntigenSpec>> euAntigenSpecCache,
                               CollectionReader<EUAntigenSpec, String> euAntigenSpecReader,
                               boolean euAntigenSpecUseCache) {
        this.euAntigenSpecCache = euAntigenSpecCache;
        this.euAntigenSpecReader = euAntigenSpecReader;
        this.useCache = euAntigenSpecUseCache;
    }

    public boolean isLegalForEU(LabResult.AntigenResult agResult) {
        return isIdDeviceLegalAtTime(agResult.manufacturer(), agResult.samplingDateTime());
    }

    public boolean isIdDeviceLegalAtTime(String idDevice, LocalDateTime sampleDateTime) {
        if (StringUtils.isEmpty(idDevice)) {
            LOGGER.debug("Rejecting test with empty idDevice");
            return false;
        } else if (sampleDateTime == null) {
            LOGGER.debug("Rejecting test with null sampleDateTime");
            return false;
        }

        Collection<EUAntigenSpec> idDeviceHistory = getEuAntigenSpecs(idDevice);

        final boolean anyMatch = idDeviceHistory.stream().anyMatch(spec -> isPeriodValid(spec, sampleDateTime));
        if (LOGGER.isDebugEnabled() && !anyMatch)
            LOGGER.debug(String.format("Test on sampleDateTime=%s with idDevice=%s is not valid for EU use", sampleDateTime, idDevice));
        return anyMatch;
    }

    private Collection<EUAntigenSpec> getEuAntigenSpecs(String idDevice) {
        if (!useCache) {
            return euAntigenSpecReader.search(idDevice);
        }

        try {
            return euAntigenSpecCache.get(idDevice);
        } catch (ExecutionException e) {
            throw new CacheException(e);
        }
    }

    private boolean isPeriodValid(EUAntigenSpec spec, LocalDateTime sampleDateTime) {
        return sampleDateTime.isAfter(spec.validFrom()) &&
            (spec.validTo() == null || sampleDateTime.isBefore(spec.validTo()));
    }

}
