/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.util;

import dk.sundhedsdatastyrelsen.cbs.dao.PersonRecord;
import org.springframework.lang.NonNull;
import se.digg.dgc.payload.v1.DGCSchemaVersion;
import se.digg.dgc.payload.v1.DigitalGreenCertificate;
import se.digg.dgc.payload.v1.PersonName;

import java.time.Instant;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZonedDateTime;
import java.util.Arrays;
import java.util.Objects;

import static no.kantega.id.dk.DanishIdNumber.birthday;
import static no.kantega.id.dk.DanishIdNumber.forId;

public class PassportUtils {
    private PassportUtils() {
    }

    public static LocalDate cprToDate(@NonNull String cpr) {
        return birthday(forId(cpr)).orElseThrow(() -> new IllegalArgumentException("CPR string could not be parsed into a date"));
    }

    public static ZonedDateTime earliestDate(ZonedDateTime... dates) {
        return
            Arrays
                .stream(dates)
                .filter(Objects::nonNull)
                .min(ZonedDateTime::compareTo).orElseThrow(() -> new IllegalArgumentException("vararg argument did not contain any non-null values"));
    }

    public static DigitalGreenCertificate buildCertificateFrom(PersonRecord person) {
        return buildDigitalGreenCertificate(
            person.givenName(),
            person.familyName(),
            cprToDate(person.personIdentifier())
        );
    }

    public static DigitalGreenCertificate buildDigitalGreenCertificate(String givenName, String familyName, LocalDate dateOfBirth) {
        DigitalGreenCertificate certificate = new DigitalGreenCertificate();
        certificate.setNam(new PersonName()
            .withGn(givenName)
            .withFn(familyName));
        certificate.setDob(dateOfBirth);
        certificate.setVer(DGCSchemaVersion.DGC_SCHEMA_VERSION);
        return certificate;
    }

    public static Instant earliestInstant(Instant... instants) {
        return
            Arrays
                .stream(instants)
                .filter(Objects::nonNull)
                .min(Instant::compareTo).orElseThrow(() -> new IllegalArgumentException("vararg argument did not contain any non-null values"));
    }
}
