/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.webservice;


import dk.sundhedsdatastyrelsen.cbs.logging.AuditData;
import dk.sundhedsdatastyrelsen.cbs.logging.AuditLogger;
import dk.sundhedsdatastyrelsen.cbs.service.handler.EUCovidTestCertificateSigningHandler;
import dk.sundhedsdatastyrelsen.cbs.service.model.SignedCertificateCWT;
import dk.sundhedsdatastyrelsen.cbs.service.security.WsSecurity;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.EUCovidTestCertificateSigningPort;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.GetSignedEUCovidTestCertificateRequest;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.GetSignedEUCovidTestCertificateResponse;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.ObjectFactory;
import org.apache.cxf.annotations.SchemaValidation;
import org.springframework.context.ApplicationContext;
import org.springframework.lang.NonNull;
import se.digg.dgc.payload.v1.DGCSchemaVersion;

import javax.jws.HandlerChain;
import javax.jws.WebService;
import javax.jws.soap.SOAPBinding;
import javax.xml.bind.annotation.XmlSeeAlso;

import static dk.sds.nsp.security.Security.getSecurityContext;

@SchemaValidation(type = SchemaValidation.SchemaValidationType.BOTH)
@WebService(
    name = "EUCovidTestCertificateSigningPort",
    serviceName = "EUCovidTestCertificateSigningService",
    portName = "Soap",
    targetNamespace = "http://sundhedsdatastyrelsen.dk/coronapass/2021/07/eu-covid-test-certificate-signing",
    wsdlLocation = "wsdl/eu_covid_test_certificate_signing_service.wsdl"
)
@SOAPBinding(parameterStyle = SOAPBinding.ParameterStyle.BARE)
@HandlerChain(file = "handler-chain.xml")
@XmlSeeAlso(ObjectFactory.class)
public class EUCovidTestCertificateSigningService extends AbstractWebService implements EUCovidTestCertificateSigningPort {

    public static final String SERVICE_IDENTIFIER = "EUCovidTestCertificateSigningService";
    private WsSecurity wsSecurity;
    private EUCovidTestCertificateSigningHandler certificateSigningHandler;
    private AuditLogger auditLogger;

    @Override
    protected void initializeSpringDependencies(@NonNull final ApplicationContext context) {
        wsSecurity = context.getBean(WsSecurity.class);
        certificateSigningHandler = context.getBean(EUCovidTestCertificateSigningHandler.class);
        auditLogger = context.getBean(AuditLogger.class);
    }

    // IntelliJ complains about 'Web Method problem: Class java.time.LocalDate not public or does not allow instantiation',
    // but that is handled in package dk.sundhedsdatastyrelsen.xmladapter
    @SuppressWarnings("ValidExternallyBoundObject")
    @Override
    public GetSignedEUCovidTestCertificateResponse getSignedEUCovidTestCertificate(GetSignedEUCovidTestCertificateRequest request) {
        final AuditData auditData = new AuditData()
            .withSensitivePersonalInformation("personIdentifier", request.getPersonIdentifier());
        try {
            wsSecurity.checkSecurityTicket(getSecurityContext(), request.getPersonIdentifier(), SERVICE_IDENTIFIER);

            final SignedCertificateCWT signedCertificateCWT = certificateSigningHandler.getSignedCertificateCWT(request);

            auditData
                .withNonPersonalInformation("CI", signedCertificateCWT.getCi())
                .withSensitivePersonalInformation("CWT", signedCertificateCWT.getCwt());

            final GetSignedEUCovidTestCertificateResponse response = new GetSignedEUCovidTestCertificateResponse();
            response.setCWT(signedCertificateCWT.getCwt());
            response.setCertificateIdentifier(signedCertificateCWT.getCi());
            response.setSchemaVersion(DGCSchemaVersion.DGC_SCHEMA_VERSION);
            response.setDocumentation("https://github.com/ehn-dcc-development/ehn-dcc-schema");

            return response;
        } catch (Exception e) {
            auditData.withNonPersonalInformation("error", e.getMessage());
            throw e;
        } finally {
            auditLogger.log(auditData, SERVICE_IDENTIFIER);
        }
    }
}
