/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Iterables;
import com.google.common.collect.MoreCollectors;
import com.upokecenter.cbor.CBORObject;
import dk.sundhedsdatastyrelsen.cbs.dao.PersonRecord;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.PersonRecordImpl;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultsBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.VaccinationBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.kafka.CbsKafkaProducer;
import dk.sundhedsdatastyrelsen.cbs.logging.AuditLogger;
import dk.sundhedsdatastyrelsen.cbs.service.cache.Data;
import dk.sundhedsdatastyrelsen.cbs.service.cache.DataReader;
import dk.sundhedsdatastyrelsen.cbs.service.config.*;
import dk.sundhedsdatastyrelsen.cbs.service.model.CoronaPass;
import dk.sundhedsdatastyrelsen.cbs.service.model.CoronaPassResult;
import dk.sundhedsdatastyrelsen.cbs.service.model.PassType;
import dk.sundhedsdatastyrelsen.cbs.service.passports.CombinedPassportFactory;
import dk.sundhedsdatastyrelsen.cbs.service.passports.DkPassportFactory;
import dk.sundhedsdatastyrelsen.cbs.service.passports.EuPassportFactory;
import dk.sundhedsdatastyrelsen.cbs.service.passports.VaccineMapper;
import dk.sundhedsdatastyrelsen.cbs.service.webservice.MockDataService;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import dk.sundhedsdatastyrelsen.cbs.test.CwtTestUtils;
import dk.sundhedsdatastyrelsen.cbs.test.Data.Results;
import dk.sundhedsdatastyrelsen.cbs.test.KafkaTestUtils;
import dk.sundhedsdatastyrelsen.cbs.test.RandomData;
import org.apache.log4j.Logger;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.mockito.MockedStatic;
import org.mockito.Mockito;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Primary;
import org.springframework.jdbc.datasource.DriverManagerDataSource;
import org.springframework.jdbc.datasource.lookup.JndiDataSourceLookup;
import org.springframework.lang.NonNull;
import org.springframework.mock.jndi.SimpleNamingContextBuilder;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.TestPropertySource;
import org.springframework.test.context.junit.jupiter.SpringJUnitConfig;
import se.digg.dgc.payload.v1.DigitalGreenCertificate;
import se.digg.dgc.payload.v1.RecoveryEntry;
import se.digg.dgc.service.impl.MrzParser;
import se.digg.dgc.signatures.cwt.Cwt;

import javax.naming.NamingException;
import java.time.*;
import java.time.temporal.ChronoUnit;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.UUID;
import java.util.stream.Collectors;

import static com.google.common.truth.Truth.assertThat;
import static dk.sundhedsdatastyrelsen.cbs.test.Data.Vaccinations;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

@CbsTest
@SpringJUnitConfig({EUAntigenSpecConfiguration.class, CoronaPassServiceTest.MockConfig.class, CombinedPassportFactory.class, DkPassportFactory.class, EuPassportFactory.class, SignatureConfiguration.class, DkPassportConfiguration.class, EuPassportConfiguration.class, VaccineMapperConfiguration.class})
@ActiveProfiles(SignatureConfiguration.LOCAL_KEY_STORE)
@TestPropertySource(properties = {
    "cwt.eu.ag-result.maximum-age=P7D",
    "cwt.eu.pcr-result.maximum-age=P14D",
    "cwt.dkinfo.lab-result.maximum-age=P7D",
    "cwt.dk.maximum-validity=PT24H",
    "cwt.eu.maximum-validity=PT48H",
    "eu.recovery.validator.minimum-period-after-recovery=P11D",
    "eu.recovery.validator.maximum-period-after-recovery=P180D"
})
class CoronaPassServiceTest {
    private static final Logger logger = Logger.getLogger(CoronaPassServiceTest.class);
    private static final RandomData randomData = new RandomData(System.currentTimeMillis());
    private final String idwsMessageId = UUID.randomUUID().toString();
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${cwt.dk.maximum-validity}')}")
    private Duration dkMaximumValidity;
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${cwt.eu.maximum-validity}')}")
    private Duration euMaximumValidity;
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${cwt.eu.pcr-result.maximum-age}')}")
    private Duration euPcrMaxAge;
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${cwt.dkinfo.lab-result.maximum-age:P7D}')}")
    private Duration cwtDkInfoLabResultMaxAge;
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.vaccination.validator.maximum-amount-after-first-vaccination-of-partially-completed-vaccination-schedule:P42D}')}")
    private Duration maximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule;
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.test.validator.maximum-period-after-test.pcr:PT96H}')}")
    private Duration maximumPeriodAfterTestPcr;
    @Mock
    private CbsKafkaProducer cbsKafkaProducer;
    @Autowired
    private CombinedPassportFactory passportFactory;
    @Autowired
    private AuditLogger auditLogger;

    @Mock
    private DataReader dataReader;

    private CoronaPassService coronaPassService;

    @Configuration
    public static class MockConfig {
        @Bean
        @Primary
        AuditLogger auditLogger() {
            return Mockito.mock(AuditLogger.class);
        }

    }

    @BeforeEach
    void setUp() {
        coronaPassService = new CoronaPassService(cbsKafkaProducer, passportFactory, auditLogger, dataReader);
    }

    @BeforeAll
    static void setUpJNDI() throws NamingException {
        //noinspection deprecation
        SimpleNamingContextBuilder.emptyActivatedContextBuilder();
        new JndiDataSourceLookup().getJndiTemplate().bind("java:jboss/datasources/euantigenspec", new DriverManagerDataSource("jdbc:h2:mem:testdb"));
    }

    @Test
    void does_not_use_messageId_as_jobId() {
        String messageId = randomData.uuidString();
        final String cpr = randomData.cpr();
        String jobId = coronaPassService.createDataFetcherJob(cpr, messageId, ImmutableMap.of(cpr, KafkaTestUtils.getRandomFetchType()));
        assertNotEquals(messageId, jobId);
    }

    private PersonRecord toPersonRecord(MockDataService.Person person, String cpr) {
        return new PersonRecordImpl()
            .personIdentifier(cpr)
            .givenName(person.first)
            .familyName(person.last);
    }

    @NonNull
    private String cprWithCacheHitOn(@NonNull final String cpr) {
        final Vaccination vaccination = Vaccinations.completedSingleDose(cpr, LocalDateTime.now().minusWeeks(4));
        final LabResults labResults = mock(LabResults.class);
        given(cpr, vaccination, labResults);
        return cpr;
    }


    @Test
    void test_getCoronaPass_cache_hit() {
        final String cprPrefix = "040587010";
        for (int i = 0; i < 10; i++) {
            final String cpr = cprWithCacheHitOn(cprPrefix + 1);
            verifyPass(cpr);
        }
    }

    @Test
    void test_get_no_passes_except_dkinfo_case_0405870114() {
        String cpr = "0405870118";

        given(cpr, VaccinationBuilder.vaccination(cpr, false).build(), LabResultsBuilder.cpr(cpr).build());

        CoronaPassResult coronaPassResult = coronaPassService.getCoronaPass(cpr, idwsMessageId);
        assertTrue(coronaPassResult.isCacheHit());
        assertEquals(1, coronaPassResult.getCoronaPasses().size());
        assertEquals(PassType.DKINFO, coronaPassResult.getCoronaPasses().iterator().next().getPassType());
    }


    @Test
    void test_verify_DK_and_DKPlus_and_DKInfo_content_and_that_max_validity_accounts_for_DST_change() {
        String cpr = "0405870113";
        MockDataService.Person person = MockDataService.person(cpr);

        final ZoneId zoneId = ZoneId.of("Europe/Copenhagen");
        final ZonedDateTime beforeDstEnds = ZonedDateTime.of(2021, 10, 31, 0, 0, 0, 0, zoneId);

        final Vaccination vaccination = Vaccinations.completedMultiDose(cpr, beforeDstEnds.minusWeeks(2).toLocalDateTime());
        given(cpr, vaccination, noLabResults(cpr));

        CoronaPassResult coronaPassResult;
        Clock fixedClock = Clock.fixed(beforeDstEnds.toInstant(), zoneId);
        try (MockedStatic<Clock> mockedClock = Mockito.mockStatic(Clock.class)) {
            mockedClock.when(Clock::systemUTC).thenReturn(fixedClock); //called by Instant.now()
            mockedClock.when(Clock::systemDefaultZone).thenReturn(fixedClock); //called by LocalDateTime.now() and ZonedDateTime.now()
            coronaPassResult = coronaPassService.getCoronaPass(cpr, idwsMessageId);
        }
        assertTrue(coronaPassResult.isCacheHit());

        Cwt dkContent = CwtTestUtils.getCwt(coronaPassResult.getCoronaPasses(), PassType.DK);
        assertThat(dkContent.getIssuedAt()).isEqualTo(beforeDstEnds.toInstant());
        assertThat(dkContent.getExpiration()).isEqualTo(beforeDstEnds.plus(dkMaximumValidity).toInstant());
        assertNotNull(UUID.fromString(new String(dkContent.getCwtId())));

        Cwt dkPlusContent = CwtTestUtils.getCwt(coronaPassResult.getCoronaPasses(), PassType.DKPLUS);
        assertThat(dkPlusContent.getIssuedAt()).isEqualTo(beforeDstEnds.toInstant());
        assertThat(dkPlusContent.getExpiration()).isEqualTo(beforeDstEnds.plus(dkMaximumValidity).toInstant());
        assertEquals("1987-05-04", dkPlusContent.getClaim("dateOfBirth").AsString());
        assertEquals(person.first + " " + person.last, dkPlusContent.getClaim("legalName").AsString());
    }

    @Test
    void test_verify_DKInfo_cwtDkInfoLabResultMaxAge_and_test_and_vac_expiration_accounts_for_DST_change() {
        String cpr = "0405870113";

        final ZoneId zoneId = ZoneId.of("Europe/Copenhagen");
        final ZonedDateTime afterDstEnds = ZonedDateTime.of(2021, 10, 31, 10, 0, 0, 0, zoneId);

        final ZonedDateTime pcrTestExpiringIn15Minutes = afterDstEnds.minus(maximumPeriodAfterTestPcr).plusMinutes(15);
        final ZonedDateTime pcrTest15MinutesOlderThanMaxAge = afterDstEnds.minus(cwtDkInfoLabResultMaxAge).minusMinutes(15);
        final LabResults labResults = Results.negativePcr(cpr, pcrTestExpiringIn15Minutes.toLocalDateTime(), pcrTest15MinutesOlderThanMaxAge.toLocalDateTime());
        final Vaccination vaccination = Vaccinations.partial(cpr, afterDstEnds.minus(maximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule).plusDays(1).minusMinutes(15).toLocalDateTime());
        given(cpr, vaccination, labResults);

        CoronaPassResult coronaPassResult;
        Clock fixedClock = Clock.fixed(afterDstEnds.toInstant(), zoneId);
        try (MockedStatic<Clock> mockedClock = Mockito.mockStatic(Clock.class)) {
            mockedClock.when(Clock::systemUTC).thenReturn(fixedClock); //called by Instant.now()
            mockedClock.when(Clock::systemDefaultZone).thenReturn(fixedClock); //called by LocalDateTime.now() and ZonedDateTime.now()
            coronaPassResult = coronaPassService.getCoronaPass(cpr, idwsMessageId);
        }
        assertTrue(coronaPassResult.isCacheHit());

        Cwt dkInfoContent = CwtTestUtils.getCwt(coronaPassResult.getCoronaPasses(), PassType.DKINFO);
        assertThat(dkInfoContent.getIssuedAt()).isEqualTo(afterDstEnds.toInstant());
        final CBORObject t = dkInfoContent.getClaim(-260).get(1).get("t");
        final CBORObject v = dkInfoContent.getClaim(-260).get(1).get("v");
        assertNotNull(t); //t should only have 'pcrTestExpiringIn15Minutes' and not 'pcrTest15MinutesOlderThanMaxAge'
        assertEquals(1, t.size());
        final Instant testExpiration = Instant.parse(t.get(0).get("xdu").AsString());
        assertEquals(afterDstEnds.plusMinutes(15).toInstant(), testExpiration);

        assertNotNull(v); //v should be included and its xdu property should be 2021-10-31T23:00:00Z (midnight local time)
        assertEquals(1, v.size());
        final Instant vacExpiration = Instant.parse(v.get(0).get("xdu").AsString());
        assertEquals(Instant.parse("2021-10-31T23:00:00Z"), vacExpiration);
    }

    private Vaccination noVaccination(String cpr) {
        return VaccinationBuilder.vaccination(cpr, false).build();
    }

    private LabResults noLabResults(String cpr) {
        return LabResultsBuilder.cpr(cpr).build();
    }

    private void given(String cpr, Vaccination vaccination, LabResults labResults) {
        when(dataReader.getData(cpr)).thenReturn(Collections.singleton(new Data(vaccination, labResults, toPersonRecord(MockDataService.person(cpr), cpr))));
    }

    @Test
    void test_positive_test_which_ends_up_as_eu_recovery_pass() {
        String cpr = "0405870115";
        LabResults labresults = Results.positivePcr(cpr, LocalDateTime.now().minusDays(11));
        given(cpr, noVaccination(cpr), labresults);

        CoronaPassResult coronaPassResult = coronaPassService.getCoronaPass(cpr, idwsMessageId);
        assertTrue(coronaPassResult.isCacheHit());
        assertNotNull(coronaPassResult.getCoronaPasses());
        DigitalGreenCertificate dgc = payloadToEUpasses(coronaPassResult.getCoronaPasses()).stream().collect(MoreCollectors.onlyElement());
        verifyEuImmunity(labresults, dgc);
    }

    private static void verifyEuImmunity(LabResults labresults, DigitalGreenCertificate dgc) {
        assertNull(dgc.getV());
        assertNull(dgc.getT());
        assertFalse(dgc.getR().isEmpty());
        RecoveryEntry recovery = Iterables.getOnlyElement(dgc.getR());
        LabResult.PositivePcrResult pcrResult = labresults.mostRecentPositivePcrResult();
        assertNotNull(pcrResult);
        assertEquals(pcrResult.samplingDateTime().toLocalDate(), recovery.getFr());
        assertEquals(pcrResult.samplingDateTime().plusDays(11).toLocalDate(), recovery.getDf());
        assertEquals(pcrResult.samplingDateTime().plusDays(180).toLocalDate(), recovery.getDu());
        assertEquals("DK", recovery.getCo());
        assertEquals(VaccineMapper.SNOMED_CT_COVID_CONCEPT_ID, recovery.getTg());
        assertEquals(VaccineMapper.ISSUER, recovery.getIs());
        assertNotNull(recovery.getCi());
    }

    @Test
    void test_positive_tests_and_only_immunization_eu_passes() {
        String cpr = "0405870115";

        given(cpr, noVaccination(cpr), Results.positivePcr(cpr, LocalDateTime.now().minusHours(14 * 24).minusMinutes(15)));

        CoronaPassResult coronaPassResult = coronaPassService.getCoronaPass(cpr, idwsMessageId);
        assertTrue(coronaPassResult.isCacheHit());
        assertNotNull(coronaPassResult.getCoronaPasses());
        DigitalGreenCertificate dgc = payloadToEUpasses(coronaPassResult.getCoronaPasses()).stream().collect(MoreCollectors.onlyElement());
        assertFalse(dgc.getR().isEmpty());
        assertNull(dgc.getV());
        assertNull(dgc.getT());
    }


    @Test
    void test_getCoronaPassOnlyEU() {
        String cpr = "0405870115";

        given(cpr, Vaccinations.completedSingleDose(cpr, LocalDateTime.now().minusWeeks(3)), noLabResults(cpr));

        CoronaPassResult coronaPassResult = coronaPassService.getCoronaPass(cpr, idwsMessageId);
        assertTrue(coronaPassResult.isCacheHit());
        assertNotNull(coronaPassResult.getCoronaPasses());
        assertEquals(4, coronaPassResult.getCoronaPasses().size());

        verifyDKContent(coronaPassResult, Instant.now());
        DigitalGreenCertificate vaccineDGC = payloadToEUpasses(coronaPassResult.getCoronaPasses()).stream().collect(MoreCollectors.onlyElement());

        String expectedGnt = MrzParser.toMrz(vaccineDGC.getNam().getGn(), -1);
        String expectedFnt = MrzParser.toMrz(vaccineDGC.getNam().getFn(), -1);
        assertEquals(expectedGnt, vaccineDGC.getNam().getGnt());
        assertEquals(expectedFnt, vaccineDGC.getNam().getFnt());

    }

    @Test
    void test_getCoronaPassEUMRZnames() {
        String cpr = cprWithCacheHitOn("2011451636");
        CoronaPassResult coronaPassResult = coronaPassService.getCoronaPass(cpr, idwsMessageId);
        assertTrue(coronaPassResult.isCacheHit());
        assertNotNull(coronaPassResult.getCoronaPasses());
        assertEquals(4, coronaPassResult.getCoronaPasses().size());
        List<DigitalGreenCertificate> listDGC = payloadToEUpasses(coronaPassResult.getCoronaPasses());
        assertEquals(1, listDGC.size());
        listDGC.forEach(certificate -> {
            String expectedGnt = MrzParser.toMrz(certificate.getNam().getGn(), -1);
            String expectedFnt = MrzParser.toMrz(certificate.getNam().getFn(), -1);
            assertEquals(expectedGnt, certificate.getNam().getGnt());
            assertEquals(expectedFnt, certificate.getNam().getFnt());
        });
    }

    @Test
    void test_mrzParsing() {
        String result = MrzParser.toMrz("ÍÐÅgeå bárðurþÞ", -1);
        assertEquals("IDAAGEAA<BARDURTHTH", result);
    }

    @Test
    void verify_euPcrMaxAge_takes_DST_change_into_account_and_sets_expiration_accordingly() {
        String cpr = "0405870117";
        final ZoneId zoneId = ZoneId.of("Europe/Copenhagen");
        final ZonedDateTime dayAfterDstEnds = ZonedDateTime.of(2021, 11, 1, 10, 0, 0, 0, zoneId);
        LocalDateTime samplingDateTime = dayAfterDstEnds.minus(euPcrMaxAge).plusMinutes(15).toLocalDateTime(); //test that change in DST is handled correctly, down to the minute
        given(cpr, noVaccination(cpr), Results.negativePcr(cpr, samplingDateTime));

        CoronaPassResult coronaPassResult;
        Clock fixedClock = Clock.fixed(dayAfterDstEnds.toInstant(), zoneId);
        try (MockedStatic<Clock> mockedClock = Mockito.mockStatic(Clock.class)) {
            mockedClock.when(Clock::systemUTC).thenReturn(fixedClock); //called by Instant.now()
            mockedClock.when(Clock::systemDefaultZone).thenReturn(fixedClock); //called by LocalDateTime.now()
            //Run getCoronaPass with fixed clock
            coronaPassResult = coronaPassService.getCoronaPass(cpr, idwsMessageId);
        }

        assertTrue(coronaPassResult.isCacheHit());
        assertNotNull(coronaPassResult.getCoronaPasses());

        List<Cwt> cwts = payloadToRawEUCWT(coronaPassResult.getCoronaPasses());
        assertEquals(1, cwts.size());

        for (Cwt cwt : cwts) {
            logger.info("CWT Expiration: " + cwt.getExpiration());
            assertTrue(timestampsAreApproxEqual(dayAfterDstEnds.toInstant(), cwt.getExpiration().minus(15, ChronoUnit.MINUTES)));
        }
    }

    @Test
    void verify_euPcrMaxAge_takes_DST_change_into_account_and_removes_TestEntry() {
        String cpr = "0405870117";
        final ZoneId zoneId = ZoneId.of("Europe/Copenhagen");
        final ZonedDateTime dayAfterDstEnds = ZonedDateTime.of(2021, 11, 1, 10, 0, 0, 0, zoneId);
        LocalDateTime samplingDateTime = dayAfterDstEnds.minus(euPcrMaxAge).minusMinutes(15).toLocalDateTime(); //test that change in DST is handled correctly, down to the minute
        given(cpr, noVaccination(cpr), Results.negativePcr(cpr, samplingDateTime));

        CoronaPassResult coronaPassResult;
        Clock fixedClock = Clock.fixed(dayAfterDstEnds.toInstant(), zoneId);
        try (MockedStatic<Clock> mockedClock = Mockito.mockStatic(Clock.class)) {
            mockedClock.when(Clock::systemUTC).thenReturn(fixedClock); //called by Instant.now()
            mockedClock.when(Clock::systemDefaultZone).thenReturn(fixedClock); //called by LocalDateTime.now() and ZonedDateTime.now()
            //Run getCoronaPass with fixed clock
            coronaPassResult = coronaPassService.getCoronaPass(cpr, idwsMessageId);
        }
        assertTrue(coronaPassResult.isCacheHit());
        assertNotNull(coronaPassResult.getCoronaPasses());

        List<Cwt> rawCwts = payloadToRawEUCWT(coronaPassResult.getCoronaPasses());
        assertEquals(0, rawCwts.size());
    }

    @Test
    void verify_euMaximumValidity_takes_DST_change_into_account() {
        String cpr = "0405870107";

        final ZoneId zoneId = ZoneId.of("Europe/Copenhagen");
        final ZonedDateTime dayBeforeDstEnds = ZonedDateTime.of(2021, 10, 30, 10, 0, 0, 0, zoneId); //test that a token expiration running into a DST change is handled correctly, down to the minute
        LocalDateTime effectuatedAt = dayBeforeDstEnds.minusDays(20).toLocalDateTime(); //note: the used effecutation time not important for this test

        final Vaccination vaccination = Vaccinations.completedSingleDose(cpr, effectuatedAt);

        given(cpr, vaccination, noLabResults(cpr));

        CoronaPassResult coronaPassResult;
        Clock fixedClock = Clock.fixed(dayBeforeDstEnds.toInstant(), zoneId);
        try (MockedStatic<Clock> mockedClock = Mockito.mockStatic(Clock.class)) {
            mockedClock.when(Clock::systemUTC).thenReturn(fixedClock); //called by Instant.now()
            mockedClock.when(Clock::systemDefaultZone).thenReturn(fixedClock); //called by LocalDateTime.now() and ZonedDateTime.now()
            //Run getCoronaPass with fixed clock
            coronaPassResult = coronaPassService.getCoronaPass(cpr, idwsMessageId);
        }
        assertTrue(coronaPassResult.isCacheHit());
        assertNotNull(coronaPassResult.getCoronaPasses());

        verifyDKContent(coronaPassResult, dayBeforeDstEnds.toInstant());

        List<Cwt> euCwts = payloadToRawEUCWT(coronaPassResult.getCoronaPasses());
        assertEquals(1, euCwts.size());

        for (Cwt cwt : euCwts) {
            assertThat(cwt.getExpiration()).isLessThan(dayBeforeDstEnds.toInstant().plus(euMaximumValidity).plus(1, ChronoUnit.MINUTES));
        }
    }


    private void verifyPass(String cpr) {
        CoronaPassResult coronaPassResult = coronaPassService.getCoronaPass(cpr, idwsMessageId);
        assertTrue(coronaPassResult.isCacheHit());
        assertNotNull(coronaPassResult.getCoronaPasses());
    }

    @Test
    void test_getCoronaPass_cache_miss() {
        final String cpr = "0405870101";

        given(cpr, null, null);
        CoronaPassResult coronaPassResult = coronaPassService.getCoronaPass(cpr, idwsMessageId);
        assertFalse(coronaPassResult.isCacheHit());
        assertNull(coronaPassResult.getCoronaPasses());
        assertNotNull(coronaPassResult.getJobId());
    }


    private void verifyDKContent(CoronaPassResult coronaPassResult, Instant issuedAtOrBefore) {
        Cwt dkContent = CwtTestUtils.getCwt(coronaPassResult.getCoronaPasses(), PassType.DK);
        Cwt dkPlusContent = CwtTestUtils.getCwt(coronaPassResult.getCoronaPasses(), PassType.DKPLUS);

        assertTrue(dkContent.getExpiration().isAfter(issuedAtOrBefore));
        assertThat(dkContent.getExpiration()).isLessThan(issuedAtOrBefore.plus(dkMaximumValidity).plus(1, ChronoUnit.MINUTES));
        assertFalse(dkContent.getIssuedAt().isAfter(issuedAtOrBefore));
        assertNotNull(UUID.fromString(new String(dkContent.getCwtId())));

        assertTrue(dkPlusContent.getExpiration().isAfter(issuedAtOrBefore));
        assertThat(dkPlusContent.getExpiration()).isLessThan(issuedAtOrBefore.plus(dkMaximumValidity).plus(1, ChronoUnit.MINUTES));
        assertFalse(dkPlusContent.getIssuedAt().isAfter(issuedAtOrBefore));
        assertNotNull(UUID.fromString(new String(dkPlusContent.getCwtId())));
        assertNotNull(dkPlusContent.getClaim("dateOfBirth"));
        assertNotNull(dkPlusContent.getClaim("legalName"));
    }

    private static List<DigitalGreenCertificate> payloadToEUpasses(Collection<CoronaPass> passes) {
        return passes.stream().filter(p -> p.getPassType() == PassType.EU).map(CoronaPass::getToken).map(s -> s.substring(4)).map(CwtTestUtils::getDigitalGreenCertificate).collect(Collectors.toList());
    }

    private static List<Cwt> payloadToRawEUCWT(Collection<CoronaPass> passes) {
        return passes.stream().filter(p -> p.getPassType() == PassType.EU).map(CoronaPass::getToken).map(CwtTestUtils::getCwt).collect(Collectors.toList());
    }

    private boolean timestampInRange(Instant start, Instant end, Instant subject) {
        return !subject.isBefore(start) && !subject.isAfter(end);
    }

    private boolean timestampIsApproxNow(Instant subject) {
        return timestampInRange(Instant.now().minusSeconds(2), Instant.now().plusSeconds(2), subject);
    }

    private boolean timestampsAreApproxEqual(Instant i1, Instant i2) {
        return timestampInRange(i1.minusSeconds(2), i1.plusSeconds(2), i2);
    }
}
