/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.cache;

import dk.sundhedsdatastyrelsen.cbs.dao.LabResultRecord;
import dk.sundhedsdatastyrelsen.cbs.dao.VaccinationRecord;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.LabResultRecordImpl;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.VaccinationRecordImpl;
import dk.sundhedsdatastyrelsen.cbs.datacache.Writer;
import dk.sundhedsdatastyrelsen.cbs.datacache.test.InMemoryDbExtension;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultsBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.EffectuatedVaccinationBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.VaccinationBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.service.config.TestConfiguration;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import dk.sundhedsdatastyrelsen.cbs.test.RandomData;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.test.context.TestPropertySource;
import org.springframework.test.context.junit.jupiter.SpringJUnitConfig;

import java.time.LocalDateTime;
import java.util.Map;
import java.util.Objects;
import java.util.UUID;
import java.util.function.Function;
import java.util.stream.Collectors;

import static com.google.common.truth.Truth.assertThat;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertTrue;

@SpringJUnitConfig(TestConfiguration.class)
@CbsTest
@ExtendWith(InMemoryDbExtension.class)
@TestPropertySource(properties = "cwt.children.include-in-response=true")
class DataReaderTest {
    public static final String PARENT = "2502801010";
    public static final String CHILD_1 = "0706128662";
    public static final String CHILD_2 = "2503107875";
    public static final String CHILD_3 = "1005189655";
    @Autowired
    private DataReader dataReader;
    @Autowired
    private Writer<LabResultRecord<LabResults>> labResultWriter;
    @Autowired
    private Writer<VaccinationRecord<Vaccination>> vaccinationWriter;
    @Autowired
    private JdbcTemplate jdbcTemplate;

    private final RandomData randomData = new RandomData(System.currentTimeMillis());

    @BeforeEach
    void setUp() {
        jdbcTemplate.execute("TRUNCATE TABLE " + LabResultRecord.INSERT_TABLE_NAME);
        jdbcTemplate.execute("TRUNCATE TABLE " + VaccinationRecord.INSERT_TABLE_NAME);
    }

    @Test
    void should_include_persons_children_when_enabled() {
        final Map<String, Data> data = getData(PARENT);

        assertThat(data.keySet()).containsExactly(PARENT, CHILD_1, CHILD_2, CHILD_3);
    }

    @Nested
    @TestPropertySource(properties = "cwt.children.include-in-response=false")
    @SpringJUnitConfig(TestConfiguration.class)
    class NoChildrenInResponse {
        @Autowired
        private DataReader noChildrenDataReader;

        @Test
        void should_not_include_persons_children_if_disabled() {
            final Map<String, Data> data = noChildrenDataReader.getData(PARENT).stream()
                .collect(Collectors.toMap(d -> d.personRecord().personIdentifier(), Function.identity()));

            assertThat(data.keySet()).containsExactly(PARENT);
        }
    }

    @Test
    void should_only_include_person_if_they_have_no_children() {
        final Map<String, Data> data = getData(CHILD_1);

        assertThat(data.keySet()).containsExactly(CHILD_1);
    }

    @Test
    void should_report_cache_miss_if_data_is_missing() {
        final Map<String, Data> data = getData(PARENT);

        assertThat(data.keySet()).containsExactly(PARENT, CHILD_1, CHILD_2, CHILD_3);

        assertTrue(data.values().stream().allMatch(Data::isCacheMiss));
        assertTrue(data.values().stream().map(Data::vaccination).allMatch(Objects::isNull));
        assertTrue(data.values().stream().map(Data::labResults).allMatch(Objects::isNull));
    }

    @Test
    void should_report_cache_miss_if_vaccinations_are_missing() {
        addLabResult(PARENT, CHILD_1, CHILD_2, CHILD_3);

        final Map<String, Data> data = getData(PARENT);

        assertThat(data.keySet()).containsExactly(PARENT, CHILD_1, CHILD_2, CHILD_3);

        assertTrue(data.values().stream().allMatch(Data::isCacheMiss));
        assertTrue(data.values().stream().map(Data::vaccination).allMatch(Objects::isNull));
        assertTrue(data.values().stream().map(Data::labResults).allMatch(Objects::nonNull));
    }

    @Test
    void should_report_cache_miss_if_lab_results_are_missing() {
        addVaccination(PARENT, CHILD_1, CHILD_2, CHILD_3);

        final Map<String, Data> data = getData(PARENT);

        assertThat(data.keySet()).containsExactly(PARENT, CHILD_1, CHILD_2, CHILD_3);

        assertTrue(data.values().stream().allMatch(Data::isCacheMiss));
        assertTrue(data.values().stream().map(Data::vaccination).allMatch(Objects::nonNull));
        assertTrue(data.values().stream().map(Data::labResults).allMatch(Objects::isNull));
    }

    @Test
    void should_not_report_cache_miss_if_data_is_present() {
        addVaccination(PARENT, CHILD_1, CHILD_2, CHILD_3);
        addLabResult(PARENT, CHILD_1, CHILD_2, CHILD_3);

        final Map<String, Data> data = getData(PARENT);

        assertThat(data.keySet()).containsExactly(PARENT, CHILD_1, CHILD_2, CHILD_3);

        assertTrue(data.values().stream().noneMatch(Data::isCacheMiss));
        assertTrue(data.values().stream().map(Data::vaccination).allMatch(Objects::nonNull));
        assertTrue(data.values().stream().map(Data::labResults).allMatch(Objects::nonNull));
    }

    @Test
    void all_combinations() {
        addVaccination(PARENT, CHILD_1);
        addLabResult(PARENT, CHILD_2);

        final Map<String, Data> data = getData(PARENT);
        assertThat(data.keySet()).containsExactly(PARENT, CHILD_1, CHILD_2, CHILD_3);

        assertFalse(data.get(PARENT).isCacheMiss());
        assertNotNull(data.get(PARENT).vaccination());
        assertNotNull(data.get(PARENT).labResults());

        assertTrue(data.get(CHILD_1).isCacheMiss());
        assertNotNull(data.get(CHILD_1).vaccination());
        assertNull(data.get(CHILD_1).labResults());

        assertTrue(data.get(CHILD_2).isCacheMiss());
        assertNull(data.get(CHILD_2).vaccination());
        assertNotNull(data.get(CHILD_2).labResults());

        assertTrue(data.get(CHILD_3).isCacheMiss());
        assertNull(data.get(CHILD_3).vaccination());
        assertNull(data.get(CHILD_3).labResults());
    }

    private void addVaccination(String... cprs) {
        for (String cpr : cprs) {

            vaccinationWriter.write(new VaccinationRecordImpl<Vaccination>()
                .cpr(cpr)
                .traceId(UUID.randomUUID().toString())
                .creationTime(LocalDateTime.now())
                .setData(VaccinationBuilder.vaccination(cpr, randomData.nextBoolean())
                    .add(EffectuatedVaccinationBuilder.effectuatedVaccination(randomData.nextLong(), randomData.localDateTime(), randomData.countryCode()).build())
                    .build())
            );
        }
    }

    private void addLabResult(String... cprs) {
        for (String cpr : cprs) {
            labResultWriter.write(new LabResultRecordImpl<LabResults>()
                .cpr(cpr)
                .traceId(UUID.randomUUID().toString())
                .creationTime(LocalDateTime.now())
                .setData(LabResultsBuilder.cpr(cpr)
                    .add(LabResultBuilder.pcrResult(randomData.nextInt(), randomData.localDateTime(), randomData.localDateTime()).negative().build())
                    .build()
                )
            );
        }
    }

    private Map<String, Data> getData(String cpr) {
        return dataReader.getData(cpr).stream()
            .collect(Collectors.toMap(data -> data.personRecord().personIdentifier(), Function.identity()));
    }


}
