/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.handler;

import dk.sundhedsdatastyrelsen.cbs.service.config.EUCovidTestCertificateSigningConfiguration;
import dk.sundhedsdatastyrelsen.cbs.service.config.SignatureConfiguration;
import dk.sundhedsdatastyrelsen.cbs.service.exceptions.CertificateSigningException;
import dk.sundhedsdatastyrelsen.cbs.service.model.SignedCertificateCWT;
import dk.sundhedsdatastyrelsen.cbs.service.util.EUAntigenFiltration;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.*;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.mockito.Mock;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.TestPropertySource;
import org.springframework.test.context.junit.jupiter.SpringJUnitConfig;
import se.digg.dgc.service.DGCEncoder;

import java.text.MessageFormat;
import java.time.*;
import java.util.stream.Stream;

import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.lenient;

@SpringJUnitConfig({SignatureConfiguration.class, EUCovidTestCertificateSigningConfiguration.class})
@ActiveProfiles(SignatureConfiguration.LOCAL_KEY_STORE)
@CbsTest
@TestPropertySource(properties = {
    "eu.covid.test.certificate.signing.cwt.test-result.maximum-validity=P7D"
})
class EUCovidTestCertificateSigningHandlerTest {

    public static final String APPROVED_MANUFACTURER = "1833";
    @Autowired
    private DGCEncoder encoder;

    @Autowired
    private EUCovidTestCertificateSigningConfiguration configuration;
    @Mock
    private EUAntigenFiltration euAntigenFiltration;

    private static final Duration testCwtMaxValidity = Duration.ofDays(7L);
    private EUCovidTestCertificateSigningHandler signingHandler;
    private static final ZonedDateTime now = ZonedDateTime.now(ZoneId.of("Europe/Copenhagen"));


    @BeforeEach
    void setUp() {
        signingHandler = new EUCovidTestCertificateSigningHandler(encoder, euAntigenFiltration, configuration);
        lenient().when(euAntigenFiltration.isIdDeviceLegalAtTime(eq(APPROVED_MANUFACTURER), any())).thenReturn(true);
    }

    @ParameterizedTest(name = "{index}: {1}")
    @MethodSource("validRequests")
    void should_issue_certificate_for_valid_request(GetSignedEUCovidTestCertificateRequest request, @SuppressWarnings("unused") String testName) {
        final SignedCertificateCWT signedCertificateCWT = signingHandler.getSignedCertificateCWT(request);

        assertNotNull(signedCertificateCWT.getCwt());
        assertNotNull(signedCertificateCWT.getCi());
    }

    static Stream<Arguments> validRequests() {
        return Stream.of(
            createRequest(TestType.PCR, TestResult.NEGATIVE, now, null),
            createRequest(TestType.PCR, TestResult.POSITIVE, now, null),
            createRequest(TestType.RAT, TestResult.NEGATIVE, now, APPROVED_MANUFACTURER),
            createRequest(TestType.PCR, TestResult.POSITIVE, now.minus(testCwtMaxValidity.plusMinutes(5L)), null)
            ).map(request -> Arguments.arguments(request, name(request)));
    }

    @ParameterizedTest(name = "{index}: {1}")
    @MethodSource("invalidRequests")
    void should_not_issue_certificate_for_invalid_request(GetSignedEUCovidTestCertificateRequest request, @SuppressWarnings("unused") String testName) {
        assertThrows(CertificateSigningException.class, () -> signingHandler.getSignedCertificateCWT(request));
    }

    static Stream<Arguments> invalidRequests() {
        return Stream.of(
            createRequest(TestType.PCR, TestResult.NEGATIVE, now.minus(testCwtMaxValidity.plusMinutes(5L)), null),
            createRequest(TestType.PCR, TestResult.POSITIVE, now.minusDays(200), null),
            createRequest(TestType.RAT, TestResult.POSITIVE, now, APPROVED_MANUFACTURER),
            createRequest(TestType.RAT, TestResult.NEGATIVE, now.minus(testCwtMaxValidity.plusMinutes(5L)), APPROVED_MANUFACTURER),
            createRequest(TestType.RAT, TestResult.POSITIVE, now.minus(testCwtMaxValidity.plusMinutes(5L)), APPROVED_MANUFACTURER),
            createRequest(TestType.RAT, TestResult.POSITIVE, now, "1234"),
            createRequest(TestType.RAT, TestResult.NEGATIVE, now, "1234")
        ).map(request -> Arguments.arguments(request, name(request)));
    }

    private static String name(GetSignedEUCovidTestCertificateRequest request) {
        return MessageFormat.format("Type: {0} Result: {1} Manufacturer: {2} SampleCollected: {3}",
            request.getPayload().getTypeOfTest(),
            request.getPayload().getTestResult(),
            request.getPayload().getManufacturer(),
            request.getPayload().getSampleCollected());
    }


    private static GetSignedEUCovidTestCertificateRequest createRequest(TestType testType, TestResult testResult, ZonedDateTime sampleCollected, String manufacturer) {
        final GetSignedEUCovidTestCertificateRequest request = new GetSignedEUCovidTestCertificateRequest();
        final TestSample payload = new TestSample();
        payload.setGivenName("Brian");
        payload.setFamilyName("London");
        payload.setDateOfBirth(LocalDate.of(1992, Month.DECEMBER, 17));
        payload.setCountryOfTest(CountryOfTest.DK);
        payload.setDiseaseAgentTargeted(DiseaseAgent.COVID_19);
        payload.setTestingCenter("Billund Lufthavn - Falck A/S");
        payload.setManufacturer(manufacturer);
        payload.setSampleCollected(sampleCollected);
        payload.setTypeOfTest(testType);
        payload.setTestResult(testResult);

        request.setPersonIdentifier("12345678");
        request.setPayload(payload);
        return request;
    }
}
