/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.passports;

import com.google.common.collect.Range;
import com.google.common.truth.Truth;
import dk.sundhedsdatastyrelsen.cbs.dao.EUAntigenSpec;
import dk.sundhedsdatastyrelsen.cbs.dao.PersonRecord;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.EUAntigenSpecImpl;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.PersonRecordImpl;
import dk.sundhedsdatastyrelsen.cbs.datacache.CollectionReader;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultsBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.VaccinationBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.logging.AuditData;
import dk.sundhedsdatastyrelsen.cbs.rules.EuRecoveryPassportEvaluator;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.CombinedPassportEvaluator;
import dk.sundhedsdatastyrelsen.cbs.service.config.DkPassportConfiguration;
import dk.sundhedsdatastyrelsen.cbs.service.config.EuPassportConfiguration;
import dk.sundhedsdatastyrelsen.cbs.service.config.SignatureConfiguration;
import dk.sundhedsdatastyrelsen.cbs.service.config.VaccineMapperConfiguration;
import dk.sundhedsdatastyrelsen.cbs.service.model.CoronaPass;
import dk.sundhedsdatastyrelsen.cbs.service.model.PassType;
import dk.sundhedsdatastyrelsen.cbs.service.passports.model.Custody;
import dk.sundhedsdatastyrelsen.cbs.service.passports.model.DkRules;
import dk.sundhedsdatastyrelsen.cbs.service.passports.model.EuRules;
import dk.sundhedsdatastyrelsen.cbs.service.util.EUAntigenFiltration;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import dk.sundhedsdatastyrelsen.cbs.test.CwtTestUtils;
import dk.sundhedsdatastyrelsen.cbs.test.Data.Result;
import dk.sundhedsdatastyrelsen.cbs.test.Data.Results;
import dk.sundhedsdatastyrelsen.cbs.test.Data.Vaccinations;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Primary;
import org.springframework.context.annotation.Profile;
import org.springframework.jdbc.datasource.DriverManagerDataSource;
import org.springframework.jdbc.datasource.lookup.JndiDataSourceLookup;
import org.springframework.mock.jndi.SimpleNamingContextBuilder;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.NestedTestConfiguration;
import org.springframework.test.context.TestPropertySource;
import org.springframework.test.context.junit.jupiter.SpringJUnitConfig;
import se.digg.dgc.payload.v1.DGCSchemaVersion;
import se.digg.dgc.payload.v1.Eudgc;
import se.digg.dgc.service.DGCEncoder;
import se.digg.dgc.service.impl.DefaultDKEncoder;
import se.digg.dgc.signatures.cwt.Cwt;

import javax.naming.NamingException;
import java.time.Duration;
import java.time.Instant;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static java.time.LocalDateTime.now;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

@SpringJUnitConfig({CombinedPassportFactoryTest.EuAgFilterMockConfiguration.class, CombinedPassportFactory.class, DkPassportFactory.class, EuPassportFactory.class, SignatureConfiguration.class, DkPassportConfiguration.class, EuPassportConfiguration.class, VaccineMapperConfiguration.class})
@ActiveProfiles(SignatureConfiguration.LOCAL_KEY_STORE)
@TestPropertySource(properties = {
    "cwt.eu.ag-result.maximum-age=P7D",
    "cwt.eu.pcr-result.maximum-age=P14D",
    "cwt.eu.ag-result.limit=5",
    "cwt.eu.pcr-result.limit=5",
    "cwt.children.include-in-response=true",
    "cwt.children.include-in-response.dk-passes=false"
})
@CbsTest
class CombinedPassportFactoryTest {
    @Value("${cwt.children.include-in-response.dk-passes}")
    private boolean makeChildrenDkPasses;
    @Autowired
    private EuPassportConfiguration euPassportConfiguration;
    @Autowired
    private DGCEncoder euEncoder;
    @Autowired
    private VaccineMapper vaccineMapper;
    @Autowired
    @EuRules
    private EuRecoveryPassportEvaluator euRecoveryPassportEvaluator;
    @Autowired
    private CombinedPassportFactory combinedPassportFactory;
    @Autowired
    private DkPassportFactory dkPassportFactory;
    @Autowired
    private EuPassportFactory euPassportFactory;
    @Autowired
    private EUAntigenFiltration euAntigenFiltrationWithMockReader;
    public static final String CPR = "0506504003";
    public static final PersonRecord PERSON_RECORD = new PersonRecordImpl()
        .personIdentifier(CPR)
        .givenName("John Johnny")
        .familyName("Johnson");
    private final Vaccination noVaccinations = VaccinationBuilder.vaccination(CPR, false).build();
    private final LabResults noLabResults = LabResultsBuilder.cpr(CPR).build();

    @Profile(SignatureConfiguration.LOCAL_KEY_STORE)
    @Configuration
    static class EuAgFilterMockConfiguration {
        @Bean
        @Primary
        public EUAntigenFiltration euAntigenFiltrationWithMockReader() {
            CollectionReader<EUAntigenSpec, String> euAntigenSpecReader = mock(CollectionReader.class);
            when(euAntigenSpecReader.search(TestData.UNAPPROVED_MANUFACTURER)).thenReturn(Collections.emptyList());
            when(euAntigenSpecReader.search(TestData.APPROVED_MANUFACTURER)).thenReturn(Collections.singletonList(new EUAntigenSpecImpl().validFrom(now().minusDays(100)).validTo(now().plusDays(1))));
            return new EUAntigenFiltration(null, euAntigenSpecReader, false);
        }
    }

    @BeforeAll
    static void beforeAll() throws NamingException {
        //noinspection deprecation
        SimpleNamingContextBuilder.emptyActivatedContextBuilder();
        new JndiDataSourceLookup().getJndiTemplate().bind("java:jboss/datasources/euantigenspec", new DriverManagerDataSource("jdbc:h2:mem:testdb"));
    }

    @Test
    void grant_no_passes_for_person_with_no_vaccination_and_no_test_results() {
        final Collection<CoronaPass> pass = getCoronaPasses(noVaccinations, noLabResults);

        assertEquals(1, getDKInfoPass(pass).size());
        assertEquals(0, getDKPass(pass).size());
        assertEquals(0, getEUPass(pass, "vac").size());
        assertEquals(0, getEUPass(pass, "rec").size());
        assertEquals(0, getEUPass(pass, "lab").size());
    }

    @Test
    void grant_pass_for_recovery() {
        final LabResults labResultsWithRecovery = Results.positivePcr(now().minusWeeks(8).minusDays(1));
        final Collection<CoronaPass> pass = getCoronaPasses(noVaccinations, labResultsWithRecovery);

        assertEquals(2, getDKPass(pass).size());
        assertEquals(0, getEUPass(pass, "vac").size());
        assertEquals(1, getEUPass(pass, "rec").size());
        assertEquals(0, getEUPass(pass, "lab").size());
    }

    @Test
    void grant_pass_for_negative_lab_result() {
        final LabResults negativeLabResults = Results.negativePcr(LocalDateTime.now().minusDays(1));
        final Collection<CoronaPass> pass = getCoronaPasses(noVaccinations, negativeLabResults);

        assertEquals(2, getDKPass(pass).size());
        assertEquals(0, getEUPass(pass, "vac").size());
        assertEquals(0, getEUPass(pass, "rec").size());
        assertEquals(1, getEUPass(pass, "lab").size());
    }

    @Test
    void grant_pass_for_incomplete_vaccination() {
        final Vaccination vaccination = Vaccinations.partial(LocalDateTime.now().minusDays(15));
        final Collection<CoronaPass> pass = getCoronaPasses(vaccination, noLabResults);

        assertEquals(2, getDKPass(pass).size());
        assertEquals(1, getEUPass(pass, "vac").size());
        assertEquals(0, getEUPass(pass, "rec").size());
        assertEquals(0, getEUPass(pass, "lab").size());
    }

    @Test
    void grant_pass_for_complete_single_dose_vaccination() {
        final Vaccination vaccination = Vaccinations.completedSingleDose(LocalDateTime.now().minusDays(15));
        final Collection<CoronaPass> pass = getCoronaPasses(vaccination, noLabResults);

        assertEquals(2, getDKPass(pass).size());
        assertEquals(1, getEUPass(pass, "vac").size());
        assertEquals(0, getEUPass(pass, "rec").size());
        assertEquals(0, getEUPass(pass, "lab").size());
    }

    @Test
    void grant_pass_for_complete_single_dose_vaccination_with_additional_vaccination() {
        final Vaccination vaccination = Vaccinations.completedSingleDoseWithAdditionalVaccinations(LocalDateTime.now().minusDays(15), 1);
        final Collection<CoronaPass> pass = getCoronaPasses(vaccination, noLabResults);

        assertEquals(2, getDKPass(pass).size());
        assertEquals(2, getEUPass(pass, "vac").size()); //includes additional vaccinations
        assertEquals(0, getEUPass(pass, "rec").size());
        assertEquals(0, getEUPass(pass, "lab").size());
    }


    @Test
    void grant_pass_for_complete_double_dose_vaccination() {
        final Vaccination vaccination = Vaccinations.completedMultiDose(LocalDateTime.now().minusDays(15));
        final Collection<CoronaPass> pass = getCoronaPasses(vaccination, noLabResults);

        assertEquals(2, getDKPass(pass).size());
        assertEquals(2, getEUPass(pass, "vac").size());
        assertEquals(0, getEUPass(pass, "rec").size());
        assertEquals(0, getEUPass(pass, "lab").size());
    }

    @Test
    void grant_pass_for_complete_double_dose_vaccination_with_additional_vaccination() {
        final Vaccination vaccination = Vaccinations.completedMultiDoseWithAdditionalVaccinations(LocalDateTime.now().minusDays(15), 1);
        final Collection<CoronaPass> pass = getCoronaPasses(vaccination, noLabResults);

        assertEquals(2, getDKPass(pass).size());
        assertEquals(3, getEUPass(pass, "vac").size()); //includes additional vaccinations
        assertEquals(0, getEUPass(pass, "rec").size());
        assertEquals(0, getEUPass(pass, "lab").size());
    }

    @Test
    void grant_all_possible_passes() {
        final Vaccination vaccination = Vaccinations.completedSingleDoseWithAdditionalVaccinations(LocalDateTime.now().minusDays(20), 2);
        final LabResults labResults = LabResultsBuilder.cpr(CPR)
            .add(Result.negativePcr(LocalDateTime.now().minusDays(1)))
            .add(Result.positivePcr(LocalDateTime.now().minusWeeks(8).minusDays(1)))
            .build();

        final Collection<CoronaPass> pass = getCoronaPasses(vaccination, labResults);
        assertEquals(2, getDKPass(pass).size());
        assertEquals(3, getEUPass(pass, "vac").size()); //includes additional vaccinations
        assertEquals(1, getEUPass(pass, "rec").size());
        assertEquals(1, getEUPass(pass, "lab").size());
    }

    @Test
    void include_old_pcr_results_until_max_age() {
        final LocalDateTime now = LocalDateTime.now();
        assertTrue(now.minus(euPassportConfiguration.getPcrResultsMaxAge()).isEqual(now.minusWeeks(2)), "precondition failed");
        final LabResults labResults = Results.negativePcr(now.minusDays(1), now.minusWeeks(1), now.minusWeeks(3));

        final Collection<CoronaPass> pass = getCoronaPasses(noVaccinations, labResults);
        assertEquals(2, getDKPass(pass).size());
        assertEquals(0, getEUPass(pass, "vac").size());
        assertEquals(0, getEUPass(pass, "rec").size());
        assertEquals(2, getEUPass(pass, "lab").size());
    }

    @Test
    void include_old_ag_results_until_max_age() {
        final LocalDateTime now = LocalDateTime.now();
        assertTrue(now.minus(euPassportConfiguration.getAntiGenResultsMaxAge()).isEqual(now.minusWeeks(1)), "precondition failed");
        final LabResults labResults = Results.negativeAntigen(now.minusDays(1), now.minusDays(6), now.minusDays(8));

        final Collection<CoronaPass> pass = getCoronaPasses(noVaccinations, labResults);
        assertEquals(2, getDKPass(pass).size());
        assertEquals(0, getEUPass(pass, "vac").size());
        assertEquals(0, getEUPass(pass, "rec").size());
        assertEquals(2, getEUPass(pass, "lab").size());
    }

    @Test
    void do_not_include_positive_results_when_not_configured_to_do_so() {
        assertFalse(euPassportConfiguration.isIncludePositiveResults(), "precondition failed");
        final LabResults labResults = LabResultsBuilder.cpr(CPR)
            .add(Result.positiveAntigen(LocalDateTime.now().minusDays(1)))
            .add(Result.negativeAntigen(LocalDateTime.now().minusDays(2)))
            .build();

        final Collection<CoronaPass> pass = getCoronaPasses(noVaccinations, labResults);
        assertEquals(0, getDKPass(pass).size());
        assertEquals(0, getEUPass(pass, "vac").size());
        assertEquals(0, getEUPass(pass, "rec").size());
        assertEquals(1, getEUPass(pass, "lab").size());
    }

    @Test
    void ag_result_and_pcr_result_max_ages_are_independent() {
        final LocalDateTime now = LocalDateTime.now();
        assertTrue(now.minus(euPassportConfiguration.getAntiGenResultsMaxAge()).isEqual(now.minusWeeks(1)), "precondition failed");
        assertTrue(now.minus(euPassportConfiguration.getPcrResultsMaxAge()).isEqual(now.minusWeeks(2)), "precondition failed");

        final LabResults labResults = LabResultsBuilder.cpr(CPR)
            .add(Result.negativeAntigen(now.minusDays(1))) // < ag-max-age
            .add(Result.negativePcr(now.minusDays(1)))     // < pcr-max-age
            .add(Result.negativeAntigen(now.minusDays(8))) // > ag-max-age
            .add(Result.negativePcr(now.minusDays(8)))     // > pcr-max-age
            .build();

        final Collection<CoronaPass> pass = getCoronaPasses(noVaccinations, labResults);

        assertEquals(2, getDKPass(pass).size());
        assertEquals(0, getEUPass(pass, "vac").size());
        assertEquals(0, getEUPass(pass, "rec").size());
        assertEquals(3, getEUPass(pass, "lab").size());
    }

    @Test
    void passport_is_version_1_3_0() {
        final Vaccination vaccination = Vaccinations.completedSingleDose(LocalDateTime.now().minusDays(15));

        final Collection<CoronaPass> pass = getCoronaPasses(vaccination, noLabResults);

        assertTrue(getPassByType(pass, PassType.EU)
            .map(CoronaPass::getToken)
            .map(CwtTestUtils::getDigitalGreenCertificate)
            .map(Eudgc::getVer)
            .allMatch(v -> v.equalsIgnoreCase(DGCSchemaVersion.DGC_SCHEMA_VERSION)));
    }

    @Test
    void should_limit_test_certificate_amount() {
        final LocalDateTime sampleTime1 = LocalDateTime.now().minusHours(72);
        final LocalDateTime sampleTime2 = LocalDateTime.now().minusHours(71);
        final LocalDateTime sampleTime3 = LocalDateTime.now().minusHours(70);
        final LocalDateTime sampleTime4 = LocalDateTime.now().minusHours(69);
        final LocalDateTime sampleTime5 = LocalDateTime.now().minusHours(68);
        final LocalDateTime sampleTime6 = LocalDateTime.now().minusHours(67);
        final LocalDateTime sampleTime7 = LocalDateTime.now().minusHours(66);
        final LocalDateTime sampleTime8 = LocalDateTime.now().minusHours(65);
        final LocalDateTime sampleTime9 = LocalDateTime.now().minusHours(64);

        final LabResults labResults = LabResultsBuilder.cpr(CPR)
            .addAll(Results.negativePcr(sampleTime1, sampleTime2, sampleTime3, sampleTime4, sampleTime5, sampleTime6).results())
            .add(Result.positivePcr(sampleTime7))
            .addAll(Results.negativeAntigen(sampleTime1, sampleTime2, sampleTime3, sampleTime4, sampleTime5, sampleTime6, sampleTime7, sampleTime8).results())
            .add(Result.positiveAntigen(sampleTime9))
            .build();

        final Collection<CoronaPass> coronaPasses = getCoronaPasses(noVaccinations, labResults);
        final Collection<CoronaPass> lab = getEUPass(coronaPasses, "lab");
        assertEquals(10, lab.size());
    }

    @Test
    void do_not_include_children_dk_passes_when_not_configured_to_do_so() {
        assertFalse(makeChildrenDkPasses, "precondition failed");
        final Vaccination completedMultiDoseVaccination = Vaccinations.completedMultiDose(now());
        final LabResults negativePcr = Results.negativePcr(now());
        final Set<CoronaPass> pass = combinedPassportFactory.createPassportFrom(completedMultiDoseVaccination, negativePcr, PERSON_RECORD, new AuditData(), Custody.child(1));
        assertEquals(0, getDKPass(pass).size());
        assertEquals(1, getDKInfoPass(pass).size());
        assertEquals(2, getEUPass(pass, "vac").size());
        assertEquals(0, getEUPass(pass, "rec").size());
        assertEquals(1, getEUPass(pass, "lab").size());
    }

    @Nested
    @NestedTestConfiguration(NestedTestConfiguration.EnclosingConfiguration.INHERIT)
    @TestPropertySource(properties = {
        "cwt.children.include-in-response=true",
        "cwt.children.include-in-response.dk-passes=true"
    })
    @SpringJUnitConfig({CombinedPassportFactoryTest.EuAgFilterMockConfiguration.class, CombinedPassportFactory.class, DkPassportFactory.class, EuPassportFactory.class, SignatureConfiguration.class, DkPassportConfiguration.class, EuPassportConfiguration.class, VaccineMapperConfiguration.class})
    class IncludesChildrenDkPasses {
        @Autowired
        private CombinedPassportFactory nestedCombinedPassportFactory;
        @Value("${cwt.children.include-in-response.dk-passes}")
        private boolean makeChildrenDkPasses;

        @Test
        void include_children_dk_passes_when_configured() {
            assertTrue(makeChildrenDkPasses, "precondition failed");

            final LocalDateTime now = now();
            final Vaccination vaccinations = Vaccinations.completedMultiDoseWithAdditionalVaccinations(now, 2);
            final LabResults negativePcr = Results.negativePcr(now);
            final Set<CoronaPass> pass = nestedCombinedPassportFactory.createPassportFrom(vaccinations, negativePcr, PERSON_RECORD, new AuditData(), Custody.child(1));
            assertEquals(2, getDKPass(pass).size());
            assertEquals(1, getDKInfoPass(pass).size());
            assertEquals(4, getEUPass(pass, "vac").size());
            assertEquals(0, getEUPass(pass, "rec").size());
            assertEquals(1, getEUPass(pass, "lab").size());
        }
    }

    @Nested
    @NestedTestConfiguration(NestedTestConfiguration.EnclosingConfiguration.INHERIT)
    @TestPropertySource(properties = {
        "cwt.eu.include-positive-results=true"
    })
    @SpringJUnitConfig(EuPassportConfiguration.class)
    class IncludesPositiveResults {
        @Autowired
        private EuPassportConfiguration nestedEuPassportConfiguration;

        @Test
        void include_positive_results_when_configured() {
            final ArrayList<PassportFactory> passportFactories = new ArrayList<>();
            passportFactories.add(new EuPassportFactory(euEncoder, vaccineMapper, euRecoveryPassportEvaluator, euAntigenFiltrationWithMockReader, nestedEuPassportConfiguration));
            passportFactories.add(dkPassportFactory);
            combinedPassportFactory = new CombinedPassportFactory(passportFactories);

            final LabResults labResults = LabResultsBuilder.cpr(CPR)
                .add(Result.positiveAntigen(LocalDateTime.now().minusDays(1)))
                .add(Result.negativeAntigen(LocalDateTime.now().minusDays(2)))
                .build();
            final Collection<CoronaPass> pass = getCoronaPasses(noVaccinations, labResults);

            assertEquals(0, getDKPass(pass).size());
            assertEquals(0, getEUPass(pass, "vac").size());
            assertEquals(0, getEUPass(pass, "rec").size());
            assertEquals(2, getEUPass(pass, "lab").size());
        }
    }

    @Test
    void ag_test_eu_approved_included() {
        final LabResults labResults = LabResultsBuilder.cpr(CPR)
            .add(Result.negativeEuApprovedAntigen(LocalDateTime.now().minusDays(1)))
            .build();
        final Collection<CoronaPass> pass = getCoronaPasses(noVaccinations, labResults);
        assertEquals(1, getEUPass(pass, "lab").size());
    }

    @Test
    void ag_test_not_eu_approved_not_included() {
        final LabResults labResults = LabResultsBuilder.cpr(CPR)
            .add(Result.negativeEuUnapprovedAntigen(LocalDateTime.now().minusDays(1)))
            .build();
        final Collection<CoronaPass> pass = getCoronaPasses(noVaccinations, labResults);
        assertEquals(0, getEUPass(pass, "lab").size());
    }

    @Nested
    @NestedTestConfiguration(NestedTestConfiguration.EnclosingConfiguration.INHERIT)
    @TestPropertySource(properties = {
        "cwt.dk.maximum-validity=PT10H"
    })
    @SpringJUnitConfig({DkPassportConfiguration.class, SignatureConfiguration.class})
    class LongExpirationTest {

        @Autowired
        private DkPassportConfiguration nestedDkPassportConfiguration;

        @Autowired
        private DefaultDKEncoder dkEncoder;

        @Autowired
        @DkRules
        private CombinedPassportEvaluator dkEvaluator;

        @Test
        void expiration_of_DK_cwt_is_based_on_the_test_that_grants_it() {
            final ArrayList<PassportFactory> passportFactories = new ArrayList<>();
            passportFactories.add(euPassportFactory);
            passportFactories.add(new DkPassportFactory(dkEncoder, dkEvaluator, nestedDkPassportConfiguration));
            combinedPassportFactory = new CombinedPassportFactory(passportFactories);

            // Vaccination becomes valid in 4 days
            final Vaccination vaccination = Vaccinations.partial("0506504003", now().minusDays(10L));

            // Test is valid for 4 more hours
            final LabResults labResults = Results.negativePcr("0506504003", now().minusHours(92L));

            final Collection<CoronaPass> coronaPasses = getCoronaPasses(vaccination, labResults);
            final Cwt cwt = CwtTestUtils.getCwt(coronaPasses, PassType.DK);
            final Instant expiration = cwt.getExpiration();

            // DK token should be valid for 4 hours
            Truth.assertThat(expiration).isIn(Range.closed(Instant.now(), Instant.now().plus(Duration.ofHours(5L))));
        }
    }

    private Collection<CoronaPass> getEUPass(Collection<CoronaPass> pass, String euType) {
        return getPassByType(pass, PassType.EU)
            .filter(coronaPass -> coronaPass.hasProperty("eu:keytype", euType))
            .collect(Collectors.toList());
    }

    private Collection<CoronaPass> getDKPass(Collection<CoronaPass> pass) {
        return Stream.concat(getPassByType(pass, PassType.DK), getPassByType(pass, PassType.DKPLUS)).collect(Collectors.toList());
    }

    private Collection<CoronaPass> getDKInfoPass(Collection<CoronaPass> pass) {
        return getPassByType(pass, PassType.DKINFO).collect(Collectors.toList());
    }

    private Stream<CoronaPass> getPassByType(Collection<CoronaPass> pass, PassType type) {
        return pass.stream().filter(coronaPass -> coronaPass.getPassType() == type);
    }

    private Collection<CoronaPass> getCoronaPasses(Vaccination vaccination, LabResults labResults) {
        return combinedPassportFactory.createPassportFrom(vaccination, labResults, PERSON_RECORD, new AuditData(), null);
    }
}
