/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.passports;

import com.upokecenter.cbor.CBORObject;
import dk.sundhedsdatastyrelsen.cbs.dao.PersonRecord;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.PersonRecordImpl;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultsBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.VaccinationBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.EffectuatedVaccination;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.logging.AuditData;
import dk.sundhedsdatastyrelsen.cbs.service.config.DkPassportConfiguration;
import dk.sundhedsdatastyrelsen.cbs.service.config.SignatureConfiguration;
import dk.sundhedsdatastyrelsen.cbs.service.config.VaccineMapperConfiguration;
import dk.sundhedsdatastyrelsen.cbs.service.model.CoronaPass;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import dk.sundhedsdatastyrelsen.cbs.test.CwtTestUtils;
import dk.sundhedsdatastyrelsen.cbs.test.Data.Result;
import dk.sundhedsdatastyrelsen.cbs.test.Data.Results;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.junit.jupiter.SpringJUnitConfig;
import se.digg.dgc.signatures.cwt.Cwt;

import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.temporal.TemporalAmount;
import java.util.List;
import java.util.Set;

import static dk.sundhedsdatastyrelsen.cbs.test.Data.Vaccinations;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.argThat;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.verify;

@CbsTest
@SpringJUnitConfig({DkPassportConfiguration.class, VaccineMapperConfiguration.class, SignatureConfiguration.class})
@ActiveProfiles(SignatureConfiguration.LOCAL_KEY_STORE)
class DkInfoPassportFactoryTest {
    public static final String CPR = "0506504003";
    public static final String CPR_UNDER_18 = "0506205002";
    public static final String DOB = "1950-06-05";

    public static final PersonRecord PERSON_RECORD = new PersonRecordImpl()
        .personIdentifier(CPR)
        .givenName("John Johnny")
        .familyName("Johnson");

    public static final PersonRecord PERSON_RECORD_UNDER_18 = new PersonRecordImpl()
        .personIdentifier(CPR_UNDER_18)
        .givenName("John Johnny Junior")
        .familyName("Johnson");

    @Autowired
    private DkInfoPassportFactory dkInfoPassportFactory;

    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.vaccination.validator.maximum-amount-after-first-vaccination-of-partially-completed-vaccination-schedule:P42D}')}")
    private TemporalAmount maximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule;
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.vaccination.validator.maximum-amount-after-completed-vaccination-schedule:P7M}')}")
    private TemporalAmount maximumAmountAfterCompletedVaccinationSchedule;
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.vaccination.validator.maximum-amount-after-completed-vaccination-schedule-under-18:P5Y}')}")
    private TemporalAmount maximumAmountAfterCompletedVaccinationScheduleUnder18;
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.revaccination.validator.maximum-period-after-revaccination:P3Y}')}")
    private TemporalAmount maximumPeriodAfterRevaccination;
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.recovery.validator.maximum-period-after-recovery}')}")
    private TemporalAmount maximumPeriodAfterRecovery;
    @Value("#{T(dk.sundhedsdatastyrelsen.cbs.service.util.TemporalAmountParser).parse('${dk.recovery.validator.suspension-period-after-recovery}')}")
    private TemporalAmount suspensionPeriodAfterRecovery;


    @Test
    void audit_logs_DKINFO_token() {
        final Vaccination vaccination = Vaccinations.completedSingleDoseWithAdditionalVaccinations(CPR, LocalDateTime.now().minusDays(20), LocalDateTime.now().minusDays(10));

        final LocalDateTime recoveryPcrSamplingDate = LocalDateTime.now().minus(suspensionPeriodAfterRecovery).plusDays(5); //such that positive test within last 5 days doesnt suspend recovery pass
        final LabResults labResults = LabResultsBuilder.cpr(CPR)
            .add(Result.negativePcr(LocalDateTime.now().minusDays(1)))
            .add(Result.positivePcr(LocalDateTime.now().minusDays(2)))
            .add(Result.positivePcr(recoveryPcrSamplingDate))
            .build();

        final AuditData mockAuditData = mock(AuditData.class);
        final Set<CoronaPass> passports = dkInfoPassportFactory.createPassportFrom(vaccination, labResults, PERSON_RECORD, mockAuditData, null);

        assertEquals(1, passports.size());
        final CoronaPass coronaPass = passports.iterator().next();
        final String token = coronaPass.getToken();

        verify(mockAuditData).withSensitivePersonalInformation(eq("CWT_DKINFO"), argThat((List<String> arg) -> arg.contains(token)));
    }

    @ParameterizedTest
    @ValueSource(booleans = {true, false})
    void sets_vaccination_expiration_xdu_property(boolean under18) {
        final Vaccination vaccination = Vaccinations.completedMultiDoseWithAdditionalVaccinations(under18 ? CPR_UNDER_18 : CPR, LocalDateTime.now(), 2);

        final Set<CoronaPass> passports = dkInfoPassportFactory.createPassportFrom(vaccination, Results.EMPTY, under18 ? PERSON_RECORD_UNDER_18 : PERSON_RECORD, new AuditData(), null);
        assertEquals(1, passports.size());
        final CoronaPass coronaPass = passports.iterator().next();
        final Cwt cwt = CwtTestUtils.getCwt(coronaPass.getToken());
        final CBORObject v = cwt.getClaim(-260).get(1).get("v");
        assertNotNull(v);
        assertEquals(4, v.size());

        final EffectuatedVaccination[] effectuations = vaccination.vaccinations().toArray(new EffectuatedVaccination[0]);
        final EffectuatedVaccination[] revaccinations = vaccination.additionalVaccinations().toArray(new EffectuatedVaccination[0]);

        final Instant oneOfTwoExpiration = getExpiration(effectuations[1].effectuated(), maximumAmountAfterFirstVaccinationOfPartiallyCompletedVaccinationSchedule);
        final Instant twoOfTwoExpiration = getExpiration(effectuations[0].effectuated(), under18 ? maximumAmountAfterCompletedVaccinationScheduleUnder18 : maximumAmountAfterCompletedVaccinationSchedule);
        final Instant firstRevaccinationExpiration = getExpiration(revaccinations[1].effectuated(), maximumPeriodAfterRevaccination);
        final Instant secondRevaccinationExpiration = getExpiration(revaccinations[0].effectuated(), maximumPeriodAfterRevaccination);

        assertEquals(oneOfTwoExpiration, Instant.parse(v.get(0).get("xdu").AsString()));
        assertEquals(twoOfTwoExpiration, Instant.parse(v.get(1).get("xdu").AsString()));
        assertEquals(firstRevaccinationExpiration, Instant.parse(v.get(2).get("xdu").AsString()));
        assertEquals(secondRevaccinationExpiration, Instant.parse(v.get(3).get("xdu").AsString()));
    }

    private Instant getExpiration(LocalDateTime effectuated, TemporalAmount validityAmount) {
        final LocalDateTime localExpirationDateTime = effectuated.toLocalDate().atStartOfDay().plus(validityAmount);
        return localExpirationDateTime.toInstant(ZoneId.systemDefault().getRules().getOffset(localExpirationDateTime));
    }

    @Test
    void test_full_DKINFO_token() {
        final Vaccination vaccination = Vaccinations.completedMultiDoseWithAdditionalVaccinations(CPR, LocalDateTime.now().minusDays(20), 1);

        final LocalDateTime recoveryPcrSamplingDate = LocalDateTime.now().minus(suspensionPeriodAfterRecovery).plusDays(5); //such that positive test within last 5 days doesnt suspend recovery pass
        final LabResults labResults = LabResultsBuilder.cpr(CPR)
            .add(Result.negativePcr(LocalDateTime.now().minusDays(1)))
            .add(Result.positivePcr(LocalDateTime.now().minusDays(2)))
            .add(Result.positivePcr(recoveryPcrSamplingDate))
            .build();

        final Set<CoronaPass> passports = dkInfoPassportFactory.createPassportFrom(vaccination, labResults, PERSON_RECORD, new AuditData(), null);

        assertEquals(1, passports.size());
        final CoronaPass coronaPass = passports.iterator().next();

        final String token = coronaPass.getToken();
        final Cwt cwt = CwtTestUtils.getCwt(token);

        final CBORObject t = cwt.getClaim(-260).get(1).get("t");
        final CBORObject v = cwt.getClaim(-260).get(1).get("v");
        final CBORObject r = cwt.getClaim(-260).get(1).get("r");
        assertNotNull(t);
        assertNotNull(v);
        assertNotNull(r);
        assertEquals(1, t.size());
        assertEquals(3, v.size());
        assertEquals(1, r.size());

        final String recoveryValidUntil = r.get(0).get("du").AsString();
        assertEquals(recoveryPcrSamplingDate.plus(maximumPeriodAfterRecovery).toLocalDate().toString(), recoveryValidUntil);

        assertEquals(PERSON_RECORD.familyName(), cwt.getClaim(-260).get(1).get("nam").get("fn").AsString());
        assertEquals(PERSON_RECORD.givenName(), cwt.getClaim(-260).get(1).get("nam").get("gn").AsString());
        assertEquals(PERSON_RECORD.givenName(), cwt.getClaim(-260).get(1).get("nam").get("gn").AsString());
        assertEquals(DOB, cwt.getClaim(-260).get(1).get("dob").AsString());
    }

    @Test
    void test_no_recovery() {
        final Vaccination vaccination = Vaccinations.completedSingleDose(CPR, LocalDateTime.now().minusDays(20));

        final LocalDateTime recoveryPcrSamplingDate = LocalDateTime.now().minus(suspensionPeriodAfterRecovery).minusDays(5); //such that positive test within last 5 days suspends recovery pass
        final LabResults labResults = LabResultsBuilder.cpr(CPR)
            .add(Result.negativePcr(LocalDateTime.now().minusDays(1)))
            .add(Result.positivePcr(LocalDateTime.now().minusDays(2)))
            .add(Result.positivePcr(recoveryPcrSamplingDate))
            .build();

        final Set<CoronaPass> passports = dkInfoPassportFactory.createPassportFrom(vaccination, labResults, PERSON_RECORD, new AuditData(), null);

        assertEquals(1, passports.size());
        final CoronaPass coronaPass = passports.iterator().next();

        final String token = coronaPass.getToken();
        final Cwt cwt = CwtTestUtils.getCwt(token);

        final CBORObject t = cwt.getClaim(-260).get(1).get("t");
        final CBORObject v = cwt.getClaim(-260).get(1).get("v");
        final CBORObject r = cwt.getClaim(-260).get(1).get("r");
        assertNotNull(t);
        assertNotNull(v);
        assertNull(r);
        assertEquals(1, t.size());
        assertEquals(1, v.size());
    }


    @Test
    void contains_name_even_when_no_data() {
        final Vaccination noVaccinations = VaccinationBuilder.vaccination(CPR, false).build();
        final LabResults noLabResults = LabResultsBuilder.cpr(CPR).build();
        final Set<CoronaPass> passports = dkInfoPassportFactory.createPassportFrom(noVaccinations, noLabResults, PERSON_RECORD, new AuditData(), null);

        assertEquals(1, passports.size());
        final CoronaPass coronaPass = passports.iterator().next();

        final String token = coronaPass.getToken();
        final Cwt cwt = CwtTestUtils.getCwt(token);

        final CBORObject t = cwt.getClaim(-260).get(1).get("t");
        final CBORObject v = cwt.getClaim(-260).get(1).get("v");
        final CBORObject r = cwt.getClaim(-260).get(1).get("r");
        assertNull(t);
        assertNull(v);
        assertNull(r);

        assertEquals(PERSON_RECORD.familyName(), cwt.getClaim(-260).get(1).get("nam").get("fn").AsString());
        assertEquals(PERSON_RECORD.givenName(), cwt.getClaim(-260).get(1).get("nam").get("gn").AsString());
        assertEquals(PERSON_RECORD.givenName(), cwt.getClaim(-260).get(1).get("nam").get("gn").AsString());
        assertEquals(DOB, cwt.getClaim(-260).get(1).get("dob").AsString());
    }
}
