/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.passports;

import com.google.common.cache.LoadingCache;
import dk.sundhedsdatastyrelsen.cbs.dao.EUAntigenSpec;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.EUAntigenSpecImpl;
import dk.sundhedsdatastyrelsen.cbs.datacache.CollectionReader;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultsBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.service.config.*;
import dk.sundhedsdatastyrelsen.cbs.service.model.CoronaPass;
import dk.sundhedsdatastyrelsen.cbs.service.util.EUAntigenFiltration;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import org.junit.jupiter.api.Test;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Primary;
import org.springframework.context.annotation.Profile;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.junit.jupiter.SpringJUnitConfig;

import java.time.LocalDateTime;
import java.util.Collection;
import java.util.Collections;
import java.util.concurrent.ExecutionException;

import static java.time.LocalDateTime.now;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

@SpringJUnitConfig({CombinedPassportFactory.class, EUAntigenFiltrationValidFromValidToTest.EUAGFilterConfiguration.class, DkPassportFactory.class, EuPassportFactory.class, SignatureConfiguration.class, DkPassportConfiguration.class, EuPassportConfiguration.class, EUAntigenSpecConfiguration.class, VaccineMapperConfiguration.class})
@ActiveProfiles(SignatureConfiguration.LOCAL_KEY_STORE)
@CbsTest
class EUAntigenFiltrationValidFromValidToTest extends EUAntigenTest {

    // Master data only valid until 5 days ago. So 6 days old test provides pass, and 4 days old does not

    @Profile(SignatureConfiguration.LOCAL_KEY_STORE)
    @Configuration
    static class EUAGFilterConfiguration {

        @Bean("eufilter")
        @Primary
        public EUAntigenFiltration euAntigenFiltration() throws ExecutionException {

            CollectionReader<EUAntigenSpec, String> euAntigenSpecReader = mock(CollectionReader.class);
            when(euAntigenSpecReader.search(TestData.UNAPPROVED_MANUFACTURER)).thenReturn(Collections.emptyList());
            when(euAntigenSpecReader.search(TestData.APPROVED_MANUFACTURER)).thenReturn(Collections.singletonList(new EUAntigenSpecImpl().validFrom(now().minusDays(100)).validTo(now().plusDays(1))));

            LoadingCache<String, Collection<EUAntigenSpec>> euAntigenSpecCache = mock(LoadingCache.class);
            when(euAntigenSpecCache.get(TestData.UNAPPROVED_MANUFACTURER)).thenReturn(Collections.emptyList());
            when(euAntigenSpecCache.get(TestData.APPROVED_MANUFACTURER)).thenReturn(Collections.singletonList(new EUAntigenSpecImpl().validFrom(now().minusDays(100)).validTo(now().minusDays(5)))); // ValidTo = 5 days ago

            return new EUAntigenFiltration(euAntigenSpecCache, euAntigenSpecReader, true);
        }
    }

    @Test
    void ag_test_eu_valid_in_time() {
        final LabResults labResults = LabResultsBuilder.cpr(TestData.CPR)
            .add(LabResultBuilder.antigenResult(1, LocalDateTime.now().minusDays(6), LocalDateTime.now().minusDays(5)).negative().manufacturer(TestData.APPROVED_MANUFACTURER).build())
            .build();
        final Collection<CoronaPass> pass = getCoronaPasses(noVaccinations, labResults);
        assertEquals(1, getEUPass(pass, "lab").size()); // master data valid until 5 days ago, and this test is 6 days old - so ok
    }

    @Test
    void ag_test_eu_excluded_no_longer_valid() {
        final LabResults labResults = LabResultsBuilder.cpr(TestData.CPR)
            .add(LabResultBuilder.antigenResult(1, LocalDateTime.now().minusDays(4), LocalDateTime.now().minusDays(5)).negative().manufacturer(TestData.APPROVED_MANUFACTURER).build())
            .build();
        final Collection<CoronaPass> pass = getCoronaPasses(noVaccinations, labResults);
        assertEquals(0, getEUPass(pass, "lab").size()); // master data valid until 5 days ago, and this test is only 4 days old - so not ok
    }

}
