/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.passports;

import dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.time.LocalDateTime;
import java.time.Period;
import java.time.ZonedDateTime;
import java.time.temporal.TemporalAmount;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.Stream;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;

@CbsTest
class LabResultFilterTest {

    private ZonedDateTime now;

    @BeforeEach
    void setUp() {
        now = ZonedDateTime.now();
    }

    @Test
    void should_filter_results_older_than_max_age() {
        final List<LabResult> results = IntStream.range(1, 11)
            .unordered()
            .mapToObj(i -> labResult(i, Period.ofDays(i), false))
            .collect(Collectors.toList());

        final Stream<LabResult> labResultStream = LabResultFilter.filterLabResults(results, now.minusWeeks(1), 10, false);

        assertTrue(labResultStream.allMatch(labResult -> labResult.id() <= 7));
    }

    @Test
    void should_limit_results_when_limit_is_reached_discarding_oldest_first() {
        final List<LabResult> results = IntStream.range(1, 11)
            .unordered()
            .mapToObj(i -> labResult(i, Period.ofDays(i), false))
            .collect(Collectors.toList());

        final Stream<LabResult> labResultStream = LabResultFilter.filterLabResults(results, now.minusMonths(12), 5, false);

        assertTrue(labResultStream.allMatch(labResult -> labResult.id() <= 5));
    }

    @Test
    void should_include_positive_results_if_configured_to() {
        final List<LabResult> results = IntStream.range(1, 6)
            .unordered()
            .mapToObj(i -> labResult(i, Period.ofDays(1), i > 3))
            .collect(Collectors.toList());

        final Stream<LabResult> labResultStream = LabResultFilter.filterLabResults(results, now.minusMonths(12), 5, true);

        assertEquals(5, labResultStream.count());
    }

    @Test
    void should_not_include_positive_results_if_configured_not_to() {
        final List<LabResult> results = IntStream.range(1, 6)
            .unordered()
            .mapToObj(i -> labResult(i, Period.ofDays(1), i > 3))
            .collect(Collectors.toList());

        final Stream<LabResult> labResultStream = LabResultFilter.filterLabResults(results, now.minusMonths(12), 5, false);

        assertEquals(3, labResultStream.count());
    }

    private LabResult.AntigenResult labResult(int id, TemporalAmount age, boolean positive) {
        final ZonedDateTime time = now.minus(age);
        return LabResultBuilder.antigenResult(id, time.toLocalDateTime(), time.toLocalDateTime()).positive(positive).build();
    }
}
