/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.passports;

import dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultImpl;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult;
import dk.sundhedsdatastyrelsen.cbs.service.exceptions.LabResultMapperException;
import dk.sundhedsdatastyrelsen.cbs.test.RandomData;
import org.junit.jupiter.api.Test;
import se.digg.dgc.payload.v1.TestEntry;

import java.net.URI;
import java.time.LocalDateTime;
import java.time.ZoneId;

import static dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultBuilder.antigenResult;
import static dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultBuilder.pcrResult;
import static org.junit.jupiter.api.Assertions.*;


class LabResultMapperTest {

    private static final RandomData RANDOM_DATA = new RandomData("CBS".hashCode());

    @Test
    void testMapOfResults_unknownLabResultType() {
        LabResultImpl labResult = new LabResultImpl(
            0,
            false,
            "q",
            "manufacturer",
            "",
            "",
            LocalDateTime.now(),
            LocalDateTime.now()) {
        };
        Exception exception = assertThrows(LabResultMapperException.class, () -> LabResultMapper.toTestEntry(labResult));
        assertTrue(exception.getMessage().contains("Unknown LabResult.type: 'q' - labresult cannot be mapped"));
    }

    @Test
    void testMapOfLabResultPCR() {
        LocalDateTime samplingDateTime = LocalDateTime.now(ZoneId.systemDefault()).minusDays(1);
        LocalDateTime resultDateTime = LocalDateTime.now(ZoneId.systemDefault()).minusHours(1);

        LabResult labResult = pcrResult(
            RANDOM_DATA.nextLong(),
            samplingDateTime,
            resultDateTime)
            .positive(false)
            .center(RANDOM_DATA.string(10))
            .manufacturer(RANDOM_DATA.string(10))
            .name(RANDOM_DATA.string(10))
            .build();

        TestEntry tst = LabResultMapper.toTestEntry(labResult);
        assertEquals("DK", tst.getCo());
        assertEquals(LabResultMapper.LOINC_PCR, tst.getTt());
        assertEquals(VaccineMapper.SNOMED_CT_COVID_CONCEPT_ID, tst.getTg());
        assertEquals(labResult.center(), tst.getTc());
        assertEquals(labResult.manufacturer(), tst.getMa());
        assertEquals(labResult.name(), tst.getNm());

        assertEquals(LabResultMapper.SNOMED_CT_DETECTED_NEGATIVE, tst.getTr());
        assertEquals(samplingDateTime, LocalDateTime.ofInstant(tst.getSc(), ZoneId.systemDefault()));
        assertEquals("Danish Health Data Authority", tst.getIs());
        assertDoesNotThrow(() -> new URI(tst.getCi()));
    }

    @Test
    void testMapOfLabResultAntigen() {
        LocalDateTime samplingDateTime = LocalDateTime.now(ZoneId.systemDefault()).minusDays(1);
        LocalDateTime resultDateTime = LocalDateTime.now(ZoneId.systemDefault()).minusHours(1);

        LabResult labResult = antigenResult(
            RANDOM_DATA.nextLong(),
            samplingDateTime,
            resultDateTime)
            .positive(true)
            .center(RANDOM_DATA.string(10))
            .manufacturer(RANDOM_DATA.string(10))
            .name(RANDOM_DATA.string(10))
            .build();

        TestEntry tst = LabResultMapper.toTestEntry(labResult);
        assertEquals("DK", tst.getCo());
        assertEquals(LabResultMapper.LOINC_ANTIGEN, tst.getTt());
        assertEquals(VaccineMapper.SNOMED_CT_COVID_CONCEPT_ID, tst.getTg());
        assertEquals(labResult.center(), tst.getTc());
        assertEquals(labResult.manufacturer(), tst.getMa());
        assertEquals(labResult.name(), tst.getNm());

        assertEquals(LabResultMapper.SNOMED_CT_DETECTED_POSITIVE, tst.getTr());
        assertEquals(samplingDateTime, LocalDateTime.ofInstant(tst.getSc(), ZoneId.systemDefault()));
    }

}
