/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.passports;

import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.EffectuatedVaccination;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.service.exceptions.VaccineMapperException;
import dk.sundhedsdatastyrelsen.cbs.test.RandomData;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import se.digg.dgc.payload.v1.VaccinationEntry;

import java.net.URI;
import java.time.LocalDateTime;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import static dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.EffectuatedVaccinationBuilder.effectuatedVaccination;
import static dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.VaccinationBuilder.vaccination;
import static java.time.LocalDateTime.now;
import static java.util.Objects.requireNonNull;
import static org.junit.jupiter.api.Assertions.*;

class VaccineMapperTest {
    private static final RandomData RANDOM_DATA = new RandomData("CBS".hashCode());
    private VaccineMapper vaccineMapper;
    private VaccineMappingConfiguration vaccineMappingConfiguration;
    private final static long DRUG_ID = 1;
    private final static long DRUG_ID_2 = 2;

    @BeforeEach
    void beforeEach() {
        Map<Long, VaccineMappingConfiguration.VaccineInformation> mappingConfiguration = new HashMap<>();
        mappingConfiguration.put(DRUG_ID, new VaccineMappingConfiguration.VaccineInformation(UUID.randomUUID().toString(), UUID.randomUUID().toString(), UUID.randomUUID().toString()));
        mappingConfiguration.put(DRUG_ID_2, new VaccineMappingConfiguration.VaccineInformation(UUID.randomUUID().toString(), UUID.randomUUID().toString(), UUID.randomUUID().toString()));
        vaccineMappingConfiguration = new VaccineMappingConfiguration(mappingConfiguration);
        vaccineMapper = new VaccineMapper(vaccineMappingConfiguration);
    }

    @Test
    void testIncompleteConfiguration() {
        LocalDateTime effectuated = now();

        final EffectuatedVaccination effectuatedVaccination = effectuatedVaccination(300L, effectuated, "DK")
            .build();
        final Vaccination vaccination = vaccination(RANDOM_DATA.cpr(), true).add(effectuatedVaccination).build();

        VaccineMapperException exception = assertThrows(VaccineMapperException.class, () -> vaccineMapper.getAllVaccinationEntries(vaccination));
        assertTrue(exception.getMessage().contains("Unknown drugIdentifier: '300' - vaccine cannot be mapped"));
    }

    @Test
    void mapCompletedSingleVaccine() {
        LocalDateTime effectuated = now();

        final EffectuatedVaccination effectuatedVaccination = effectuatedVaccination(DRUG_ID, effectuated, "DK")
            .build();
        Vaccination vaccination = vaccination(RANDOM_DATA.cpr(), true).add(effectuatedVaccination).build();

        final List<VaccinationEntry> vaccinationEntries = vaccineMapper.getAllVaccinationEntries(vaccination);

        assertEquals(1, vaccinationEntries.size());

        verifyVaccinationEntryContains(vaccinationEntries.get(0), requireNonNull(effectuated), DRUG_ID, "DK", 1, 1);
    }

    @Test
    void mapCompletedSingleVaccineWithAdditionalVaccinations() {
        final LocalDateTime completedDate = now().minusMonths(6);
        final LocalDateTime additionalEffectuationDate1 = completedDate.plusMonths(2);
        final LocalDateTime additionalEffectuationDate2 = completedDate.plusMonths(4);

        Vaccination vaccination = vaccination(RANDOM_DATA.cpr(), true)
            .add(effectuatedVaccination(DRUG_ID, completedDate, "DK").build())
            .addAdditional(effectuatedVaccination(DRUG_ID_2, additionalEffectuationDate1, "DK").build())
            .addAdditional(effectuatedVaccination(DRUG_ID, additionalEffectuationDate2, "GL").build())
            .build();

        final List<VaccinationEntry> vaccinationEntries = vaccineMapper.getAllVaccinationEntries(vaccination);

        assertEquals(3, vaccinationEntries.size());
        verifyVaccinationEntryContains(vaccinationEntries.get(0), completedDate, DRUG_ID, "DK", 1, 1);
        verifyVaccinationEntryContains(vaccinationEntries.get(1), additionalEffectuationDate1, DRUG_ID_2, "DK", 2, 1);
        verifyVaccinationEntryContains(vaccinationEntries.get(2), additionalEffectuationDate2, DRUG_ID, "GL", 3, 1);

    }

    @Test
    void mapCompletedVaccineWithTwoEffectuations() {
        LocalDateTime effectuated = now();
        final LocalDateTime firstEffectuation = effectuated.minusDays(20);

        Vaccination vaccination = vaccination(RANDOM_DATA.cpr(), true)
            .add(effectuatedVaccination(DRUG_ID, firstEffectuation, "GL")
                .build())
            .add(effectuatedVaccination(DRUG_ID, effectuated, "DK")
                .build()).build();

        final List<VaccinationEntry> vaccinationEntries = vaccineMapper.getAllVaccinationEntries(vaccination);

        assertEquals(2, vaccinationEntries.size());
        verifyVaccinationEntryContains(vaccinationEntries.get(0), firstEffectuation, DRUG_ID, "GL", 1, 2);
        verifyVaccinationEntryContains(vaccinationEntries.get(1), effectuated, DRUG_ID, "DK", 2, 2);
    }

    @Test
    void mapCompletedVaccineWithTwoEffectuationsAndAdditionalVaccination() {
        LocalDateTime secondEffectuated = now();
        final LocalDateTime firstEffectuation = secondEffectuated.minusDays(20);
        final LocalDateTime additionalEffectuationDate1 = secondEffectuated.plusDays(20);
        final LocalDateTime additionalEffectuationDate2 = secondEffectuated.plusDays(40);

        Vaccination vaccination = vaccination(RANDOM_DATA.cpr(), true)
            .add(effectuatedVaccination(DRUG_ID, firstEffectuation, "GL").build())
            .add(effectuatedVaccination(DRUG_ID, secondEffectuated, "DK").build())
            .addAdditional(effectuatedVaccination(DRUG_ID_2, additionalEffectuationDate1, "FO").build())
            .addAdditional(effectuatedVaccination(DRUG_ID, additionalEffectuationDate2, "DK").build())
            .build();

        final List<VaccinationEntry> vaccinationEntries = vaccineMapper.getAllVaccinationEntries(vaccination);

        assertEquals(4, vaccinationEntries.size());
        verifyVaccinationEntryContains(vaccinationEntries.get(0), firstEffectuation, DRUG_ID, "GL", 1, 2);
        verifyVaccinationEntryContains(vaccinationEntries.get(1), secondEffectuated, DRUG_ID, "DK", 2, 2);
        verifyVaccinationEntryContains(vaccinationEntries.get(2), additionalEffectuationDate1, DRUG_ID_2, "FO", 3, 3);
        verifyVaccinationEntryContains(vaccinationEntries.get(3), additionalEffectuationDate2, DRUG_ID, "DK", 4, 4);
    }

    @Test
    void mapIncompleteVaccineWithOneEffectuatedVaccine() {
        LocalDateTime effectuated = now();

        final EffectuatedVaccination effectuatedVaccination = effectuatedVaccination(DRUG_ID, effectuated, "GL")
            .build();
        Vaccination vaccination = vaccination(RANDOM_DATA.cpr(), false).add(effectuatedVaccination).build();

        final List<VaccinationEntry> vaccinationEntries = vaccineMapper.getAllVaccinationEntries(vaccination);

        assertEquals(1, vaccinationEntries.size());
        verifyVaccinationEntryContains(vaccinationEntries.get(0), requireNonNull(vaccination.mostRecentlyEffectuated()).effectuated(), DRUG_ID, "GL", 1, 2);
    }

    @Test
    @DisplayName("CR-242: Effektueringer fra 'Delvist vaccineret' DDV-svar skal alle have nummerangivelsen '1/2' i EU vac CWTs")
    void mapIncompleteVaccination() {
        final LocalDateTime firstDt = now().minusDays(30);
        final LocalDateTime secondDt = now().minusDays(29);
        final LocalDateTime thirdDt = now().minusDays(10);

        Vaccination vaccination = vaccination(RANDOM_DATA.cpr(), false)
            .add(effectuatedVaccination(DRUG_ID, firstDt, "GL").build())
            .add(effectuatedVaccination(DRUG_ID, secondDt, "DK").build())
            .add(effectuatedVaccination(DRUG_ID_2, thirdDt, "FO").build())
            .build();

        final VaccineMappingConfiguration.VaccineInformation vaccine1 = vaccineMappingConfiguration.getVaccineIdValues(DRUG_ID);
        final VaccineMappingConfiguration.VaccineInformation vaccine2 = vaccineMappingConfiguration.getVaccineIdValues(DRUG_ID_2);

        final List<VaccinationEntry> vaccinationEntries = vaccineMapper.getAllVaccinationEntries(vaccination);

        assertEquals(3, vaccinationEntries.size());
        verifyVaccinationEntryContains(vaccinationEntries.get(0), firstDt, DRUG_ID, "GL", 1, 2);
        verifyVaccinationEntryContains(vaccinationEntries.get(1), secondDt, DRUG_ID, "DK", 1, 2);
        verifyVaccinationEntryContains(vaccinationEntries.get(2), thirdDt, DRUG_ID_2, "FO", 1, 2);
    }

    private void verifyVaccinationEntryContains(VaccinationEntry vaccinationEntry,
                                                LocalDateTime expectedEffectuationTime,
                                                long expectedDrugId,
                                                String expectedCountryCode,
                                                int expectedDoseNumber,
                                                int expectedDosesInSeries) {
        assertEquals(expectedEffectuationTime.toLocalDate(), vaccinationEntry.getDt());
        assertEquals(vaccineMappingConfiguration.getVaccineIdValues(expectedDrugId).ma, vaccinationEntry.getMa());
        assertEquals(vaccineMappingConfiguration.getVaccineIdValues(expectedDrugId).vp, vaccinationEntry.getVp());
        assertEquals(vaccineMappingConfiguration.getVaccineIdValues(expectedDrugId).mp, vaccinationEntry.getMp());
        assertEquals(expectedCountryCode, vaccinationEntry.getCo());
        assertEquals(VaccineMapper.SNOMED_CT_COVID_CONCEPT_ID, vaccinationEntry.getTg());
        assertEquals(expectedDoseNumber, vaccinationEntry.getDn());
        assertEquals(expectedDosesInSeries, vaccinationEntry.getSd());
        assertEquals("Danish Health Data Authority", vaccinationEntry.getIs());
        assertDoesNotThrow(() -> new URI(vaccinationEntry.getCi()));
    }

}
