/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.security;

import com.google.common.collect.ImmutableSet;
import dk.sds.nsp.security.SecurityContext;
import dk.sundhedsdatastyrelsen.cbs.dao.ChildRelation;
import dk.sundhedsdatastyrelsen.cbs.dao.ServiceClient;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.ChildRelationImpl;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.ServiceClientImpl;
import dk.sundhedsdatastyrelsen.cbs.datacache.CollectionReader;
import dk.sundhedsdatastyrelsen.cbs.service.ErrorMessages;
import dk.sundhedsdatastyrelsen.cbs.service.ServiceConstants;
import dk.sundhedsdatastyrelsen.cbs.service.exceptions.SecurityRuntimeException;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;

import java.util.*;

import static dk.sds.nsp.security.SecurityContext.User.UserType.Citizen;
import static dk.sds.nsp.security.SecurityContext.User.UserType.HealthcareProfessional;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

@CbsTest
class WsSecurityTest {
    public static final String SERVICE_IDENTIFIER = "CoronaPassWebService";
    @Mock
    private SecurityContext.Ticket ticket;
    @Mock
    private SecurityContext.User user;
    @Mock
    private SecurityContext securityContext;

    @Mock
    private SecurityContext.Client client;

    @Mock
    private SecurityContext.Organisation organisation;

    @Mock
    private CollectionReader<ServiceClient, String> serviceClientReader;

    private Map<String, Collection<ChildRelation>> childRelations;
    private WsSecurity wsSecurity;

    public static final String CPR = "0506504003";
    private final String persistedUniqueKey = UUID.randomUUID().toString();
    private final String powerOfAttorneyPrivilegeName = "my-cbs-power-of-attorney-privilege-name";

    @BeforeEach
    void setUp() {
        childRelations = new HashMap<>();
        wsSecurity = new WsSecurity(cpr -> childRelations.getOrDefault(cpr, Collections.emptyList()), serviceClientReader, powerOfAttorneyPrivilegeName);
    }

    @Nested
    @DisplayName("IDWS")
    class IdwsSpecificTests {

        @BeforeEach
        void setUp() {
            lenient().when(ticket.isValid()).thenReturn(true);
            lenient().when(ticket.getAudience()).thenReturn(Optional.of(ServiceConstants.AUDIENCE));
            when(securityContext.getTicket()).thenReturn(Optional.of(ticket));
            lenient().when(securityContext.getActingUser()).thenReturn(Optional.of(user));
            lenient().when(user.getIdentifier()).thenReturn(CPR);
            lenient().when(user.getUserType()).thenReturn(Citizen);
        }

        @Test
        void throws_exception_when_no_ticket() {
            when(securityContext.getTicket()).thenReturn(Optional.empty());

            SecurityRuntimeException exception = assertThrows(SecurityRuntimeException.class, () -> wsSecurity.checkSecurityTicket(securityContext, CPR, SERVICE_IDENTIFIER));
            assertEquals(ErrorMessages.NO_SECURITY_TICKET, exception.getMessage());
        }

        @Test
        void throws_exception_on_invalid_ticket() {

            when(ticket.isValid()).thenReturn(false);

            SecurityRuntimeException exception = assertThrows(SecurityRuntimeException.class, () -> wsSecurity.checkSecurityTicket(securityContext, CPR, SERVICE_IDENTIFIER));
            assertEquals(ErrorMessages.INVALID_SECURITY_TICKET, exception.getMessage());
        }

        @Test
        void throws_exception_on_missing_audience() {

            when(ticket.getAudience()).thenReturn(Optional.empty());

            SecurityRuntimeException exception = assertThrows(SecurityRuntimeException.class, () -> wsSecurity.checkSecurityTicket(securityContext, CPR, SERVICE_IDENTIFIER));
            assertEquals(ErrorMessages.MISSING_AUDIENCE, exception.getMessage());
        }

        @Test
        void throws_exception_on_invalid_audience() {

            when(ticket.getAudience()).thenReturn(Optional.of("jamaica"));

            SecurityRuntimeException exception = assertThrows(SecurityRuntimeException.class, () -> wsSecurity.checkSecurityTicket(securityContext, CPR, SERVICE_IDENTIFIER));
            assertEquals(ErrorMessages.INVALID_AUDIENCE, exception.getMessage());
        }

        @Test
        void throws_exception_on_invalid_user_type() {

            when(user.getUserType()).thenReturn(HealthcareProfessional);

            SecurityRuntimeException exception = assertThrows(SecurityRuntimeException.class, () -> wsSecurity.checkSecurityTicket(securityContext, CPR, SERVICE_IDENTIFIER));
            assertEquals(ErrorMessages.USER_TYPE_NOT_ALLOWED, exception.getMessage());
        }


        @Test
        void throws_exception_on_null_user_type() {

            when(user.getUserType()).thenReturn(null);

            SecurityRuntimeException exception = assertThrows(SecurityRuntimeException.class, () -> wsSecurity.checkSecurityTicket(securityContext, CPR, SERVICE_IDENTIFIER));
            assertEquals(ErrorMessages.USER_TYPE_NOT_ALLOWED, exception.getMessage());
        }

        @Test
        void throws_exception_on_cpr_mismatch() {
            when(user.getIdentifier()).thenReturn("0101012222");

            SecurityRuntimeException exception = assertThrows(SecurityRuntimeException.class, () -> wsSecurity.checkSecurityTicket(securityContext, CPR, SERVICE_IDENTIFIER));
            assertEquals(ErrorMessages.NOT_AUTHORIZED_CPR_MISMATCH, exception.getMessage());
        }

        @Test
        void should_pass_if_cpr_matches() {
            assertDoesNotThrow(() -> wsSecurity.checkSecurityTicket(securityContext, CPR, SERVICE_IDENTIFIER));
        }

        @Test
        void should_pass_if_idws_user_has_power_of_attorney() {
            String principalUserCpr = "0101012233";

            //before power of attorney
            assertThrows(SecurityRuntimeException.class, () -> wsSecurity.checkSecurityTicket(securityContext, principalUserCpr, SERVICE_IDENTIFIER));

            //add power of attorney
            SecurityContext.User.Credentials userCredentials = mock(SecurityContext.User.Credentials.class);
            when(userCredentials.getPowerOfAttorneyPrivileges()).thenReturn(ImmutableSet.of("foo", "bar", powerOfAttorneyPrivilegeName));
            when(user.getCredentials()).thenReturn(Optional.of(userCredentials));
            SecurityContext.User principalUser = mock(SecurityContext.User.class);
            when(principalUser.getIdentifier()).thenReturn(principalUserCpr);
            when(securityContext.getPrincipalUser()).thenReturn(Optional.of(principalUser));

            assertDoesNotThrow(() -> wsSecurity.checkSecurityTicket(securityContext, principalUserCpr, SERVICE_IDENTIFIER));
        }

        @Test
        void should_pass_if_idws_user_has_child_matching_cpr() {
            String childCpr = "0101212223";
            childRelations.put(CPR, Collections.singletonList(new ChildRelationImpl().personIdentifier(CPR).childIdentifier(childCpr)));

            assertDoesNotThrow(() -> wsSecurity.checkSecurityTicket(securityContext, childCpr, SERVICE_IDENTIFIER));
        }
    }

    @Nested
    @DisplayName("DGWS")
    class DgwsSpecificTest {
        @BeforeEach
        void setUp() {
            when(securityContext.getTicket()).thenReturn(Optional.of(ticket));
            lenient().when(securityContext.getClient()).thenReturn(Optional.of(client));
            lenient().when(securityContext.getOrganisation()).thenReturn(Optional.of(organisation));

            lenient().when(ticket.isValid()).thenReturn(true);
            lenient().when(client.getPersistentUniqueKey()).thenReturn(Optional.of(() -> persistedUniqueKey));
            lenient().when(serviceClientReader.search(persistedUniqueKey)).thenReturn(Collections.singletonList(new ServiceClientImpl().persistentUniqueKey(persistedUniqueKey).serviceIdentifier(SERVICE_IDENTIFIER)));
        }

        @Test
        void should_pass_if_cpr_matches() {
            assertDoesNotThrow(() -> wsSecurity.checkSecurityTicket(securityContext, CPR, SERVICE_IDENTIFIER));
        }

        @Test
        void should_reject_if_ticket_is_missing() {
            when(securityContext.getTicket()).thenReturn(Optional.empty());

            SecurityRuntimeException exception = assertThrows(SecurityRuntimeException.class, () -> wsSecurity.checkSecurityTicket(securityContext, CPR, "service"));

            assertEquals(ErrorMessages.NO_SECURITY_TICKET, exception.getMessage());

        }

        @Test
        void should_reject_if_ticket_is_not_valid() {
            when(ticket.isValid()).thenReturn(false);

            SecurityRuntimeException exception = assertThrows(SecurityRuntimeException.class, () -> wsSecurity.checkSecurityTicket(securityContext, CPR, "service"));

            assertEquals(ErrorMessages.INVALID_SECURITY_TICKET, exception.getMessage());
        }

        @Test
        void should_reject_if_client_is_missing() {
            when(securityContext.getClient()).thenReturn(Optional.empty());

            SecurityRuntimeException exception = assertThrows(SecurityRuntimeException.class, () -> wsSecurity.checkSecurityTicket(securityContext, CPR, "service"));

            assertEquals(ErrorMessages.CLIENT_NOT_PRESENT, exception.getMessage());
        }

        @Test
        void should_reject_if_client_unique_key_is_missing() {
            when(client.getPersistentUniqueKey()).thenReturn(Optional.empty());

            SecurityRuntimeException exception = assertThrows(SecurityRuntimeException.class, () -> wsSecurity.checkSecurityTicket(securityContext, CPR, "service"));

            assertEquals(ErrorMessages.PERSISTENT_KEY_NOT_PRESENT, exception.getMessage());
        }

        @Test
        void should_reject_if_organization_is_not_present() {
            when(securityContext.getOrganisation()).thenReturn(Optional.empty());

            final SecurityRuntimeException exception = assertThrows(SecurityRuntimeException.class, () -> wsSecurity.checkSecurityTicket(securityContext, CPR, "service"));

            assertEquals(ErrorMessages.ORGANIZATION_NOT_PRESENT, exception.getMessage());
        }

        @Test
        void should_reject_if_client_unique_key_is_not_whitelisted() {
            when(client.getPersistentUniqueKey()).thenReturn(Optional.of(() -> UUID.randomUUID().toString()));

            SecurityRuntimeException exception = assertThrows(SecurityRuntimeException.class, () -> wsSecurity.checkSecurityTicket(securityContext, CPR, "service"));

            assertEquals(ErrorMessages.CLIENT_NOT_WHITELISTED, exception.getMessage());

        }

        @Test
        void should_pass_if_client_unique_key_is_whitelisted_for_service() {
            String persistedUniqueKey = UUID.randomUUID().toString();
            String serviceIdentifier = "service";

            when(client.getPersistentUniqueKey()).thenReturn(Optional.of(() -> persistedUniqueKey));

            when(serviceClientReader.search(persistedUniqueKey)).thenReturn(Collections.singletonList(new ServiceClientImpl().persistentUniqueKey(persistedUniqueKey).serviceIdentifier(serviceIdentifier)));

            assertDoesNotThrow(() -> wsSecurity.checkSecurityTicket(securityContext, CPR, serviceIdentifier));
        }

        @Test
        void should_reject_if_client_unique_key_is_whitelisted_for_another_service() {
            String persistedUniqueKey = UUID.randomUUID().toString();

            when(client.getPersistentUniqueKey()).thenReturn(Optional.of(() -> persistedUniqueKey));

            when(serviceClientReader.search(persistedUniqueKey)).thenReturn(Collections.singletonList(new ServiceClientImpl().persistentUniqueKey(persistedUniqueKey).serviceIdentifier("otherService")));

            SecurityRuntimeException exception = assertThrows(SecurityRuntimeException.class, () -> wsSecurity.checkSecurityTicket(securityContext, CPR, "service"));

            assertEquals(ErrorMessages.CLIENT_NOT_WHITELISTED, exception.getMessage());

        }

        @Test
        void should_pass_if_client_unique_key_is_whitelisted_for_service_among_many_services() {
            String persistedUniqueKey = UUID.randomUUID().toString();
            String serviceIdentifier = "service1";

            when(client.getPersistentUniqueKey()).thenReturn(Optional.of(() -> persistedUniqueKey));

            final List<ServiceClient> serviceClients = Arrays.asList(
                new ServiceClientImpl().persistentUniqueKey(persistedUniqueKey).serviceIdentifier(serviceIdentifier),
                new ServiceClientImpl().persistentUniqueKey(persistedUniqueKey).serviceIdentifier("service2"),
                new ServiceClientImpl().persistentUniqueKey(persistedUniqueKey).serviceIdentifier("service3"),
                new ServiceClientImpl().persistentUniqueKey(persistedUniqueKey).serviceIdentifier("service4"),
                new ServiceClientImpl().persistentUniqueKey(persistedUniqueKey).serviceIdentifier("service5")
            );
            when(serviceClientReader.search(persistedUniqueKey)).thenReturn(serviceClients);

            assertDoesNotThrow(() -> wsSecurity.checkSecurityTicket(securityContext, CPR, serviceIdentifier));
        }
    }


}
