/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.util;

import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;

import java.time.Duration;
import java.time.LocalDateTime;
import java.time.Month;
import java.time.temporal.TemporalAmount;
import java.util.stream.Stream;

import static org.junit.jupiter.api.Assertions.*;

class CprFormatBasedAgeCalculatorImplTest {

    private final CprFormatBasedAgeCalculatorImpl ageCalculator = new CprFormatBasedAgeCalculatorImpl();

    @ParameterizedTest
    @MethodSource("isUnder18Arguments")
    void testIsUnder18(String cpr, LocalDateTime now, boolean expected) {
        assertEquals(expected, ageCalculator.isUnder18(cpr, now));
    }

    private static Stream<Arguments> isUnder18Arguments() {
        final LocalDateTime now = LocalDateTime.of(2021, Month.DECEMBER, 9, 16, 14);
        return Stream.of(
            Arguments.of("0101011234", now, false),
            Arguments.of("0912034000", now, false), // 18th birthday today
            Arguments.of("1012034000", now, true),  // 18th birthday tomorrow
            Arguments.of("1012033000", now, false)  // born 1903 (118th birthday tomorrow)
        );
    }

    @ParameterizedTest
    @MethodSource("timeUntil18Arguments")
    void testTimeUntil18(String cpr, LocalDateTime now, TemporalAmount expected) {

        final TemporalAmount timeUntil18 = ageCalculator.timeUntil18(cpr, now);

        assertEquals(expected, timeUntil18);
    }

    private static Stream<Arguments> timeUntil18Arguments() {
        final LocalDateTime now = LocalDateTime.of(2021, Month.DECEMBER, 9, 16, 14); // 7 hours and 46 minutes before midnight
        return Stream.of(
            Arguments.of("0101011234", now, Duration.ZERO),
            Arguments.of("0912034000", now, Duration.ZERO), // 18th birthday today
            Arguments.of("1012033000", now, Duration.ZERO),  // born 1903 (118th birthday tomorrow)

            // the following durations are calculated as: duration = 24H * [days until 18th birthday] + [time until midnight]
            Arguments.of("1012034000", now, Duration.ofHours(7).plusMinutes(46)),    // 18th birthday tomorrow   - 24H *  0 + 7H + 46M =   7H + 46M
            Arguments.of("0401047166", now, Duration.ofHours(607).plusMinutes(46)),  // 18th birthday in 25 days - 24H * 25 + 7H + 46M = 607H + 46M
            Arguments.of("0501048577", now, Duration.ofHours(631).plusMinutes(46)),  // 18th birthday in 26 days - 24H * 26 + 7H + 46M = 631H + 46M
            Arguments.of("0601048175", now, Duration.ofHours(655).plusMinutes(46))   // 18th birthday in 27 days - 24H * 27 + 7H + 46M = 655H + 46M
        );
    }

    @Test
    void testInvalidCprIsUnder18returnsFalse() {
        final LocalDateTime now = LocalDateTime.of(2021, Month.DECEMBER, 9, 16, 14);
        assertFalse(ageCalculator.isUnder18("9999999999", now));
    }

    @Test
    void testInvalidCprTimeUntil18IsZero() {
        final LocalDateTime now = LocalDateTime.of(2021, Month.DECEMBER, 9, 16, 14);
        assertEquals(Duration.ZERO, ageCalculator.timeUntil18("9999999999", now));
    }
}
