/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.webservice;

import com.google.common.truth.Truth;
import dk.sds.nsp.security.DevelopmentSecurityProvider;
import dk.sds.nsp.security.SecurityContext.Ticket;
import dk.sds.nsp.security.SecurityProvider;
import dk.sundhedsdatastyrelsen.cbs.service.CoronaPassService;
import dk.sundhedsdatastyrelsen.cbs.service.model.CoronaPass;
import dk.sundhedsdatastyrelsen.cbs.service.model.CoronaPassResult;
import dk.sundhedsdatastyrelsen.cbs.service.model.PassType;
import dk.sundhedsdatastyrelsen.cbs.service.security.WsSecurity;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassRequest;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassResponse;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mockito;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Primary;
import org.springframework.lang.NonNull;
import org.springframework.test.context.junit.jupiter.SpringJUnitConfig;

import java.util.Collections;
import java.util.UUID;

import static dk.sds.nsp.security.SecurityContext.User;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.mockito.Mockito.*;

@CbsTest
@SpringJUnitConfig({CoronaPassWebServiceTest.MockConfig.class})
class CoronaPassWebServiceTest {
    private static final User user = mock(User.class);
    @Autowired
    private ApplicationContext testContext;
    @Autowired
    private WsSecurity mockWsSecurity;
    @Autowired
    private CoronaPassService mockCoronaPassService;

    private final CoronaPassWebService coronaPassWebService = new CoronaPassWebService();

    @BeforeAll
    static void beforeAll() {
        final DevelopmentSecurityProvider securityProvider = (DevelopmentSecurityProvider) SecurityProvider.provider();
        securityProvider.setGlobalTicket(when(mock(Ticket.class).isValid()).thenReturn(true).getMock());
        securityProvider.setGlobalActingUser(user);
    }

    public static class MockConfig {
        @Bean
        @Primary
        public WsSecurity mockSecurity() {
            return mock(WsSecurity.class);
        }

        @Primary
        @Bean
        public CoronaPassService service() {
            CoronaPassService mockCoronaPassService = mock(CoronaPassService.class);
            Mockito.lenient().when(mockCoronaPassService.createDataFetcherJob(anyString(), anyString(), any())).thenAnswer(i -> UUID.randomUUID().toString());
            return mockCoronaPassService;
        }

    }

    @BeforeEach
    void setUp() {
        coronaPassWebService.initializeSpringDependencies(testContext);
    }

    @NonNull
    private GetCoronaPassRequest request(String cpr) {
        GetCoronaPassRequest request = new GetCoronaPassRequest();
        request.setIdentifier(cpr);
        when(user.getIdentifier()).thenReturn(cpr);
        return request;
    }

    @Test
    void test_GetCoronaPass_cache_miss() {
        String jobId = UUID.randomUUID().toString();
        when(mockCoronaPassService.getCoronaPass(anyString(), anyString())).thenReturn(new CoronaPassResult(jobId));

        String cpr = "0101012233";
        GetCoronaPassRequest request = request(cpr);
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request);
        assertEquals(jobId, coronaPass.getFutureResult().getJobId());
        assertNull(coronaPass.getCoronaPasses());

        verify(mockWsSecurity).checkSecurityTicket(any(), eq(cpr), anyString());
        verify(mockCoronaPassService).getCoronaPass(eq(cpr), anyString());
    }

    @Test
    void test_GetCoronaPass_cache_hit() {
        String token = "DK1:NCFJ70S90-7WF.9.60+O4W79F%GZM9NTVO K*6B0XK1JCYL92F3*8JPUF2F3IVI7HBJ6BWW6J%6-96N-C8X57076L64S6ZX68X50EC9OC3UCK%6OF6SIC5:6EM6FBBZ10/ITMII2AH-H77LUN 35*T.*K:/AH HMV0V+2R J %LV+2BPAXQAJVRVSSG70GD0I5JMPK8FHS6HS873$SRC7+J6LND.ES+4T99B4DN$GQ:DAT-1S*4I9S085+N3. KU/83TEYDKKLEN8G*PJDWB-AV+.IWTNK0W$5M-85R*IC93H/BISN$/B1KNNMK611X.KQ29O:KZ43 6DD.RTL5PC3I9459GYAGV V%9TARA-32H6EUMS12OK*P/H1CE2O46FBN.F4ISEWTTH916JB-GCCU4+5VHV5VCQDFJQKH I50TAWTA2%F5QBV68:YD2HLLSL:A7.RO.9TFEPH%PL:V7PHGH3PX35ENFW5S0O/6S*N0EVGZRKJTLCZ7PBBHO2M5DQWO%0";
        when(mockCoronaPassService.getCoronaPass(anyString(), anyString())).thenReturn(new CoronaPassResult(Collections.singleton(new CoronaPass(PassType.DK, token))));

        String cpr = "0202023344";
        GetCoronaPassRequest request = request(cpr);
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request);
        assertNull(coronaPass.getFutureResult());
        Truth.assertThat(coronaPass.getCoronaPasses().getPass()).hasSize(1);
        assertEquals(token, coronaPass.getCoronaPasses().getPass().get(0).getJWT());

        verify(mockWsSecurity).checkSecurityTicket(any(), eq(cpr), anyString());
        verify(mockCoronaPassService).getCoronaPass(eq(cpr), anyString());
    }
}
