/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.test;

import com.google.common.collect.MoreCollectors;
import dk.sundhedsdatastyrelsen.cbs.service.exceptions.CwtException;
import dk.sundhedsdatastyrelsen.cbs.service.model.CoronaPass;
import dk.sundhedsdatastyrelsen.cbs.service.model.PassType;
import org.apache.commons.lang3.exception.ExceptionUtils;
import se.digg.dgc.encoding.Base45;
import se.digg.dgc.encoding.Zlib;
import se.digg.dgc.payload.v1.DigitalGreenCertificate;
import se.digg.dgc.service.impl.DefaultDGCDecoder;
import se.digg.dgc.signatures.CertificateProvider;
import se.digg.dgc.signatures.cose.CoseSign1_Object;
import se.digg.dgc.signatures.cwt.Cwt;
import se.digg.dgc.signatures.impl.DefaultDGCSignatureVerifier;

import java.security.KeyStore;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import static org.junit.jupiter.api.Assertions.assertTrue;

public class CwtTestUtils {
    public static final X509Certificate certificate;
    public static final DefaultDGCSignatureVerifier verifier;
    public static final CertificateProvider certificateProvider;

    static {
        try {
            KeyStore keystore = KeyStore.getInstance("PKCS12");
            keystore.load(CwtTestUtils.class.getClassLoader().getResourceAsStream("Sundhedsdatastyrelsen - COVID-19 Borgerstatus.p12"), "Test1234".toCharArray());
            certificate = (X509Certificate) keystore.getCertificate("covid-19 borgerstatus (funktionscertifikat)");
        } catch (Exception e) {
            throw new CwtException(e);
        }

        verifier = new DefaultDGCSignatureVerifier();
        certificateProvider = (country, kid) -> {
            List<X509Certificate> listOfCertificate = new ArrayList<>();
            listOfCertificate.add(certificate);
            return listOfCertificate;
        };
    }

    public static DigitalGreenCertificate getDigitalGreenCertificate(String token) {
        try {
            return new DefaultDGCDecoder(verifier, certificateProvider).decode(token);
        } catch (Exception e) {
            return ExceptionUtils.rethrow(e);
        }
    }


    public static Cwt getCwt(String token) {
        String data = token.substring(4);
        try {
            byte[] cborBytes = Zlib.decompress(Base45.getDecoder().decode(data), true);
            CoseSign1_Object coseObject = new CoseSign1_Object(cborBytes);
            coseObject.verifySignature(certificate.getPublicKey());
            return coseObject.getCwt();
        } catch (Exception e) {
            return ExceptionUtils.rethrow(e);
        }
    }

    public static Cwt getCwt(Collection<CoronaPass> coronaPass, PassType passType) {
        if (passType == PassType.DK) {
            CoronaPass dkPass = coronaPass.stream().filter(p -> p.getPassType() == PassType.DK).collect(MoreCollectors.onlyElement());
            assertTrue(dkPass.getToken().startsWith("DK1:"));
            return getCwt(dkPass.getToken());
        } else if (passType == PassType.DKPLUS) {
            CoronaPass dkPlusPass = coronaPass.stream().filter(p -> p.getPassType() == PassType.DKPLUS).collect(MoreCollectors.onlyElement());
            assertTrue(dkPlusPass.getToken().startsWith("DK2:"));
            return getCwt(dkPlusPass.getToken());
        } else if (passType == PassType.DKINFO) {
            CoronaPass dkInfoPass = coronaPass.stream().filter(p -> p.getPassType() == PassType.DKINFO).collect(MoreCollectors.onlyElement());
            assertTrue(dkInfoPass.getToken().startsWith("DK3:"));
            return getCwt(dkInfoPass.getToken());
        }
        throw new RuntimeException();
    }

}
