/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.test;

import org.opentest4j.TestAbortedException;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;
import java.time.*;
import java.time.format.DateTimeFormatter;
import java.util.*;
import java.util.concurrent.ThreadLocalRandom;
import java.util.function.Supplier;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static java.lang.String.format;
import static java.lang.String.valueOf;
import static java.nio.charset.StandardCharsets.UTF_8;
import static java.time.LocalDate.ofYearDay;
import static java.time.format.DateTimeFormatter.ofPattern;
import static java.util.Collections.*;
import static java.util.stream.Stream.empty;
import static java.util.stream.Stream.generate;
import static org.apache.commons.lang3.StringUtils.leftPad;

public class RandomData extends Random {

    private static final String EMPTY = "";
    private static final DateTimeFormatter CPR_FORMATTER = ofPattern("ddMMyy");

    public RandomData() {
        super("CBS".hashCode());
    }

    public RandomData(final long seed) {
        super(seed);
    }

    @NonNull
    public String cpr() {
        return ofYearDay(nextInt(100) + 1900, nextInt(365) + 1)
            .format(CPR_FORMATTER) + leftPad(valueOf(nextInt(10000)), 4, '0');
    }

    @NonNull
    public String uuidString() {
        return UUID.randomUUID().toString();
    }

    @NonNull
    public <T> Stream<T> stream(final int bound, @NonNull final Supplier<T> constructor) {
        return bound == 0 ? empty() : generate(constructor).limit(nextInt(bound));
    }

    @NonNull
    public <T> Stream<T> stream(int lowerBound, int upperBound, @NonNull final Supplier<T> constructor) {
        return upperBound == 0 ? empty() : generate(constructor).limit(nextInt(lowerBound, upperBound));
    }

    @NonNull
    public <T> Set<T> set(final int bound, @NonNull final Supplier<T> constructor) {
        return bound == 0 ? emptySet() : stream(1, bound, constructor).collect(Collectors.toSet());
    }

    @NonNull
    public <T> Set<T> set(int lowerBound, int upperBound, @NonNull final Supplier<T> constructor) {
        return upperBound == 0 ? emptySet() : stream(lowerBound, upperBound, constructor).collect(Collectors.toSet());
    }

    @NonNull
    public <T> List<T> list(final int bound, @NonNull final Supplier<T> constructor) {
        return bound == 0 ? emptyList() : stream(1, bound, constructor).collect(Collectors.toList());
    }

    @NonNull
    public <T> List<T> list(int lowerBound, int upperBound, @NonNull final Supplier<T> constructor) {
        return upperBound == 0 ? emptyList() : stream(lowerBound, upperBound, constructor).collect(Collectors.toList());
    }

    @NonNull
    public <K, V> Map<K, V> map(int bound, @NonNull final Supplier<Map.Entry<K, V>> mappingsSupplier) {
        return bound == 0 ? emptyMap() : stream(1, bound, mappingsSupplier).collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue));
    }

    @NonNull
    public <K, V> Map<K, V> map(int lowerBound, int upperBound, @NonNull final Supplier<Map.Entry<K, V>> mappingsSupplier) {
        return upperBound == 0 ? emptyMap() : stream(lowerBound, upperBound, mappingsSupplier).collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue));
    }

    @NonNull
    public XMLGregorianCalendar xmlGregorianCalendar() {
        final GregorianCalendar calendar = new GregorianCalendar();
        calendar.setTime(new Date());
        try {
            final DatatypeFactory factory = DatatypeFactory.newInstance();
            return factory.newXMLGregorianCalendar(calendar);
        } catch (DatatypeConfigurationException e) {
            throw new TestAbortedException(format("Failed to create instance of %s", XMLGregorianCalendar.class), e);
        }
    }

    @NonNull
    public ZonedDateTime zonedDateTime() {
        return ZonedDateTime.ofInstant(Instant.ofEpochSecond(ThreadLocalRandom.current().nextLong(0, Instant.now().getEpochSecond())), ZoneId.systemDefault());
    }

    public int nextInt(final int lowerBound, final int upperBound) {
        return upperBound == lowerBound ? lowerBound : nextInt(upperBound - lowerBound) + lowerBound;
    }

    @NonNull
    public String string(final int maxLength) {
        if (maxLength == 0) return EMPTY;
        final byte[] bytes = new byte[nextInt(maxLength)];
        nextBytes(bytes);
        return new String(bytes, UTF_8);
    }

    @NonNull
    public LocalDateTime localDateTime() {
        return LocalDateTime.ofEpochSecond(nextInt(), nextInt(1000000000), ZoneOffset.UTC);
    }

    @NonNull
    public LocalDateTime localDateTimeAfter(LocalDateTime that) {
        final int randomSeconds = next(Integer.SIZE - 1); //fastest way to get a non-negative random int
        final long maxSecondsPossible = LocalDateTime.MAX.toEpochSecond(ZoneOffset.MAX) - that.toEpochSecond(ZoneOffset.MAX);
        if (maxSecondsPossible <= 0) throw new IllegalArgumentException("cannot create a LocalDateTime later than " + that);
        return that
            .plusSeconds(Math.min(randomSeconds, maxSecondsPossible))
            .plusNanos(10000000); //in case random int was 0
    }

    @NonNull
    public String countryCode() {
        final String[] isoCountries = Locale.getISOCountries();
        return isoCountries[nextInt(isoCountries.length)];
    }

    @NonNull
    public Stream<String> strings(final int maxLength, final int maxLimit) {
        return maxLimit == 0 ? empty() : generate(() -> string(maxLength)).limit(nextInt(maxLimit));
    }

    @NonNull
    public Stream<String> strings(final int maxLength, final int minLimit, final int maxLimit) {
        return maxLimit == 0 ? empty() : generate(() -> string(maxLength)).limit(nextInt(minLimit, maxLimit));
    }

    @NonNull
    public Stream<LocalDateTime> localDateTimes(final int maxLimit) {
        return maxLimit == 0 ? empty() : generate(this::localDateTime).limit(nextInt(maxLimit));
    }

    @NonNull
    public Stream<LocalDateTime> localDateTimes(final int minLimit, final int maxLimit) {
        return maxLimit == 0 ? empty() : generate(this::localDateTime).limit(nextInt(minLimit, maxLimit));
    }

    @NonNull
    public Stream<Long> longs(final int maxLimit) {
        return maxLimit == 0 ? empty() : generate(this::nextLong).limit(nextInt(maxLimit));
    }

    @NonNull
    public Stream<Long> longs(final int minLimit, final int maxLimit) {
        return maxLimit == 0 ? empty() : generate(this::nextLong).limit(nextInt(minLimit, maxLimit));
    }

    @SafeVarargs
    @Nullable
    public final <T> T choose(@NonNull final T... values) {
        return values[nextInt(values.length)];
    }

    @Nullable
    public final <T> T choose(@NonNull final Collection<T> values) {
        final int itemNumber = nextInt(values.size()) + 1;
        final Iterator<T> iterator = values.iterator();
        T value = null;
        for (int i = 0; i < itemNumber; i++) value = iterator.next();
        return value;
    }

    @Nullable
    public final <E extends Enum<E>> E choose(@NonNull Class<E> enumeration) {
        return choose(EnumSet.allOf(enumeration));
    }

}
