/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.test;

import com.google.common.truth.Truth;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.junit.jupiter.params.provider.ValueSource;
import org.opentest4j.TestAbortedException;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import javax.xml.datatype.DatatypeFactory;
import java.lang.reflect.Field;
import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.time.format.DateTimeFormatter;
import java.util.List;
import java.util.Objects;
import java.util.Random;
import java.util.Set;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Stream;

import static java.lang.String.format;
import static java.lang.reflect.Modifier.FINAL;
import static java.time.LocalDate.parse;
import static java.time.format.DateTimeFormatter.ofPattern;
import static java.util.stream.Stream.generate;
import static java.util.stream.Stream.of;
import static javax.xml.datatype.DatatypeFactory.DATATYPEFACTORY_IMPLEMENTATION_CLASS;
import static org.junit.jupiter.api.Assertions.*;
import static org.junit.jupiter.api.Assumptions.assumeTrue;
import static org.junit.jupiter.params.provider.Arguments.arguments;

class RandomDataTest {

    private static final int SEED = "SEED".hashCode();
    private static final int CPR_LENGTH = 10;
    private static final RandomData RANDOM_DATA = new RandomData(SEED);
    private static final Random RANDOM = new Random(SEED);
    private static final DateTimeFormatter CPR_FORMATTER = ofPattern("ddMMyy");
    public static final int MAX_REPEATS = 100;
    public static final int MAX_TESTS = 100;


    @DisplayName("cpr numbers")
    @ParameterizedTest(name = "cpr: \"{0}\"")
    @MethodSource("cpr")
    void cpr(@NonNull final String cpr) {
        assertNotNull(cpr);
        assertEquals(CPR_LENGTH, cpr.length(), format("Cpr number %s has invalid length", cpr));
        assertDoesNotThrow(() -> parse(cpr.substring(0, 6), CPR_FORMATTER), "Invalid date");
    }

    @NonNull
    private static Stream<Arguments> cpr() {
        return generate(RANDOM_DATA::cpr).limit(100).map(Arguments::arguments);
    }


    @ParameterizedTest(name = "sets({0}, {1}, () -> item)")
    @MethodSource("bounds")
    <T> void set(final int lowerBound, final int upperBound) {
        final AtomicInteger count = new AtomicInteger();
        for (int i = RANDOM.nextInt(MAX_REPEATS); i > 0; i--) {
            count.set(0);
            RANDOM_DATA.set(lowerBound, upperBound, Object::new).forEach(element -> {
                count.incrementAndGet();
                assertNotNull(element);
            });
            assertTrue(count.get() <= upperBound);
            assertTrue(count.get() >= lowerBound);
        }
    }

    @ParameterizedTest
    @DisplayName("set with positive bound")
    @ValueSource(ints = {0, 1, 2, 4, 8, 16, 32, 64, 128})
    void setValidBound(final int bound) {
        final Set<Integer> set = RANDOM_DATA.set(bound, () -> bound);
        assertTrue(bound >= set.size());
        set.forEach((element) -> assertEquals(bound, element));
    }

    @NonNull
    private static Stream<Arguments> bounds() {
        return generate(() -> {
            final int minLimit = RANDOM.nextInt(10);
            return arguments(minLimit, minLimit + RANDOM.nextInt(10));
        }).limit(MAX_TESTS);
    }


    @ParameterizedTest
    @DisplayName("list with positive bound")
    @ValueSource(ints = {0, 1, 2, 4, 8, 16, 32, 64, 128})
    void listValidBound(final int bound) {
        final List<Integer> list = RANDOM_DATA.list(bound, () -> bound);
        assertTrue(bound >= list.size());
        list.forEach((element) -> assertEquals(bound, element));
    }

    @ParameterizedTest
    @DisplayName("list with negative bound throws IllegalArgumentException")
    @ValueSource(ints = {-1, -2, -4, -8, -16, -32, -64, -128})
    void listNegativeBound(final int bound) {
        assertThrows(IllegalArgumentException.class, () -> RANDOM_DATA.list(bound, Object::new));
    }

    @ParameterizedTest(name = "lists({0}, {1}, () -> item)")
    @MethodSource("list3")
    <T> void list(final int lowerBound, final int upperBound, @NonNull final T item) {
        final AtomicInteger count = new AtomicInteger();
        for (int i = RANDOM.nextInt(MAX_REPEATS); i > 0; i--) {
            count.set(0);
            RANDOM_DATA.list(lowerBound, upperBound, () -> item).forEach(element -> {
                count.incrementAndGet();
                assertSame(element, item);
            });
            assertTrue(count.get() <= upperBound);
            assertTrue(count.get() >= lowerBound);
        }
    }

    @NonNull
    private static Stream<Arguments> list3() {
        return generate(() -> {
            final int minLimit = RANDOM.nextInt(10);
            return arguments(minLimit, minLimit + RANDOM.nextInt(10), new Object());
        }).limit(MAX_TESTS);
    }

    @ParameterizedTest(name = "longs({0})")
    @ValueSource(ints = {0, 1, 2, 4, 8, 16, 32, 64, 128})
    void longs(final int bound) {
        final AtomicInteger count = new AtomicInteger();
        for (int i = RANDOM.nextInt(MAX_REPEATS); i > 0; i--) {
            count.set(0);
            RANDOM_DATA.longs(bound).forEach(localDateTime -> {
                count.incrementAndGet();
                assertNotNull(localDateTime);
            });
            assertTrue(count.get() <= bound);
        }
    }

    @ParameterizedTest(name = "longs({0}, {1})")
    @MethodSource("lowerAndUpperBounds")
    @SuppressWarnings("EndlessStream")
    void longs(final int lowerBound, final int upperBound) {
        final AtomicInteger count = new AtomicInteger();
        for (int i = RANDOM.nextInt(MAX_REPEATS); i > 0; i--) {
            count.set(0);
            RANDOM_DATA.longs(lowerBound, upperBound).forEach(localDateTime -> {
                count.incrementAndGet();
                assertNotNull(localDateTime);
            });
            assertTrue(count.get() <= upperBound);
            assertTrue(count.get() >= lowerBound);
        }
    }

    @ParameterizedTest(name = "LocalDateTime.ofEpochSecond({0}, 0, ZoneOffset.UTC)")
    @ValueSource(longs = {-31557014135596800L /*LocalDateTime.MIN*/, 123456789L, 31556889832780798L/*LocalDateTime.MAX - 1*/})
    void localDateTimeAfter(long seconds) {
        final LocalDateTime input = LocalDateTime.ofEpochSecond(seconds, 0, ZoneOffset.UTC);
        final LocalDateTime output = RANDOM_DATA.localDateTimeAfter(input);
        Truth.assertThat(output).isGreaterThan(input);
    }

    @ParameterizedTest(name = "localDateTimes({0})")
    @ValueSource(ints = {0, 1, 2, 4, 8, 16, 32, 64, 128})
    void localDateTimes(final int bound) {
        final AtomicInteger count = new AtomicInteger();
        for (int i = RANDOM.nextInt(MAX_REPEATS); i > 0; i--) {
            count.set(0);
            RANDOM_DATA.localDateTimes(bound).forEach(localDateTime -> {
                count.incrementAndGet();
                assertNotNull(localDateTime);
            });
            assertTrue(count.get() <= bound);
        }
    }

    @ParameterizedTest(name = "localDateTimes({0}, {1})")
    @MethodSource("lowerAndUpperBounds")
    void localDateTimes(final int lowerBound, final int upperBound) {
        final AtomicInteger count = new AtomicInteger();
        for (int i = RANDOM.nextInt(MAX_REPEATS); i > 0; i--) {
            count.set(0);
            RANDOM_DATA.localDateTimes(lowerBound, upperBound).forEach(localDateTime -> {
                count.incrementAndGet();
                assertNotNull(localDateTime);
            });
            assertTrue(count.get() <= upperBound);
            assertTrue(count.get() >= lowerBound);
        }
    }

    @NonNull
    private static Stream<Arguments> lowerAndUpperBounds() {
        return generate(() -> {
            final int minLimit = RANDOM.nextInt(10);
            return arguments(minLimit, minLimit + RANDOM.nextInt(10));
        }).limit(MAX_TESTS);
    }

    @ParameterizedTest(name = "{index}: max length = {0}")
    @DisplayName("string with positive max length")
    @ValueSource(ints = {0, 1, 2, 4, 8, 16, 32, 64, 128})
    void stringValidBound(final int bound) {
        assertTrue(bound >= RANDOM_DATA.string(bound).length());
    }

    @ParameterizedTest(name = "{index}: max length = {1}")
    @DisplayName("string with negative max length throws IllegalArgumentException")
    @ValueSource(ints = {-1, -2, -4, -8, -16, -32, -64, -128})
    void stringNegativeBound(final int bound) {
        assertThrows(IllegalArgumentException.class, () -> RANDOM_DATA.string(bound));
    }

    @ParameterizedTest(name = "strings({0}, {1})")
    @MethodSource("strings2")
    void strings(final int maxLength, final int maxLimit) {
        final AtomicInteger count = new AtomicInteger();
        for (int i = RANDOM.nextInt(MAX_REPEATS); i > 0; i--) {
            count.set(0);
            RANDOM_DATA.strings(maxLength, maxLimit).forEach(string -> {
                count.incrementAndGet();
                assertTrue(string.length() <= maxLength);
            });
            assertTrue(count.get() <= maxLimit);
        }
    }

    @ParameterizedTest(name = "strings({0}, {1}, {2})")
    @MethodSource("strings3")
    void strings(final int maxLength, final int minLimit, final int maxLimit) {
        final AtomicInteger count = new AtomicInteger();
        for (int i = RANDOM.nextInt(MAX_REPEATS); i > 0; i--) {
            count.set(0);
            RANDOM_DATA.strings(maxLength, minLimit, maxLimit).forEach(string -> {
                count.incrementAndGet();
                assertTrue(string.length() <= maxLength);
            });
            assertTrue(count.get() <= maxLimit);
            assertTrue(count.get() >= minLimit);
        }
    }

    @NonNull
    private static Stream<Arguments> strings2() {
        return generate(() -> arguments(RANDOM.nextInt(10), RANDOM.nextInt(10))).limit(MAX_TESTS);
    }

    @NonNull
    private static Stream<Arguments> strings3() {
        return generate(() -> {
            final int minLimit = RANDOM.nextInt(10);
            return arguments(RANDOM.nextInt(10), minLimit, minLimit + RANDOM.nextInt(10));
        }).limit(MAX_TESTS);
    }

    @ParameterizedTest(name = "choose({0}, {1}, {2})")
    @MethodSource("choose1")
    void choose(
        @Nullable final Object obj1,
        @Nullable final Object obj2,
        @Nullable final Object obj3) {
        final Object chosen = RANDOM_DATA.choose(obj1, obj2, obj3);
        final Object match = of(obj1, obj2, obj3)
            .filter(Objects::nonNull)
            .filter(o -> o == chosen)
            .findAny()
            .orElse(null);
        assertEquals(match, chosen);
    }

    @ParameterizedTest(name = "choose({0})")
    @MethodSource("choose2")
    void choose(@NonNull final Class<Enum> enumeration) {
        final Enum chosen = RANDOM_DATA.choose(Enum.class);
        final Object match = of(Enum.values())
            .filter(o -> o == chosen)
            .findAny()
            .orElse(null);
        assertEquals(match, chosen, format("Failed for enum %s", enumeration));
    }

    @NonNull
    private static Stream<Arguments> choose1() {
        return
            of(null, new Object()).flatMap(obj1 ->
                of(null, new Object()).flatMap(obj2 ->
                    of(null, new Object()).map(obj3 ->
                        arguments(obj1, obj2, obj3))));

    }

    @NonNull
    private static Stream<Arguments> choose2() {
        return generate(() -> Enum.class)
            .limit(10)
            .map(Arguments::arguments);
    }


    @Test
    @DisplayName("XML Gregorian Calendar")
    void xmlCalendar() {
        assertNotNull(RANDOM_DATA.xmlGregorianCalendar());
    }

    @Test
    @DisplayName("XML Gregorian Calendar - DatatypeConfigurationException")
    @SuppressWarnings({"ConstantConditions", "UnnecessaryLocalVariable", "java:S5778"})
    void xmlCalendarException() throws IllegalAccessException, NoSuchFieldException {
        final String value = DATATYPEFACTORY_IMPLEMENTATION_CLASS;
        setDataTypeFactoryImplementationClass(null);
        assumeTrue(DATATYPEFACTORY_IMPLEMENTATION_CLASS == null);
        assertThrows(TestAbortedException.class, () -> {
            try {
                RANDOM_DATA.xmlGregorianCalendar();
            } finally {
                setDataTypeFactoryImplementationClass(value);
            }
        });
    }

    @Test
    void zonedDateTime() {
        assertNotNull(RANDOM_DATA.zonedDateTime());
    }

    private void setDataTypeFactoryImplementationClass(@Nullable String value) throws NoSuchFieldException, IllegalAccessException {
        final Field field = DatatypeFactory.class.getDeclaredField("DATATYPEFACTORY_IMPLEMENTATION_CLASS");
        final Field modifiersField = Field.class.getDeclaredField("modifiers");
        modifiersField.setAccessible(true);
        modifiersField.setInt(field, field.getModifiers() & ~FINAL);
        field.setAccessible(true);
        field.set(null, value);
    }

    private enum Enum {
        ONE, TWO, THREE
    }

}
