/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package se.digg.dgc.service;

import java.io.IOException;
import java.security.SignatureException;
import java.security.cert.CertificateExpiredException;

import se.digg.dgc.payload.v1.DGCSchemaException;
import se.digg.dgc.payload.v1.DigitalGreenCertificate;

/**
 * Service for decoding a Digital Green Certificate from its image representation into the actual DGC payload.
 * 
 * @author Martin Lindström (martin@idsec.se)
 * @author Henrik Bengtsson (extern.henrik.bengtsson@digg.se)
 * @author Henric Norlander (extern.henric.norlander@digg.se)
 */
public interface DGCDecoder {
  
  /**
   * Given the base45-encoding of a signed DGC the method verifies and decodes it into the DGC payload representation. 
   * 
   * @param base45
   *          the base45-encoding of the signed DGC (including the HCERT header)
   * @return the DGC payload
   * @throws DGCSchemaException
   *           for DGC schema errors
   * @throws SignatureException
   *           for signature verification errors
   * @throws CertificateExpiredException
   *           if the DGC has expired
   * @throws IOException
   *           for errors decoding data, for example CBOR related errors
   * @see #decodeBarcodeToBytes(byte[])
   */
  DigitalGreenCertificate decode(final String base45)
      throws DGCSchemaException, SignatureException, CertificateExpiredException, IOException;

  /**
   * Given the base45-encoding of a signed DGC the method verifies and decodes it into the CBOR encoding of the DGC payload.
   * 
   * @param base45
   *          the base45-encoding of the signed DGC (including the HCERT header)
   * @return the CBOR encoding of the DGC payload
   * @throws SignatureException
   *           for signature verification errors
   * @throws CertificateExpiredException
   *           if the DGC has expired
   * @throws IOException
   *           for errors decoding data, for example CBOR related errors
   * @see #decodeBarcode(byte[])
   */
  byte[] decodeToBytes(final String base45) throws SignatureException, CertificateExpiredException, IOException;

  /**
   * Verifies a "raw" DGC (i.e., a signed CWT holding the DGC payload) and decodes it to the actual DGC payload.
   * 
   * @param cwt
   *          the signed CWT holding the DGC
   * @return the DGC payload
   * @throws DGCSchemaException
   *           for DGC schema errors
   * @throws SignatureException
   *           for signature verification errors
   * @throws CertificateExpiredException
   *           if the DGC has expired
   * @throws IOException
   *           for errors decoding data, for example CBOR related errors
   * @see #decodeRawToBytes(byte[])
   */
  DigitalGreenCertificate decodeRaw(final byte[] cwt)
      throws DGCSchemaException, SignatureException, CertificateExpiredException, IOException;

  /**
   * Verifies a "raw" DGC (i.e., a signed CWT holding the DGC payload) and returns the CBOR encoding of the DGC payload.
   * 
   * @param cwt
   *          the signed CWT holding the DGC
   * @return the CBOR encoded DGC payload
   * @throws SignatureException
   *           for signature verification errors
   * @throws CertificateExpiredException
   *           if the DGC has expired
   * @throws IOException
   *           for errors decoding data, for example CBOR related errors
   * @see #decodeRaw(byte[])
   */
  byte[] decodeRawToBytes(final byte[] cwt) throws SignatureException, CertificateExpiredException, IOException;

}
