/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package se.digg.dgc.service.impl;

import com.upokecenter.cbor.CBORException;
import com.upokecenter.cbor.CBORObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import se.digg.dgc.encoding.Base45;
import se.digg.dgc.encoding.DKConstants;
import se.digg.dgc.encoding.Zlib;
import se.digg.dgc.service.DKEncoder;
import se.digg.dgc.signatures.impl.DefaultDKSigner;

import java.io.IOException;
import java.security.SignatureException;
import java.time.Instant;
import java.util.Optional;

/**
 * A bean implementing the {@link DKEncoder} interface.
 */
public class DefaultDKEncoder implements DKEncoder {

  /** Logger */
  private static final Logger log = LoggerFactory.getLogger(DefaultDKEncoder.class);

  /** The DGC signer instance. */
  private final DefaultDKSigner dgcSigner;

  /**
   * Constructor.
   *
   * @param dgcSigner
   *          a signer for signing the DGC:s
   */
  public DefaultDKEncoder(final DefaultDKSigner dgcSigner) {
    this.dgcSigner = Optional.ofNullable(dgcSigner).orElseThrow(() -> new IllegalArgumentException("dgcSigner must not be null"));
  }

  /** {@inheritDoc} */

  public String encodeDk(final CBORObject cborObject, final Instant expiration, String cti) throws IOException, SignatureException {
    return encode(cborObject, expiration, DKConstants.DK_HEADER, cti);
  }

  /** {@inheritDoc} */

  public String encodeDkPlus(final CBORObject cborObject, final Instant expiration, String cti) throws IOException, SignatureException {
    return encode(cborObject, expiration, DKConstants.DKPLUS_HEADER, cti);
  }

  private String encode(final CBORObject cborObject, final Instant expiration, String prefix,  String cti) throws IOException, SignatureException {


    // Create a signed CWT ...
    //
    byte[] cwt = this.sign(cborObject, expiration, cti);

    // Compression and Base45 encoding ...
    //
    log.trace("Compressing the signed CWT of length {} ...", cwt.length);
    cwt = Zlib.compress(cwt);
    log.trace("Signed CWT was compressed into {} bytes", cwt.length);

    log.trace("Base45 encoding compressed CWT ...");
    final String base45 = Base45.getEncoder().encodeToString(cwt);
    log.trace("Base45 encoding: {}", base45);

    return prefix + base45; // DK version med andre præfix
  }

  /** {@inheritDoc} */

  public byte[] sign(final CBORObject cborObject, final Instant expiration, String cti) throws IOException, SignatureException {

    try {
      // Sign the bytes ...
      //

      return this.dgcSigner.sign(cborObject, expiration, cti);
    }
    catch (final CBORException e) {
      log.info("Internal CBOR error - {}", e.getMessage(), e);
      throw new IOException("Internal CBOR error - " + e.getMessage(), e);
    }
  }

}
