/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package se.swedenconnect.security.credential;

import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.cert.X509Certificate;
import java.util.UUID;

/**
 * A basic implementation of the {@link PkiCredential} interface.
 *
 * @author Martin Lindström (martin@idsec.se)
 * @author Stefan Santesson (stefan@idsec.se)
 */
public class BasicCredential extends AbstractPkiCredential {

    /**
     * Default constructor.
     */
    public BasicCredential() {
        super();
    }

    /**
     * Constructor setting the public and private keys.
     *
     * @param publicKey
     *          the public key
     * @param privateKey
     *          the private key
     */
    public BasicCredential(final PublicKey publicKey, final PrivateKey privateKey) {
        this.setPublicKey(publicKey);
        this.setPrivateKey(privateKey);
    }

    /**
     * Constructor setting the certificate and private key.
     *
     * @param certificate
     *          the certificate
     * @param privateKey
     *          the private key
     */
    public BasicCredential(final X509Certificate certificate, final PrivateKey privateKey) {
        this.setCertificate(certificate);
        this.setPrivateKey(privateKey);
    }



    /**
     * Gets the subject DN of the certificate and if no certificate is available an UUID is used.
     */
    @Override
    protected String getDefaultName() {
        final X509Certificate cert = this.getCertificate();
        if (cert != null) {
            return cert.getSubjectX500Principal().getName();
        }
        final PublicKey key = this.getPublicKey();
        if (key != null) {
            return String.format("%s-%s", key.getAlgorithm(), UUID.randomUUID());
        }
        return "BasicCredential-" + UUID.randomUUID().toString();
    }

}
