/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package se.digg.dgc.cose;

import com.upokecenter.cbor.CBORObject;
import com.upokecenter.numbers.EInteger;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.junit.Assert;
import org.junit.Ignore;
import org.junit.Test;
import se.digg.dgc.signatures.cose.CoseSign1_Object;
import se.digg.dgc.signatures.cose.HeaderParameterKey;
import se.digg.dgc.signatures.cose.SignatureAlgorithm;
import se.digg.dgc.signatures.cwt.Cwt;
import se.swedenconnect.security.credential.BasicCredential;
import se.swedenconnect.security.credential.PkiCredential;

import java.security.*;
import java.security.cert.Certificate;
import java.security.cert.X509Certificate;
import java.security.spec.ECGenParameterSpec;
import java.time.Instant;

import static se.digg.dgc.service.impl.DefaultDGCBarcodeEncoderDecoderTest.generateCertificate;
import static se.digg.dgc.signatures.cose.CoseSign1_Object.MESSAGE_TAG;

/**
 * Test cases for CoseSign1_Object.
 *
 * @author Martin Lindström (martin@idsec.se)
 * @author Henrik Bengtsson (extern.henrik.bengtsson@digg.se)
 * @author Henric Norlander (extern.henric.norlander@digg.se)
 */
public class CoseSign1_ObjectTest {

  private PkiCredential rsa;
  private PkiCredential ecdsa;

  private static final char[] password = "secret".toCharArray();
    public CoseSign1_ObjectTest() throws Exception {

       // this.rsa = new KeyStoreCredential(new ClassPathResource("rsa.jks"), password, "signer", password);
       // this.rsa.init();

        Security.insertProviderAt(new BouncyCastleProvider(), 1);
        KeyPairGenerator keyPairGenerator = KeyPairGenerator.getInstance("RSA", "BC");
        keyPairGenerator.initialize(2048, new SecureRandom());

        KeyPair keyPair = keyPairGenerator.generateKeyPair();
        Certificate certificate = generateCertificate(keyPair);
        rsa = new BasicCredential((X509Certificate) certificate, keyPair.getPrivate());

        KeyPairGenerator g = KeyPairGenerator.getInstance("ECDSA", "BC");
        ECGenParameterSpec spec = new ECGenParameterSpec("secp256r1");
        g.initialize(spec);
        KeyPair eckeyPair = g.generateKeyPair();
        ecdsa = new BasicCredential(eckeyPair.getPublic(), eckeyPair.getPrivate());
    }


  @Test
  @Ignore
  public void testSignVerifyEc() throws Exception {
    final Cwt cwt = Cwt.builder().issuer("SE").issuedAt(Instant.now()).build();
    final CoseSign1_Object sign = CoseSign1_Object.builder()

      .protectedAttribute(HeaderParameterKey.ALG.getCborObject(), SignatureAlgorithm.ES256.getCborObject())
      .protectedAttribute(HeaderParameterKey.KID.getCborObject(), CBORObject.FromObject(getKeyId(this.ecdsa.getCertificate())))
      .content(cwt.encode())
      .build();

    sign.sign(this.ecdsa.getPrivateKey(), null);

    final byte[] encoding = sign.encode();

      EInteger tag = CBORObject.DecodeFromBytes(encoding).getMostOuterTag();

    CoseSign1_Object object2 = CoseSign1_Object.decode(encoding);

        // In a real life scenario, this is where I would ask for the KID (in order to locate the public key to use).
    final byte[] kid = object2.getKeyIdentifier();
    Assert.assertTrue(kid.length == 8);

    object2.verifySignature(this.ecdsa.getPublicKey());

    final Cwt cwt2 = object2.getCwt();
    Assert.assertEquals("SE", cwt2.getIssuer());
  }

  @Test
  public void testSignVerifyRsa() throws Exception {

    final Cwt cwt = Cwt.builder().issuer("SE").issuedAt(Instant.now()).build();
    final CoseSign1_Object sign = CoseSign1_Object.builder()
        .includeMessageTag(true)
      .protectedAttribute(HeaderParameterKey.ALG.getCborObject(), SignatureAlgorithm.PS256.getCborObject())
      .protectedAttribute(HeaderParameterKey.KID.getCborObject(), CBORObject.FromObject(getKeyId(this.rsa.getCertificate())))
      .content(cwt.encode())
      .build();

    sign.sign(this.rsa.getPrivateKey(), null);

    final byte[] encoding = sign.encode();

    CoseSign1_Object object2 = CoseSign1_Object.decode(encoding);

    Assert.assertEquals(String.valueOf(MESSAGE_TAG), CBORObject.DecodeFromBytes(encoding).getMostOuterTag().toString());

    object2.verifySignature(this.rsa.getPublicKey());

    final Cwt cwt2 = object2.getCwt();
    Assert.assertEquals("SE", cwt2.getIssuer());
  }

  @Test(expected = SignatureException.class)
  @Ignore
  public void testFailedVerify() throws Exception {
    final Cwt cwt = Cwt.builder().issuer("SE").issuedAt(Instant.now()).build();
    final CoseSign1_Object sign = CoseSign1_Object.builder()
      .protectedAttribute(HeaderParameterKey.ALG.getCborObject(), SignatureAlgorithm.ES256.getCborObject())
      .protectedAttribute(HeaderParameterKey.KID.getCborObject(), CBORObject.FromObject(getKeyId(this.ecdsa.getCertificate())))
      .content(cwt.encode())
      .build();

    sign.sign(this.ecdsa.getPrivateKey(), null);

    final byte[] encoding = sign.encode();

    // Change the signature
    encoding[encoding.length - 1] = 0x00;

    CoseSign1_Object object2 = CoseSign1_Object.decode(encoding);
    object2.verifySignature(this.ecdsa.getPublicKey());
  }

  private static byte[] getKeyId(final X509Certificate cert) throws Exception {
    final MessageDigest digest = MessageDigest.getInstance("SHA-256");
    final byte[] sha256 = digest.digest(cert.getEncoded());
    final byte[] kid = new byte[8];
    System.arraycopy(sha256, 0, kid, 0, 8);
    return kid;
  }

}
