/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package se.digg.dgc.service.impl;

import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.x509.X509V1CertificateGenerator;
import org.junit.Assert;
import org.junit.Test;
import se.digg.dgc.encoding.Barcode;
import se.digg.dgc.encoding.impl.DefaultBarcodeCreator;
import se.digg.dgc.encoding.impl.DefaultBarcodeDecoder;
import se.digg.dgc.payload.v1.DigitalGreenCertificate;
import se.digg.dgc.payload.v1.PersonName;
import se.digg.dgc.payload.v1.VaccinationEntry;
import se.digg.dgc.signatures.impl.DefaultDGCSignatureVerifier;
import se.digg.dgc.signatures.impl.DefaultDGCSigner;
import se.swedenconnect.security.credential.BasicCredential;
import se.swedenconnect.security.credential.PkiCredential;

import javax.security.auth.x500.X500Principal;
import java.math.BigInteger;
import java.security.*;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.time.Duration;
import java.time.Instant;
import java.time.LocalDate;
import java.util.Arrays;
import java.util.Date;
import java.util.UUID;

/**
 * Test cases for {@link DefaultDGCBarcodeEncoder}.
 *
 * @author Martin Lindström (martin@idsec.se)
 * @author Henrik Bengtsson (extern.henrik.bengtsson@digg.se)
 * @author Henric Norlander (extern.henric.norlander@digg.se)
 */
public class DefaultDGCBarcodeEncoderDecoderTest {

    private PkiCredential rsa;

    private static final char[] password = "secret".toCharArray();


    public DefaultDGCBarcodeEncoderDecoderTest() throws NoSuchAlgorithmException, NoSuchProviderException, CertificateException, SignatureException, InvalidKeyException {

        Security.insertProviderAt(new BouncyCastleProvider(), 1);
        KeyPairGenerator keyPairGenerator = KeyPairGenerator.getInstance("RSA", "BC");
        keyPairGenerator.initialize(2048, new SecureRandom());

        KeyPair keyPair = keyPairGenerator.generateKeyPair();
        Certificate certificate = generateCertificate(keyPair);
        rsa = new BasicCredential((X509Certificate) certificate, keyPair.getPrivate());
    }

    public static Certificate generateCertificate(KeyPair keyPair) throws CertificateException, NoSuchAlgorithmException, SignatureException, NoSuchProviderException, InvalidKeyException {
        // yesterday
        Date validityBeginDate = new Date(System.currentTimeMillis() - 24L * 60 * 60 * 1000);
        // in 2 years
        Date validityEndDate = new Date(System.currentTimeMillis() + 2L * 365 * 24 * 60 * 60 * 1000); //use long, according to https://errorprone.info/bugpattern/ConstantOverflow

        X509V1CertificateGenerator certGen = new X509V1CertificateGenerator();
        X500Principal dnName = new X500Principal("CN=John Doe, C=DK");
        certGen.setSerialNumber(BigInteger.valueOf(System.currentTimeMillis()));
        certGen.setSubjectDN(dnName);
        certGen.setIssuerDN(dnName); // use the same
        certGen.setNotBefore(validityBeginDate);
        certGen.setNotAfter(validityEndDate);
        certGen.setPublicKey(keyPair.getPublic());
        certGen.setSignatureAlgorithm("SHA256WithRSAEncryption");
        return certGen.generate(keyPair.getPrivate(), "BC");

    }

    @Test
    public void testEncodeDecodeBarcode() throws Exception {

        final Instant expire = Instant.now().plus(Duration.ofDays(30));
        final DigitalGreenCertificate dgc = getTestDGC();

        final DefaultDGCBarcodeEncoder encoder = new DefaultDGCBarcodeEncoder(new DefaultDGCSigner(this.rsa), new DefaultBarcodeCreator());

        final Barcode barcode = encoder.encodeToBarcode(dgc, expire);

        final DefaultDGCBarcodeDecoder decoder = new DefaultDGCBarcodeDecoder(
            new DefaultDGCSignatureVerifier(), (x, y) -> Arrays.asList(this.rsa.getCertificate()), new DefaultBarcodeDecoder());

        final DigitalGreenCertificate dgc2 = decoder.decodeBarcode(barcode.getImage());
        Assert.assertEquals(dgc, dgc2);
    }


    private DigitalGreenCertificate getTestDGC() {
        DigitalGreenCertificate dgc = new DigitalGreenCertificate();
        dgc.setVer("1.0.0");

        dgc.setNam(new PersonName().withGn("Karl Mårten").withFn("Lindström"));
        dgc.setDob(LocalDate.parse("1969-11-29"));


        VaccinationEntry vac = new VaccinationEntry();
        vac
            .withCi(UUID.randomUUID().toString())
            .withTg("840539006")
            .withDt(LocalDate.parse("2021-04-02"))
            .withCo("SE")
            .withDn(Integer.valueOf(1))
            .withSd(Integer.valueOf(2))
            .withIs("eHälsomyndigheten")
            .withMa("ORG-100001699")
            .withMp("CVnCoV")
            .withVp("1119349007");

        dgc.setV(Arrays.asList(vac));

        return dgc;
    }

}
