/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.fluent.spring.jdbc;

import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.jdbc.core.namedparam.SqlParameterSource;
import org.springframework.lang.NonNull;

import java.sql.ResultSet;
import java.sql.Timestamp;
import java.sql.Types;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.function.BiConsumer;
import java.util.function.BiFunction;
import java.util.function.Function;
import java.util.function.Supplier;

public final class Mappers {

    private Mappers() {
    }

    @NonNull
    public static <T, K, M extends MapperBuilder<T, K, M>> MapperBuilder<T, K, M> mapper(
        @NonNull final Class<K> keyClass,
        @NonNull final TableMetaData insertTableMetaData,
        @NonNull final TableMetaData selectTableMetaData,
        @NonNull final Supplier<T> constructor) {
        return new MapperBuilder<>(
            keyClass,
            insertTableMetaData,
            selectTableMetaData,
            constructor);
    }

    public static final class SearchMapperBuilder<T, S> extends MapperBuilder<T, S, SearchMapperBuilder<T, S>> {

        @NonNull private final String searchColumnLabel;
        @NonNull private final Function<S, ?> searchSerializer;
        private final int sqlType;

        protected SearchMapperBuilder(
            @NonNull final TableMetaData insertTableMetaData,
            @NonNull final TableMetaData selectTableMetaData,
            @NonNull final Supplier<T> constructor,
            @NonNull final List<BiConsumer<T, ResultSet>> deserializers,
            @NonNull final Map<String, Serializer<T, ?>> serializers,
            @NonNull final String searchColumnLabel,
            final int sqlType,
            @NonNull final Function<S, ?> searchSerializer) {
            super(insertTableMetaData, selectTableMetaData, constructor, deserializers, serializers);
            this.searchColumnLabel = searchColumnLabel;
            this.searchSerializer = searchSerializer;
            this.sqlType = sqlType;
        }

        public static <T, S> SearchMapperBuilder<T, S> of(
            @NonNull final MapperBuilder<T, S, ?> mapperBuilder,
            @NonNull final String searchColumnLabel,
            final int sqlType,
            @NonNull final Function<S, ?> searchSerializer) {
            return new SearchMapperBuilder<>(
                mapperBuilder.insertTableMetaData,
                mapperBuilder.selectTableMetaData,
                mapperBuilder.constructor,
                mapperBuilder.deserializers,
                mapperBuilder.serializers,
                searchColumnLabel,
                sqlType,
                searchSerializer);
        }

        @NonNull
        @Override
        public SearchMapper<T, S> build() {
            return new SearchMapperImpl<>(
                insertTableMetaData,
                selectTableMetaData,
                constructor,
                deserializers,
                serializers,
                searchColumnLabel,
                sqlType,
                searchSerializer);
        }
    }

    public static final class KeyMapperBuilder<T, K> extends MapperBuilder<T, K, KeyMapperBuilder<T, K>> {

        @NonNull private final String keyColumnLabel;
        @NonNull private final Function<K, ?> keySerializer;
        private final int sqlType;

        protected KeyMapperBuilder(
            @NonNull final TableMetaData insertTableMetaData,
            @NonNull final TableMetaData selectTableMetaData,
            @NonNull final Supplier<T> constructor,
            @NonNull final List<BiConsumer<T, ResultSet>> deserializers,
            @NonNull final Map<String, Serializer<T, ?>> serializers,
            @NonNull final String keyColumnLabel,
            @NonNull final Function<K, ?> keySerializer,
            final int sqlType) {
            super(insertTableMetaData, selectTableMetaData, constructor, deserializers, serializers);
            this.keyColumnLabel = keyColumnLabel;
            this.keySerializer = keySerializer;
            this.sqlType = sqlType;
        }

        public static <T, K> KeyMapperBuilder<T, K> of(
            @NonNull final MapperBuilder<T, K, ?> mapperBuilder,
            @NonNull final String keyColumnName,
            final int sqlType,
            @NonNull final Function<K, ?> keySerializer) {
            return new KeyMapperBuilder<>(
                mapperBuilder.insertTableMetaData,
                mapperBuilder.selectTableMetaData,
                mapperBuilder.constructor,
                mapperBuilder.deserializers,
                mapperBuilder.serializers,
                keyColumnName,
                keySerializer,
                sqlType);
        }

        @NonNull
        @Override
        public KeyMapper<T, K> build() {
            return new KeyMapperImpl<>(
                insertTableMetaData,
                selectTableMetaData,
                constructor,
                deserializers,
                serializers,
                keyColumnLabel,
                sqlType,
                keySerializer
            );
        }
    }

    public static class MapperBuilder<T, S, M extends MapperBuilder<T, S, M>> {

        @NonNull protected final TableMetaData insertTableMetaData;
        @NonNull protected final TableMetaData selectTableMetaData;
        @NonNull protected final Supplier<T> constructor;
        @NonNull protected final List<BiConsumer<T, ResultSet>> deserializers;
        @NonNull protected final Map<String, Serializer<T, ?>> serializers;
        @NonNull protected final M actual;

        @SuppressWarnings({"unused", "unchecked"})
        private MapperBuilder(
            @NonNull final Class<S> keyClass,
            @NonNull final TableMetaData insertTableMetaData,
            @NonNull final TableMetaData selectTableMetaData,
            @NonNull final Supplier<T> constructor) {
            this.insertTableMetaData = insertTableMetaData;
            this.selectTableMetaData = selectTableMetaData;
            this.constructor = constructor;
            deserializers = new ArrayList<>();
            serializers = new HashMap<>();
            actual = (M) this;
        }

        @SuppressWarnings("unchecked")
        protected MapperBuilder(
            @NonNull final TableMetaData insertTableMetaData,
            @NonNull final TableMetaData selectTableMetaData,
            @NonNull final Supplier<T> constructor,
            @NonNull final List<BiConsumer<T, ResultSet>> deserializers,
            @NonNull final Map<String, Serializer<T, ?>> serializers) {
            this.insertTableMetaData = insertTableMetaData;
            this.selectTableMetaData = selectTableMetaData;
            this.constructor = constructor;
            this.deserializers = deserializers;
            this.serializers = serializers;
            actual = (M) this;
        }

        @NonNull
        public <U, V> SearchMapperBuilder<T, S> addSearchMapping(
            @NonNull final String columnLabel,
            final int sqlType,
            @NonNull final Deserializer<T, U> deserializer,
            @NonNull final Serializer<T, V> serializer,
            @NonNull final Function<S, ?> searchSerializer) {
            addMapping(columnLabel, deserializer, serializer);
            return SearchMapperBuilder.of(this, columnLabel, sqlType, searchSerializer);
        }

        @NonNull
        public <U, V> KeyMapperBuilder<T, S> addKeyMapping(
            @NonNull final String columnLabel,
            final int sqlType,
            @NonNull final Deserializer<T, U> deserializer,
            @NonNull final Serializer<T, V> serializer,
            @NonNull final Function<S, ?> keySerializer) {
            addMapping(columnLabel, deserializer, serializer);
            return KeyMapperBuilder.of(this, columnLabel, sqlType, keySerializer);
        }

        @NonNull
        public <U, V> M addMapping(
            @NonNull final String columnLabel,
            @NonNull final Deserializer<T, U> deserializer,
            @NonNull final Serializer<T, V> serializer) {
            deserializers.add(
                (obj, resultSet) -> deserializer.setter.accept(obj, deserializer.mapper.apply(resultSet, columnLabel)));
            serializers.put(columnLabel, serializer);
            return actual;
        }

        @NonNull
        public M addTimestampMapping(
            @NonNull final String columnLabel,
            @NonNull final BiConsumer<T, Timestamp> setter,
            @NonNull final Function<T, Timestamp> getter) {
            addMapping(columnLabel, timestamp(setter), timestamp(getter));
            return actual;
        }

        @NonNull
        public M addTimestampToLocalDateTimeMapping(
            @NonNull final String columnLabel,
            @NonNull final BiConsumer<T, LocalDateTime> setter,
            @NonNull final Function<T, LocalDateTime> getter) {
            addMapping(columnLabel, timestampToLocalDateTime(setter), localDateTime(getter));
            return actual;
        }

        @NonNull
        public M addStringMapping(
            @NonNull final String columnLabel,
            @NonNull final BiConsumer<T, String> setter,
            @NonNull final Function<T, String> getter) {
            addMapping(columnLabel, string(setter), string(getter));
            return actual;
        }

        @NonNull
        public M addLongMapping(
            @NonNull final String columnLabel,
            @NonNull final BiConsumer<T, Long> setter,
            @NonNull final Function<T, Long> getter) {
            addMapping(columnLabel, longNumber(setter), longNumber(getter));
            return actual;
        }

        @NonNull
        public KeyMapperBuilder<T, S> addKeyStringMapping(
            @NonNull final String columnLabel,
            @NonNull final BiConsumer<T, String> setter,
            @NonNull final Function<T, String> getter,
            @NonNull final Function<S, String> keySerializer) {
            return addKeyMapping(columnLabel, Types.VARCHAR, string(setter), string(getter), keySerializer);
        }

        @NonNull
        public SearchMapperBuilder<T, S> addSearchStringMapping(
            @NonNull final String columnLabel,
            @NonNull final BiConsumer<T, String> setter,
            @NonNull final Function<T, String> getter,
            @NonNull final Function<S, String> searchSerializer) {
            return addSearchMapping(columnLabel, Types.VARCHAR, string(setter), string(getter), searchSerializer);
        }

        @NonNull
        public M addIntegerMapping(
            @NonNull final String columnLabel,
            @NonNull final BiConsumer<T, Integer> setter,
            @NonNull final Function<T, Integer> getter) {
            addMapping(columnLabel, integer(setter), integer(getter));
            return actual;
        }

        @NonNull
        public M addBytesMapping(
            @NonNull final String columnLabel,
            @NonNull final BiConsumer<T, UUID> setter,
            @NonNull final Function<T, UUID> getter) {
            addMapping(columnLabel, bytes(setter), uuid(getter));
            return actual;
        }

        @NonNull
        public Mapper<T> build() {
            return new MapperImpl<>(
                insertTableMetaData,
                selectTableMetaData,
                constructor,
                deserializers,
                serializers);
        }
    }

    private static final class KeyMapperImpl<T, K> extends MapperImpl<T> implements KeyMapper<T, K> {

        @NonNull private final Function<K, ?> keySerializer;
        @NonNull private final String keyColumnLabel;
        private final int sqlType;

        private KeyMapperImpl(
            @NonNull final TableMetaData insertTableMetaData,
            @NonNull final TableMetaData selectTableMetaData,
            @NonNull final Supplier<T> constructor,
            @NonNull final List<BiConsumer<T, ResultSet>> deserializers,
            @NonNull final Map<String, Serializer<T, ?>> serializers,
            @NonNull final String keyColumnLabel,
            int sqlType, @NonNull final Function<K, ?> keySerializer) {
            super(insertTableMetaData, selectTableMetaData, constructor, deserializers, serializers);
            this.keyColumnLabel = keyColumnLabel;
            this.keySerializer = keySerializer;
            this.sqlType = sqlType;
        }

        @Override
        @NonNull
        public Object serializeKey(@NonNull final K key) {
            return keySerializer.apply(key);
        }

        @Override
        @NonNull
        public String keyColumnLabel() {
            return keyColumnLabel;
        }

        @Override
        public int sqlType() {
            return sqlType;
        }
    }

    private static final class SearchMapperImpl<T, S> extends MapperImpl<T> implements SearchMapper<T, S> {

        @NonNull private final Function<S, ?> searchSerializer;
        @NonNull private final String searchColumnLabel;
        private final int sqlType;

        private SearchMapperImpl(
            @NonNull final TableMetaData insertTableMetaData,
            @NonNull final TableMetaData selectTableMetaData,
            @NonNull final Supplier<T> constructor,
            @NonNull final List<BiConsumer<T, ResultSet>> deserializers,
            @NonNull final Map<String, Serializer<T, ?>> serializers,
            @NonNull final String searchColumnLabel,
            final int sqlType,
            @NonNull final Function<S, ?> searchSerializer) {
            super(insertTableMetaData, selectTableMetaData, constructor, deserializers, serializers);
            this.searchColumnLabel = searchColumnLabel;
            this.searchSerializer = searchSerializer;
            this.sqlType = sqlType;
        }

        @Override
        @NonNull
        public Object serializeSearchValue(@NonNull final S value) {
            return searchSerializer.apply(value);
        }

        @Override
        @NonNull
        public String searchColumnLabel() {
            return searchColumnLabel;
        }

        @Override
        public int sqlType() {
            return sqlType;
        }
    }

    private static class MapperImpl<T> implements Mapper<T> {

        @NonNull
        private final TableMetaData insertTableMetaData;
        @NonNull
        private final TableMetaData selectTableMetaData;
        @NonNull
        private final Supplier<T> constructor;
        @NonNull
        private final List<BiConsumer<T, ResultSet>> deserializers;
        @NonNull
        private final Map<String, Serializer<T, ?>> serializers;

        private MapperImpl(
            @NonNull final TableMetaData insertTableMetaData,
            @NonNull final TableMetaData selectTableMetaData,
            @NonNull final Supplier<T> constructor,
            @NonNull final List<BiConsumer<T, ResultSet>> deserializers,
            @NonNull final Map<String, Serializer<T, ?>> serializers) {
            this.insertTableMetaData = insertTableMetaData;
            this.selectTableMetaData = selectTableMetaData;
            this.constructor = constructor;
            this.deserializers = deserializers;
            this.serializers = serializers;
        }

        @Override
        @NonNull
        public T mapRow(@NonNull final ResultSet resultSet, final int rowNum) {
            final T obj = constructor.get();
            deserializers.forEach(fieldMapper -> fieldMapper.accept(obj, resultSet));
            return obj;
        }

        @Override
        @NonNull
        public TableMetaData insertTableMetaData() {
            return insertTableMetaData;
        }

        @Override
        @NonNull
        public TableMetaData selectTableMetaData() {
            return selectTableMetaData;
        }

        @Override
        @NonNull
        public SqlParameterSource parameterSource(@NonNull final T object) {
            final MapSqlParameterSource source = new MapSqlParameterSource();
            serializers.forEach(
                (columnLabel, serializer) ->
                    source.addValue(
                        columnLabel,
                        serializer.getter.apply(object),
                        serializer.sqlType));
            return source;
        }
    }

    public static <T, U> SerializerChain<T, U> serializer(@NonNull final Function<T, U> getter) {
        return new SerializerChain<>(getter);
    }

    public static final class SerializerChain<T, U> {
        @NonNull
        private final Function<? super T, U> getter;

        private SerializerChain(@NonNull final Function<? super T, U> getter) {
            this.getter = getter;
        }

        @NonNull
        public <V> SerializerChain<T, V> map(@NonNull final Function<? super U, V> mapper) {
            return new SerializerChain<>(getter.andThen(mapper));
        }

        @NonNull
        public <V> Serializer<T, V> get(@NonNull final Function<? super U, V> mapper, final int sqlType) {
            return new Serializer<>(getter.andThen(mapper), sqlType);
        }
    }

    public static final class Serializer<T, U> {
        private final int sqlType;
        @NonNull
        private final Function<? super T, U> getter;

        private Serializer(@NonNull final Function<? super T, U> getter, final int sqlType) {
            this.sqlType = sqlType;
            this.getter = getter;
        }
    }

    @NonNull
    public static <T> Serializer<T, String> string(@NonNull final Function<T, String> getter) {
        return new Serializer<>(getter, Types.VARCHAR);
    }

    @NonNull
    public static <T> Serializer<T, Timestamp> timestamp(@NonNull final Function<T, Timestamp> getter) {
        return new Serializer<>(getter, Types.TIMESTAMP);
    }

    @NonNull
    public static <T> Serializer<T, Timestamp> localDateTime(@NonNull final Function<T, LocalDateTime> getter) {
        return timestamp(owner -> {
            final LocalDateTime dateTime = getter.apply(owner);
            return dateTime == null ? null : Timestamp.valueOf(dateTime);
        });
    }

    @NonNull
    public static <T> Serializer<T, Long> longNumber(@NonNull final Function<T, Long> getter) {
        return new Serializer<>(getter, Types.BIGINT);
    }

    @NonNull
    public static <T> Serializer<T, Integer> integer(@NonNull final Function<T, Integer> getter) {
        return new Serializer<>(getter, Types.INTEGER);
    }

    @NonNull
    public static <T> Serializer<T, byte[]> uuid(@NonNull final Function<T, UUID> getter) {
        return new Serializer<>(getter.andThen(UUIDs::toBytes), Types.BINARY);
    }

    @NonNull
    public static <T> DeserializerChain<T> deserializer(@NonNull final BiFunction<ResultSet, String, T> function) {
        return new DeserializerChain<>(function);
    }

    public static final class DeserializerChain<T> {

        @NonNull
        private final BiFunction<ResultSet, String, T> mapper;

        private DeserializerChain(@NonNull final BiFunction<ResultSet, String, T> mapper) {
            this.mapper = mapper;
        }

        @NonNull
        public <U> DeserializerChain<U> map(@NonNull final Function<T, U> mapper) {
            return new DeserializerChain<>(this.mapper.andThen(mapper));
        }

        @NonNull
        public <U> Deserializer<U, T> set(@NonNull final BiConsumer<U, ? super T> setter) {
            return new Deserializer<>(mapper, setter);
        }
    }

    public static final class Deserializer<T, U> {

        @NonNull
        private final BiFunction<ResultSet, String, U> mapper;
        @NonNull
        private final BiConsumer<T, ? super U> setter;

        private Deserializer(
            @NonNull final BiFunction<ResultSet, String, U> mapper,
            @NonNull final BiConsumer<T, ? super U> setter) {
            this.mapper = mapper;
            this.setter = setter;
        }
    }

    @NonNull
    public static <T> Deserializer<T, Timestamp> timestamp(@NonNull final BiConsumer<T, Timestamp> setter) {
        return deserializer(ResultSets::getTimestamp).set(setter);
    }

    @NonNull
    public static <T> Deserializer<T, LocalDateTime> timestampToLocalDateTime(@NonNull final BiConsumer<T, LocalDateTime> setter) {
        return deserializer(ResultSets::getTimestamp).map(ts -> ts == null ? null : ts.toLocalDateTime()).set(setter);
    }

    @NonNull
    public static <T> Deserializer<T, String> string(@NonNull final BiConsumer<T, String> setter) {
        return deserializer(ResultSets::getString).set(setter);
    }

    @NonNull
    public static <T> Deserializer<T, Long> longNumber(@NonNull final BiConsumer<T, Long> setter) {
        return deserializer(ResultSets::getLong).set(setter);
    }

    @NonNull
    public static <T> Deserializer<T, Integer> integer(@NonNull final BiConsumer<T, Integer> setter) {
        return deserializer(ResultSets::getInteger).set(setter);
    }

    @NonNull
    public static <T> Deserializer<T, UUID> bytes(@NonNull final BiConsumer<T, UUID> setter) {
        return deserializer(ResultSets::getBytes).map(UUIDs::of).set(setter);
    }
}
