/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.fluent.spring.jdbc.core;

import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.metadata.TableMetaDataContext;
import org.springframework.jdbc.core.simple.AbstractJdbcInsert;
import org.springframework.jdbc.core.simple.SimpleJdbcInsert;
import org.springframework.jdbc.core.simple.SimpleJdbcInsertOperations;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import javax.sql.DataSource;
import java.lang.reflect.Field;

import static java.lang.String.format;
import static java.lang.reflect.Modifier.FINAL;

@SuppressWarnings({"java:S3011"})
public class JdbcInsertIgnore extends SimpleJdbcInsert implements SimpleJdbcInsertOperations {

    private static final Field tableMetaDataContextField;
    private static final String TABLE_META_DATA_CONTEXT_FIELD_NAME = "tableMetaDataContext";

    static {
        try {
            tableMetaDataContextField = AbstractJdbcInsert.class.getDeclaredField(TABLE_META_DATA_CONTEXT_FIELD_NAME);
            final Field modifiers = Field.class.getDeclaredField("modifiers");
            modifiers.setAccessible(true);
            tableMetaDataContextField.setAccessible(true);
            modifiers.set(tableMetaDataContextField, tableMetaDataContextField.getModifiers() & ~FINAL);
        } catch (NoSuchFieldException|IllegalAccessException|SecurityException e) {
            throw new ExceptionInInitializerError(format(
                "Failed to obtain accessible non-final instance of %s %s of %s",
                Field.class.getName(),
                TABLE_META_DATA_CONTEXT_FIELD_NAME,
                AbstractJdbcInsert.class.getName()));
        }
    }

    public JdbcInsertIgnore(@NonNull final DataSource dataSource) {
        super(dataSource);
        setTableMetaDataContext();
    }

    public JdbcInsertIgnore(@NonNull final JdbcTemplate jdbcTemplate) {
        super(jdbcTemplate);
        setTableMetaDataContext();
    }

    private void setTableMetaDataContext() {
        try {
            tableMetaDataContextField.set(this, new InsertIgnoreTableMetadataContext());
        } catch (IllegalAccessException e) {
            throw new ExceptionInInitializerError(format(
                "Failed to set value of %s %s for instance %s",
                Field.class.getName(),
                TABLE_META_DATA_CONTEXT_FIELD_NAME,
                this));
        }
    }

    private static class InsertIgnoreTableMetadataContext extends TableMetaDataContext {
        @Override
        @NonNull
        public String createInsertString(@NonNull String... generatedKeyNames) {
            return
                "INSERT IGNORE" +
                super.createInsertString(generatedKeyNames).substring("INSERT".length());
        }
    }

    @Override
    @NonNull
    public JdbcInsertIgnore withTableName(@Nullable final String tableName) {
        setTableName(tableName);
        return this;
    }

    @Override
    @NonNull
    public JdbcInsertIgnore withSchemaName(@Nullable final String schemaName) {
        setSchemaName(schemaName);
        return this;
    }

    @Override
    @NonNull
    public JdbcInsertIgnore withCatalogName(@Nullable final String catalogName) {
        setCatalogName(catalogName);
        return this;
    }
}
