/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.fluent.spring.jdbc.core;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import org.mockito.junit.jupiter.MockitoSettings;
import org.opentest4j.TestAbortedException;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.simple.SimpleJdbcInsert;
import org.springframework.lang.NonNull;

import javax.sql.DataSource;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.Types;
import java.util.concurrent.atomic.AtomicBoolean;

import static java.lang.String.format;
import static org.junit.jupiter.api.Assertions.assertDoesNotThrow;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.when;
import static org.mockito.quality.Strictness.LENIENT;

@ExtendWith(MockitoExtension.class)
@MockitoSettings(strictness = LENIENT)
class JdbcInsertIgnoreTest {

    @NonNull private static final String CATALOG_NAME = "my_catalog";
    @NonNull private static final String SCHEMA_NAME = "my_schema";
    @NonNull private static final String TABLE_NAME = "my_table";
    @NonNull private static final String COLUMN_NAME = "my_column";

    @NonNull private final DataSource dataSource;
    private final AtomicBoolean nextTable;
    private final AtomicBoolean nextColumn;

    JdbcInsertIgnoreTest(@Mock @NonNull final DataSource dataSource,
                         @Mock @NonNull final Connection connection,
                         @Mock @NonNull final DatabaseMetaData metaData,
                         @Mock @NonNull final ResultSet tableResultSet,
                         @Mock @NonNull final ResultSet columnResultSet) {
        this.dataSource = dataSource;
        nextTable = new AtomicBoolean(true);
        nextColumn = new AtomicBoolean(true);
        try {
            when(dataSource.getConnection()).thenReturn(connection);
            when(connection.getMetaData()).thenReturn(metaData);
            when(metaData.getTables(any(), any(), any(), any())).thenReturn(tableResultSet);
            when(tableResultSet.next()).thenAnswer(i -> nextTable.getAndSet(false));
            when(tableResultSet.getString("TABLE_SCHEM")).thenReturn(SCHEMA_NAME);
            when(tableResultSet.getString("TABLE_NAME")).thenReturn(TABLE_NAME);
            when(metaData.getColumns(any(), any(), any(), any())).thenReturn(columnResultSet);
            when(columnResultSet.next()).thenAnswer(i -> nextColumn.getAndSet(false));
            when(columnResultSet.getString("COLUMN_NAME")).thenReturn(COLUMN_NAME);
            when(columnResultSet.getInt("DATA_TYPE")).thenReturn(Types.VARCHAR);
            when(columnResultSet.getBoolean("NULLABLE")).thenReturn(false);
        } catch (Exception e) {
            throw new ExceptionInInitializerError(e);
        }

    }

    @Test
    @DisplayName("Construct new JDBCInsertIgnore(DataSource)")
    void constructViaDataSource() {
        assertDoesNotThrow(() -> new JdbcInsertIgnore(dataSource));
    }

    @Test
    @DisplayName("Construct new JDBCInsertIgnore(JdbcTemplate)")
    void constructViaJdbcTemplate() {
        assertDoesNotThrow(() -> new JdbcInsertIgnore(new JdbcTemplate(dataSource)));
    }

    @Test
    @DisplayName("Check compilation for IGNORE")
    void compileInternal() {
        final JdbcInsertIgnore insertIgnore = construct(dataSource)
            .withCatalogName(CATALOG_NAME)
            .withSchemaName(SCHEMA_NAME)
            .withTableName(TABLE_NAME);
        final SimpleJdbcInsert simpleInsert = new SimpleJdbcInsert(dataSource)
            .withCatalogName(CATALOG_NAME)
            .withSchemaName(SCHEMA_NAME)
            .withTableName(TABLE_NAME);

        insertIgnore.compile();

        nextTable.set(true);
        nextColumn.set(true);

        simpleInsert.compile();

        final String insertIgnoreString = insertIgnore.getInsertString();
        final String simpleInsertString = simpleInsert.getInsertString();

        assertTrue(insertIgnoreString.startsWith("INSERT IGNORE INTO "),
            "statement does not contain 'INSERT IGNORE INTO'");
        assertEquals(
            simpleInsertString.substring("INSERT".length()),
            insertIgnoreString.substring("INSERT IGNORE".length()));
    }

    @NonNull
    private static JdbcInsertIgnore construct(@NonNull final DataSource dataSource) {
        try {
            return new JdbcInsertIgnore(dataSource);
        } catch (ExceptionInInitializerError e) {
            throw new TestAbortedException(format("Failed to construct new instance of %s", JdbcInsertIgnore.class.getName()));
        }
    }
}
