/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs;

import com.google.common.collect.Range;
import com.google.common.truth.Truth;
import dk.sundhedsdatastyrelsen.cbs.annotations.RequiresLocalDB;
import dk.sundhedsdatastyrelsen.cbs.configuration.IntegrationTestConfiguration;
import dk.sundhedsdatastyrelsen.cbs.datacache.DataCacheDriver;
import dk.sundhedsdatastyrelsen.cbs.ddvdriver.DDVDriver;
import dk.sundhedsdatastyrelsen.cbs.kidsdriver.KidsDriver;
import dk.sundhedsdatastyrelsen.cbs.test.TestWatcherLogger;
import dk.sundhedsdatastyrelsen.cbs.util.CwtTestUtils;
import dk.sundhedsdatastyrelsen.cbs.webservice.idws.IdwsUser;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassResponse;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.PassJWT;
import dk.sundhedsdatastyrelsen.coronapass.ws.JobStatus;
import dk.sundhedsdatastyrelsen.coronapass.ws.JobStatusNotification;
import dk.vaccinationsregister.schemas._2013._12._01.DGWSFault;
import org.junit.jupiter.api.*;
import org.junit.jupiter.api.extension.ExtendWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit.jupiter.SpringExtension;
import se.digg.dgc.payload.v1.RecoveryEntry;
import se.digg.dgc.signatures.cwt.Cwt;

import javax.xml.ws.soap.SOAPFaultException;
import java.time.Duration;
import java.time.LocalDateTime;
import java.time.Period;
import java.time.ZoneId;
import java.util.List;

import static dk.sundhedsdatastyrelsen.cbs.TestConstants.MANUFACTURER_HSC;
import static dk.sundhedsdatastyrelsen.cbs.util.CwtTestUtils.getCwtFromToken;
import static dk.sundhedsdatastyrelsen.cbs.webservice.ResponseTester.assertThat;
import static dk.sundhedsdatastyrelsen.cbs.webservice.idws.IdwsUser.IN_NSP_STAMDATA_BUT_NOT_IN_CPR;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.TestInstance.Lifecycle;

@ExtendWith({SpringExtension.class, TestWatcherLogger.class})
@ContextConfiguration(classes = {IntegrationTestConfiguration.class})
@TestInstance(Lifecycle.PER_CLASS)
@DisplayNameGeneration(DisplayNameGenerator.ReplaceUnderscores.class)
class CoronaPassFTCase {
    @Autowired
    private TestConductor testConductor;
    @Autowired
    private KidsDriver kidsDriver;
    @Autowired
    private DDVDriver ddvDriver;
    @Autowired
    private DataCacheDriver dataCacheDriver;

    @Test
    @RequiresLocalDB
    void test_A3_failed_jobs_due_to_integration_error_are_sent_to_NAS() throws InterruptedException {
        final IdwsUser user = IdwsUser.CPR_2401010010;
        kidsDriver.withUser(user)
            .addResult(true, "ASDF", LocalDateTime.now(), ""); //invalid test data (unknown test type 'ASDF')
        dataCacheDriver.withUser(user)
            .invalidateLabCache();

        GetCoronaPassResponse response = testConductor.waitForInvalidationAndGetFutureResult(user);

        assertThat(response).isFutureResult();

        final JobStatusNotification jobStatusNotification = testConductor.fetchDataAndWaitForJobStatusNotification(response.getFutureResult().getJobId());
        assertEquals(JobStatus.FAILED, jobStatusNotification.getJobStatus(), "Unexpected JobStatus for JobId: " + jobStatusNotification.getJobId());
    }

    @Test
    @RequiresLocalDB
    void test_A11_user_not_in_CPR_db() {
        final SOAPFaultException exception = assertThrows(SOAPFaultException.class, () -> testConductor.getCoronaPasses(IN_NSP_STAMDATA_BUT_NOT_IN_CPR));
        Truth.assertThat(exception).hasMessageThat().contains("Person data is missing");
    }

    @Test
    @RequiresLocalDB
    void test_DK_passes_gets_removed_after_positive_test() throws dk.ssi.nsi.xml_schema._2013._01._01.DGWSFault, DGWSFault, InterruptedException {
        final IdwsUser user = IdwsUser.CPR_2401010018;
        testConductor.clearDdvAndWaitForInvalidationAndDataFetch(user);
        kidsDriver
            .withUser(user)
            .clearKids()
            .addNegativePcr(LocalDateTime.now().minusDays(1));
        dataCacheDriver
            .withUser(user)
            .invalidateLabCache();

        GetCoronaPassResponse response = testConductor.getCoronaPassesAfterInvalidation(user);
        List<PassJWT> passList = response.getCoronaPasses().getPass();
        testConductor.validateTokensAreValidAndFresh(passList);

        assertThat(response)
            .hasDkPasses()
            .hasEuVacPasses(0)
            .hasEuLabPasses(1)
            .hasEuRecPasses(0);

        kidsDriver
            .withUser(user)
            .addPositiveAg(LocalDateTime.now().minusHours(2), MANUFACTURER_HSC);

        dataCacheDriver
            .withUser(user)
            .invalidateLabCache();

        response = testConductor.getCoronaPassesAfterInvalidation(user);
        assertThat(response)
            .hasNoDkPasses()
            .hasEuVacPasses(0)
            .hasEuLabPasses(1)
            .hasEuRecPasses(0);

        testConductor.validateTokensAreValidAndFresh(response.getCoronaPasses().getPass());
    }

    @Test
    @DisplayName("CR-194.1 - PCR with longer validity")
    @RequiresLocalDB
    void longestValidTestDeterminesCertificateValidity() throws Exception {
        final IdwsUser user = IdwsUser.CPR_2401010010;

        ddvDriver.withCitizen(user).clear();

        dataCacheDriver.withUser(user)
            .invalidateLabCache()
            .invalidateVacCache();

        final LocalDateTime time = LocalDateTime.now();
        final LocalDateTime test1SampleTime = time.minusHours(95);
        final LocalDateTime test2SampleTime = time.minusHours(71).minusMinutes(30);

        kidsDriver.withUser(user).clearKids()
            .addNegativePcr(test1SampleTime)
            .addNegativeAg(test2SampleTime, MANUFACTURER_HSC);

        final GetCoronaPassResponse coronaPasses = testConductor.getCoronaPassesAfterInvalidation(user);

        assertThat(coronaPasses)
            .hasDkPasses()
            .hasEuLabPasses(2);

        final PassJWT dkPlusPass = CwtTestUtils.getDKPlusPass(coronaPasses);

        final Cwt dkCwt = getCwtFromToken(dkPlusPass.getJWT().substring(4));

        final Duration expiresIn = Duration.between(time.atZone(ZoneId.of("Europe/Copenhagen")).toInstant(), dkCwt.getExpiration());

        Truth.assertThat(expiresIn)
            .isIn(Range.closed(Duration.ofHours(1).minusSeconds(60L), Duration.ofHours(1)));
    }

    @Test
    @DisplayName("CR-194.3 - PCR with longer validity determines token expiry")
    @RequiresLocalDB
    void longestValidTestDeterminesCertificateValidityPCRLongest() throws Exception {
        final IdwsUser user = IdwsUser.CPR_2401010010;

        ddvDriver.withCitizen(user).clear();

        dataCacheDriver.withUser(user)
            .invalidateLabCache()
            .invalidateVacCache();

        final LocalDateTime now = LocalDateTime.now();
        final LocalDateTime test1SampleTime = now.minus(Duration.ofHours(95L).plusMinutes(15L)); // valid for 45 minutes
        final LocalDateTime test2SampleTime = now.minus(Duration.ofHours(71L).plusMinutes(30L)); // valid for 30 minutes

        kidsDriver.withUser(user).clearKids()
            .addNegativePcr(test1SampleTime)
            .addNegativeAg(test2SampleTime, MANUFACTURER_HSC);

        final GetCoronaPassResponse coronaPasses = testConductor.getCoronaPassesAfterInvalidation(user);

        assertThat(coronaPasses)
            .hasDkPasses()
            .hasEuLabPasses(2);

        final PassJWT dkPlusPass = CwtTestUtils.getDKPlusPass(coronaPasses);

        final Cwt cwt = CwtTestUtils.getCwtFromToken(dkPlusPass.getJWT().substring(4));

        final Duration expiresIn = Duration.between(now.atZone(ZoneId.of("Europe/Copenhagen")).toInstant(), cwt.getExpiration());

        Truth.assertThat(expiresIn)
            .isIn(Range.closed(Duration.ofMinutes(44), Duration.ofMinutes(45)));
    }

    @Test
    @DisplayName("CR-194.4 - AG test with longer validity determines token expiry")
    @RequiresLocalDB
    void longestValidTestDeterminesCertificateValidityAgLongest() throws Exception {
        final IdwsUser user = IdwsUser.CPR_2401010010;

        ddvDriver.withCitizen(user).clear();

        dataCacheDriver.withUser(user)
            .invalidateLabCache()
            .invalidateVacCache();

        final LocalDateTime time = LocalDateTime.now();
        final LocalDateTime test1SampleTime = time.minus(Duration.ofHours(95L).plusMinutes(55L)); // valid for 5 minutes
        final LocalDateTime test2SampleTime = time.minus(Duration.ofHours(71L).plusMinutes(20L)); // valid for 40 minutes

        kidsDriver.withUser(user).clearKids()
            .addNegativePcr(test1SampleTime)
            .addNegativeAg(test2SampleTime, MANUFACTURER_HSC);

        final GetCoronaPassResponse coronaPasses = testConductor.getCoronaPassesAfterInvalidation(user);

        assertThat(coronaPasses)
            .hasDkPasses()
            .hasEuLabPasses(2);

        testConductor.getEuTestEntriesOrderedBySampleTime(CwtTestUtils.getEuLabPasses(coronaPasses));

        final PassJWT dkPlusPass = CwtTestUtils.getDKPlusPass(coronaPasses);

        final Cwt cwt = CwtTestUtils.getCwtFromToken(dkPlusPass.getJWT().substring(4));

        final Duration expiresIn = Duration.between(time.atZone(ZoneId.of("Europe/Copenhagen")).toInstant(), cwt.getExpiration());

        Truth.assertThat(expiresIn)
            .isIn(Range.closed(Duration.ofMinutes(39), Duration.ofMinutes(40)));
    }

    final Duration TEST_OFFSET = Duration.ofMinutes(1);
    final Period TEST_DISPLAY_MAX = Period.ofWeeks(1);

    @Test
    @DisplayName("2.1.4 - Test certificate 7 days plus 1 minute after test")
    @RequiresLocalDB
    void case_2_1_4() throws InterruptedException {
        final LocalDateTime test1SampleTime = LocalDateTime.now().minus(TEST_DISPLAY_MAX).minus(TEST_OFFSET);

        final IdwsUser user = IdwsUser.CPR_2401010010;

        dataCacheDriver.withUser(user).invalidateLabCache();

        kidsDriver.withUser(user).clearKids()
            .addNegativePcr(test1SampleTime);

        final GetCoronaPassResponse coronaPasses = testConductor.getCoronaPassesAfterInvalidation(user);

        assertThat(coronaPasses).hasNoEuLabPasses();
    }

    final Period RECOVERY_VALID_START = Period.ofDays(11);
    final Period RECOVERY_VALID_END = Period.ofDays(180);
    final Duration RECOVERY_OFFSET = Duration.ofMinutes(1L);

    @Test
    @DisplayName("3.1.3 - Recovery after positive PCR, other valid time")
    @RequiresLocalDB
    void case_3_1_3() throws InterruptedException {

        final IdwsUser user = IdwsUser.CPR_2401010010;
        final LocalDateTime test1SampleTime = LocalDateTime.now().minus(RECOVERY_VALID_START).minus(RECOVERY_OFFSET);

        dataCacheDriver.withUser(user).invalidateLabCache();

        kidsDriver.withUser(user)
            .clearKids()
            .addPositivePcr(test1SampleTime);

        final GetCoronaPassResponse coronaPasses = testConductor.getCoronaPassesAfterInvalidation(user);

        final List<PassJWT> euRecPasses = CwtTestUtils.getEuRecPasses(coronaPasses);
        assertEquals(1, euRecPasses.size());

        final RecoveryEntry recoveryEntry = testConductor.getRecoveryEntry(euRecPasses);

        assertEquals(test1SampleTime.toLocalDate(), recoveryEntry.getFr());
    }

    @Test
    @DisplayName("3.1.4 - Recovery after positive PCR, last valid time")
    @RequiresLocalDB
    void case_3_1_4() throws InterruptedException {
        final IdwsUser user = IdwsUser.CPR_2401010010;
        final LocalDateTime test1SampleTime = LocalDateTime.now().minus(RECOVERY_VALID_END).plus(RECOVERY_OFFSET);

        dataCacheDriver.withUser(user).invalidateLabCache();

        kidsDriver.withUser(user)
            .clearKids()
            .addPositivePcr(test1SampleTime);

        final GetCoronaPassResponse coronaPasses = testConductor.getCoronaPassesAfterInvalidation(user);

        final List<PassJWT> euRecPasses = CwtTestUtils.getEuRecPasses(coronaPasses);
        assertEquals(1, euRecPasses.size());

        final RecoveryEntry recoveryEntry = testConductor.getRecoveryEntry(euRecPasses);

        assertEquals(test1SampleTime.toLocalDate(), recoveryEntry.getFr());
    }

}
