/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs;

import com.google.common.truth.Truth;
import dk.sundhedsdatastyrelsen.cbs.annotations.RequiresEmbeddedChildPassesEnabled;
import dk.sundhedsdatastyrelsen.cbs.cbsdriver.CBSDGWSDriver;
import dk.sundhedsdatastyrelsen.cbs.cbsdriver.CBSDriver;
import dk.sundhedsdatastyrelsen.cbs.configuration.IntegrationTestConfiguration;
import dk.sundhedsdatastyrelsen.cbs.ddvdriver.DDVDriver;
import dk.sundhedsdatastyrelsen.cbs.ddvdriver.DDVDriver.Citizen;
import dk.sundhedsdatastyrelsen.cbs.test.TestWatcherLogger;
import dk.sundhedsdatastyrelsen.cbs.util.CoronaPassAssertions;
import dk.sundhedsdatastyrelsen.cbs.util.CwtTestUtils;
import dk.sundhedsdatastyrelsen.cbs.webservice.ResponseTester;
import dk.sundhedsdatastyrelsen.cbs.webservice.idws.IdwsUser;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassResponse;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.PassJWT;
import dk.sundhedsdatastyrelsen.coronapass.ws.JobStatus;
import dk.sundhedsdatastyrelsen.coronapass.ws.JobStatusNotification;
import dk.vaccinationsregister.schemas._2013._12._01.DGWSFault;
import org.apache.log4j.Logger;
import org.awaitility.Awaitility;
import org.junit.jupiter.api.*;
import org.junit.jupiter.api.TestInstance.Lifecycle;
import org.junit.jupiter.api.extension.ExtendWith;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.opentest4j.AssertionFailedError;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit.jupiter.EnabledIf;
import org.springframework.test.context.junit.jupiter.SpringExtension;
import se.digg.dgc.payload.v1.DigitalGreenCertificate;
import se.digg.dgc.payload.v1.RecoveryEntry;
import se.digg.dgc.payload.v1.TestEntry;
import se.digg.dgc.payload.v1.VaccinationEntry;
import se.digg.dgc.service.DGCDecoder;

import javax.xml.ws.soap.SOAPFaultException;
import java.time.*;
import java.time.temporal.ChronoUnit;
import java.time.temporal.TemporalAmount;
import java.util.*;
import java.util.stream.Collectors;

import static dk.sundhedsdatastyrelsen.cbs.TestConstants.MANUFACTURER_CE;
import static dk.sundhedsdatastyrelsen.cbs.TestConstants.MANUFACTURER_HSC;
import static dk.sundhedsdatastyrelsen.cbs.test.AssertionTestUtil.waitForAsserted;
import static dk.sundhedsdatastyrelsen.cbs.util.LabResultTester.assertThatTestEntry;
import static dk.sundhedsdatastyrelsen.cbs.util.RecoveryTester.assertThatRecoveryEntry;
import static dk.sundhedsdatastyrelsen.cbs.util.VaccinationTester.assertThatVaccination;
import static dk.sundhedsdatastyrelsen.cbs.webservice.ResponseTester.assertThat;
import static org.junit.jupiter.api.Assertions.*;

@ExtendWith({SpringExtension.class, TestWatcherLogger.class})
@ContextConfiguration(classes = {IntegrationTestConfiguration.class})
@TestInstance(Lifecycle.PER_CLASS)
@DisplayNameGeneration(DisplayNameGenerator.ReplaceUnderscores.class)
class CoronaPassITCase {
    private static final Logger logger = Logger.getLogger(CoronaPassITCase.class);
    private static final Period RECOVERY_VALID_START = Period.ofDays(11);
    private static final Period RECOVERY_VALID_END = Period.ofDays(180);

    @Autowired
    private TestConductor conductor;
    @Autowired
    private CBSDriver cbsDriver;
    @Autowired
    private DGCDecoder decoder;
    @Autowired
    private CBSDGWSDriver cbsdgwsDriver;
    @Autowired
    private DDVDriver ddvDriver;
    private Instant now;

    private final Queue<IdwsUser> userQueue = new LinkedList<>();

    @BeforeAll
    public void setupAll() {
        final List<IdwsUser> idwsUsers = Arrays.asList(
            IdwsUser.CPR_2401010011,
            IdwsUser.CPR_2401010012,
            IdwsUser.CPR_2401010013,
            IdwsUser.CPR_2401010014,
            IdwsUser.CPR_2401010015,
            IdwsUser.CPR_2401010018
        );
        Collections.shuffle(idwsUsers);
        userQueue.addAll(idwsUsers);
    }

    private IdwsUser nextUser() {
        final IdwsUser user = userQueue.poll();
        userQueue.offer(user);
        logger.info("Using IdwsUser: " + user);
        return user;
    }

    @BeforeEach
    void setUp() {
        now = Instant.now();
        conductor.setNow(now);
    }

    @Nested
    @DisplayName("CBS Acceptance Tests based on https://www.nspop.dk/display/CR/User+Stories+og+Scenarier")
    class AcceptanceTests {
        @Test
        void test_A1_when_cache_is_empty_future_is_returned() throws DGWSFault, InterruptedException {
            final IdwsUser user = IdwsUser.NO_LAB_RESULTS;
            ddvDriver.withCitizen(user).addJohnsonVaccination(new Date());

            final GetCoronaPassResponse response = conductor.waitForInvalidationAndGetFutureResult(user);
            assertNull(response.getCoronaPasses());
            assertNotNull(response.getFutureResult());
        }

        @Test
        void test_A2_repeated_calls_returns_freshly_issued_tokens() throws InterruptedException {
            final IdwsUser user = IdwsUser.NC_CPR_0405870106;
            GetCoronaPassResponse response = conductor.getCoronaPasses(user);
            assertThat(response).hasDkPasses();
            final List<PassJWT> passList = response.getCoronaPasses().getPass();
            conductor.validateTokensAreValidAndFresh(passList);

            Awaitility.waitAtMost(CoronaPassAssertions.TOKEN_ISSUED_AT_SLACK.plusSeconds(10L))
                .untilAsserted(() -> assertThrows(AssertionFailedError.class, () -> conductor.validateTokensAreValidAndFresh(passList)));

            response = conductor.getCoronaPasses(user);
            assertThat(response).hasDkPasses();
            conductor.validateTokensAreValidAndFresh(response.getCoronaPasses().getPass());
        }

        @Test
        void test_A5_user_with_only_test_result() throws Exception {
            GetCoronaPassResponse response = conductor.getCoronaPasses(IdwsUser.NC_CPR_2009171897);

            List<PassJWT> passList = response.getCoronaPasses().getPass();
            conductor.validateTokensAreValidAndFresh(passList);

            assertThat(response)
                .hasDkPasses()
                .hasEuVacPasses(0)
                .hasEuLabPasses(1)
                .hasEuRecPasses(0);
        }

        @Test
        void test_A6_user_with_only_vaccination() throws Exception {
            final IdwsUser user = IdwsUser.NO_LAB_RESULTS;
            Citizen citizen = ddvDriver.withCitizen(user).clear();
            citizen.addPfizerVaccination(Date.from(now.minus(48, ChronoUnit.DAYS)));
            citizen.addPfizerVaccination(Date.from(now.minus(14, ChronoUnit.DAYS)));

            GetCoronaPassResponse response = conductor.getCoronaPassesAfterInvalidation(user);

            assertThat(response)
                .hasDkPasses()
                .hasEuVacPasses(2)
                .hasEuLabPasses(0)
                .hasEuRecPasses(0);
            conductor.validateTokensAreValidAndFresh(response.getCoronaPasses().getPass());
        }


        @Test
        void test_A7_user_with_only_immunity() throws Exception {
            GetCoronaPassResponse response = conductor.getCoronaPasses(IdwsUser.NC_CPR_1801293574);
            List<PassJWT> passList = response.getCoronaPasses().getPass();
            conductor.validateTokensAreValidAndFresh(passList);

            assertThat(response)
                .hasDkPasses()
                .hasEuVacPasses(0)
                .hasEuLabPasses(0)
                .hasEuRecPasses(1);
        }

        @Test
        void test_A8_user_with_too_early_vac_gets_no_DK_passes() throws Exception {
            final IdwsUser user = IdwsUser.NO_LAB_RESULTS;
            ddvDriver
                .withCitizen(user).clear()
                .addPfizerVaccination(Date.from(now.minus(1, ChronoUnit.DAYS)));
            GetCoronaPassResponse response = conductor.getCoronaPassesAfterInvalidation(user);

            assertThat(response)
                .hasNoDkPasses()
                .hasEuVacPasses(1)
                .hasEuLabPasses(0)
                .hasEuRecPasses(0);
        }

        @Test
        void test_A8_user_with_positive_test_gets_no_passes() throws Exception {
            GetCoronaPassResponse response = conductor.getCoronaPasses(IdwsUser.NC_CPR_2310901374);

            assertThat(response)
                .hasNoDkPasses()
                .hasEuVacPasses(0)
                .hasEuLabPasses(0)
                .hasEuRecPasses(0);
        }

        @Test
        void test_A8_user_with_expired_immunity_gets_no_passes() throws InterruptedException {
            GetCoronaPassResponse response = conductor.getCoronaPasses(IdwsUser.NC_CPR_0208193805);

            assertThat(response)
                .hasNoDkPasses()
                .hasEuVacPasses(0)
                .hasEuLabPasses(0)
                .hasEuRecPasses(0);
        }

        @Test
        void test_A9_user_with_no_tests_or_vaccinations_gets_no_passes_returned() throws Exception {
            conductor.clearDdvAndWaitForInvalidationAndDataFetch(IdwsUser.NO_LAB_RESULTS);
            GetCoronaPassResponse response = conductor.getCoronaPasses(IdwsUser.NO_LAB_RESULTS);

            assertThat(response)
                .hasNoDkPasses()
                .hasEuVacPasses(0)
                .hasEuLabPasses(0)
                .hasEuRecPasses(0);
        }

        @Test
        void test_A10_CPRMismatch() {
            // Create cpr mismatch:
            assertThrows(SOAPFaultException.class, () -> cbsDriver.getCoronaPass(IdwsUser.CPR_2401010010, "1111111118"));
        }

        @Test
        @Disabled("Not reliably triggering error")
        void test_A12_failed_jobs_due_to_internal_error_are_sent_to_NAS() throws InterruptedException, dk.ssi.nsi.xml_schema._2013._01._01.DGWSFault, DGWSFault {
            conductor.clearDdvAndWaitForInvalidationAndDataFetch(IdwsUser.NO_LAB_RESULTS);
            //trigger CR-198 ("LabResult ETL FAILED" due to primary key violation)
            for (int i = 0; i < 10; i++) {
                final GetCoronaPassResponse futureResult = cbsDriver.getCoronaPass(IdwsUser.NO_LAB_RESULTS);
                assertNotNull(futureResult.getFutureResult());
            }
            final GetCoronaPassResponse futureResult = cbsDriver.getCoronaPass(IdwsUser.NO_LAB_RESULTS);
            assertNotNull(futureResult.getFutureResult());
            final JobStatusNotification jobStatusNotification = conductor.fetchDataAndWaitForJobStatusNotification(futureResult.getFutureResult().getJobId());
            assertEquals(JobStatus.FAILED, jobStatusNotification.getJobStatus());
        }

        @Test
        void test_passport_with_all_tokens() throws InterruptedException {
            final IdwsUser user = IdwsUser.NC_CPR_0405870101;

            GetCoronaPassResponse response = conductor.getCoronaPasses(user);
            List<PassJWT> passList = response.getCoronaPasses().getPass();
            conductor.validateTokensAreValidAndFresh(passList);

            assertThat(response)
                .hasDkPasses()
                .hasEuVacPasses(2)
                .hasEuLabPasses(1)
                .hasEuRecPasses(1);
        }

        @Test
        void test_parent_requests_child_passport() throws Exception {
            final IdwsUser parent = IdwsUser.PARENT_A;
            final IdwsUser child = IdwsUser.CHILD_OF_A;

            Citizen citizen = ddvDriver.withCitizen(child).clear();
            citizen.addPfizerVaccination(Date.from(now.minus(48, ChronoUnit.DAYS)));
            citizen.addPfizerVaccination(Date.from(now.minus(14, ChronoUnit.DAYS)));

            GetCoronaPassResponse response = conductor.waitForInvalidationAndGetFutureResult(parent, child);
            assertNull(response.getCoronaPasses());
            assertNotNull(response.getFutureResult());

            String jobId = response.getFutureResult().getJobId();
            final JobStatusNotification jobStatusNotification = conductor.fetchDataAndWaitForJobStatusNotification(jobId);
            assertEquals(JobStatus.OK, jobStatusNotification.getJobStatus());
            GetCoronaPassResponse passport = cbsDriver.getCoronaPass(parent, child.cpr());

            ResponseTester.assertThat(passport)
                .hasDkPasses()
                .hasEuVacPasses(2);
            conductor.validateTokensAreValidAndFresh(passport.getCoronaPasses().getPass());
        }

        @Test
        @DisplayName("Get passport for user with EU-valid test data in little KIDS")
        void checkEUPassForEUValidTest() throws InterruptedException {
            final GetCoronaPassResponse coronaPasses = conductor.getCoronaPasses(IdwsUser.NC_CPR_2009171897);
            assertThat(coronaPasses)
                .hasDkPasses()
                .hasNoEuVacPasses()
                .hasNoEuRecPass()
                .hasEuLabPasses(1);
        }

        @Test
        @DisplayName("Get passport for user with EU-invalid test data in little KIDS cases no EU test passport")
        void checkNoEUPassForEUInvalidTest() throws InterruptedException {
            final GetCoronaPassResponse coronaPasses = conductor.getCoronaPasses(IdwsUser.NC_CPR_0405870106);
            assertThat(coronaPasses)
                .hasDkPasses()
                .hasNoEuVacPasses()
                .hasNoEuRecPass()
                .hasNoEuLabPasses();
        }

        @Test
        @DisplayName("CR-182 - All Certificates (except DK) contains legal name of person")
        void checkCorrectNameInPasses() throws Exception {
            final IdwsUser user = IdwsUser.NC_CPR_0405870101;

            GetCoronaPassResponse response = conductor.getCoronaPasses(user);
            List<PassJWT> passList = response.getCoronaPasses().getPass();
            conductor.validateTokensAreValidAndFresh(passList);

            assertThat(response, decoder)
                .hasDkPasses()
                .hasDkInfoWithName(user.commonName())
                .hasEuVacPasses(2)
                .hasEuLabPasses(1)
                .hasEuRecPasses(1)
                .allCwtsWithNameHasName(user.commonName());

        }

        @Test
        @DisplayName("CR-182 - DKINFO Pass contains legal name of person")
        void checkCorrectNameInDKINFO() throws Exception {
            final IdwsUser user = IdwsUser.NC_CPR_0510740700;
            GetCoronaPassResponse response = conductor.getCoronaPasses(user);

            assertThat(response, decoder)
                .hasDkInfoWithName(user.commonName());
        }

        @Test
        @DisplayName("CR-194.2 - No DK pass if test certificates are too old")
        void expiredTests() throws Exception {
            IdwsUser user = IdwsUser.NC_CPR_1703501676;
            final GetCoronaPassResponse coronaPasses = conductor.getCoronaPasses(user);
            assertThat(coronaPasses)
                .hasNoDkPasses()
                .hasEuLabPasses(1);
        }
    }

    @Nested
    @DisplayName("EU Vaccination Corona Pass Rules")
    class VaccinationTests {
        final Duration VAC_OFFSET = Duration.ofMinutes(1);
        final Duration VAC_VALID_START = Duration.ofDays(14);
        final Period VAC_VALID_END_COMPLETED = Period.ofMonths(7);
        final Duration VAC_VALID_END_FIRST_EFFECTUATION = Duration.ofDays(42);

        @Test
        void gets_country_code_from_ddv() throws DGWSFault, dk.ssi.nsi.xml_schema._2013._01._01.DGWSFault, InterruptedException {
            final IdwsUser user = nextUser();
            final Instant effectuationTime = now.minus(VAC_VALID_START).plus(Duration.ofDays(5));

            final String greenland = "GL";
            ddvDriver.withCitizen(user)
                .clear()
                .addJohnsonVaccinationEffectuatedInCountry(greenland, Date.from(effectuationTime));

            final GetCoronaPassResponse coronaPass = conductor.getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> vaccinationPasses = CwtTestUtils.getEuVacPasses(coronaPass);
            assertEquals(1, vaccinationPasses.size());

            final List<VaccinationEntry> vaccinationEntries = conductor.getEuVaccinationEntriesOrderedByDate(vaccinationPasses);

            assertThatVaccination(vaccinationEntries.get(0))
                .hasCountryCode(greenland)
                .isDoseOf(1, 1)
                .hasDateOfVaccination(effectuationTime)
                .isJohnsonVaccination();

        }

        @Test
        @DisplayName("1.1.1 - First effectuation with AstraZeneca at first valid time")
        void case_1_1_1() throws Exception {
            final IdwsUser user = nextUser();

            final Instant effectuationTime = now.minus(VAC_OFFSET);

            ddvDriver.withCitizen(user)
                .clear()
                .addAstraZenecaVaccination(Date.from(effectuationTime));

            final GetCoronaPassResponse coronaPass = conductor.getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> vaccinationPasses = CwtTestUtils.getEuVacPasses(coronaPass);

            assertEquals(1, vaccinationPasses.size());

            final List<VaccinationEntry> vaccinationEntries = conductor.getEuVaccinationEntriesOrderedByDate(vaccinationPasses);

            assertThatVaccination(vaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(effectuationTime)
                .isAstraZenicaVaccination();
        }

        @Test
        @DisplayName("1.1.2 - First effectuation with AstraZeneca at latest valid time")
        void case_1_1_2() throws Exception {
            final IdwsUser user = nextUser();

            final Instant effectuationTime = now.minus(VAC_VALID_END_FIRST_EFFECTUATION.minus(VAC_OFFSET));

            ddvDriver.withCitizen(user)
                .clear()
                .addAstraZenecaVaccination(Date.from(effectuationTime));

            final GetCoronaPassResponse coronaPass = conductor.getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> vaccinationPasses = CwtTestUtils.getEuVacPasses(coronaPass);

            assertEquals(1, vaccinationPasses.size());

            final List<VaccinationEntry> vaccinationEntries = conductor.getEuVaccinationEntriesOrderedByDate(vaccinationPasses);

            assertThatVaccination(vaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(effectuationTime)
                .isAstraZenicaVaccination();
        }

        @Test
        @DisplayName("1.1.3 - First effectuation with AstraZeneca at latest valid time threshold ")
        void case_1_1_3() throws Exception {
            final IdwsUser user = nextUser();

            final Instant effectuationTime = now.minus(VAC_VALID_END_FIRST_EFFECTUATION);

            ddvDriver.withCitizen(user)
                .clear()
                .addAstraZenecaVaccination(Date.from(effectuationTime));

            final GetCoronaPassResponse coronaPass = conductor.getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> vaccinationPasses = CwtTestUtils.getEuVacPasses(coronaPass);

            assertEquals(1, vaccinationPasses.size());

            final List<VaccinationEntry> vaccinationEntries = conductor.getEuVaccinationEntriesOrderedByDate(vaccinationPasses);

            assertThatVaccination(vaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(effectuationTime)
                .isAstraZenicaVaccination();
        }

        @Test
        @DisplayName("1.1.4 - First effectuation with AstraZeneca at latest valid time + 1 minute")
        void case_1_1_4() throws Exception {
            final IdwsUser user = nextUser();

            final Instant effectuationTime = now.minus(VAC_VALID_END_FIRST_EFFECTUATION.plus(VAC_OFFSET));

            ddvDriver.withCitizen(user)
                .clear()
                .addAstraZenecaVaccination(Date.from(effectuationTime));

            final GetCoronaPassResponse coronaPass = conductor.getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> vaccinationPasses = CwtTestUtils.getEuVacPasses(coronaPass);

            assertEquals(1, vaccinationPasses.size());

            final List<VaccinationEntry> vaccinationEntries = conductor.getEuVaccinationEntriesOrderedByDate(vaccinationPasses);

            assertThatVaccination(vaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(effectuationTime)
                .isAstraZenicaVaccination();
        }

        @Test
        @DisplayName("1.2.1 - Completed Vaccination with Johnson & Johnson at first valid time")
        void case_1_2_1() throws Exception {
            final IdwsUser user = nextUser();

            final Instant effectuationTime = now.minus(VAC_VALID_START);

            ddvDriver.withCitizen(user)
                .clear()
                .addJohnsonVaccination(Date.from(effectuationTime));

            final GetCoronaPassResponse coronaPass = conductor.getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> vaccinationPasses = CwtTestUtils.getEuVacPasses(coronaPass);

            assertEquals(1, vaccinationPasses.size());

            final List<VaccinationEntry> vaccinationEntries = conductor.getEuVaccinationEntriesOrderedByDate(vaccinationPasses);

            assertThatVaccination(vaccinationEntries.get(0))
                .isDoseOf(1, 1)
                .hasDateOfVaccination(effectuationTime)
                .isJohnsonVaccination();
        }

        @Test
        @DisplayName("1.2.2 - Completed Vaccination with Johnson & Johnson at expiration time")
        void case_1_2_2() throws Exception {
            final IdwsUser user = IdwsUser.NO_LAB_RESULTS;

            final Instant effectuationTime = now.atZone(ZoneId.of("Europe/Copenhagen")).minus(VAC_VALID_END_COMPLETED).minusDays(1).toInstant();

            ddvDriver.withCitizen(user)
                .clear()
                .addJohnsonVaccination(Date.from(effectuationTime));

            final GetCoronaPassResponse coronaPass = conductor.getCoronaPassesAfterInvalidation(user);

            assertThat(coronaPass)
                .hasNoDkPasses()
                .hasEuVacPasses(1);

            final List<PassJWT> vaccinationPasses = CwtTestUtils.getEuVacPasses(coronaPass);

            assertEquals(1, vaccinationPasses.size());

            final List<VaccinationEntry> vaccinationEntries = conductor.getEuVaccinationEntriesOrderedByDate(vaccinationPasses);

            assertThatVaccination(vaccinationEntries.get(0))
                .isDoseOf(1, 1)
                .hasDateOfVaccination(effectuationTime)
                .isJohnsonVaccination();
        }

        @Test
        @DisplayName("1.3.1 - Completed Vaccination with AstraZeneca at first valid time")
        void case_1_3_1() throws Exception {
            final IdwsUser user = nextUser();

            final Instant firstEffectuation = now.minus(VAC_VALID_END_FIRST_EFFECTUATION);
            final Instant secondEffectuation = now.minus(VAC_VALID_START);

            final Citizen citizen = ddvDriver.withCitizen(user)
                .clear();

            citizen.addAstraZenecaVaccination(Date.from(firstEffectuation));
            citizen.addAstraZenecaVaccination(Date.from(secondEffectuation));

            final GetCoronaPassResponse coronaPass = conductor.getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> vaccinationPasses = CwtTestUtils.getEuVacPasses(coronaPass);

            assertEquals(2, vaccinationPasses.size());

            final List<VaccinationEntry> vaccinationEntries = conductor.getEuVaccinationEntriesOrderedByDate(vaccinationPasses);

            assertThatVaccination(vaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(firstEffectuation)
                .isAstraZenicaVaccination();

            assertThatVaccination(vaccinationEntries.get(1))
                .isDoseOf(2, 2)
                .hasDateOfVaccination(secondEffectuation)
                .isAstraZenicaVaccination();
        }

        @Test
        @DisplayName("1.4.1 - Completed Vaccination with AstraZeneca and Moderna")
        void case_1_4_1() throws Exception {
            final IdwsUser user = nextUser();

            final Instant firstEffectuation = now.minus(VAC_VALID_END_FIRST_EFFECTUATION);
            final Instant secondEffectuation = now.minus(VAC_VALID_START);

            final Citizen citizen = ddvDriver.withCitizen(user)
                .clear();

            citizen.addAstraZenecaVaccination(Date.from(firstEffectuation));
            citizen.addModernaVaccination(Date.from(secondEffectuation));

            final GetCoronaPassResponse coronaPass = conductor.getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> vaccinationPasses = CwtTestUtils.getEuVacPasses(coronaPass);

            assertEquals(2, vaccinationPasses.size());

            final List<VaccinationEntry> vaccinationEntries = conductor.getEuVaccinationEntriesOrderedByDate(vaccinationPasses);

            assertThatVaccination(vaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(firstEffectuation)
                .isAstraZenicaVaccination();

            assertThatVaccination(vaccinationEntries.get(1))
                .isDoseOf(2, 2)
                .hasDateOfVaccination(secondEffectuation)
                .isModernaVaccination();
        }

        @Test
        @DisplayName("1.5.1 - 2 Vaccinations too close to each other to be considered completed")
        void case_1_5_1() throws Exception {
            final IdwsUser user = nextUser();

            final Instant firstEffectuation = now.minus(VAC_VALID_START.plus(VAC_OFFSET));
            final Instant secondEffectuation = now.minus(VAC_VALID_START);

            final Citizen citizen = ddvDriver.withCitizen(user)
                .clear();

            citizen.addPfizerVaccination(Date.from(firstEffectuation));
            citizen.addPfizerVaccination(Date.from(secondEffectuation));

            final GetCoronaPassResponse coronaPass = conductor.getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> vaccinationPasses = CwtTestUtils.getEuVacPasses(coronaPass);

            assertEquals(2, vaccinationPasses.size());

            final List<VaccinationEntry> vaccinationEntries = conductor.getEuVaccinationEntriesOrderedByDate(vaccinationPasses);

            assertThatVaccination(vaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(firstEffectuation)
                .isPfizerVaccination();

            assertThatVaccination(vaccinationEntries.get(1))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(secondEffectuation)
                .isPfizerVaccination();
        }

        @Test
        @DisplayName("1.5.2 - 2 Vaccinations too far from each other")
        void case_1_5_2() throws Exception {
            final IdwsUser user = nextUser();

            final Instant firstEffectuation = now.atZone(ZoneId.of("Europe/Copenhagen")).minus(VAC_VALID_END_COMPLETED).toInstant();
            final Instant secondEffectuation = now.minus(VAC_OFFSET);

            final Citizen citizen = ddvDriver.withCitizen(user)
                .clear();

            citizen.addPfizerVaccination(Date.from(firstEffectuation));
            citizen.addPfizerVaccination(Date.from(secondEffectuation));

            final GetCoronaPassResponse coronaPass = conductor.getCoronaPassesAfterInvalidation(user);

            final List<PassJWT> vaccinationPasses = CwtTestUtils.getEuVacPasses(coronaPass);

            assertEquals(2, vaccinationPasses.size());

            final List<VaccinationEntry> vaccinationEntries = conductor.getEuVaccinationEntriesOrderedByDate(vaccinationPasses);

            assertThatVaccination(vaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(firstEffectuation)
                .isPfizerVaccination();

            assertThatVaccination(vaccinationEntries.get(1))
                .isDoseOf(2, 2)
                .hasDateOfVaccination(secondEffectuation)
                .isPfizerVaccination();
        }

    }

    @Nested
    @DisplayName("EU Test Corona Pass Rules")
    class LabResultTests {

        @Test
        @DisplayName("2.1.1 - PCR Test certificate after test")
        void case_2_1_1() throws InterruptedException {
            final LocalDateTime test1SampleTime = zeroHourMinus(Duration.ofHours(4));

            final IdwsUser user = IdwsUser.NC_CPR_2011451636;
            final GetCoronaPassResponse coronaPasses = conductor.getCoronaPasses(user);
            final List<PassJWT> euLabPasses = CwtTestUtils.getEuLabPasses(coronaPasses);
            assertEquals(1, euLabPasses.size());

            final List<TestEntry> testEntries = conductor.getEuTestEntriesOrderedBySampleTime(euLabPasses);
            assertEquals(1, testEntries.size());

            final TestEntry firstTest = testEntries.get(0);
            assertThatTestEntry(firstTest)
                .isNegative()
                .isPcr()
                .hasSampleTime(test1SampleTime);
        }

        @Test
        @DisplayName("2.1.2 - PCR Test certificate more than 96 hours old and less than 7 days old")
        void case_2_1_2() throws InterruptedException {
            final LocalDateTime test1SampleTime = zeroHourMinus(Duration.ofHours(107));

            final IdwsUser user = IdwsUser.NC_CPR_1703501676;
            final GetCoronaPassResponse coronaPasses = conductor.getCoronaPasses(user);

            assertThat(coronaPasses)
                .hasNoDkPasses()
                .hasEuLabPasses(1);

            final List<PassJWT> euLabPasses = CwtTestUtils.getEuLabPasses(coronaPasses);
            assertEquals(1, euLabPasses.size());

            final List<TestEntry> testEntries = conductor.getEuTestEntriesOrderedBySampleTime(euLabPasses);
            assertEquals(1, testEntries.size());

            final TestEntry firstTest = testEntries.get(0);
            assertThatTestEntry(firstTest)
                .isNegative()
                .isPcr()
                .hasSampleTime(test1SampleTime);
        }

        @Test
        @DisplayName("2.1.3 - PCR Test certificate more than 7 days after test")
        void case_2_1_3() throws InterruptedException {
            final IdwsUser user = IdwsUser.NC_CPR_1511212961; //has a positive PCR test ~11 days old

            final GetCoronaPassResponse coronaPasses = conductor.getCoronaPasses(user);

            assertThat(coronaPasses)
                .hasNoEuLabPasses();
        }

        @Test
        @DisplayName("2.2.1 - Test certificate for negative PCR and approved AG tests and filtering of unapproved negative AG test and positive PCR and AG tests")
        void case_2_2_1() throws InterruptedException {
            final IdwsUser user = IdwsUser.NC_CPR_3012283023;
            final LocalDateTime test1_negativePcr = zeroHourMinus(Duration.ofHours(4));
//            final LocalDateTime test2_positivePcr = zeroHourMinus(Duration.ofHours(8));
            final LocalDateTime test3_negativePcr = zeroHourMinus(Duration.ofHours(12));
            final LocalDateTime test4_negativeHscAg = zeroHourMinus(Duration.ofHours(16));
//            final LocalDateTime test5_positiveHscAg = zeroHourMinus(Duration.ofHours(20));
//            final LocalDateTime test6_negativeNotHscAg = zeroHourMinus(Duration.ofHours(24));

            final GetCoronaPassResponse coronaPasses = conductor.getCoronaPasses(user);
            final List<PassJWT> euLabPasses = CwtTestUtils.getEuLabPasses(coronaPasses);
            assertEquals(3, euLabPasses.size());
            final List<TestEntry> testEntries = conductor.getEuTestEntriesOrderedBySampleTime(euLabPasses);
            assertEquals(3, testEntries.size());

            final TestEntry firstTest = testEntries.get(0);
            final TestEntry secondTest = testEntries.get(1);
            final TestEntry thirdTest = testEntries.get(2);

            assertThatTestEntry(firstTest)
                .isNegative()
                .isAG()
                .hasSampleTime(test4_negativeHscAg);

            assertThatTestEntry(secondTest)
                .isNegative()
                .isPcr()
                .hasSampleTime(test3_negativePcr);

            assertThatTestEntry(thirdTest)
                .isNegative()
                .isPcr()
                .hasSampleTime(test1_negativePcr);
        }

        @Test
        @DisplayName("2.3.1 - Test certificate approved AG tests manufacturers")
        void case_2_3_1() throws InterruptedException {
            IdwsUser user = IdwsUser.NC_CPR_0301301728;//user has 5 AG with manufacturer (CE / CE / unknown / HSC / HSC), taken (-100 / -90 / -80 / -70 / -60 hours from 10:00 DK time)
            final LocalDateTime test1SampleTime = zeroHourMinus(Duration.ofHours(100));
            final LocalDateTime test2SampleTime = zeroHourMinus(Duration.ofHours(90));
            final LocalDateTime test3SampleTime = zeroHourMinus(Duration.ofHours(80));
            final LocalDateTime test4SampleTime = zeroHourMinus(Duration.ofHours(70));
            final LocalDateTime test5SampleTime = zeroHourMinus(Duration.ofHours(60));

            final GetCoronaPassResponse coronaPasses = conductor.getCoronaPasses(user);
            assertThat(coronaPasses).hasEuLabPasses(2);

            final List<PassJWT> dkInfoPass = CwtTestUtils.getDKInfoPass(coronaPasses);
            Truth.assertThat(dkInfoPass).hasSize(1);
            final DigitalGreenCertificate dkinfoPayload = conductor.getDKINFOPayload(dkInfoPass.get(0).getJWT());
            final List<TestEntry> testEntriesFromDKINFO = CwtTestUtils.getTestEntriesOrderedBySampleTime(dkinfoPayload);

            assertEquals(5, testEntriesFromDKINFO.size());

            assertThatTestEntry(testEntriesFromDKINFO.get(0))
                .isNegative()
                .isAG()
                .hasSampleTime(test1SampleTime)
                .hasManufacturer(MANUFACTURER_CE);

            assertThatTestEntry(testEntriesFromDKINFO.get(1))
                .isNegative()
                .isAG()
                .hasSampleTime(test2SampleTime)
                .hasManufacturer(MANUFACTURER_CE);

            assertThatTestEntry(testEntriesFromDKINFO.get(2))
                .isNegative()
                .isAG()
                .hasSampleTime(test3SampleTime)
                .hasManufacturer(null);

            assertThatTestEntry(testEntriesFromDKINFO.get(3))
                .isNegative()
                .isAG()
                .hasSampleTime(test4SampleTime)
                .hasManufacturer(MANUFACTURER_HSC);

            assertThatTestEntry(testEntriesFromDKINFO.get(4))
                .isNegative()
                .isAG()
                .hasSampleTime(test5SampleTime)
                .hasManufacturer(MANUFACTURER_HSC);

            final List<TestEntry> testEntriesFromEU = conductor.getEuTestEntriesOrderedBySampleTime(CwtTestUtils.getEuLabPasses(coronaPasses));

            assertEquals(2, testEntriesFromEU.size());

            assertThatTestEntry(testEntriesFromEU.get(0))
                .isNegative()
                .isAG()
                .hasSampleTime(test4SampleTime)
                .hasManufacturer(MANUFACTURER_HSC);

            assertThatTestEntry(testEntriesFromEU.get(1))
                .isNegative()
                .isAG()
                .hasSampleTime(test5SampleTime)
                .hasManufacturer(MANUFACTURER_HSC);
        }
    }

    @Nested
    @DisplayName("EU Recovery Corona Pass Rules")
    class RecoveryTests {
        @Test
        @DisplayName("3.1.1 - Recovery after positive PCR, invalid time")
        void case_3_1_1() throws InterruptedException {
            final IdwsUser user = IdwsUser.NC_CPR_1704080278; //has positive PCR 4 hours old from 10:00 DK time
            final GetCoronaPassResponse coronaPasses = conductor.getCoronaPasses(user);
            assertThat(coronaPasses).hasNoEuRecPass();
        }

        @Test
        @DisplayName("3.1.2 - Recovery after positive PCR, first valid time")
        void case_3_1_2() throws InterruptedException {
            final IdwsUser user = IdwsUser.NC_CPR_1301071035;
            final LocalDate test1SampleDate = zeroHourMinus(Duration.ofHours(275)).toLocalDate();

            final GetCoronaPassResponse coronaPasses = conductor.getCoronaPasses(user);
            final List<PassJWT> euRecPasses = CwtTestUtils.getEuRecPasses(coronaPasses);
            assertEquals(1, euRecPasses.size());
            final RecoveryEntry recoveryEntry = conductor.getRecoveryEntry(euRecPasses);

            assertThatRecoveryEntry(recoveryEntry)
                .hasSampleDate(test1SampleDate)
                .hasStartDate(test1SampleDate.plus(RECOVERY_VALID_START))
                .hasEndDate(test1SampleDate.plus(RECOVERY_VALID_END));
        }

        @Test
        @DisplayName("3.1.5 - Recovery after positive PCR, first invalid time")
        void case_3_1_5() throws InterruptedException {
            final IdwsUser user = IdwsUser.NC_CPR_1403993158; //has positive PCR -4331 hours from 10:00 DK time
            final GetCoronaPassResponse coronaPasses = conductor.getCoronaPasses(user);
            assertThat(coronaPasses)
                .hasNoDkPasses()
                .hasNoEuRecPass();
        }

        @Test
        @DisplayName("3.1.6 - Recovery after positive PCR, other invalid time")
        void case_3_1_6() throws InterruptedException {
            final IdwsUser user = IdwsUser.NC_CPR_1403993158; //has a positive test 180 days old
            final GetCoronaPassResponse coronaPasses = conductor.getCoronaPasses(user);
            assertThat(coronaPasses)
                .hasNoDkPasses()
                .hasNoEuRecPass();
        }

        //Note: Case 3.2.1 has been deleted due to not being relevant (there is no case where a new positive PCR would remove an existing EU recovery pass)

        @Test
        @DisplayName("3.2.2 test + test that recovery is based on latest positive PCR at least 11 days old")
        void case_3_2_2() throws InterruptedException {
            final IdwsUser user = IdwsUser.NC_CPR_2104513363; //two positive PCRs, -347 hours / -528 hours
            final LocalDate earliestSampleDate = zeroHourMinus(Duration.ofHours(347)).toLocalDate(); //recovery should be based on this test

            final GetCoronaPassResponse coronaPasses = conductor.getCoronaPasses(user);

            final List<PassJWT> euRecPasses = CwtTestUtils.getEuRecPasses(coronaPasses);
            assertEquals(1, euRecPasses.size());
            final RecoveryEntry recoveryEntry = conductor.getRecoveryEntry(euRecPasses);
            assertEquals(earliestSampleDate, recoveryEntry.getFr());
        }

        @Test
        @DisplayName("3.2.3 - Positive AG does not suspend EU recovery pass")
        void case_3_2_3() throws InterruptedException {
            final IdwsUser user = IdwsUser.NC_CPR_2506893339;
//            final LocalDateTime test1SampleTime = zeroHourMinus(Duration.ofHours(4));
            final LocalDateTime test2SampleTime = zeroHourMinus(Duration.ofHours(1440));

            final GetCoronaPassResponse coronaPasses = conductor.getCoronaPasses(user);

            final List<PassJWT> euRecPasses = CwtTestUtils.getEuRecPasses(coronaPasses);
            assertEquals(1, euRecPasses.size());
            final RecoveryEntry recoveryEntry = conductor.getRecoveryEntry(euRecPasses);
            assertThatRecoveryEntry(recoveryEntry)
                .hasSampleDate(test2SampleTime.toLocalDate());
        }

        @Test
        @DisplayName("New positive PCR does not suspend EU recovery pass")
        void new_positive_pcr_does_not_affect_existing_valid_eu_recovery() throws InterruptedException {
            final IdwsUser user = IdwsUser.NC_CPR_1010742357;
//            final LocalDateTime test1SampleTime = zeroHourMinus(Duration.ofHours(1));
            final LocalDateTime test2SampleTime = zeroHourMinus(Duration.ofHours(288));

            final GetCoronaPassResponse coronaPasses = conductor.getCoronaPasses(user);

            final List<PassJWT> euRecPasses = CwtTestUtils.getEuRecPasses(coronaPasses);
            assertEquals(1, euRecPasses.size());
            final RecoveryEntry recoveryEntry = conductor.getRecoveryEntry(euRecPasses);
            assertThatRecoveryEntry(recoveryEntry)
                .hasSampleDate(test2SampleTime.toLocalDate());
        }
    }

    @Nested
    @DisplayName("Child Pass Cases")
    @RequiresEmbeddedChildPassesEnabled
    class ChildPasses {
        @Value("${cwt.children.include-in-response.dk-passes}")
        private boolean makeChildrenDkPasses;

        @Test
        @DisplayName("243.1 - Child and Parent has Test Certificates")
        void childAndParentHasPassport() throws InterruptedException {
            final IdwsUser parent = IdwsUser.NC_CPR_2109624288;
            final IdwsUser child = IdwsUser.NC_CPR_0106107878;
            final LocalDateTime parentSampleTime = zeroHourMinus(Duration.ofHours(4));
            final LocalDateTime childSampleTime1 = zeroHourMinus(Duration.ofHours(4));
            final LocalDateTime childSampleTime2 = zeroHourMinus(Duration.ofHours(97));

            final GetCoronaPassResponse response = conductor.getCoronaPasses(parent);

            assertThat(response, decoder, "Parent")
                .hasDkPasses()
                .hasDkInfoWithName(parent.commonName())
                .hasEuLabPassWithSampleTime(parentSampleTime)
                .allCwtsWithNameHasName(parent.commonName());
            final List<PassJWT> parentPasses = CwtTestUtils.getPassesForCustody(response, "Parent");
            final List<TestEntry> parentTestEntries = conductor.getEuTestEntriesOrderedBySampleTime(CwtTestUtils.getEuLabPasses(parentPasses));
            assertEquals(1, parentTestEntries.size());
            assertThatTestEntry(parentTestEntries.get(0))
                .isAG()
                .isNegative()
                .hasSampleTime(parentSampleTime)
                .hasManufacturer(MANUFACTURER_HSC);

            final String childLabel = custodyLabels(response).get(child);
            if (makeChildrenDkPasses) {
                assertThat(response, childLabel).hasDkPasses();
            } else {
                assertThat(response, childLabel).hasNoDkPasses();
            }
            assertThat(response, decoder, childLabel)
                .hasDkInfoWithName(child.commonName())
                .hasEuLabPassWithSampleTime(childSampleTime1)
                .hasEuLabPassWithSampleTime(childSampleTime2)
                .allCwtsWithNameHasName(child.commonName());
            final List<PassJWT> childPasses = CwtTestUtils.getPassesForCustody(response, childLabel);
            final List<TestEntry> childTestEntries = conductor.getEuTestEntriesOrderedBySampleTime(CwtTestUtils.getEuLabPasses(childPasses));
            assertEquals(2, childTestEntries.size());
            assertThatTestEntry(childTestEntries.get(0))
                .isPcr()
                .isNegative()
                .hasSampleTime(childSampleTime2);
            assertThatTestEntry(childTestEntries.get(1))
                .isAG()
                .isNegative()
                .hasSampleTime(childSampleTime1)
                .hasManufacturer(MANUFACTURER_HSC);
        }

        @Test
        @DisplayName("243.2 - Parent has Test Certificate, Child does not")
        void onlyParentHasPassport() throws InterruptedException {
            final IdwsUser parent = IdwsUser.NC_CPR_2109624288;
            final IdwsUser child = IdwsUser.NC_CPR_2906087879;
            final LocalDateTime parentSampleTime = zeroHourMinus(Duration.ofHours(4));

            final GetCoronaPassResponse response = conductor.getCoronaPasses(parent);

            assertThat(response, decoder, "Parent")
                .hasEuLabPassWithSampleTime(parentSampleTime)
                .hasDkPasses();

            final String childLabel = custodyLabels(response).get(child);
            if (makeChildrenDkPasses) {
                assertThat(response, childLabel).hasDkPasses();
            } else {
                assertThat(response, childLabel).hasNoDkPasses();
            }
            assertThat(response, decoder, childLabel)
                .hasDkInfoWithName(child.commonName())
                .hasNoEuLabPasses()
                .allCwtsWithNameHasName(child.commonName());
        }

        @Test
        @DisplayName("243.3 - Child has Test Certificate, Parent does not")
        void onlyChildHasPassport() throws InterruptedException {
            final IdwsUser parent = IdwsUser.NC_CPR_2302471279;
            final IdwsUser child = IdwsUser.NC_CPR_0106107878;
            final LocalDateTime childSampleTime1 = zeroHourMinus(Duration.ofHours(4));
            final LocalDateTime childSampleTime2 = zeroHourMinus(Duration.ofHours(97));

            final GetCoronaPassResponse response = conductor.getCoronaPasses(parent);

            assertThat(response, decoder, "Parent")
                .hasDkInfoWithName(parent.commonName())
                .hasNoEuLabPasses()
                .allCwtsWithNameHasName(parent.commonName());

            final String childLabel = custodyLabels(response).get(child);
            if (makeChildrenDkPasses) {
                assertThat(response, childLabel).hasDkPasses();
            } else {
                assertThat(response, childLabel).hasNoDkPasses();
            }
            assertThat(response, decoder, childLabel)
                .hasDkInfoWithName(child.commonName())
                .hasEuLabPassWithSampleTime(childSampleTime1)
                .hasEuLabPassWithSampleTime(childSampleTime2)
                .allCwtsWithNameHasName(child.commonName());
            final List<PassJWT> childPasses = CwtTestUtils.getPassesForCustody(response, childLabel);
            final List<TestEntry> childTestEntries = conductor.getEuTestEntriesOrderedBySampleTime(CwtTestUtils.getEuLabPasses(childPasses));
            assertEquals(2, childTestEntries.size());
            assertThatTestEntry(childTestEntries.get(0))
                .isPcr()
                .isNegative()
                .hasSampleTime(childSampleTime2);
            assertThatTestEntry(childTestEntries.get(1))
                .isAG()
                .isNegative()
                .hasSampleTime(childSampleTime1)
                .hasManufacturer(MANUFACTURER_HSC);
        }

        @Test
        @DisplayName("243.4 - Child has Test Certificate, Parent has Vaccination Certificate")
        void childHasTestParentHasVac() throws Exception {
            final IdwsUser parent = IdwsUser.NC_CPR_2302471279;
            final IdwsUser child = IdwsUser.NC_CPR_0106107878;
            final LocalDate parentEffectutationDate = LocalDate.of(2021, 6, 1);
            final LocalDateTime childSampleTime1 = zeroHourMinus(Duration.ofHours(4));
            final LocalDateTime childSampleTime2 = zeroHourMinus(Duration.ofHours(97));

            final GetCoronaPassResponse response = conductor.getCoronaPasses(parent);

            assertThat(response, decoder, "Parent")
                .hasDkInfoWithName(parent.commonName())
                .hasEuVacPassWithEffectuationDate(parentEffectutationDate)
                .allCwtsWithNameHasName(parent.commonName());

            final String childLabel = custodyLabels(response).get(child);
            if (makeChildrenDkPasses) {
                assertThat(response, childLabel).hasDkPasses();
            } else {
                assertThat(response, childLabel).hasNoDkPasses();
            }
            assertThat(response, decoder, childLabel)
                .hasDkInfoWithName(child.commonName())
                .hasEuLabPassWithSampleTime(childSampleTime1)
                .hasEuLabPassWithSampleTime(childSampleTime2)
                .allCwtsWithNameHasName(child.commonName());
            final List<PassJWT> childPasses = CwtTestUtils.getPassesForCustody(response, childLabel);
            final List<TestEntry> childTestEntries = conductor.getEuTestEntriesOrderedBySampleTime(CwtTestUtils.getEuLabPasses(childPasses));
            assertEquals(2, childTestEntries.size());
            assertThatTestEntry(childTestEntries.get(0))
                .isPcr()
                .isNegative()
                .hasSampleTime(childSampleTime2);
            assertThatTestEntry(childTestEntries.get(1))
                .isAG()
                .isNegative()
                .hasSampleTime(childSampleTime1)
                .hasManufacturer(MANUFACTURER_HSC);
        }

        @Test
        @DisplayName("243.6 - Multiple children with different certificates")
        void multipleChildren() throws Exception {
            final IdwsUser parent = IdwsUser.NC_CPR_2302471279; //has vaccination
            final IdwsUser child1 = IdwsUser.NC_CPR_2906087879; //has vaccination
            final IdwsUser child2 = IdwsUser.NC_CPR_0106107878; //has 2 negative tests (PCR -97 timer / HSC  -4 timer)
            final IdwsUser child3 = IdwsUser.NC_CPR_3006077879; //has recovery

            final LocalDate parentEffectutationDate = LocalDate.of(2021, 6, 1);
            final LocalDate child1EffectutationDate1 = LocalDate.of(2021, 5, 1);
            final LocalDate child1EffectutationDate2 = LocalDate.of(2021, 6, 1);
            final LocalDateTime child2SampleTime1 = zeroHourMinus(Duration.ofHours(97));
            final LocalDateTime child2SampleTime2 = zeroHourMinus(Duration.ofHours(4));
            final LocalDate child3SampleDate = zeroHourMinus(Duration.ofHours(336)).toLocalDate();

            final GetCoronaPassResponse response = conductor.getCoronaPasses(parent);
            final Map<IdwsUser, String> custodyLabels = custodyLabels(response);

            //Parent
            assertThat(response, decoder, "Parent")
                .hasDkInfoWithName(parent.commonName())
                .hasNoEuLabPasses()
                .hasEuVacPasses(1)
                .hasNoEuRecPass();
            final List<VaccinationEntry> parentVaccinationEntriesOrderedByDate = conductor.getEuVaccinationEntriesOrderedByDate(CwtTestUtils.getEuVacPassesForCustody(response, "Parent"));
            Truth.assertThat(parentVaccinationEntriesOrderedByDate).hasSize(1);
            assertThatVaccination(parentVaccinationEntriesOrderedByDate.get(0))
                .isDoseOf(1, 1)
                .hasDateOfVaccination(parentEffectutationDate)
                .isJohnsonVaccination();

            //Child1
            final String child1Label = custodyLabels.get(child1);
            if (makeChildrenDkPasses) {
                assertThat(response, child1Label).hasDkPasses();
            } else {
                assertThat(response, child1Label).hasNoDkPasses();
            }
            assertThat(response, decoder, child1Label)
                .hasDkInfoWithName(child1.commonName())
                .hasNoEuLabPasses()
                .hasEuVacPasses(2)
                .hasNoEuRecPass()
                .allCwtsWithNameHasName(child1.commonName());
            final List<VaccinationEntry> child1VaccinationEntriesOrderedByDate = conductor.getEuVaccinationEntriesOrderedByDate(CwtTestUtils.getEuVacPassesForCustody(response, child1Label));
            Truth.assertThat(child1VaccinationEntriesOrderedByDate).hasSize(2);
            assertThatVaccination(child1VaccinationEntriesOrderedByDate.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(child1EffectutationDate1)
                .isPfizerVaccination();
            assertThatVaccination(child1VaccinationEntriesOrderedByDate.get(1))
                .isDoseOf(2, 2)
                .hasDateOfVaccination(child1EffectutationDate2)
                .isPfizerVaccination();

            //Child2
            final String child2Label = custodyLabels.get(child2);
            if (makeChildrenDkPasses) {
                assertThat(response, child2Label).hasDkPasses();
            } else {
                assertThat(response, child2Label).hasNoDkPasses();
            }
            assertThat(response, decoder, child2Label)
                .hasDkInfoWithName(child2.commonName())
                .hasEuLabPasses(2)
                .hasNoEuVacPasses()
                .hasNoEuRecPass()
                .allCwtsWithNameHasName(child2.commonName());
            final List<TestEntry> child2TestEntriesOrderedBySampleTime = conductor.getEuTestEntriesOrderedBySampleTime(CwtTestUtils.getEuLabPassesForCustody(response, child2Label));
            Truth.assertThat(child1VaccinationEntriesOrderedByDate).hasSize(2);
            assertThatTestEntry(child2TestEntriesOrderedBySampleTime.get(0))
                .isPcr()
                .isNegative()
                .hasSampleTime(child2SampleTime1);
            assertThatTestEntry(child2TestEntriesOrderedBySampleTime.get(1))
                .isAG()
                .hasManufacturer(MANUFACTURER_HSC)
                .isNegative()
                .hasSampleTime(child2SampleTime2);

            //Child3
            final String child3Label = custodyLabels.get(child3);
            if (makeChildrenDkPasses) {
                assertThat(response, child3Label).hasDkPasses();
            } else {
                assertThat(response, child3Label).hasNoDkPasses();
            }
            assertThat(response, decoder, child3Label)
                .hasDkInfoWithName(child3.commonName())
                .hasNoEuLabPasses()
                .hasNoEuLabPasses()
                .hasEuRecPasses(1);
            final RecoveryEntry recoveryEntry = conductor.getRecoveryEntry(CwtTestUtils.getEuRecPassesForCustody(response, child3Label));
            assertThatRecoveryEntry(recoveryEntry)
                .hasSampleDate(child3SampleDate)
                .hasStartDate(child3SampleDate.plus(RECOVERY_VALID_START))
                .hasEndDate(child3SampleDate.plus(RECOVERY_VALID_END));
        }

        @Test
        @DisplayName("243.7 - Children older than 15 are not included in Parents response")
        void childOver15NotIncluded() throws InterruptedException {
            final IdwsUser parent = IdwsUser.NC_CPR_2104262681;
            final IdwsUser childUnder15 = IdwsUser.NC_CPR_2405125656; //no data
            final IdwsUser childOver15 = IdwsUser.NC_CPR_0107066666;

            final GetCoronaPassResponse response = conductor.getCoronaPasses(parent);

            assertThat(response, decoder, "Parent")
                .hasDkInfoWithName(parent.commonName())
                .allCwtsWithNameHasName(parent.commonName());

            Truth.assertThat(custodyLabels(response)).doesNotContainKey(childOver15);

            final String childLabel = custodyLabels(response).get(childUnder15);
            assertThat(response, childLabel).hasNoDkPasses();
            assertThat(response, decoder, childLabel)
                .hasDkInfoWithName(childUnder15.commonName())
                .hasNoDkPasses()
                .hasNoEuLabPasses()
                .hasNoEuVacPasses();
        }

        private Map<IdwsUser, String> custodyLabels(GetCoronaPassResponse coronaPass) {
            return CwtTestUtils.getDKInfoPass(coronaPass).stream()
                .collect(Collectors.toMap(passJWT -> IdwsUser.fromName(getName(conductor.getDKINFOPayload(passJWT.getJWT()))), CwtTestUtils::getCustody));
        }

        private String getName(DigitalGreenCertificate certificate) {
            return certificate.getNam().getGn() + " " + certificate.getNam().getFn();
        }
    }

    @Nested
    @DisplayName("EU Corona Pass Rules Special Cases")
    class SpecialCases {

        @Test
        void power_of_attorney_can_get_passes() throws dk.ssi.nsi.xml_schema._2013._01._01.DGWSFault, DGWSFault, InterruptedException {
            final IdwsUser attorney = IdwsUser.POWER_OF_ATTORNEY_ACTING_USER;
            final IdwsUser principal = IdwsUser.POWER_OF_ATTORNEY_PRINCIPAL_USER;

            final Instant attorneyVaccinationTime = now.minus(Duration.ofDays(30));
            final Instant principalVaccinationTime = now.minus(Duration.ofDays(60));

            //attorney gets principal's corona passes
            ddvDriver.withCitizen(principal)
                .clear()
                .addJohnsonVaccination(Date.from(principalVaccinationTime));
            GetCoronaPassResponse response1 = conductor.getCoronaPassesAfterInvalidation(attorney, principal);
            assertThat(response1, decoder)
                .hasDkPasses()
                .hasDkInfoWithName(principal.commonName())
                .hasEuVacPasses(1);
            final List<PassJWT> principalVaccinationPasses = CwtTestUtils.getEuVacPasses(response1);
            final List<VaccinationEntry> principalVaccinationEntries = conductor.getEuVaccinationEntriesOrderedByDate(principalVaccinationPasses);
            assertThatVaccination(principalVaccinationEntries.get(0))
                .isDoseOf(1, 1)
                .hasDateOfVaccination(principalVaccinationTime)
                .isJohnsonVaccination();

            //attorney gets own corona passes
            ddvDriver.withCitizen(attorney)
                .clear()
                .addPfizerVaccination(Date.from(attorneyVaccinationTime));
            final GetCoronaPassResponse response2 = conductor.getCoronaPassesAfterInvalidation(attorney);
            assertThat(response2, decoder)
                .hasDkPasses()
                .hasDkInfoWithName(attorney.commonName())
                .hasEuVacPasses(1);
            final List<PassJWT> attorneyVaccinationPasses = CwtTestUtils.getEuVacPasses(response2);
            final List<VaccinationEntry> attorneyVaccinationEntries = conductor.getEuVaccinationEntriesOrderedByDate(attorneyVaccinationPasses);
            assertThatVaccination(attorneyVaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(attorneyVaccinationTime)
                .isPfizerVaccination();
        }

        @Test
        void power_of_attorney_specifies_wrong_input_cpr() {
            final IdwsUser attorney = IdwsUser.POWER_OF_ATTORNEY_ACTING_USER;
            final IdwsUser notPrincipal = IdwsUser.CPR_2401010010;

            String principalCpr = notPrincipal.cpr();
            assertThrows(SOAPFaultException.class, () -> cbsDriver.getCoronaPass(attorney, principalCpr));
        }

        @Test
        void user_does_not_have_power_of_attorney() {
            final IdwsUser notAttorney = IdwsUser.CPR_2401010014;
            final IdwsUser principal = IdwsUser.POWER_OF_ATTORNEY_PRINCIPAL_USER;

            String principalCpr = principal.cpr();
            assertThrows(SOAPFaultException.class, () -> cbsDriver.getCoronaPass(notAttorney, principalCpr));
        }

        @Test
        @DisplayName("Special Case - Issue Test Certificates for positive test results")
        @EnabledIf(expression = "${cwt.eu.include-positive-results}",
            reason = "Requires that positive test certificates are enabled on CBS",
            loadContext = true)
        void issue_test_certificates_for_positive_results() throws Exception {
            final IdwsUser user = IdwsUser.NC_CPR_3012283023;
            final LocalDateTime test1_negativePcr = zeroHourMinus(Duration.ofHours(4));
            final LocalDateTime test2_positivePcr = zeroHourMinus(Duration.ofHours(8));
            final LocalDateTime test3_negativePcr = zeroHourMinus(Duration.ofHours(12));
            final LocalDateTime test4_negativeHscAg = zeroHourMinus(Duration.ofHours(16));
            final LocalDateTime test5_positiveHscAg = zeroHourMinus(Duration.ofHours(20));
//            final LocalDateTime test6_negativeNotHscAg = zeroHourMinus(Duration.ofHours(24));

            final GetCoronaPassResponse coronaPasses = conductor.getCoronaPasses(user);
            final List<PassJWT> euLabPasses = CwtTestUtils.getEuLabPasses(coronaPasses);
            assertEquals(5, euLabPasses.size());
            final List<TestEntry> testEntries = conductor.getEuTestEntriesOrderedBySampleTime(euLabPasses);
            assertEquals(5, testEntries.size());

            assertThatTestEntry(testEntries.get(0))
                .isPositive()
                .isAG()
                .hasSampleTime(test5_positiveHscAg);

            assertThatTestEntry(testEntries.get(1))
                .isNegative()
                .isAG()
                .hasManufacturer(MANUFACTURER_HSC)
                .hasSampleTime(test4_negativeHscAg);

            assertThatTestEntry(testEntries.get(2))
                .isNegative()
                .isPcr()
                .hasSampleTime(test3_negativePcr);

            assertThatTestEntry(testEntries.get(3))
                .isPositive()
                .isPcr()
                .hasSampleTime(test2_positivePcr);

            assertThatTestEntry(testEntries.get(4))
                .isNegative()
                .isPcr()
                .hasSampleTime(test1_negativePcr);
        }

        @Test
        @DisplayName("Special Case - Issue at most 5 test certificates of each type within the last 7 days")
        void max_5_test_passes_for_each_type() throws InterruptedException {
            final LocalDateTime pcrTest1SampleTime = zeroHourMinus(Duration.ofHours(24));
            final LocalDateTime pcrTest2SampleTime = pcrTest1SampleTime.minusHours(1);
            final LocalDateTime pcrTest3SampleTime = pcrTest1SampleTime.minusHours(2);
            final LocalDateTime pcrTest4SampleTime = pcrTest1SampleTime.minusHours(3);
            final LocalDateTime pcrTest5SampleTime = pcrTest1SampleTime.minusHours(4);
//            final LocalDateTime pcrTest6SampleTime = pcrTest1SampleTime.minusHours(5);
            final LocalDateTime agTest1SampleTime = pcrTest1SampleTime.minusHours(6);
            final LocalDateTime agTest2SampleTime = pcrTest1SampleTime.minusHours(7);
            final LocalDateTime agTest3SampleTime = pcrTest1SampleTime.minusHours(8);
            final LocalDateTime agTest4SampleTime = pcrTest1SampleTime.minusHours(9);
            final LocalDateTime agTest5SampleTime = pcrTest1SampleTime.minusHours(10);
//            final LocalDateTime agTest6SampleTime = pcrTest1SampleTime.minusHours(11);

            final GetCoronaPassResponse coronaPasses = conductor.getCoronaPasses(IdwsUser.NC_CPR_0712081266);
            assertThat(coronaPasses).hasEuLabPasses(10);
            final List<PassJWT> euLabPasses = CwtTestUtils.getEuLabPasses(coronaPasses);
            final List<TestEntry> testEntries = conductor.getEuTestEntriesOrderedBySampleTime(euLabPasses);
            assertEquals(10, testEntries.size());

            assertThatTestEntry(testEntries.get(9)).isNegative().isPcr().hasSampleTime(pcrTest1SampleTime);
            assertThatTestEntry(testEntries.get(8)).isNegative().isPcr().hasSampleTime(pcrTest2SampleTime);
            assertThatTestEntry(testEntries.get(7)).isNegative().isPcr().hasSampleTime(pcrTest3SampleTime);
            assertThatTestEntry(testEntries.get(6)).isNegative().isPcr().hasSampleTime(pcrTest4SampleTime);
            assertThatTestEntry(testEntries.get(5)).isNegative().isPcr().hasSampleTime(pcrTest5SampleTime);
            assertThatTestEntry(testEntries.get(4)).isNegative().isAG().hasSampleTime(agTest1SampleTime);
            assertThatTestEntry(testEntries.get(3)).isNegative().isAG().hasSampleTime(agTest2SampleTime);
            assertThatTestEntry(testEntries.get(2)).isNegative().isAG().hasSampleTime(agTest3SampleTime);
            assertThatTestEntry(testEntries.get(1)).isNegative().isAG().hasSampleTime(agTest4SampleTime);
            assertThatTestEntry(testEntries.get(0)).isNegative().isAG().hasSampleTime(agTest5SampleTime);
        }

        @Test
        @DisplayName("Special Case - Vaccination pass suspended due to positive PCR test")
        void vaccination_pass_suspended_due_to_positive_PCR() throws Exception {
            final IdwsUser user = IdwsUser.NC_CPR_1704080278;
            final LocalDate effectuationDate1 = LocalDate.of(2021, 7, 16);
            final LocalDate effectuationDate2 = LocalDate.of(2021, 8, 26);
//            final LocalDateTime positivePcrSampleTime = zeroHourMinus(Duration.ofHours(4));

            final GetCoronaPassResponse coronaPasses = conductor.getCoronaPasses(user);

            assertThat(coronaPasses)
                .hasNoDkPasses() //a positive test within last 14 days suspends the DK vaccination pass
                .hasNoEuLabPasses()
                .hasEuVacPasses(2)
                .hasNoEuRecPass();

            final List<VaccinationEntry> vaccinationEntriesOrderedByDate = conductor.getEuVaccinationEntriesOrderedByDate(CwtTestUtils.getEuVacPasses(coronaPasses));
            Truth.assertThat(vaccinationEntriesOrderedByDate).hasSize(2);
            assertThatVaccination(vaccinationEntriesOrderedByDate.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(effectuationDate1)
                .isAstraZenicaVaccination();
            assertThatVaccination(vaccinationEntriesOrderedByDate.get(1))
                .isDoseOf(2, 2)
                .hasDateOfVaccination(effectuationDate2)
                .isAstraZenicaVaccination();
        }

        @Test
        @DisplayName("Special Case - Vaccination pass suspended due to positive AG test without subsequent negative PCR")
        void vaccination_pass_suspended_due_to_positive_AG() throws Exception {
            final IdwsUser user = IdwsUser.NC_CPR_2310901373;
            final LocalDate effectuationDate1 = LocalDate.of(2021, 7, 16);
            final LocalDate effectuationDate2 = LocalDate.of(2021, 8, 26);
//            final LocalDate positiveAgSampleDate = zeroHourMinus(Duration.ofHours(4)).toLocalDate();

            final GetCoronaPassResponse coronaPasses = conductor.getCoronaPasses(user);

            assertThat(coronaPasses)
                .hasNoDkPasses() //a positive test within last 14 days suspends the DK vaccination pass
                .hasNoEuLabPasses()
                .hasEuVacPasses(2)
                .hasNoEuRecPass();

            final List<VaccinationEntry> vaccinationEntriesOrderedByDate = conductor.getEuVaccinationEntriesOrderedByDate(CwtTestUtils.getEuVacPasses(coronaPasses));
            Truth.assertThat(vaccinationEntriesOrderedByDate).hasSize(2);
            assertThatVaccination(vaccinationEntriesOrderedByDate.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(effectuationDate1)
                .isPfizerVaccination();
            assertThatVaccination(vaccinationEntriesOrderedByDate.get(1))
                .isDoseOf(2, 2)
                .hasDateOfVaccination(effectuationDate2)
                .isPfizerVaccination();
        }

        @Test
        @DisplayName("Special Case - Vaccination pass suspension lifted due to positive PCR test older han suspension period")
        void vaccination_pass_suspension_lifted_due_to_time_passed_after_positive_PCR() throws Exception {
            final IdwsUser user = IdwsUser.NC_CPR_0811220950;
            //TODO test with more recent vaccinations - this test is not actually testing that vaccination pass suspension was lifted (vaccination pass has expired)
            final LocalDate effectuationDate1 = LocalDate.of(2021, 3, 16);
            final LocalDate effectuationDate2 = LocalDate.of(2021, 4, 26);
            final LocalDate positivePcrSampleDate = zeroHourMinus(Duration.ofHours(347)).toLocalDate();

            final GetCoronaPassResponse coronaPasses = conductor.getCoronaPasses(user);

            assertThat(coronaPasses)
                .hasDkPasses()
                .hasNoEuLabPasses()
                .hasEuVacPasses(2)
                .hasEuRecPasses(1);

            final List<VaccinationEntry> vaccinationEntriesOrderedByDate = conductor.getEuVaccinationEntriesOrderedByDate(CwtTestUtils.getEuVacPasses(coronaPasses));
            Truth.assertThat(vaccinationEntriesOrderedByDate).hasSize(2);
            assertThatVaccination(vaccinationEntriesOrderedByDate.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(effectuationDate1)
                .isAstraZenicaVaccination();
            assertThatVaccination(vaccinationEntriesOrderedByDate.get(1))
                .isDoseOf(2, 2)
                .hasDateOfVaccination(effectuationDate2)
                .isAstraZenicaVaccination();

            final RecoveryEntry recoveryEntry = conductor.getRecoveryEntry(CwtTestUtils.getEuRecPasses(coronaPasses));
            assertThatRecoveryEntry(recoveryEntry)
                .hasSampleDate(positivePcrSampleDate)
                .hasStartDate(positivePcrSampleDate.plus(RECOVERY_VALID_START))
                .hasEndDate(positivePcrSampleDate.plus(RECOVERY_VALID_END));
        }

        @Test
        @DisplayName("Special Case - Vaccination pass suspension lifted due to positive AG test older han suspension period")
        void revaccination_pass_suspension_lifted_due_to_time_passed_after_positive_AG() throws Exception {
            final IdwsUser user = IdwsUser.NC_CPR_1206103647; //has completed vaccination (24/4/2021) + revaccination (15/9/2021) + immunity via positive AG -336 hours
            final LocalDate effectuationDate1 = LocalDate.of(2021, 3, 16);
            final LocalDate effectuationDate2 = LocalDate.of(2021, 4, 26);
            final LocalDate revaccinationDate = LocalDate.of(2021, 9, 15);
            final LocalDate positiveAgSampleDate = zeroHourMinus(Duration.ofHours(336)).toLocalDate();

            final GetCoronaPassResponse coronaPasses = conductor.getCoronaPasses(user);

            assertThat(coronaPasses)
                .hasDkPasses() //from revaccination
                .hasNoEuLabPasses()
                .hasEuVacPasses(3)
                .hasNoEuRecPass();

            final List<VaccinationEntry> vaccinationEntriesOrderedByDate = conductor.getEuVaccinationEntriesOrderedByDate(CwtTestUtils.getEuVacPasses(coronaPasses));
            Truth.assertThat(vaccinationEntriesOrderedByDate).hasSize(3);
            assertThatVaccination(vaccinationEntriesOrderedByDate.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(effectuationDate1)
                .isPfizerVaccination();
            assertThatVaccination(vaccinationEntriesOrderedByDate.get(1))
                .isDoseOf(2, 2)
                .hasDateOfVaccination(effectuationDate2)
                .isPfizerVaccination();
            assertThatVaccination(vaccinationEntriesOrderedByDate.get(2))
                .isDoseOf(3, 3)
                .hasDateOfVaccination(revaccinationDate)
                .isPfizerVaccination();
        }

        @Test
        @DisplayName("Special Case - Vaccination pass suspension lifted due to negative PCR after positive AG")
        void vaccination_pass_suspension_lifted_due_to_negative_PCR_after_positive_AG() throws Exception {
            final IdwsUser user = IdwsUser.NC_CPR_0706293072;
            final LocalDate effectuationDate1 = LocalDate.of(2021, 8, 1);
            final LocalDate effectuationDate2 = LocalDate.of(2021, 8, 30);
//            final LocalDateTime positiveAgSampleDate = zeroHourMinus(Duration.ofHours(168));
            final LocalDateTime negativePcrSampleDate = zeroHourMinus(Duration.ofHours(144));

            final GetCoronaPassResponse coronaPasses = conductor.getCoronaPasses(user);

            assertThat(coronaPasses, decoder)
                .hasDkPasses()
                .hasEuLabPassWithSampleTime(negativePcrSampleDate)
                .hasEuVacPasses(2)
                .hasNoEuRecPass();

            final List<VaccinationEntry> vaccinationEntriesOrderedByDate = conductor.getEuVaccinationEntriesOrderedByDate(CwtTestUtils.getEuVacPasses(coronaPasses));
            Truth.assertThat(vaccinationEntriesOrderedByDate).hasSize(2);
            assertThatVaccination(vaccinationEntriesOrderedByDate.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(effectuationDate1)
                .isPfizerVaccination();
            assertThatVaccination(vaccinationEntriesOrderedByDate.get(1))
                .isDoseOf(2, 2)
                .hasDateOfVaccination(effectuationDate2)
                .isPfizerVaccination();
        }
    }

    @Nested
    @DisplayName("CR-321 + CR-349 - Additional Vaccinations")
    class AdditionalVaccinations {

        @ParameterizedTest(name = "3stik_1 (AKA CR-349 #07 + #08) - 3 stik, alle pfizer - færdigvaccineret for {0} måneder siden")
        @ValueSource(ints = {5, 11})
        void case1(int monthsSinceCompletedVaccination) throws dk.ssi.nsi.xml_schema._2013._01._01.DGWSFault, DGWSFault, InterruptedException {
            final Instant effectuationTime1 = LocalDateTime.now().minusMonths(monthsSinceCompletedVaccination + 1).toInstant(ZoneOffset.UTC);
            final Instant effectuationTime2 = LocalDateTime.now().minusMonths(monthsSinceCompletedVaccination).toInstant(ZoneOffset.UTC);
            final Instant effectuationTime3 = LocalDateTime.now().minusMonths(1).toInstant(ZoneOffset.UTC);

            IdwsUser user = nextUser();
            final Citizen citizen = ddvDriver.withCitizen(user).clear();
            citizen.addPfizerVaccination(Date.from(effectuationTime1)); //program vaccination
            citizen.addPfizerVaccination(Date.from(effectuationTime2)); //program vaccination
            citizen.addPfizerVaccination(Date.from(effectuationTime3)); //additional vaccination

            final GetCoronaPassResponse response = conductor.getCoronaPassesAfterInvalidation(user);

            assertThat(response)
                .hasDkPasses()
                .hasEuVacPasses(3)
                .hasNoEuLabPasses()
                .hasNoEuRecPass();

            final List<VaccinationEntry> euVaccinationEntries = conductor.getEuVaccinationEntriesOrderedByDate(CwtTestUtils.getEuVacPasses(response));

            assertThatVaccination(euVaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(effectuationTime1)
                .isPfizerVaccination();
            assertThatVaccination(euVaccinationEntries.get(1))
                .isDoseOf(2, 2)
                .hasDateOfVaccination(effectuationTime2)
                .isPfizerVaccination();
            assertThatVaccination(euVaccinationEntries.get(2))
                .isDoseOf(3, 3)
                .hasDateOfVaccination(effectuationTime3)
                .isPfizerVaccination();

            final List<VaccinationEntry> dkInfoVaccinationEntries = conductor.getDKINFOPayload(CwtTestUtils.getDKInfoPass(response).get(0).getJWT()).getV();
            assertEquals(3, dkInfoVaccinationEntries.size());
            assertThatVaccination(dkInfoVaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(effectuationTime1)
                .isPfizerVaccination();
            assertThatVaccination(dkInfoVaccinationEntries.get(1))
                .isDoseOf(2, 2)
                .hasDateOfVaccination(effectuationTime2)
                .isPfizerVaccination();
            assertThatVaccination(dkInfoVaccinationEntries.get(2))
                .isDoseOf(3, 3)
                .hasDateOfVaccination(effectuationTime3)
                .isPfizerVaccination();
        }

        @Test
        @DisplayName("3stik_2 - Ældre Vaccination med 3 stik, alle pfizer")
        void case2() throws DGWSFault, dk.ssi.nsi.xml_schema._2013._01._01.DGWSFault, InterruptedException {
            final Instant effectuationTime1 = LocalDateTime.of(2020, 12, 1, 14, 0).toInstant(ZoneOffset.UTC);
            final Instant effectuationTime2 = LocalDateTime.of(2021, 1, 1, 14, 0).toInstant(ZoneOffset.UTC);
            final Instant effectuationTime3 = LocalDateTime.of(2021, 2, 1, 14, 0).toInstant(ZoneOffset.UTC);

            IdwsUser user = nextUser();
            final Citizen citizen = ddvDriver.withCitizen(user).clear();
            citizen.addPfizerVaccination(Date.from(effectuationTime1)); //program vaccination
            citizen.addPfizerVaccination(Date.from(effectuationTime2)); //program vaccination
            citizen.addPfizerVaccination(Date.from(effectuationTime3)); //additional vaccination

            final GetCoronaPassResponse response = conductor.getCoronaPassesAfterInvalidation(user);
            final List<VaccinationEntry> euVaccinationEntries = conductor.getEuVaccinationEntriesOrderedByDate(CwtTestUtils.getEuVacPasses(response));
            assertEquals(3, euVaccinationEntries.size());

            assertThatVaccination(euVaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(effectuationTime1)
                .isPfizerVaccination();
            assertThatVaccination(euVaccinationEntries.get(1))
                .isDoseOf(2, 2)
                .hasDateOfVaccination(effectuationTime2)
                .isPfizerVaccination();
            assertThatVaccination(euVaccinationEntries.get(2))
                .isDoseOf(3, 3)
                .hasDateOfVaccination(effectuationTime3)
                .isPfizerVaccination();

            final List<VaccinationEntry> dkInfoVaccinationEntries = conductor.getDKINFOPayload(CwtTestUtils.getDKInfoPass(response).get(0).getJWT()).getV();
            assertEquals(3, dkInfoVaccinationEntries.size());
            assertThatVaccination(dkInfoVaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(effectuationTime1)
                .isPfizerVaccination();
            assertThatVaccination(dkInfoVaccinationEntries.get(1))
                .isDoseOf(2, 2)
                .hasDateOfVaccination(effectuationTime2)
                .isPfizerVaccination();
            assertThatVaccination(dkInfoVaccinationEntries.get(2))
                .isDoseOf(3, 3)
                .hasDateOfVaccination(effectuationTime3)
                .isPfizerVaccination();
        }

        @Test
        @DisplayName("3stik_3 - Ny Vaccination med 4 stik, alle pfizer")
        void case3() throws DGWSFault, dk.ssi.nsi.xml_schema._2013._01._01.DGWSFault, InterruptedException {
            final Instant effectuationTime1 = LocalDateTime.of(2021, 2, 1, 14, 0).toInstant(ZoneOffset.UTC);
            final Instant effectuationTime2 = LocalDateTime.of(2021, 3, 1, 14, 0).toInstant(ZoneOffset.UTC);
            final Instant effectuationTime3 = LocalDateTime.of(2021, 8, 1, 14, 0).toInstant(ZoneOffset.UTC);
            final Instant effectuationTime4 = LocalDateTime.of(2021, 9, 1, 14, 0).toInstant(ZoneOffset.UTC);

            IdwsUser user = nextUser();
            final Citizen citizen = ddvDriver.withCitizen(user).clear();
            citizen.addPfizerVaccination(Date.from(effectuationTime1)); //program vaccination
            citizen.addPfizerVaccination(Date.from(effectuationTime2)); //program vaccination
            citizen.addPfizerVaccination(Date.from(effectuationTime3)); //additional vaccination
            citizen.addPfizerVaccination(Date.from(effectuationTime4)); //additional vaccination

            final GetCoronaPassResponse response = conductor.getCoronaPassesAfterInvalidation(user);
            final List<VaccinationEntry> euVaccinationEntries = conductor.getEuVaccinationEntriesOrderedByDate(CwtTestUtils.getEuVacPasses(response));
            assertEquals(4, euVaccinationEntries.size());

            assertThatVaccination(euVaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(effectuationTime1)
                .isPfizerVaccination();
            assertThatVaccination(euVaccinationEntries.get(1))
                .isDoseOf(2, 2)
                .hasDateOfVaccination(effectuationTime2)
                .isPfizerVaccination();
            assertThatVaccination(euVaccinationEntries.get(2))
                .isDoseOf(3, 3)
                .hasDateOfVaccination(effectuationTime3)
                .isPfizerVaccination();
            assertThatVaccination(euVaccinationEntries.get(3))
                .isDoseOf(4, 4)
                .hasDateOfVaccination(effectuationTime4)
                .isPfizerVaccination();

            final List<VaccinationEntry> dkInfoVaccinationEntries = conductor.getDKINFOPayload(CwtTestUtils.getDKInfoPass(response).get(0).getJWT()).getV();
            assertEquals(4, dkInfoVaccinationEntries.size());
            assertThatVaccination(dkInfoVaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(effectuationTime1)
                .isPfizerVaccination();
            assertThatVaccination(dkInfoVaccinationEntries.get(1))
                .isDoseOf(2, 2)
                .hasDateOfVaccination(effectuationTime2)
                .isPfizerVaccination();
            assertThatVaccination(dkInfoVaccinationEntries.get(2))
                .isDoseOf(3, 3)
                .hasDateOfVaccination(effectuationTime3)
                .isPfizerVaccination();
            assertThatVaccination(dkInfoVaccinationEntries.get(3))
                .isDoseOf(4, 4)
                .hasDateOfVaccination(effectuationTime4)
                .isPfizerVaccination();
        }

        @Test
        @DisplayName("3stik_4 - Kombi med 3 stik færdigvacineret")
        void case4() throws DGWSFault, dk.ssi.nsi.xml_schema._2013._01._01.DGWSFault, InterruptedException {
            final Instant effectuationTime1 = LocalDateTime.of(2021, 2, 1, 14, 0).toInstant(ZoneOffset.UTC);
            final Instant effectuationTime2 = LocalDateTime.of(2021, 3, 1, 14, 0).toInstant(ZoneOffset.UTC);
            final Instant effectuationTime3 = LocalDateTime.of(2021, 8, 1, 14, 0).toInstant(ZoneOffset.UTC);

            IdwsUser user = nextUser();
            final Citizen citizen = ddvDriver.withCitizen(user).clear();
            citizen.addAstraZenecaVaccination(Date.from(effectuationTime1)); //program vaccination
            citizen.addPfizerVaccination(Date.from(effectuationTime2)); //program vaccination
            citizen.addPfizerVaccination(Date.from(effectuationTime3)); //additional vaccination

            final GetCoronaPassResponse response = conductor.getCoronaPassesAfterInvalidation(user);
            final List<VaccinationEntry> euVaccinationEntries = conductor.getEuVaccinationEntriesOrderedByDate(CwtTestUtils.getEuVacPasses(response));
            assertEquals(3, euVaccinationEntries.size());

            assertThatVaccination(euVaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(effectuationTime1)
                .isAstraZenicaVaccination();
            assertThatVaccination(euVaccinationEntries.get(1))
                .isDoseOf(2, 2)
                .hasDateOfVaccination(effectuationTime2)
                .isPfizerVaccination();
            assertThatVaccination(euVaccinationEntries.get(2))
                .isDoseOf(3, 3)
                .hasDateOfVaccination(effectuationTime3)
                .isPfizerVaccination();

            final List<VaccinationEntry> dkInfoVaccinationEntries = conductor.getDKINFOPayload(CwtTestUtils.getDKInfoPass(response).get(0).getJWT()).getV();
            assertEquals(3, dkInfoVaccinationEntries.size());
            assertThatVaccination(dkInfoVaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(effectuationTime1)
                .isAstraZenicaVaccination();
            assertThatVaccination(dkInfoVaccinationEntries.get(1))
                .isDoseOf(2, 2)
                .hasDateOfVaccination(effectuationTime2)
                .isPfizerVaccination();
            assertThatVaccination(dkInfoVaccinationEntries.get(2))
                .isDoseOf(3, 3)
                .hasDateOfVaccination(effectuationTime3)
                .isPfizerVaccination();
        }

        @Test
        @DisplayName("3stik_5 - Færdigvaccination med 3. stik af anden type")
        void case5() throws DGWSFault, dk.ssi.nsi.xml_schema._2013._01._01.DGWSFault, InterruptedException {
            final Instant effectuationTime1 = LocalDateTime.of(2021, 2, 1, 14, 0).toInstant(ZoneOffset.UTC);
            final Instant effectuationTime2 = LocalDateTime.of(2021, 5, 1, 14, 0).toInstant(ZoneOffset.UTC);
            final Instant effectuationTime3 = LocalDateTime.of(2021, 8, 1, 14, 0).toInstant(ZoneOffset.UTC);

            IdwsUser user = nextUser();
            final Citizen citizen = ddvDriver.withCitizen(user).clear();
            citizen.addAstraZenecaVaccination(Date.from(effectuationTime1)); //program vaccination
            citizen.addAstraZenecaVaccination(Date.from(effectuationTime2)); //program vaccination
            citizen.addPfizerVaccination(Date.from(effectuationTime3)); //additional vaccination

            final GetCoronaPassResponse response = conductor.getCoronaPassesAfterInvalidation(user);
            final List<VaccinationEntry> euVaccinationEntries = conductor.getEuVaccinationEntriesOrderedByDate(CwtTestUtils.getEuVacPasses(response));
            assertEquals(3, euVaccinationEntries.size());

            assertThatVaccination(euVaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(effectuationTime1)
                .isAstraZenicaVaccination();
            assertThatVaccination(euVaccinationEntries.get(1))
                .isDoseOf(2, 2)
                .hasDateOfVaccination(effectuationTime2)
                .isAstraZenicaVaccination();
            assertThatVaccination(euVaccinationEntries.get(2))
                .isDoseOf(3, 3)
                .hasDateOfVaccination(effectuationTime3)
                .isPfizerVaccination();

            final List<VaccinationEntry> dkInfoVaccinationEntries = conductor.getDKINFOPayload(CwtTestUtils.getDKInfoPass(response).get(0).getJWT()).getV();
            assertEquals(3, dkInfoVaccinationEntries.size());
            assertThatVaccination(dkInfoVaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(effectuationTime1)
                .isAstraZenicaVaccination();
            assertThatVaccination(dkInfoVaccinationEntries.get(1))
                .isDoseOf(2, 2)
                .hasDateOfVaccination(effectuationTime2)
                .isAstraZenicaVaccination();
            assertThatVaccination(dkInfoVaccinationEntries.get(2))
                .isDoseOf(3, 3)
                .hasDateOfVaccination(effectuationTime3)
                .isPfizerVaccination();
        }

        @Test
        @DisplayName("3stik_6 - Færdigvaccination med 3 og 4 stik af anden men samme type")
        void case6() throws DGWSFault, dk.ssi.nsi.xml_schema._2013._01._01.DGWSFault, InterruptedException {
            final Instant effectuationTime1 = LocalDateTime.of(2020, 2, 1, 14, 0).toInstant(ZoneOffset.UTC);
            final Instant effectuationTime2 = LocalDateTime.of(2021, 5, 1, 14, 0).toInstant(ZoneOffset.UTC);
            final Instant effectuationTime3 = LocalDateTime.of(2021, 8, 1, 14, 0).toInstant(ZoneOffset.UTC);
            final Instant effectuationTime4 = LocalDateTime.of(2021, 9, 1, 14, 0).toInstant(ZoneOffset.UTC);

            IdwsUser user = nextUser();
            final Citizen citizen = ddvDriver.withCitizen(user).clear();
            citizen.addAstraZenecaVaccination(Date.from(effectuationTime1)); //program vaccination
            citizen.addAstraZenecaVaccination(Date.from(effectuationTime2)); //program vaccination
            citizen.addPfizerVaccination(Date.from(effectuationTime3)); //additional vaccination
            citizen.addPfizerVaccination(Date.from(effectuationTime4)); //additional vaccination

            final GetCoronaPassResponse response = conductor.getCoronaPassesAfterInvalidation(user);
            final List<VaccinationEntry> euVaccinationEntries = conductor.getEuVaccinationEntriesOrderedByDate(CwtTestUtils.getEuVacPasses(response));
            assertEquals(4, euVaccinationEntries.size());

            assertThatVaccination(euVaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(effectuationTime1)
                .isAstraZenicaVaccination();
            assertThatVaccination(euVaccinationEntries.get(1))
                .isDoseOf(2, 2)
                .hasDateOfVaccination(effectuationTime2)
                .isAstraZenicaVaccination();
            assertThatVaccination(euVaccinationEntries.get(2))
                .isDoseOf(3, 3)
                .hasDateOfVaccination(effectuationTime3)
                .isPfizerVaccination();
            assertThatVaccination(euVaccinationEntries.get(3))
                .isDoseOf(4, 4)
                .hasDateOfVaccination(effectuationTime4)
                .isPfizerVaccination();

            final List<VaccinationEntry> dkInfoVaccinationEntries = conductor.getDKINFOPayload(CwtTestUtils.getDKInfoPass(response).get(0).getJWT()).getV();
            assertEquals(4, dkInfoVaccinationEntries.size());
            assertThatVaccination(dkInfoVaccinationEntries.get(0))
                .isDoseOf(1, 2)
                .hasDateOfVaccination(effectuationTime1)
                .isAstraZenicaVaccination();
            assertThatVaccination(dkInfoVaccinationEntries.get(1))
                .isDoseOf(2, 2)
                .hasDateOfVaccination(effectuationTime2)
                .isAstraZenicaVaccination();
            assertThatVaccination(dkInfoVaccinationEntries.get(2))
                .isDoseOf(3, 3)
                .hasDateOfVaccination(effectuationTime3)
                .isPfizerVaccination();
            assertThatVaccination(dkInfoVaccinationEntries.get(3))
                .isDoseOf(4, 4)
                .hasDateOfVaccination(effectuationTime4)
                .isPfizerVaccination();
        }

        @Test
        @DisplayName("3stik_7 - Færdigvaccination med 1stik og 2. stik af samme")
        void case7() throws DGWSFault, dk.ssi.nsi.xml_schema._2013._01._01.DGWSFault, InterruptedException {
            final Instant effectuationTime1 = LocalDateTime.of(2020, 12, 1, 14, 0).toInstant(ZoneOffset.UTC);
            final Instant effectuationTime2 = LocalDateTime.of(2021, 9, 1, 14, 0).toInstant(ZoneOffset.UTC);

            IdwsUser user = nextUser();
            final Citizen citizen = ddvDriver.withCitizen(user).clear();
            citizen.addJohnsonVaccination(Date.from(effectuationTime1)); //program vaccination
            citizen.addJohnsonVaccination(Date.from(effectuationTime2)); //additional vaccination

            final GetCoronaPassResponse response = conductor.getCoronaPassesAfterInvalidation(user);
            final List<VaccinationEntry> euVaccinationEntries = conductor.getEuVaccinationEntriesOrderedByDate(CwtTestUtils.getEuVacPasses(response));
            assertEquals(2, euVaccinationEntries.size());

            assertThatVaccination(euVaccinationEntries.get(0))
                .isDoseOf(1, 1)
                .hasDateOfVaccination(effectuationTime1)
                .isJohnsonVaccination();
            assertThatVaccination(euVaccinationEntries.get(1))
                .isDoseOf(2, 1)
                .hasDateOfVaccination(effectuationTime2)
                .isJohnsonVaccination();

            final List<VaccinationEntry> dkInfoVaccinationEntries = conductor.getDKINFOPayload(CwtTestUtils.getDKInfoPass(response).get(0).getJWT()).getV();
            assertEquals(2, dkInfoVaccinationEntries.size());
            assertThatVaccination(dkInfoVaccinationEntries.get(0))
                .isDoseOf(1, 1)
                .hasDateOfVaccination(effectuationTime1)
                .isJohnsonVaccination();
            assertThatVaccination(dkInfoVaccinationEntries.get(1))
                .isDoseOf(2, 1)
                .hasDateOfVaccination(effectuationTime2)
                .isJohnsonVaccination();
        }

        @Test
        @DisplayName("3stik_8 - Færdigvaccination med 1stik og 2. stik af anden type")
        void case8() throws DGWSFault, dk.ssi.nsi.xml_schema._2013._01._01.DGWSFault, InterruptedException {
            final Instant effectuationTime1 = LocalDateTime.of(2020, 12, 1, 14, 0).toInstant(ZoneOffset.UTC);
            final Instant effectuationTime2 = LocalDateTime.of(2021, 9, 1, 14, 0).toInstant(ZoneOffset.UTC);

            IdwsUser user = nextUser();
            final Citizen citizen = ddvDriver.withCitizen(user).clear();
            citizen.addJohnsonVaccination(Date.from(effectuationTime1)); //program vaccination
            citizen.addPfizerVaccination(Date.from(effectuationTime2)); //additional vaccination

            final GetCoronaPassResponse response = conductor.getCoronaPassesAfterInvalidation(user);
            final List<VaccinationEntry> euVaccinationEntries = conductor.getEuVaccinationEntriesOrderedByDate(CwtTestUtils.getEuVacPasses(response));
            assertEquals(2, euVaccinationEntries.size());

            assertThatVaccination(euVaccinationEntries.get(0))
                .isDoseOf(1, 1)
                .hasDateOfVaccination(effectuationTime1)
                .isJohnsonVaccination();
            assertThatVaccination(euVaccinationEntries.get(1))
                .isDoseOf(2, 1)
                .hasDateOfVaccination(effectuationTime2)
                .isPfizerVaccination();

            final List<VaccinationEntry> dkInfoVaccinationEntries = conductor.getDKINFOPayload(CwtTestUtils.getDKInfoPass(response).get(0).getJWT()).getV();
            assertEquals(2, dkInfoVaccinationEntries.size());
            assertThatVaccination(dkInfoVaccinationEntries.get(0))
                .isDoseOf(1, 1)
                .hasDateOfVaccination(effectuationTime1)
                .isJohnsonVaccination();
            assertThatVaccination(dkInfoVaccinationEntries.get(1))
                .isDoseOf(2, 1)
                .hasDateOfVaccination(effectuationTime2)
                .isPfizerVaccination();
        }

    }

    @Nested
    @DisplayName("CBS DGWS Test")
    class DGWSServiceTest {

//        @Test
//        void dgws_service_ecpr_normal_flow() {
//            cbsdgwsDriver.getCoronaPasses("1208751RY9");
//        }

        @Test
        void dgws_service_normal_flow() throws Exception {
            final IdwsUser user = nextUser();

            final Instant effectuationTime = now.minus(Duration.ofDays(28L));

            ddvDriver.withCitizen(user)
                .clear()
                .addPfizerVaccination(Date.from(effectuationTime));
            conductor.getCoronaPassesAfterInvalidation(user);

            waitForAsserted(() -> {
                final GetCoronaPassResponse response = cbsdgwsDriver.getCoronaPasses(user.cpr());
                assertThat(response)
                    .hasEuVacPasses(1);
            });
        }
    }


    private LocalDateTime zeroHourMinus(TemporalAmount amount) {
        LocalDateTime zeroHour = LocalDateTime.now().withHour(10).withMinute(0).withSecond(0).withNano(0);
        if (LocalDateTime.now().getHour() >= 23) //KIDS data is reloaded on test environments every day at 23:00 setting "zero hour" as 10:00 next day
            zeroHour = zeroHour.plusDays(1);
        return zeroHour.minus(amount);
    }


}
