/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs;

import com.google.common.collect.Range;
import dk.sundhedsdatastyrelsen.cbs.cbsdriver.CBSEuSigningDriver;
import dk.sundhedsdatastyrelsen.cbs.cbsdriver.RequestModifyingSoapHandler;
import dk.sundhedsdatastyrelsen.cbs.configuration.IntegrationTestConfiguration;
import dk.sundhedsdatastyrelsen.cbs.test.TestWatcherLogger;
import dk.sundhedsdatastyrelsen.cbs.util.CwtTestUtils;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.CountryOfTest;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.DiseaseAgent;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.GetSignedEUCovidTestCertificateRequest;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.GetSignedEUCovidTestCertificateResponse;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.TestResult;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.TestSample;
import dk.sundhedsdatastyrelsen.coronapass._2021._07.eu_covid_test_certificate_signing.TestType;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.extension.ExtendWith;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit.jupiter.SpringExtension;
import se.digg.dgc.payload.v1.DigitalGreenCertificate;
import se.digg.dgc.payload.v1.RecoveryEntry;
import se.digg.dgc.payload.v1.TestEntry;
import se.digg.dgc.service.DGCDecoder;

import javax.xml.ws.soap.SOAPFaultException;
import java.time.Duration;
import java.time.Instant;
import java.time.LocalDate;
import java.time.Month;
import java.time.Period;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.temporal.TemporalAmount;
import java.util.List;
import java.util.function.Consumer;
import java.util.stream.Stream;

import static com.google.common.truth.Truth.assertThat;
import static dk.sundhedsdatastyrelsen.cbs.TestConstants.MANUFACTURER_HSC;
import static dk.sundhedsdatastyrelsen.cbs.TestConstants.MANUFACTURER_NOT_HSC;
import static dk.sundhedsdatastyrelsen.cbs.util.LabResultTester.assertThatTestEntry;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertThrows;

@ExtendWith({SpringExtension.class, TestWatcherLogger.class})
@ContextConfiguration(classes = {IntegrationTestConfiguration.class})
@TestInstance(TestInstance.Lifecycle.PER_CLASS)
@DisplayName("EU Signing Service Tests")
class EUSigningITCase {

    public static final ZoneId ZONE_ID = ZoneId.of("Europe/Copenhagen");
    public static final String TESTING_CENTER = "Århus Ø Svæveflyverklub - Falck A/S";

    @Autowired
    private CBSEuSigningDriver cbsEuSigningDriver;

    @Autowired
    private DGCDecoder decoder;


    @Test
    @DisplayName("EUSS-1.1 - Test Certificate is issued for Negative PCR test sampled now, expires in 1 week")
    void case_euss_1_1() {
        final ZonedDateTime sampleCollected = ZonedDateTime.now(ZONE_ID);

        final GetSignedEUCovidTestCertificateRequest request = createRequest(
            TestType.PCR,
            TestResult.NEGATIVE,
            sampleCollected,
            MANUFACTURER_HSC);

        final GetSignedEUCovidTestCertificateResponse response = cbsEuSigningDriver.getSignedEUCovidTestCertificate(request);
        verifyResponse(response);

        final TestEntry testEntry = getTestEntry(response);

        assertThatTestEntry(testEntry)
            .isNegative()
            .isPcr()
            .hasManufacturer(MANUFACTURER_HSC)
            .hasTestCenter(TESTING_CENTER)
            .hasSampleTime(sampleCollected.toLocalDateTime());

        assertExpiresIn(response, Period.ofWeeks(1));
    }

    @Test
    @DisplayName("EUSS-1.2 - Test Certificate is issued for Negative PCR test sampled 3 days ago, expires in 4 days")
    void case_euss_1_2() {
        final ZonedDateTime sampleCollected = ZonedDateTime.now(ZONE_ID).minus(Duration.ofDays(3));

        final GetSignedEUCovidTestCertificateRequest request = createRequest(
            TestType.PCR,
            TestResult.NEGATIVE,
            sampleCollected,
            MANUFACTURER_HSC);

        final GetSignedEUCovidTestCertificateResponse response = cbsEuSigningDriver.getSignedEUCovidTestCertificate(request);
        verifyResponse(response);

        final TestEntry testEntry = getTestEntry(response);

        assertThatTestEntry(testEntry)
            .isNegative()
            .isPcr()
            .hasManufacturer(MANUFACTURER_HSC)
            .hasTestCenter(TESTING_CENTER)
            .hasSampleTime(sampleCollected.toLocalDateTime());

        assertExpiresIn(response, Duration.ofDays(4L));
    }

    @Test
    @DisplayName("EUSS-2.1 - Test Certificate is issued for Negative AG test from approved manufacturer sampled now, expires in 1 week")
    void case_euss_2_1() {
        final ZonedDateTime sampleCollected = ZonedDateTime.now(ZONE_ID);

        final GetSignedEUCovidTestCertificateRequest request = createRequest(
            TestType.RAT,
            TestResult.NEGATIVE,
            sampleCollected,
            MANUFACTURER_HSC);

        final GetSignedEUCovidTestCertificateResponse response = cbsEuSigningDriver.getSignedEUCovidTestCertificate(request);
        verifyResponse(response);

        final TestEntry testEntry = getTestEntry(response);

        assertThatTestEntry(testEntry)
            .isNegative()
            .isAG()
            .hasManufacturer(MANUFACTURER_HSC)
            .hasTestCenter(TESTING_CENTER)
            .hasSampleTime(sampleCollected.toLocalDateTime());

        assertExpiresIn(response, Period.ofWeeks(1));
    }

    @Test
    @DisplayName("EUSS-2.2 - Test Certificate is issued for Negative AG test from approved manufacturer sampled 3 days ago, expires in 4 days")
    void case_euss_2_2() {
        final ZonedDateTime sampleCollected = ZonedDateTime.now(ZONE_ID).minus(Duration.ofDays(3));

        final GetSignedEUCovidTestCertificateRequest request = createRequest(
            TestType.RAT,
            TestResult.NEGATIVE,
            sampleCollected,
            MANUFACTURER_HSC);

        final GetSignedEUCovidTestCertificateResponse response = cbsEuSigningDriver.getSignedEUCovidTestCertificate(request);
        verifyResponse(response);

        final TestEntry testEntry = getTestEntry(response);

        assertThatTestEntry(testEntry)
            .isNegative()
            .isAG()
            .hasManufacturer(MANUFACTURER_HSC)
            .hasTestCenter(TESTING_CENTER)
            .hasSampleTime(sampleCollected.toLocalDateTime());

        assertExpiresIn(response, Duration.ofDays(4L));
    }

    @Test
    @DisplayName("EUSS-3.1 - Recovery Certificate is issued for Positive PCR test sampled now, expires in 180 days")
    void case_euss_3_1() {
        final ZonedDateTime sampleCollected = ZonedDateTime.now(ZONE_ID);

        final GetSignedEUCovidTestCertificateRequest request = createRequest(
            TestType.PCR,
            TestResult.POSITIVE,
            sampleCollected,
            "");

        final GetSignedEUCovidTestCertificateResponse response = cbsEuSigningDriver.getSignedEUCovidTestCertificate(request);
        verifyResponse(response);

        final RecoveryEntry recoveryEntry = getRecoveryEntry(response);

        assertEquals(sampleCollected.toLocalDate(), recoveryEntry.getFr());
        assertEquals(sampleCollected.plusDays(11L).toLocalDate(), recoveryEntry.getDf());
        assertEquals(sampleCollected.plusDays(180L).toLocalDate(), recoveryEntry.getDu());

        assertExpiresIn(response, Duration.ofDays(180L));
    }

    @Test
    @DisplayName("EUSS-3.2 - Recovery Certificate is issued for Positive PCR test sampled 10 days ago, expires in 170 days")
    void case_euss_3_2() {
        final ZonedDateTime sampleCollected = ZonedDateTime.now(ZONE_ID).minus(Duration.ofDays(10));

        final GetSignedEUCovidTestCertificateRequest request = createRequest(
            TestType.PCR,
            TestResult.POSITIVE,
            sampleCollected,
            "");

        final GetSignedEUCovidTestCertificateResponse response = cbsEuSigningDriver.getSignedEUCovidTestCertificate(request);
        verifyResponse(response);

        final RecoveryEntry recoveryEntry = getRecoveryEntry(response);

        assertEquals(sampleCollected.toLocalDate(), recoveryEntry.getFr());
        assertEquals(sampleCollected.plusDays(11L).toLocalDate(), recoveryEntry.getDf());
        assertEquals(sampleCollected.plusDays(180L).toLocalDate(), recoveryEntry.getDu());

        assertExpiresIn(response, Duration.ofDays(170L));
    }

    @Test
    @DisplayName("EUSS-4.1 - Certificate is not issued for Positive PCR test sampled 181 days ago, with reason 'Test is expired'")
    void case_euss_4_1() {
        final ZonedDateTime sampleCollected = ZonedDateTime.now(ZONE_ID).minusDays(181L);

        final GetSignedEUCovidTestCertificateRequest request = createRequest(
            TestType.PCR,
            TestResult.POSITIVE,
            sampleCollected,
            "");

        final SOAPFaultException exception = assertThrows(SOAPFaultException.class, () -> cbsEuSigningDriver.getSignedEUCovidTestCertificate(request));

        assertEquals("Test is expired", exception.getFault().getFaultString());
    }

    @Test
    @DisplayName("EUSS-4.2 - Certificate is not issued for Negative PCR test sampled 8 days ago, with reason 'Test is expired'")
    void case_euss_4_2() {
        final ZonedDateTime sampleCollected = ZonedDateTime.now(ZONE_ID).minusDays(8L);

        final GetSignedEUCovidTestCertificateRequest request = createRequest(
            TestType.PCR,
            TestResult.NEGATIVE,
            sampleCollected,
            "");

        final SOAPFaultException exception = assertThrows(SOAPFaultException.class, () -> cbsEuSigningDriver.getSignedEUCovidTestCertificate(request));

        assertEquals("Test is expired", exception.getFault().getFaultString());
    }

    @Test
    @DisplayName("EUSS-4.3 - Certificate is not issued for Negative AG test sampled 8 days ago, with reason 'Test is expired'")
    void case_euss_4_3() {
        final ZonedDateTime sampleCollected = ZonedDateTime.now(ZONE_ID).minusDays(8L);

        final GetSignedEUCovidTestCertificateRequest request = createRequest(
            TestType.RAT,
            TestResult.NEGATIVE,
            sampleCollected,
            MANUFACTURER_HSC);

        final SOAPFaultException exception = assertThrows(SOAPFaultException.class, () -> cbsEuSigningDriver.getSignedEUCovidTestCertificate(request));

        assertEquals("Test is expired", exception.getFault().getFaultString());
    }

    @Test
    @DisplayName("EUSS-4.4 - Certificate is not issued for Positive PCR test sampled in the future, with reason 'SampleCollected cannot be in the future'")
    void case_euss_4_4() {
        final ZonedDateTime sampleCollected = ZonedDateTime.now(ZONE_ID).plusDays(1L);

        final GetSignedEUCovidTestCertificateRequest request = createRequest(
            TestType.PCR,
            TestResult.POSITIVE,
            sampleCollected,
            "");

        final SOAPFaultException exception = assertThrows(SOAPFaultException.class, () -> cbsEuSigningDriver.getSignedEUCovidTestCertificate(request));

        assertEquals("SampleCollected cannot be in the future", exception.getFault().getFaultString());
    }

    @Test
    @DisplayName("EUSS-4.5 - Certificate is not issued for Negative PCR test sampled in the future, with reason 'SampleCollected cannot be in the future'")
    void case_euss_4_5() {
        final ZonedDateTime sampleCollected = ZonedDateTime.now(ZONE_ID).plusDays(1L);

        final GetSignedEUCovidTestCertificateRequest request = createRequest(
            TestType.PCR,
            TestResult.NEGATIVE,
            sampleCollected,
            "");

        final SOAPFaultException exception = assertThrows(SOAPFaultException.class, () -> cbsEuSigningDriver.getSignedEUCovidTestCertificate(request));

        assertEquals("SampleCollected cannot be in the future", exception.getFault().getFaultString());
    }

    @Test
    @DisplayName("EUSS-4.6 - Certificate is not issued for Negative AG test sampled in the future, with reason 'SampleCollected cannot be in the future'")
    void case_euss_4_6() {
        final ZonedDateTime sampleCollected = ZonedDateTime.now(ZONE_ID).plusDays(1L);

        final GetSignedEUCovidTestCertificateRequest request = createRequest(
            TestType.RAT,
            TestResult.NEGATIVE,
            sampleCollected,
            MANUFACTURER_HSC);

        final SOAPFaultException exception = assertThrows(SOAPFaultException.class, () -> cbsEuSigningDriver.getSignedEUCovidTestCertificate(request));

        assertEquals("SampleCollected cannot be in the future", exception.getFault().getFaultString());
    }

    @Test
    @DisplayName("EUSS-5.1 - Certificate is not issued for Positive AG test sampled now, with reason 'Cannot issue covid test certificate for positive test'")
    void case_euss_5_1() {
        final ZonedDateTime sampleCollected = ZonedDateTime.now(ZONE_ID);

        final GetSignedEUCovidTestCertificateRequest request = createRequest(
            TestType.RAT,
            TestResult.POSITIVE,
            sampleCollected,
            MANUFACTURER_HSC);

        final SOAPFaultException exception = assertThrows(SOAPFaultException.class, () -> cbsEuSigningDriver.getSignedEUCovidTestCertificate(request));

        assertEquals("Cannot issue covid test certificate for positive test", exception.getFault().getFaultString());
    }

    @Test
    @DisplayName("EUSS-5.2 - Certificate is not issued for Negative AG test from unapproved manufacturer sampled now, with reason 'Test is not HSC Approved'")
    void case_euss_5_2() {
        final ZonedDateTime sampleCollected = ZonedDateTime.now(ZONE_ID);

        final GetSignedEUCovidTestCertificateRequest request = createRequest(
            TestType.RAT,
            TestResult.NEGATIVE,
            sampleCollected,
            MANUFACTURER_NOT_HSC);

        final SOAPFaultException exception = assertThrows(SOAPFaultException.class, () -> cbsEuSigningDriver.getSignedEUCovidTestCertificate(request));

        assertEquals("Test is not HSC Approved", exception.getFault().getFaultString());
    }

    @DisplayName("EUSS-6 - Required Fields")
    @ParameterizedTest(name = "EUSS-6.{index} - Field [{1}] is required")
    @MethodSource("missingFieldsRequests")
    void requiredFields(GetSignedEUCovidTestCertificateRequest request, String field) {
        final SOAPFaultException exception = assertThrows(SOAPFaultException.class, () -> cbsEuSigningDriver.getSignedEUCovidTestCertificate(request));

        assertThat(exception.getFault().getFaultString()).contains("Unmarshalling Error");
        assertThat(exception.getFault().getFaultString()).contains(field);
    }

    @DisplayName("EUSS-7 - Restricted Fields")
    @ParameterizedTest(name = "EUSS-7.{index} - Field [{0}] cannot be [{1}] must be in {2}")
    @MethodSource("restrictedFields")
    void restrictedFieldsTest(String field, String value, String enumeration) {
        final GetSignedEUCovidTestCertificateRequest request = validRequest();

        RequestModifyingSoapHandler.CONTEXT.setReplace(true).setElementName(field).setValue(value);

        final SOAPFaultException exception = assertThrows(SOAPFaultException.class, () -> cbsEuSigningDriver.getSignedEUCovidTestCertificate(request));

        assertEquals(String.format("Unmarshalling Error: cvc-enumeration-valid: Value '%s' is not facet-valid with respect to enumeration '%s'. It must be a value from the enumeration. ", value, enumeration) ,exception.getFault().getFaultString());

        RequestModifyingSoapHandler.CONTEXT.setReplace(false);
    }

    @SuppressWarnings("unused") // used by ParameterizedTest
    private static Stream<Arguments> missingFieldsRequests() {
        return Stream.of(
            Arguments.of(validRequestBut(r -> r.setPersonIdentifier(null)), "PersonIdentifier"),
            Arguments.of(validRequestBut(r -> r.setPayload(null)), "Payload"),
            Arguments.of(validRequestBut(r -> r.getPayload().setTypeOfTest(null)), "TypeOfTest"),
            Arguments.of(validRequestBut(r -> r.getPayload().setTestResult(null)), "TestResult"),
            Arguments.of(validRequestBut(r -> r.getPayload().setManufacturer(null)), "Manufacturer"),
            Arguments.of(validRequestBut(r -> r.getPayload().setSampleCollected(null)), "SampleCollected"),
            Arguments.of(validRequestBut(r -> r.getPayload().setTestingCenter(null)), "TestingCenter"),
            Arguments.of(validRequestBut(r -> r.getPayload().setDiseaseAgentTargeted(null)), "DiseaseAgentTargeted"),
            Arguments.of(validRequestBut(r -> r.getPayload().setCountryOfTest(null)), "CountryOfTest"),
            Arguments.of(validRequestBut(r -> r.getPayload().setFamilyName(null)), "FamilyName"),
            Arguments.of(validRequestBut(r -> r.getPayload().setGivenName(null)), "GivenName"),
            Arguments.of(validRequestBut(r -> r.getPayload().setDateOfBirth(null)), "DateOfBirth")
        );
    }

    @SuppressWarnings("unused") // used by ParameterizedTest
    private static Stream<Arguments> restrictedFields() {
        return Stream.of(
            Arguments.of("TypeOfTest", "LP-456456-2", "[LP217198-3, LP6464-4]"),
            Arguments.of("TestResult", "1234", "[260373001, 260415000]"),
            Arguments.of("DiseaseAgentTargeted", "840539023", "[840539006]"),
            Arguments.of("CountryOfTest", "GB", "[DK]")
        );
    }

    private static GetSignedEUCovidTestCertificateRequest validRequestBut(Consumer<GetSignedEUCovidTestCertificateRequest> consumer) {
        final GetSignedEUCovidTestCertificateRequest request = validRequest();
        consumer.accept(request);
        return request;
    }

    private static GetSignedEUCovidTestCertificateRequest validRequest() {
        return createRequest(
                TestType.RAT,
                TestResult.NEGATIVE,
                ZonedDateTime.now(ZONE_ID),
                MANUFACTURER_HSC);
    }


    private static GetSignedEUCovidTestCertificateRequest createRequest(TestType testType, TestResult testResult, ZonedDateTime sampleCollected, String manufacturer) {
        final GetSignedEUCovidTestCertificateRequest request = new GetSignedEUCovidTestCertificateRequest();
        final TestSample payload = new TestSample();
        payload.setGivenName("Brian");
        payload.setFamilyName("London");
        payload.setDateOfBirth(LocalDate.of(1992, Month.DECEMBER, 17));
        payload.setCountryOfTest(CountryOfTest.DK);
        payload.setDiseaseAgentTargeted(DiseaseAgent.COVID_19);
        payload.setTestingCenter(TESTING_CENTER);
        payload.setManufacturer(manufacturer);
        payload.setSampleCollected(sampleCollected);
        payload.setTypeOfTest(testType);
        payload.setTestResult(testResult);

        request.setPersonIdentifier("12345678");
        request.setPayload(payload);
        return request;
    }

    private void assertExpiresIn(GetSignedEUCovidTestCertificateResponse response, TemporalAmount duration) {
        final Instant expiration = CwtTestUtils.getCwtFromToken(response.getCWT()).getExpiration();
        final Instant expected = Instant.now().plus(duration);
        assertThat(expiration).isIn(Range.closed(expected.minusSeconds(10L), expected.plusSeconds(10L)));
    }

    private RecoveryEntry getRecoveryEntry(GetSignedEUCovidTestCertificateResponse response) {
        final DigitalGreenCertificate certificate = CwtTestUtils.getEUPayload(response.getCWT(), decoder);
        return CwtTestUtils.getRecoveryEntry(certificate);
    }

    private TestEntry getTestEntry(GetSignedEUCovidTestCertificateResponse response) {
        final DigitalGreenCertificate certificate = CwtTestUtils.getEUPayload(response.getCWT(), decoder);

        final List<TestEntry> testEntries = CwtTestUtils.getTestEntriesOrderedBySampleTime(certificate);

        assertEquals(1, testEntries.size());

        return testEntries.get(0);
    }

    private void verifyResponse(GetSignedEUCovidTestCertificateResponse response) {
        assertNotNull(response);
        assertEquals("1.3.0", response.getSchemaVersion());
        assertEquals("https://github.com/ehn-dcc-development/ehn-dcc-schema", response.getDocumentation());
        assertNotNull(response.getCertificateIdentifier());
        assertNotNull(response.getCWT());
    }

}
