/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs;

import dk.sundhedsdatastyrelsen.cbs.cbsdriver.CBSDriver;
import dk.sundhedsdatastyrelsen.cbs.datacache.DataCacheInvalidatorDriver;
import dk.sundhedsdatastyrelsen.cbs.datacache.DataFetcherDriver;
import dk.sundhedsdatastyrelsen.cbs.ddvdriver.DDVDriver;
import dk.sundhedsdatastyrelsen.cbs.nasdriver.NASDriver;
import dk.sundhedsdatastyrelsen.cbs.util.CoronaPassAssertions;
import dk.sundhedsdatastyrelsen.cbs.util.CwtTestUtils;
import dk.sundhedsdatastyrelsen.cbs.util.Wait;
import dk.sundhedsdatastyrelsen.cbs.webservice.idws.IdwsUser;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassResponse;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.PassJWT;
import dk.sundhedsdatastyrelsen.coronapass.ws.JobStatus;
import dk.sundhedsdatastyrelsen.coronapass.ws.JobStatusNotification;
import dk.vaccinationsregister.schemas._2013._12._01.DGWSFault;
import se.digg.dgc.encoding.DGCConstants;
import se.digg.dgc.payload.v1.*;
import se.digg.dgc.service.DGCDecoder;
import se.digg.dgc.signatures.cwt.Cwt;

import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.Comparator;
import java.util.Date;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;

import static com.google.common.collect.MoreCollectors.onlyElement;
import static org.junit.jupiter.api.Assertions.*;

public class TestConductor {

    private final DDVDriver ddvDriver;
    private final NASDriver nasDriver;
    private final CBSDriver cbsDriver;
    private final DGCDecoder decoder;
    private final DataFetcherDriver dataFetcherDriver;
    private final DataCacheInvalidatorDriver dataCacheInvalidatorDriver;

    private Instant now;

    public TestConductor(DDVDriver ddvDriver, NASDriver nasDriver, CBSDriver cbsDriver, DGCDecoder decoder, DataFetcherDriver dataFetcherDriver, DataCacheInvalidatorDriver dataCacheInvalidatorDriver) {
        this.ddvDriver = ddvDriver;
        this.nasDriver = nasDriver;
        this.cbsDriver = cbsDriver;
        this.decoder = decoder;
        this.dataFetcherDriver = dataFetcherDriver;
        this.dataCacheInvalidatorDriver = dataCacheInvalidatorDriver;
        now = Instant.now();
    }

    public TestConductor setNow(Instant now) {
        this.now = now;
        return this;
    }

    public List<TestEntry> getEuTestEntriesOrderedBySampleTime(List<PassJWT> passes) {
        return passes.stream()
            .filter(pass -> pass.getType().equals("EU"))
            .filter(pass -> pass.getProperties().stream().anyMatch(prop -> prop.getKey().equals("eu:keytype") && prop.getValue().equals("lab")))
            .map(PassJWT::getJWT)
            .map(this::getEUPayload)
            .map(Eudgc::getT)
            .flatMap(List::stream)
            .sorted(Comparator.comparing(TestEntry::getSc))
            .collect(Collectors.toList());
    }


    public List<VaccinationEntry> getEuVaccinationEntriesOrderedByDate(List<PassJWT> passes) {
        return passes.stream()
            .filter(pass -> pass.getType().equals("EU"))
            .filter(pass -> pass.getProperties().stream().anyMatch(prop -> prop.getKey().equals("eu:keytype") && prop.getValue().equals("vac")))
            .map(PassJWT::getJWT)
            .map(this::getEUPayload)
            .flatMap(cert -> cert.getV().stream())
            .sorted(Comparator.comparing(VaccinationEntry::getDt))
            .collect(Collectors.toList());
    }

    public GetCoronaPassResponse waitForInvalidationAndGetFutureResult(IdwsUser idwsUser) throws InterruptedException {
        return waitForInvalidationAndGetFutureResult(idwsUser, idwsUser);
    }

    public GetCoronaPassResponse waitForInvalidationAndGetFutureResult(IdwsUser caller, IdwsUser requestUser) throws InterruptedException {
        return Wait.waitFor("passport to become invalidated", () -> {
            dataCacheInvalidatorDriver.triggerInvalidation();
            return cbsDriver.getCoronaPass(caller, requestUser.cpr());
        }, r -> r.getFutureResult() != null);
    }

    public GetCoronaPassResponse waitForInvalidationAndGetFutureResult(IdwsUser caller, IdwsUser requestUser, IdwsUser principalUser) throws InterruptedException {
        return Wait.waitFor("passport to become invalidated", () -> {
            dataCacheInvalidatorDriver.triggerInvalidation();
            return cbsDriver.getCoronaPass(caller, requestUser.cpr(), principalUser.cpr());
        }, r -> r.getFutureResult() != null);
    }

    public void clearDdvAndWaitForInvalidationAndDataFetch(IdwsUser user) throws dk.ssi.nsi.xml_schema._2013._01._01.DGWSFault, DGWSFault, InterruptedException {
        ddvDriver.withCitizen(user)
            .clear()
            .addPfizerVaccination(Date.from(now.minus(48, ChronoUnit.DAYS)))
            .deleteVaccination();

        final GetCoronaPassResponse response = waitForInvalidationAndGetFutureResult(user);
        fetchDataAndWaitForJobStatusNotification(response.getFutureResult().getJobId());
    }

    public GetCoronaPassResponse getCoronaPassesAfterInvalidation(IdwsUser idwsUser) throws InterruptedException {
        final String jobId = waitForInvalidationAndGetFutureResult(idwsUser).getFutureResult().getJobId();
        fetchDataAndWaitForJobStatusNotification(jobId);
        return getCoronaPasses(idwsUser);
    }

    public GetCoronaPassResponse getCoronaPassesAfterInvalidation(IdwsUser actingUser, IdwsUser principalUser) throws InterruptedException {
        final String jobId = waitForInvalidationAndGetFutureResult(actingUser, principalUser, principalUser).getFutureResult().getJobId();
        fetchDataAndWaitForJobStatusNotification(jobId);
        return getCoronaPasses(actingUser, principalUser.cpr(), principalUser.cpr());
    }

    public GetCoronaPassResponse getCoronaPasses(IdwsUser idwsUser) throws InterruptedException {
        GetCoronaPassResponse coronaPass = cbsDriver.getCoronaPass(idwsUser);

        if (coronaPass.getFutureResult() != null) {
            final JobStatusNotification jobStatusNotification = fetchDataAndWaitForJobStatusNotification(coronaPass.getFutureResult().getJobId());
            assertEquals(JobStatus.OK, jobStatusNotification.getJobStatus(), "Unexpected JobStatus for JobId: " + jobStatusNotification.getJobId());
            coronaPass = Wait.waitFor("passport to be returned", () -> cbsDriver.getCoronaPass(idwsUser), r -> r.getCoronaPasses() != null);
        }

        return coronaPass;
    }

    public GetCoronaPassResponse getCoronaPasses(IdwsUser actingUser, String requestCpr, String principalUserCpr) throws InterruptedException {
        GetCoronaPassResponse coronaPass = cbsDriver.getCoronaPass(actingUser, requestCpr, principalUserCpr);

        if (coronaPass.getFutureResult() != null) {
            final JobStatusNotification jobStatusNotification = fetchDataAndWaitForJobStatusNotification(coronaPass.getFutureResult().getJobId());
            assertEquals(JobStatus.OK, jobStatusNotification.getJobStatus(), "Unexpected JobStatus for JobId: " + jobStatusNotification.getJobId());
            coronaPass = cbsDriver.getCoronaPass(actingUser, requestCpr, principalUserCpr);
        }

        return coronaPass;
    }


    public JobStatusNotification fetchDataAndWaitForJobStatusNotification(String jobId) throws InterruptedException {
        return Wait.waitFor("NAS message with jobId " + jobId + " to arrive", () -> {
            dataFetcherDriver.triggerDataFetch();
            return nasDriver.getNotifications().stream().filter(m -> m.getJobId().equals(jobId)).findAny().orElse(null);

        }, Objects::nonNull);
    }


    public void validateTokensAreValidAndFresh(List<PassJWT> passList) {
        for (PassJWT passJWT : passList) {

            if (passJWT.getType().startsWith("eu")) {
                // This will validate the signature:
                DigitalGreenCertificate dgc = getEUPayload(passJWT.getJWT());
                assertNotNull(dgc);
            }

            Cwt cwt = CwtTestUtils.getCwtFromToken(passJWT.getJWT().substring(4));

            assertTrue(
                cwt.getIssuedAt().isAfter(Instant.now().minus(CoronaPassAssertions.TOKEN_ISSUED_AT_SLACK)),
                "Token of type " + passJWT.getType() + " doesn't seem to be a freshly issued token - was issued at " + cwt.getIssuedAt()
            );
        }
    }

    public RecoveryEntry getRecoveryEntry(List<PassJWT> euRecPasses) {
        return euRecPasses.stream()
            .map(PassJWT::getJWT)
            .map(this::getEUPayload)
            .map(Eudgc::getR)
            .flatMap(List::stream)
            .collect(onlyElement());
    }

    public DigitalGreenCertificate getEUPayload(String token) {
        try {
            return decoder.decode(token);
        } catch (Exception e) {
            throw new RuntimeException(e.getMessage(), e);
        }
    }

    public DigitalGreenCertificate getDKINFOPayload(String jwt) {
        return getEUPayload(jwt.replaceFirst("DK3:", DGCConstants.DGC_V1_HEADER));
    }


}
