/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.configuration;

import dk.sosi.seal.model.IdentityToken;
import dk.sosi.seal.model.LibertyRequestDOMEnhancer;
import dk.sundhedsdatastyrelsen.cbs.webservice.idws.BootstrapHelper;
import dk.sundhedsdatastyrelsen.cbs.webservice.idws.IdwsTokenHelper;
import dk.sundhedsdatastyrelsen.cbs.webservice.idws.IdwsUser;
import org.apache.commons.io.IOUtils;
import org.apache.cxf.binding.soap.SoapMessage;
import org.apache.cxf.binding.soap.interceptor.AbstractSoapInterceptor;
import org.apache.cxf.interceptor.Fault;
import org.apache.cxf.io.CachedOutputStream;
import org.apache.cxf.phase.Phase;
import org.w3c.dom.Document;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import java.io.OutputStream;

public class LibertyRequestDecoratorInterceptor extends AbstractSoapInterceptor implements BorgerContext, RequestContext {

    private final BootstrapHelper bootstrapHelper;

    private final IdwsTokenHelper idwsTokenHelper;

    private IdwsUser user;

    private String audience;

    private String procuration;

    private String messageId;

    public LibertyRequestDecoratorInterceptor(IdwsTokenHelper idwsTokenHelper) {
        super(Phase.PRE_STREAM);

        this.bootstrapHelper = new BootstrapHelper();
        this.idwsTokenHelper = idwsTokenHelper;
    }


    @Override
    public void clearContext() {

        audience = null;
        procuration = null;
        messageId = null;
        user = null;
    }


    @Override
    public void setContext(String audience, IdwsUser user, String procuration) {
        this.audience = audience;
        this.procuration = procuration;
        this.user = user;
    }


    private IdentityToken getIdentityToken() {
        if (audience != null && user != null) {
            return idwsTokenHelper.getIdentityToken(user.cpr(), audience, procuration, bootstrapHelper.buildBootstrapAssertion(user));
        }
        return null;
    }

    @Override
    public void handleMessage(SoapMessage message) throws Fault {
        OutputStream os = message.getContent(OutputStream.class);
        CachedOutputStream cs = new CachedOutputStream();
        message.setContent(OutputStream.class, cs);
        message.getInterceptorChain().doIntercept(message);

        try {
            cs.flush();
            IOUtils.closeQuietly(cs);
            CachedOutputStream csnew = (CachedOutputStream) message.getContent(OutputStream.class);
            DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();
            documentBuilderFactory.setNamespaceAware(true);

            Document doc = documentBuilderFactory.newDocumentBuilder().parse(csnew.getInputStream());
            handleDocument(doc);

            csnew.flush();
            IOUtils.closeQuietly(csnew);


            TransformerFactory tf = TransformerFactory.newInstance();
            Transformer trans = tf.newTransformer();
            trans.transform(new DOMSource(doc), new StreamResult(os));
            os.flush();
            message.setContent(OutputStream.class, os);
            IOUtils.closeQuietly(os);


        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    protected void handleDocument(Document doc) {
        IdentityToken id = getIdentityToken();
        if (id != null) {
            LibertyRequestDOMEnhancer enhancer = idwsTokenHelper.getOioIdwsFactory().createRequestDOMEnhancer(doc);
            enhancer.setIdentityToken(id);
            if (messageId != null) {
                enhancer.setWSAddressingMessageID(messageId);
            }
            enhancer.setWSAddressingAction("action");
            enhancer.enhanceAndSign();
        }
    }


    @Override
    public void setRequestContext(String messageId) {
        this.messageId = messageId;
    }
}
