/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.ddvdriver;

import dk.ssi.nsi.xml_schema._2013._01._01.DumpRestore;
import dk.ssi.nsi.xml_schema._2013._01._01.DumpRestorePortType;
import dk.ssi.nsi.xml_schema._2013._01._01.IdentifierType;
import dk.ssi.nsi.xml_schema._2013._01._01.ResetPatientsRequestType;
import dk.sundhedsdatastyrelsen.cbs.ddvdriver.dgws.DgwsClientMessageHandler;
import dk.sundhedsdatastyrelsen.cbs.ddvdriver.dgws.IdCardFactory;
import dk.sundhedsdatastyrelsen.cbs.ddvdriver.dgws.WhitelistingClientMessageHandler;
import dk.sundhedsdatastyrelsen.cbs.dgws.STSConfig;
import dk.sundhedsdatastyrelsen.cbs.dgws.http.HttpClientServiceImpl;
import dk.sundhedsdatastyrelsen.cbs.dgws.http.HttpExecutor;
import dk.sundhedsdatastyrelsen.cbs.webservice.idws.IdwsUser;
import dk.vaccinationsregister.schemas._2013._12._01.*;
import org.apache.cxf.ext.logging.LoggingFeature;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import javax.xml.bind.JAXBElement;
import javax.xml.namespace.QName;
import javax.xml.soap.SOAPException;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.handler.Handler;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.Date;
import java.util.List;

import static dk.sundhedsdatastyrelsen.cbs.configuration.IntegrationTestConfiguration.CBS_IT_LOG_DDV;
import static dk.sundhedsdatastyrelsen.cbs.configuration.IntegrationTestConfiguration.CBS_IT_LOG_STS;

public class DDVDriver {
    private final VaccineData vaccineData;
    private final VaccinationsPortType20131201 ddvPort;
    private final DumpRestorePortType dumpRestorePort;

    public DDVDriver(String stsUrl, String ddvUrl, String ddvDumpRestoreUrl, VaccineData vaccineData) throws SOAPException {
        this.vaccineData = vaccineData;

        STSConfig stsConfig = new STSConfig(
            stsUrl + "/NewSecurityTokenService",
            "Sundhedsdatastyrelsen",
            "20921897",
            "Sundhedsdatastyrelsen - COVID-19 Borgerstatus.jks",
            "Test1234",
            "CBS"
        );

        if (CBS_IT_LOG_STS) {
            Logger.getLogger("org.apache.http.wire").setLevel(Level.DEBUG);
        }
        HttpClientServiceImpl httpClient = new HttpClientServiceImpl(new HttpExecutor());
        IdCardFactory idCardFactory = new IdCardFactory(stsConfig, httpClient);
        ddvPort = setupDdvPort(idCardFactory, stsConfig, ddvUrl);
        dumpRestorePort = setupDumpRestorePort(idCardFactory, stsConfig, ddvDumpRestoreUrl);
    }

    private VaccinationsPortType20131201 setupDdvPort(IdCardFactory idCardFactory, STSConfig stsConfig, String endpointURL) throws SOAPException {
        VaccinationsService20131201 service;
        if (CBS_IT_LOG_DDV) {
            final LoggingFeature loggingFeature = new LoggingFeature();
            loggingFeature.setPrettyLogging(true);
            service = new VaccinationsService20131201(loggingFeature);
        } else {
            service = new VaccinationsService20131201();
        }

        VaccinationsPortType20131201 port = service.getVaccinationsPort();
        BindingProvider bp = (BindingProvider) port;
        bp.getRequestContext().put(BindingProvider.ENDPOINT_ADDRESS_PROPERTY, endpointURL);

        List<Handler> handlerChain = bp.getBinding().getHandlerChain();
        handlerChain.add(new DgwsClientMessageHandler(idCardFactory));
        handlerChain.add(new WhitelistingClientMessageHandler(stsConfig.getOrganizationName(), stsConfig.getSystemName(), "1.0.0", stsConfig.getOrganizationName()));
        bp.getBinding().setHandlerChain(handlerChain);

        return port;
    }

    private DumpRestorePortType setupDumpRestorePort(IdCardFactory idCardFactory, STSConfig stsConfig, String endpointURL) throws SOAPException {
        DumpRestore service;
        if (CBS_IT_LOG_DDV) {
            final LoggingFeature loggingFeature = new LoggingFeature();
            loggingFeature.setPrettyLogging(true);
            service = new DumpRestore(loggingFeature);
        } else {
            service = new DumpRestore();
        }

        DumpRestorePortType port = service.getDumpRestorePort();
        BindingProvider bp = (BindingProvider) port;
        bp.getRequestContext().put(BindingProvider.ENDPOINT_ADDRESS_PROPERTY, endpointURL);

        List<Handler> handlerChain = bp.getBinding().getHandlerChain();
        handlerChain.add(new DgwsClientMessageHandler(idCardFactory));
        handlerChain.add(new WhitelistingClientMessageHandler(stsConfig.getOrganizationName(), stsConfig.getSystemName(), "1.0.0", stsConfig.getOrganizationName()));
        bp.getBinding().setHandlerChain(handlerChain);

        return port;
    }

    public Citizen withCitizen(IdwsUser user) {
        if (!user.allowDdvChanges()) throw new UnsupportedOperationException("IdwsUser " + user + " does not allow data to be changed in DDV");
        return new Citizen(user.cpr());
    }


    private ModifiedType createModifier() {
        ModifiedType modifier = new ModifiedType();

        modifier.setModifiedDateTime(ZonedDateTime.now());
        ModificatorType modificator = new ModificatorType();

        AuthorisedHealthcareProfessionalType authPerson = new AuthorisedHealthcareProfessionalType();

        authPerson.setAuthorisationIdentifier("J0184");
        authPerson.setName("Lars Larsen");

        OrganisationType org = new OrganisationType();
        org.setName("Ole Abrahamsen");
        org.setType(PredefinedOrganisationTypeType.YDER.value());
        OrganisationIdentifierType orgId = new OrganisationIdentifierType();
        orgId.setSource(OrganisationIdentifierPredefinedSourceType.YDER.value());
        orgId.setValue("19");
        org.setIdentifier(orgId);


        modificator.getContent().add(new JAXBElement<>(new QName("http://vaccinationsregister.dk/schemas/2013/12/01", "AuthorisedHealthcareProfessional"),
            AuthorisedHealthcareProfessionalType.class,
            authPerson));
        modificator.getContent().add(new JAXBElement<>(new QName("http://vaccinationsregister.dk/schemas/2013/12/01", "Organisation"),
            OrganisationType.class,
            org));
        modifier.setModificator(modificator);
        return modifier;
    }


    public class Citizen {
        public final String cpr;

        public Citizen(String cpr) {
            this.cpr = cpr;
        }

        public Citizen clear() throws dk.ssi.nsi.xml_schema._2013._01._01.DGWSFault {
            ResetPatientsRequestType request = new ResetPatientsRequestType();
            IdentifierType identifier = new IdentifierType();
            identifier.getPersonIdentifier().add(cpr);
            request.getIdentifier().add(identifier);
            dumpRestorePort.resetPatients(request);

            return this;
        }

        public Vaccination addVaccination(long vaccineId, long drugId, Date effectuationDateTime) throws DGWSFault {
            CreateVaccinationRequestType request = new CreateVaccinationRequestType();

            request.setCreated(createModifier());
            request.setPersonCivilRegistrationIdentifier(cpr);

            VaccinationCreateType vaccinationCreate = new VaccinationCreateType();

            vaccinationCreate.setBatchNumber("123");
            vaccinationCreate.setVaccineIdentifier(vaccineId);
            vaccinationCreate.setDrugIdentifier(drugId);

            vaccinationCreate.setEffectuatedDateTime(ZonedDateTime.ofInstant(effectuationDateTime.toInstant(), ZoneId.systemDefault()));
            request.setVaccinationCreate(vaccinationCreate);

            CreateVaccinationResponseType response = ddvPort.createVaccination20131201(request);
            return new Vaccination(this, response.getVaccination().get(0));
        }

        public Vaccination addVaccination(long vaccineId, long drugId, Date effectuationDateTime, String countryCode) throws DGWSFault {
            CreatePreviousVaccinationRequestType request = new CreatePreviousVaccinationRequestType();
            ModifiedType modifier = new ModifiedType();
            modifier.setModifiedDateTime(ZonedDateTime.now());

            PartlyDefinedEffectuatorType partlyDefinedEffectuator = new PartlyDefinedEffectuatorType();
            partlyDefinedEffectuator.setEffectuatedInCountryCode(countryCode);

            ModificatorType modificator = new ModificatorType();
            modificator.getContent().add(new JAXBElement<>(
                new QName("http://vaccinationsregister.dk/schemas/2013/12/01", "PartlyDefinedEffectuator"),
                PartlyDefinedEffectuatorType.class,
                partlyDefinedEffectuator)
            );

            modifier.setModificator(modificator);
            request.setCreated(modifier);
            request.setPersonCivilRegistrationIdentifier(cpr);

            final PreviousVaccinationCreateType previousVaccinationCreate = new PreviousVaccinationCreateType();
            previousVaccinationCreate.setBatchNumber("123");
            previousVaccinationCreate.setVaccineIdentifier(vaccineId);
            previousVaccinationCreate.setDrugIdentifier(drugId);
            previousVaccinationCreate.setEffectuatedDateTime(ZonedDateTime.ofInstant(effectuationDateTime.toInstant(), ZoneId.systemDefault()));

            request.setPreviousVaccinationCreate(previousVaccinationCreate);

            CreatePreviousVaccinationResponseType response = ddvPort.createPreviousVaccination20131201(request);
            return new Vaccination(this, response.getVaccination().get(0));
        }

        public Vaccination addPfizerVaccination(Date effectuationDateTime) throws DGWSFault {
            return addVaccination(vaccineData.pfizerVaccineId, vaccineData.pfizerDrugId, effectuationDateTime);
        }

        public Vaccination addAstraZenecaVaccination(Date effectuationDateTime) throws DGWSFault {
            return addVaccination(vaccineData.azVaccineId, vaccineData.azDrugId, effectuationDateTime);
        }

        public Vaccination addJohnsonVaccination(Date effectuationDateTime) throws DGWSFault {
            return addVaccination(vaccineData.jjVaccineId, vaccineData.jjDrugId, effectuationDateTime);
        }

        public Vaccination addJohnsonVaccinationEffectuatedInCountry(String countryCode, Date effectuationDateTime) throws DGWSFault {
            return addVaccination(vaccineData.jjVaccineId, vaccineData.jjDrugId, effectuationDateTime, countryCode);
        }

        public Vaccination addModernaVaccination(Date effectuationDateTime) throws DGWSFault {
            return addVaccination(vaccineData.modernaVaccineId, vaccineData.modernaDrugId, effectuationDateTime);
        }

    }

    public class Vaccination {

        public final VaccinationType vaccination;
        public final Citizen citizen;

        public Vaccination(Citizen citizen, VaccinationType vaccination) {
            this.citizen = citizen;
            this.vaccination = vaccination;
        }

        public void deleteVaccination() throws DGWSFault {
            DeleteVaccinationRequestType request = new DeleteVaccinationRequestType();
            request.setModified(createModifier());
            request.setPersonCivilRegistrationIdentifier(citizen.cpr);

            DeleteVaccinationType deleteVaccination = new DeleteVaccinationType();
            deleteVaccination.setVaccinationIdentifier(vaccination.getVaccinationIdentifier());
            deleteVaccination.setVaccinationVersionIdentifier(vaccination.getVaccinationVersionIdentifier());

            request.setDeleteVaccination(deleteVaccination);

            ddvPort.deleteVaccination20131201(request);
        }

    }

}
