/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.ddvdriver.dgws;

import java.io.IOException;
import java.io.StringReader;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;
import java.util.UUID;

import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.soap.SOAPEnvelope;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPHeader;
import javax.xml.soap.SOAPMessage;
import javax.xml.ws.handler.MessageContext;
import javax.xml.ws.handler.soap.SOAPHandler;
import javax.xml.ws.handler.soap.SOAPMessageContext;

import org.joda.time.DateTime;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

public class DgwsClientMessageHandler implements SOAPHandler<SOAPMessageContext> {

    DocumentBuilderFactory documentBuilderFactory;

    IdCardFactory idCardFactory;

    private static final String WSSE_NS = "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd";
    private static final String WSSE_NS_PREFIX = "wsse";
    private static final QName WSSE_SECURITY_QNAME = new QName(WSSE_NS, "Security", WSSE_NS_PREFIX);
    private static final String WSU_NS = "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd";
    private static final String MEDCOM_NS = "http://www.medcom.dk/dgws/2006/04/dgws-1.0.xsd";

    public DgwsClientMessageHandler(IdCardFactory idCardFactory) throws SOAPException {
        this.idCardFactory = idCardFactory;
        documentBuilderFactory  = DocumentBuilderFactory.newInstance();
        documentBuilderFactory.setNamespaceAware(true);
    }

    @Override
    public void close(MessageContext arg0) {
    }

    @Override
    public boolean handleFault(SOAPMessageContext arg0) {
        return true;
    }

    @Override
    public boolean handleMessage(SOAPMessageContext messageContext) {
        try {
            SOAPMessage message = messageContext.getMessage();

            SOAPEnvelope envelope = message.getSOAPPart().getEnvelope();
            SOAPHeader header = envelope.getHeader();
            if (header == null) {
                envelope.addHeader();
            }

            SOAPHeader headerElement = message.getSOAPHeader();
            Document doc = headerElement.getOwnerDocument();

            Element securityElement = createSecurityHeader(doc, headerElement);
            headerElement.appendChild(securityElement);

            addTimestampHeader(doc, securityElement, new DateTime());

            addAssertionHeader(doc, securityElement, idCardFactory.serialize2DOMDocument(doc));

            addMedcomHeader(doc, headerElement, UUID.randomUUID().toString());

            return true;
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    public static void addMedcomHeader(Document doc, SOAPHeader headerElement, String flowId) {
        Element medcomHeaderElement = doc.createElementNS(MEDCOM_NS, "Header");

        Element medcomSecurityLevelElement = doc.createElementNS(MEDCOM_NS, "SecurityLevel");
        medcomSecurityLevelElement.appendChild(doc.createTextNode("1"));
        medcomHeaderElement.appendChild(medcomSecurityLevelElement);

        Element medcomLinkingElement = doc.createElementNS(MEDCOM_NS, "Linking");
        Node messageIdElement = medcomLinkingElement.appendChild(doc.createElementNS(MEDCOM_NS, "MessageID"));

        messageIdElement.setTextContent(flowId);

        medcomHeaderElement.appendChild(medcomLinkingElement);

        headerElement.appendChild(medcomHeaderElement);
    }

    private Element createSecurityHeader(Document doc, Element headerElement) {
        Element securityElement = doc.createElementNS(WSSE_SECURITY_QNAME.getNamespaceURI(), WSSE_SECURITY_QNAME.getLocalPart());

        return securityElement;
    }

    private void addTimestampHeader(Document doc, Element securityElement, DateTime dateTime) {
        Element timestampElement = doc.createElementNS(WSU_NS, "Timestamp");
        Element createdElement = doc.createElementNS(WSU_NS, "Created");
        createdElement.appendChild(doc.createTextNode(dateTime.toString()));
        timestampElement.appendChild(createdElement);
        securityElement.appendChild(timestampElement);
    }

    private void addAssertionHeader(Document doc, Element securityElement, Element assertion) {
        securityElement.appendChild(doc.importNode(assertion, true));
    }

    public Element parseXml(String xml) throws ParserConfigurationException, SAXException, IOException {

        DocumentBuilder builder = documentBuilderFactory.newDocumentBuilder();
        Document document = builder.parse(new InputSource(new StringReader(xml)));

        return document.getDocumentElement();
    }

    @Override
    public Set<QName> getHeaders() {
        return new HashSet<QName>(Arrays.asList(WSSE_SECURITY_QNAME));
    }

}
