/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.ddvdriver.dgws;

import java.security.cert.X509Certificate;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.Date;
import java.util.Properties;

import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.AuthenticationLevel;
import dk.sosi.seal.model.CareProvider;
import dk.sosi.seal.model.SecurityTokenRequest;
import dk.sosi.seal.model.SecurityTokenResponse;
import dk.sosi.seal.model.SignatureConfiguration;
import dk.sosi.seal.model.SignatureUtil;
import dk.sosi.seal.model.UserIDCard;
import dk.sosi.seal.model.UserInfo;
import dk.sosi.seal.model.constants.IDValues;
import dk.sosi.seal.model.constants.SubjectIdentifierTypeValues;
import dk.sosi.seal.pki.SOSITestFederation;
import dk.sosi.seal.pki.SignatureProviderFactory;
import dk.sosi.seal.vault.ClasspathCredentialVault;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.GenericCredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import dk.sundhedsdatastyrelsen.cbs.dgws.STSConfig;
import dk.sundhedsdatastyrelsen.cbs.dgws.http.HttpClientService;

/**
 * Factory for DGWS requests, which caches a SystemIDCard instance such that we don't need to request a new one from
 * STS for each service invocation.
 */
public class IdCardFactory {

    private static final Logger log = Logger.getLogger(IdCardFactory.class);

    private final STSConfig stsConfig;
    private final HttpClientService httpClientService;
    private final SOSIFactory sosiFactory;

    private UserIDCard stsSignedIdCard;

    private Properties sealProperties;

    public IdCardFactory(STSConfig stsConfig, HttpClientService httpClientService) {
        this.stsConfig = stsConfig;
        this.httpClientService = httpClientService;

        sealProperties = SignatureUtil.setupCryptoProviderForJVM();
        sealProperties.setProperty(SOSIFactory.PROPERTYNAME_SOSI_VALIDATE, Boolean.toString(true));
        CredentialVault vault = new ClasspathCredentialVault(sealProperties, stsConfig.getKeystorePath(), stsConfig.getKeystorePassword());
        sosiFactory = new SOSIFactory(new SOSITestFederation(new Properties()), vault, sealProperties);
    }

    public void ensureValidIdCard() {
        if (stsSignedIdCard != null) {
            if (isCurrentIdCardValid()) {
                // We have a valid idcard with at least 5 hours remaining.
                return;
            } else {
                log.info("Renewing IDCard from " + stsConfig.getStsEndpoint());
            }
        } else {
            log.info("Getting first IDCard from " + stsConfig.getStsEndpoint());
        }

        Document securityTokenRequest = buildSecurityTokenRequest();
        String requestXml = XmlUtil.node2String(securityTokenRequest, false, true);

        String responseXml = httpClientService.soapRequest(stsConfig.getStsEndpoint(), "\"Issue\"", requestXml);

        stsSignedIdCard = buildUserIDCard(responseXml);

        log.info("Got new IDCard based on '" + stsSignedIdCard.getAlternativeIdentifier() + "' issued by " + stsSignedIdCard.getIssuer());
    }

    private boolean isCurrentIdCardValid() {
        Date expiryDate = stsSignedIdCard.getExpiryDate();
        return expiryDate.toInstant().minus(5, ChronoUnit.HOURS).isAfter(Instant.now());
    }

    public UserIDCard getIDCard() {
        ensureValidIdCard();
        
        return stsSignedIdCard;
    }

//    public String getIDCardAsString() {
//        UserIDCard idCard = getIDCard();
//        
//        try {
//            DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();
//            documentBuilderFactory.setNamespaceAware(true);
//
//            Document doc = documentBuilderFactory.newDocumentBuilder().newDocument();
//
//            return XmlUtil.node2String(idCard.serialize2DOMDocument(sosiFactory, doc));
//
//            //                doc.appendChild(idCardElement);
//            //                
//            //                private void documentToOutputStream(Document doc, OutputStream out) throws TransformerFactoryConfigurationError, TransformerException {
//            //                    DOMSource domSource = new DOMSource(doc);
//            //                    Transformer transformer = TransformerFactory.newInstance().newTransformer();
//            //                    transformer.setOutputProperty(OutputKeys.METHOD, "xml");
//            //                    transformer.setOutputProperty(OutputKeys.ENCODING, "UTF-8");
//            //                    transformer.transform(domSource, new StreamResult(out));
//            //                }
//            //
//            //                return sealContext.documentToString(doc);
//        } catch (ParserConfigurationException e) {
//            throw new RuntimeException(e);
//        }
//
//    }

    private UserIDCard buildUserIDCard(String responseXml) {
        SecurityTokenResponse securityTokenResponse = sosiFactory.deserializeSecurityTokenResponse(responseXml);
        UserIDCard signedIdCard = (UserIDCard) securityTokenResponse.getIDCard();

        if (securityTokenResponse.isFault() || signedIdCard == null) {
            log.error("Got a SOAP Fault from STS. The detailed message can be seen in the xml response");
            if (securityTokenResponse.isFault()) {
                log.error("A SOAP Fault was received: " + securityTokenResponse.getFaultString());
            }
            throw new IllegalStateException("SOAP Fault from STS");
        }
        return signedIdCard;
    }

    public Element serialize2DOMDocument(Document doc) {
        UserIDCard idCard = getIDCard();

        return idCard.serialize2DOMDocument(sosiFactory, doc);
    }

    private Document buildSecurityTokenRequest() {
        CareProvider careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, stsConfig.getOrganizationCode(), stsConfig.getOrganizationName());
        X509Certificate certificate = sosiFactory.getCredentialVault().getSystemCredentialPair().getCertificate();

        UserInfo userInfo = new UserInfo("0501792275", "Lars", "Larsen", "lars@larsen.dk", "Læge", "Læge", "J0184");
        UserIDCard userIDCard = sosiFactory.createNewUserIDCard(stsConfig.getSystemName(), userInfo, careProvider, AuthenticationLevel.MOCES_TRUSTED_USER, null, null, certificate, null);

        SecurityTokenRequest securityTokenRequest = sosiFactory.createNewSecurityTokenRequest();
        securityTokenRequest.setIDCard(userIDCard);

        GenericCredentialVault userVault = new GenericCredentialVault(sealProperties, "Test1234");
        userVault.setSystemCredentialPair(getClass().getResourceAsStream("/Lars Larsen.p12"), "Test1234");

        certificate = sosiFactory.getCredentialVault().getSystemCredentialPair().getCertificate();

        SignatureConfiguration signatureConfiguration = new SignatureConfiguration(new String[] { IDValues.IDCARD }, IDValues.IDCARD, IDValues.id);
        SignatureUtil.sign(SignatureProviderFactory.fromCredentialVault(userVault), securityTokenRequest.serialize2DOMDocument(), signatureConfiguration);
        
        return securityTokenRequest.serialize2DOMDocument();
    }

}
