/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.nasdriver;

import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.nsi.advis.v10.NotifyContent;
import dk.sundhedsdatastyrelsen.cbs.dgws.DGWSHeaders;
import dk.sundhedsdatastyrelsen.cbs.dgws.DGWSRequestFactory;
import dk.sundhedsdatastyrelsen.cbs.dgws.GenericDGWSHeaderProvider;
import dk.sundhedsdatastyrelsen.cbs.dgws.STSConfig;
import dk.sundhedsdatastyrelsen.cbs.dgws.http.HttpExecutor;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.EndpointReferences;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.InvalidSOAPFaultHandler;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.NasClientImpl;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.PullPointFactory;
import dk.sundhedsdatastyrelsen.coronapass.ws.JobStatusNotification;
import org.apache.cxf.ext.logging.LoggingFeature;
import org.apache.log4j.Logger;
import org.oasis_open.docs.wsn.b_2.DestroyPullPoint;
import org.oasis_open.docs.wsn.b_2.NotificationMessageHolderType;
import org.oasis_open.docs.wsn.b_2.Unsubscribe;
import org.oasis_open.docs.wsn.bw_2.CreatePullPoint;
import org.oasis_open.docs.wsn.bw_2.NotificationBrokerService;
import org.oasis_open.docs.wsn.bw_2.NotificationConsumer;
import org.oasis_open.docs.wsn.bw_2.NotificationProducer;
import org.oasis_open.docs.wsn.bw_2.PullPoint;
import org.oasis_open.docs.wsn.bw_2.PullPointFactoryService;
import org.oasis_open.docs.wsn.bw_2.SubscriptionManagerService;
import org.oasis_open.docs.wss._2004._01.oasis_200401_wss_wssecurity_secext_1_0.Security;
import org.w3c.dom.Node;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.Holder;
import javax.xml.ws.WebServiceFeature;
import javax.xml.ws.handler.Handler;
import javax.xml.ws.soap.AddressingFeature;
import javax.xml.ws.wsaddressing.W3CEndpointReference;
import java.util.List;
import java.util.UUID;
import java.util.stream.Collectors;

import static dk.sundhedsdatastyrelsen.cbs.configuration.IntegrationTestConfiguration.CBS_IT_LOG_NAS;

public class NASDriver {
    private static final Logger logger = Logger.getLogger(NASDriver.class);
    public static final String JOB_COMPLETED_TOPIC = "http://nspop.dk/GreenCertificate/2021/04/01:JobCompleted";
    private static JAXBContext jaxbContext;

    static {
        try {
            jaxbContext = JAXBContext.newInstance(JobStatusNotification.class);
        } catch (JAXBException e) {
            logger.error("Unable to create JAXBContext for JobStatusNotification");
        }
    }

    private GenericDGWSHeaderProvider<Security, Header> headerProvider;
    private PullPointFactory pullPointFactory;
    private W3CEndpointReference pullPointReference;
    private W3CEndpointReference subscribe;
    private NasClientImpl nasClient;
    private final String stsUrl;
    private final String notificationBrokerUrl;
    private final String pullPointFactoryUrl;
    private final String subscriptionManagerUrl;

    public NASDriver(String stsUrl, String notificationBrokerUrl, String pullPointFactoryUrl, String subscriptionManagerUrl) {
        this.stsUrl = stsUrl;
        this.notificationBrokerUrl = notificationBrokerUrl;
        this.pullPointFactoryUrl = pullPointFactoryUrl;
        this.subscriptionManagerUrl = subscriptionManagerUrl;

        System.setProperty(DGWSRequestFactory.class.getName() + ".test", "true");
    }

    public void setUp() throws JAXBException {
        STSConfig stsConfig = new STSConfig(
            stsUrl + "/NewSecurityTokenService",
            "Sundhedsdatastyrelsen",
            "33257872",
            "Sundhedsdatastyrelsen - COVID-19 Borgerstatus.jks",
            "Test1234",
            "CBS"
        );

        final NotificationBrokerService notificationBrokerService;
        if (CBS_IT_LOG_NAS) {
            final LoggingFeature loggingFeature = new LoggingFeature();
            loggingFeature.setPrettyLogging(true);
            notificationBrokerService = new NotificationBrokerService(loggingFeature, new AddressingFeature(false));
        } else {
            notificationBrokerService = new NotificationBrokerService(new AddressingFeature(false));
        }

        NotificationConsumer notificationConsumer = notificationBrokerService.getSoap();
        setEndpoint((BindingProvider) notificationConsumer, notificationBrokerUrl);

        headerProvider = new GenericDGWSHeaderProvider<>(new DGWSRequestFactory(stsConfig, new dk.sundhedsdatastyrelsen.cbs.dgws.http.HttpClientServiceImpl(new HttpExecutor())), Security.class, Header.class);
        pullPointFactory = new PullPointFactory();

        final SubscriptionManagerService subscriptionManagerService;
        if (CBS_IT_LOG_NAS) {
            final LoggingFeature loggingFeature = new LoggingFeature();
            loggingFeature.setPrettyLogging(true);
            subscriptionManagerService = new SubscriptionManagerService(loggingFeature, new AddressingFeature(false));
        } else {
            subscriptionManagerService = new SubscriptionManagerService(new AddressingFeature(false));
        }
        NotificationProducer notificationProducer = subscriptionManagerService.getSoap(new AddressingFeature(false));
        setEndpoint((BindingProvider) notificationProducer, subscriptionManagerUrl);

        final PullPointFactoryService pullPointFactoryService;
        if (CBS_IT_LOG_NAS) {
            final LoggingFeature loggingFeature = new LoggingFeature();
            loggingFeature.setPrettyLogging(true);
            pullPointFactoryService = new PullPointFactoryService(loggingFeature, new AddressingFeature(false));
        } else {
            pullPointFactoryService = new PullPointFactoryService(new AddressingFeature(false));
        }
        CreatePullPoint createPullPoint = pullPointFactoryService.getSoap(new AddressingFeature(false));
        setEndpoint((BindingProvider) createPullPoint, pullPointFactoryUrl);

        nasClient = new NasClientImpl(headerProvider, notificationProducer, createPullPoint, pullPointFactory);

        pullPointReference = nasClient.createPullPoint();
        subscribe = nasClient.subscribe(pullPointReference, JOB_COMPLETED_TOPIC);
    }

    public void tearDown() throws Exception {
        DGWSHeaders<Security, Header> headers = headerProvider.getHeaders();

        if (subscribe != null) {
            String subscribeAddress = EndpointReferences.extractAddress(subscribe);
            NotificationProducer notificationProducer = new SubscriptionManagerService().getSoap(new AddressingFeature(false));
            setEndpoint((BindingProvider) notificationProducer, subscribeAddress);
            notificationProducer.unsubscribe(new Unsubscribe(), headers.getSecurity(), new Holder<>(headers.getMedComHeader()));
        }

        if (pullPointReference != null) {
            final PullPoint pullPoint = pullPointFactory.getPullPoint(EndpointReferences.extractAddress(pullPointReference), new AddressingFeature(false));
            pullPoint.destroyPullPoint(new DestroyPullPoint(), headers.getSecurity(), new Holder<>(headers.getMedComHeader()));
        }
    }

    public List<JobStatusNotification> getNotifications() {
        final WebServiceFeature[] features;
        if (CBS_IT_LOG_NAS) {
            final LoggingFeature loggingFeature = new LoggingFeature();
            loggingFeature.setPrettyLogging(true);
            features = new WebServiceFeature[]{new AddressingFeature(false), loggingFeature};
        } else {
            features = new WebServiceFeature[]{new AddressingFeature(false)};
        }
        List<NotificationMessageHolderType> messages = nasClient.getMessages(pullPointReference, 1000, UUID.randomUUID(), features);
        return messages.stream().map(mes -> unmarshal(mes)).collect(Collectors.toList());
    }

    private JobStatusNotification unmarshal(NotificationMessageHolderType mes) {
        try {
            NotifyContent notifyContent = (NotifyContent) mes.getMessage().getAny();
            Object any = notifyContent.getAny();
            Unmarshaller jaxbUnmarshaller = jaxbContext.createUnmarshaller();
            JAXBElement<JobStatusNotification> jaxbElement = jaxbUnmarshaller.unmarshal((Node) any, JobStatusNotification.class);
            return jaxbElement.getValue();

        } catch (JAXBException e) {
            throw new RuntimeException("Unable to unmarshal", e);
        }

    }

    private void setEndpoint(BindingProvider bindingProvider, String url) {
        bindingProvider.getRequestContext().put(BindingProvider.ENDPOINT_ADDRESS_PROPERTY, url);

        final List<Handler> handlerChain = bindingProvider.getBinding().getHandlerChain();
        handlerChain.add(new InvalidSOAPFaultHandler());
        bindingProvider.getBinding().setHandlerChain(handlerChain);
    }


}
