/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.util;

import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.PassJWT;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.exception.ExceptionUtils;
import se.digg.dgc.encoding.Base45;
import se.digg.dgc.encoding.DGCConstants;
import se.digg.dgc.encoding.Zlib;
import se.digg.dgc.payload.v1.DigitalGreenCertificate;
import se.digg.dgc.service.DGCDecoder;
import se.digg.dgc.signatures.cose.CoseSign1_Object;
import se.digg.dgc.signatures.cwt.Cwt;

import java.time.Duration;
import java.util.List;
import java.util.function.Consumer;
import java.util.zip.ZipException;

import static org.junit.jupiter.api.Assertions.assertAll;
import static org.junit.jupiter.api.Assertions.assertEquals;

public class CoronaPassAssertions {
    public static final Duration TOKEN_ISSUED_AT_SLACK = Duration.ofSeconds(10L);  // Clocks might be out of sync

    public static Consumer<DigitalGreenCertificate> assertHasName(String name) {
        final String given = StringUtils.substringBeforeLast(name, " ");
        final String family = StringUtils.substringAfterLast(name, " ");
        return assertHasName(given, family);
    }

    public static Consumer<DigitalGreenCertificate> assertHasName(String given, String family) {
        return certificate -> assertAll(
            () -> assertEquals(given, certificate.getNam().getGn()),
            () -> assertEquals(family, certificate.getNam().getFn())
        );
    }

    public static void assertHasName(String name, DigitalGreenCertificate certificate) {
        final String given = StringUtils.substringBeforeLast(name, " ");
        final String family = StringUtils.substringAfterLast(name, " ");

        assertEquals(given, certificate.getNam().getGn());
        assertEquals(family, certificate.getNam().getFn());
    }

    public static void assertHasName(String given, String family, DigitalGreenCertificate certificate) {
        assertEquals(given, certificate.getNam().getGn());
        assertEquals(family, certificate.getNam().getFn());
    }

    public static void assertHasName(String name, DGCDecoder decoder, List<PassJWT> passes) {
        final String given = StringUtils.substringBeforeLast(name, " ");
        final String family = StringUtils.substringAfterLast(name, " ");
        assertHasName(given, family, decoder, passes);
    }

    public static void assertHasName(String given, String family, DGCDecoder decoder, List<PassJWT> passes) {
        passes.forEach(p -> {
            if (p.getType().equals("DKPLUS")) {
                String token = p.getJWT().substring(4);
                byte[] cborBytes;
                try {
                    cborBytes = Zlib.decompress(Base45.getDecoder().decode(token), true);
                } catch (ZipException e) {
                    throw new RuntimeException("Zip decompression exception", e);
                }
                CoseSign1_Object coseObject = new CoseSign1_Object(cborBytes);
                Cwt cwt = coseObject.getCwt();
                String name = cwt.getClaim("legalName").AsString();
                assertEquals(given + " " + family, name);
            } else if (!p.getType().equals("DK")) {
                try {
                    final DigitalGreenCertificate digitalGreenCertificate = decoder.decode(p.getJWT().replaceFirst("DK3:", DGCConstants.DGC_V1_HEADER));
                    assertHasName(given, family, digitalGreenCertificate);
                } catch (Exception e) {
                    ExceptionUtils.wrapAndThrow(e);
                }
            }
        });
    }
}
