/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.util;

import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassResponse;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.KeyValuePair;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.PassJWT;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.PassList;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;
import se.digg.dgc.encoding.Base45;
import se.digg.dgc.encoding.DGCConstants;
import se.digg.dgc.encoding.Zlib;
import se.digg.dgc.payload.v1.DigitalGreenCertificate;
import se.digg.dgc.payload.v1.RecoveryEntry;
import se.digg.dgc.payload.v1.TestEntry;
import se.digg.dgc.service.DGCDecoder;
import se.digg.dgc.signatures.cose.CoseSign1_Object;
import se.digg.dgc.signatures.cwt.Cwt;

import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.zip.ZipException;

public class CwtTestUtils {
    private CwtTestUtils() {
    }

    public static List<PassJWT> getDkPasses(GetCoronaPassResponse response) {
        return getDkPasses(response.getCoronaPasses().getPass());
    }

    public static List<PassJWT> getDkPasses(List<PassJWT> passList) {
        return passList.stream()
            .filter(pass -> pass.getType().equals("DK") || pass.getType().equals("DKPLUS"))
            .collect(Collectors.toList());
    }

    public static PassJWT getDKPass(GetCoronaPassResponse response) {
        return getDKPass(response.getCoronaPasses().getPass());
    }

    public static PassJWT getDKPass(List<PassJWT> passList) {
        return passList.stream()
            .filter(pass -> pass.getType().equals("DK"))
            .findFirst()
            .orElse(null);
    }

    public static PassJWT getDKPlusPass(GetCoronaPassResponse response) {
        return getDKPlusPass(response.getCoronaPasses().getPass());
    }

    public static PassJWT getDKPlusPass(List<PassJWT> passList) {
        return passList.stream()
            .filter(pass -> pass.getType().equals("DKPLUS"))
            .findFirst()
            .orElse(null);
    }

    public static List<PassJWT> getDKInfoPass(GetCoronaPassResponse response) {
        return response.getCoronaPasses().getPass().stream()
            .filter(pass -> pass.getType().equals("DKINFO"))
            .collect(Collectors.toList());
    }

    public static List<PassJWT> getDKInfoPass(List<PassJWT> passList) {
        return passList.stream()
            .filter(pass -> pass.getType().equals("DKINFO"))
            .collect(Collectors.toList());
    }


    @NonNull
    public static List<PassJWT> getEuVacPasses(@NonNull GetCoronaPassResponse passport) {
        //scope to parent by default (if cwt.children.include-in-response=true)
        final List<PassJWT> passList = hasCustodyLabels(passport) ? CwtTestUtils.getPassesForCustody(passport, "Parent") : passport.getCoronaPasses().getPass();
        return getEuVacPasses(passList);
    }

    @NonNull
    public static List<PassJWT> getEuVacPassesForCustody(@NonNull GetCoronaPassResponse passport, @Nullable String custody) {
        return getEuVacPasses(CwtTestUtils.getPassesForCustody(passport, custody));
    }

    @NonNull
    public static List<PassJWT> getEuVacPasses(@Nullable List<PassJWT> passList) {
        return passList != null ? passList.stream()
            .filter(pass -> pass.getType().equals("EU"))
            .filter(pass -> pass.getProperties().stream().anyMatch(prop -> prop.getKey().equals("eu:keytype") && prop.getValue().equals("vac")))
            .collect(Collectors.toList()) : Collections.emptyList();
    }

    @NonNull
    public static List<PassJWT> getEuLabPasses(@NonNull GetCoronaPassResponse passport) {
        //scope to parent by default (if cwt.children.include-in-response=true)
        final List<PassJWT> passList = hasCustodyLabels(passport) ? CwtTestUtils.getPassesForCustody(passport, "Parent") : passport.getCoronaPasses().getPass();
        return getEuLabPasses(passList);
    }

    @NonNull
    public static List<PassJWT> getEuLabPassesForCustody(@NonNull GetCoronaPassResponse passport, @Nullable String custody) {
        return getEuLabPasses(CwtTestUtils.getPassesForCustody(passport, custody));
    }

    @NonNull
    public static List<PassJWT> getEuLabPasses(@Nullable List<PassJWT> passList) {
        return passList != null ? passList.stream()
            .filter(pass -> pass.getType().equals("EU"))
            .filter(pass -> pass.getProperties().stream().anyMatch(prop -> prop.getKey().equals("eu:keytype") && prop.getValue().equals("lab")))
            .collect(Collectors.toList()) : Collections.emptyList();
    }

    @NonNull
    public static List<PassJWT> getEuRecPasses(@NonNull GetCoronaPassResponse passport) {
        //scope to parent by default (if cwt.children.include-in-response=true)
        final List<PassJWT> passList = hasCustodyLabels(passport) ? CwtTestUtils.getPassesForCustody(passport, "Parent") : passport.getCoronaPasses().getPass();
        return getEuRecPasses(passList);
    }

    @NonNull
    public static List<PassJWT> getEuRecPassesForCustody(@NonNull GetCoronaPassResponse passport, @Nullable String custody) {
        return getEuRecPasses(CwtTestUtils.getPassesForCustody(passport, custody));
    }

    @NonNull
    public static List<PassJWT> getEuRecPasses(@Nullable List<PassJWT> passList) {
        return passList != null ? passList.stream()
            .filter(pass -> pass.getType().equals("EU"))
            .filter(pass -> pass.getProperties().stream().anyMatch(prop -> prop.getKey().equals("eu:keytype") && prop.getValue().equals("rec")))
            .collect(Collectors.toList()) : Collections.emptyList();
    }

    public static boolean hasCustodyLabels(@NonNull GetCoronaPassResponse response) {
        final List<PassJWT> parent = CwtTestUtils.getPassesForCustody(response, "Parent");
        return parent != null && !parent.isEmpty(); //there is always for all non-future responses a parent DKINFO token, if we are using custody labels
    }

    @Nullable
    public static List<PassJWT> getPassesForCustody(@NonNull GetCoronaPassResponse passport, @Nullable String custody) {
        final PassList coronaPasses = passport.getCoronaPasses();
        return coronaPasses != null ? getPassesForCustody(coronaPasses.getPass(), custody) : null;
    }

    @Nullable
    public static List<PassJWT> getPassesForCustody(@NonNull List<PassJWT> passList, @Nullable String custody) {
        if (custody == null) return null;
        final List<PassJWT> custodyList = passList.stream()
            .filter(pass -> pass.getProperties().stream().anyMatch(prop -> prop.getKey().equals("Custody") && prop.getValue().equals(custody)))
            .collect(Collectors.toList());
        if (custodyList.isEmpty()) throw new RuntimeException("No Jwts found for custody=" + custody);
        return custodyList;
    }

    @NonNull
    public static Set<String> getCustodies(@NonNull List<PassJWT> passList) {
        return passList.stream()
            .map(CwtTestUtils::getCustody)
            .collect(Collectors.toSet());
    }

    @NonNull
    public static String getCustody(@NonNull PassJWT passJWT) {
        return passJWT.getProperties().stream()
            .filter(p -> p.getKey().equals("Custody"))
            .findAny()
            .map(KeyValuePair::getValue).orElseThrow(() -> new RuntimeException("No CWTs with custody label found"));
    }

    @NonNull
    public static List<TestEntry> getTestEntriesOrderedBySampleTime(@NonNull DigitalGreenCertificate certificate) {
        return certificate.getT().stream()
            .sorted(Comparator.comparing(TestEntry::getSc))
            .collect(Collectors.toList());
    }

    @NonNull
    public static Cwt getCwtFromToken(@NonNull String token) {
        if (token.startsWith(DGCConstants.DGC_V1_HEADER)) {
            token = token.substring(4);
        }

        byte[] compressedCwt = Base45.getDecoder().decode(token);

        byte[] cwtBytes;
        try {
            cwtBytes = Zlib.decompress(compressedCwt, false);
        } catch (ZipException e) {
            throw new RuntimeException("Zip decompression failed", e);
        }
        CoseSign1_Object coseObject = CoseSign1_Object.decode(cwtBytes);

        return coseObject.getCwt();
    }

    public static DigitalGreenCertificate getEUPayload(String token, DGCDecoder decoder) {
        try {
            return decoder.decode(token);
        } catch (Exception e) {
            throw new RuntimeException(e.getMessage(), e);
        }
    }

    public static RecoveryEntry getRecoveryEntry(DigitalGreenCertificate certificate) {
        return certificate.getR().stream().findAny().orElse(null);
    }
}
