/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.util;

import org.apache.log4j.Logger;

import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.concurrent.TimeUnit;
import java.util.function.Function;
import java.util.function.Supplier;

import static org.junit.jupiter.api.Assertions.fail;

public class Wait {
    private static final Logger log = Logger.getLogger(Wait.class);

    private static final int MAX_WAIT_DEFAULT = 300000;

    public static <T> T waitFor(String what, Supplier<T> supplier, Function<T, Boolean> evaluator) throws InterruptedException {
        return waitFor(what, supplier, evaluator, MAX_WAIT_DEFAULT);
    }

    public static <T> T waitFor(String what, Supplier<T> supplier, Function<T, Boolean> evaluator, TimeUnit unit, long amount) throws InterruptedException {
        return waitFor(what, supplier, evaluator, unit.toMillis(amount));
    }

    public static <T> T waitFor(String what, Supplier<T> supplier, Function<T, Boolean> evaluator, long maxWaitMillis) throws InterruptedException {
        Instant start = Instant.now();

        while (start.plus(maxWaitMillis, ChronoUnit.MILLIS).isAfter(Instant.now())) {
            log.debug("Checking for " + what);
            T result = supplier.get();
            if (evaluator.apply(result)) {
                log.debug("Got expected result");
                return result;
            }
            Thread.sleep(1000);
        }

        fail("Timeout waiting for " + what);

        return null;
    }

}
