/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.webservice;

import com.google.common.truth.Correspondence;
import com.google.common.truth.Truth;
import dk.sundhedsdatastyrelsen.cbs.util.CoronaPassAssertions;
import dk.sundhedsdatastyrelsen.cbs.util.CwtTestUtils;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.FutureResult;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassResponse;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.PassJWT;
import org.apache.commons.lang3.StringUtils;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;
import se.digg.dgc.payload.v1.DigitalGreenCertificate;
import se.digg.dgc.service.DGCDecoder;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.ZoneOffset;
import java.time.temporal.ChronoUnit;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.function.Consumer;

import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertNull;

@SuppressWarnings({"unused", "UnusedReturnValue"})
public class ResponseTester {
    private final List<PassJWT> passList;
    private final FutureResult futureResult;
    private final DGCDecoder decoder;

    private ResponseTester(@NonNull GetCoronaPassResponse response, @Nullable DGCDecoder decoder, @Nullable String custody) {
        this.futureResult = response.getFutureResult();
        if (futureResult != null) {
            this.passList = null;
        } else {
            this.passList = custody != null ? Objects.requireNonNull(CwtTestUtils.getPassesForCustody(response, custody)) : response.getCoronaPasses().getPass();
        }
        this.decoder = decoder;
    }

    private ResponseTester(FutureResult futureResult, List<PassJWT> passList, DGCDecoder decoder) {
        this.passList = passList;
        this.futureResult = futureResult;
        this.decoder = decoder;
    }


    @NonNull
    public static ResponseTester assertThat(@NonNull GetCoronaPassResponse response) {
        return assertThat(response, null, null);
    }

    @NonNull
    public static ResponseTester assertThat(@NonNull GetCoronaPassResponse response, @Nullable String custody) {
        return assertThat(response, null, custody);
    }

    @NonNull
    public static ResponseTester assertThat(@NonNull GetCoronaPassResponse response, @Nullable DGCDecoder decoder) {
        return assertThat(response, decoder, null);
    }


    @NonNull
    public static ResponseTester assertThat(@NonNull GetCoronaPassResponse response, @Nullable DGCDecoder decoder, @Nullable String custody) {
        if (custody == null && CwtTestUtils.hasCustodyLabels(response)) {
            return new ResponseTester(response, decoder, "Parent");
        } else {
            return new ResponseTester(response, decoder, custody);
        }
    }

    @NonNull
    public ResponseTester withDecoder(@NonNull Consumer<ResponseTester> tester, @Nullable DGCDecoder decoder) {
        tester.accept(new ResponseTester(this.futureResult, this.passList, decoder));
        return this;
    }


//
//
//    public ResponseTester forParent(Consumer<ResponseTester> tester) {
//        tester.accept(new ResponseTester(futureResult, CwtTestUtils.getPassesForCustody(passList, "Parent"), decoder));
//        return this;
//    }
//
//    public ResponseTester forCustody(String custody, Consumer<ResponseTester> tester) {
//        tester.accept(new ResponseTester(futureResult, CwtTestUtils.getPassesForCustody(passList, custody), decoder));
//        return this;
//    }

    public ResponseTester isNotFutureResult() {
        assertNull(futureResult, "Response contained a FutureResult");
        return this;
    }

    public ResponseTester isFutureResult() {
        assertNotNull(futureResult, "Response did not contain a FutureResult");
        return this;
    }

    public ResponseTester hasDkPasses() {
        isNotFutureResult();
        Truth.assertThat(CwtTestUtils.getDkPasses(passList)).hasSize(2);
        return this;
    }

    public ResponseTester hasDkInfoWithName(String name) {
        final String given = StringUtils.substringBeforeLast(name, " ");
        final String family = StringUtils.substringAfterLast(name, " ");
        return hasDkInfoWithName(given, family);
    }

    public ResponseTester hasDkInfoWithName(String given, String family) {
        isNotFutureResult();
        final List<PassJWT> dkInfoPass = CwtTestUtils.getDKInfoPass(passList);
        Truth.assertThat(dkInfoPass).hasSize(1);
        CoronaPassAssertions.assertHasName(given, family, decoder, dkInfoPass);
        return this;
    }

    public ResponseTester hasNoCustodyChildPasses() {
        Truth.assertThat(CwtTestUtils.getCustodies(passList)).containsExactly("Parent");
        return this;
    }

    public ResponseTester hasNoDkPasses() {
        isNotFutureResult();
        Truth.assertThat(CwtTestUtils.getDkPasses(passList)).isEmpty();
        return this;
    }

    public ResponseTester hasEuVacPasses(int number) {
        isNotFutureResult();
        Truth.assertThat(CwtTestUtils.getEuVacPasses(passList)).hasSize(number);
        return this;
    }

    public ResponseTester hasNoEuVacPasses() {
        isNotFutureResult();
        Truth.assertThat(CwtTestUtils.getEuVacPasses(passList)).isEmpty();
        return this;
    }

    public ResponseTester hasEuVacPassWithEffectuationDate(LocalDate effectuationDate) {
        Objects.requireNonNull(decoder, "cannot compare sample times, decoder is null. Instantiate the ResponseTester with a decoder");
        isNotFutureResult();
        final List<PassJWT> euVacPasses = CwtTestUtils.getEuVacPasses(passList);
        Correspondence<PassJWT, LocalDate> matchingEffectuationDate = Correspondence.from((PassJWT p, LocalDate s) -> {
            try {
                final DigitalGreenCertificate digitalGreenCertificate = decoder.decode(p.getJWT());
                return digitalGreenCertificate.getV().stream().anyMatch(t -> t.getDt().equals(effectuationDate));
            } catch (Exception e) {
                throw new RuntimeException("decoding failed", e);
            }
        }, "EU vac PassJWT has effectuation date equal to");

        Truth.assertThat(euVacPasses).comparingElementsUsing(matchingEffectuationDate).contains(effectuationDate);
        return this;
    }

    public ResponseTester hasEuLabPasses(int number) {
        isNotFutureResult();
        Truth.assertThat(CwtTestUtils.getEuLabPasses(passList)).hasSize(number);
        return this;
    }

    public ResponseTester hasEuLabPassWithSampleTime(LocalDateTime samplingTime) {
        Objects.requireNonNull(decoder, "cannot compare sample times, decoder is null. Instantiate the ResponseTester with a decoder");
        ZoneOffset samplingZoneOffset = ZoneId.systemDefault().getRules().getOffset(samplingTime);
        isNotFutureResult();
        final List<PassJWT> euLabPasses = CwtTestUtils.getEuLabPasses(passList);
        Correspondence<PassJWT, LocalDateTime> matchingSamplingTime = Correspondence.from((PassJWT p, LocalDateTime s) -> {
            try {
                final DigitalGreenCertificate digitalGreenCertificate = decoder.decode(p.getJWT());
                return digitalGreenCertificate.getT().stream().anyMatch(t ->
                    t.getSc().truncatedTo(ChronoUnit.SECONDS).equals(samplingTime.toInstant(samplingZoneOffset).truncatedTo(ChronoUnit.SECONDS))
                );
            } catch (Exception e) {
                throw new RuntimeException("decoding failed", e);
            }
        }, "EU lab PassJWT has samplingTime equal to");

        Truth.assertThat(euLabPasses).comparingElementsUsing(matchingSamplingTime).contains(samplingTime);
        return this;
    }

    public ResponseTester hasNoEuLabPasses() {
        isNotFutureResult();
        Truth.assertThat(CwtTestUtils.getEuLabPasses(passList)).isEmpty();
        return this;
    }


    public ResponseTester hasEuRecPasses(int number) {
        isNotFutureResult();
        Truth.assertThat(CwtTestUtils.getEuRecPasses(passList)).hasSize(number);
        return this;
    }

    public ResponseTester hasNoEuRecPass() {
        isNotFutureResult();
        Truth.assertThat(CwtTestUtils.getEuRecPasses(passList)).isEmpty();
        return this;
    }


    public ResponseTester allCwtsWithNameHasName(String name) {
        final ArrayList<PassJWT> passListCopy = new ArrayList<>(passList);
        passListCopy.remove(CwtTestUtils.getDKPass(passListCopy));
        CoronaPassAssertions.assertHasName(name, decoder, passListCopy);
        return this;
    }

}
