/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.webservice.idws;

import dk.sosi.seal.OIOIDWSFactory;
import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.model.IdentityToken;
import dk.sosi.seal.model.OIOBootstrapToIdentityTokenResponse;
import dk.sosi.seal.model.OIOBootstrapToken;
import dk.sosi.seal.model.OIOSAMLAssertion;
import dk.sosi.seal.model.SignatureUtil;
import dk.sosi.seal.model.dombuilders.OIOBootstrapToIdentityTokenRequestDOMBuilder;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.pki.SOSITestFederation;
import dk.sosi.seal.vault.ClasspathCredentialVault;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import org.apache.commons.lang3.tuple.Triple;
import org.apache.log4j.Logger;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpMethod;
import org.springframework.http.MediaType;
import org.springframework.http.RequestEntity;
import org.springframework.http.ResponseEntity;
import org.springframework.http.converter.StringHttpMessageConverter;
import org.springframework.web.client.RestTemplate;
import org.w3c.dom.Document;

import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import static java.nio.charset.StandardCharsets.UTF_8;

public class IdwsTokenHelper {

    public static final String SERVICE_BOOTSTRAP_2_IDWS = "/Bst2Idws";

    private static final Logger logger = Logger.getLogger(IdwsTokenHelper.class);

    private final CredentialVault credentialVault;
    private final Federation federation;
    private final String stsUrl;
    private final Map<Triple<String, String, String>, OIOBootstrapToIdentityTokenResponse> tokenCache = new HashMap<>();
    private final boolean logRequestResponse;

    public IdwsTokenHelper(String stsUrl, boolean logRequestResponse) {
        this.logRequestResponse = logRequestResponse;
        this.federation = new SOSITestFederation(SignatureUtil.setupCryptoProviderForJVM());
        this.credentialVault = new ClasspathCredentialVault(new Properties(), "TEST whitelisted SP SOSI alias.jks", "Test1234");
        this.stsUrl = stsUrl;
    }

    public IdentityToken getIdentityToken(String cpr, String audience, String procurationCpr, OIOSAMLAssertion bootstrapTokenAssertion) {
        final Triple<String, String, String> cacheKey = Triple.of(cpr, audience, procurationCpr);
        final OIOBootstrapToIdentityTokenResponse tokenResponse = tokenCache.get(cacheKey);
        if (tokenResponse != null && tokenResponse.getExpires().after(new Date())) {
            return tokenResponse.getIdentityToken();
        }
        OIOBootstrapToken bootstrapToken = new OIOBootstrapToken(bootstrapTokenAssertion.getDOM().getDocumentElement());
        OIOBootstrapToIdentityTokenRequestDOMBuilder oioBootstrapToIdentityTokenRequestDOMBuilder = new OIOBootstrapToIdentityTokenRequestDOMBuilder();
        oioBootstrapToIdentityTokenRequestDOMBuilder.setSigningVault(credentialVault);
        oioBootstrapToIdentityTokenRequestDOMBuilder.setAudience(audience);
        oioBootstrapToIdentityTokenRequestDOMBuilder.setCPRNumberClaim(cpr);
        oioBootstrapToIdentityTokenRequestDOMBuilder.setOIOBootstrapToken(bootstrapToken);
        oioBootstrapToIdentityTokenRequestDOMBuilder.setProcurationByCprClaim(procurationCpr);
        Document bootstrapRequest = oioBootstrapToIdentityTokenRequestDOMBuilder.build();
        Document stsBootstrapResponse = callSts(SERVICE_BOOTSTRAP_2_IDWS, bootstrapRequest);

        OIOSAMLFactory oiosamlFactory = new OIOSAMLFactory();
        OIOBootstrapToIdentityTokenResponse bootstrapToIdentityTokenResponse = oiosamlFactory.createOIOBootstrapToIdentityTokenResponseModelBuilder().build(stsBootstrapResponse);

        if (bootstrapToIdentityTokenResponse.getIdentityToken() == null) {
            throw new RuntimeException("Error received from STS: " + bootstrapToIdentityTokenResponse.getFaultCode() + " - " + bootstrapToIdentityTokenResponse.getFaultString());
        }
        tokenCache.put(cacheKey, bootstrapToIdentityTokenResponse);

		return bootstrapToIdentityTokenResponse.getIdentityToken();
	}

	public OIOIDWSFactory getOioIdwsFactory() {
		return new OIOIDWSFactory(federation, credentialVault);
	}

	public Document callSts(String issueService, Document wstRequest) {
        String stsRequestBody = XmlUtil.node2String(wstRequest, false, true);
        final HttpHeaders requestHeaders = new HttpHeaders();
        requestHeaders.setContentType(new MediaType("application", "xml", UTF_8));
        RequestEntity<String> stsRequest = new RequestEntity<>(stsRequestBody, requestHeaders, HttpMethod.POST, null);
        final RestTemplate restTemplate = new RestTemplate();
        restTemplate.getMessageConverters().add(new StringHttpMessageConverter(UTF_8));
        if (logRequestResponse) logger.info(stsUrl + issueService + " request body: \n" + stsRequestBody);
        ResponseEntity<String> stsResult = restTemplate.postForEntity(stsUrl + issueService, stsRequest, String.class);
        String responseBody = stsResult.getBody();
        if (logRequestResponse) logger.info(stsUrl + issueService + " response body: \n" + responseBody);
        return XmlUtil.readXml(new Properties(), responseBody, false);
    }
}
